/**************************************************************************
 * BASF2 (Belle Analysis Framework 2)                                     *
 * Copyright(C) 2018 - Belle II Collaboration                             *
 *                                                                        *
 * Author: The Belle II Collaboration                                     *
 * Contributors: Felix Metzner                                            *
 *                                                                        *
 * This software is provided "as is" without any warranty.                *
 **************************************************************************/

#include <analysis/modules/CurlerCloneMarker/CurlerCloneMarkerModule.h>

#include <framework/gearbox/Const.h>
#include <framework/logging/Logger.h>
#include <framework/datastore/RelationArray.h>

#include <mdst/dataobjects/TrackFitResult.h>
#include <mdst/dataobjects/HitPatternVXD.h>

#include <analysis/dataobjects/Particle.h>

#include <TVector3.h>

#include <cmath>
#include <algorithm>
#include <TMatrixFSym.h>

namespace Belle2 {

  REG_MODULE(CurlerCloneMarker)

  CurlerCloneMarkerModule::CurlerCloneMarkerModule() : Module() {
    setDescription("This module will loop over all tracks in an event and mark such tracks "
                   "which fulfill the defined criteria to be considered as clone resulting "
                   "from a curling track.\n"
                   "The module evaluates pairs of tracks regarding their transverse (has to be "
                   "lower than maxPt), the difference in their transverse momenta "
                   "(abs(p_t1 - p_t2) < maxPtDiff) and the angle between their momenta "
                   "(abs(cos(p1.Angle(p2)) > cos(maxAngle)).\n"
                   "If a pair of tracks fulfilling those criteria is found, the one that does "
                   "not have SVD hits is marked as the clone. If both or none have SVD hits, "
                   "the track with the higher impact parameters ((d0/errd0)^2 + (z0/errz0)^2) "
                   "is marked as clone.");
    setPropertyFlags(c_ParallelProcessingCertified);

    addParam("maxPt", m_maxPt,
             "Maximal value of the transverse momentum of a track to be considered as curler.",
             m_maxPt);
    addParam("maxPtDiff", m_maxPtDiff,
             "Maximal difference in transverse momentum of two tracks to be considered as resulting "
             "from the same curler.",
             m_maxPtDiff);
    addParam("maxAngle", m_maxAngle,
             "Maximal angle between the momentum vectors of the two tracks to be considered as "
             "resulting from the same curler. Abs(cosine()",
             m_maxAngle);
  }


  void CurlerCloneMarkerModule::initialize() {
    B2ASSERT("Parameter maxPt is not within the accepted bounds of 0.0 < maxPt < 1.0 GeV.",
             (m_maxPt > 0.0) and (m_maxPt < 1.0));
    B2ASSERT("Parameter maxPtDiff is not within the accepted bounds of 0.0 < maxPtDiff < 1.0 GeV.",
             (m_maxPtDiff > 0.0) and (m_maxPtDiff < 1.0));
    B2ASSERT("Parameter maxAngle is not within the accepted bounds of 0.0 < maxPtDiff < 20.0 degrees.",
             (m_maxPtDiff > 0.0) and (m_maxPtDiff < 20.0));

    m_maxAbsCosAngle = cos(m_maxAngle * M_PI / 180.);

    B2INFO("Using CurlerCloneMarkerModule with settings:\n"
           "\tmax Pt = " << m_maxPt << " GeV\n"
           "\tmax difference in Pt = " << m_maxPtDiff << " GeV\n"
           "\tmax angle = " << m_maxAngle << " degrees.\n"
           "\tmax absolut value of cosine of angle = " << m_maxAbsCosAngle << "."
    );
  }


  void CurlerCloneMarkerModule::event() {
    StoreArray<Track> Tracks;

    if (Tracks.getEntries() < 2) {
      return;
    }

    for (int i = 0; i < Tracks.getEntries() - 1; i++) {
      for (int j = i + 1; j < Tracks.getEntries(); j++) {
        checkTrackPair(Tracks[i], Tracks[j]);
      }
    }
  }


  bool CurlerCloneMarkerModule::checkTrackPair(Track *firstTrack, Track *secondTrack) {
    const auto firstTrackFit = firstTrack->getTrackFitResultWithClosestMass(Const::pion);
    const auto secondTrackFit = secondTrack->getTrackFitResultWithClosestMass(Const::pion);

    const double firstPt = firstTrackFit->getTransverseMomentum();
    const double secondPt = secondTrackFit->getTransverseMomentum();

    if (isnan(firstPt) or isnan(secondPt)) {
      return false;
    }

    // Check if transverse momenta a below maximal bound.
    if ((firstPt > m_maxPt) or (secondPt > m_maxPt)) {
      return false;
    }

    // Check if difference in transverse momenta is below acceptance bound.
    if (abs(firstPt - secondPt) > m_maxPtDiff) {
      return false;
    }

    // Check if theta angle agree with each other, as curling tracks should be parallel.
    TVector3 first3Momentum = firstTrackFit->get4Momentum().Vect();
    TVector3 second3Momentum = secondTrackFit->get4Momentum().Vect();

    if (abs(cos(first3Momentum.Angle(second3Momentum))) < m_maxAbsCosAngle) {
      return false;
    }

    // Check if the charge fits:
    const double arg = first3Momentum.Dot(second3Momentum) / first3Momentum.Mag2() * second3Momentum.Mag2();
    const bool parallelTracks = arg > 0;
    const auto firstCharge = firstTrackFit->getChargeSign();
    const auto secondCharge = secondTrackFit->getChargeSign();
    const bool sameCharge = (firstCharge * secondCharge) > 0;
    if (not((sameCharge and parallelTracks) or ((not sameCharge) and (not parallelTracks)))) {
      return false;
    }

    // Find the clone and mark it.
    const unsigned short firstNSVD = firstTrackFit->getHitPatternVXD().getNSVDHits();
    const unsigned short secondNSVD = secondTrackFit->getHitPatternVXD().getNSVDHits();
    if ((firstNSVD == 0) and (secondNSVD >= 1)) {
      firstTrack->setCurlerCloneIndicator();
    } else if ((firstNSVD >= 1) and (secondNSVD == 0)) {
      secondTrack->setCurlerCloneIndicator();
    } else {
      const double firstImpact = (pow(firstTrackFit->getD0(), 2) / firstTrackFit->getCovariance5()[0][0])
                                 + (pow(firstTrackFit->getZ0(), 2) / firstTrackFit->getCovariance5()[3][3]);
      const double secondImpact = (pow(secondTrackFit->getD0(), 2) / secondTrackFit->getCovariance5()[0][0])
                                  + (pow(secondTrackFit->getZ0(), 2) / secondTrackFit->getCovariance5()[3][3]);
      if (firstImpact > secondImpact) {
        firstTrack->setCurlerCloneIndicator();
      } else {
        secondTrack->setCurlerCloneIndicator();
      }
    }

    return true;
  }
} // end Belle2 namespace
