/**************************************************************************
 * BASF2 (Belle Analysis Framework 2)                                     *
 * Copyright(C) 2010 - Belle II Collaboration                             *
 *                                                                        *
 * Author: The Belle II Collaboration                                     *
 * Contributors: Felix Metzner, Max Welsch, Patrick Ecker                 *
 *                                                                        *
 * File for untracked private Variables for B -> D(*(*)) l nu             *
 * Created on October 17th 2018                                           *
 *                                                                        *
 * This software is provided "as is" without any warranty.                *
 **************************************************************************/

#pragma once

#include <TLorentzVector.h>

#include <analysis/VariableManager/Manager.h>

namespace Belle2 {
  class Particle;

  namespace Variable {

    /**
     *
     * @param part Particle
     * @param minDaughters Minimal number of daughters the particle has to have.
     *                      Default is 3 for MC B meson decay with neutrino.
     * @return True if give particle is a B meson with three daughters.
     */
    bool isValidBMeson(const Particle* part, unsigned int minDaughters = 3);
    bool isValidBMesonMC(const MCParticle* mcPart, unsigned int minDaughters = 3);

    /**
     *
     * @param part Particle
     * @return True if give particle is a Upsilon(4S) meson with two B mesons as daughters.
     */
    bool isValidUpsilon4S(const Particle* part);

    /**
     * Calculates Magnitude of lepton momentum in the signal B rest frame, where the latter
     * is calculated using the reconstructed tag side B meson for which the energy is set
     * to half of the beam energy.
     * @param part Expected to be a Upsilon(4S) meson.
     * @return Magnitude of the lepton momentum in the signal B rest frame
     */
    double pStarL(const Particle* part);

    /**
     * @param part Expected to be a generator level B meson.
     * @return Squared difference of four-momenta of particle (= B meson) and its first daughter.
     */
    double qSquaredMC(const Particle* part);

    /**
     * @param part Expected to be a Upsilon(4S) meson.
     * @return Squared difference of four-momenta of the signal B meson and its first daughter.
     *         The signal B meson must be the second daughter of the Upsilon(4S)!
     */
    double qSquaredReco(const Particle* part);
    double qSquaredReco2(const Particle* part);
    double qSquaredReco3(const Particle* part);
    double qSquaredReco4(const Particle* part);

    /**
     * @param part Expected to be a Upsilon(4S) meson.
     * @return Squared difference of four-momenta of the signal B meson and its first and third daughter.
     *         The third daughter is a pion which together with the first daughter forms a D** candidate.
     *         The signal B meson must be the second daughter of the Upsilon(4S)!
     */
    double qSquaredRecoDStarStar(const Particle* part);

    /**
     * @param part Expected to be a gen level B meson
     * @return Product of four vector velocities of particle and its first daughter
     */
    double w(const Particle* part);
    double wMC(const Particle* part);

    /**
     * The signal B meson must be the second daughter of the Upsilon(4S)!
     * @param part Expected to be a Upsilon(4S) meson
     * @return Product of four vector velocities of B meson reconstructed as first daughter of the Y(4S)
     * and the first daughter of this B meson
     */
    double wReco(const Particle* part);

    /**
     * @param part Expected to be a gen level B meson
     * @return Cosine of the angle between the lepton and virtual W boson in a specific Lorentz frame
     */
    double cosThetaL(const Particle* part);
    double cosThetaLMC(const Particle* part);
    double cosThetaLWOPhotonsMC(const Particle* part);

    /**
     * The signal B meson must be the second daughter of the Upsilon(4S)!
     * @param part Expected to be a Upsilon(4S) meson
     * @return Cosine of the angle between the lepton and virtual W boson in a specific Lorentz frame
     */
    double cosThetaLReco(const Particle* part);
    double cosThetaLRecoOld(const Particle* part);

    /**
     * @param part Expected to be a gen level B meson
     * @return Cosine of the angle between the D*(*) meson and its first daughter in a specific Lorentz frame
     */
    double cosThetaV(const Particle* part);
    double cosThetaVMC(const Particle* part);

    /**
     * The signal B meson must be the second daughter of the Upsilon(4S)!
     * @param part Expected to be a Upsilon(4S) meson
     * @return Cosine of the angle between the D*(*) meson and its first daughter in a specific Lorentz frame
     */
    double cosThetaVReco(const Particle* part);

    /**
     *
     * @param part Expected to be a gen level B meson
     * @return Cosine of the angle between decay planes of the D*(*) meson and virtual W boson in rad
     */
    double chiPrime(const Particle* part);
    double chiPrimeMC(const Particle* part);
    double chiPrimeWOPhotonsMC(const Particle* part);
    double altChi(const Particle* part);
    double chiMC(const Particle* part);
    double chiWOPhotonsMC(const Particle* part);

    /**
     * The signal B meson must be the second daughter of the Upsilon(4S)!
     * @param part Expected to be a Upsilon(4S) meson
     * @return Cosine of the angle between decay planes of the D*(*) meson and virtual W boson in rad
     */
    double chiPrimeReco(const Particle* part);
    double chiReco(const Particle* part);

    /**
     * Convenience function to sum the 4Momenta of all descendant particles of the given
     * particle matching one of the PDG codes provided via the iterators.
     * @tparam Iterator A Iterator type of the container used to provide the PDG codes.
     * @param mcPart MCParticle for which the descendants shall be evaluated.
     * @param begin Iterator begin of the container containing the PDG codes.
     * @param end Iterator end of the container containing the PDG codes.
     * @return TLorentzVector sum of all 4Momenta of the descendants with matchin PDG code.
     */
    template<typename Iterator>
    TLorentzVector sumMomentumOfAllDescendantsWithPDG(
      const MCParticle* mcPart,
      Iterator begin,
      Iterator end
    );

    /**
     * Calculates the MC-truth of the squared missing mass by calculating the
     * squared invariant mass of all neutrinos on the signal side.
     * @return True Squared Missing Mass
     */
    double mcMissingMassSquared(const Particle*);

    /**
     * Calculates and returns the missing Momentum of the Event in the
     * lab frame. The signal B meson must be the second daughter
     * of the Upsilon(4S) to which the variable should be applied.
     */
    double missingMomentum(const Particle* part);

    /**
    * Calculates and returns the missing Momentum of the Event in the
    * lab frame. The signal B meson must be the second daughter
    * of the Upsilon(4S) to which the variable should be applied.
    * The Energy and Momentum of the Tag Side B Meson is thereby rescaled so E_tag = E_beam/2 and |p_tag| according
     * to the energy momentum relation
    */
    double missingMomentumRescaledTag(const Particle* part);

    /**
     * Calculates the MC-truth of the missing momentum in the lab frame
     * by summing over all neutrinos on the signal side.
     * @return True Missing Momentum in lab frame
     */
    double mcMissingMomentum(const Particle*);

    /**
     * Calculates and returns the missing Momentum of the Event in the
     * center of mass frame. The signal B meson must be the second daughter
     * of the Upsilon(4S) to which the variable should be applied.
     */
    double missingMomentumInCMSFrame(const Particle* part);

    /**
    * Calculates and returns the missing Momentum of the Event in the
    * center of mass frame. The signal B meson must be the second daughter
    * of the Upsilon(4S) to which the variable should be applied.
    * The Energy and Momentum of the Tag Side B Meson is thereby rescaled so E_tag = E_beam/2 and |p_tag| according
    * to the energy momentum relation
    */
    double missingMomentumRescaledTagInCMSFrame(const Particle* part);

    /**
     * Calculates and returns the missing Momentum of the Event in the
     * rest frame of the signal B which must be the second daughter
     * of the Upsilon(4S) to which the variable should be applied.
     */
    double missingMomentumInBSigRestFrame(const Particle* part);

    /**
    * Calculates and returns the missing Momentum of the Event in the
    * rest frame of the signal B which must be the second daughter
    * of the Upsilon(4S) to which the variable should be applied.
    * The Energy and Momentum of the Tag Side B Meson is thereby rescaled so E_tag = E_beam/2 and |p_tag| according
    * to the energy momentum relation
    */
    double missingMomentumRescaledTagInBSigRestFrame(const Particle* part);

    /**
     * Calculates the MC-truth of the missing momentum in the signal B meson's
     * rest frame by summing over all neutrinos on the signal side.
     * @return True Missing Momentum in signal B rest frame
     */
    double mcMissingMomentumInBSigRestFrame(const Particle*);

    /**
     * Calculates and returns the angle between the particle that is
     * specified via the two indices which are provided as arguments
     * and the missing Momentum of the Event in the rest frame of the
     * signal B which must be the second daughter of the Upsilon(4S)
     * to which the variable should be applied.
     * The provided indices are the daughter-indices of the particle
     * that should be specified. The first is the daughter indices
     * that decides between the B_sig and B_tag and the second for
     * the decay products of the B.
     */
    Manager::FunctionPtr angleWithMissingMomentum(const std::vector<std::string>& arguments);

    /**
     * returns the squared missing mass of the signal side which is calculated in the CMS frame under the assumption that the signal and
     * tag side are produced back to back and the tag side energy equals the beam energy and the tag-side momentum is rescaled accordingly.
     * The variable must be applied to the Upsilon and
     * the tag side must be the first, the signal side the second daughter!
     */
    double m2RecoilSignalSideRescaledTag(const Particle*);

    /**
    * returns the squared missing mass of the signal side which is calculated in the CMS frame under the assumption that the signal and
    * tag side are produced back to back.
    * The variable must be applied to the Upsilon and
    * the tag side must be the first, the signal side the second daughter!
    */
    double m2RecoilSignalSideNoScaling(const Particle*);

    /**
     * returns the missing energy in the event with the assumption that the energy of the tag-side B meson equals the beamEnergy
    * The variable must be applied to the Upsilon and
    * the tag side must be the first, the signal side the second daughter!
     */
    double missingEnergyCMS(const Particle*);

    /**
    * returns the missing energy in the event
    * The variable must be applied to the Upsilon and
    * the tag side must be the first, the signal side the second daughter!
    */
    double missingEnergyNoScalingCMS(const Particle*);

    /**
    * returns the missing energy in the event with the assumption that the energy of the tag-side B meson equals the beamEnergy
    * The variable must be applied to the Upsilon and
    * the tag side must be the first, the signal side the second daughter!
    */
    double missingEnergyLab(const Particle*);

    /**
    * returns the missing energy in the event
    * The variable must be applied to the Upsilon and
    * the tag side must be the first, the signal side the second daughter!
    */
    double missingEnergyNoScalingLab(const Particle*);

    /**
    * returns the missing energy in the event with the assumption that the energy of the tag-side B meson equals the beamEnergy and the momentum is rescaled accordingly.
    * The variable must be applied to the Upsilon and
    * the tag side must be the first, the signal side the second daughter!
    */
    double missingEnergyRescaledTagLab(const Particle*);

    /**
     * Returns if a given Particle has a FSR corrected Particle somewhere in his decay chain
     */
    double wasFSRCorrected(const Particle*);

    /**
     * returns the 4-vector of the FSR momentum which a given particle has
     */
    TLorentzVector getFSRMomentum(const Particle*);

    /**
     * returns the squared missing mass with no fsr correction applied
     */
    double m2RecoilSignalSideNoFSRCorrection(const Particle*);

    /**
     * returns the pStarL variable with no fsr correction applied
     */
    double pStarLNoFSR(const Particle*);

    /**
    * returns the magnitude of the FSR momentum
    */
    double fsrMomentum(const Particle*);

    /**
     * returns the components of the FSR four momentum
     */
    double fsrMomentumX(const Particle*);
    double fsrMomentumY(const Particle*);
    double fsrMomentumZ(const Particle*);
    double fsrE(const Particle*);

    /**
     * returns difference in invariant mass of reconstructed D** Meson and its daughter or granddaughter
     */
    double deltaMDStarStarVsDaughter(const Particle*);
    double deltaMDStarStarVsGrandDaughter(const Particle*);
  }
} // Belle2 namespace