"""
Tool for the Reweighting of Branching Fractions of B -> D^(*)lnu

Patrick Ecker 2020
"""

import numpy as np
import pandas as pd

from typing import Dict, List, Optional

from rdstar.offline_analysis.rdstar_sample_weights import signal_mc_bfs

from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name

from rdstar.branching_fractions import belle_charged_b_meson_decay_info as belle_charged_b_bfs
from rdstar.branching_fractions import belle_neutral_b_meson_decay_info as belle_neutral_b_bfs


__all__ = [
    "BFReweighter",
]


class BFReweighter:
    weight_total_col_name: str = "weight_bf_total"
    weight_sig_col_name: str = "weight_bf_sig"
    weight_tag_col_name: str = "weight_bf_tag"

    b_sig_gen_id_col: str = "B_sig_gen_id_for_bf_reweighting"
    b_tag_gen_id_col: str = "B_tag_gen_id_for_bf_reweighting"

    _all_weight_columns: List[str] = [
        weight_total_col_name,
        weight_sig_col_name,
        weight_tag_col_name,
    ]

    b_mdst_col_name: str = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
    b_mdst_tau_col_name: str = (
        "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"
    )
    b1_gen_id_col_name: str = gen_b1_id_col_name
    b2_gen_id_col_name: str = gen_b2_id_col_name

    _required_cols: List[str] = [
        b_mdst_col_name,
        b_mdst_tau_col_name,
        b1_gen_id_col_name,
        b2_gen_id_col_name,
    ]

    default_b_pdg_col_name: str = "genUpsilon4S__bomcDaughter__bo1__cm__spPDG__bc__bc"

    # Belle BF taken form the Belle_DEVAY.DEC
    _gen_id_Belle_bf_dict: Dict[str, Dict[int, float]] = {
        "b_charged": {
            1: belle_charged_b_bfs.belle_bf_bp_to_d_tau_nu,  # r"$D \tau \nu$"
            2: belle_charged_b_bfs.belle_bf_bp_to_d_st_tau_nu,  # r"$D^* \tau \nu$"
            3: belle_charged_b_bfs.belle_bf_bp_to_d_l_nu,  # r"$D \ell \nu$"
            4: belle_charged_b_bfs.belle_bf_bp_to_d_st_l_nu,  # r"$D^* \ell \nu$"
        },
        "b_neutral": {
            1: belle_neutral_b_bfs.belle_bf_b0_to_d_tau_nu,  # r"$D \tau \nu$"
            2: belle_neutral_b_bfs.belle_bf_b0_to_d_st_tau_nu,  # r"$D^* \tau \nu$"
            3: belle_neutral_b_bfs.belle_bf_b0_to_d_l_nu,  # r"$D \ell \nu$"
            4: belle_neutral_b_bfs.belle_bf_b0_to_d_st_l_nu,  # r"$D^* \ell \nu$"
        },
    }

    # Updated branching fraction, taken from the rdstar_sample_weights.py.
    _gen_id_updated_bf_dict: Dict[str, Dict[int, float]] = {
        "b_charged": {
            1: signal_mc_bfs["DZTau"].bf,  # r"$D \tau \nu$"
            2: signal_mc_bfs["DSZTau"].bf,  # r"$D^* \tau \nu$"
            3: signal_mc_bfs["DZE"].bf,  # r"$D \ell \nu$"
            4: signal_mc_bfs["DSZE"].bf,  # r"$D^* \ell \nu$"
        },
        "b_neutral": {
            1: signal_mc_bfs["DMTau"].bf,  # r"$D \tau \nu$"
            2: signal_mc_bfs["DSMTau"].bf,  # r"$D^* \tau \nu$"
            3: signal_mc_bfs["DME"].bf,  # r"$D \ell \nu$"
            4: signal_mc_bfs["DSME"].bf,  # r"$D^* \ell \nu$"
        },
    }

    def __init__(self, signal_side_only: bool = False) -> None:
        self.signal_side_only = signal_side_only

    def _generate_gen_id_weight_dict(self) -> Dict[str, Dict[int, float]]:
        gen_id_weight_dict: Dict[str, Dict[int, float]] = {}
        assert set(self._gen_id_Belle_bf_dict.keys()) == set(self._gen_id_updated_bf_dict.keys()), [
            set(self._gen_id_Belle_bf_dict.keys()),
            set(self._gen_id_updated_bf_dict.keys()),
        ]
        for b_charge, bf_dict in self._gen_id_Belle_bf_dict.items():
            temp_dict: Dict[int, float] = {}
            assert set(self._gen_id_Belle_bf_dict[b_charge].keys()) == set(self._gen_id_updated_bf_dict[b_charge].keys())
            for gen_id in bf_dict.keys():
                weight = self._gen_id_updated_bf_dict[b_charge][gen_id] / self._gen_id_Belle_bf_dict[b_charge][gen_id]
                assert gen_id not in temp_dict.keys(), (gen_id, temp_dict.keys())
                temp_dict.update({gen_id: weight})

            assert b_charge not in gen_id_weight_dict.keys(), (b_charge, gen_id_weight_dict.keys())
            gen_id_weight_dict.update({b_charge: temp_dict})

        return gen_id_weight_dict

    def _get_b_gen_id(self, df: pd.DataFrame) -> None:
        new_cols = [self.b_sig_gen_id_col, self.b_tag_gen_id_col]
        assert all([col not in df.columns for col in new_cols]), [c for c in new_cols]

        temp_sig_b_mdst_index_col: str = "temp_sig_b_mdst_index"
        temp_columns = [temp_sig_b_mdst_index_col]
        assert all([col not in df.columns for col in temp_columns]), [c for c in temp_columns]

        cond = df[self.b_mdst_col_name].isin([1, 2])
        cond_tau = df[self.b_mdst_tau_col_name].isin([1, 2])

        df.loc[:, temp_sig_b_mdst_index_col] = np.nan
        df.loc[cond, temp_sig_b_mdst_index_col] = df[self.b_mdst_col_name]
        df.loc[cond_tau, temp_sig_b_mdst_index_col] = df[self.b_mdst_tau_col_name]
        df.loc[(~cond) & (~cond_tau), temp_sig_b_mdst_index_col] = 0
        assert not pd.isnull(df[temp_sig_b_mdst_index_col]).any(), temp_sig_b_mdst_index_col

        cond_b1 = df[temp_sig_b_mdst_index_col] == 1
        cond_b2 = df[temp_sig_b_mdst_index_col] == 2

        # Signal Side
        df.loc[:, self.b_sig_gen_id_col] = np.nan
        df.loc[cond_b1, self.b_sig_gen_id_col] = df[self.b1_gen_id_col_name]
        df.loc[cond_b2, self.b_sig_gen_id_col] = df[self.b2_gen_id_col_name]
        df.loc[(~cond_b1) & (~cond_b2), self.b_sig_gen_id_col] = -999
        assert not pd.isnull(df[self.b_sig_gen_id_col]).any(), self.b_sig_gen_id_col

        # Tag Side
        df.loc[:, self.b_tag_gen_id_col] = np.nan
        df.loc[cond_b1, self.b_tag_gen_id_col] = df[self.b2_gen_id_col_name]
        df.loc[cond_b2, self.b_tag_gen_id_col] = df[self.b1_gen_id_col_name]
        df.loc[(~cond_b1) & (~cond_b2), self.b_tag_gen_id_col] = -999
        assert not pd.isnull(df[self.b_tag_gen_id_col]).any(), self.b_tag_gen_id_col

        for col in temp_columns:
            del df[col]

    def _get_weights(
        self,
        df: pd.DataFrame,
        weight_col: str,
        gen_id_col: str,
        b_pdg_col: Optional[str] = None,
    ) -> None:
        if b_pdg_col is None:
            _b_pdg_col: str = self.default_b_pdg_col_name
        else:
            _b_pdg_col = b_pdg_col

        assert _b_pdg_col in df.columns, _b_pdg_col
        gen_id_weight_dict = self._generate_gen_id_weight_dict()

        df.loc[:, weight_col] = np.nan
        cond_rw = df[gen_id_col].isin(gen_id_weight_dict["b_charged"].keys())
        df.loc[cond_rw, weight_col] = df[gen_id_col]
        charge_cond = abs(df[_b_pdg_col]) == 521

        df[weight_col] = np.where(
            charge_cond,
            df[weight_col].replace(gen_id_weight_dict["b_charged"]),
            df[weight_col].replace(gen_id_weight_dict["b_neutral"]),
        )

        df.loc[~cond_rw, weight_col] = 1
        assert not pd.isnull(df[weight_col]).any(), weight_col

    def add_weights(self, df: pd.DataFrame) -> None:
        assert all(col not in df.columns for col in self._all_weight_columns)
        assert all(col in df.columns for col in self._required_cols)

        self._get_b_gen_id(df=df)

        self._get_weights(
            df=df,
            gen_id_col=self.b_sig_gen_id_col,
            weight_col=self.weight_sig_col_name,
        )
        self._get_weights(
            df=df,
            gen_id_col=self.b_tag_gen_id_col,
            weight_col=self.weight_tag_col_name,
        )

        if self.signal_side_only:
            df.loc[:, self.weight_total_col_name] = df[self.weight_sig_col_name]
        else:
            df.loc[:, self.weight_total_col_name] = df[self.weight_sig_col_name] * df[self.weight_tag_col_name]

        assert not pd.isnull(df[self.weight_total_col_name]).any(), self.weight_total_col_name

    @staticmethod
    def add_neutral_weights(df: pd.DataFrame) -> None:
        assert all(col not in df.columns for col in BFReweighter._all_weight_columns)
        assert all(col in df.columns for col in BFReweighter._required_cols)

        for new_col_name in BFReweighter._all_weight_columns:
            df.loc[:, new_col_name] = 1.0
