"""
Useful functions for the D BF Reweighting

Patrick Ecker 2020
"""

import pdg

from typing import List, Tuple, Dict, Union, Optional, NamedTuple
from rdstar.online_analysis.decay_channels import D_decay_channels


__all__ = [
    "DMesonBFInfo",
    "BFTupleType",
    "DecayProductsType",
    "BFDictType",
    "PDGBFDictType",
    "generate_decay_products_list",
    "generate_bf_dict",
    "generate_charge_conjugate_bf_weight_dict",
    "generate_bf_weight_dict",
    "convert_weight_dict_for_comparison",
    "get_d_reco_decays",
]


class DMesonBFInfo(NamedTuple):
    decay_products: str
    bf: float


BFTupleType = Tuple[DMesonBFInfo, ...]
DecayProductsType = Tuple[str, ...]
DecayProductsPDGType = Tuple[int, ...]
BFDictType = Dict[DecayProductsType, float]
PDGBFDictType = Dict[DecayProductsPDGType, float]
WeightDictComparisonType = Dict[Union[int, Tuple[int, ...]], float]


# this is necessary because of changing order in the anti-D Decay file
special_cc_decays = {
    # D+ to D-
    (-311, 321, -321, -211): (311, 321, -321, 211),
    (-311, 311, 211): (-311, 311, -211),
    (-311, 321, -321, 211): (311, 321, -321, -211),
    (211, 211, -211): (-211, 211, -211),
    (211, 211, -211, 111): (-211, 211, -211, 111),
    (331, 211, 211, -211): (331, -211, 211, -211),
    (221, 211, 211, -211): (221, -211, 211, -211),
    # D0 to anti-D0
    (-313, 211, -211): (313, 211, -211),
    (-321, 211, 211, -211): (321, -211, 211, -211),
    (-311, 211, -211, 111): (311, 211, -211, 111),
    (-321, 211, 211, -211, 111): (321, -211, 211, -211, 111),
    (-313, 321, -321): (313, 321, -321),
    (-311, 321, -321): (311, 321, -321),
    (321, -321): (321, -321),
    (333, 211, -211): (333, 211, -211),
    (321, -321, 113): (321, -321, 113),
    (321, -321, 111, 111): (321, -321, 111, 111),
    (-311, 311, 211, -211): (-311, 311, 211, -211),
    (321, -321, -321, 211): (321, -321, 321, -211),
    (321, -321, -311, 111): (321, -321, 311, 111),
    (321, -321, 211, -211, 111): (321, -321, 211, -211, 111),
    (211, -211): (211, -211),
    (211, 211, -211, -211): (211, 211, -211, -211),
    (211, -211, 111, 111): (211, -211, 111, 111),
    (211, -211, 211, -211, 111): (211, -211, 211, -211, 111),
    (211, -211, 111, 111, 111): (211, -211, 111, 111, 111),
    (211, -211, 211, -211, 211, -211): (211, -211, 211, -211, 211, -211),
    (213, -221): (213, -221),
    (-213, 221): (-213, -221),
}  # type: Dict[Tuple[int, ...], Tuple[int, ...]]


def generate_decay_products_list(decay_str: str) -> List[str]:
    return list(decay_str.split())


def generate_bf_dict(bf_tuple: BFTupleType) -> PDGBFDictType:
    bf_dict = {}  # type: PDGBFDictType
    for decay in bf_tuple:
        decay_products = generate_decay_products_list(decay_str=decay.decay_products)
        key = pdg.from_names(decay_products)
        assert tuple(key) not in bf_dict.keys(), tuple(key)
        bf_dict.update({tuple(key): decay.bf})
    return bf_dict


def conjugate_pdg_list(pdg_codes: List[int]) -> Tuple[int, ...]:
    names = pdg.to_names(pdg_codes)
    cc_names = [pdg.conjugate(name) for name in names]
    cc_pdgs = tuple(pdg.from_names(cc_names))
    return cc_pdgs


def generate_charge_conjugate_bf_weight_dict(
    bf_weight_dict: PDGBFDictType,
    special_weights: Optional[PDGBFDictType] = None,
    additional_weights: Optional[PDGBFDictType] = None,
) -> PDGBFDictType:
    cc_bf_weight_dict = {}  # type: PDGBFDictType
    for decay, weight in bf_weight_dict.items():
        if decay not in special_cc_decays.keys():
            cc_decay = conjugate_pdg_list(pdg_codes=list(decay))  # type: Tuple[int, ...]
            cc_bf_weight_dict.update({tuple(cc_decay): weight})
        else:
            cc_decay = special_cc_decays[decay]
            cc_bf_weight_dict.update({cc_decay: weight})

    if special_weights:
        for decay, weight in special_weights.items():
            assert decay in cc_bf_weight_dict.keys(), decay
            cc_bf_weight_dict[decay] = weight

    if additional_weights:
        for decay, weight in additional_weights.items():
            assert decay not in cc_bf_weight_dict.keys(), decay
            cc_bf_weight_dict.update({decay: weight})

    return cc_bf_weight_dict


def make_bf_dict_compatible(bf_dict: BFDictType) -> PDGBFDictType:
    compatible_bf_dict = {}  # type: PDGBFDictType
    for decay, bf in bf_dict.items():
        decay_pdgs = tuple(pdg.from_names(list(decay)))  # type: Tuple[int, ...]
        assert decay_pdgs not in compatible_bf_dict.keys(), decay_pdgs
        compatible_bf_dict.update({decay_pdgs: bf})
    return compatible_bf_dict


def generate_bf_weight_dict(
    belle_bf_dict: PDGBFDictType,
    updated_bf_dict: BFDictType,
) -> PDGBFDictType:
    _updated_bf_dict = make_bf_dict_compatible(bf_dict=updated_bf_dict)  # type: PDGBFDictType
    belle_bf_sum = 0.0  # type: float
    updated_bf_sum = 0.0  # type: float
    for decay in belle_bf_dict.keys():
        if decay in _updated_bf_dict.keys() and _updated_bf_dict[decay] > 0:
            belle_bf_sum += belle_bf_dict[decay]
            updated_bf_sum += _updated_bf_dict[decay]

    scale_factor = belle_bf_sum / updated_bf_sum
    bf_weight_dict = {}
    for decay in belle_bf_dict.keys():
        if decay not in _updated_bf_dict.keys() or _updated_bf_dict[decay] <= 0:
            weight = 1.0  # type: float
        else:
            weight = scale_factor * _updated_bf_dict[decay] / belle_bf_dict[decay]
        bf_weight_dict.update({decay: weight})
    return bf_weight_dict


def convert_weight_dict_for_comparison(weight_dict: PDGBFDictType) -> WeightDictComparisonType:
    weight_comparison_dict = {-999: 1}  # type: WeightDictComparisonType
    for decay, weight in weight_dict.items():
        decay_list = list(decay)
        while len(decay_list) < 7:
            decay_list.append(-999)
        assert tuple(decay_list) not in weight_comparison_dict.keys(), tuple(decay_list)
        weight_comparison_dict.update({tuple(decay_list): weight})
    return weight_comparison_dict


def get_d_reco_decays(d_charge: str) -> Tuple[List[str], ...]:
    assert d_charge in ["D+", "D0"], f"d_charge has to be D+ or D0 but was {d_charge}"
    d_reco_decays = []  # type: List[List[str]]
    for dec_name, decay_str in D_decay_channels.items():
        if dec_name.startswith(d_charge):
            decay_products = decay_str.split("->")[1]
            if "K_S0" in decay_products:
                for k in ["K0", "anti-K0"]:
                    tmp_decay_products = decay_products.replace("K_S0", k)
                    d_reco_decays.append(generate_decay_products_list(decay_str=tmp_decay_products))
            d_reco_decays.append(generate_decay_products_list(decay_str=decay_products))
    return tuple(d_reco_decays)


if __name__ == "__main__":
    print(get_d_reco_decays(d_charge="D+"))
    print(get_d_reco_decays(d_charge="D0"))
