"""
B -> D l nu FF Reweighting (CLN -> BGL) with eFFORT
"""

import numpy as np
import pandas as pd

from uncertainties import ufloat
from typing import Tuple, List, Optional

from eFFORT.utility import PDG
from eFFORT.SLBToC.BToDLNu import BToDLNuCLNBelle, BToDLNuBGL, BToDLNu

from rdstar.utilities.logging import log_to_default

from rdstar.mc_reweighting.utility import FFReweightingBase
from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_info import GenMCIDs
from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name


__all__ = [
    "BtoDFFReweighting",
]


logger = log_to_default(name=__name__)


class BtoDFFReweighting(FFReweightingBase):
    """
    Applies form factor reweighting for B->Dstlv decays for a given
    DataFrame.
    """

    weight_name: str = "weight_ff_btod_"
    nom_weight_name: str = f"{weight_name}nom"
    base_var_weight_name: str = f"{weight_name}var_"

    vcb_cln: float = 41.1e-3
    # self.vcb_bgl: float = 40.83e-3

    # Parameters are produced by Florian using a private fit and data from https://arxiv.org/pdf/1510.03657v3.pdf
    params: np.ndarray = np.array(
        [
            ufloat(4.11837e-02, 1.15178e-03),  # V_cb
            ufloat(1.26147e-02, 9.78879e-05),  # bgl fplus coeff 0
            ufloat(-9.62084e-02, 3.34148e-03),  # bgl fplus coeff 1
            ufloat(4.13884e-01, 9.44471e-02),  # bgl fplus coeff 2
            ufloat(-1.73699e-01, 8.91809e-01),  # bgl fplus coeff 3
        ]
    )
    param_central_values: np.array = np.array([cv.nominal_value for cv in params])
    param_uncertainties: np.array = np.array([cv.std_dev for cv in params])
    param_correlation_matrix: np.ndarray = np.array(
        [
            [1.000, -0.402, -0.238, -0.110, 0.047],
            [-0.402, 1.000, 0.245, -0.161, 0.020],
            [-0.238, 0.245, 1.000, -0.654, 0.272],
            [-0.110, -0.161, -0.654, 1.000, -0.770],
            [0.047, 0.020, 0.272, -0.770, 1.000],
        ]
    )
    covariance_matrix: np.matrix = np.outer(param_uncertainties, param_uncertainties) * param_correlation_matrix

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._weight_name_str: str = BtoDFFReweighting.weight_name

        self.B1_is_charged_btodl: Optional[pd.Series] = None
        self.B2_is_charged_btodl: Optional[pd.Series] = None
        self.B1_is_mixed_btodl: Optional[pd.Series] = None
        self.B2_is_mixed_btodl: Optional[pd.Series] = None

        self.B1_reweighting_var = self.b1_w
        self.B2_reweighting_var = self.b2_w

        self.varied_central_values = self.get_varied_central_values(
            central=self.param_central_values, covariance_matrix=self.covariance_matrix
        )

        self._btod_charged_cln = BToDLNuCLNBelle(m_B=PDG.m_Bplus, m_D=PDG.m_Dzero, V_cb=self.vcb_cln)

        self._btod_charged_bgl_nom = BToDLNuBGL(
            m_B=PDG.m_Bplus,
            m_D=PDG.m_Dzero,
            V_cb=self.param_central_values[0],
            bgl_fplus_coefficients=self.param_central_values[1:],
        )

        self._btod_mixed_cln = BToDLNuCLNBelle(m_B=PDG.m_Bzero, m_D=PDG.m_Dplus, V_cb=self.vcb_cln)

        self._btod_mixed_bgl_nom = BToDLNuBGL(
            m_B=PDG.m_Bzero,
            m_D=PDG.m_Dplus,
            V_cb=self.param_central_values[0],
            bgl_fplus_coefficients=self.param_central_values[1:],
        )

        self._btod_charged_bgl_var: List[Tuple[BToDLNuBGL, BToDLNuBGL]] = list()
        self._btod_mixed_bgl_var: List[Tuple[BToDLNuBGL, BToDLNuBGL]] = list()
        for variation in self.varied_central_values:
            btod_charged_bgl_var_up = BToDLNuBGL(
                m_B=PDG.m_Bplus, m_D=PDG.m_Dzero, V_cb=variation[0][0], bgl_fplus_coefficients=variation[0][1:]
            )
            btod_charged_bgl_var_down = BToDLNuBGL(
                m_B=PDG.m_Bplus, m_D=PDG.m_Dzero, V_cb=variation[1][0], bgl_fplus_coefficients=variation[1][1:]
            )
            btod_mixed_bgl_var_up = BToDLNuBGL(
                m_B=PDG.m_Bzero, m_D=PDG.m_Dplus, V_cb=variation[0][0], bgl_fplus_coefficients=variation[0][1:]
            )
            btod_mixed_bgl_var_down = BToDLNuBGL(
                m_B=PDG.m_Bzero, m_D=PDG.m_Dplus, V_cb=variation[1][0], bgl_fplus_coefficients=variation[1][1:]
            )

            assert len(self._btod_charged_bgl_var) == len(self._btod_mixed_bgl_var)

            self._btod_charged_bgl_var.append((btod_charged_bgl_var_up, btod_charged_bgl_var_down))
            self._btod_mixed_bgl_var.append((btod_mixed_bgl_var_up, btod_mixed_bgl_var_down))
        assert len(self._btod_charged_bgl_var) == len(self._btod_mixed_bgl_var)

    def get_condition_tuples_and_vars(self) -> Tuple[List[Tuple[pd.Series, pd.Series]], Tuple[str, str]]:
        b_condition_tuples: List[Tuple[pd.Series, pd.Series]] = [
            (self.B1_is_charged_btodl, self.B1_is_mixed_btodl),
            (self.B2_is_charged_btodl, self.B2_is_mixed_btodl),
        ]
        b_variables: Tuple[str, str] = (self.B1_reweighting_var, self.B2_reweighting_var)

        assert len(b_condition_tuples) == 2, len(b_condition_tuples)
        assert all(len(s.index) == len(b_condition_tuples[0][0].index) for t in b_condition_tuples for s in t)
        return b_condition_tuples, b_variables

    def _add_nominal_weights(self, df: pd.DataFrame) -> None:
        self.nominal_weight_key = f"{self.weight_name_str}nom"
        df.loc[:, self.nominal_weight_key] = np.nan

        b_condition_tuples, b_variables = self.get_condition_tuples_and_vars()

        for b_condition_tuple, b_variable in zip(b_condition_tuples, b_variables):
            df.loc[b_condition_tuple[0], self.nominal_weight_key] = self._calculate_weights(
                old=self._btod_charged_cln,
                new=self._btod_charged_bgl_nom,
                w=df.loc[b_condition_tuple[0], b_variable].values,
            )
            df.loc[b_condition_tuple[1], self.nominal_weight_key] = self._calculate_weights(
                old=self._btod_mixed_cln,
                new=self._btod_mixed_bgl_nom,
                w=df.loc[b_condition_tuple[1], b_variable].values,
            )

        df.loc[df[self.nominal_weight_key].isnull(), self.nominal_weight_key] = 1.0
        logger.info("Nominal B->Dlnu FF weights added.")

    def _add_varied_weights(self, df: pd.DataFrame) -> None:
        self.variation_weight_keys = list()

        b_condition_tuples, b_variables = self.get_condition_tuples_and_vars()

        for i, (charged_var, mixed_var) in enumerate(zip(self._btod_charged_bgl_var, self._btod_mixed_bgl_var)):
            up_charged = charged_var[0]
            down_charged = charged_var[1]
            up_mixed = mixed_var[0]
            down_mixed = mixed_var[1]

            up_key = f"{self.weight_name_str}var_{i}_up"
            down_key = f"{self.weight_name_str}var_{i}_down"

            self.variation_weight_keys.append(up_key)
            self.variation_weight_keys.append(down_key)

            for var_key, new_var in zip([up_key, down_key], [(up_charged, up_mixed), (down_charged, down_mixed)]):
                df.loc[:, var_key] = np.nan
                for b_condition_tuple, b_variable in zip(b_condition_tuples, b_variables):
                    df.loc[b_condition_tuple[0], var_key] = self._calculate_weights(
                        old=self._btod_charged_cln,
                        new=new_var[0],
                        w=df.loc[b_condition_tuple[0], b_variable].values,
                    )
                    df.loc[b_condition_tuple[1], var_key] = self._calculate_weights(
                        old=self._btod_mixed_cln,
                        new=new_var[1],
                        w=df.loc[b_condition_tuple[1], b_variable].values,
                    )
                df.loc[df[var_key].isnull(), var_key] = 1.0

            logger.info(f"Variation {i} B->Dlnu FF weights added.")

    @staticmethod
    def _calculate_weights(old: BToDLNu, new: BToDLNu, w: np.ndarray) -> np.ndarray:
        old_d_gamma = old.dGamma_dw(w)
        new_d_gamma = new.dGamma_dw(w)

        return (old.Gamma() / new.Gamma()) * (new_d_gamma / old_d_gamma)

    def add_weights(self, df: pd.DataFrame) -> None:
        btodlnu_gen_id: int = GenMCIDs.Dl

        self.B1_is_charged_btodl = (
            (df[gen_b1_id_col_name] == btodlnu_gen_id)
            & ((df[self.b_index_var] == 1) | (df[self.b_via_tau_index_var] == 1))
            & (abs(df[self.get_b_meson_mc_pdg_var(b_index=0)]) == self.charged_b_meson_pdg_code)
        )
        self.B2_is_charged_btodl = (
            (df[gen_b2_id_col_name] == btodlnu_gen_id)
            & ((df[self.b_index_var] == 2) | (df[self.b_via_tau_index_var] == 2))
            & (abs(df[self.get_b_meson_mc_pdg_var(b_index=1)]) == self.charged_b_meson_pdg_code)
        )
        self.B1_is_mixed_btodl = (
            (df[gen_b1_id_col_name] == btodlnu_gen_id)
            & ((df[self.b_index_var] == 1) | (df[self.b_via_tau_index_var] == 1))
            & (abs(df[self.get_b_meson_mc_pdg_var(b_index=0)]) == self.mixed_b_meson_pdg_code)
        )
        self.B2_is_mixed_btodl = (
            (df[gen_b2_id_col_name] == btodlnu_gen_id)
            & ((df[self.b_index_var] == 2) | (df[self.b_via_tau_index_var] == 2))
            & (abs(df[self.get_b_meson_mc_pdg_var(b_index=1)]) == self.mixed_b_meson_pdg_code)
        )

        assert (self.B1_is_charged_btodl & self.B1_is_mixed_btodl).sum() == 0
        assert (self.B2_is_charged_btodl & self.B2_is_mixed_btodl).sum() == 0

        logger.info(
            f"Number of candidates to reweight:\n\t#Charged B meson = "
            f"{len(self.B1_is_charged_btodl.index) + len(self.B2_is_charged_btodl.index)}"
            f"\n\t#Mixed B meson = {len(self.B1_is_mixed_btodl.index) + len(self.B2_is_mixed_btodl.index)}"
        )

        self._add_nominal_weights(df)
        self._add_varied_weights(df)

        b1_cond = self.B1_is_charged_btodl | self.B1_is_mixed_btodl
        b2_cond = self.B2_is_charged_btodl | self.B2_is_mixed_btodl
        self._write_weight_info(df, conditions=(b1_cond, b2_cond))

    @staticmethod
    def finalize_weights(
        df: pd.DataFrame, weight_base_name: str = "weight_ff_btod_", threshold: float = 3.0, reset_special_to: float = 0.0
    ) -> None:
        nom_weight_col = f"{weight_base_name}nom"
        assert nom_weight_col in df.columns

        wcs = [c for c in df.columns if c.startswith(weight_base_name)]

        btodlnu_gen_ids: List[int] = [GenMCIDs.Dl]

        BtoDFFReweighting.reset_problematic_special_mc(
            df=df, weight_base_name=weight_base_name, gen_ids=btodlnu_gen_ids, reset_special_to=reset_special_to
        )

        assert not pd.isnull(df[wcs]).any().any()
        assert (df.loc[:, nom_weight_col] >= 0.0).all()
        assert (df.loc[:, nom_weight_col] < threshold).all()
