"""
Automated generation of weights for reweighting of B -> D** l nu decays from ISGW2 to LLSW model.
The weights are calculated based on MC simulations with both models.

Felix Metzner, Maximilian Welsch (2018 - 2020)
"""

import os
import pdg
import math
import basf2
import shutil
import rdstar
import uproot
import b2luigi
import itertools
import pandas as pd

from multiprocessing import pool
from b2luigi.basf2_helper import Basf2PathTask
from typing import Union, Tuple, List, Dict, Generator

from analysistools.reweighting.rbf_reweighter import RbfReweighter
from analysistools.dataframe_tools.uproot_compatibility import uproottree2df

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.mc_reweighting.BtoDststlnu.event_generation import DststReweightingEventGenerator
from rdstar.mc_reweighting.BtoDststlnu.reweighting_variables import calculate_reweighting_variables
from rdstar.mc_reweighting.BtoDststlnu.calculate_llsw_variations import DecayFileWithLLSWVariationsCreator

__all__ = [
    "DststReweightingMaster",
    "RbfCreationTask",
    "ReweightingDataMergerTask",
    "MergeIntoCommonDataFrameTask",
]


logger = log_to_default(name=__name__)


class LLSWDecayFileVariationsCreatorTask(b2luigi.Task):
    log_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    result_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str

    output_dir_name = "llsw_decay_file_variations_check"  # type: str
    output_key = "llsw_decay_file_variation_check"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

        self.base_output_dir = os.path.join(self.result_dir, self.output_dir_name)
        self.check_file_path = os.path.join(self.base_output_dir, f"{self.output_key}.info")

        decay_file_directory = DecayFileWithLLSWVariationsCreator.decay_files_dir_path
        self.dec_file_creator = DecayFileWithLLSWVariationsCreator(decay_file_directory=decay_file_directory)

    def output(self) -> Dict[str, b2luigi.LocalTarget]:
        output_dict = self.dec_file_creator.get_future_output()
        output_dict.update({self.output_key: self.check_file_path})
        return {key: b2luigi.LocalTarget(path) for key, path in output_dict.items()}

    def run(self) -> None:
        force_overwrite = False  # type: bool
        self.dec_file_creator.create_llsw_variations(force_overwrite=force_overwrite)

        with open(file=self.check_file_path, mode="w") as new_file:
            new_file.write(f"Files have been created or {'replaced' if force_overwrite else 'checked'}.")


class EventGenerationTask(Basf2PathTask):
    result_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    log_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    n_events = b2luigi.IntParameter(positional=False)  # type: int
    decay_file_path = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    model_key = b2luigi.Parameter(positional=False)  # type: str

    output_key = "events.root"  # type: str

    d1prime_pdg = 20423  # type: int
    updated_d1prime_mass = 2.427  # type: float

    model_to_evt_pdl_mapping = {
        "ISGW2": "my_belleII_evt.pdl",
        "LLSW": "my_belleII_evt.pdl",
    }  # type: Dict[str, PathType]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

        assert all(k in RbfCreationTask.models for k in self.model_to_evt_pdl_mapping.keys()), [
            k for k in self.model_to_evt_pdl_mapping.keys() if k not in RbfCreationTask.models
        ]
        assert self.model_key in RbfCreationTask.models, self.model_key
        assert self.model_key in self.model_to_evt_pdl_mapping.keys(), self.model_key

    @staticmethod
    def get_base_directory() -> PathType:
        return os.path.dirname(os.path.abspath(__file__))

    def get_evt_pdl_directory(self) -> PathType:
        return os.path.join(self.get_base_directory(), "evt_pdl")

    def _get_evt_pdl_file_path(self) -> PathType:
        evt_pdl_path = os.path.join(self.get_evt_pdl_directory(), self.model_to_evt_pdl_mapping[self.model_key])

        logger.info(
            f"Using evt pdl file from\n\t{evt_pdl_path}\nfor the decay file\n"
            f"\t{self.decay_file_path}\n which is simulated using the {self.model_key} model."
        )

        assert os.path.isfile(path=evt_pdl_path), evt_pdl_path
        return evt_pdl_path

    def output(self) -> Generator[Dict[str, b2luigi.LocalTarget], None, None]:
        yield self.add_to_output(self.output_key)

    def create_path(self) -> basf2.Path:
        pdg.load(self._get_evt_pdl_file_path())

        d1prime_mass_for_check = pdg.get(pdg.to_name(EventGenerationTask.d1prime_pdg)).Mass()
        if not d1prime_mass_for_check == EventGenerationTask.updated_d1prime_mass:
            raise RuntimeError(
                f"The wrong evt.pdl file is being used for the MC generation!\n"
                f"\tD'_1 mass used = {d1prime_mass_for_check}\n"
                f"\tExpected D'_1 mass = {EventGenerationTask.updated_d1prime_mass}"
            )

        event_generator = DststReweightingEventGenerator(
            number_of_events=self.n_events,
            decay_file_path=os.path.join(self.get_base_directory(), self.decay_file_path),
            output_path=self.get_output_file_name(self.output_key),
        )
        path = event_generator.get_path()
        return path

    def requires(self) -> Generator[Generator[b2luigi.Task, None, None], None, None]:
        yield self.clone(LLSWDecayFileVariationsCreatorTask)


class MergeIntoCommonDataFrameTask(b2luigi.Task):
    result_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    log_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    n_events = b2luigi.IntParameter(positional=False)  # type: int
    model_variation = b2luigi.Parameter(positional=False)  # type: str
    decay_channel = b2luigi.Parameter(positional=False)  # type: str

    output_key = "merged.pkl"  # type: str
    variation_col_name = "variation"  # type: str
    model_col_name = "model"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

        decay_files_dir_path = DecayFileWithLLSWVariationsCreator.decay_files_dir_path
        self._decay_files_directory = os.path.join(decay_files_dir_path, self.model_variation, self.decay_channel)

    def output(self) -> Generator[Dict[str, b2luigi.LocalTarget], None, None]:
        yield self.add_to_output(self.output_key)

    def _get_variation_index(self) -> str:
        key_str = "LLSW_variation_"
        if self.model_variation.startswith(key_str):
            return self.model_variation.replace(key_str, "")
        else:
            return "nominal"

    def _get_model_name(self) -> str:
        return self.model_variation.split(sep="_", maxsplit=1)[0]

    def run(self) -> None:
        input_file_paths = self.get_input_file_names(EventGenerationTask.output_key)

        dfs = []
        number_of_events = 0
        for file_path in input_file_paths:
            assert os.path.isfile(path=file_path), file_path
            assert os.path.splitext(p=file_path)[1] == ".root", os.path.splitext(p=file_path)
            data_file = uproot.open(file_path)
            tree = data_file[data_file.keys()[0]]
            dfs.append(uproottree2df(tree))
            number_of_events += len(dfs[-1].index)

        df = pd.concat(dfs, ignore_index=True, sort=False)
        assert len(df.index) == number_of_events, (len(df.index), number_of_events)

        assert self.variation_col_name not in df.columns, (self.variation_col_name, df.columns)
        df.loc[:, self.variation_col_name] = self._get_variation_index()
        assert self.model_col_name not in df.columns, (self.model_col_name, df.columns)
        df.loc[:, self.model_col_name] = self._get_model_name()

        df.to_pickle(path=self.get_output_file_name(self.output_key))

    def get_decay_file_names(self) -> Tuple[PathType, ...]:
        var_index = self._get_variation_index()
        variation_tag = "nominal" if var_index == "nominal" else f"variation_{var_index}"
        decay_file_names = []
        for (charge, lepton) in itertools.product(("zero", "minus"), ("electron", "muon")):
            other_charge = "minus" if charge == "zero" else "zero"
            decay_tag = f"{self._get_model_name().lower()}_B{charge}2{self.decay_channel}{other_charge}_{lepton}"
            decay_file_names.append(f"{decay_tag}_{variation_tag}.dec")

        if variation_tag == "nominal":
            assert all(os.path.isfile(os.path.join(self._decay_files_directory, dfn)) for dfn in decay_file_names), [
                dfn for dfn in decay_file_names if not os.path.isfile(os.path.join(self._decay_files_directory, dfn))
            ]
        return tuple(decay_file_names)

    def requires(self) -> Generator[b2luigi.Task, None, None]:
        for decay_file_name in self.get_decay_file_names():
            yield self.clone(
                EventGenerationTask,
                decay_file_path=os.path.join(self._decay_files_directory, decay_file_name),
                model_key=self._get_model_name(),
            )


@b2luigi.requires(MergeIntoCommonDataFrameTask)
class ReweightingVariablesCalculationTask(b2luigi.Task):
    output_key = "reweighting_vars.h5"  # type: str
    df_key = "df"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

    def output(self) -> Generator[Generator[b2luigi.Task, None, None], None, None]:
        yield self.add_to_output(self.output_key)

    def run(self) -> None:
        # There is only one input file so take only the 0th component
        input_file = self.get_input_file_names(MergeIntoCommonDataFrameTask.output_key)[0]
        df = pd.read_pickle(path=input_file)

        if self.decay_channel == "Dprimeone":
            col_name_dict = DststReweightingEventGenerator.get_predefined_root_compatible_variables_dict()
            d_mass_col = col_name_dict["d_meson_mass"]
            d_delta_mass_col = col_name_dict["d_meson_delta_mass"]
            unique_masses = (df[d_mass_col] - df[d_delta_mass_col]).unique()

            assert len(unique_masses) == 1, unique_masses
            assert math.isclose(
                a=unique_masses[0],
                b=EventGenerationTask.updated_d1prime_mass,
                abs_tol=1e-5,
                rel_tol=1e-6,
            ), (unique_masses[0], EventGenerationTask.updated_d1prime_mass)

        calculate_reweighting_variables(df)
        df.to_hdf(path_or_buf=self.get_output_file_name(self.output_key), key=self.df_key, mode="w")


class ReweightingDataMergerTask(b2luigi.Task):
    result_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    log_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    n_events = b2luigi.IntParameter(positional=False)  # type: int
    decay_channel = b2luigi.Parameter(positional=False)  # type: str
    model_variations = b2luigi.ListParameter(positional=False, hashed=True)  # type: List[str]

    df_key = "df"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

    def _get_model_name(self) -> str:
        model_name = self.model_variations[0].split(sep="_", maxsplit=1)[0]
        assert all(mod_var.split("_", 1)[0] == model_name for mod_var in self.model_variations), self.model_variations
        return model_name

    @staticmethod
    def get_df_output_key(decay_channel: str, model: str) -> str:
        assert model in RbfCreationTask.models, (model, RbfCreationTask.models)
        assert decay_channel in DststReweightingMaster.decay_channels, (
            decay_channel,
            DststReweightingMaster.decay_channels,
        )
        return f"{decay_channel}_{model}"

    def do_merge(self, dry_run: bool = False) -> Dict[str, PathType]:
        base_result_dir_path = os.path.join(self.result_dir, self._get_model_name())  # type: PathType
        output_key = self.get_df_output_key(decay_channel=self.decay_channel, model=self._get_model_name())
        file_name = f"{output_key}_{ReweightingVariablesCalculationTask.output_key}"
        output_path = os.path.join(base_result_dir_path, file_name)
        temp_output_path = os.path.join(base_result_dir_path, f"temp_{file_name}")

        output_dict = {output_key: output_path}  # type: Dict[str, PathType]

        if not dry_run:
            os.makedirs(name=base_result_dir_path, exist_ok=True)
            input_paths = self.get_input_file_names(ReweightingVariablesCalculationTask.output_key)
            dfs = [
                pd.read_hdf(path_or_buf=input_path, key=ReweightingVariablesCalculationTask.df_key)
                for input_path in input_paths
            ]

            assert all(isinstance(d, pd.DataFrame) for d in dfs), [type(d) for d in dfs]

            df = pd.concat(dfs, ignore_index=True, sort=False)
            df.to_hdf(path_or_buf=temp_output_path, key=self.df_key, mode="w")
            os.rename(src=temp_output_path, dst=output_path)

        return output_dict

    def output(self) -> Dict[str, b2luigi.LocalTarget]:
        output_dict = self.do_merge(dry_run=True)
        return {output_key: b2luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def run(self) -> None:
        self.do_merge()

    def requires(self) -> Generator[b2luigi.Task, None, None]:
        for model_variation in self.model_variations:
            yield self.clone(
                ReweightingVariablesCalculationTask,
                model_variation=model_variation,
            )


class RbfCreationTask(b2luigi.Task):
    result_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    log_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    decay_channel = b2luigi.Parameter()  # type: str
    n_events = b2luigi.IntParameter(positional=False)  # type: int

    models = (
        "ISGW2",
        "LLSW",
    )  # type: Tuple[str, ...]

    cols_dict = DststReweightingEventGenerator.get_predefined_root_compatible_variables_dict()  # type: Dict[str, str]
    w_col_name = cols_dict["wMC"]  # type: str
    costhetal_col_name = cols_dict["costhetalMC"]  # type: str
    costhetanu_col_name = cols_dict["costhetavMC"]  # type: str
    chi_col_name = cols_dict["chi_MC"]  # type: str

    rbf_arguments = {
        "Done": {
            "variables": (w_col_name, costhetal_col_name, costhetanu_col_name),
            "num_bins": (10, 10, 10),
            "limits": ((1.0, 1.42), (-1.0, 1.0), (-1.0, 1.0)),
        },
        "Dtwostar": {
            "variables": (w_col_name, costhetal_col_name),
            "num_bins": (20, 20),
            "limits": ((1.0, 1.42), (-1.0, 1.0)),
        },
        "Dprimeone": {
            "variables": (w_col_name, costhetal_col_name),
            "num_bins": (20, 20),
            "limits": ((1.0, 1.425), (-1.0, 1.0)),
        },
        "Dzerostar": {
            "variables": (w_col_name,),
            "num_bins": (50,),
            "limits": ((1.0, 1.425),),
        },
    }  # type: Dict[str, Dict[str, Tuple[Union[str, int, Tuple[float, float]], ...]]]

    output_dir_name = "rbf_output"  # type: str
    package_data_dir_name = "data/dstst_reweighting_data"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

        self.base_output_dir = os.path.join(self.result_dir, self.output_dir_name)
        self.package_target_path = os.path.join(os.path.dirname(rdstar.__file__), self.package_data_dir_name)
        assert os.path.exists(self.package_target_path)

        assert set(RbfCreationTask.rbf_arguments.keys()) == set(DststReweightingMaster.decay_channels), (
            RbfCreationTask.rbf_arguments.keys(),
            DststReweightingMaster.decay_channels,
        )

    @staticmethod
    def get_model_variations(model_key: str, decay_channel: str) -> Tuple[str, ...]:
        assert model_key in RbfCreationTask.models, (model_key, RbfCreationTask.models)
        assert decay_channel in RbfCreationTask.rbf_arguments, (decay_channel, RbfCreationTask.rbf_arguments.keys())

        variation_list = [f"{model_key}_nominal"]

        if model_key == "LLSW":
            n_vars = DecayFileWithLLSWVariationsCreator.get_number_of_variations_for_meson(d_meson_name=decay_channel)
            for i_var in range(n_vars):
                variation_list.extend(
                    [
                        DecayFileWithLLSWVariationsCreator.get_llsw_variation_name(var_index=i_var, direction="down"),
                        DecayFileWithLLSWVariationsCreator.get_llsw_variation_name(var_index=i_var, direction="up"),
                    ]
                )

        return tuple(variation_list)

    def get_llsw_variation_keys(self, remove_model_name: bool = False) -> Tuple[str, ...]:
        keys = self.get_model_variations(model_key="LLSW", decay_channel=self.decay_channel)
        if remove_model_name:
            keys = tuple([k.replace("LLSW_", "").replace("variation_", "") for k in keys])
        return keys

    def load_llsw_data(self) -> Dict[str, pd.DataFrame]:
        input_key = ReweightingDataMergerTask.get_df_output_key(decay_channel=self.decay_channel, model="LLSW")
        path_to_hdf = self.input()["LLSW"][input_key].path
        data = pd.read_hdf(path_to_hdf, ReweightingDataMergerTask.df_key)
        assert isinstance(data, pd.DataFrame), type(pd.DataFrame)
        variations = list(data[MergeIntoCommonDataFrameTask.variation_col_name].unique())

        return {v: data.query(f"{MergeIntoCommonDataFrameTask.variation_col_name}=='{v}'") for v in variations}

    def load_isgw2_data(self) -> pd.DataFrame:
        input_key = ReweightingDataMergerTask.get_df_output_key(decay_channel=self.decay_channel, model="ISGW2")
        path_to_hdf = self.input()["ISGW2"][input_key].path
        data = pd.read_hdf(path_to_hdf, ReweightingDataMergerTask.df_key)
        assert isinstance(data, pd.DataFrame), type(pd.DataFrame)
        return data

    def _get_output_path(self, key: str, temporary: bool = False) -> PathType:
        name = f"rbf_{self.decay_channel}_{key}"
        temp_prefix = "temp_" if temporary else ""
        path = os.path.join(self.base_output_dir, f"{temp_prefix}{name}.json")
        return path

    def copy_files_to_package(self, file_path: PathType) -> None:
        assert os.path.isfile(file_path), file_path
        file_name = str(os.path.basename(file_path))
        if file_name.startswith("temp_"):
            file_name = file_name.split(sep="temp_", maxsplit=1)[1]
        target_path = os.path.join(self.package_target_path, file_name)
        if os.path.exists(target_path):
            os.remove(target_path)
        shutil.copyfile(src=file_path, dst=target_path)

    def perform_rbf_creation(self, dry_run: bool = False) -> Dict[str, PathType]:
        llsw_variation_list = self.get_llsw_variation_keys(remove_model_name=True)
        output_dict = {key: self._get_output_path(key=key) for key in llsw_variation_list}

        if dry_run:
            return output_dict

        llsw_sample = self.load_llsw_data()
        isgw2_sample = self.load_isgw2_data()

        os.makedirs(self.base_output_dir, exist_ok=True)

        cols = list(self.rbf_arguments[self.decay_channel]["variables"])
        args = [
            (self.decay_channel, key, isgw2_sample[cols], df[cols], self._get_output_path(key=key, temporary=True))
            for key, df in llsw_sample.items()
        ]
        p_out = pool.Pool(processes=4).starmap(RbfCreationTask._create_rbf, args)

        assert len(p_out) == len(output_dict), (len(p_out), len(output_dict))

        for key in llsw_variation_list:
            temp_path = self._get_output_path(key=key, temporary=True)
            real_path = self._get_output_path(key=key, temporary=False)
            assert os.path.exists(temp_path), temp_path
            assert not os.path.exists(real_path), real_path
            self.copy_files_to_package(file_path=temp_path)
            os.rename(src=temp_path, dst=real_path)

        assert all(os.path.isfile(p) for p in output_dict.values()), [
            p for p in output_dict.values() if not os.path.isfile(p)
        ]

        return output_dict

    @staticmethod
    def _create_rbf(
        dss_type: str,
        variation: str,
        isgw2_df: pd.DataFrame,
        llsw_df: pd.DataFrame,
        rbf_path: PathType,
    ) -> Tuple[str, RbfReweighter]:
        logger.info(f"Creating RBF for {dss_type} for {variation} and writing it to\n\t{rbf_path}")

        rbf = RbfReweighter(**(RbfCreationTask.rbf_arguments[dss_type]))
        rbf.create_interpolation(origin=isgw2_df, target=llsw_df)
        rbf.export_to_json(output_path=rbf_path)

        return variation, rbf

    def run(self) -> None:
        os.makedirs(name=self.base_output_dir, exist_ok=True)
        self.perform_rbf_creation(dry_run=False)

    def output(self) -> Dict[str, b2luigi.LocalTarget]:
        output_dict = self.perform_rbf_creation(dry_run=True)
        return {k: b2luigi.LocalTarget(p) for k, p in output_dict.items()}

    def requires(self) -> Dict[str, b2luigi.Task]:
        task_dict = {}
        for model in self.models:
            model_variations = self.get_model_variations(model_key=model, decay_channel=self.decay_channel)
            task_dict.update({model: self.clone(ReweightingDataMergerTask, model_variations=list(model_variations))})

        return task_dict


class DststReweightingMaster(b2luigi.WrapperTask):
    """
    Main Tasks for the monte carlo production. Requires all other tasks.
    You have to generate the decay files for the varied parameters by executing
    the `calculate_llsw_variations.py` script.
    """

    result_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    log_dir = b2luigi.Parameter(positional=False, hashed=True)  # type: str
    n_events = b2luigi.IntParameter(positional=False)  # type: int

    decay_channels = (
        "Done",
        "Dprimeone",
        "Dtwostar",
        "Dzerostar",
    )  # type: Tuple[str, ...]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        b2luigi.set_setting(key="result_dir", value=self.result_dir)
        b2luigi.set_setting(key="log_dir", value=self.log_dir)

    def requires(self) -> Generator[b2luigi.Task, None, None]:
        for decay_channel in self.decay_channels:
            yield self.clone(RbfCreationTask, decay_channel=decay_channel)
