"""
Defines MC bin composition plots.
These allow to analyse the composition of the bins of a histogram
with respect to a second variable which defines the components.

Felix Metzner 2021
"""

import os
import numpy as np
import pandas as pd

from typing import Optional, Tuple, List, Dict, Any, NamedTuple, Callable, Type

from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.rdstar_samples import main_sample_keys
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.validation.observable_validation import plot_bin_composition
from rdstar.offline_analysis.analysis_plots.analysis_plot_base import AnalysisPlotBaseClass
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.observables.offline_observable_definition_base import BinnedOfflineObservableBase
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import missing_mass_hist_var

from rdstar.offline_analysis.observables.reco_modes_observables import RecoIDObservables, BTagMesonRecoIDDefinition
from rdstar.offline_analysis.observables.binned_mc_observables import (
    OfflineBinnedMCObservables,
    LeptonMCPDGBinID,
    DMesonMCPDGBinID,
    MCDataTypeBinID,
    Q2TruthSigSide,
    WTruthSigSide,
)


__all__ = [
    "MCBinCompositionDefinition",
    "MCBinCompositionPlotter",
]


mbc_fine_hist_var = HistVariable(
    df_label="daughter__bo0__cm__spMbc__bc",
    n_bins=16,
    scope=(5.270, 5.290),
    var_name=LatexObsNames.b_tag_mbc,
    unit=LatexUnits.gev,
    use_log_scale=False,
)

mbc_fine_binning = Binning(
    bins=mbc_fine_hist_var.n_bins,
    dimensions=1,
    scope=mbc_fine_hist_var.scope,
    log_scale=False,
)

fei_probability_fine_hist_var = HistVariable(
    df_label="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc",
    n_bins=20,
    scope=(0.0, 1.0),
    var_name=LatexObsNames.fei_prob,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

fei_probability_fine_binning = Binning(
    bins=fei_probability_fine_hist_var.n_bins,
    dimensions=1,
    scope=fei_probability_fine_hist_var.scope,
    log_scale=False,
)

fei_probability_special_hist_var = HistVariable(
    df_label="daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc",
    n_bins=20,
    scope=(0.001, 1.0),
    var_name=LatexObsNames.fei_prob,
    unit=LatexUnits.no_unit,
    use_log_scale=False,
)

__base: float = 1000.0
__bins: int = fei_probability_special_hist_var.n_bins
__scale: float = (__base - 1) / __bins
_fei_log_1000_bins: np.array = np.flip(1 - (np.log(np.arange(1.0, __base + __scale, __scale)) / np.log(__base)))
_fei_log_1000_bins[0] = 0.001  # setting lower bound to cut value.
fei_probability_log1000_binning = Binning(
    bins=tuple(_fei_log_1000_bins),
    dimensions=1,
    scope=(_fei_log_1000_bins[0], _fei_log_1000_bins[-1]),
    log_scale=False,
)

_fei_prob_eqf_bins: Tuple[float, ...] = tuple(
    [
        0.001,
        0.00119469,
        0.00144025,
        0.00174844,
        0.00214756,
        0.00266676,
        0.00336498,
        0.004309,
        0.00561211,
        0.00742395,
        0.01008662,
        0.01401974,
        0.019896,
        0.02899445,
        0.04356232,
        0.06810507,
        0.1106317,
        0.18925892,
        0.34744963,
        0.67438808,
        1.0,
    ]
)

fei_probability_eqf_binning = Binning(
    bins=_fei_prob_eqf_bins,
    dimensions=1,
    scope=(_fei_prob_eqf_bins[0], _fei_prob_eqf_bins[-1]),
    log_scale=False,
)


class MCBinCompositionDefinition(NamedTuple):
    name: str
    main_hist_var: HistVariable
    secondary_hist_var: HistVariable
    secondary_binning: Binning
    variable_definition: Optional[Callable] = None
    variable_definition_class: Optional[Type[BinnedOfflineObservableBase]] = None
    tick_label_mapping: Optional[Dict[int, str]] = None


m_bc_fine_binning_based_definition = MCBinCompositionDefinition(
    name="m_bc_fine_binning_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=mbc_fine_hist_var,
    secondary_binning=mbc_fine_binning,
)

fei_probability_fine_binning_based_definition = MCBinCompositionDefinition(
    name="fei_prob_fine_binning_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=fei_probability_fine_hist_var,
    secondary_binning=fei_probability_fine_binning,
)

fei_probability_log1000_binning_based_definition = MCBinCompositionDefinition(
    name="fei_prob_log1000_binning_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=fei_probability_special_hist_var,
    secondary_binning=fei_probability_log1000_binning,
)

fei_probability_eqf_binning_based_definition = MCBinCompositionDefinition(
    name="fei_prob_eqf_binning_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=fei_probability_special_hist_var,
    secondary_binning=fei_probability_eqf_binning,
)

lepton_candidate_pdg_based_definition = MCBinCompositionDefinition(
    name=f"{LeptonMCPDGBinID.new_column_name}_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=LeptonMCPDGBinID.hist_var,
    secondary_binning=LeptonMCPDGBinID.binning,
    variable_definition=OfflineBinnedMCObservables.get_definition_function(name=LeptonMCPDGBinID.new_column_name),
    variable_definition_class=OfflineBinnedMCObservables.get_mc_obs_definition(name=LeptonMCPDGBinID.new_column_name),
    tick_label_mapping=LeptonMCPDGBinID.get_tick_label_mapping(),
)

first_b_daughter_pdg_based_definition = MCBinCompositionDefinition(
    name=f"{DMesonMCPDGBinID.new_column_name}_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=DMesonMCPDGBinID.hist_var,
    secondary_binning=DMesonMCPDGBinID.binning,
    variable_definition=OfflineBinnedMCObservables.get_definition_function(name=DMesonMCPDGBinID.new_column_name),
    variable_definition_class=OfflineBinnedMCObservables.get_mc_obs_definition(name=DMesonMCPDGBinID.new_column_name),
    tick_label_mapping=DMesonMCPDGBinID.get_tick_label_mapping(),
)

mc_type_based_definition = MCBinCompositionDefinition(
    name=f"{MCDataTypeBinID.new_column_name}_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=MCDataTypeBinID.hist_var,
    secondary_binning=MCDataTypeBinID.binning,
    variable_definition=OfflineBinnedMCObservables.get_definition_function(name=MCDataTypeBinID.new_column_name),
    variable_definition_class=OfflineBinnedMCObservables.get_mc_obs_definition(name=MCDataTypeBinID.new_column_name),
    tick_label_mapping=MCDataTypeBinID.get_tick_label_mapping(),
)

q2_truth_based_definition = MCBinCompositionDefinition(
    name=f"{Q2TruthSigSide.new_column_name}_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=Q2TruthSigSide.hist_var,
    secondary_binning=Q2TruthSigSide.binning,
    variable_definition=OfflineBinnedMCObservables.get_definition_function(name=Q2TruthSigSide.new_column_name),
    variable_definition_class=OfflineBinnedMCObservables.get_mc_obs_definition(name=Q2TruthSigSide.new_column_name),
    tick_label_mapping=Q2TruthSigSide.get_tick_label_mapping(),
)

w_truth_based_definition = MCBinCompositionDefinition(
    name=f"{WTruthSigSide.new_column_name}_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=WTruthSigSide.hist_var,
    secondary_binning=WTruthSigSide.binning,
    variable_definition=OfflineBinnedMCObservables.get_definition_function(name=WTruthSigSide.new_column_name),
    variable_definition_class=OfflineBinnedMCObservables.get_mc_obs_definition(name=WTruthSigSide.new_column_name),
    tick_label_mapping=WTruthSigSide.get_tick_label_mapping(),
)

b_tag_decay_id_based_definition = MCBinCompositionDefinition(
    name=f"{BTagMesonRecoIDDefinition.new_column_name}_based_composition_definition",
    main_hist_var=missing_mass_hist_var,
    secondary_hist_var=BTagMesonRecoIDDefinition.hist_var,
    secondary_binning=BTagMesonRecoIDDefinition.binning,
    variable_definition=RecoIDObservables.get_definition_function(name=BTagMesonRecoIDDefinition.new_column_name),
    variable_definition_class=RecoIDObservables.get_reco_id_definition(name=BTagMesonRecoIDDefinition.new_column_name),
    tick_label_mapping=BTagMesonRecoIDDefinition.get_tick_label_mapping(),
)


class MCBinCompositionPlotter(AnalysisPlotBaseClass):
    _plotter_id_key: str = "MC_Bin_Composition"
    base_dir_name: str = "MC_Bin_Composition_Plots"

    input_appendix: str = f"{WeightManager.finalized_weights_appendix}"
    input_with_res_corr_appendix: str = f"{input_appendix}_{ResolutionCorrectionTask.appendix}"

    relevant_selections: Tuple[str, ...] = tuple(sel_setup.selection_name for sel_setup in SelectionSetups.my_selections)

    composition_definitions: Tuple[MCBinCompositionDefinition, ...] = (
        m_bc_fine_binning_based_definition,
        fei_probability_fine_binning_based_definition,
        fei_probability_log1000_binning_based_definition,
        fei_probability_eqf_binning_based_definition,
        lepton_candidate_pdg_based_definition,
        first_b_daughter_pdg_based_definition,
        mc_type_based_definition,
        q2_truth_based_definition,
        w_truth_based_definition,
        b_tag_decay_id_based_definition,
    )

    weight_column_name_options: List[Optional[str]] = [
        None,
        WeightManager.final_weight_col,
        WeightManager.final_weight_with_additions_col,
    ]

    _default_n_legend_columns: int = 1
    _entries_per_legend_column: int = 12

    def __init__(
        self,
        output_dir_path: PathType,
        input_dictionary: Dict[str, PathType],
        selection_name: str,
        cs_tag: str,
        composition_definition_name: str,
        weight_col_name: Optional[str],
    ) -> None:
        super().__init__(
            output_dir_path=output_dir_path,
            input_dictionary=input_dictionary,
            selection_name=selection_name,
            cs_tag=cs_tag,
        )

        assert len(set([c.name for c in self.composition_definitions])) == len(self.composition_definitions)

        self.composition_definition_name: str = composition_definition_name
        self.weight_col_name: Optional[str] = weight_col_name
        _weight_str: str = "no_weights" if self.weight_col_name is None else self.weight_col_name
        self.sub_dir_name: str = f"{self.composition_definition_name}_with_{_weight_str}"

    @property
    def composition_definition(self) -> MCBinCompositionDefinition:
        _these_cds: List[MCBinCompositionDefinition] = [
            cd for cd in self.composition_definitions if cd.name == self.composition_definition_name
        ]
        assert len(_these_cds) == 1, (_these_cds, len(_these_cds))
        [this_cd] = _these_cds
        return this_cd

    @staticmethod
    def composition_definition_names() -> List[str]:
        return [cd.name for cd in MCBinCompositionPlotter.composition_definitions]

    @staticmethod
    def plotter_id_key() -> str:
        if not MCBinCompositionPlotter._plotter_id_key:
            raise RuntimeError(
                f"Plotter {MCBinCompositionPlotter.__class__.__name__} has no _plotter_id_key"
                f" {MCBinCompositionPlotter._plotter_id_key}"
            )
        return MCBinCompositionPlotter._plotter_id_key

    @staticmethod
    def get_plotter_options(**kwargs) -> Dict[str, List[Any]]:
        selection_name_argument_name: str = "selection_name"
        if selection_name_argument_name not in kwargs.keys():
            raise TypeError(f"Missing keyword argument '{selection_name_argument_name}' in call to get_plotter_options.")

        selection_name: str = kwargs[selection_name_argument_name]

        if selection_name not in MCBinCompositionPlotter.relevant_selections:
            return {}

        return {
            "composition_definition_name": MCBinCompositionPlotter.composition_definition_names(),
            "weight_col_name": MCBinCompositionPlotter.weight_column_name_options,
        }

    @staticmethod
    def get_required_task_type() -> Callable:
        return ResolutionCorrectionTask

    def get_relevant_input(self) -> Dict[str, PathType]:
        key_beginnings: List[str] = [f"{mc_key}_{self.input_appendix}_" for mc_key in main_sample_keys]
        key_ending: str = f"_{self.selection_name}_with_CS_{self.cs_tag}"

        relevant_input_dict = {
            k: p
            for k, p in self.input_dictionary.items()
            if (
                any(k.startswith(kb) for kb in key_beginnings)
                and k.endswith(key_ending)
                and os.path.splitext(p)[1] == ".h5"
            )
        }
        assert len(relevant_input_dict) == len(main_sample_keys), (
            relevant_input_dict,
            len(relevant_input_dict),
            key_beginnings,
            len(key_beginnings),
            key_ending,
            self.input_dictionary.keys(),
        )
        return relevant_input_dict

    def create_plots(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict: Dict[str, PathType] = {}

        n_legend_columns: int = self._default_n_legend_columns
        df: pd.DataFrame = self.load_df(dry_run=dry_run)

        if not dry_run:
            if self.weight_col_name is not None:
                assert self.weight_col_name in df.columns, self.weight_col_name

            if self.composition_definition.variable_definition is not None:
                assert self.composition_definition.variable_definition_class is not None
                obs_class: Type[BinnedOfflineObservableBase] = self.composition_definition.variable_definition_class
                if obs_class.new_column_name not in df.columns:
                    self.composition_definition.variable_definition(df=df)

            assert (
                self.composition_definition.main_hist_var.df_label in df.columns
            ), self.composition_definition.main_hist_var.df_label
            assert (
                self.composition_definition.secondary_hist_var.df_label in df.columns
            ), self.composition_definition.secondary_hist_var.df_label

            _sec_binning: Optional[Binning] = self.composition_definition.secondary_binning
            _sec_h_var_n_bins: int = self.composition_definition.secondary_hist_var.n_bins
            _n_bins: int = _sec_binning.num_bins_total if _sec_binning is not None else _sec_h_var_n_bins
            n_legend_columns = int(np.ceil((1.0 * _n_bins) / self._entries_per_legend_column))

        new_output_entries = plot_bin_composition(
            mc_df=df,
            main_hist_var=self.composition_definition.main_hist_var,
            secondary_hist_var=self.composition_definition.secondary_hist_var,
            secondary_binning=self.composition_definition.secondary_binning,
            secondary_hist_var_label_mapping=self.composition_definition.tick_label_mapping,
            weight_col=self.weight_col_name,
            plot_shape=True,
            dec_modes={k: v for k, v in BSigRecoModes.main_modes.main_reco_mode_infos_as_tuple},
            dec_titles=BSigRecoModes.main_modes.get_reco_mode_name_to_latex_label_mapping(short=False),
            legend_columns=n_legend_columns,
            export_path=os.path.join(self.output_dir_path, self.sub_dir_name),
            dry_run=dry_run,
        )
        assert all(k not in output_dict for k in new_output_entries.keys())
        assert all(v not in output_dict.values() for v in new_output_entries.values())
        output_dict.update(new_output_entries)

        return output_dict
