"""
Corrections for reconstruction channel dependent difference in reconstruction efficiency on MC and data.

Felix Metzner 2021
"""

import os
import json
import numpy as np
import pandas as pd

from dataclasses import dataclass, asdict
from typing import Tuple, Sequence, Dict, ClassVar, Any, Generator

from rdstar.utilities import PathType


__all__ = [
    "RecoEffCorrInfoContainer",
    "ReconstructionModeEfficiencyCorrector",
]


@dataclass(frozen=True)
class RecoEffCorrInfoContainer:
    reco_mode_id_column_name: str
    reco_mode_keys: Tuple[str, ...]
    reco_mode_ids: Tuple[Tuple[int, ...], ...]
    base_weight_column_name: str
    eval_hist_var_column_names: Tuple[str, ...]
    eval_hist_var_scopes: Tuple[Tuple[float, float], ...]
    correction_factors: Tuple[float, ...]
    correction_factor_errors: Tuple[float, ...]

    def __post_init__(self) -> None:
        assert len(self.reco_mode_keys) == len(self.reco_mode_ids), (len(self.reco_mode_keys), len(self.reco_mode_ids))
        assert len(self.correction_factors) == len(self.reco_mode_ids), (
            len(self.correction_factors),
            len(self.reco_mode_ids),
        )
        assert len(self.correction_factors) == len(self.correction_factor_errors), (
            len(self.correction_factors),
            len(self.correction_factor_errors),
        )

        assert all(isinstance(v, int) for vv in self.reco_mode_ids for v in vv), self.reco_mode_ids

        assert len(self.all_reco_mode_ids) == len(set(self.all_reco_mode_ids)), (
            self.reco_mode_ids,
            self.all_reco_mode_ids,
        )

    @property
    def all_reco_mode_ids(self) -> Tuple[int, ...]:
        return tuple(v for vv in self.reco_mode_ids for v in vv)

    def as_dict(self) -> Dict[str, Any]:
        return asdict(self)

    def write_to_file(self, target_file_path: PathType) -> None:
        if os.path.exists(target_file_path):
            raise FileExistsError(f"Trying to overwrite {target_file_path}!")

        result_dict = self.as_dict()  # type: Dict[str, Any]

        tmp_target_path = os.path.join(os.path.dirname(target_file_path), f"tmp_{os.path.basename(target_file_path)}")
        with open(tmp_target_path, "w") as output_file:
            json.dump(result_dict, output_file, indent=2)
        os.rename(tmp_target_path, target_file_path)

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "RecoEffCorrInfoContainer":
        return RecoEffCorrInfoContainer(
            **{
                k: tuple([tuple(vv) if isinstance(vv, list) else vv for vv in v]) if isinstance(v, list) else v
                for k, v in input_dict.items()
            }
        )

    @staticmethod
    def load_from_file(file_path: PathType) -> "RecoEffCorrInfoContainer":
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"RecoEffCorrInfoContainer file {file_path} does not exist!")

        with open(file_path, "r") as in_file:
            input_dict = json.load(in_file)

        return RecoEffCorrInfoContainer.from_dict(input_dict=input_dict)


@dataclass(frozen=True)
class ScaleInfoItem:
    reco_mode_key: str
    reco_mode_ids: Tuple[int, ...]
    scale_factor: float

    def __post_init__(self) -> None:
        assert len(self.reco_mode_ids)


class ReconstructionModeEfficiencyCorrector:
    reco_eff_correction_weight_column_name: ClassVar[str] = "__reconstruction_efficiency_correction_weight__"

    def __init__(
        self,
        reco_efficiency_correction_info_container: RecoEffCorrInfoContainer,
        relative_rescaling_only: bool = True,
    ) -> None:
        self._correction_info_container = reco_efficiency_correction_info_container  # type: RecoEffCorrInfoContainer
        self._relative_rescaling_only = relative_rescaling_only  # type: bool

    @staticmethod
    def initialize_with_container_path(
        reco_efficiency_correction_info_container_path: PathType,
        relative_rescaling_only: bool = True,
    ) -> "ReconstructionModeEfficiencyCorrector":
        container = RecoEffCorrInfoContainer.load_from_file(file_path=reco_efficiency_correction_info_container_path)
        return ReconstructionModeEfficiencyCorrector(
            reco_efficiency_correction_info_container=container,
            relative_rescaling_only=relative_rescaling_only,
        )

    @property
    def reco_mode_id_col(self) -> str:
        return self._correction_info_container.reco_mode_id_column_name

    @property
    def all_reco_mode_ids(self) -> Tuple[int, ...]:
        return self._correction_info_container.all_reco_mode_ids

    @property
    def scale_infos(self) -> Generator[ScaleInfoItem, None, None]:
        for reco_mode_key, reco_mode_ids, scale_factor in zip(
            self._correction_info_container.reco_mode_keys,
            self._correction_info_container.reco_mode_ids,
            self._correction_info_container.correction_factors,
        ):
            yield ScaleInfoItem(
                reco_mode_key=reco_mode_key,
                reco_mode_ids=reco_mode_ids,
                scale_factor=scale_factor,
            )

    def add_correction_weight(
        self,
        df: pd.DataFrame,
        base_weight_column_name: str,
    ) -> None:
        assert self.reco_mode_id_col in df.columns, self.reco_mode_id_col
        assert base_weight_column_name in df.columns, base_weight_column_name
        assert base_weight_column_name == self._correction_info_container.base_weight_column_name, (
            base_weight_column_name,
            self._correction_info_container.base_weight_column_name,
        )

        new_weight_column_name = self.reco_eff_correction_weight_column_name  # type: str
        assert new_weight_column_name not in df.columns, new_weight_column_name

        df.loc[:, new_weight_column_name] = 1.0

        rescale_info = dict()  # type: Dict[str, float]
        for scale_info in self.scale_infos:
            reco_mode_cond = df[self.reco_mode_id_col].isin(scale_info.reco_mode_ids)  # type: pd.Series

            rescale_info.update({scale_info.reco_mode_key: df.loc[reco_mode_cond, base_weight_column_name].sum()})
            df.loc[reco_mode_cond, new_weight_column_name] *= scale_info.scale_factor

        if self._relative_rescaling_only:
            all_modes_cond = df[self.reco_mode_id_col].isin(self.all_reco_mode_ids)  # type: pd.Series
            original_sow = df.loc[all_modes_cond, base_weight_column_name].sum()  # type: float

            new_sow = sum([si.scale_factor * rescale_info[si.reco_mode_key] for si in self.scale_infos])

            inverted_norm_weight = original_sow / new_sow  # type: float
            df.loc[all_modes_cond, new_weight_column_name] *= inverted_norm_weight

    def add_weight_to_existing_weight_columns(
        self,
        df: pd.DataFrame,
        weight_columns: Sequence[str],
    ) -> None:
        assert len(weight_columns) > 0
        assert all(wc in df.columns for wc in weight_columns), [wc for wc in weight_columns if wc not in df.columns]
        assert self._relative_rescaling_only

        assert self._correction_info_container.base_weight_column_name in weight_columns, (
            self._correction_info_container.base_weight_column_name,
            weight_columns,
        )
        self.add_correction_weight(
            df=df,
            base_weight_column_name=self._correction_info_container.base_weight_column_name,
        )
        corr_weight_col_name = self.reco_eff_correction_weight_column_name  # type: str
        assert corr_weight_col_name in df.columns, corr_weight_col_name

        for weight_column in weight_columns:
            new_weight_values = df[weight_column].values * df[corr_weight_col_name].values  # type: np.ndarray
            df.loc[:, weight_column] = new_weight_values
