import numpy as np
import matplotlib.pyplot as plt
from typing import Dict, Tuple

from rdstar.utilities import PathType

from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetup

from templatefitter.plotter import plot_style, KITColors
from templatefitter.plotter.plot_utilities import export


__all__ = [
    "MVABFDictType",
    "MVAOptimizer",
]


MVABFDictType = Dict[float, Dict[str, Dict[str, float]]]


class MVAOptimizer:
    def __init__(
        self,
        mva_cut_infos: Tuple[float, ...],
        bf_dict: MVABFDictType,
        significance_dict: Dict[float, Dict[str, float]],
        fitting_setup: BFFittingSetup,
    ):
        self._mva_cut_infos = mva_cut_infos
        self._bf_dict = bf_dict
        self._significance_dict = significance_dict
        self._fitting_setup = fitting_setup

    def plot_branching_fractions(
        self,
        yield_parameter: str,
        filename: str,
        output_dir: PathType,
    ) -> None:
        plot_style.set_matplotlibrc_params()
        fig, ax = plt.subplots(figsize=(6.5, 5))

        bfs = np.array([self._bf_dict[mci][yield_parameter]["branching_fraction"] for mci in self._mva_cut_infos])
        lower_errors = np.array([self._bf_dict[mci][yield_parameter]["lower_bf_error"] for mci in self._mva_cut_infos])
        upper_errors = np.array([self._bf_dict[mci][yield_parameter]["upper_bf_error"] for mci in self._mva_cut_infos])

        ax.errorbar(
            x=np.array(self._mva_cut_infos),
            y=np.array(bfs) * 100,
            xerr=0,
            yerr=(np.array(lower_errors) * 100, np.array(upper_errors) * 100),
            ls="",
            marker=".",
            color="black",
            capsize=3,
            label=self._fitting_setup.fit_case.plot_info_text,
        )

        plt.legend(loc="best")
        plt.title(self._fitting_setup.fit_case.plot_labels[yield_parameter], loc="right")
        ax.set_ylabel(r"$\mathcal{B}(B\rightarrow D^{**} \ell \nu)$ in %", plot_style.ylabel_pos)
        export(fig=fig, filename=filename, target_dir=output_dir)

    def plot_branching_fraction_errors(
        self,
        yield_parameter: str,
        filename: str,
        output_dir: PathType,
    ) -> None:
        plot_style.set_matplotlibrc_params()
        fig, ax = plt.subplots(figsize=(6.5, 5))

        lower_errors = np.array([self._bf_dict[mci][yield_parameter]["lower_bf_error"] for mci in self._mva_cut_infos])
        upper_errors = np.array([self._bf_dict[mci][yield_parameter]["upper_bf_error"] for mci in self._mva_cut_infos])

        ax.plot(
            np.array(self._mva_cut_infos),
            lower_errors * 100,
            ls="",
            marker=".",
            color=KITColors.kit_blue,
            label="Lower Error",
        )
        ax.plot(
            np.array(self._mva_cut_infos),
            upper_errors * 100,
            ls="",
            marker=".",
            color=KITColors.kit_cyan,
            label="Upper Error",
        )

        plt.legend(loc="best")
        plt.title(self._fitting_setup.fit_case.plot_labels[yield_parameter], loc="right")
        ax.set_ylabel(r"$\sigma$ in %", plot_style.ylabel_pos)
        ax.set_xlabel(r"MVA Cut", plot_style.xlabel_pos)
        export(fig=fig, filename=filename, target_dir=output_dir)

    def plot_significance(
        self,
        yield_parameter: str,
        filename: str,
        output_dir: PathType,
    ) -> None:
        plot_style.set_matplotlibrc_params()
        fig, ax = plt.subplots(figsize=(6.5, 5))

        significance = np.array([self._significance_dict[mci][yield_parameter] for mci in self._mva_cut_infos])

        ax.plot(
            np.array(self._mva_cut_infos),
            significance,
            ls="",
            marker=".",
            color=KITColors.kit_blue,
        )

        plt.title(self._fitting_setup.fit_case.plot_labels[yield_parameter], loc="right")
        ax.set_ylabel(r"Significance", plot_style.ylabel_pos)
        ax.set_xlabel(r"MVA Cut", plot_style.xlabel_pos)
        export(fig=fig, filename=filename, target_dir=output_dir)
