from typing import Dict, List, Optional, Tuple

from rdstar.offline_analysis.dss_fitting.setup.dss_fit_cases import (
    FitCaseDict,
    reduced_dss_asimov_fit_case_dict_for_DPiZ,
    reduced_dss_asimov_fit_case_dict_for_DSPiZ,
    reduced_dss_asimov_fit_case_dict_for_DPiP,
    reduced_dss_asimov_fit_case_dict_for_DSPiP,
    alternative_dss_asimov_fit_case_dict_for_DPiZ,
    alternative_dss_asimov_fit_case_dict_for_DSPiZ,
    alternative_dss_asimov_fit_case_dict_for_DPiP,
    alternative_dss_asimov_fit_case_dict_for_DSPiP,
    dss_pStarL_reduced_asimov_fit_case_dict_for_DPiZ,
    dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiZ,
    dss_pStarL_reduced_asimov_fit_case_dict_for_DPiP,
    dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiP,
    dss_pStarL_alternative_asimov_fit_case_dict_for_DPiZ,
    dss_pStarL_alternative_asimov_fit_case_dict_for_DSPiZ,
    dss_pStarL_alternative_asimov_fit_case_dict_for_DPiP,
    dss_pStarL_alternative_asimov_fit_case_dict_for_DSPiP,
)

from rdstar.offline_analysis.dss_fitting.utilities.dss_ff_study_utils import (
    q_square_advanced_binning_for_fitting_var,
    q_square_advanced_binning_scopes,
    pStarL_advanced_binning_scopes,
    pStarL_advanced_binning_for_fitting_var,
)

from templatefitter.plotter.histogram_variable import HistVariable

__all__ = [
    "FittingSetup",
    "FittingSetupCollection",
]


class FittingSetup:
    def __init__(
        self,
        name: str,
        fit_case_dict: FitCaseDict,
        fit_in_bins_of: HistVariable,
        ratio_yield_parameters: List[str],
        fit_in_bins_of_scopes: Optional[Tuple[float, ...]] = None,
        fix_nui_params: Optional[bool] = True,
    ) -> None:
        if fit_in_bins_of_scopes:
            assert len(fit_in_bins_of_scopes) == fit_in_bins_of.n_bins + 1

        self._name = name
        self._fit_case_dict = fit_case_dict
        self._fit_in_bins_of = fit_in_bins_of
        self._ratio_yield_parameters = ratio_yield_parameters
        self._fit_in_bins_of_scopes = fit_in_bins_of_scopes
        self._fix_nui_params = fix_nui_params

    @property
    def name(self) -> str:
        return self._name

    @property
    def fit_case_dict(self) -> FitCaseDict:
        return self._fit_case_dict

    @property
    def fit_in_bins_of(self) -> HistVariable:
        return self._fit_in_bins_of

    @property
    def ratio_yield_parameters(self) -> List[str]:
        return self._ratio_yield_parameters

    @property
    def fit_in_bins_of_scopes(self) -> Tuple[float, ...]:
        if self._fit_in_bins_of_scopes:
            return self._fit_in_bins_of_scopes
        else:
            bins = self.fit_in_bins_of.n_bins
            increment = (self.fit_in_bins_of.scope[1] - self.fit_in_bins_of.scope[0]) / bins
            bin_edges = [self.fit_in_bins_of.scope[0] + i * increment for i in range(0, bins)]
            bin_edges.append(self.fit_in_bins_of.scope[1])
            assert len(bin_edges) == bins + 1
            return tuple(bin_edges)

    @property
    def fix_nui_params(self) -> bool:
        assert self._fix_nui_params is not None
        return self._fix_nui_params


# Asimov Fitting Setups
# q^2
q_square_reduced_asimov_fitting_setup_for_DPiZ = FittingSetup(
    name="q_square_reduced_asimov_fitting_setup_for_DPiZ",
    fit_case_dict=reduced_dss_asimov_fit_case_dict_for_DPiZ,
    fit_in_bins_of=q_square_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DPiZ"],
    fit_in_bins_of_scopes=q_square_advanced_binning_scopes,
)

q_square_reduced_asimov_fitting_setup_for_DSPiZ = FittingSetup(
    name="q_square_reduced_asimov_fitting_setup_for_DSPiZ",
    fit_case_dict=reduced_dss_asimov_fit_case_dict_for_DSPiZ,
    fit_in_bins_of=q_square_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DSPiZ"],
    fit_in_bins_of_scopes=q_square_advanced_binning_scopes,
)

q_square_reduced_asimov_fitting_setup_for_DPiP = FittingSetup(
    name="q_square_reduced_asimov_fitting_setup_for_DPiP",
    fit_case_dict=reduced_dss_asimov_fit_case_dict_for_DPiP,
    fit_in_bins_of=q_square_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DPiP"],
    fit_in_bins_of_scopes=q_square_advanced_binning_scopes,
)

q_square_reduced_asimov_fitting_setup_for_DSPiP = FittingSetup(
    name="q_square_reduced_asimov_fitting_setup_for_DSPiP",
    fit_case_dict=reduced_dss_asimov_fit_case_dict_for_DSPiP,
    fit_in_bins_of=q_square_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DSPiP"],
    fit_in_bins_of_scopes=q_square_advanced_binning_scopes,
)

q_square_alternative_asimov_fitting_setup_for_DPiZ = FittingSetup(
    name="q_square_alternative_asimov_fitting_setup_for_DPiZ",
    fit_case_dict=alternative_dss_asimov_fit_case_dict_for_DPiZ,
    fit_in_bins_of=q_square_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DPiZ"],
    fit_in_bins_of_scopes=q_square_advanced_binning_scopes,
)

q_square_alternative_asimov_fitting_setup_for_DSPiZ = FittingSetup(
    name="q_square_alternative_asimov_fitting_setup_for_DSPiZ",
    fit_case_dict=alternative_dss_asimov_fit_case_dict_for_DSPiZ,
    fit_in_bins_of=q_square_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DSPiZ"],
    fit_in_bins_of_scopes=q_square_advanced_binning_scopes,
)

q_square_alternative_asimov_fitting_setup_for_DPiP = FittingSetup(
    name="q_square_alternative_asimov_fitting_setup_for_DPiP",
    fit_case_dict=alternative_dss_asimov_fit_case_dict_for_DPiP,
    fit_in_bins_of=q_square_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DPiP"],
    fit_in_bins_of_scopes=q_square_advanced_binning_scopes,
)

q_square_alternative_asimov_fitting_setup_for_DSPiP = FittingSetup(
    name="q_square_alternative_asimov_fitting_setup_for_DSPiP",
    fit_case_dict=alternative_dss_asimov_fit_case_dict_for_DSPiP,
    fit_in_bins_of=q_square_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DSPiP"],
    fit_in_bins_of_scopes=q_square_advanced_binning_scopes,
)

# pStarL
pStarL_reduced_asimov_fitting_setup_for_DPiZ = FittingSetup(
    name="pStarL_reduced_asimov_fitting_setup_for_DPiZ",
    fit_case_dict=dss_pStarL_reduced_asimov_fit_case_dict_for_DPiZ,
    fit_in_bins_of=pStarL_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DPiZ"],
    fit_in_bins_of_scopes=pStarL_advanced_binning_scopes,
)

pStarL_reduced_asimov_fitting_setup_for_DSPiZ = FittingSetup(
    name="pStarL_reduced_asimov_fitting_setup_for_DSPiZ",
    fit_case_dict=dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiZ,
    fit_in_bins_of=pStarL_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DSPiZ"],
    fit_in_bins_of_scopes=pStarL_advanced_binning_scopes,
)

pStarL_reduced_asimov_fitting_setup_for_DPiP = FittingSetup(
    name="pStarL_reduced_asimov_fitting_setup_for_DPiP",
    fit_case_dict=dss_pStarL_reduced_asimov_fit_case_dict_for_DPiP,
    fit_in_bins_of=pStarL_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DPiP"],
    fit_in_bins_of_scopes=pStarL_advanced_binning_scopes,
)

pStarL_reduced_asimov_fitting_setup_for_DSPiP = FittingSetup(
    name="pStarL_reduced_asimov_fitting_setup_for_DSPiP",
    fit_case_dict=dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiP,
    fit_in_bins_of=pStarL_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DSPiP"],
    fit_in_bins_of_scopes=pStarL_advanced_binning_scopes,
)

pStarL_alternative_asimov_fitting_setup_for_DPiZ = FittingSetup(
    name="pStarL_alternative_asimov_fitting_setup_for_DPiZ",
    fit_case_dict=dss_pStarL_alternative_asimov_fit_case_dict_for_DPiZ,
    fit_in_bins_of=pStarL_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DPiZ"],
    fit_in_bins_of_scopes=pStarL_advanced_binning_scopes,
)

pStarL_alternative_asimov_fitting_setup_for_DSPiZ = FittingSetup(
    name="pStarL_alternative_asimov_fitting_setup_for_DSPiZ",
    fit_case_dict=dss_pStarL_alternative_asimov_fit_case_dict_for_DSPiZ,
    fit_in_bins_of=pStarL_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DSPiZ"],
    fit_in_bins_of_scopes=pStarL_advanced_binning_scopes,
)

pStarL_alternative_asimov_fitting_setup_for_DPiP = FittingSetup(
    name="pStarL_alternative_asimov_fitting_setup_for_DPiP",
    fit_case_dict=dss_pStarL_alternative_asimov_fit_case_dict_for_DPiP,
    fit_in_bins_of=pStarL_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DPiP"],
    fit_in_bins_of_scopes=pStarL_advanced_binning_scopes,
)

pStarL_alternative_asimov_fitting_setup_for_DSPiP = FittingSetup(
    name="pStarL_alternative_asimov_fitting_setup_for_DSPiP",
    fit_case_dict=dss_pStarL_alternative_asimov_fit_case_dict_for_DSPiP,
    fit_in_bins_of=pStarL_advanced_binning_for_fitting_var,
    ratio_yield_parameters=["DSPiP"],
    fit_in_bins_of_scopes=pStarL_advanced_binning_scopes,
)

# Data Fitting Setups
# todo: implement once we can look onto data


class FittingSetupCollection:
    available_setups = {
        q_square_reduced_asimov_fitting_setup_for_DPiZ.name: q_square_reduced_asimov_fitting_setup_for_DPiZ,
        q_square_reduced_asimov_fitting_setup_for_DSPiZ.name: q_square_reduced_asimov_fitting_setup_for_DSPiZ,
        q_square_reduced_asimov_fitting_setup_for_DPiP.name: q_square_reduced_asimov_fitting_setup_for_DPiP,
        q_square_reduced_asimov_fitting_setup_for_DSPiP.name: q_square_reduced_asimov_fitting_setup_for_DSPiP,
        q_square_alternative_asimov_fitting_setup_for_DPiZ.name: q_square_alternative_asimov_fitting_setup_for_DPiZ,
        q_square_alternative_asimov_fitting_setup_for_DSPiZ.name: q_square_alternative_asimov_fitting_setup_for_DSPiZ,
        q_square_alternative_asimov_fitting_setup_for_DPiP.name: q_square_alternative_asimov_fitting_setup_for_DPiP,
        q_square_alternative_asimov_fitting_setup_for_DSPiP.name: q_square_alternative_asimov_fitting_setup_for_DSPiP,
        pStarL_reduced_asimov_fitting_setup_for_DPiZ.name: pStarL_reduced_asimov_fitting_setup_for_DPiZ,
        pStarL_reduced_asimov_fitting_setup_for_DSPiZ.name: pStarL_reduced_asimov_fitting_setup_for_DSPiZ,
        pStarL_reduced_asimov_fitting_setup_for_DPiP.name: pStarL_reduced_asimov_fitting_setup_for_DPiP,
        pStarL_reduced_asimov_fitting_setup_for_DSPiP.name: pStarL_reduced_asimov_fitting_setup_for_DSPiP,
        pStarL_alternative_asimov_fitting_setup_for_DPiZ.name: pStarL_alternative_asimov_fitting_setup_for_DPiZ,
        pStarL_alternative_asimov_fitting_setup_for_DSPiZ.name: pStarL_alternative_asimov_fitting_setup_for_DSPiZ,
        pStarL_alternative_asimov_fitting_setup_for_DPiP.name: pStarL_alternative_asimov_fitting_setup_for_DPiP,
        pStarL_alternative_asimov_fitting_setup_for_DSPiP.name: pStarL_alternative_asimov_fitting_setup_for_DSPiP,
    }  # type: Dict[str, FittingSetup]
