import pandas as pd
import numpy as np

from typing import List, Tuple

from rdstar.mc_reweighting.BtoDststlnu.BtoDststlnu_weights import BtoDststFFReweighting as BtoDss_FF

from rdstar.offline_analysis.dss_fitting.utilities.utilities import dss_ff_systematics
from rdstar.offline_analysis.dss_fitting.setup.fitting_setup import FittingSetup
from rdstar.offline_analysis.systematics import SysColsInfo

from rdstar.offline_analysis.offline_observables import observables


class DSSFFSystematics:
    temp_dss_nom_weight_col = "temp_dss_nom_weight"
    temp_dss_var_weight_col = "temp_dss_var_weight"
    var_final_weight_column = "var_final_weight"

    def __init__(
        self,
        df: pd.DataFrame,
        fitting_setup: FittingSetup,
    ) -> None:
        assert self.temp_dss_var_weight_col not in df.columns
        assert self.temp_dss_nom_weight_col not in df.columns
        self._df = df
        self._fitting_setup = fitting_setup

        self.dim = self._fitting_setup.fit_in_bins_of.n_bins
        self._fit_case = self._fitting_setup.fit_case_dict.pivot_fit_case

    @staticmethod
    def get_up_var_cols(
        sys_col_info: SysColsInfo,
        var_no: int,
    ) -> str:
        up_cols = [col for col in observables if col.startswith(sys_col_info.var) and col.endswith(f"{var_no}_up")]
        assert len(up_cols) == 1, (len(up_cols), up_cols)
        return up_cols[0]

    @staticmethod
    def get_down_var_cols(
        sys_col_info: SysColsInfo,
        var_no: int,
    ) -> str:
        down_cols = [col for col in observables if col.startswith(sys_col_info.var) and col.endswith(f"{var_no}_up")]
        assert len(down_cols) == 1, (len(down_cols), down_cols)
        return down_cols[0]

    @staticmethod
    def get_dss_sys_infos() -> List[Tuple[Tuple[SysColsInfo, SysColsInfo], int]]:
        return [
            (
                (
                    SysColsInfo(
                        var=BtoDss_FF.get_var_dstst_weight_name(dstst_type="Done"),
                        ratio=BtoDss_FF.get_nominal_dstst_weight_name(dstst_type="Done"),
                        from_df=True,
                    ),
                    SysColsInfo(
                        var=BtoDss_FF.get_var_dstst_weight_name(dstst_type="Dtwostar"),
                        ratio=BtoDss_FF.get_nominal_dstst_weight_name(dstst_type="Dtwostar"),
                        from_df=True,
                    ),
                ),
                4,
            ),
            (
                (
                    SysColsInfo(
                        var=BtoDss_FF.get_var_dstst_weight_name(dstst_type="Dzerostar"),
                        ratio=BtoDss_FF.get_nominal_dstst_weight_name(dstst_type="Dzerostar"),
                        from_df=True,
                    ),
                    SysColsInfo(
                        var=BtoDss_FF.get_var_dstst_weight_name(dstst_type="Dprimeone"),
                        ratio=BtoDss_FF.get_nominal_dstst_weight_name(dstst_type="Dprimeone"),
                        from_df=True,
                    ),
                ),
                3,
            ),
        ]

    def split_df(self, df: pd.DataFrame) -> List[pd.DataFrame]:
        split_var = self._fitting_setup.fit_in_bins_of
        bins = split_var.n_bins
        increment = (split_var.scope[1] - split_var.scope[0]) / bins
        split_dfs = [
            df.query(
                f"{i * increment + split_var.scope[0]} <= {split_var.df_label} < {(i + 1) * increment + split_var.scope[0]}"
            )
            for i in range(0, bins)
        ]

        return split_dfs

    def calculate_cov_for_var(
        self,
        sys_col_infos: Tuple[SysColsInfo, SysColsInfo],
        var_no: int,
        up_down: str,
    ) -> np.ndarray:
        assert up_down in ["up", "down"], up_down
        var_cols = []  # type: List[str]
        for sys_col_info in sys_col_infos:
            if up_down == "up":
                var_cols.append(self.get_up_var_cols(sys_col_info=sys_col_info, var_no=var_no))
            elif up_down == "down":
                var_cols.append(self.get_down_var_cols(sys_col_info=sys_col_info, var_no=var_no))
        assert len(var_cols) == 2, (len(var_cols), var_cols)

        self._df.loc[:, self.var_final_weight_column] = self._df[self._fit_case.weight_column_name]

        assert all(isinstance(sys_col_info.ratio, str) for sys_col_info in sys_col_infos)
        nom_cols = [
            sys_col_info.ratio for sys_col_info in sys_col_infos if isinstance(sys_col_info.ratio, str)
        ]  # type: List[str]
        for col in nom_cols:
            self._df[self.var_final_weight_column] /= self._df[col]

        for var_col in var_cols:
            self._df[self.var_final_weight_column] *= self._df[var_col]

        split_dfs = self.split_df(df=self._df)
        assert len(split_dfs) == self.dim
        cov = np.zeros((self.dim, self.dim))

        for i in range(0, self.dim):
            for j in range(0, self.dim):
                var_i = np.sum(split_dfs[i][self.var_final_weight_column]) - np.sum(
                    split_dfs[i][self._fit_case.weight_column_name]
                )
                var_j = np.sum(split_dfs[j][self.var_final_weight_column]) - np.sum(
                    split_dfs[j][self._fit_case.weight_column_name]
                )
                cov[i][j] += var_i * var_j

        return cov

    def get_total_cov(self) -> np.ndarray:
        total_cov = np.zeros((self.dim, self.dim))
        for combination in self.get_dss_sys_infos():
            sys_infos = combination[0]
            number_of_vars = combination[1]

            total_cov_up = np.zeros((self.dim, self.dim))
            total_cov_down = np.zeros((self.dim, self.dim))

            for i in range(0, number_of_vars):
                cov_up = self.calculate_cov_for_var(sys_col_infos=sys_infos, up_down="up", var_no=i)
                cov_down = self.calculate_cov_for_var(sys_col_infos=sys_infos, up_down="down", var_no=i)

                total_cov_up += cov_up
                total_cov_down += cov_down

            mixed_cov = 0.5 * (total_cov_up + total_cov_down)
            total_cov += mixed_cov

        return total_cov


if __name__ == "__main__":

    for sys in dss_ff_systematics:
        print(DSSFFSystematics.get_up_var_cols(sys_col_info=sys, var_no=1))
