import os
import b2luigi as luigi

from luigi.util import inherits
from typing import Union, Optional, Dict, Generator

from templatefitter.minimizer import MinimizeResult

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.dss_fitting.bf_plotter import BFPlotter
from rdstar.offline_analysis.dss_fitting.bf_calculator import BFCalculator
from rdstar.offline_analysis.dss_fitting.tasks.bf_fitting_task import BFFittingTask
from rdstar.offline_analysis.dss_fitting.setup.dss_bf_fit_cases import BFFitCaseDict
from rdstar.offline_analysis.dss_fitting.tasks.preparation_task import FittingPreparationTask
from rdstar.offline_analysis.dss_fitting.setup.bf_fitting_setup import BFFittingSetupCollection
from rdstar.offline_analysis.efficiency_calculation.simple_efficiency_calculator import EfficiencyNumbersContainer


__all__ = [
    "BFCalculationTask",
]


@inherits(FittingPreparationTask)
class BFCalculationTask(utilities.BaseOfflineTask):
    fitting_setup_name = luigi.Parameter(positional=False)  # type: str
    plot_output = luigi.BoolParameter(positional=False)  # type: bool
    mva_cut_info = luigi.Parameter(positional=False)  # type: Optional[float]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if not self.mva_cut_info:
            self.base_dir = os.path.join(self.working_dir, "bf_results", self.fitting_setup_name)  # type: PathType
        else:
            self.base_dir = os.path.join(
                self.working_dir, "bf_results", self.fitting_setup_name, f"mva_cut_{self.mva_cut_info}"
            )

        self.fitting_setup = BFFittingSetupCollection.available_setups[self.fitting_setup_name]

        self.fit_case = BFFitCaseDict.get_fit_case(fit_case_name=self.fitting_setup.fit_case_name)
        self.fit_output_tag = self.fit_case.output_tag

        self.yield_parameters = self.fitting_setup.yield_parameters

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_efficiencies_dict(self) -> EfficiencyNumbersContainer:
        eff_path = self.input_dict[f"efficiency_{self.fit_case.name}"]
        efficiency_numbers_dict = EfficiencyNumbersContainer()
        efficiency_numbers_dict.from_file(file_path=eff_path)
        return efficiency_numbers_dict

    def do_plots(
        self,
        bf_calculator: Union[BFCalculator, None],
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]

        bf_output_key = f"{self.fitting_setup_name}_branching_fractions"
        bf_output_path = os.path.join(self.base_dir, f"{bf_output_key}.pdf")
        output_dict.update({bf_output_key: bf_output_path})

        yield_output_key = f"{self.fitting_setup_name}_yields"
        yield_output_path = os.path.join(self.base_dir, f"{yield_output_key}.pdf")
        output_dict.update({yield_output_key: yield_output_path})

        eff_output_key = f"{self.fitting_setup_name}_efficiencies"
        eff_output_path = os.path.join(self.base_dir, f"{eff_output_key}.pdf")
        output_dict.update({eff_output_key: eff_output_path})

        if dry_run:
            return output_dict
        else:
            assert bf_calculator is not None

        bf_plotter = BFPlotter(
            fitting_setup=self.fitting_setup,
            bf_calculator=bf_calculator,
        )
        bf_plotter.plot_branching_fractions(
            file_name=bf_output_key,
            output_dir=self.base_dir,
        )
        bf_plotter.plot_yields(
            file_name=yield_output_key,
            output_dir=self.base_dir,
        )
        bf_plotter.plot_efficiencies(
            file_name=eff_output_key,
            output_dir=self.base_dir,
        )

        return output_dict

    def run_bf_calculation(self, dry_run: bool = False) -> Dict[str, PathType]:

        output_dict = dict()  # type: Dict[str, PathType]

        if not self.mva_cut_info:
            yield_output_key = f"yields_{self.fitting_setup_name}"
            yield_tex_output_key = f"yields_tex_{self.fitting_setup_name}"
            efficiency_output_key = f"efficiencies_{self.fitting_setup_name}"
            eff_tex_output_key = f"efficiencies_tex_{self.fitting_setup_name}"
            bf_output_key = f"branching_fractions_{self.fitting_setup_name}"
        else:
            yield_output_key = f"yields_{self.fitting_setup_name}_mva_cut_{self.mva_cut_info}"
            yield_tex_output_key = f"yields_tex_{self.fitting_setup_name}_mva_cut_{self.mva_cut_info}"
            efficiency_output_key = f"efficiencies_{self.fitting_setup_name}_mva_cut_{self.mva_cut_info}"
            eff_tex_output_key = f"efficiencies_tex_{self.fitting_setup_name}_mva_cut_{self.mva_cut_info}"
            bf_output_key = f"branching_fractions_{self.fitting_setup_name}_mva_cut_{self.mva_cut_info}"

        yield_output_path = os.path.join(self.base_dir, f"{yield_output_key}.json")
        output_dict.update({yield_output_key: yield_output_path})

        yield_tex_output_path = os.path.join(self.base_dir, f"{yield_tex_output_key}.tex")
        output_dict.update({yield_tex_output_key: yield_tex_output_path})

        efficiency_output_path = os.path.join(self.base_dir, f"{efficiency_output_key}.json")
        output_dict.update({efficiency_output_key: efficiency_output_path})

        efficiency_tex_output_path = os.path.join(self.base_dir, f"{eff_tex_output_key}.tex")
        output_dict.update({eff_tex_output_key: efficiency_tex_output_path})

        branching_fractions_output_path = os.path.join(self.base_dir, f"{bf_output_key}.json")
        output_dict.update({bf_output_key: branching_fractions_output_path})

        if dry_run:
            bf_calculator = None
        else:
            efficiencies_dict = self.get_efficiencies_dict()
            if not self.mva_cut_info:
                key = f"{self.fit_output_tag}_FitResult_{self.fit_case.name}"
            else:
                key = f"{self.fit_output_tag}_FitResult_{self.fit_case.name}_mva_cut_{self.mva_cut_info}"
            fit_result_path = self.input_dict[key]
            fit_result = MinimizeResult.load_from(path=fit_result_path)

            bf_calculator = BFCalculator(
                fitting_setup=self.fitting_setup,
                fit_result=fit_result,
                efficiency_dict=efficiencies_dict,
            )
            bf_calculator.calculate()
            bf_calculator.save_yields_to_file(file_path=yield_output_path)
            bf_calculator.save_yields_to_latex(
                file_path=yield_tex_output_path,
                output_order=tuple(self.fitting_setup.yield_parameters + [self.fitting_setup.norm_parameter]),
            )
            bf_calculator.save_efficiencies_to_file(file_path=efficiency_output_path)
            bf_calculator.save_efficiencies_to_latex(
                file_path=efficiency_tex_output_path,
                output_order=tuple(self.fitting_setup.yield_parameters + [self.fitting_setup.norm_parameter]),
            )
            bf_calculator.save_branching_fractions_to_file(file_path=branching_fractions_output_path)

        plot_output_dict = self.do_plots(
            bf_calculator=bf_calculator,
            dry_run=dry_run,
        )
        output_dict.update(plot_output_dict)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_bf_calculation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_bf_calculation(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(
            BFFittingTask,
            fit_case_name=self.fitting_setup.fit_case_name,
            plot_output=self.plot_output,
            mva_cut_info=self.mva_cut_info,
        )
