import os
import json
import pandas as pd
import b2luigi as luigi
import matplotlib.pyplot as plt

from luigi.util import inherits
from typing import Optional, Dict, Generator, Tuple

from templatefitter.fitter import TemplateFitter
from templatefitter.plotter.fit_result_plots import FitResultPlotter

from rdstar.utilities.logging import log_to_default
from rdstar.utilities import PathType, merge_plot_pdf_files

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.fitting.fit_model_creator import FitModelCreator
from rdstar.offline_analysis.dss_fitting.tasks.splitter_task import SampleSplitterTask
from rdstar.offline_analysis.dss_fitting.setup.fitting_setup import FittingSetupCollection


__all__ = [
    "FittingTask",
]


logger = log_to_default(name=__name__)


@inherits(SampleSplitterTask)
class FittingTask(utilities.BaseOfflineTask):
    bin_number = luigi.IntParameter(positional=False)  # type: int

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_dir = os.path.join(self.working_dir, "ff_fit_results", self.fitting_setup_name)  # type: PathType

        self.fitting_setup = FittingSetupCollection.available_setups[self.fitting_setup_name]

        self.fit_case = self.fitting_setup.fit_case_dict.get_fit_case_in_bin(bin_number=self.bin_number)
        self.fit_output_tag = self.fit_case.output_tag  # type: str

        self.mc_key = "combined_mc"  # type: str
        self.data_key = "data"  # type: str

        self.yield_parameters = self.fitting_setup.ratio_yield_parameters

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_data_path(self, data_key: str) -> PathType:
        input_key = f"splitted_{data_key}_{self.selection}_{self.fitting_setup_name}_bin_{self.bin_number}"
        return self.input_dict[input_key]

    def get_fit_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        key = f"{self.fit_output_tag}_{file_tag}_{self.fitting_setup_name}_bin_{self.bin_number}"
        return key, f"{key}.json"

    def get_fit_plot_output_key_file_name_pair(self, file_tag: str) -> Tuple[str, str]:
        key = f"{self.fit_output_tag}_{file_tag}_{self.fitting_setup_name}_bin_{self.bin_number}"
        return key, f"{key}.pdf"

    def initialize_model_creator(self) -> FitModelCreator:
        FitModelCreator.fit_cases = {self.fit_case.name: self.fit_case}

        model_creator = FitModelCreator(
            fit_case=self.fit_case.name,
        )

        return model_creator

    def run_fit(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]
        fit_result_key, fit_result_file_name = self.get_fit_output_key_file_name_pair(file_tag="FitResult")
        temp_fit_result_path = os.path.join(self.base_dir, f"temp_{fit_result_file_name}")
        fit_result_path = os.path.join(self.base_dir, fit_result_file_name)
        output_dict.update({fit_result_key: fit_result_path})

        pre_fit_plot_key, pre_fit_plot_file_name = self.get_fit_plot_output_key_file_name_pair(file_tag="PreFitPlot")
        temp_pre_fit_plot_path = os.path.join(self.base_dir, f"temp_{pre_fit_plot_file_name}")
        pre_fit_plot_path = os.path.join(self.base_dir, pre_fit_plot_file_name)
        output_dict.update({pre_fit_plot_key: pre_fit_plot_path})

        post_fit_plot_key, post_fit_plot_file_name = self.get_fit_plot_output_key_file_name_pair(file_tag="PostFitPlot")
        temp_post_fit_plot_path = os.path.join(self.base_dir, f"temp_{post_fit_plot_file_name}")
        post_fit_plot_path = os.path.join(self.base_dir, post_fit_plot_file_name)
        output_dict.update({post_fit_plot_key: post_fit_plot_path})

        significance_key = f"significance_{self.fitting_setup_name}_bin_{self.bin_number}"
        significance_path = os.path.join(self.base_dir, f"{significance_key}.json")
        output_dict.update({significance_key: significance_path})

        if not dry_run:
            df_mc = pd.read_hdf(path_or_buf=self.get_data_path(data_key=self.mc_key))
            assert isinstance(df_mc, pd.DataFrame)

            df_data = (
                pd.read_hdf(path_or_buf=self.get_data_path(data_key=self.data_key))
                if not self.fit_case.asimov_data
                else None
            )

            logger.debug("Setting up Fit Model...")
            model_creator = self.initialize_model_creator()
            model_creator.setup_model(
                mc_df=df_mc,
                data_df=df_data,
                use_asimov_data=self.fit_case.asimov_data,
                round_asimov_data_bin_counts=self.fit_case.round_asimov_data,
            )
            fit_model = model_creator.fit_model
            logger.debug("Fit Model set up!")

            plt.switch_backend("Agg")
            fit_result_plotter = FitResultPlotter(
                variables_by_channel=model_creator.hist_variables_by_mode_dict,
                fit_model=fit_model,
                reference_dimension=0,
                channel_label_dict=self.fit_case.reco_mode_labels_dict,
                data_label=self.fit_case.plot_info_text,
            )

            pre_fit_plot_paths = fit_result_plotter.plot_fit_result(
                use_initial_values=True,
                output_dir_path=self.base_dir,
                output_name_tag=f"{self.fit_output_tag}_{self.fitting_setup_name}_{self.bin_number}",
            )
            assert len(pre_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["png"]), (
                len(pre_fit_plot_paths["pdf"]),
                len(pre_fit_plot_paths["png"]),
            )
            merge_plot_pdf_files(input_paths=pre_fit_plot_paths["pdf"], output_path=temp_pre_fit_plot_path)
            os.rename(src=temp_pre_fit_plot_path, dst=pre_fit_plot_path)
            logger.debug("Pre-Fit Plots performed!")

            fitter = TemplateFitter(fit_model=fit_model, minimizer_id="iminuit")
            fit_result = fitter.do_fit(
                update_templates=True,
                get_hesse=True,
                verbose=True,
                fix_nui_params=self.fitting_setup.fix_nui_params,
            )
            fit_result.save_to(path=temp_fit_result_path, overwrite=True)
            os.rename(src=temp_fit_result_path, dst=fit_result_path)
            logger.debug("Fit performed!")

            post_fit_plot_paths = fit_result_plotter.plot_fit_result(
                output_dir_path=self.base_dir,
                output_name_tag=f"{self.fit_output_tag}_{self.fitting_setup_name}_{self.bin_number}",
            )
            assert len(post_fit_plot_paths["pdf"]) == len(post_fit_plot_paths["png"]), (
                len(post_fit_plot_paths["pdf"]),
                len(post_fit_plot_paths["png"]),
            )
            assert len(post_fit_plot_paths["pdf"]) == len(pre_fit_plot_paths["pdf"]), (
                len(post_fit_plot_paths["pdf"]),
                len(pre_fit_plot_paths["pdf"]),
            )
            merge_plot_pdf_files(input_paths=post_fit_plot_paths["pdf"], output_path=temp_post_fit_plot_path)
            os.rename(src=temp_post_fit_plot_path, dst=post_fit_plot_path)
            logger.debug("Post-Fit Plots performed!")

            logger.debug("Calculate Significance...")
            significance_dict = dict()  # type: Dict[str, float]
            for yield_parameter in self.yield_parameters:
                significance = fitter.get_significance(yield_parameter=f"yield_{yield_parameter}", fix_nui_params=True)
                significance_dict.update({yield_parameter: significance})

            tmp_significance_path = os.path.join(self.base_dir, f"tmp_{significance_key}.json")
            with open(tmp_significance_path, "w") as fp:
                json.dump(significance_dict, fp, indent=2)
            os.rename(src=tmp_significance_path, dst=significance_path)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_fit()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.run_fit(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(SampleSplitterTask)
