"""
This file contains functions used for the plotting of the B->D**lv signal extraction
analysis

Max Welsch 2019
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.stats import norm

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.plot_style import KITColors, xlabel_pos, ylabel_pos

from rdstar.offline_analysis.dstst_analysis.template_definitions import channel_to_title, template_label, \
    template_colors

import maxtemplatefitter as tf


def plot_all_channels(mct, export_path=None, tag=None, yscale=1.3,
                      left_title=None, plot_text=None,
                      xlabel=r"$\mathrm{M}^2_\mathrm{miss}$ in $\mathrm{GeV}^2$",
                      lumi_text=(0.6, 0.5)):
    for channel in mct.channels.values():
        fig, axis = plt.subplots(1, 1, figsize=(4, 4), dpi=300)
        bin_width = channel.plot_stacked_on(axis)
        axis.legend(frameon=False)
        axis.set_title(channel_to_title[channel.name], loc="right")
        axis.set_ylabel(r"Events / (%.2f $\mathrm{GeV}^2$)" % bin_width, ylabel_pos)
        axis.ticklabel_format(axis="y", style="sci", scilimits=(-3, 3))
        axis.set_xlabel(xlabel, xlabel_pos)
        axis.set_ylim([0, axis.get_ylim()[1] * yscale])
        # axis.yaxis.get_offset_text().set_x(-0.12)
        h, labels = axis.get_legend_handles_labels()
        axis.legend(h, get_latex_labels(labels), frameon=False, ncol=2)

        if left_title:
            axis.set_title(left_title, loc="left")

        if lumi_text:
            axis.text(*lumi_text, r"$\int \mathcal{L} \,dt=711\,\mathrm{fb}^{-1}$", transform=axis.transAxes)

        if plot_text:
            axis.text(*plot_text, transform=axis.transAxes, fontweight="bold")

        if export_path:
            if tag:
                fig_name = f"template_{channel.name}_{tag}"
            else:
                fig_name = f"template_{channel.name}"
            export(fig, fig_name, export_path)


def plot_templates_per_channel(mct, export_path=None, tag=None):
    for ch_name, channel in mct.channels.items():
        for name, template in channel.templates.items():
            fig, axis = plt.subplots(1, 1, figsize=(3.5, 3.5), dpi=300)
            template.plot_on(axis)
            axis.legend(frameon=False)
            axis.set_title(channel_to_title[channel.name], loc="right")
            axis.set_ylabel(r"Events / (%.2f $\mathrm{GeV}^2$)" % template.bin_widths[0], **ylabel_pos)
            axis.ticklabel_format(axis="y", style="sci", scilimits=(-3, 3))
            axis.set_xlabel(r"$\mathrm{M}^2_\mathrm{miss}$ in $\mathrm{GeV}^2$", **xlabel_pos)
            axis.set_ylim([0, axis.get_ylim()[1] * 1.2])

            h, labels = axis.get_legend_handles_labels()
            axis.legend(h, get_latex_labels(labels), frameon=False)
            plt.tight_layout()
            if export_path:
                if tag:
                    fig_name = f"template_{ch_name}_{name}_{tag}"
                else:
                    fig_name = f"template_{ch_name}_{name}"
                export(fig, fig_name, export_path)


def plot_bin_nui_params(res, channel_name, template_names, num_bins,
                        legend=True, export_path=None, tag=None, left_title=None):
    nui_params = {}
    for template_name in template_names:
        nui_param_names = [f"{channel_name}_{template_name}_bin_{i}" for i in range(num_bins)]
        nui_params_vals = [res.params[name][0] for name in nui_param_names]
        nui_params_errs = [res.params[name][1] for name in nui_param_names]

        nui_params[template_name] = {"vals": np.array(nui_params_vals), "errs": np.array(nui_params_errs)}

    fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)
    ax.axvline(0, color=KITColors.kit_black)
    for i, items in enumerate(nui_params.items()):
        name, val = items[0], items[1]
        stepsize = 1
        ax.errorbar(x=val["vals"], y=np.arange(i * num_bins * stepsize, (i + 1) * num_bins * stepsize, stepsize),
                    xerr=val["errs"], ls="", marker=".",
                    label=template_label[name], color=template_colors[name])
    ax.set_xlabel(r"$(\hat{\theta} - \theta_0)/\sigma_{\theta_0}$", **xlabel_pos)
    ax.set_ylabel("Bins", **ylabel_pos)
    ax.set_yticklabels([])
    ax.set_title(channel_to_title[channel_name], loc="right")

    if legend:
        # ax.set_ylim(ax.get_ylim()[0], ax.get_ylim()[1]*1.2)
        ax.legend(frameon=True, loc="best", fontsize="small")

    if left_title:
        ax.set_title(left_title, loc="left")

    # plt.tight_layout()
    if export_path:
        if tag:
            fig_name = f"nui_params_{channel_name}_{tag}"
        else:
            fig_name = f"nui_params_{channel_name}_{name}"
        export(fig, fig_name, export_path)


def plot_nll_profile(points, values, hesse, stat_points, stat_values, fit_res, param_id, process_label,
                     export_path=None, tag=None, left_title=None):
    fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)

    sig_yield = fit_res.params[param_id]

    ax.plot(stat_points, stat_values, lw=2, label="stat. only", color=KITColors.kit_green)
    ax.plot(points, values, lw=2, label=r"stat. + sys.")
    ax.plot(points, hesse, lw=2, ls="--", label="Hesse Approximation", color=KITColors.kit_orange)

    ax.set_ylabel(r"$-\log(\mathrm{L}(\nu, \hat{\hat{\theta}}))$", ylabel_pos)
    ax.set_xlabel(r"$\nu$", xlabel_pos)
    ax.set_title(process_label, loc="right")

    ax.set_ylim([0, ax.get_ylim()[1] * 1.1])
    ax.errorbar(sig_yield[0], y=0.5, xerr=sig_yield[1], ls="", marker=".", capsize=1.5, color=KITColors.kit_red,
                zorder=10, label="Fit Result")

    ax.axhline(0.5, ls="--", color=KITColors.grey)
    ax.axhline(2, ls="--", color=KITColors.grey)
    ax.text(.93, 0.23, r"$1\sigma$", color=KITColors.grey, transform=ax.transAxes)
    ax.text(.93, 0.87, r"$2\sigma$", color=KITColors.grey, transform=ax.transAxes)

    ax.legend(frameon=True, loc="best")

    if left_title:
        ax.set_title(left_title, loc="left")

    if export_path:
        if tag:
            fig_name = f"nll_profile_{param_id}_{tag}"
        else:
            fig_name = f"nll_profile_{param_id}"
        export(fig, fig_name, export_path)


def lin1d(x, m, c):
    return x * m + c


def plot_linearity_test_result(x, y, yerr, process, process_name, ntoys=200, export_path=None, tag=None):
    fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)
    ax.errorbar(x, y, yerr=yerr, ls="", marker=".", label=f"Mean of {ntoys} Toys")
    popt, pcov = curve_fit(lin1d, x, y, p0=(1, 0), sigma=yerr)
    ax.plot(x, lin1d(x, *popt), label=r"$m\cdot x+c$")
    ax.legend(frameon=True)

    ax.text(0.05, 0.75, r"$m={:.2f}\pm{:.2f}$".format(popt[0], np.sqrt(pcov[0, 0])),
            transform=ax.transAxes)
    ax.text(0.05, 0.7, r"$c={:.2f}\pm{:.2f}$".format(popt[1], np.sqrt(pcov[1, 1])),
            transform=ax.transAxes)
    ax.set_ylabel(r"Extracted $\nu_{%s}$ Yield" % process, ylabel_pos)
    ax.set_xlabel(r"Input $\nu_{%s}$ Yield" % process, xlabel_pos)
    ax.ticklabel_format(axis="x", style="sci", scilimits=(-5, 5))

    if export_path:
        if tag:
            fig_name = f"linearity_test_{process_name}_{tag}"
        else:
            fig_name = f"linearity_test_{process_name}"
        export(fig, fig_name, export_path)


def plot_background_linearity_test_result(x, y, yerr, sig_process, sig_process_name, bkg_process, bkg_process_name,
                                          ntoys=200, export_path=None, tag=None):
    fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)
    ax.errorbar(x, y, yerr=yerr, ls="", marker=".", label=f"Mean of {ntoys} Toys")
    popt, pcov = curve_fit(lin1d, x, y, p0=(1, 0), sigma=yerr)
    ax.plot(x, lin1d(x, *popt), label=r"$m\cdot x+c$")
    ax.legend(frameon=True)

    yerr_max = np.amax(yerr)
    ax.set_ylim([popt[1] - 1.5 * yerr_max, popt[1] + 1.5 * yerr_max])

    ax.text(0.05, 0.1, r"$m={:.2f}\pm{:.2f}$".format(popt[0], np.sqrt(pcov[0, 0])),
            transform=ax.transAxes)
    ax.text(0.05, 0.05, r"$c={:.2f}\pm{:.2f}$".format(popt[1], np.sqrt(pcov[1, 1])),
            transform=ax.transAxes)
    ax.set_ylabel(r"Extracted $\nu_{%s}$ Yield" % sig_process, ylabel_pos)
    ax.set_xlabel(r"Input $\nu_{%s}$ Yield" % bkg_process, xlabel_pos)
    ax.ticklabel_format(axis="x", style="sci", scilimits=(-5, 5))

    if export_path:
        if tag:
            fig_name = f"linearity_test_{sig_process_name}_{bkg_process_name}_{tag}"
        else:
            fig_name = f"linearity_test_{sig_process_name}_{bkg_process_name}"
        export(fig, fig_name, export_path)


def plot_toy_pulls(ts, process_label, param, ntoys=None, export_path=None, tag=None):
    pulls = ts.get_toy_result_pulls(param)

    fig, ax = plt.subplots(1, 1, figsize=(4, 4), dpi=200)
    binc, bine, _ = ax.hist(pulls, bins=25, density=True,
                            edgecolor=KITColors.kit_black, histtype="stepfilled", label="Toy Pulls")
    binm = tf.histograms.bin_mids(bine)

    def gauss_helper(x, loc, scale):
        return norm.pdf(x, loc, scale)

    popt, pcov = curve_fit(gauss_helper, binm, binc)

    x_values = np.linspace(np.amin(pulls), np.amax(pulls), 100)
    ax.plot(x_values, norm.pdf(x_values, popt[0], popt[1]), label=r"$\mathcal{N}(\mu, \sigma^2)$")

    if ntoys:
        ax.legend(frameon=True, title=r"$N_\mathrm{toys}=%d$" % ntoys)
    else:
        ax.legend(frameon=True)
    ax.set_ylabel("Normalized in arb. units", ylabel_pos)
    ax.set_xlabel(r"Pull $\nu_{%s}$" % process_label, xlabel_pos)
    ax.set_ylim([0, 1.2 * ax.get_ylim()[1]])

    median = np.median(pulls)
    ax.text(0.05, 0.9, r"$\mu={:.2f}\pm{:.2f}$".format(popt[0], np.sqrt(pcov[0, 0])),
            transform=ax.transAxes)
    ax.text(0.05, 0.85, r"$\sigma={:.2f}\pm{:.2f}$".format(popt[1], np.sqrt(pcov[1, 1])),
            transform=ax.transAxes)
    ax.text(0.05, 0.8, r"$\tilde{\nu}=%.2f$" % median,
            transform=ax.transAxes)

    if export_path:
        if tag:
            fig_name = f"toy_pulls_{param}_{tag}"
        else:
            fig_name = f"toy_pulls_{param}"
        export(fig, fig_name, export_path)


def get_latex_labels(l):
    latex_labels = []
    other = {
        "Data": "Data",
        "Uncertainty": "Uncertainty"
    }
    for label in l:

        if label in template_label.keys():
            latex_labels.append(template_label[label])
        elif label in other.keys():
            latex_labels.append(other[label])
        else:
            latex_labels.append("")
    return latex_labels
