"""
This file contains helper functions used in the B->D**lv signal extraction analysis.
"""
import os
import json
import itertools

import numpy as np
import pandas as pd

import maxtemplatefitter as tf

from rdstar.utilities.logging import log_to_default
from rdstar.offline_analysis.dstst_analysis.template_definitions import *


logger = log_to_default(name=__name__)


def load_dataframes(source_dir="/storage/9/fmetzner/rdstar/data_set_prod_18thApr2019_stream0/CombinedSamples"):
    """
    Loads final dataframes for all samples.

    :param source_dir: Path to the source directory of the analysis dataframes
    :return: Dictionary of pd.DataFrames
    """

    dataframes = {}
    dataframes["dss_df"] = pd.read_hdf(os.path.join(source_dir, "Dss_first_selection_with_CS_CS1.h5"))
    dataframes["gap_df"] = pd.read_hdf(os.path.join(source_dir, "Gap_first_selection_with_CS_CS1.h5"))
    dataframes["mixed_df"] = pd.read_hdf(os.path.join(source_dir, "Charged_first_selection_with_CS_CS1.h5"))
    dataframes["charged_df"] = pd.read_hdf(os.path.join(source_dir, "Mixed_first_selection_with_CS_CS1.h5"))
    dataframes["charm_df"] = pd.read_hdf(os.path.join(source_dir, "Charm_first_selection_with_CS_CS1.h5"))
    dataframes["uds_df"] = pd.read_hdf(os.path.join(source_dir, "UDS_first_selection_with_CS_CS1.h5"))
    # dataframes["psi_df"] = pd.read_hdf(os.path.join(source_dir, "SpecialPsi_first_selection_with_CS_CS1.h5"))
    # dataframes["special_df"] = pd.read_hdf(os.path.join(source_dir, "Special_first_selection_with_CS_CS1.h5"))

    return dataframes


def combine_dataframe_dict(df_dict):
    """
    Combines a dictionary of pd.DataFrames into a common pd.DataFrame.
    """
    return pd.concat(list(df_dict.values()), sort=True)


def split_dataframe_into_modes(df, mode_dict, mode_var):
    """
    Splits a dataframe into different groups.
    """
    df_per_mode = {}
    for mode_name, mode_ids in mode_dict.items():
        df_per_mode[mode_name] = df.loc[df[mode_var].isin(mode_ids)]

    return df_per_mode


def prepare_template_dataframes(
        df,
        channel_mode_dict,
        template_id_dict,
        reco_mode_var="daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc",
        sig_id_var="SIG_ID",
        weight_column="__scaling_weight__"
):
    """
    Prepares a dict of dict of pd.DataFrames which hold the split data set.
    The keys of the first and the second dict denote the decay channels and
    the template names, respectively.

    :param df: Total data set as pd.DataFrame
    :param channel_mode_dict: Dictionary of str: list(int). The list
    holds the decay mode ids which will be combined into one channel
    :param template_id_dict: Dictionary of str: list(int). The list
    holds the SIG_IDs which are combined into one template.
    :param reco_mode_var: A string denoting the variable used to identify
    the decay mode of the B meson.
    :param sig_id_var: A string denoting the variables used to identify
    the signal id of the B meson
    :param weight_column: Name of the column that holds the event weight.
    :type weight_column: str
    :return: dict of dict of pd.DataFrames
    """
    df = df.loc[df[reco_mode_var].isin(list(itertools.chain.from_iterable(channel_mode_dict.values()))), :]
    template_dfs = split_dataframe_into_modes(df, template_id_dict, sig_id_var)
    per_mode_templates = dict()
    per_mode_template_effs = dict()

    for temp_name, template_df in template_dfs.items():
        per_mode_templates[temp_name] = split_dataframe_into_modes(template_df, channel_mode_dict, reco_mode_var)

        # get efficiencies
        per_mode_template_effs[temp_name] = dict()
        for mode_name, template_per_mode_df in per_mode_templates[temp_name].items():
            per_mode_template_effs[temp_name][mode_name] = np.sum(template_per_mode_df[weight_column].values) / np.sum(
                template_df[weight_column].values)

    return per_mode_templates, per_mode_template_effs


def prepare_mct(total_df, signal_var="m2RecoilSignalSide", weight="__scaling_weight__",
                definitions="missingmass_signal", add_systematics=True, export_path=None):
    """
    Prepares the tf.MultiChannelTemplate used for the fit.

    :param total_df: DataSet used to build the templates
    :type total_df: pd.DataFrame
    :param signal_var: Name of the column that holds the signal variable.
    :type signal_var: str
    :param weight: Name of the column that holds the event weight.
    :type weight: str
    :return: A prepared tf.MultiChannelTemplate.
    """
    logger.info("Preparing MultiChannelTemplate...")
    per_mode_templates, per_mode_template_effs = prepare_template_dataframes(total_df, channel_dec_modes, template_ids)
    logger.info("Process per decay mode efficiency:")
    logger.info(per_mode_template_effs)

    if export_path:
        with open(os.path.join(export_path, "template_effs.json"), "w") as outfile:
            json.dump(per_mode_template_effs, outfile, indent=2)

    mct = tf.templates.MultiChannelTemplate()

    channel_def = channel_defs[definitions]

    for name, vals in channel_def.items():
        mct.define_channel(name, *vals)

    for process in per_mode_templates.keys():
        mct.define_process(process)

    hist_dict = dict()
    template_dict = dict()

    for name, df_mode_dict in per_mode_templates.items():
        hist_dict[name] = dict()
        template_dict[name] = dict()
        for mode, df in df_mode_dict.items():
            hist = tf.histograms.Hist1d(
                bins=channel_def[mode][0],
                range=channel_def[mode][1],
                data=df[signal_var],
                weights=df[weight]
            )
            hist_dict[name][mode] = hist

            template = tf.templates.Template1d(name, "missingMass", hist, color=template_colors[name])

            if add_systematics:
                for uncertainty in sys_uncertainties[name]:
                    template.add_variation(df[signal_var], df[uncertainty[1]], df[uncertainty[0]])

            template_dict[name][mode] = template

            mct.add_template(mode, name, template, per_mode_template_effs[name][mode])

    return mct, hist_dict, template_dict


def get_mode(df, mode_ids):
    return df.loc[df.daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc.isin(mode_ids), :]


def prepare_data_hists(data, signal_var="m2RecoilSignalSide", definitions="missingmass_signal"):
    data_per_mode = split_dataframe_into_modes(
        data, channel_dec_modes, "daughter__bo1__cm__spextraInfo__bodecayModeID__bc__bc"
    )

    channel_def = channel_defs[definitions]

    hdata_per_mode = {}
    for mode, df in data_per_mode.items():
        hdata_per_mode[mode] = tf.Hist1d(*channel_def[mode], df[signal_var])

    return hdata_per_mode


def prepare_weights(total_df):
    all_weights = [
        "__scaling_weight__",
        'weight_ff_Done_nom',
        'weight_ff_Dprimeone_nom',
        'weight_ff_Dtwostar_nom',
        'weight_ff_Dzerostar_nom',
        'weight_ff_btod_nom',
        'weight_ff_btodst_nom',
        '__weight_pid_corr_dd_0_ratio__',
        '__weight_pid_corr_dd_1_ratio__',
        '__weight_pid_corr_dd_2_ratio__',
        '__weight_pid_corr_dd_3_ratio__',
        '__weight_pid_corr_dd_4_ratio__',
        '__weight_pid_corr_dst_pi_ratio__',
        '__weight_pid_corr_lepton_ratio__',
    ]

    set_negative_pid_weights_to_one(total_df, pid_weights, pid_errors_stat, pid_errors_sys)

    for nominal_value, error_stat, error_sys in zip(pid_weights, pid_errors_stat, pid_errors_sys):
        add_pid_variations(total_df, nominal_value, error_stat, error_sys, "var_up", "var_down")
        prepare_varied_weights(
            total_df, all_weights, nominal_value, [nominal_value + "var_up", nominal_value + "var_down"]
        )

    for ff_weight_nom, ff_weight_var in ff_weight_variations.items():
        prepare_varied_weights(total_df, all_weights, ff_weight_nom, ff_weight_var)

    total_df["weight"] = total_df[all_weights].product(axis=1)

    return total_df


def set_negative_pid_weights_to_one(total_df, pid_weight_list, pid_errors_stat_list, pid_errors_sys_list):
    """
    At the moment it might happen that some pid weight
    values are set to negative values. This functions sets
    the values to the intended value of 1.0.
    """
    for pid_weight in pid_weight_list:
        total_df.loc[total_df[pid_weight] < 0., pid_weight] = 1.0

    for pid_error in pid_errors_stat_list:
        total_df.loc[total_df[pid_error] < 0., pid_error] = 0.05

    for pid_error in pid_errors_sys_list:
        total_df.loc[total_df[pid_error] < 0., pid_error] = 0.05


def add_pid_variations(total_df, pid_nom, error_stat, error_sys, up_tag="var_up", down_tag="var_down"):
    """
    Calculates the variations for the pid weights given
    the nominal value and its uncertainty
    """
    assert np.all(total_df[error_sys].values >= 0)
    assert np.all(total_df[error_stat].values >= 0)
    total_error = np.sqrt(total_df[error_stat] ** 2 + total_df[error_sys] ** 2)
    total_df[pid_nom + "total_error"] = total_error
    total_df[pid_nom + up_tag] = total_df[pid_nom] + total_error / 2
    total_df[pid_nom + down_tag] = total_df[pid_nom] - total_error / 2


def prepare_varied_weights(total_df, all_weights, nom_weight, varied_weights):
    """
    Replaces the nominal value of a weight in the product of all weights
    with its varied value and stores the varied value in the same column
    in the pd.DataFrame as before.
    """
    all_weights_copy = all_weights.copy()
    all_weights_copy.remove(nom_weight)

    for var in varied_weights:
        total_df[var] = total_df[all_weights_copy + [var]].product(axis=1)
