"""
Collection of continuum calibration fitters.

Felix Metzner 2023
"""

from typing import Tuple, Dict, Type

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.comparison_tools.select_utilities import SimpleCutInfo
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.offline_analysis.histogram_variables import RDStarHistVars

from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters.cont_cal_fitter_base import ContCalFitterBase

from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters import fitter_cont_cal_first_approach
from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters import fitter_cont_cal_second_approach
from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters import fitter_cont_cal_third_approach
from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters import fitter_cont_cal_fourth_approach

from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters import fitter_cont_cal_for_dss_1st_approach

from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters import fitter_cont_cal_in_b_tag_modes
from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters import fitter_cont_cal_for_dss_in_b_tag_modes


__all__ = [
    "n_e_ecl_extra_selection",
    "cont_cal_fitter_dict",
    "keys_to_evaluate",
]


n_e_ecl_extra_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"{RDStarHistVars.neextra_for_fit.df_label} > 0.75",
    cut_latex_label=rf"{LatexObsNames.roe_n_e_ecl_extra[:-1]} > 0.75\,{LatexUnits.gev[1:]}",
    cut_var_col_name=RDStarHistVars.neextra_for_fit.df_label,
)


cont_cal_fitter_dict: Dict[str, Type[ContCalFitterBase]] = {
    # Used Tag Cal Setups:
    "cont_cal_main": fitter_cont_cal_third_approach.ContCalFitter,
    "cont_cal_main_for_dss": fitter_cont_cal_for_dss_1st_approach.ContCalFitter,
    #
    # First approach for Continuum fits: 8 components
    "cont_cal_first_approach": fitter_cont_cal_first_approach.ContCalFitter,
    #
    # Second approach for Continuum fits: 12 components
    "cont_cal_second_approach": fitter_cont_cal_second_approach.ContCalFitter,
    #
    # Third approach for Continuum fits: in 2 reco channels and 4 components
    "cont_cal_third_approach": fitter_cont_cal_third_approach.ContCalFitter,
    #
    # Fourth approach for Continuum fits: in 2 reco channels (using D** reco channels) and 4 components
    "cont_cal_fourth_approach": fitter_cont_cal_fourth_approach.ContCalFitter,
    #
    # 1st approach for Continuum fits in Dss: in 2 reco channels and 4 components
    "cont_cal_for_dss_1st_approach": fitter_cont_cal_for_dss_1st_approach.ContCalFitter,
    #
    # Setups via B Tag Modes
    "cont_cal_in_b_tag_modes": fitter_cont_cal_in_b_tag_modes.ContCalFitter,
    "cont_cal_for_dss_in_b_tag_modes": fitter_cont_cal_for_dss_in_b_tag_modes.ContCalFitter,
}


keys_to_evaluate: Tuple[str, ...] = (
    "cont_cal_main",
    "cont_cal_main_for_dss",
)


assert all(k in cont_cal_fitter_dict for k in keys_to_evaluate), [
    k for k in keys_to_evaluate if k not in cont_cal_fitter_dict
]
