"""
Containers for fit information

Felix Metzner 2022
"""

import numpy as np
import pandas as pd

from uncertainties import UFloat
from dataclasses import dataclass
from typing import Tuple, List, Dict, Union, Optional, Callable, Iterator, Iterable, Mapping

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.reco_modes.reco_mode_utilities import RecoModeIDInfo, RecoModeGroupInfo
from rdstar.offline_analysis.fitting.dedicated_fit_approach.utility import immutable_cached_property

from rdstar.offline_analysis.efficiency_calculation.rdstar_efficiency_definitions import SigIDToDTMappingType


__all__ = [
    "DataSetupInfo",
    "FitObservableInfo",
    "WeightInfo",
    "ComponentInfo",
    "TagEfficiencyCorrectionInfo",
    "TagEfficiencyCorrectionInfoItem",
    "RecoChannelInfo",
    "FitSetupInfoContainer",
]


logger = log_to_default(name=__name__)


@dataclass(frozen=True)
class DataSetupInfo:
    recorded_data: bool = False
    toy_data: bool = False
    asimov_data: bool = False
    round_bin_count: bool = False
    toy_size: Optional[int] = None
    toy_seed: Optional[int] = None

    def __post_init__(self) -> None:
        _types: Dict[str, bool] = {
            "recorded_data": self.recorded_data,
            "toy_data": self.toy_data,
            "asimov_data": self.asimov_data,
        }

        if not sum([v for v in _types.values()]) == 1:
            raise ValueError(
                "Exactly one data type must be selected, but the following have been selected:\n\t-"
                + "\n\t-".join([k for k, v in _types.items() if v])
                + "\n"
            )

        if self.toy_data:
            if self.toy_seed is None:
                raise ValueError("A seed must be provided, if toy_data is selected.")
            if self.toy_size is None:
                raise ValueError("A size for the toy study must be provided, if toy_data is selected.")

        if self.recorded_data and self.round_bin_count:
            raise RuntimeError("The option round_bin_count cannot be used together with the setup type recorded_data!")

    @property
    def info_str(self) -> str:
        if self.recorded_data:
            return "Data"
        elif self.asimov_data:
            return "Asimov"
        elif self.toy_data:
            return f"Toys{self.toy_size}s{self.toy_seed}"
        else:
            raise RuntimeError(f"Unknown data setup:\n\t{self.__repr__()}...")


@dataclass(frozen=True)
class FitObservableInfo:
    hist_var: HistVariable
    alt_scope: Optional[Tuple[float, float]]
    alt_binning: Optional[Tuple[float, ...]]

    def __post_init__(self) -> None:
        if self.alt_binning is not None:
            assert self.alt_binning[0] == self.scope[0], (self.alt_binning, self.scope)
            assert self.alt_binning[-1] == self.scope[1], (self.alt_binning, self.scope)

    @property
    def col_name(self) -> str:
        return self.hist_var.df_label

    @property
    def latex_label(self) -> str:
        return self.hist_var.variable_name

    @property
    def scope(self) -> Tuple[float, float]:
        if self.alt_scope is not None:
            return self.alt_scope
        return self.hist_var.scope

    @property
    def bin_edges(self) -> Tuple[float, ...]:
        if self.alt_binning is not None:
            return self.alt_binning

        return tuple(np.linspace(*self.scope, self.hist_var.n_bins + 1))

    @property
    def n_bins(self) -> int:
        return len(self.bin_edges) - 1

    def reset(
        self,
        scope: Optional[Tuple[float, float]],
        binning: Union[int, Tuple[float, ...], None],
    ) -> "FitObservableInfo":
        _scope: Tuple[float, float] = self.scope if scope is None else scope
        if binning is None:
            _binning: Tuple[float, ...] = self.bin_edges
        elif isinstance(binning, int):
            _binning = tuple(np.linspace(_scope[0], _scope[1], binning))
        else:
            _binning = binning

        return FitObservableInfo(
            hist_var=self.hist_var,
            alt_scope=_scope,
            alt_binning=_binning,
        )


@dataclass(frozen=True)
class WeightInfo:
    name: str
    col_name: str


@dataclass(frozen=True)
class ComponentInfo:
    name: str
    comp_index: int
    latex_label: str
    sig_ids: Tuple[int, ...]
    color: Optional[str] = None

    def get_component_mask(
        self,
        df: pd.DataFrame,
        component_id_column_name: str,
    ) -> np.ndarray:
        return df[component_id_column_name].isin(self.sig_ids)


@dataclass(frozen=True)
class TagEfficiencyCorrectionInfoItem:
    key_name: str
    value: UFloat

    @property
    def correction(self) -> float:
        return self.value.nominal_value

    @property
    def uncertainty(self) -> float:
        return self.value.std_dev


class TagEfficiencyCorrectionInfo(Mapping):
    def __init__(
        self,
        items: Iterable[TagEfficiencyCorrectionInfoItem],
        key_order: Tuple[str, ...],
    ) -> None:
        _item_names: List[str] = [i.key_name for i in items]
        assert len(_item_names) == len(set(_item_names)), (len(_item_names), len(set(_item_names)), _item_names)
        assert set(key_order) == set(_item_names), (key_order, _item_names)

        self.key_order: Tuple[str, ...] = key_order
        self._info: Dict[str, TagEfficiencyCorrectionInfoItem] = {item.key_name: item for item in items}

    def __getitem__(self, item: str) -> TagEfficiencyCorrectionInfoItem:
        return self._info[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self._info)

    def __len__(self) -> int:
        return len(self._info)

    @property
    def ordered_values(self) -> Tuple[UFloat, ...]:
        return tuple(self[k].value for k in self.key_order)

    @property
    def ordered_nominal_values(self) -> Tuple[float, ...]:
        return tuple(self[k].value.nominal_value for k in self.key_order)

    @property
    def ordered_std_dev_values(self) -> Tuple[float, ...]:
        return tuple(self[k].value.std_dev for k in self.key_order)


@dataclass(frozen=True)
class RecoChannelInfo:
    name: str
    ch_index: int
    reco_mode_info: Union[RecoModeIDInfo, RecoModeGroupInfo]
    components: Tuple[ComponentInfo, ...]
    observable_infos: Tuple[FitObservableInfo, ...]
    tag_eff_corr_factor: UFloat

    @property
    def latex_label(self) -> str:
        return self.reco_mode_info.latex_label

    @property
    def reco_channel_ids(self) -> Tuple[int, ...]:
        if isinstance(self.reco_mode_info, RecoModeIDInfo):
            return (self.reco_mode_info.id,)
        else:
            return self.reco_mode_info.ids

    def __post_init__(self) -> None:
        _all_ids: List[int] = [i for c in self.components for i in c.sig_ids]
        assert len(_all_ids) == len(set(_all_ids)), _all_ids

    @property
    def n_bins_per_obs(self) -> Tuple[int, ...]:
        return tuple(o.n_bins for o in self.observable_infos)

    @property
    def n_bins_total(self) -> int:
        return np.product(self.n_bins_per_obs)

    @property
    def scope_per_obs(self) -> Tuple[Tuple[float, float], ...]:
        return tuple(o.scope for o in self.observable_infos)

    @property
    def observable_col_names(self) -> Tuple[str, ...]:
        return tuple(o.col_name for o in self.observable_infos)

    def get_mask(
        self,
        df: pd.DataFrame,
        reco_mode_column_str: str,
    ) -> np.ndarray:
        ch_mask: np.ndarray = df[reco_mode_column_str].isin(self.reco_channel_ids).values

        obs_query: str = " & ".join(
            [f"({obs.scope[0]} <= {obs.col_name} <= {obs.scope[1]})" for obs in self.observable_infos]
        )
        obs_mask: np.ndarray = df.eval(obs_query).values

        return ch_mask & obs_mask


class FitSetupInfoContainer:
    def __init__(
        self,
        components: Tuple[ComponentInfo, ...],
        reco_channels: Tuple[RecoChannelInfo, ...],
        observable_infos: Tuple[FitObservableInfo, ...],
        weight_info: WeightInfo,
        data_setup_info: DataSetupInfo,
        remove_empty_bins: bool,
        use_alternative_gap: bool,
        tag_efficiency_correction_info: TagEfficiencyCorrectionInfo,
        reco_mode_column_name: str,
        component_id_column_name: str,
        component_id_function: Optional[Callable] = None,
        sig_id_to_data_type_map_function: Optional[Callable] = None,
    ) -> None:

        _all_ids: List[int] = [i for c in components for i in c.sig_ids]
        assert len(_all_ids) == len(set(_all_ids)), _all_ids

        self._components: Tuple[ComponentInfo, ...] = components
        self._reco_channels: Tuple[RecoChannelInfo, ...] = reco_channels
        self._observable_infos: Tuple[FitObservableInfo, ...] = observable_infos

        self._weight_info: WeightInfo = weight_info
        self._reco_mode_col_name: str = reco_mode_column_name
        self._component_id_column_name: str = component_id_column_name

        self._component_id_function: Optional[Callable] = component_id_function

        self._data_setup_info: DataSetupInfo = data_setup_info
        self._remove_empty_bins: bool = remove_empty_bins

        self._use_alternative_gap: bool = use_alternative_gap

        self.sig_id_to_data_type_map_function: Optional[Callable] = sig_id_to_data_type_map_function

        self._tag_efficiency_correction_info: TagEfficiencyCorrectionInfo = tag_efficiency_correction_info

        self._check_consistency()

        self._comp_per_rc_mask: np.ndarray = self._get_comp_per_rc_mask()

    @property
    def components(self) -> Tuple[ComponentInfo, ...]:
        return self._components

    @property
    def reco_channels(self) -> Tuple[RecoChannelInfo, ...]:
        return self._reco_channels

    @property
    def observable_infos(self) -> Tuple[FitObservableInfo, ...]:
        return self._observable_infos

    @property
    def weight_info(self) -> WeightInfo:
        return self._weight_info

    @property
    def reco_mode_col(self) -> str:
        return self._reco_mode_col_name

    @property
    def component_id_column_name(self) -> str:
        return self._component_id_column_name

    @property
    def data_setup_info(self) -> DataSetupInfo:
        return self._data_setup_info

    @property
    def remove_empty_bins(self) -> bool:
        return self._remove_empty_bins

    @property
    def use_alternative_gap(self) -> bool:
        return self._use_alternative_gap

    @property
    def n_components(self) -> int:
        return len(self.components)

    @property
    def n_reco_channels(self) -> int:
        return len(self.reco_channels)

    @property
    def n_bins_per_reco_ch(self) -> Tuple[int, ...]:
        return tuple(rc.n_bins_total for rc in self.reco_channels)

    @property
    def n_bins_total(self) -> int:
        return sum(self.n_bins_per_reco_ch)

    @property
    def max_n_flattened_bins_per_ch(self) -> int:
        return max(self.n_bins_per_reco_ch)

    @property
    def max_n_bins_total(self) -> int:
        return self.max_n_flattened_bins_per_ch * self.n_reco_channels

    def get_max_n_bins_per_observable(self) -> Tuple[int, ...]:
        _n_obs: int = len(self._observable_infos)
        _n_max_bins_per_obs: List[int] = [0 for _ in range(_n_obs)]

        for reco_ch in self.reco_channels:
            for obs_id in range(_n_obs):
                _n_max_bins_per_obs[obs_id] = max(reco_ch.observable_infos[obs_id].n_bins, _n_max_bins_per_obs[obs_id])

        return tuple(_n_max_bins_per_obs)

    @property
    def tag_efficiency_correction_per_reco_channel(self) -> Tuple[UFloat, ...]:
        return tuple(rc.tag_eff_corr_factor for rc in self.reco_channels)

    @property
    def tag_efficiency_correction_info(self) -> TagEfficiencyCorrectionInfo:
        return self._tag_efficiency_correction_info

    @immutable_cached_property
    def sig_id_to_data_type_mapping(self) -> SigIDToDTMappingType:
        if self.sig_id_to_data_type_map_function is None:
            raise RuntimeError("No sig_id_to_data_type_mapping was provided, but it is required for the fit!")
        _map: SigIDToDTMappingType = self.sig_id_to_data_type_map_function(use_alternative_gap=self.use_alternative_gap)

        for sig_id in _map:
            assert any(sig_id in c.sig_ids for c in self.components), (sig_id, [c.sig_ids for c in self.components])

        return _map

    def prepare_mc_df(
        self,
        df: pd.DataFrame,
        fail_if_already_prepared: bool = True,
    ) -> None:
        if self._component_id_function is not None:
            if fail_if_already_prepared:
                assert self.component_id_column_name not in df.columns, self.component_id_column_name
            if self.component_id_column_name not in df.columns:
                self._component_id_function(df=df)

        assert self.component_id_column_name in df.columns, self.component_id_column_name

    def _check_consistency(self) -> None:
        reco_ch_indices: Tuple[int, ...] = tuple(ch.ch_index for ch in self.reco_channels)
        assert reco_ch_indices == tuple(range(self.n_reco_channels)), (self.n_reco_channels, reco_ch_indices)

        _rc_comps: Tuple[Tuple[ComponentInfo, ...], ...] = tuple([rc.components for rc in self.reco_channels])
        assert all(c in self.components for _comps in _rc_comps for c in _comps)

        _base_comp_indices: List[int] = [c.comp_index for c in self.components]
        assert _base_comp_indices == sorted(_base_comp_indices), _base_comp_indices

        _comp_sets: Tuple[Tuple[ComponentInfo, ...], ...] = tuple([self.components, *_rc_comps])
        for i, _comp_set in enumerate(_comp_sets):
            _comp_indices: List[int] = [c.comp_index for c in _comp_set]
            assert _comp_indices == sorted(_comp_indices), (i, _comp_indices)

        obs_col_names: Tuple[str, ...] = self.reco_channels[0].observable_col_names
        for _ch in self.reco_channels:
            assert obs_col_names == _ch.observable_col_names, (obs_col_names, _ch.observable_col_names, _ch.name)

    def _get_comp_per_rc_mask(self) -> np.ndarray:
        m = np.array([[c in rc.components for c in self.components] for rc in self.reco_channels])
        assert m.shape == (self.n_reco_channels, self.n_components), (m.shape, self.n_reco_channels, self.n_components)
        return m

    @property
    def components_name_to_id_mapping(self) -> Dict[int, str]:
        return {c.comp_index: c.name for c in self.components}

    def create_setup_with_new_data_setup(self, data_setup: DataSetupInfo) -> "FitSetupInfoContainer":
        return FitSetupInfoContainer(
            components=self.components,
            reco_channels=self.reco_channels,
            observable_infos=self.observable_infos,
            weight_info=self.weight_info,
            data_setup_info=data_setup,
            remove_empty_bins=self.remove_empty_bins,
            use_alternative_gap=self.use_alternative_gap,
            tag_efficiency_correction_info=self.tag_efficiency_correction_info,
            reco_mode_column_name=self.reco_mode_col,
            component_id_column_name=self.component_id_column_name,
            component_id_function=self._component_id_function,
            sig_id_to_data_type_map_function=self.sig_id_to_data_type_map_function,
        )
