"""
Fit setup utilities

Felix Metzner 2023
"""

from dataclasses import dataclass
from typing import Tuple, List, Dict, Optional, Union, Iterable, Callable

from rdstar.utilities.logging import log_to_default
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.weight_management.weight_information import WeightManager

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.observables.mc_observables import (
    RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID,
    DssRecoChAndSigDTruthChargeFitSignalID,
    DssRecoChAndMainSigDTruthChargeFitSignalID,
    FEIBTagRecoChAndSignalID,
    FEIBTagAndSigDRecoChAndSignalID,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import tag_cal_obs_setups
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import dss_cal_obs_setups
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import cont_cal_obs_setups

from rdstar.offline_analysis.fitting.dedicated_fit_approach.utility import generate_sig_id_application_function
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    DataSetupInfo,
    WeightInfo,
)


__all__ = [
    "base_asimov_data_setup_info",
    "FitSetupTriplet",
    #
    "old_rdstar_weight_info",
    "base_rdstar_weight_info",
    #
    # Tag Calibration fit
    "tag_cal_fit_obs_setup",
    "base_tag_cal_weight_info",
    "tag_cal_dstst_l_bkg_sig_ids",
    "tag_cal_bbbar_bkg_sig_ids",
    "common_tag_cal_fit_component_id_column_name",
    "add_common_tag_cal_fit_component_id_to",
    #
    # Tag Calibration per Tag Mode fit
    "tag_cal_per_tag_mode_fit_obs_setup",
    "fei_tag_b_reco_based_fit_component_id_column_name",
    "add_fei_tag_b_reco_based_fit_component_id",
    #
    # Tag Calibration Closure Test fit
    "closure_test_a_fit_component_id_column_name",
    "add_closure_test_a_fit_component_id",
    #
    # D** Calibration fit
    "dss_cal_fit_obs_setup",
    "base_dss_cal_weight_info",
    "dss_cal_bbbar_bkg_sig_ids",
    "common_dss_cal_fit_component_id_column_name",
    "add_common_dss_cal_fit_component_id_to",
    "dss_specific_fit_component_id_column_name",
    "add_dss_specific_fit_component_id_to",
    "main_dss_cal_fit_component_id_column_name",
    "add_main_dss_cal_fit_component_id_to",
    #
    # Continuum Calibration fit
    "cont_cal_fit_obs_setup",
    "cont_cal_b_charge_fit_obs_setup",
    "base_cont_cal_weight_info",
    "cont_cal_bbbar_bkg_sig_ids",
    "cont_cal_dss_bkg_sig_ids",
    "cont_cal_other_bbbar_bkg_sig_ids",
    "common_cont_cal_fit_component_id_column_name",
    "add_common_cont_cal_fit_component_id_to",
    #
    # Generic methods
    "get_sig_ids_with_charge_info",
    "get_sig_ids_with_d_charge_info",
    "get_sig_ids_with_main_d_charge_info",
    "get_sig_ids_with_tag_b_reco_info",
    "get_sig_ids_for_tag_cal_closure_a_info",
]


logger = log_to_default(name=__name__)


base_asimov_data_setup_info = DataSetupInfo(
    recorded_data=False,
    toy_data=False,
    asimov_data=True,
    round_bin_count=False,
    toy_size=None,
    toy_seed=None,
)


@dataclass(frozen=True)
class FitSetupTriplet:
    asimov: FitSetupInfoContainer

    def get_toy_setup(
        self,
        toy_size: int = 100,
        toy_seed: int = 13374748,
    ) -> FitSetupInfoContainer:
        _toy_data_setup_info = DataSetupInfo(
            recorded_data=False,
            toy_data=True,
            asimov_data=False,
            round_bin_count=False,
            toy_size=toy_size,
            toy_seed=toy_seed,
        )
        return self.asimov.create_setup_with_new_data_setup(data_setup=_toy_data_setup_info)

    @property
    def data(self) -> FitSetupInfoContainer:
        _data_data_setup_info = DataSetupInfo(
            recorded_data=True,
            toy_data=False,
            asimov_data=False,
            round_bin_count=False,
            toy_size=None,
            toy_seed=None,
        )
        return self.asimov.create_setup_with_new_data_setup(data_setup=_data_data_setup_info)


# RDStar Fit Specific stuff:

old_rdstar_weight_info = WeightInfo(
    name=WeightManager.final_weight_with_bf_corrections_col,
    col_name=WeightManager.final_weight_with_bf_corrections_col,
)

base_rdstar_weight_info = WeightInfo(
    name=Cols.hammered_weight_col_name,
    col_name=Cols.hammered_weight_col_name,
)


# Tag Cal Fit Specific stuff:
tag_cal_fit_obs_setup = tag_cal_obs_setups.TagCalObsSetup2p2

base_tag_cal_weight_info = WeightInfo(
    name=Cols.hammered_weight_col_name,
    col_name=Cols.hammered_weight_col_name,
)

tag_cal_dstst_l_bkg_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.DSSl.id,
    OldMCMatchingIDsQueryInfo.DSSlbkg.id,
)

tag_cal_bbbar_bkg_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.HBKG.id,
    OldMCMatchingIDsQueryInfo.GBKG.id,
    OldMCMatchingIDsQueryInfo.PiLNu.id,
)


common_tag_cal_fit_component_id_column_name: str = RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID.new_column_name
add_common_tag_cal_fit_component_id_to: Callable = generate_sig_id_application_function(
    sig_col_obs_definition_class=RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID,
    new_col_name=common_tag_cal_fit_component_id_column_name,
)


# Tag Cal per Tag Mode Fit Specific stuff:

tag_cal_per_tag_mode_fit_obs_setup = tag_cal_obs_setups.TagCalPerBTagModeObsSetup1

fei_tag_b_reco_based_fit_component_id_column_name: str = FEIBTagRecoChAndSignalID.new_column_name
add_fei_tag_b_reco_based_fit_component_id: Callable = generate_sig_id_application_function(
    sig_col_obs_definition_class=FEIBTagRecoChAndSignalID,
    new_col_name=fei_tag_b_reco_based_fit_component_id_column_name,
    add_modified_fei_reco_mode_col=True,
)


# Tag Cal Closure Test Fits Specific stuff:
closure_test_a_fit_component_id_column_name: str = FEIBTagAndSigDRecoChAndSignalID.new_column_name
add_closure_test_a_fit_component_id: Callable = generate_sig_id_application_function(
    sig_col_obs_definition_class=FEIBTagAndSigDRecoChAndSignalID,
    new_col_name=closure_test_a_fit_component_id_column_name,
    add_modified_fei_reco_mode_col=True,
    add_combined_tag_and_sig_reco_mode_col=True,
)


# D** Cal Fit Specific stuff:

dss_cal_fit_obs_setup = dss_cal_obs_setups.DssCalObsSetup3

base_dss_cal_weight_info = WeightInfo(
    name=Cols.hammered_weight_col_name,
    col_name=Cols.hammered_weight_col_name,
)

dss_cal_bbbar_bkg_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.HBKG.id,
    OldMCMatchingIDsQueryInfo.GBKG.id,
    OldMCMatchingIDsQueryInfo.PiLNu.id,
)

common_dss_cal_fit_component_id_column_name: str = RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID.new_column_name
add_common_dss_cal_fit_component_id_to: Callable = generate_sig_id_application_function(
    sig_col_obs_definition_class=RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID,
    new_col_name=common_dss_cal_fit_component_id_column_name,
)

dss_specific_fit_component_id_column_name: str = DssRecoChAndSigDTruthChargeFitSignalID.new_column_name
add_dss_specific_fit_component_id_to: Callable = generate_sig_id_application_function(
    sig_col_obs_definition_class=DssRecoChAndSigDTruthChargeFitSignalID,
    new_col_name=dss_specific_fit_component_id_column_name,
)

main_dss_cal_fit_component_id_column_name: str = DssRecoChAndMainSigDTruthChargeFitSignalID.new_column_name
add_main_dss_cal_fit_component_id_to: Callable = generate_sig_id_application_function(
    sig_col_obs_definition_class=DssRecoChAndMainSigDTruthChargeFitSignalID,
    new_col_name=main_dss_cal_fit_component_id_column_name,
)


# Continuum Cal Fit Specific stuff:

cont_cal_fit_obs_setup = cont_cal_obs_setups.ContCalObsSetup1
cont_cal_b_charge_fit_obs_setup = cont_cal_obs_setups.ContCalBChargeObsSetup1

base_cont_cal_weight_info = WeightInfo(
    name=Cols.hammered_weight_col_name,
    col_name=Cols.hammered_weight_col_name,
)

cont_cal_bbbar_bkg_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.DSSl.id,
    OldMCMatchingIDsQueryInfo.DSSlbkg.id,
    OldMCMatchingIDsQueryInfo.HBKG.id,
    OldMCMatchingIDsQueryInfo.GBKG.id,
    OldMCMatchingIDsQueryInfo.PiLNu.id,
)

cont_cal_dss_bkg_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.DSSl.id,
    OldMCMatchingIDsQueryInfo.DSSlbkg.id,
)

cont_cal_other_bbbar_bkg_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.HBKG.id,
    OldMCMatchingIDsQueryInfo.GBKG.id,
    OldMCMatchingIDsQueryInfo.PiLNu.id,
)

common_cont_cal_fit_component_id_column_name: str = RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID.new_column_name
add_common_cont_cal_fit_component_id_to: Callable = generate_sig_id_application_function(
    sig_col_obs_definition_class=RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID,
    new_col_name=common_cont_cal_fit_component_id_column_name,
)


# Generic Stuff


def get_sig_ids_with_charge_info(
    sig_ids: Union[int, Iterable[int]],
    reco_type: Optional[str] = None,  # ["c_d", "n_d", "c_dst", "n_dst", "charged", "neutral", None="all"]
    charge_type: Optional[str] = None,  # ["charged", "neutral", None="both"]
    cross_feed_type: Optional[str] = None,  # ["cf", "wwo_cf", "wo_cf", None]
) -> Tuple[int, ...]:
    SigObsClass = RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID

    _valid_charge_values: Tuple[Optional[str], ...] = ("charged", "neutral", None, "both")
    assert charge_type in _valid_charge_values, (charge_type, _valid_charge_values)

    _reco_dict: Dict[Optional[str], Tuple[int, ...]] = {
        "c_d": (SigObsClass.charged_d_offsets,),
        "c_dst": (SigObsClass.charged_dst_offsets,),
        "n_d": (SigObsClass.neutral_d_offsets,),
        "n_dst": (SigObsClass.neutral_dst_offsets,),
        "charged": SigObsClass.charged_offsets,
        "neutral": SigObsClass.neutral_offsets,
        None: tuple(list(SigObsClass.charged_offsets) + list(SigObsClass.neutral_offsets)),
        "all": tuple(list(SigObsClass.charged_offsets) + list(SigObsClass.neutral_offsets)),
    }
    _valid_reco_values: Tuple[Optional[str], ...] = tuple(_reco_dict.keys())
    assert reco_type in _valid_reco_values, (reco_type, _valid_reco_values)

    _valid_cf_values: Tuple[Optional[str], ...] = ("cf", "wwo_cf", "wo_cf", None)
    assert cross_feed_type in _valid_cf_values, (cross_feed_type, _valid_cf_values)

    if isinstance(sig_ids, int):
        _ids: List[int] = [sig_ids]
    elif isinstance(sig_ids, Iterable) and all(isinstance(v, int) for v in sig_ids):
        _ids = list(sig_ids)
    else:
        raise ValueError(f"Parameter sig_ids must be integer or iterable if integers! Provided was {sig_ids}")

    if reco_type in _reco_dict:
        _res = [v + o for v in _ids for o in _reco_dict[reco_type]]
    else:
        raise ValueError(f"Provided reco_type '{reco_type}' is not one of {_valid_reco_values}!")

    if cross_feed_type is not None:
        assert all(i in SigObsClass.cross_feed_dict for i in _ids), (
            _ids,
            SigObsClass.cross_feed_dict,
        )

    if cross_feed_type == "cf":
        _new_res = [v + SigObsClass.cross_feed_offset for v in _res]
        _res = _new_res
    elif cross_feed_type == "wwo_cf":
        _new_res = [v + SigObsClass.cross_feed_offset for v in _res]
        _res.extend(_new_res)
    elif cross_feed_type == "wo_cf" or cross_feed_type is None:
        pass  # Nothing to do for this case
    else:
        raise ValueError(f"Provided cross_feed_type '{cross_feed_type}' is not one of {_valid_cf_values}!")

    if charge_type == "charged":
        return tuple(_res)
    elif charge_type == "neutral":
        return tuple([v * SigObsClass.sig_id_neutral_sign_flip for v in _res])
    elif charge_type == "both" or charge_type is None:
        return tuple(_res + [v * SigObsClass.sig_id_neutral_sign_flip for v in _res])
    else:
        raise ValueError(f"Provided charge_type '{charge_type}' is not one of {_valid_charge_values}!")


def get_sig_ids_with_d_charge_info(
    sig_ids: Union[int, Iterable[int]],
    reco_type: Optional[str] = None,  # ["c_d", "n_d", "c_dst", "n_dst", "charged", "neutral", None="all"]
    d_charge_type: Optional[str] = None,  # ["charged", "neutral", None="both"]
) -> Tuple[int, ...]:
    SigObsClass = DssRecoChAndSigDTruthChargeFitSignalID

    _valid_d_charge_values: Tuple[Optional[str], ...] = ("charged", "neutral", None, "both")
    assert d_charge_type in _valid_d_charge_values, (d_charge_type, _valid_d_charge_values)

    _reco_dict: Dict[Optional[str], Tuple[int, ...]] = {
        "c_d": (SigObsClass.charged_d_offsets,),
        "c_dst": (SigObsClass.charged_dst_offsets,),
        "n_d": (SigObsClass.neutral_d_offsets,),
        "n_dst": (SigObsClass.neutral_dst_offsets,),
        "charged": SigObsClass.charged_offsets,
        "neutral": SigObsClass.neutral_offsets,
        None: tuple(list(SigObsClass.charged_offsets) + list(SigObsClass.neutral_offsets)),
        "all": tuple(list(SigObsClass.charged_offsets) + list(SigObsClass.neutral_offsets)),
    }
    _valid_reco_values: Tuple[Optional[str], ...] = tuple(_reco_dict.keys())
    assert reco_type in _valid_reco_values, (reco_type, _valid_reco_values)

    if isinstance(sig_ids, int):
        _ids: List[int] = [sig_ids]
    elif isinstance(sig_ids, Iterable) and all(isinstance(v, int) for v in sig_ids):
        _ids = list(sig_ids)
    else:
        raise ValueError(f"Parameter sig_ids must be integer or iterable if integers! Provided was {sig_ids}")

    if reco_type in _reco_dict:
        _res = [v + o for v in _ids for o in _reco_dict[reco_type]]
    else:
        raise ValueError(f"Provided reco_type '{reco_type}' is not one of {_valid_reco_values}!")

    if d_charge_type == "charged":
        return tuple(_res)
    elif d_charge_type == "neutral":
        return tuple([v * SigObsClass.sig_id_neutral_sign_flip for v in _res])
    elif d_charge_type == "both" or d_charge_type is None:
        return tuple(_res + [v * SigObsClass.sig_id_neutral_sign_flip for v in _res])
    else:
        raise ValueError(f"Provided d_charge_type '{d_charge_type}' is not one of {_valid_d_charge_values}!")


def get_sig_ids_with_main_d_charge_info(
    sig_ids: Union[int, Iterable[int]],
    reco_type: Optional[str],
    # ["d_cpi", "d_npi", "dst_cpi", "dst_npi", "c_pi", "n_pi", "d", "dst", None="all", "other", "all_and_other"]
    main_d_type: Optional[str],  # ["c_d", "n_d", "c_dst", "n_dst", "d", "dst", "charged", "neutral", "all", None]
    true_dss_type: Optional[str] = None,  # ["narrow", "wide", "both", "other", None]
) -> Tuple[int, ...]:
    SigObsClass = DssRecoChAndMainSigDTruthChargeFitSignalID

    _main_d_dict: Dict[Optional[str], Tuple[int, ...]] = {
        "c_d": (SigObsClass.true_cd_offset,),
        "c_dst": (SigObsClass.true_cdst_offset,),
        "n_d": (SigObsClass.true_nd_offset,),
        "n_dst": (SigObsClass.true_ndst_offset,),
        "charged": SigObsClass.true_charged_offsets,
        "neutral": SigObsClass.true_neutral_offsets,
        "d": SigObsClass.true_d_offsets,
        "dst": SigObsClass.true_dst_offsets,
        "all": SigObsClass.true_charged_offsets + SigObsClass.true_neutral_offsets,
        None: (0,),
    }

    _valid_main_d_values: Tuple[Optional[str], ...] = tuple(_main_d_dict.keys())
    assert main_d_type in _valid_main_d_values, (main_d_type, _valid_main_d_values)

    if isinstance(sig_ids, int):
        if main_d_type is None:
            assert sig_ids >= 7, sig_ids
        if sig_ids >= 7:
            assert main_d_type is None, (sig_ids, main_d_type)
    elif isinstance(sig_ids, Iterable):
        if main_d_type is None:
            assert all(i >= 7 for i in sig_ids), sig_ids
        if any(_id >= 7 for _id in sig_ids):
            assert main_d_type is None, (sig_ids, main_d_type)
    else:
        raise ValueError(f"Parameter sig_ids must be integer or iterable if integers! Provided was {sig_ids}")

    _true_dss_dict: Dict[Optional[str], Tuple[int, ...]] = {
        "narrow": (SigObsClass.narrow_dss_offset,),
        "wide": (SigObsClass.wide_dss_offset,),
        "both": (SigObsClass.narrow_dss_offset, SigObsClass.wide_dss_offset),
        "other": (SigObsClass.no_dss_offset,),
        None: SigObsClass.all_dss_offsets,
    }

    _valid_true_dss_values: Tuple[Optional[str], ...] = tuple(_true_dss_dict.keys())
    assert true_dss_type in _valid_true_dss_values, (true_dss_type, _valid_true_dss_values)

    if true_dss_type not in (None, "other"):
        if isinstance(sig_ids, int):
            assert sig_ids in (5, 6), sig_ids
        elif isinstance(sig_ids, Iterable):
            assert all(i in (5, 6) for i in sig_ids), sig_ids
        else:
            raise ValueError(f"Parameter sig_ids must be integer or iterable if integers! Provided was {sig_ids}")

    _reco_dict: Dict[Optional[str], Tuple[int, ...]] = {
        "d_cpi": (SigObsClass.reco_d_charged_pi_offset,),
        "d_npi": (SigObsClass.reco_d_neutral_pi_offset,),
        "dst_cpi": (SigObsClass.reco_dst_charged_pi_offset,),
        "dst_npi": (SigObsClass.reco_dst_neutral_pi_offset,),
        "c_pi": SigObsClass.reco_charged_pi_offsets,
        "n_pi": SigObsClass.reco_neutral_pi_offsets,
        "d": SigObsClass.reco_d_offsets,
        "dst": SigObsClass.reco_dst_offsets,
        None: SigObsClass.reco_all_offsets,
        "all": SigObsClass.reco_all_offsets,
        "other": (SigObsClass.reco_others_offset,),
        "all_and_other": SigObsClass.reco_all_offsets + (SigObsClass.reco_others_offset,),
    }

    _valid_reco_values: Tuple[Optional[str], ...] = tuple(_reco_dict.keys())
    assert reco_type in _valid_reco_values, (reco_type, _valid_reco_values)

    # Sig IDs
    if isinstance(sig_ids, int):
        _ids: List[int] = [sig_ids]
    elif isinstance(sig_ids, Iterable) and all(isinstance(v, int) for v in sig_ids):
        _ids = list(sig_ids)
    else:
        raise ValueError(f"Parameter sig_ids must be integer or iterable if integers! Provided was {sig_ids}")

    # Reco Type
    _res: List[int] = [v + o for v in _ids for o in _reco_dict[reco_type]]

    # True D** Meson Type
    _res_2nd_step: List[int] = [v + o for v in _res for o in _true_dss_dict[true_dss_type]]

    # True Main D Meson Type
    return tuple([v + md_of for v in _res_2nd_step for md_of in _main_d_dict[main_d_type]])


def get_sig_ids_with_tag_b_reco_info(
    sig_ids: Union[int, Iterable[int]],
    fei_reco_type: Optional[str] = None,  # ["d_1cpi", "d_mcpi", "ocd_pipi", "dd", "ds_1cpi", "ds_mcpi", "dsd",
    #                                     #  "j_psi", "dsd_and_j_psi", None="all"]
    fei_reco_charge_type: Optional[str] = None,  # ["charged", "neutral", None="both"]
) -> Tuple[int, ...]:
    SigObsClass = FEIBTagRecoChAndSignalID

    _valid_tag_charge_values: Tuple[Optional[str], ...] = ("charged", "neutral", None, "both")
    assert fei_reco_charge_type in _valid_tag_charge_values, (fei_reco_charge_type, _valid_tag_charge_values)

    _reco_dict: Dict[Optional[str], Tuple[int, ...]] = {
        "d_1cpi": (SigObsClass.b_to_d_1cpi_tag_offset,),
        "d_mcpi": (SigObsClass.b_to_d_mcpi_tag_offset,),
        "ocd_pipi": (SigObsClass.b_to_ocd_pipi_tag_offset,),
        "dd": (SigObsClass.b_to_dd_tag_offset,),
        "ds_1cpi": (SigObsClass.b_to_ds_1cpi_tag_offset,),
        "ds_mcpi": (SigObsClass.b_to_ds_mcpi_tag_offset,),
        "dsd": (SigObsClass.b_to_dsd_tag_offset,),
        "j_psi": (SigObsClass.b_to_j_psi_tag_offset,),
        "dd_and_dsd": (SigObsClass.b_to_dd_tag_offset, SigObsClass.b_to_dsd_tag_offset),
        "dd_dsd_and_j_psi": (
            SigObsClass.b_to_dd_tag_offset,
            SigObsClass.b_to_dsd_tag_offset,
            SigObsClass.b_to_j_psi_tag_offset,
        ),
        None: SigObsClass.all_offset_values,
        "all": SigObsClass.all_offset_values,
    }

    _valid_fei_tag_reco_values: Tuple[Optional[str], ...] = tuple(_reco_dict.keys())
    assert fei_reco_type in _valid_fei_tag_reco_values, (fei_reco_type, _valid_fei_tag_reco_values)

    if isinstance(sig_ids, int):
        _ids: List[int] = [sig_ids]
    elif isinstance(sig_ids, Iterable) and all(isinstance(v, int) for v in sig_ids):
        _ids = list(sig_ids)
    else:
        raise ValueError(f"Parameter sig_ids must be integer or iterable if integers! Provided was {sig_ids}")

    if fei_reco_type in _reco_dict:
        _res = [v + o for v in _ids for o in _reco_dict[fei_reco_type]]
    else:
        raise ValueError(f"Provided reco_type '{fei_reco_type}' is not one of {_valid_fei_tag_reco_values}!")

    if fei_reco_charge_type == "charged":
        return tuple(_res)
    elif fei_reco_charge_type == "neutral":
        return tuple([v * SigObsClass.sig_id_neutral_b_tag_sign_flip for v in _res])
    elif fei_reco_charge_type == "both" or fei_reco_charge_type is None:
        return tuple(_res + [v * SigObsClass.sig_id_neutral_b_tag_sign_flip for v in _res])
    else:
        raise ValueError(f"Provided charge_type '{fei_reco_charge_type}' is not one of {_valid_tag_charge_values}!")


def get_sig_ids_for_tag_cal_closure_a_info(
    sig_ids: Union[int, Iterable[int]],
    fei_reco_type: Optional[str] = None,  # ["d_1cpi", "d_mcpi", "ocd_pipi", "dd", "ds_1cpi", "ds_mcpi", "dsd",
    #                                     #  "j_psi", "dsd_and_j_psi", None="all"]
    fei_reco_charge_type: Optional[str] = None,  # ["charged", "neutral", None="both"]
    sig_b_reco_via_ds: Optional[bool] = None,  # [True, False, None="both"]
) -> Tuple[int, ...]:
    SigObsClass = FEIBTagAndSigDRecoChAndSignalID

    _valid_tag_charge_values: Tuple[Optional[str], ...] = ("charged", "neutral", None, "both")
    assert fei_reco_charge_type in _valid_tag_charge_values, (fei_reco_charge_type, _valid_tag_charge_values)

    _reco_dict: Dict[Optional[str], Tuple[int, ...]] = {
        "d_1cpi": (SigObsClass.b_to_d_1cpi_tag_offset,),
        "d_mcpi": (SigObsClass.b_to_d_mcpi_tag_offset,),
        "ocd_pipi": (SigObsClass.b_to_ocd_pipi_tag_offset,),
        "dd": (SigObsClass.b_to_dd_tag_offset,),
        "ds_1cpi": (SigObsClass.b_to_ds_1cpi_tag_offset,),
        "ds_mcpi": (SigObsClass.b_to_ds_mcpi_tag_offset,),
        "dsd": (SigObsClass.b_to_dsd_tag_offset,),
        "j_psi": (SigObsClass.b_to_j_psi_tag_offset,),
        "dd_and_dsd": (SigObsClass.b_to_dd_tag_offset, SigObsClass.b_to_dsd_tag_offset),
        "dd_dsd_and_j_psi": (
            SigObsClass.b_to_dd_tag_offset,
            SigObsClass.b_to_dsd_tag_offset,
            SigObsClass.b_to_j_psi_tag_offset,
        ),
        None: SigObsClass.all_offset_values,
        "all": SigObsClass.all_offset_values,
    }

    _valid_fei_tag_reco_values: Tuple[Optional[str], ...] = tuple(_reco_dict.keys())
    assert fei_reco_type in _valid_fei_tag_reco_values, (fei_reco_type, _valid_fei_tag_reco_values)

    _valid_sig_b_reco_via_ds_values: Tuple[Optional[bool], ...] = (True, False, None)
    assert sig_b_reco_via_ds in _valid_sig_b_reco_via_ds_values, (sig_b_reco_via_ds, _valid_sig_b_reco_via_ds_values)

    if isinstance(sig_ids, int):
        _ids: List[int] = [sig_ids]
    elif isinstance(sig_ids, Iterable) and all(isinstance(v, int) for v in sig_ids):
        _ids = list(sig_ids)
    else:
        raise ValueError(f"Parameter sig_ids must be integer or iterable if integers! Provided was {sig_ids}")

    if fei_reco_type in _reco_dict:
        _base_res = [v + o for v in _ids for o in _reco_dict[fei_reco_type]]
    else:
        raise ValueError(f"Provided reco_type '{fei_reco_type}' is not one of {_valid_fei_tag_reco_values}!")

    if sig_b_reco_via_ds is True:
        _res = [_i + SigObsClass.sig_d_star_offset for _i in _base_res]
    elif sig_b_reco_via_ds is False:
        _res = _base_res
    elif sig_b_reco_via_ds is None:
        _res = [i for i in _base_res]
        _res.extend([_i + SigObsClass.sig_d_star_offset for _i in _base_res])
    else:
        raise ValueError(f"sig_b_reco_via_ds='{sig_b_reco_via_ds}' is not one of {_valid_sig_b_reco_via_ds_values}!")

    if fei_reco_charge_type == "charged":
        return tuple(_res)
    elif fei_reco_charge_type == "neutral":
        return tuple([v * SigObsClass.sig_id_neutral_b_tag_sign_flip for v in _res])
    elif fei_reco_charge_type == "both" or fei_reco_charge_type is None:
        return tuple(_res + [v * SigObsClass.sig_id_neutral_b_tag_sign_flip for v in _res])
    else:
        raise ValueError(f"Provided charge_type '{fei_reco_charge_type}' is not one of {_valid_tag_charge_values}!")
