"""
Tools for rdstar fit component plots results.

Felix Metzner 2023
"""


import os
import pandas as pd

from dataclasses import dataclass
from typing import Tuple, Dict, Optional

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.rdstar_plots import plot_stacked_for_all_modes

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.rdstar_fit_obs_setups import FitObsSetupMain
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import base_rdstar_weight_info
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.rdstar_fit_info import (
    original__rdstar_fit__component_info_collection,
)


__all__ = [
    "RDStarFitComponentPlots",
]


logger = log_to_default(name=__name__)


@dataclass(frozen=True)
class LegendInfo:
    loc: int
    n_cols: int


@dataclass(frozen=True)
class InfoTextDef:
    text: str
    x_pos: float
    y_pos: float


@dataclass(frozen=True)
class ComponentPlotInfo:
    hist_var: HistVariable
    query_str: Optional[str] = None
    special_binning: Optional[Dict[str, Tuple[float, ...]]] = None
    output_name_appendix: Optional[str] = None
    legend_info: Optional[LegendInfo] = None
    info_text: Optional[InfoTextDef] = None

    @property
    def output_name(self) -> str:
        if self.output_name_appendix is None:
            return self.hist_var.df_label
        else:
            return f"{self.hist_var.df_label}_{self.output_name_appendix}"


# region component plot infos

_neextra_hv: HistVariable = RDStarHistVars.neextra_for_main_fit
_mms_hv: HistVariable = RDStarHistVars.missing_mass_res_corr_for_main_fit

base_neextra_query: str = f"{_neextra_hv.scope[0]} <= {_neextra_hv.df_label} <= {_neextra_hv.scope[1]}"
base_mms_query: str = f"{_mms_hv.scope[0]} <= {_mms_hv.df_label} <= {_mms_hv.scope[1]}"
signal_zoom_mms_query: str = f"1.5 <= {_mms_hv.df_label} <= {_mms_hv.scope[1]}"


mms_signal_zoom_info_text_def: InfoTextDef = InfoTextDef(
    text=rf"{_mms_hv.variable_name} $> 1.5\,${_mms_hv.unit}",
    x_pos=0.74,
    y_pos=0.74,
)
mms_signal_zoom_legend_info: LegendInfo = LegendInfo(loc=2, n_cols=3)


neextra_comp_plot_info = ComponentPlotInfo(
    hist_var=_neextra_hv,
    query_str=base_mms_query,
    special_binning=None,
    output_name_appendix="with_10bin_binning",
    legend_info=None,
    info_text=None,
)

neextra_comp_fit_binning_plot_info = ComponentPlotInfo(
    hist_var=_neextra_hv,
    query_str=base_mms_query,
    special_binning=FitObsSetupMain.binning_for_reco_modes_obs_fit_neextra,
    output_name_appendix="with_fit_binning",
    legend_info=None,
    info_text=None,
)

neextra_comp_with_mms_sig_zoom_plot_info = ComponentPlotInfo(
    hist_var=_neextra_hv,
    query_str=signal_zoom_mms_query,
    special_binning=None,
    output_name_appendix="with_10bin_binning_MMSZoom",
    legend_info=mms_signal_zoom_legend_info,
    info_text=mms_signal_zoom_info_text_def,
)

neextra_comp_fit_binning_with_mms_sig_zoom_plot_info = ComponentPlotInfo(
    hist_var=_neextra_hv,
    query_str=signal_zoom_mms_query,
    special_binning=FitObsSetupMain.binning_for_reco_modes_obs_fit_neextra,
    output_name_appendix="with_fit_binning_MMSZoom",
    legend_info=mms_signal_zoom_legend_info,
    info_text=mms_signal_zoom_info_text_def,
)


mms_comp_plot_info = ComponentPlotInfo(
    hist_var=_mms_hv,
    query_str=base_neextra_query,
    special_binning=None,
    output_name_appendix="with_10bin_binning",
    legend_info=None,
    info_text=None,
)

mms_comp_fit_binning_plot_info = ComponentPlotInfo(
    hist_var=_mms_hv,
    query_str=base_neextra_query,
    special_binning=FitObsSetupMain.binning_for_reco_modes_obs_fit_mms,
    output_name_appendix="with_fit_binning",
    legend_info=None,
    info_text=None,
)

rdstar_component_plot_infos: Tuple[ComponentPlotInfo, ...] = (
    neextra_comp_plot_info,
    neextra_comp_fit_binning_plot_info,
    neextra_comp_with_mms_sig_zoom_plot_info,
    neextra_comp_fit_binning_with_mms_sig_zoom_plot_info,
    mms_comp_plot_info,
    mms_comp_fit_binning_plot_info,
)

# endregion


# region RDStar fit component plotter class
class RDStarFitComponentPlots:

    plot_dir_name: str = "RDStarComponentPlots"

    plot_infos: Tuple[ComponentPlotInfo, ...] = rdstar_component_plot_infos

    def __init__(
        self,
        evaluation_name: str,
        base_output_path: PathType,
        using_old_ratio_values: bool,
    ) -> None:

        self.evaluation_name: str = evaluation_name
        self.base_output_path: PathType = base_output_path
        self.using_old_ratio_values: bool = using_old_ratio_values

        self.output_path: PathType = os.path.join(self.base_output_path, self.plot_dir_name, self.evaluation_name)

        self.weight_col_name: str = base_rdstar_weight_info.col_name

    def produce_fit_component_plots(
        self,
        mc_sample: pd.DataFrame,
        close_figures: bool = False,
    ) -> None:
        _output_dir: PathType = os.path.join(
            self.output_path,
            "Belle2015" if self.using_old_ratio_values else "This2023",
        )
        os.makedirs(_output_dir, exist_ok=True)

        assert self.weight_col_name in mc_sample.columns, self.weight_col_name

        for component_plot_info in self.plot_infos:
            self.produce_component_plots(
                mc_sample=mc_sample,
                component_plot_info=component_plot_info,
                output_dir_path=_output_dir,
                close_figures=close_figures,
            )

    def produce_component_plots(
        self,
        mc_sample: pd.DataFrame,
        component_plot_info: ComponentPlotInfo,
        output_dir_path: PathType,
        close_figures: bool = False,
    ) -> None:
        for this_fig, this_ax, this_output_key, _ in plot_stacked_for_all_modes(
            df=mc_sample.query(component_plot_info.query_str),
            variable=component_plot_info.hist_var,
            weight_column=self.weight_col_name,
            special_binning=component_plot_info.special_binning,
            sig_mc_matching_collection=original__rdstar_fit__component_info_collection,
            reco_mode_info_collection=None,
            fig_dpi=300,
            output_name=component_plot_info.output_name,
            base_output_path=output_dir_path,
            close_figure=close_figures,
        ):
            assert this_fig is not None
            assert this_ax is not None
            assert this_output_key is not None

            if component_plot_info.legend_info is not None:
                this_ax.legend(
                    frameon=False,
                    loc=component_plot_info.legend_info.loc,
                    ncol=component_plot_info.legend_info.n_cols,
                )

            if component_plot_info.info_text is not None:
                this_ax.text(
                    x=component_plot_info.info_text.x_pos,
                    y=component_plot_info.info_text.y_pos,
                    s=component_plot_info.info_text.text,
                    transform=this_ax.transAxes,
                    va="top",
                    linespacing=1.5,
                )

            if (component_plot_info.legend_info is not None) or (component_plot_info.info_text is not None):
                export(
                    fig=this_fig,
                    filename=this_output_key,
                    target_dir=output_dir_path,
                    close_figure=close_figures,
                )


# endregion
