"""
Fitter class for extraction of tagging efficiency correction factors.

Felix Metzner 2022
"""

import numpy as np
import pandas as pd

from typing import Optional, Tuple, Dict

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters.tag_cal_fitter_base import (
    TagCorrFitterBase,
    TagCalCorrIDType,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import FitSetupCollection
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet


__all__ = [
    "TagCorrectionFitter",
]


logger = log_to_default(name=__name__)


class TagCorrectionFitter(TagCorrFitterBase):

    this_fitter_setup: FitSetupTriplet = FitSetupCollection.tc_main__fs

    ordered_parameter_names: Tuple[str, ...] = (
        "bp_to_dwwos_l_yield",
        "bz_to_dwwos_l_yield",
        "charged_b_to_d_bkg_yield",
        "neutral_b_to_d_bkg_yield",
        "charged_b_to_ds_bkg_yield",
        "neutral_b_to_ds_bkg_yield",
        "bp_to_dsl_df_yield",
        "bz_to_dsl_df_yield",
    )

    bp_to_dwwos_l_yield_index: int = 0
    bz_to_dwwos_l_yield_index: int = 1

    charged_b_to_d_bkg_yield_index: int = 2
    neutral_b_to_d_bkg_yield_index: int = 3
    charged_b_to_ds_bkg_yield_index: int = 4
    neutral_b_to_ds_bkg_yield_index: int = 5

    bp_to_dsl_df_yield_index: int = 6
    bz_to_dsl_df_yield_index: int = 7

    end_yield_params_index: int = bz_to_dsl_df_yield_index

    main_floating_params_slice = np.s_[bp_to_dwwos_l_yield_index : end_yield_params_index + 1 : 1]

    number_of_hardcoded_floating_params: int = end_yield_params_index + 1

    fixed_params_indices: Tuple[int, ...] = tuple([])
    manual_nuisances_params_indices: Tuple[int, ...] = tuple([])

    background_indices: Tuple[int, ...] = (
        charged_b_to_d_bkg_yield_index,
        neutral_b_to_d_bkg_yield_index,
        charged_b_to_ds_bkg_yield_index,
        neutral_b_to_ds_bkg_yield_index,
    )

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        **kwargs,
    ) -> None:
        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=None,
            **kwargs,
        )

        assert len(self.fit_setup_info.components) == self.end_yield_params_index + 1, (
            [(c.comp_index, c.name) for c in self.fit_setup_info.components],
            len(self.fit_setup_info.components),
            self.end_yield_params_index + 1,
        )

        self.set_initial_parameters(values=self._calculate_initial_parameter_vector())

    # region Cost function
    def cost_fcn(self, params: np.ndarray) -> float:
        return self.chi2_fcn(params=params) + self.gaussian_constrains(params=params)

    def chi2_fcn(self, params: np.ndarray) -> float:
        _chi2 = np.sum(
            (self.expected_bin_counts(params=params) - self.data_bin_counts()) ** 2
            / (self.data_bin_stat_errors_squared + self.mc_bin_stat_errors_squared),
            axis=None,
        )
        assert isinstance(_chi2, float)
        return _chi2

    def gaussian_constrains(self, params: np.ndarray) -> float:
        # TODO!!!!!!
        return 0.0

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            self.get_yields(params=params) * self.local_efficiencies * self.template_shapes,
            axis=1,
        )

    # endregion

    # region Yields relations
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        return params[self.main_floating_params_slice]

    # endregion

    # region Parameter information
    @property
    def end_of_floating_params(self) -> int:
        _number_of_automated_nuisances: int = 0  # TODO!
        return self.number_of_hardcoded_floating_params + _number_of_automated_nuisances

    def _calculate_initial_parameter_vector(self) -> np.ndarray:

        # Yields
        _normalization_factors = self._calculate_normalization_factor_per_component(
            templates_per_ch=self.base_mc_bin_counts,
        )

        # Nuisances:
        automated_nuisances_initial_values: Tuple[float, ...] = tuple([])  # TODO

        _initial_params = np.array(
            [
                _normalization_factors[self.bp_to_dwwos_l_yield_index],
                _normalization_factors[self.bz_to_dwwos_l_yield_index],
                _normalization_factors[self.charged_b_to_d_bkg_yield_index],
                _normalization_factors[self.neutral_b_to_d_bkg_yield_index],
                _normalization_factors[self.charged_b_to_ds_bkg_yield_index],
                _normalization_factors[self.neutral_b_to_ds_bkg_yield_index],
                #
                _normalization_factors[self.bp_to_dsl_df_yield_index],
                _normalization_factors[self.bz_to_dsl_df_yield_index],
                #
                *automated_nuisances_initial_values,
            ]
        )

        assert len(_initial_params) == self.end_of_floating_params, (
            len(_initial_params),
            self.end_of_floating_params,
            _initial_params,
        )

        return _initial_params

    # endregion

    # region Correction application info
    @classmethod
    def get_correction_mapping(cls) -> Dict[TagCalCorrIDType, int]:
        df_comp_indices = (cls.bp_to_dsl_df_yield_index, cls.bz_to_dsl_df_yield_index)

        c_df_ids: Tuple[int, ...] = cls.this_fitter_setup.asimov.components[cls.bp_to_dsl_df_yield_index].sig_ids
        n_df_ids: Tuple[int, ...] = cls.this_fitter_setup.asimov.components[cls.bz_to_dsl_df_yield_index].sig_ids
        other_ids: Tuple[int, ...] = tuple(
            set(
                sig_id
                for c_id, c in enumerate(cls.this_fitter_setup.asimov.components)
                if c_id not in df_comp_indices
                for sig_id in c.sig_ids
            )
        )

        charged_b: TagCalCorrIDType = (RecoModeInfos.charged_reco_modes.ids, other_ids)
        neutral_b: TagCalCorrIDType = (RecoModeInfos.neutral_reco_modes.ids, other_ids)
        charged_b_to_d_df: TagCalCorrIDType = (RecoModeInfos.charged_b_to_d_with_dss_modes.ids, c_df_ids)
        neutral_b_to_d_df: TagCalCorrIDType = (RecoModeInfos.neutral_b_to_d_with_dss_modes.ids, n_df_ids)

        return {
            charged_b: cls.bp_to_dwwos_l_yield_index,
            neutral_b: cls.bz_to_dwwos_l_yield_index,
            charged_b_to_d_df: cls.bp_to_dsl_df_yield_index,
            neutral_b_to_d_df: cls.bz_to_dsl_df_yield_index,
        }

    @classmethod
    def get_simple_correction_mapping(cls) -> Dict[Tuple[int, ...], int]:
        return {
            RecoModeInfos.charged_b_to_d_with_dss_modes.ids: cls.bp_to_dwwos_l_yield_index,
            RecoModeInfos.neutral_b_to_d_with_dss_modes.ids: cls.bz_to_dwwos_l_yield_index,
        }

    @classmethod
    def get_full_correction_mapping(cls) -> Dict[TagCalCorrIDType, int]:
        df_comp_indices: Tuple[int, ...] = (cls.bp_to_dsl_df_yield_index, cls.bz_to_dsl_df_yield_index)
        bkg_comp_indices: Tuple[int, ...] = (
            cls.charged_b_to_d_bkg_yield_index,
            cls.neutral_b_to_d_bkg_yield_index,
            cls.charged_b_to_ds_bkg_yield_index,
            cls.neutral_b_to_ds_bkg_yield_index,
        )

        cb_to_ds_df_ids: Tuple[int, ...] = cls.this_fitter_setup.asimov.components[cls.bp_to_dsl_df_yield_index].sig_ids
        nb_to_ds_df_ids: Tuple[int, ...] = cls.this_fitter_setup.asimov.components[cls.bz_to_dsl_df_yield_index].sig_ids

        cb_to_d_bkg_ids = cls.this_fitter_setup.asimov.components[cls.charged_b_to_d_bkg_yield_index].sig_ids
        nb_to_d_bkg_ids = cls.this_fitter_setup.asimov.components[cls.neutral_b_to_d_bkg_yield_index].sig_ids
        cb_to_ds_bkg_ids = cls.this_fitter_setup.asimov.components[cls.charged_b_to_ds_bkg_yield_index].sig_ids
        nb_to_ds_bkg_ids = cls.this_fitter_setup.asimov.components[cls.neutral_b_to_ds_bkg_yield_index].sig_ids

        other_ids: Tuple[int, ...] = tuple(
            set(
                sig_id
                for c_id, c in enumerate(cls.this_fitter_setup.asimov.components)
                if (c_id not in df_comp_indices and c_id not in bkg_comp_indices)
                for sig_id in c.sig_ids
            )
        )

        charged_b_to_d: TagCalCorrIDType = (RecoModeInfos.charged_reco_modes.ids, other_ids)
        neutral_b_to_d: TagCalCorrIDType = (RecoModeInfos.neutral_reco_modes.ids, other_ids)
        charged_b_to_ds: TagCalCorrIDType = (RecoModeInfos.charged_b_to_ds_with_dss_modes.ids, other_ids)
        neutral_b_to_ds: TagCalCorrIDType = (RecoModeInfos.neutral_b_to_ds_with_dss_modes.ids, other_ids)

        charged_b_to_ds_df: TagCalCorrIDType = (RecoModeInfos.charged_b_to_d_with_dss_modes.ids, cb_to_ds_df_ids)
        neutral_b_to_ds_df: TagCalCorrIDType = (RecoModeInfos.neutral_b_to_d_with_dss_modes.ids, nb_to_ds_df_ids)

        charged_b_to_d_bkg: TagCalCorrIDType = (RecoModeInfos.charged_b_to_d_with_dss_modes.ids, cb_to_d_bkg_ids)
        neutral_b_to_d_bkg: TagCalCorrIDType = (RecoModeInfos.neutral_b_to_d_with_dss_modes.ids, nb_to_d_bkg_ids)
        charged_b_to_ds_bkg: TagCalCorrIDType = (RecoModeInfos.charged_b_to_ds_with_dss_modes.ids, cb_to_ds_bkg_ids)
        neutral_b_to_ds_bkg: TagCalCorrIDType = (RecoModeInfos.neutral_b_to_ds_with_dss_modes.ids, nb_to_ds_bkg_ids)

        return {
            charged_b_to_d: cls.bp_to_dwwos_l_yield_index,
            neutral_b_to_d: cls.bz_to_dwwos_l_yield_index,
            charged_b_to_ds: cls.bp_to_dwwos_l_yield_index,
            neutral_b_to_ds: cls.bz_to_dwwos_l_yield_index,
            charged_b_to_ds_df: cls.bp_to_dsl_df_yield_index,
            neutral_b_to_ds_df: cls.bz_to_dsl_df_yield_index,
            charged_b_to_d_bkg: cls.charged_b_to_d_bkg_yield_index,
            neutral_b_to_d_bkg: cls.neutral_b_to_d_bkg_yield_index,
            charged_b_to_ds_bkg: cls.charged_b_to_ds_bkg_yield_index,
            neutral_b_to_ds_bkg: cls.neutral_b_to_ds_bkg_yield_index,
        }

    @classmethod
    def get_correction_mapping_for_bkg(cls) -> Dict[TagCalCorrIDType, int]:
        cb_to_d_bkg_ids = cls.this_fitter_setup.asimov.components[cls.charged_b_to_d_bkg_yield_index].sig_ids
        nb_to_d_bkg_ids = cls.this_fitter_setup.asimov.components[cls.neutral_b_to_d_bkg_yield_index].sig_ids
        cb_to_ds_bkg_ids = cls.this_fitter_setup.asimov.components[cls.charged_b_to_ds_bkg_yield_index].sig_ids
        nb_to_ds_bkg_ids = cls.this_fitter_setup.asimov.components[cls.neutral_b_to_ds_bkg_yield_index].sig_ids

        charged_b_to_d_bkg: TagCalCorrIDType = (RecoModeInfos.charged_b_to_d_with_dss_modes.ids, cb_to_d_bkg_ids)
        neutral_b_to_d_bkg: TagCalCorrIDType = (RecoModeInfos.neutral_b_to_d_with_dss_modes.ids, nb_to_d_bkg_ids)
        charged_b_to_ds_bkg: TagCalCorrIDType = (RecoModeInfos.charged_b_to_ds_with_dss_modes.ids, cb_to_ds_bkg_ids)
        neutral_b_to_ds_bkg: TagCalCorrIDType = (RecoModeInfos.neutral_b_to_ds_with_dss_modes.ids, nb_to_ds_bkg_ids)

        return {
            charged_b_to_d_bkg: cls.bp_to_dwwos_l_yield_index,
            neutral_b_to_d_bkg: cls.bz_to_dwwos_l_yield_index,
            charged_b_to_ds_bkg: cls.bp_to_dwwos_l_yield_index,
            neutral_b_to_ds_bkg: cls.bz_to_dwwos_l_yield_index,
        }

    @classmethod
    def _get_tag_cal_corr_eta_values(cls) -> Dict[str, int]:
        raise NotImplementedError("The '_get_tag_cal_corr_eta_values' is WIP!")
        # eta_mapping = Dict[str, int] = {
        #     "Bz_to_Dm": ,
        #     "Bm_to_Dz": ,
        #     "Bz_to_Dsm": ,
        #     "Bm_to_Dsz": ,
        # }
        # return eta_mapping

    # endregion
