"""
Fitter class for extraction of tagging efficiency correction factors.

Felix Metzner 2022
"""

import numpy as np
import pandas as pd

from typing import Optional, Tuple, Dict

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.reco_modes.reco_mode_infos import TagRecoModeInfos

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters.tag_cal_fitter_base import (
    TagCorrFitterBase,
    TagCalCorrIDType,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import FitSetupCollection
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet


__all__ = [
    "TagCorrectionFitter",
]


logger = log_to_default(name=__name__)


class TagCorrectionFitter(TagCorrFitterBase):

    this_fitter_setup: FitSetupTriplet = FitSetupCollection.tc_ptm_charge_split_coupled_d__fs

    ordered_parameter_names: Tuple[str, ...] = (
        "tag_cb_to_d_1cpi_sig_yield",
        "tag_cb_to_d_mcpi_sig_yield",
        "tag_cb_to_ocd_pipi_sig_yield",
        "tag_cb_to_ds_1cpi_sig_yield",
        "tag_cb_to_ds_mcpi_sig_yield",
        "tag_cb_to_j_psi_sig_yield",
        "tag_cb_to_all_dd_sig_yield",
        #
        "tag_nb_to_d_1cpi_sig_yield",
        "tag_nb_to_d_mcpi_sig_yield",
        "tag_nb_to_ocd_pipi_sig_yield",
        "tag_nb_to_ds_1cpi_sig_yield",
        "tag_nb_to_ds_mcpi_sig_yield",
        "tag_nb_to_j_psi_sig_yield",
        "tag_nb_to_all_dd_sig_yield",
        #
        "tag_cb_to_d_1cpi_bkg_yield",
        "tag_cb_to_d_mcpi_bkg_yield",
        "tag_cb_to_ocd_pipi_bkg_yield",
        "tag_cb_to_ds_1cpi_bkg_yield",
        "tag_cb_to_ds_mcpi_bkg_yield",
        "tag_cb_to_j_psi_bkg_yield",
        "tag_cb_to_all_dd_bkg_yield",
        #
        "tag_nb_to_d_1cpi_bkg_yield",
        "tag_nb_to_d_mcpi_bkg_yield",
        "tag_nb_to_ocd_pipi_bkg_yield",
        "tag_nb_to_ds_1cpi_bkg_yield",
        "tag_nb_to_ds_mcpi_bkg_yield",
        "tag_nb_to_j_psi_bkg_yield",
        "tag_nb_to_all_dd_bkg_yield",
    )

    tag_cb_to_d_1cpi_sig_yield_index: int = 0
    tag_cb_to_d_mcpi_sig_yield_index: int = 1
    tag_cb_to_ocd_pipi_sig_yield_index: int = 2
    tag_cb_to_ds_1cpi_sig_yield_index: int = 3
    tag_cb_to_ds_mcpi_sig_yield_index: int = 4
    tag_cb_to_j_psi_sig_yield_index: int = 5
    tag_cb_to_all_dd_sig_yield_index: int = 6

    tag_nb_to_d_1cpi_sig_yield_index: int = 7
    tag_nb_to_d_mcpi_sig_yield_index: int = 8
    tag_nb_to_ocd_pipi_sig_yield_index: int = 9
    tag_nb_to_ds_1cpi_sig_yield_index: int = 10
    tag_nb_to_ds_mcpi_sig_yield_index: int = 11
    tag_nb_to_j_psi_sig_yield_index: int = 12
    tag_nb_to_all_dd_sig_yield_index: int = 13

    tag_cb_to_d_1cpi_bkg_yield_index: int = 14
    tag_cb_to_d_mcpi_bkg_yield_index: int = 15
    tag_cb_to_ocd_pipi_bkg_yield_index: int = 16
    tag_cb_to_ds_1cpi_bkg_yield_index: int = 17
    tag_cb_to_ds_mcpi_bkg_yield_index: int = 18
    tag_cb_to_j_psi_bkg_yield_index: int = 19
    tag_cb_to_all_dd_bkg_yield_index: int = 20

    tag_nb_to_d_1cpi_bkg_yield_index: int = 21
    tag_nb_to_d_mcpi_bkg_yield_index: int = 22
    tag_nb_to_ocd_pipi_bkg_yield_index: int = 23
    tag_nb_to_ds_1cpi_bkg_yield_index: int = 24
    tag_nb_to_ds_mcpi_bkg_yield_index: int = 25
    tag_nb_to_j_psi_bkg_yield_index: int = 26
    tag_nb_to_all_dd_bkg_yield_index: int = 27

    end_yield_params_index: int = tag_nb_to_all_dd_bkg_yield_index

    main_floating_params_slice = np.s_[tag_cb_to_d_1cpi_sig_yield_index : end_yield_params_index + 1 : 1]

    number_of_hardcoded_floating_params: int = end_yield_params_index + 1

    fixed_params_indices: Tuple[int, ...] = tuple([])
    manual_nuisances_params_indices: Tuple[int, ...] = tuple([])

    def __init__(
        self,
        *ignore,
        fit_setup_info: FitSetupInfoContainer,
        mc_df: pd.DataFrame,
        data_df: Optional[pd.DataFrame],
        **kwargs,
    ) -> None:
        super().__init__(
            *ignore,
            fit_setup_info=fit_setup_info,
            mc_df=mc_df,
            data_df=data_df,
            file_dict_dir_path=None,
            **kwargs,
        )

        assert len(self.fit_setup_info.components) == self.end_yield_params_index + 1, (
            [(c.comp_index, c.name) for c in self.fit_setup_info.components],
            len(self.fit_setup_info.components),
            self.end_yield_params_index + 1,
        )

        self.set_initial_parameters(values=self._calculate_initial_parameter_vector())

    # region Cost function
    def cost_fcn(self, params: np.ndarray) -> float:
        return self.chi2_fcn(params=params) + self.gaussian_constrains(params=params)

    def chi2_fcn(self, params: np.ndarray) -> float:
        _chi2 = np.sum(
            (self.expected_bin_counts(params=params) - self.data_bin_counts()) ** 2
            / (self.data_bin_stat_errors_squared + self.mc_bin_stat_errors_squared),
            axis=None,
        )
        assert isinstance(_chi2, float)
        return _chi2

    def gaussian_constrains(self, params: np.ndarray) -> float:
        # TODO!!!!!!
        return 0.0

    # endregion

    # region Expected bin count
    def expected_bin_counts(self, params: np.ndarray) -> np.ndarray:
        return np.sum(
            self.get_yields(params=params) * self.local_efficiencies * self.template_shapes,
            axis=1,
        )

    # endregion

    # region Yields relations
    def get_yields(self, params: np.ndarray) -> np.ndarray:
        return params[self.main_floating_params_slice]

    # endregion

    # region Parameter information
    @property
    def end_of_floating_params(self) -> int:
        _number_of_automated_nuisances: int = 0  # TODO!
        return self.number_of_hardcoded_floating_params + _number_of_automated_nuisances

    def _calculate_initial_parameter_vector(self) -> np.ndarray:

        # Yields
        _normalization_factors = self._calculate_normalization_factor_per_component(
            templates_per_ch=self.base_mc_bin_counts,
        )

        # Nuisances:
        automated_nuisances_initial_values: Tuple[float, ...] = tuple([])  # TODO

        _initial_params = np.array(
            [
                _normalization_factors[self.tag_cb_to_d_1cpi_sig_yield_index],
                _normalization_factors[self.tag_cb_to_d_mcpi_sig_yield_index],
                _normalization_factors[self.tag_cb_to_ocd_pipi_sig_yield_index],
                _normalization_factors[self.tag_cb_to_ds_1cpi_sig_yield_index],
                _normalization_factors[self.tag_cb_to_ds_mcpi_sig_yield_index],
                _normalization_factors[self.tag_cb_to_j_psi_sig_yield_index],
                _normalization_factors[self.tag_cb_to_all_dd_sig_yield_index],
                #
                _normalization_factors[self.tag_nb_to_d_1cpi_sig_yield_index],
                _normalization_factors[self.tag_nb_to_d_mcpi_sig_yield_index],
                _normalization_factors[self.tag_nb_to_ocd_pipi_sig_yield_index],
                _normalization_factors[self.tag_nb_to_ds_1cpi_sig_yield_index],
                _normalization_factors[self.tag_nb_to_ds_mcpi_sig_yield_index],
                _normalization_factors[self.tag_nb_to_j_psi_sig_yield_index],
                _normalization_factors[self.tag_nb_to_all_dd_sig_yield_index],
                #
                _normalization_factors[self.tag_cb_to_d_1cpi_bkg_yield_index],
                _normalization_factors[self.tag_cb_to_d_mcpi_bkg_yield_index],
                _normalization_factors[self.tag_cb_to_ocd_pipi_bkg_yield_index],
                _normalization_factors[self.tag_cb_to_ds_1cpi_bkg_yield_index],
                _normalization_factors[self.tag_cb_to_ds_mcpi_bkg_yield_index],
                _normalization_factors[self.tag_cb_to_j_psi_bkg_yield_index],
                _normalization_factors[self.tag_cb_to_all_dd_bkg_yield_index],
                #
                _normalization_factors[self.tag_nb_to_d_1cpi_bkg_yield_index],
                _normalization_factors[self.tag_nb_to_d_mcpi_bkg_yield_index],
                _normalization_factors[self.tag_nb_to_ocd_pipi_bkg_yield_index],
                _normalization_factors[self.tag_nb_to_ds_1cpi_bkg_yield_index],
                _normalization_factors[self.tag_nb_to_ds_mcpi_bkg_yield_index],
                _normalization_factors[self.tag_nb_to_j_psi_bkg_yield_index],
                _normalization_factors[self.tag_nb_to_all_dd_bkg_yield_index],
                #
                *automated_nuisances_initial_values,
            ]
        )

        assert len(_initial_params) == self.end_of_floating_params, (
            len(_initial_params),
            self.end_of_floating_params,
            _initial_params,
        )

        return _initial_params

    # endregion

    # region Correction application info
    @classmethod
    def get_correction_mapping(cls) -> Dict[TagCalCorrIDType, int]:

        ids: Tuple[int, ...] = tuple(set(sig_id for c in cls.this_fitter_setup.asimov.components for sig_id in c.sig_ids))

        return {
            (TagRecoModeInfos.charged_b_to_d_1cpi_tag_modes.ids, ids): cls.tag_cb_to_d_1cpi_sig_yield_index,
            (TagRecoModeInfos.charged_b_to_d_mcpi_tag_modes.ids, ids): cls.tag_cb_to_d_mcpi_sig_yield_index,
            (TagRecoModeInfos.charged_b_to_ocd_pipi_tag_modes.ids, ids): cls.tag_cb_to_ocd_pipi_sig_yield_index,
            (TagRecoModeInfos.charged_b_to_ds_1cpi_tag_modes.ids, ids): cls.tag_cb_to_ds_1cpi_sig_yield_index,
            (TagRecoModeInfos.charged_b_to_ds_mcpi_tag_modes.ids, ids): cls.tag_cb_to_ds_mcpi_sig_yield_index,
            (TagRecoModeInfos.charged_b_to_j_psi_tag_modes.ids, ids): cls.tag_cb_to_j_psi_sig_yield_index,
            (TagRecoModeInfos.charged_b_to__all_dd__tag_modes.ids, ids): cls.tag_cb_to_all_dd_sig_yield_index,
            #
            (TagRecoModeInfos.neutral_b_to_d_1cpi_tag_modes.ids, ids): cls.tag_nb_to_d_1cpi_sig_yield_index,
            (TagRecoModeInfos.neutral_b_to_d_mcpi_tag_modes.ids, ids): cls.tag_nb_to_d_mcpi_sig_yield_index,
            (TagRecoModeInfos.neutral_b_to_ocd_pipi_tag_modes.ids, ids): cls.tag_nb_to_ocd_pipi_sig_yield_index,
            (TagRecoModeInfos.neutral_b_to_ds_1cpi_tag_modes.ids, ids): cls.tag_nb_to_ds_1cpi_sig_yield_index,
            (TagRecoModeInfos.neutral_b_to_ds_mcpi_tag_modes.ids, ids): cls.tag_nb_to_ds_mcpi_sig_yield_index,
            (TagRecoModeInfos.neutral_b_to_j_psi_tag_modes.ids, ids): cls.tag_nb_to_j_psi_sig_yield_index,
            (TagRecoModeInfos.neutral_b_to__all_dd__tag_modes.ids, ids): cls.tag_nb_to_all_dd_sig_yield_index,
        }

    @classmethod
    def get_simple_correction_mapping(cls) -> Dict[Tuple[int, ...], int]:
        return {
            TagRecoModeInfos.charged_b_to_d_1cpi_tag_modes.ids: cls.tag_cb_to_d_1cpi_sig_yield_index,
            TagRecoModeInfos.charged_b_to_d_mcpi_tag_modes.ids: cls.tag_cb_to_d_mcpi_sig_yield_index,
            TagRecoModeInfos.charged_b_to_ocd_pipi_tag_modes.ids: cls.tag_cb_to_ocd_pipi_sig_yield_index,
            TagRecoModeInfos.charged_b_to_ds_1cpi_tag_modes.ids: cls.tag_cb_to_ds_1cpi_sig_yield_index,
            TagRecoModeInfos.charged_b_to_ds_mcpi_tag_modes.ids: cls.tag_cb_to_ds_mcpi_sig_yield_index,
            TagRecoModeInfos.charged_b_to_j_psi_tag_modes.ids: cls.tag_cb_to_j_psi_sig_yield_index,
            TagRecoModeInfos.charged_b_to__all_dd__tag_modes.ids: cls.tag_cb_to_all_dd_sig_yield_index,
            #
            TagRecoModeInfos.neutral_b_to_d_1cpi_tag_modes.ids: cls.tag_nb_to_d_1cpi_sig_yield_index,
            TagRecoModeInfos.neutral_b_to_d_mcpi_tag_modes.ids: cls.tag_nb_to_d_mcpi_sig_yield_index,
            TagRecoModeInfos.neutral_b_to_ocd_pipi_tag_modes.ids: cls.tag_nb_to_ocd_pipi_sig_yield_index,
            TagRecoModeInfos.neutral_b_to_ds_1cpi_tag_modes.ids: cls.tag_nb_to_ds_1cpi_sig_yield_index,
            TagRecoModeInfos.neutral_b_to_ds_mcpi_tag_modes.ids: cls.tag_nb_to_ds_mcpi_sig_yield_index,
            TagRecoModeInfos.neutral_b_to_j_psi_tag_modes.ids: cls.tag_nb_to_j_psi_sig_yield_index,
            TagRecoModeInfos.neutral_b_to__all_dd__tag_modes.ids: cls.tag_nb_to_all_dd_sig_yield_index,
        }

    @classmethod
    def _get_tag_cal_corr_eta_values(cls) -> Dict[str, int]:
        raise NotImplementedError("The '_get_tag_cal_corr_eta_values' is WIP!")
        # eta_mapping = Dict[str, int] = {
        #     "Bz_to_Dm": ,
        #     "Bm_to_Dz": ,
        #     "Bz_to_Dsm": ,
        #     "Bm_to_Dsz": ,
        # }
        # return eta_mapping

    # endregion
