"""
Setup definitions for MVA cut optimization

Felix Metzner 2021
"""

import numpy as np

from typing import Tuple, List, Dict

from rdstar.mva_selection.mva_selector_setups import MVASelectorSetups
from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.fitting.mva_cut_info import MVACutInfo, MVACutInfoEntry
from rdstar.offline_analysis.fitting.mva_cut_info import MVACutRecoModeNames as MvaCRmn


__all__ = [
    "MVACutSetups",
]


# base_setup
phase_1_values: List[float] = [round(x, 2) for x in np.linspace(0.2, 0.95, 16)]
phase_2_values: List[float] = [round(x, 3) for x in np.linspace(0.8, 0.82, 11)]
phase_3_values: List[float] = [round(x, 3) for x in np.linspace(0.806, 0.814, 9)]
base_probe_values: Tuple[float, ...] = tuple(set(phase_1_values + phase_2_values + phase_3_values))


# reco_mode_4d_setup
rm_2d_phase_1_values: List[float] = [round(x, 2) for x in np.linspace(0.2, 0.95, 16)]
rm_2d_phase_2_d_values: List[float] = [round(x, 3) for x in np.linspace(0.80, 0.82, 11)]
rm_2d_phase_2_dst_values: List[float] = [round(x, 3) for x in np.linspace(0.45, 0.65, 9)]
two_d_d_reco_mode_probe_values: Tuple[float, ...] = tuple(set(rm_2d_phase_1_values + rm_2d_phase_2_d_values))
two_d_dst_reco_mode_probe_values: Tuple[float, ...] = tuple(set(rm_2d_phase_1_values + rm_2d_phase_2_dst_values))
two_d_reco_mode_key_dict: Dict[Tuple[str, ...], Tuple[float, ...]] = {
    (MvaCRmn.reco_mode_DM, MvaCRmn.reco_mode_DZ): two_d_d_reco_mode_probe_values,
    (MvaCRmn.reco_mode_DSM, MvaCRmn.reco_mode_DSZ): two_d_dst_reco_mode_probe_values,
}
two_d_reco_mode_best_cut_values_dict: Dict[Tuple[str, ...], float] = {
    (MvaCRmn.reco_mode_DM, MvaCRmn.reco_mode_DZ): 0.808,
    (MvaCRmn.reco_mode_DSM, MvaCRmn.reco_mode_DSZ): 0.575,
}


# reco_mode_3d_setup
three_d_reco_mode_keys: Tuple[str, ...] = tuple(v.name for v in BSigRecoModes.main_modes.reco_mode_infos)
rm_3d_phase_1_values: List[float] = [round(x, 2) for x in np.linspace(0.2, 0.95, 16)]
rm_3d_phase_2_d_values: List[float] = [round(x, 3) for x in np.linspace(0.80, 0.82, 11)]
rm_3d_phase_2_dst0_values: List[float] = [round(x, 3) for x in np.linspace(0.45, 0.65, 9)]
rm_3d_phase_2_dstp_values: List[float] = [round(x, 3) for x in np.linspace(0.45, 0.65, 9)]

rm_3d_d_probe_values: Tuple[float, ...] = tuple(
    {
        *rm_3d_phase_1_values,
        # *rm_3d_phase_2_d_values,
    }
)
rm_3d_dst0_probe_values: Tuple[float, ...] = tuple(
    {
        *rm_3d_phase_1_values,
        # *rm_3d_phase_2_dst0_values,
    }
)
rm_3d_dstp_probe_values: Tuple[float, ...] = tuple(
    {
        *rm_3d_phase_1_values,
        # *rm_3d_phase_2_dstp_values,
    }
)

three_d_reco_mode_key_dict: Dict[Tuple[str, ...], Tuple[float, ...]] = {
    (MvaCRmn.reco_mode_DM, MvaCRmn.reco_mode_DZ): rm_3d_d_probe_values,
    (MvaCRmn.reco_mode_DSZ,): rm_3d_dst0_probe_values,
    (MvaCRmn.reco_mode_DSM,): rm_3d_dstp_probe_values,
}
three_d_reco_mode_best_cut_values_dict: Dict[Tuple[str, ...], float] = {
    (MvaCRmn.reco_mode_DM, MvaCRmn.reco_mode_DZ): 0.808,
    (MvaCRmn.reco_mode_DSZ,): 0.575,
    (MvaCRmn.reco_mode_DSM,): 0.575,
}


# reco_mode_4d_setup
four_d_reco_mode_keys: Tuple[str, ...] = tuple(v.name for v in BSigRecoModes.main_modes.reco_mode_infos)
# rm_4d_phase_1_values: List[float] = [0.7, 0.81]
rm_4d_phase_2_values: List[float] = [round(x, 2) for x in np.linspace(0.2, 0.95, 6)]
four_d_reco_mode_probe_values: Tuple[float, ...] = tuple(set(rm_4d_phase_2_values))


# Common information
_selection_name: str = "sec_sel_mva"
base_mva_col_name: str = MVASelectorSetups.get_selection_mva_setup(
    setup_name=_selection_name
).get_selection_mva_output_column_name()


class MVACutSetups:
    # Setup without distinction of D meson types
    base_setup: Tuple[MVACutInfoEntry, ...] = tuple(
        [
            MVACutInfoEntry(
                reco_modes=None,
                cut_variable=base_mva_col_name,
                cut_value=0.81,
                probe_values=base_probe_values,
            )
        ]
    )

    # Setup differentiating between D and D*
    reco_mode_2d_setup: Tuple[MVACutInfoEntry, ...] = tuple(
        [
            MVACutInfoEntry(
                reco_modes=reco_mode_key_tuple,
                cut_variable=base_mva_col_name,
                cut_value=two_d_reco_mode_best_cut_values_dict[reco_mode_key_tuple],
                probe_values=cut_values,
            )
            for reco_mode_key_tuple, cut_values in two_d_reco_mode_key_dict.items()
        ]
    )

    # Setup differentiating between D*+, D*0 and D
    reco_mode_3d_setup: Tuple[MVACutInfoEntry, ...] = tuple(
        [
            MVACutInfoEntry(
                reco_modes=reco_mode_key_tuple,
                cut_variable=base_mva_col_name,
                cut_value=three_d_reco_mode_best_cut_values_dict[reco_mode_key_tuple],
                probe_values=cut_values,
            )
            for reco_mode_key_tuple, cut_values in three_d_reco_mode_key_dict.items()
        ]
    )

    # Setup differentiating between all 4 D(*) meson types
    reco_mode_4d_setup: Tuple[MVACutInfoEntry, ...] = tuple(
        [
            MVACutInfoEntry(
                reco_modes=(reco_mode_key,),
                cut_variable=base_mva_col_name,
                cut_value=0.81,
                probe_values=four_d_reco_mode_probe_values,
            )
            for reco_mode_key in four_d_reco_mode_keys
        ]
    )

    @staticmethod
    def get_optimizer_additional_info_entry(setup: Tuple[MVACutInfoEntry, ...]) -> Dict[str, MVACutInfo]:
        return {MVACutInfo.probing_fit_case_setting_key: MVACutInfo(cut_infos=setup)}

    @staticmethod
    def get_application_additional_info_entry(setup: Tuple[MVACutInfoEntry, ...]) -> Dict[str, MVACutInfo]:
        return {MVACutInfo.fit_case_setting_key: MVACutSetups.get_best_cuts_info(setup=setup)}

    @staticmethod
    def get_best_cuts_info(setup: Tuple[MVACutInfoEntry, ...]) -> MVACutInfo:
        return MVACutInfo(
            cut_infos=tuple(
                MVACutInfoEntry(
                    reco_modes=e.reco_modes,
                    cut_variable=e.cut_variable,
                    cut_value=e.cut_value,
                    probe_values=None,
                )
                for e in setup
            )
        )
