"""
Provides fit models based on the TemplateFitter package and tailored to this analysis.

Felix Metzner 2019
"""


from typing import Dict, Tuple, List, Union, Optional, Generator

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.binned_distributions.binning import BinsInputType

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis import rdstar_samples

from rdstar.offline_analysis.systematics import full_systematics_cols
from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.weight_management.weight_information import WeightManager

from rdstar.offline_analysis.reco_modes.reco_mode_collections import AllBMesonRecoModes, BSigRecoModes

from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit import MCShiftInfo, first_test_shift_info
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    before_vf_mbc_res_cor_setup,
    lepton_phase_space_based_res_cor_setup,
)
from rdstar.offline_analysis.efficiency_calculation.rdstar_efficiency_definitions import (
    get_rdstar_sig_id_data_type_mapping,
)

from rdstar.offline_analysis.fitting.mva_cut_info_definitions import MVACutSetups
from rdstar.offline_analysis.fitting.rdstar_fit_variables import RDStarFitVariables
from rdstar.offline_analysis.fitting.fit_case import FitCase, FitVariable, FitVarDictType
from rdstar.offline_analysis.fitting.rdstar_template_setups import RDStarFitSetupContainers


__all__ = [
    "rdstar_fit_cases",
]


logger = log_to_default(name=__name__)


def replace_mm2_hv(
    fit_case: FitCase,
    new_mm2_col_name: str,
    tag_str: str,
    new_hist_var: Optional[HistVariable] = None,
    new_binning_by_mode_dict: Union[None, bool, Dict[str, BinsInputType]] = False,
    new_scope_by_mode_dict: Union[None, bool, Dict[str, Tuple[float, float]]] = False,
) -> FitCase:
    if new_hist_var is None:
        _original_h_var_dict: Dict[str, HistVariable] = {
            reco_mode_key: fvs[0].variable for reco_mode_key, fvs in fit_case.fit_variables_by_mode_dict.items()
        }
        assert len(set([hv.df_label for hv in _original_h_var_dict.values()])) == 1, [
            hv.df_label for hv in _original_h_var_dict.values()
        ]

        _new_h_var_dict: Dict[str, HistVariable] = {
            reco_mode_key: HistVariable(
                df_label=new_mm2_col_name,
                n_bins=org_hv.n_bins,
                scope=org_hv.scope,
                var_name=org_hv.variable_name,
                unit=org_hv.unit,
            )
            for reco_mode_key, org_hv in _original_h_var_dict.items()
        }
    else:
        assert new_mm2_col_name == new_hist_var.df_label, (new_mm2_col_name, new_hist_var.df_label)
        _new_h_var_dict = {reco_mode_key: new_hist_var for reco_mode_key in fit_case.fit_variables_by_mode_dict.keys()}

    assert not (isinstance(new_binning_by_mode_dict, bool) and new_binning_by_mode_dict is True), new_binning_by_mode_dict
    assert not (isinstance(new_scope_by_mode_dict, bool) and new_scope_by_mode_dict is True), new_scope_by_mode_dict

    _new_bd: Dict[str, Optional[BinsInputType]] = {
        k: new_binning_by_mode_dict if not isinstance(new_binning_by_mode_dict, dict) else new_binning_by_mode_dict[k]
        for k in fit_case.fit_variables_by_mode_dict.keys()
        if new_binning_by_mode_dict is not False
    }
    assert not any(isinstance(v, bool) for v in _new_bd.values()), _new_bd

    _new_sd: Dict[str, Optional[Tuple[float, float]]] = {
        k: new_scope_by_mode_dict if not isinstance(new_scope_by_mode_dict, dict) else new_scope_by_mode_dict[k]
        for k in fit_case.fit_variables_by_mode_dict.keys()
        if not isinstance(new_scope_by_mode_dict, bool)
    }
    assert not any(isinstance(v, bool) for v in _new_sd.values()), _new_sd

    fit_variables_dict: FitVarDictType = {
        reco_mode_key: tuple(
            [
                fit_var
                if _i != 0
                else FitVariable(
                    variable=new_hv,
                    alternative_binning=fit_var.alternative_binning
                    if new_binning_by_mode_dict is False
                    else _new_bd[reco_mode_key],
                    alternative_scope=fit_var.alternative_scope
                    if new_scope_by_mode_dict is False
                    else _new_sd[reco_mode_key],
                )
                for _i, fit_var in enumerate(fit_case.fit_variables_by_mode_dict[reco_mode_key])
            ]
        )
        for reco_mode_key, new_hv in _new_h_var_dict.items()
    }

    return FitCase.create_modified_fit_case(
        fit_case=fit_case,
        name=fit_case.name + "_" + tag_str.lower(),
        base_dir_name=fit_case.base_dir_name + "_" + tag_str,
        output_tag=fit_case.output_tag + "_" + tag_str,
        fit_variables_by_mode_dict=fit_variables_dict,
    )


def use_alternative_gap_modeling(
    fit_case: FitCase,
) -> FitCase:
    return FitCase.create_modified_fit_case(
        fit_case=fit_case,
        name=fit_case.name + "_alternative_gap",
        base_dir_name=fit_case.base_dir_name + "_Alternative_Gap",
        output_tag=fit_case.output_tag + "_Alternative_Gap",
        sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=True),
        mc_sample_keys=rdstar_samples.alternative_sample_keys,
    )


def get_bkg_fixed_fit_cases(
    fit_case: FitCase,
    create_alt_gap_cases: bool,
) -> Generator[FitCase, None, None]:
    __dss_keys: Tuple[str, ...] = ("DSSlbkg", "DSSl")
    __bb_bkg_keys: Tuple[str, ...] = ("bbBkg", "GBKG", "HBKG", "PiLNu")
    __continuum_keys: Tuple[str, ...] = ("CBKG", "continuum")

    if isinstance(fit_case.template_setup, dict):
        template_setup_list: List[str] = [k for v in fit_case.template_setup.values() for k in v.template_setup]
    else:
        template_setup_list = [k for k in fit_case.template_setup.template_setup]

    _dss_keys: List[str] = [k for k in template_setup_list if any(bk in k for bk in __dss_keys)]
    _bb_bkg_keys: List[str] = [k for k in template_setup_list if any(bk in k for bk in __bb_bkg_keys)]
    _continuum_keys: List[str] = [k for k in template_setup_list if any(bk in k for bk in __continuum_keys)]
    _all_bkg_keys: List[str] = _dss_keys + _bb_bkg_keys + _continuum_keys

    if _dss_keys:
        dss_fixed_fit_case = FitCase.create_modified_fit_case(
            fit_case=fit_case,
            name=fit_case.name + "_dssl_fixed",
            base_dir_name=fit_case.base_dir_name + "_DSSL_Fixed",
            output_tag=fit_case.output_tag + "_DSSL_Fixed",
            fix_component_yields=_dss_keys,
        )
        yield dss_fixed_fit_case
        if create_alt_gap_cases:
            yield use_alternative_gap_modeling(fit_case=dss_fixed_fit_case)
    else:
        logger.warning(
            f"No Dstst template key to be fixed for fit setup {fit_case.name} with template setup: {fit_case.template_setup}"
        )

    if _continuum_keys:
        continuum_fixed_fit_case = FitCase.create_modified_fit_case(
            fit_case=fit_case,
            name=fit_case.name + "_continuum_fixed",
            base_dir_name=fit_case.base_dir_name + "_Continuum_Fixed",
            output_tag=fit_case.output_tag + "_Continuum_Fixed",
            fix_component_yields=_continuum_keys,
        )
        yield continuum_fixed_fit_case
        if create_alt_gap_cases:
            yield use_alternative_gap_modeling(fit_case=continuum_fixed_fit_case)
    else:
        logger.warning(
            f"No continuum template key to be fixed for fit setup {fit_case.name} with template setup: {fit_case.template_setup}"
        )

    if _bb_bkg_keys:
        bb_bkg_fixed_fit_case = FitCase.create_modified_fit_case(
            fit_case=fit_case,
            name=fit_case.name + "_bb_bkg_fixed",
            base_dir_name=fit_case.base_dir_name + "_BB_Bkg_Fixed",
            output_tag=fit_case.output_tag + "_BB_Bkg_Fixed",
            fix_component_yields=_bb_bkg_keys,
        )
        yield bb_bkg_fixed_fit_case
        if create_alt_gap_cases:
            yield use_alternative_gap_modeling(fit_case=bb_bkg_fixed_fit_case)
    else:
        logger.warning(
            f"No bb bkg template key to be fixed for fit setup {fit_case.name} with template setup: {fit_case.template_setup}"
        )

    if _all_bkg_keys:
        all_bkg_fixed_fit_case = FitCase.create_modified_fit_case(
            fit_case=fit_case,
            name=fit_case.name + "_all_bkg_fixed",
            base_dir_name=fit_case.base_dir_name + "_All_BKG_Fixed",
            output_tag=fit_case.output_tag + "_All_BKG_Fixed",
            fix_component_yields=_all_bkg_keys,
        )
        yield all_bkg_fixed_fit_case
        if create_alt_gap_cases:
            yield use_alternative_gap_modeling(fit_case=all_bkg_fixed_fit_case)
    else:
        logger.warning(
            f"No background template key to be fixed for fit setup {fit_case.name} with template setup: {fit_case.template_setup}"
        )


#############
# Fit Cases #
#############

main_reco_mode_info_tuple = BSigRecoModes.main_modes.main_reco_mode_infos_as_tuple
main_reco_mode_latex_label_mapping = BSigRecoModes.main_modes.get_reco_mode_name_to_latex_label_mapping(short=False)

# Data Signal Window Fit Case

simple_signal_fit_case = FitCase(
    name="simple_signal_fit",
    base_dir_name="Fitting_Signal_Simple",
    sample_key="",
    output_tag="Fit_Signal_Simple_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

simple_signal_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=simple_signal_fit_case)


signal_fit_case = FitCase(
    name="signal_fit",
    base_dir_name="Fitting_Signal",
    sample_key="",
    output_tag="Fit_Signal_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.default_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.default_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.default_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.default_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

signal_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=signal_fit_case)


signal_wide_fit_case = FitCase(
    name="signal_wide_fit",
    base_dir_name="Fitting_Signal_Wide",
    sample_key="",
    output_tag="Fit_Signal_Wide_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_wide_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.default_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.default_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.default_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.default_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

signal_wide_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=signal_wide_fit_case)


# Signal Fit Cases with FEI Correction

simple_signal_with_fei_corr_fit_case = FitCase(
    name="simple_signal_with_fei_corr_fit",
    base_dir_name="Fitting_Signal_Simple_WithFEICorr",
    sample_key="",
    output_tag="Fit_Signal_Simple_WithFEICorr_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
)

simple_signal_with_fei_corr_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=simple_signal_with_fei_corr_fit_case)


# Signal Fit Cases with FEI Correction and Efficiency Correction

simple_signal_with_fei_and_eff_corr_fit_case = FitCase(
    name="simple_signal_with_fei_and_eff_corr_fit",
    base_dir_name="Fitting_Signal_Simple_WithFEIAndEffCorr",
    sample_key="",
    output_tag="Fit_Signal_Simple_WithFEIAndEffCorr_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    add_reco_mode_efficiency_correction=True,
)

simple_signal_with_fei_and_eff_corr_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_with_fei_and_eff_corr_fit_case
)


# Asimov Signal Window Fit Cases

simple_signal_asimov_fit_case = FitCase(
    name="simple_signal_asimov",
    base_dir_name="Fitting_AsimovSignal_Simple",
    sample_key="",
    output_tag="AsimovFit_Signal_Simple_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

simple_signal_alt_gap_asimov_fit_case = use_alternative_gap_modeling(fit_case=simple_signal_asimov_fit_case)


simple_signal_mva_optimization_asimov_fit_case = FitCase.create_modified_fit_case(
    fit_case=simple_signal_asimov_fit_case,
    name="simple_signal_asimov_mva_optimization",
    base_dir_name=simple_signal_asimov_fit_case.base_dir_name + "_MVA_Optimization",
    output_tag=simple_signal_asimov_fit_case.output_tag + "_MVA_Optimization",
    run_significance_evaluation=False,
    create_plots=False,
    create_projection_plots=False,
    create_template_plots=False,
    create_2d_template_plots=False,
    additional_info=MVACutSetups.get_optimizer_additional_info_entry(setup=MVACutSetups.base_setup),
)

simple_signal_mva_optimization_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_mva_optimization_asimov_fit_case,
)

simple_signal_2d_mva_optimization_asimov_fit_case = FitCase.create_modified_fit_case(
    fit_case=simple_signal_asimov_fit_case,
    name="simple_signal_asimov_2d_mva_optimization",
    base_dir_name=simple_signal_asimov_fit_case.base_dir_name + "_2D_MVA_Optimization",
    output_tag=simple_signal_asimov_fit_case.output_tag + "_2D_MVA_Optimization",
    run_significance_evaluation=False,
    create_plots=False,
    create_projection_plots=False,
    create_template_plots=False,
    create_2d_template_plots=False,
    additional_info=MVACutSetups.get_optimizer_additional_info_entry(setup=MVACutSetups.reco_mode_2d_setup),
)

simple_signal_2d_mva_optimization_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_2d_mva_optimization_asimov_fit_case,
)


simple_signal_2d_mva_best_cut_asimov_fit_case = FitCase.create_modified_fit_case(
    fit_case=simple_signal_asimov_fit_case,
    name="simple_signal_asimov_2d_mva_best_cut",
    base_dir_name=simple_signal_asimov_fit_case.base_dir_name + "_2D_MVA_Best_Cut",
    output_tag=simple_signal_asimov_fit_case.output_tag + "_2D_MVA_Best_Cut",
    additional_info=MVACutSetups.get_application_additional_info_entry(setup=MVACutSetups.reco_mode_2d_setup),
)

simple_signal_2d_mva_best_cut_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_2d_mva_best_cut_asimov_fit_case,
)


simple_signal_3d_mva_optimization_asimov_fit_case = FitCase.create_modified_fit_case(
    fit_case=simple_signal_asimov_fit_case,
    name="simple_signal_asimov_3d_mva_optimization",
    base_dir_name=simple_signal_asimov_fit_case.base_dir_name + "_3D_MVA_Optimization",
    output_tag=simple_signal_asimov_fit_case.output_tag + "_3D_MVA_Optimization",
    run_significance_evaluation=False,
    create_plots=False,
    create_projection_plots=False,
    create_template_plots=False,
    create_2d_template_plots=False,
    additional_info=MVACutSetups.get_optimizer_additional_info_entry(setup=MVACutSetups.reco_mode_3d_setup),
)

simple_signal_3d_mva_optimization_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_3d_mva_optimization_asimov_fit_case,
)


simple_signal_4d_mva_optimization_asimov_fit_case = FitCase.create_modified_fit_case(
    fit_case=simple_signal_asimov_fit_case,
    name="simple_signal_asimov_4d_mva_optimization",
    base_dir_name=simple_signal_asimov_fit_case.base_dir_name + "_4D_MVA_Optimization",
    output_tag=simple_signal_asimov_fit_case.output_tag + "_4D_MVA_Optimization",
    run_significance_evaluation=False,
    create_plots=False,
    create_projection_plots=False,
    create_template_plots=False,
    create_2d_template_plots=False,
    additional_info=MVACutSetups.get_optimizer_additional_info_entry(setup=MVACutSetups.reco_mode_4d_setup),
)

simple_signal_4d_mva_optimization_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_4d_mva_optimization_asimov_fit_case,
)


simple_signal_split_bkg_asimov_fit_case = FitCase(
    name="simple_signal_split_bkg_asimov",
    base_dir_name="Fitting_AsimovSignal_Simple_SplitBKG",
    sample_key="",
    output_tag="AsimovFit_Signal_Simple_SplitBKG_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_split_bkg_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_split_bkg_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_split_bkg_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.simple_signal_split_bkg_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_split_bkg_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.simple_signal_split_bkg_setup.sig_id_customizer,
    },
)

simple_signal_split_bkg_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_split_bkg_asimov_fit_case,
)


bin_adjust_signal_asimov_fit_case = FitCase(
    name="BIN_ADJUSTMENT_signal_asimov",
    base_dir_name="Fitting_AsimovSignal_BIN_ADJUSTMENT",
    sample_key="",
    output_tag="AsimovFit_Signal_FixedNuisanceParams_BIN_ADJUSTMENT",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.bin_adjust_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)


signal_asimov_fit_case = FitCase(
    name="signal_asimov",
    base_dir_name="Fitting_AsimovSignal",
    sample_key="",
    output_tag="AsimovFit_Signal_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.default_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.default_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.default_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.default_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

signal_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=signal_asimov_fit_case)


signal_wide_asimov_fit_case = FitCase(
    name="signal_wide_asimov",
    base_dir_name="Fitting_AsimovSignal_Wide",
    sample_key="",
    output_tag="AsimovFit_Signal_Wide_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_wide_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.default_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.default_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.default_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.default_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

signal_wide_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=signal_wide_asimov_fit_case)

complete_reco_mode_collection = AllBMesonRecoModes.b_to_all_d_l_modes_by_b_charge_d_excitation_and_pi_charge

complex_signal_asimov_fit_case = FitCase(
    name="complex_signal_asimov",
    base_dir_name="Fitting_AsimovSignal_Complex",
    sample_key="",
    output_tag="AsimovFit_Signal_Complex_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.complex_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=complete_reco_mode_collection.main_reco_mode_infos_as_tuple,
    template_setup=RDStarFitSetupContainers.complex_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.complex_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.complex_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=complete_reco_mode_collection.get_reco_mode_name_to_latex_label_mapping(short=False),
    alternative_template_colors=RDStarFitSetupContainers.complex_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

complex_signal_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=complex_signal_asimov_fit_case)


complex_signal_mva_optimization_asimov_fit_case = FitCase.create_modified_fit_case(
    fit_case=complex_signal_asimov_fit_case,
    name="complex_signal_asimov_mva_optimization",
    base_dir_name=complex_signal_asimov_fit_case.base_dir_name + "_MVA_Optimization",
    output_tag=complex_signal_asimov_fit_case.output_tag + "_MVA_Optimization",
    run_significance_evaluation=False,
    create_plots=False,
    create_projection_plots=False,
    create_template_plots=False,
    create_2d_template_plots=False,
    additional_info=MVACutSetups.get_optimizer_additional_info_entry(setup=MVACutSetups.base_setup),
)
complex_signal_mva_optimization_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=complex_signal_mva_optimization_asimov_fit_case,
)


simple_signal_e_miss_p_miss_asimov_fit_case = replace_mm2_hv(
    fit_case=simple_signal_asimov_fit_case,
    new_mm2_col_name=RDStarHistVars.e_miss_minus_p_miss_res_corr_signal_var.df_label,
    tag_str="e_miss_minus_p_miss_res_corr",
    new_hist_var=RDStarHistVars.e_miss_minus_p_miss_res_corr_signal_var,
    new_binning_by_mode_dict=RDStarFitVariables.e_miss_minus_p_miss_binning_dict_signal_corr,
    new_scope_by_mode_dict=False,
)
simple_signal_e_miss_p_miss_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_e_miss_p_miss_asimov_fit_case,
)


simple_signal_mms_before_vf_mbc_res_cor_asimov_fit_case = replace_mm2_hv(
    fit_case=simple_signal_asimov_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=before_vf_mbc_res_cor_setup.name,
        use_laplace_smeared_version=False,
    ),
    tag_str="mms_before_vf_mbc_res_cor",
)
simple_signal_mms_before_vf_mbc_res_cor_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_mms_before_vf_mbc_res_cor_asimov_fit_case,
)


signal_3d_asimov_fit_case = FitCase(
    name="signal_3d_asimov",
    base_dir_name="Fitting_AsimovSignal_3D",
    sample_key="",
    output_tag="AsimovFit_Signal_3D_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.three_d_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)
signal_3d_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=signal_3d_asimov_fit_case)


signal_asimov_fit_case_scipy = FitCase(
    name="signal_asimov_scipy",
    base_dir_name="Fitting_AsimovSignal_Scipy",
    sample_key="",
    output_tag="AsimovFit_Signal_FixedNuisanceParams_Scipy",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.default_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.default_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.default_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.default_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    minimizer_id="scipy",
)


# Asimov Signal Fit Cases with FEI Correction

simple_signal_with_fei_corr_asimov_fit_case = FitCase(
    name="simple_signal_with_fei_corr_asimov",
    base_dir_name="Fitting_AsimovSignal_Simple_WithFEICorr",
    sample_key="",
    output_tag="AsimovFit_Signal_Simple_WithFEICorr_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
)

simple_signal_with_fei_corr_alt_gap_asimov_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_with_fei_corr_asimov_fit_case
)

# Asimov Signal Fit Cases with FEI Correction and Efficiency Correction

simple_signal_with_fei_and_eff_corr_asimov_fit_case = FitCase(
    name="simple_signal_with_fei_and_eff_corr_asimov",
    base_dir_name="Fitting_AsimovSignal_Simple_WithFEIAndEffCorr",
    sample_key="",
    output_tag="AsimovFit_Signal_Simple_WithFEIAndEffCorr_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_signal_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.simple_signal_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.simple_signal_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.simple_signal_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.simple_signal_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    add_reco_mode_efficiency_correction=True,
)

simple_signal_with_fei_and_eff_corr_alt_gap_asimov_fit_case = use_alternative_gap_modeling(
    fit_case=simple_signal_with_fei_and_eff_corr_asimov_fit_case
)


# Data Q2-Sideband Fit Cases

q2_sideband_fit_case = FitCase(
    name="q2_sideband_fit",
    base_dir_name="Fitting_Q2Sideband",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

q2_sideband_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_fit_case)

q2_sideband_fit_case_with_old_res_corr = replace_mm2_hv(
    fit_case=q2_sideband_fit_case,
    new_mm2_col_name=RDStarHistVars.missing_mass_res_corr_q2sb_fit_with_old_res_corr.df_label,
    tag_str="with_old_res_corr",
)
q2_sideband_alt_gap_fit_case_with_old_res_corr = use_alternative_gap_modeling(
    fit_case=q2_sideband_fit_case_with_old_res_corr
)


q2_sideband_split_d_star_modes_fit_case = FitCase(
    name="q2_sideband_split_d_star_modes",
    base_dir_name="Fitting_Q2SidebandSplitDStarModes",
    sample_key="q2_sideband",
    output_tag="Fit_SplitDStarModes_q2Sideband_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_setup.sig_id_customizer,
    },
)

q2_sideband_split_d_star_modes_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_split_d_star_modes_fit_case
)

tmp_q2_sideband_split_d_star_modes_alt_gap_fit_case_with_old_res_corr = replace_mm2_hv(
    fit_case=q2_sideband_split_d_star_modes_alt_gap_fit_case,
    new_mm2_col_name=RDStarHistVars.missing_mass_res_corr_q2sb_fit_with_old_res_corr.df_label,
    tag_str="with_old_res_corr",
)

q2_sideband_split_d_star_modes_inverted_fit_case = FitCase(
    name="q2_sideband_split_d_star_modes_inverted",
    base_dir_name="Fitting_Q2SidebandSplitDStarModesInverted",
    sample_key="q2_sideband",
    output_tag="Fit_SplitDStarModesInverted_q2Sideband_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.sig_id_customizer,
    },
)

q2_sideband_split_d_star_modes_inverted_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_split_d_star_modes_inverted_fit_case
)

# Testing mm2 shift in fits on data:
q2_sideband_fit_case_with_mm2_shift = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_fit_case,
    name=q2_sideband_fit_case.name + "_with_mm2_shift",
    base_dir_name=q2_sideband_fit_case.base_dir_name + "_WithMM2Shift",
    output_tag=q2_sideband_fit_case.output_tag + "_WithMM2Shift",
    additional_info={MCShiftInfo.extra_info_key(): first_test_shift_info},
)
q2_sideband_alt_gap_fit_case_with_mm2_shift = use_alternative_gap_modeling(fit_case=q2_sideband_fit_case_with_mm2_shift)


q2_sideband_split_d_star_modes_fit_case_with_mm2_shift = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_split_d_star_modes_fit_case,
    name=q2_sideband_split_d_star_modes_fit_case.name + "_with_mm2_shift",
    base_dir_name=q2_sideband_split_d_star_modes_fit_case.base_dir_name + "_WithMM2Shift",
    output_tag=q2_sideband_split_d_star_modes_fit_case.output_tag + "_WithMM2Shift",
    additional_info={
        **q2_sideband_split_d_star_modes_fit_case.additional_info,
        MCShiftInfo.extra_info_key(): first_test_shift_info,
    },
)
q2_sideband_split_d_star_modes_alt_gap_fit_case_with_mm2_shift = use_alternative_gap_modeling(
    fit_case=q2_sideband_split_d_star_modes_fit_case_with_mm2_shift
)

q2_sideband_split_d_star_modes_inverted_fit_case_with_mm2_shift = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_split_d_star_modes_inverted_fit_case,
    name=q2_sideband_split_d_star_modes_inverted_fit_case.name + "_with_mm2_shift",
    base_dir_name=q2_sideband_split_d_star_modes_inverted_fit_case.base_dir_name + "_WithMM2Shift",
    output_tag=q2_sideband_split_d_star_modes_inverted_fit_case.output_tag + "_WithMM2Shift",
    additional_info={
        **q2_sideband_split_d_star_modes_inverted_fit_case.additional_info,
        MCShiftInfo.extra_info_key(): first_test_shift_info,
    },
)
q2_sideband_split_d_star_modes_inverted_alt_gap_fit_case_with_mm2_shift = use_alternative_gap_modeling(
    fit_case=q2_sideband_split_d_star_modes_inverted_fit_case_with_mm2_shift
)


# Turning on Nuisance Parameters for stat only

q2_sideband_fit_case_with_stat_nuisance = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_fit_case,
    name=q2_sideband_fit_case.name + "_with_stat_nuisance",
    base_dir_name=q2_sideband_fit_case.base_dir_name + "_WithStatNuisance",
    output_tag=q2_sideband_fit_case.output_tag + "_WithStatNuisance",
    fix_nuisance_parameters=False,
    systematics_cols=tuple([]),
)
q2_sideband_alt_gap_fit_case_with_stat_nuisance = use_alternative_gap_modeling(
    fit_case=q2_sideband_fit_case_with_stat_nuisance
)


q2_sideband_fit_case_with_mm2_shift_with_stat_nuisance = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_fit_case_with_mm2_shift,
    name=q2_sideband_fit_case_with_mm2_shift.name + "_with_stat_nuisance",
    base_dir_name=q2_sideband_fit_case_with_mm2_shift.base_dir_name + "_WithStatNuisance",
    output_tag=q2_sideband_fit_case_with_mm2_shift.output_tag + "_WithStatNuisance",
    fix_nuisance_parameters=False,
    systematics_cols=tuple([]),
)
q2_sideband_alt_gap_fit_case_with_mm2_shift_with_stat_nuisance = use_alternative_gap_modeling(
    fit_case=q2_sideband_fit_case_with_mm2_shift_with_stat_nuisance
)

# Test with bin statistics nuisances and simplified binning

q2_sideband_fit_case_with_stat_nuisance_5by5binning = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_fit_case,
    name=q2_sideband_fit_case.name + "_with_stat_nuisance_5by5_binning",
    base_dir_name=q2_sideband_fit_case.base_dir_name + "_WithStatNuisance5by5Binning",
    output_tag=q2_sideband_fit_case.output_tag + "_WithStatNuisance5by5Binning",
    fit_variables_by_mode_dict=RDStarFitVariables.five_bin_q2sb_fit_with_stat_nuisances_var_dict,
    fix_nuisance_parameters=False,
    systematics_cols=tuple([]),
)

q2_sideband_fit_case_with_all_nuisance_5by5binning = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_fit_case,
    name=q2_sideband_fit_case.name + "_with_all_nuisance_5by5_binning",
    base_dir_name=q2_sideband_fit_case.base_dir_name + "_WithAllNuisance5by5Binning",
    output_tag=q2_sideband_fit_case.output_tag + "_WithAllNuisance5by5Binning",
    fit_variables_by_mode_dict=RDStarFitVariables.five_bin_q2sb_fit_with_stat_nuisances_var_dict,
    fix_nuisance_parameters=False,
)


# Other Q2 Sideband Fit Setups

q2_sideband_mm2_lep_ph_spc_fit_case = replace_mm2_hv(
    fit_case=q2_sideband_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=lepton_phase_space_based_res_cor_setup.name,
        use_laplace_smeared_version=False,
    ),
    tag_str="LepPhSpc",
)
q2_sideband_mm2_lep_ph_spc_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_mm2_lep_ph_spc_fit_case)


q2_sideband_mms_before_vf_mbc_res_cor_fit_case = replace_mm2_hv(
    fit_case=q2_sideband_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=before_vf_mbc_res_cor_setup.name,
        use_laplace_smeared_version=False,
    ),
    tag_str="mms_before_vf_mbc_res_cor",
)
q2_sideband_mms_before_vf_mbc_res_cor_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_mms_before_vf_mbc_res_cor_fit_case,
)


q2_sideband_e_miss_p_miss_fit_case = replace_mm2_hv(
    fit_case=q2_sideband_fit_case,
    new_mm2_col_name=RDStarHistVars.e_miss_minus_p_miss_res_corr_q2sb_var.df_label,
    tag_str="e_miss_minus_p_miss_res_corr",
    new_hist_var=RDStarHistVars.e_miss_minus_p_miss_res_corr_q2sb_var,
    new_binning_by_mode_dict=RDStarFitVariables.e_miss_minus_p_miss_wide_binning_dict_q2sb_corr,
    new_scope_by_mode_dict=False,
)
q2_sideband_e_miss_p_miss_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_e_miss_p_miss_fit_case)


q2_sideband_wide_fit_case = FitCase(
    name="q2_sideband_wide_fit",
    base_dir_name="Fitting_Q2Sideband_Wide",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_Wide_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_wide_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

q2_sideband_wide_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_wide_fit_case)

q2_sideband_wide_split_d_star_modes_fit_case = FitCase(
    name="q2_sideband_wide_split_d_star_modes",
    base_dir_name="Fitting_Q2Sideband_Wide_SplitDStarModes",
    sample_key="q2_sideband",
    output_tag="Fit_SplitDStarModes_q2Sideband_Wide_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_wide_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_setup.sig_id_customizer,
    },
)

q2_sideband_wide_split_d_star_modes_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_wide_split_d_star_modes_fit_case
)

q2_sideband_wide_split_d_star_modes_inverted_fit_case = FitCase(
    name="q2_sideband_wide_split_d_star_modes_inverted",
    base_dir_name="Fitting_Q2Sideband_Wide_SplitDStarModesInverted",
    sample_key="q2_sideband",
    output_tag="Fit_SplitDStarModesInverted_q2Sideband_Wide_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_wide_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.sig_id_customizer,
    },
)

q2_sideband_wide_split_d_star_modes_inverted_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_wide_split_d_star_modes_inverted_fit_case
)


q2_sideband_1d_mms_fit_case = FitCase(
    name="q2_sideband_1d_mms_fit",
    base_dir_name="Fitting_Q2Sideband_1D_MMS",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_FixedNuisanceParams_1D_MMS",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.one_d_mms_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)
q2_sideband_1d_mms_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_1d_mms_fit_case)


q2_sideband_1d_mms_before_vf_fit_case = replace_mm2_hv(
    fit_case=q2_sideband_1d_mms_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=before_vf_mbc_res_cor_setup.name,
        use_laplace_smeared_version=False,
    ),
    tag_str="mms_before_vf_mbc_res_cor",
)
q2_sideband_1d_mms_before_vf_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_1d_mms_before_vf_fit_case
)


q2_sideband_1d_e_miss_p_miss_fit_case = FitCase(
    name="q2_sideband_1d_e_miss_p_miss_fit",
    base_dir_name="Fitting_Q2Sideband_1D_E_Miss_Minus_P_Miss",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_FixedNuisanceParams_1D_E_Miss_Minus_P_Miss",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.one_d_e_miss_minus_p_miss_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)
q2_sideband_1d_e_miss_p_miss_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_1d_e_miss_p_miss_fit_case
)


# Data Q2-Sideband Fit Cases with FEI Correction

q2_sideband_with_fei_corr_fit_case = FitCase(
    name="q2_sideband_with_fei_corr_fit",
    base_dir_name="Fitting_Q2Sideband_WithFEICorr",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_WithFEICorr_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
)

q2_sideband_with_fei_corr_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_with_fei_corr_fit_case)


q2_sideband_with_fei_corr_split_d_star_modes_fit_case = FitCase(
    name="q2_sideband_with_fei_corr_split_d_star_modes",
    base_dir_name="Fitting_Q2Sideband_WithFEICorr_SplitDStarModes",
    sample_key="q2_sideband",
    output_tag="Fit_SplitDStarModes_q2Sideband_WithFEICorr_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_setup.sig_id_customizer,
    },
)

q2_sideband_with_fei_corr_split_d_star_modes_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_corr_split_d_star_modes_fit_case
)


q2_sideband_with_fei_corr_split_d_star_modes_inverted_fit_case = FitCase(
    name="q2_sideband_with_fei_corr_split_d_star_modes_inverted",
    base_dir_name="Fitting_Q2Sideband_WithFEICorr_SplitDStarModesInverted",
    sample_key="q2_sideband",
    output_tag="Fit_SplitDStarModesInverted_q2Sideband_WithFEICorr_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.sig_id_customizer,
    },
)

q2_sideband_with_fei_corr_split_d_star_modes_inverted_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_corr_split_d_star_modes_inverted_fit_case
)


# Data Q2-Sideband Fit Cases with FEI Correction and Efficiency Correction

q2_sideband_with_fei_and_eff_corr_fit_case = FitCase(
    name="q2_sideband_with_fei_and_eff_corr_fit",
    base_dir_name="Fitting_Q2Sideband_WithFEIAndEffCorr",
    sample_key="q2_sideband",
    output_tag="Fit_q2Sideband_WithFEIAndEffCorr_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    add_reco_mode_efficiency_correction=True,
)

q2_sideband_with_fei_and_eff_corr_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_and_eff_corr_fit_case
)


q2_sideband_with_fei_and_eff_corr_split_d_star_modes_fit_case = FitCase(
    name="q2_sideband_with_fei_and_eff_corr_split_d_star_modes",
    base_dir_name="Fitting_Q2Sideband_WithFEIAndEffCorr_SplitDStarModes",
    sample_key="q2_sideband",
    output_tag="Fit_SplitDStarModes_q2Sideband_WithFEIAndEffCorr_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    add_reco_mode_efficiency_correction=True,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_setup.sig_id_customizer,
    },
)

q2_sideband_with_fei_and_eff_corr_split_d_star_modes_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_and_eff_corr_split_d_star_modes_fit_case
)


q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_fit_case = FitCase(
    name="q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted",
    base_dir_name="Fitting_Q2Sideband_WithFEIAndEffCorr_SplitDStarModesInverted",
    sample_key="q2_sideband",
    output_tag="Fit_SplitDStarModesInverted_q2Sideband_WithFEIAndEffCorr_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    add_reco_mode_efficiency_correction=True,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.sig_id_customizer,
    },
)

q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_fit_case
)


# Asimov Q2-Sideband Fit Cases

q2_sideband_asimov_fit_case = FitCase(
    name="q2_sideband_asimov",
    base_dir_name="Fitting_AsimovQ2Sideband",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

q2_sideband_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_asimov_fit_case)

q2_sideband_asimov_fit_case_with_old_res_corr = replace_mm2_hv(
    fit_case=q2_sideband_asimov_fit_case,
    new_mm2_col_name=RDStarHistVars.missing_mass_res_corr_q2sb_fit_with_old_res_corr.df_label,
    tag_str="with_old_res_corr",
)
q2_sideband_asimov_alt_gap_fit_case_with_old_res_corr = use_alternative_gap_modeling(
    fit_case=q2_sideband_asimov_fit_case_with_old_res_corr
)


q2_sideband_split_d_star_modes_asimov_fit_case = FitCase(
    name="q2_sideband_asimov_split_d_star_modes",
    base_dir_name="Fitting_AsimovQ2SidebandSplitDStarModes",
    sample_key="q2_sideband",
    output_tag="AsimovFit_SplitDStarModes_q2Sideband_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_setup.sig_id_customizer,
    },
)

q2_sideband_split_d_star_modes_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_split_d_star_modes_asimov_fit_case
)

q2_sideband_split_d_star_modes_inverted_asimov_fit_case = FitCase(
    name="q2_sideband_asimov_split_d_star_modes_inverted",
    base_dir_name="Fitting_AsimovQ2SidebandSplitDStarModesInverted",
    sample_key="q2_sideband",
    output_tag="AsimovFit_SplitDStarModesInverted_q2Sideband_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.sig_id_customizer,
    },
)

q2_sideband_split_d_star_modes_inverted_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_split_d_star_modes_inverted_asimov_fit_case
)


# Testing mm2 shift in fits on Asimov:
q2_sideband_asimov_fit_case_with_mm2_shift = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_asimov_fit_case,
    name=q2_sideband_asimov_fit_case.name + "_with_mm2_shift",
    base_dir_name=q2_sideband_asimov_fit_case.base_dir_name + "_WithMM2Shift",
    output_tag=q2_sideband_asimov_fit_case.output_tag + "_WithMM2Shift",
    additional_info={MCShiftInfo.extra_info_key(): first_test_shift_info},
)
q2_sideband_asimov_alt_gap_fit_case_with_mm2_shift = use_alternative_gap_modeling(
    fit_case=q2_sideband_asimov_fit_case_with_mm2_shift
)


q2_sideband_split_d_star_modes_asimov_fit_case_with_mm2_shift = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_split_d_star_modes_asimov_fit_case,
    name=q2_sideband_split_d_star_modes_asimov_fit_case.name + "_with_mm2_shift",
    base_dir_name=q2_sideband_split_d_star_modes_asimov_fit_case.base_dir_name + "_WithMM2Shift",
    output_tag=q2_sideband_split_d_star_modes_asimov_fit_case.output_tag + "_WithMM2Shift",
    additional_info={
        **q2_sideband_split_d_star_modes_asimov_fit_case.additional_info,
        MCShiftInfo.extra_info_key(): first_test_shift_info,
    },
)
q2_sideband_split_d_star_modes_asimov_alt_gap_fit_case_with_mm2_shift = use_alternative_gap_modeling(
    fit_case=q2_sideband_split_d_star_modes_asimov_fit_case_with_mm2_shift
)

q2_sideband_split_d_star_modes_inverted_asimov_fit_case_with_mm2_shift = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_split_d_star_modes_inverted_asimov_fit_case,
    name=q2_sideband_split_d_star_modes_inverted_asimov_fit_case.name + "_with_mm2_shift",
    base_dir_name=q2_sideband_split_d_star_modes_inverted_asimov_fit_case.base_dir_name + "_WithMM2Shift",
    output_tag=q2_sideband_split_d_star_modes_inverted_asimov_fit_case.output_tag + "_WithMM2Shift",
    additional_info={
        **q2_sideband_split_d_star_modes_inverted_asimov_fit_case.additional_info,
        MCShiftInfo.extra_info_key(): first_test_shift_info,
    },
)
q2_sideband_split_d_star_modes_inverted_asimov_alt_gap_fit_case_with_mm2_shift = use_alternative_gap_modeling(
    fit_case=q2_sideband_split_d_star_modes_inverted_asimov_fit_case_with_mm2_shift
)


# Turning on Nuisance Parameters for stat only

q2_sideband_asimov_fit_case_with_stat_nuisance = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_asimov_fit_case,
    name=q2_sideband_asimov_fit_case.name + "_with_stat_nuisance",
    base_dir_name=q2_sideband_asimov_fit_case.base_dir_name + "_WithStatNuisance",
    output_tag=q2_sideband_asimov_fit_case.output_tag + "_WithStatNuisance",
    fix_nuisance_parameters=False,
    systematics_cols=tuple([]),
)
q2_sideband_alt_gap_asimov_fit_case_with_stat_nuisance = use_alternative_gap_modeling(
    fit_case=q2_sideband_asimov_fit_case_with_stat_nuisance
)


q2_sideband_asimov_fit_case_with_mm2_shift_with_stat_nuisance = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_asimov_fit_case_with_mm2_shift,
    name=q2_sideband_asimov_fit_case_with_mm2_shift.name + "_with_stat_nuisance",
    base_dir_name=q2_sideband_asimov_fit_case_with_mm2_shift.base_dir_name + "_WithStatNuisance",
    output_tag=q2_sideband_asimov_fit_case_with_mm2_shift.output_tag + "_WithStatNuisance",
    fix_nuisance_parameters=False,
    systematics_cols=tuple([]),
)
q2_sideband_alt_gap_asimov_fit_case_with_mm2_shift_with_stat_nuisance = use_alternative_gap_modeling(
    fit_case=q2_sideband_asimov_fit_case_with_mm2_shift_with_stat_nuisance
)

# Test with bin statistics nuisances and simplified binning

q2_sideband_asimov_fit_case_with_stat_nuisance_5by5binning = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_asimov_fit_case,
    name=q2_sideband_asimov_fit_case.name + "_with_stat_nuisance_5by5_binning",
    base_dir_name=q2_sideband_asimov_fit_case.base_dir_name + "_WithStatNuisance5by5Binning",
    output_tag=q2_sideband_asimov_fit_case.output_tag + "_WithStatNuisance5by5Binning",
    fit_variables_by_mode_dict=RDStarFitVariables.five_bin_q2sb_fit_with_stat_nuisances_var_dict,
    fix_nuisance_parameters=False,
    systematics_cols=tuple([]),
)

q2_sideband_asimov_fit_case_with_all_nuisance_5by5binning = FitCase.create_modified_fit_case(
    fit_case=q2_sideband_asimov_fit_case,
    name=q2_sideband_asimov_fit_case.name + "_with_all_nuisance_5by5_binning",
    base_dir_name=q2_sideband_asimov_fit_case.base_dir_name + "_WithAllNuisance5by5Binning",
    output_tag=q2_sideband_asimov_fit_case.output_tag + "_WithAllNuisance5by5Binning",
    fit_variables_by_mode_dict=RDStarFitVariables.five_bin_q2sb_fit_with_stat_nuisances_var_dict,
    fix_nuisance_parameters=False,
)


# Other Asimov Q2 Sideband Fit Setups
q2_sideband_mm2_lep_ph_spc_asimov_fit_case = replace_mm2_hv(
    fit_case=q2_sideband_asimov_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=lepton_phase_space_based_res_cor_setup.name,
        use_laplace_smeared_version=False,
    ),
    tag_str="LepPhSpc",
)

q2_sideband_mm2_lep_ph_spc_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_mm2_lep_ph_spc_asimov_fit_case,
)


q2_sideband_mms_before_vf_mbc_res_cor_asimov_fit_case = replace_mm2_hv(
    fit_case=q2_sideband_asimov_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=before_vf_mbc_res_cor_setup.name,
        use_laplace_smeared_version=False,
    ),
    tag_str="mms_before_vf_mbc_res_cor",
)
q2_sideband_mms_before_vf_mbc_res_cor_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_mms_before_vf_mbc_res_cor_asimov_fit_case,
)


q2_sideband_e_miss_p_miss_asimov_fit_case = replace_mm2_hv(
    fit_case=q2_sideband_asimov_fit_case,
    new_mm2_col_name=RDStarHistVars.e_miss_minus_p_miss_res_corr_q2sb_var.df_label,
    tag_str="e_miss_minus_p_miss_res_corr",
    new_hist_var=RDStarHistVars.e_miss_minus_p_miss_res_corr_q2sb_var,
    new_binning_by_mode_dict=RDStarFitVariables.e_miss_minus_p_miss_wide_binning_dict_q2sb_corr,
    new_scope_by_mode_dict=False,
)
q2_sideband_e_miss_p_miss_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_e_miss_p_miss_asimov_fit_case,
)


q2_sideband_wide_asimov_fit_case = FitCase(
    name="q2_sideband_wide_asimov",
    base_dir_name="Fitting_AsimovQ2Sideband_Wide",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_Wide_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_wide_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

q2_sideband_wide_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_wide_asimov_fit_case)


q2_sideband_wide_split_d_star_modes_asimov_fit_case = FitCase(
    name="q2_sideband_wide_asimov_split_d_star_modes",
    base_dir_name="Fitting_AsimovQ2Sideband_Wide_SplitDStarModes",
    sample_key="q2_sideband",
    output_tag="AsimovFit_SplitDStarModes_q2Sideband_Wide_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_wide_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_setup.sig_id_customizer,
    },
)

q2_sideband_wide_split_d_star_modes_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_wide_split_d_star_modes_asimov_fit_case
)

q2_sideband_wide_split_d_star_modes_inverted_asimov_fit_case = FitCase(
    name="q2_sideband_wide_asimov_split_d_star_modes_inverted",
    base_dir_name="Fitting_AsimovQ2Sideband_Wide_SplitDStarModesInverted",
    sample_key="q2_sideband",
    output_tag="AsimovFit_SplitDStarModesInverted_q2Sideband_Wide_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_wide_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.sig_id_customizer,
    },
)

q2_sideband_wide_split_d_star_modes_inverted_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_wide_split_d_star_modes_inverted_asimov_fit_case
)


q2_sideband_1d_mms_asimov_fit_case = FitCase(
    name="q2_sideband_1d_mms_asimov_fit",
    base_dir_name="Fitting_AsimovQ2Sideband_1D_MMS",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_FixedNuisanceParams_1D_MMS",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.one_d_mms_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)
q2_sideband_1d_mms_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=q2_sideband_1d_mms_asimov_fit_case)


q2_sideband_1d_mms_before_vf_asimov_fit_case = replace_mm2_hv(
    fit_case=q2_sideband_1d_mms_asimov_fit_case,
    new_mm2_col_name=MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
        correction_setup_name=before_vf_mbc_res_cor_setup.name,
        use_laplace_smeared_version=False,
    ),
    tag_str="mms_before_vf_mbc_res_cor",
)
q2_sideband_1d_mms_before_vf_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_1d_mms_before_vf_asimov_fit_case
)


q2_sideband_1d_e_miss_p_miss_asimov_fit_case = FitCase(
    name="q2_sideband_1d_e_miss_p_miss_asimov_fit",
    base_dir_name="Fitting_AsimovQ2Sideband_1D_E_Miss_Minus_P_Miss",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_FixedNuisanceParams_1D_E_Miss_Minus_P_Miss",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.one_d_e_miss_minus_p_miss_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)
q2_sideband_1d_e_miss_p_miss_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_1d_e_miss_p_miss_asimov_fit_case
)


# Asimov Q2-Sideband Fit Cases with FEI Correction

q2_sideband_with_fei_corr_asimov_fit_case = FitCase(
    name="q2_sideband_with_fei_corr_asimov",
    base_dir_name="Fitting_AsimovQ2Sideband_WithFEICorr",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_WithFEICorr_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
)

q2_sideband_with_fei_corr_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_corr_asimov_fit_case,
)

q2_sideband_with_fei_corr_split_d_star_modes_asimov_fit_case = FitCase(
    name="q2_sideband_with_fei_corr_asimov_split_d_star_modes",
    base_dir_name="Fitting_AsimovQ2Sideband_WithFEICorr_SplitDStarModes",
    sample_key="q2_sideband",
    output_tag="AsimovFit_SplitDStarModes_q2Sideband_WithFEICorr_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_setup.sig_id_customizer,
    },
)

q2_sideband_with_fei_corr_split_d_star_modes_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_corr_split_d_star_modes_asimov_fit_case
)

q2_sideband_with_fei_corr_split_d_star_modes_inverted_asimov_fit_case = FitCase(
    name="q2_sideband_with_fei_corr_asimov_split_d_star_modes_inverted",
    base_dir_name="Fitting_AsimovQ2Sideband_WithFEICorr_SplitDStarModesInverted",
    sample_key="q2_sideband",
    output_tag="AsimovFit_SplitDStarModesInverted_q2Sideband_WithFEICorr_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.sig_id_customizer,
    },
)

q2_sideband_with_fei_corr_split_d_star_modes_inverted_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_corr_split_d_star_modes_inverted_asimov_fit_case
)


# Asimov Q2-Sideband Fit Cases with FEI Correction and Efficiency Correction

q2_sideband_with_fei_and_eff_corr_asimov_fit_case = FitCase(
    name="q2_sideband_with_fei_and_eff_corr_asimov",
    base_dir_name="Fitting_AsimovQ2Sideband_WithFEIAndEffCorr",
    sample_key="q2_sideband",
    output_tag="AsimovFit_q2Sideband_WithFEIAndEffCorr_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    add_reco_mode_efficiency_correction=True,
)

q2_sideband_with_fei_and_eff_corr_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_and_eff_corr_asimov_fit_case,
)

q2_sideband_with_fei_and_eff_corr_split_d_star_modes_asimov_fit_case = FitCase(
    name="q2_sideband_with_fei_and_eff_corr_asimov_split_d_star_modes",
    base_dir_name="Fitting_AsimovQ2Sideband_WithFEIAndEffCorr_SplitDStarModes",
    sample_key="q2_sideband",
    output_tag="AsimovFit_SplitDStarModes_q2Sideband_WithFEIAndEffCorr_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    add_reco_mode_efficiency_correction=True,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_setup.sig_id_customizer,
    },
)

q2_sideband_with_fei_and_eff_corr_split_d_star_modes_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_and_eff_corr_split_d_star_modes_asimov_fit_case
)

q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_asimov_fit_case = FitCase(
    name="q2_sideband_with_and_eff_fei_corr_asimov_split_d_star_modes_inverted",
    base_dir_name="Fitting_AsimovQ2Sideband_WithFEIAndEffCorr_SplitDStarModesInverted",
    sample_key="q2_sideband",
    output_tag="AsimovFit_SplitDStarModesInverted_q2Sideband_WithFEIAndEffCorr_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_q2sb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    sig_queries=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.custom_sig_queries,
    alternative_template_colors=RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
    add_new_fei_calibration_factors=True,
    add_reco_mode_efficiency_correction=True,
    additional_info={
        "sig_id_customizer": RDStarFitSetupContainers.q2_sideband_split_d_star_inverted_setup.sig_id_customizer,
    },
)

q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_asimov_alt_gap_fit_case = use_alternative_gap_modeling(
    fit_case=q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_asimov_fit_case
)


# Data Mbc-Sideband Fit Cases

mbc_sideband_fit_case = FitCase(
    name="mbc_sideband_fit",
    base_dir_name="Fitting_MbcSideband",
    sample_key="mbc_sideband",
    output_tag="Fit_mbcSideband_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_mbcsb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.mbc_sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.mbc_sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.mbc_sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.mbc_sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

mbc_sideband_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=mbc_sideband_fit_case)


mbc_sideband_complex_fit_case = FitCase(
    name="mbc_sideband_complex_fit",
    base_dir_name="Fitting_MbcSideband_Complex",
    sample_key="mbc_sideband",
    output_tag="Fit_mbcSideband_Complex_FixedNuisanceParams",
    plot_info_text="Data",
    fix_nuisance_parameters=True,
    asimov_data=False,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.complex_mbcsb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.complex_mbc_sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.complex_mbc_sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.complex_mbc_sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.complex_mbc_sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

mbc_sideband_complex_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=mbc_sideband_complex_fit_case)


# Asimov Mbc-Sideband Fit Cases

mbc_sideband_asimov_fit_case = FitCase(
    name="mbc_sideband_asimov",
    base_dir_name="Fitting_AsimovMbcSideband",
    sample_key="mbc_sideband",
    output_tag="AsimovFit_mbcSideband_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.simple_mbcsb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.mbc_sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.mbc_sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.mbc_sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.mbc_sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

mbc_sideband_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=mbc_sideband_asimov_fit_case)


mbc_sideband_complex_asimov_fit_case = FitCase(
    name="mbc_sideband_complex_asimov",
    base_dir_name="Fitting_AsimovMbcSideband_Complex",
    sample_key="mbc_sideband",
    output_tag="AsimovFit_mbcSideband_Complex_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict=RDStarFitVariables.complex_mbcsb_fit_var_dict,
    weight_column_name=WeightManager.final_weight_with_bf_corrections_col,
    systematics_cols=full_systematics_cols,
    reco_mode_infos=main_reco_mode_info_tuple,
    template_setup=RDStarFitSetupContainers.complex_mbc_sideband_setup.template_setup,
    sig_id_to_data_type_mapping=get_rdstar_sig_id_data_type_mapping(use_alternative_gap=False),
    plot_colors=RDStarFitSetupContainers.complex_mbc_sideband_setup.plot_color_dict,
    plot_labels=RDStarFitSetupContainers.complex_mbc_sideband_setup.plot_labels_dict,
    reco_mode_labels_dict=main_reco_mode_latex_label_mapping,
    alternative_template_colors=RDStarFitSetupContainers.complex_mbc_sideband_setup.alternative_template_colors,
    mc_sample_keys=rdstar_samples.main_sample_keys,
)

mbc_sideband_complex_asimov_alt_gap_fit_case = use_alternative_gap_modeling(fit_case=mbc_sideband_complex_asimov_fit_case)


rdstar_fit_case_tuple: Tuple[FitCase, ...] = (
    # Signal Window -- Data  # TODO: Include one data setup for Box Opening
    # simple_signal_fit_case,
    # simple_signal_alt_gap_fit_case,
    #
    # signal_fit_case,
    # signal_alt_gap_fit_case,
    #
    # signal_wide_fit_case,
    # signal_wide_alt_gap_fit_case,
    #
    # With FEI Correction
    # simple_signal_with_fei_corr_fit_case,
    # simple_signal_with_fei_corr_alt_gap_fit_case,
    # With FEI and Efficiency Correction
    # simple_signal_with_fei_and_eff_corr_fit_case,
    # simple_signal_with_fei_and_eff_corr_alt_gap_fit_case,
    #
    # # Signal Window -- Asimov
    # simple_signal_asimov_fit_case,
    # simple_signal_alt_gap_asimov_fit_case,
    # simple_signal_mva_optimization_asimov_fit_case,
    # simple_signal_mva_optimization_asimov_alt_gap_fit_case,
    simple_signal_2d_mva_optimization_asimov_fit_case,
    # simple_signal_2d_mva_optimization_asimov_alt_gap_fit_case,
    # simple_signal_2d_mva_best_cut_asimov_fit_case,
    # simple_signal_2d_mva_best_cut_asimov_alt_gap_fit_case,
    # simple_signal_3d_mva_optimization_asimov_fit_case,
    # simple_signal_3d_mva_optimization_asimov_alt_gap_fit_case,
    # simple_signal_4d_mva_optimization_asimov_fit_case,
    # simple_signal_4d_mva_optimization_asimov_alt_gap_fit_case,
    # #
    # *list(get_bkg_fixed_fit_cases(fit_case=simple_signal_asimov_fit_case, create_alt_gap_cases=True)),
    # #
    # simple_signal_split_bkg_asimov_fit_case,
    # simple_signal_split_bkg_asimov_alt_gap_fit_case,
    # #
    # bin_adjust_signal_asimov_fit_case,
    # #
    # signal_asimov_fit_case,
    # signal_asimov_alt_gap_fit_case,
    # signal_wide_asimov_fit_case,
    # signal_wide_asimov_alt_gap_fit_case,
    # #
    # complex_signal_asimov_fit_case,
    # complex_signal_asimov_alt_gap_fit_case,
    # complex_signal_mva_optimization_asimov_fit_case,
    # complex_signal_mva_optimization_asimov_alt_gap_fit_case,
    # *list(get_bkg_fixed_fit_cases(fit_case=complex_signal_asimov_fit_case, create_alt_gap_cases=True)),
    # #
    # simple_signal_e_miss_p_miss_asimov_fit_case,
    # simple_signal_e_miss_p_miss_asimov_alt_gap_fit_case,
    # #
    # simple_signal_mms_before_vf_mbc_res_cor_asimov_fit_case,
    # simple_signal_mms_before_vf_mbc_res_cor_asimov_alt_gap_fit_case,
    # #
    # signal_3d_asimov_fit_case,
    # signal_3d_asimov_alt_gap_fit_case,
    #
    # # With FEI Correction
    # simple_signal_with_fei_corr_asimov_fit_case,
    # simple_signal_with_fei_corr_alt_gap_asimov_fit_case,
    # # With FEI and Efficiency Correction
    # simple_signal_with_fei_and_eff_corr_asimov_fit_case,
    # simple_signal_with_fei_and_eff_corr_alt_gap_asimov_fit_case,
    #
    # # q2 Sideband -- Data
    # q2_sideband_fit_case,
    # q2_sideband_alt_gap_fit_case,
    # q2_sideband_fit_case_with_old_res_corr,
    # q2_sideband_alt_gap_fit_case_with_old_res_corr,
    # # Checking Split D* modes
    # q2_sideband_split_d_star_modes_fit_case,
    # q2_sideband_split_d_star_modes_alt_gap_fit_case,
    # q2_sideband_split_d_star_modes_inverted_fit_case,
    # q2_sideband_split_d_star_modes_inverted_alt_gap_fit_case,
    # tmp_q2_sideband_split_d_star_modes_alt_gap_fit_case_with_old_res_corr,
    # # Checking mm2 shift:
    # q2_sideband_fit_case_with_mm2_shift,
    # q2_sideband_alt_gap_fit_case_with_mm2_shift,
    # q2_sideband_split_d_star_modes_fit_case_with_mm2_shift,
    # q2_sideband_split_d_star_modes_alt_gap_fit_case_with_mm2_shift,
    # q2_sideband_split_d_star_modes_inverted_fit_case_with_mm2_shift,
    # q2_sideband_split_d_star_modes_inverted_alt_gap_fit_case_with_mm2_shift,
    # # Check Nuisance
    # q2_sideband_fit_case_with_stat_nuisance,
    # q2_sideband_alt_gap_fit_case_with_stat_nuisance,
    # q2_sideband_fit_case_with_mm2_shift_with_stat_nuisance,
    # q2_sideband_alt_gap_fit_case_with_mm2_shift_with_stat_nuisance,
    # # Simple nuisance check setup in q2 sideband
    # q2_sideband_fit_case_with_stat_nuisance_5by5binning,
    # q2_sideband_fit_case_with_all_nuisance_5by5binning,
    #
    # # Wide binning q2 Sideband
    q2_sideband_wide_fit_case,
    # q2_sideband_wide_alt_gap_fit_case,
    # q2_sideband_wide_split_d_star_modes_fit_case,
    # q2_sideband_wide_split_d_star_modes_alt_gap_fit_case,
    # q2_sideband_wide_split_d_star_modes_inverted_fit_case,
    # q2_sideband_wide_split_d_star_modes_inverted_alt_gap_fit_case,
    #
    # # Other setups in q2 Sideband
    # q2_sideband_mm2_lep_ph_spc_fit_case,
    # q2_sideband_mm2_lep_ph_spc_alt_gap_fit_case,
    # #
    # q2_sideband_mms_before_vf_mbc_res_cor_fit_case,
    # q2_sideband_mms_before_vf_mbc_res_cor_alt_gap_fit_case,
    # #
    # q2_sideband_e_miss_p_miss_fit_case,
    # q2_sideband_e_miss_p_miss_alt_gap_fit_case,
    # #
    # q2_sideband_1d_mms_fit_case,
    # q2_sideband_1d_mms_alt_gap_fit_case,
    # q2_sideband_1d_mms_before_vf_fit_case,
    # q2_sideband_1d_mms_before_vf_alt_gap_fit_case,
    # q2_sideband_1d_e_miss_p_miss_fit_case,
    # q2_sideband_1d_e_miss_p_miss_alt_gap_fit_case,
    #
    # # q2 Sideband -- Data with FEI Correction
    # q2_sideband_with_fei_corr_fit_case,
    # q2_sideband_with_fei_corr_alt_gap_fit_case,
    # # Checking Split D* modes
    # q2_sideband_with_fei_corr_split_d_star_modes_fit_case,
    # q2_sideband_with_fei_corr_split_d_star_modes_alt_gap_fit_case,
    # q2_sideband_with_fei_corr_split_d_star_modes_inverted_fit_case,
    # q2_sideband_with_fei_corr_split_d_star_modes_inverted_alt_gap_fit_case,
    #
    # # q2 Sideband -- Data with FEI end Efficiency Correction
    # q2_sideband_with_fei_and_eff_corr_fit_case,
    # q2_sideband_with_fei_and_eff_corr_alt_gap_fit_case,
    # # Checking Split D* modes
    # q2_sideband_with_fei_and_eff_corr_split_d_star_modes_fit_case,
    # q2_sideband_with_fei_and_eff_corr_split_d_star_modes_alt_gap_fit_case,
    # q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_fit_case,
    # q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_alt_gap_fit_case,
    #
    # # q2 Sideband -- Asimov
    # q2_sideband_asimov_fit_case,
    # q2_sideband_asimov_alt_gap_fit_case,
    # q2_sideband_asimov_fit_case_with_old_res_corr,
    # q2_sideband_asimov_alt_gap_fit_case_with_old_res_corr,
    # # Checking Split D* modes
    # q2_sideband_split_d_star_modes_asimov_fit_case,
    # q2_sideband_split_d_star_modes_asimov_alt_gap_fit_case,
    # q2_sideband_split_d_star_modes_inverted_asimov_fit_case,
    # q2_sideband_split_d_star_modes_inverted_asimov_alt_gap_fit_case,
    # # Checking mm2 shift:
    # q2_sideband_asimov_fit_case_with_mm2_shift,
    # q2_sideband_asimov_alt_gap_fit_case_with_mm2_shift,
    # q2_sideband_split_d_star_modes_asimov_fit_case_with_mm2_shift,
    # q2_sideband_split_d_star_modes_asimov_alt_gap_fit_case_with_mm2_shift,
    # q2_sideband_split_d_star_modes_inverted_asimov_fit_case_with_mm2_shift,
    # q2_sideband_split_d_star_modes_inverted_asimov_alt_gap_fit_case_with_mm2_shift,
    # # Check Nuisance
    # q2_sideband_asimov_fit_case_with_stat_nuisance,
    # q2_sideband_alt_gap_asimov_fit_case_with_stat_nuisance,
    # q2_sideband_asimov_fit_case_with_mm2_shift_with_stat_nuisance,
    # q2_sideband_alt_gap_asimov_fit_case_with_mm2_shift_with_stat_nuisance,
    # # Simple nuisance check setup in q2 sideband
    # q2_sideband_asimov_fit_case_with_stat_nuisance_5by5binning,
    # q2_sideband_asimov_fit_case_with_all_nuisance_5by5binning,
    #
    # # Wide binning q2 Sideband
    # q2_sideband_wide_asimov_fit_case,
    # q2_sideband_wide_asimov_alt_gap_fit_case,
    # q2_sideband_wide_split_d_star_modes_asimov_fit_case,
    # q2_sideband_wide_split_d_star_modes_asimov_alt_gap_fit_case,
    # q2_sideband_wide_split_d_star_modes_inverted_asimov_fit_case,
    # q2_sideband_wide_split_d_star_modes_inverted_asimov_alt_gap_fit_case,
    #
    # # Other setups in q2 Sideband
    # q2_sideband_mm2_lep_ph_spc_asimov_fit_case,
    # q2_sideband_mm2_lep_ph_spc_asimov_alt_gap_fit_case,
    # #
    # q2_sideband_mms_before_vf_mbc_res_cor_asimov_fit_case,
    # q2_sideband_mms_before_vf_mbc_res_cor_asimov_alt_gap_fit_case,
    # #
    # q2_sideband_e_miss_p_miss_asimov_fit_case,
    # q2_sideband_e_miss_p_miss_asimov_alt_gap_fit_case,
    # #
    # q2_sideband_1d_mms_asimov_fit_case,
    # q2_sideband_1d_mms_asimov_alt_gap_fit_case,
    # q2_sideband_1d_mms_before_vf_asimov_fit_case,
    # q2_sideband_1d_mms_before_vf_asimov_alt_gap_fit_case,
    # q2_sideband_1d_e_miss_p_miss_asimov_fit_case,
    # q2_sideband_1d_e_miss_p_miss_asimov_alt_gap_fit_case,
    #
    #
    # # q2 Sideband -- Data with FEI Correction
    # q2_sideband_with_fei_corr_asimov_fit_case,
    # q2_sideband_with_fei_corr_asimov_alt_gap_fit_case,
    # # Checking Split D* modes
    # q2_sideband_with_fei_corr_split_d_star_modes_asimov_fit_case,
    # q2_sideband_with_fei_corr_split_d_star_modes_asimov_alt_gap_fit_case,
    # q2_sideband_with_fei_corr_split_d_star_modes_inverted_asimov_fit_case,
    # q2_sideband_with_fei_corr_split_d_star_modes_inverted_asimov_alt_gap_fit_case,
    #
    # q2 Sideband -- Data with FEI and Efficiency Correction
    # q2_sideband_with_fei_and_eff_corr_asimov_fit_case,
    # q2_sideband_with_fei_and_eff_corr_asimov_alt_gap_fit_case,
    # # Checking Split D* modes
    # q2_sideband_with_fei_and_eff_corr_split_d_star_modes_asimov_fit_case,
    # q2_sideband_with_fei_and_eff_corr_split_d_star_modes_asimov_alt_gap_fit_case,
    # q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_asimov_fit_case,
    # q2_sideband_with_fei_and_eff_corr_split_d_star_modes_inverted_asimov_alt_gap_fit_case,
    #
    # # Mbc Sideband -- Data
    # mbc_sideband_fit_case,
    # mbc_sideband_alt_gap_fit_case,
    # mbc_sideband_complex_fit_case,
    # mbc_sideband_complex_alt_gap_fit_case,
    #
    # # Mbc Sideband -- Asimov
    # mbc_sideband_asimov_fit_case,
    # mbc_sideband_asimov_alt_gap_fit_case,
    # mbc_sideband_complex_asimov_fit_case,
    # mbc_sideband_complex_asimov_alt_gap_fit_case,
    #
    # Signal Window --- SciPy
    # signal_asimov_fit_case_scipy,  # TODO: Does not work: Fix issues with SciPy...
)

rdstar_fit_cases: Dict[str, FitCase] = {fc.name: fc for fc in rdstar_fit_case_tuple}

assert len(rdstar_fit_cases) == len(rdstar_fit_case_tuple), (len(rdstar_fit_cases), len(rdstar_fit_case_tuple))
