"""
Simple script to run manual merge of Data n-tuple files, which is necessary because of their large size.

Felix Metzner 2021
"""

import os
import numpy as np
import pandas as pd

from typing import Tuple, List, NamedTuple

from analysistools.dataframe_tools.hdf5_utils import load_root_file

from rdstar.utilities import PathType

__all__ = [
    "ManualDataNTupleMerger",
]


class ChoppedDataInfo(NamedTuple):
    data_type: str
    streams: Tuple[int, ...]
    parts: int
    include_stream: bool
    has_underscore: bool = False
    alt_output: bool = False


class ManualDataNTupleMerger:

    file_tag = "base"  # type: str
    base_dir_path = "/ceph/fmetzner/rdstar/"  # type: PathType

    chopped_data_infos = (
        # ChoppedDataInfo(data_type="DataY4S", streams=tuple(range(1)), parts=3, include_stream=False),
        ChoppedDataInfo(data_type="Charm", streams=tuple(range(6)), parts=2, include_stream=True, has_underscore=True),
        ChoppedDataInfo(data_type="UDS", streams=tuple(range(6)), parts=3, include_stream=True, has_underscore=True),
        # ChoppedDataInfo(data_type="Charm", streams=tuple([1]), parts=3, include_stream=True),
    )  # type: Tuple[ChoppedDataInfo, ...]

    base_file_name_suffix = "Y4S_ntuple"  # type: str

    alternative_output_tag = "alternative_merge_"  # type: str

    def __init__(
        self,
        production_tag: str,
        accept_existing_parts: bool,
    ) -> None:
        self.production_tag = production_tag  # type: str
        self.accept_existing_parts = accept_existing_parts  # type: bool

    def do_merge(self) -> None:
        for chopped_data_info in self.chopped_data_infos:
            for stream in chopped_data_info.streams:
                self.do_merge_for_data_type(
                    data_type_info=chopped_data_info,
                    stream=stream,
                )

    def do_merge_for_data_type(self, data_type_info: ChoppedDataInfo, stream: int) -> None:
        data_type = data_type_info.data_type  # type: str
        parts = data_type_info.parts  # type: int
        base_dir_path = self.get_data_dir_path(data_type=data_type, stream=stream)  # type: PathType

        info_str = f"for data type {data_type} stream {stream}"  # type: str

        dfs = []  # type: List[pd.DataFrame]
        for part in range(parts):
            print(f"Loading data frame {part + 1} of {parts} {info_str}...")
            dfs.append(
                self.load_part(
                    data_type_info=data_type_info,
                    stream=stream,
                    part=int(part + 1),
                )
            )

        list_of_events = [len(df.index) for df in dfs]
        list_of_events_u = [len(np.unique(df.index)) for df in dfs]

        print(f"Merging {parts} data frames {info_str}...")
        merged_df = pd.concat(dfs, ignore_index=True, sort=False)

        for df in dfs:
            if isinstance(df, pd.DataFrame):
                del df

        merged_df.reset_index(drop=True, inplace=True)

        print(f"Checking merged data frame {info_str}...")
        assert list_of_events == list_of_events_u, (list_of_events, list_of_events_u)
        assert len(np.unique(merged_df.index)) == sum(list_of_events), (
            len(np.unique(merged_df.index)),
            sum(list_of_events),
            list_of_events,
        )

        hdf5_key = self.get_base_file_name(data_type_info=data_type_info, stream=stream)  # type: str
        alt_output_tag = self.alternative_output_tag if data_type_info.alt_output else ""  # type: str
        file_name = f"{alt_output_tag}{hdf5_key}_{self.file_tag}.h5"
        hdf5_file_path = os.path.join(base_dir_path, file_name)  # type: PathType
        tmp_hdf5_file_path = os.path.join(base_dir_path, f"tmp_{file_name}")  # type: PathType

        print(f"Writing merged data frame {info_str} to {hdf5_file_path}")
        assert not os.path.exists(hdf5_file_path), hdf5_file_path
        assert not os.path.exists(tmp_hdf5_file_path), tmp_hdf5_file_path

        merged_df.to_hdf(path_or_buf=tmp_hdf5_file_path, key=hdf5_key)
        os.rename(src=tmp_hdf5_file_path, dst=hdf5_file_path)

        print(f"Finished writing merged data frame {info_str} to {hdf5_file_path}")

    def load_part(self, data_type_info: ChoppedDataInfo, stream: int, part: int) -> pd.DataFrame:
        base_dir_path = self.get_data_dir_path(data_type=data_type_info.data_type, stream=stream)  # type: PathType
        base_file_name = self.get_base_file_name(data_type_info=data_type_info, stream=stream)  # type: str
        i_str = f"{part} of {data_type_info.parts} for data type {data_type_info.data_type} stream {stream}"  # type: str

        root_part_path = self.get_root_file_part_path(
            base_dir_path=base_dir_path,
            base_file_name=base_file_name,
            has_underscore=data_type_info.has_underscore,
            part=part,
        )  # type: PathType

        print(f"Loading part {i_str}: {root_part_path}")
        assert os.path.exists(root_part_path), root_part_path

        hdf5_key = f"{base_file_name}_part{part}"  # type: str
        file_name = f"{hdf5_key}.h5"  # type: str
        hdf5_file_path = os.path.join(base_dir_path, file_name)  # type: PathType
        tmp_hdf5_file_path = os.path.join(base_dir_path, f"tmp_{file_name}")  # type: PathType

        if self.accept_existing_parts:
            if os.path.exists(hdf5_file_path):
                print(f"Found existing hdf5 file for part {i_str}: {hdf5_file_path}\nLoading it instead of root file!")
                df = pd.read_hdf(hdf5_file_path)
                assert isinstance(df, pd.DataFrame), type(df).__name__
                return df
        else:
            assert not os.path.exists(hdf5_file_path), hdf5_file_path

        assert not os.path.exists(tmp_hdf5_file_path), tmp_hdf5_file_path

        df = load_root_file(file_path=root_part_path)

        print(f"Writing part {i_str} to hdf5 file: {hdf5_file_path}")
        df.to_hdf(path_or_buf=tmp_hdf5_file_path, key=hdf5_key)
        os.rename(src=tmp_hdf5_file_path, dst=hdf5_file_path)
        print(f"Saved part {i_str}: {hdf5_file_path}")
        return df

    @staticmethod
    def get_root_file_part_path(
        base_dir_path: PathType,
        base_file_name: str,
        has_underscore: bool,
        part: int,
    ) -> PathType:
        part_str: str = "part_" if has_underscore else "part"
        return os.path.join(base_dir_path, f"{base_file_name}_{part_str}{part}.root")

    def get_base_file_name(self, data_type_info: ChoppedDataInfo, stream: int) -> str:
        stream_tag = f"_{stream}" if data_type_info.include_stream else ""  # type: str
        return f"{data_type_info.data_type}{stream_tag}_{self.base_file_name_suffix}"

    def get_data_dir_path(self, data_type: str, stream: int) -> PathType:
        return os.path.join(self.base_dir_path, f"data_set_{self.production_tag}_stream{stream}", data_type)


if __name__ == "__main__":
    my_production_tag = "prod_24thMarch2022"  # type: str
    # my_production_tag = "prod_28thMarch2022_pilnu_reco"  # type: str

    merger = ManualDataNTupleMerger(production_tag=my_production_tag, accept_existing_parts=True)
    merger.do_merge()
