"""
Fitter for shift in missing mass squared distribution

Felix Metzner 2021
"""

import os
import copy
import json
import numpy as np
import pandas as pd

from scipy import optimize
from dataclasses import dataclass, field, asdict
from typing import Union, Optional, Tuple, List, Dict, NamedTuple, Any

from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.stats import pearson_chi2_test, cowan_binned_likelihood_gof
from templatefitter.binned_distributions.binned_distribution import BinnedDistribution, BinnedDistributionFromData

from rdstar.utilities import PathType
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit_setups import (
    ShiftEvaluationInfo,
    mm2_shift_eval_infos_dict,
)


__all__ = [
    "MM2ShiftFitter",
    "MCShiftInfo",
    "first_test_shift_info",
]


@dataclass(frozen=True)
class ShiftFitCase:
    name: str
    id_column: str
    id_cases: Tuple[Tuple[Union[int, float], ...], ...]
    id_type: str  # Can be "range", "values" or "none"

    def __post_init__(self) -> None:
        allowed_id_types = ("range", "values", "none")  # type: Tuple[str, ...]
        if self.id_type not in allowed_id_types:
            raise ValueError(f"ShiftFitCase {self.name}: id_type must be in {allowed_id_types}, but is {self.id_type}!")

        assert isinstance(self.id_cases, tuple), self.id_cases
        assert all(isinstance(v, tuple) for v in self.id_cases), self.id_cases
        if self.id_type == "range":
            assert all(isinstance(v, tuple) and len(v) == 2 for v in self.id_cases), self.id_cases

    @property
    def has_cases(self) -> bool:
        return self.id_type != "none"

    def get_condition(self, df: pd.DataFrame, id_case_values: Tuple[Union[int, float], ...]) -> pd.Series:
        if not self.has_cases:
            raise RuntimeError("No condition for ShiftFitCase without sub-cases!")

        assert id_case_values in self.id_cases, (id_case_values, self.id_cases)

        if self.id_type == "range":
            return (df[self.id_column] >= id_case_values[0]) & (df[self.id_column] <= id_case_values[1])
        elif self.id_type == "values":
            return df[self.id_column].isin(id_case_values)
        else:
            raise RuntimeError(f"id_type {self.id_type} unknown!")

    @property
    def as_dict(self) -> Dict[str, Any]:
        return asdict(self)

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "ShiftFitCase":
        return ShiftFitCase(
            name=input_dict["name"],
            id_column=input_dict["id_column"],
            id_cases=tuple(tuple(v) for v in input_dict["id_cases"]),
            id_type=input_dict["id_type"],
        )


_tag_mbc_binning = (
    (5.2700, 5.2750),
    (5.2750, 5.2780),
    (5.2780, 5.2805),
    (5.2805, 5.2835),
    (5.2835, 5.290),
)  # type: Tuple[Tuple[float, ...], ...]

# fmt: off
shift_fit_cases = (
    ShiftFitCase(name="All", id_column=Cols.decay_mode_column_name, id_cases=((11, 18),), id_type="range"),
    # ShiftFitCase(name="AllWithDss", id_column="", id_cases=(), id_type="none"),
    ShiftFitCase(name="RecoMode", id_column=Cols.decay_mode_column_name, id_cases=((11, 12), (13, 14), (15, 16), (17, 18)), id_type="values"),
    # ShiftFitCase(name="RecoDMode", id_column=Cols.decay_mode_column_name, id_cases=((11, 12, 13, 14), (15, 16, 17, 18)), id_type="values"),
    # ShiftFitCase(name="RecoBCharge", id_column=Cols.decay_mode_column_name, id_cases=((11, 12, 15, 16), (13, 14, 17, 18)), id_type="values"),
    ShiftFitCase(name="RecoModeLep", id_column=Cols.decay_mode_column_name, id_cases=((11,), (12,), (15,), (16,), (13,), (14,), (17,), (18,)), id_type="values"),
    # ShiftFitCase(name="RecoLepMode", id_column=Cols.decay_mode_column_name, id_cases=((11, 13, 15, 17), (12, 14, 16, 18)), id_type="values"),
    ShiftFitCase(name="BTagMbc", id_column=Cols.b_tag_mbc, id_cases=_tag_mbc_binning, id_type="range"),
)  # type: Tuple[ShiftFitCase, ...]
# fmt: on


class MinimizerMethodInfo(NamedTuple):
    name: str
    bounds: bool


@dataclass()
class CurrentFitStatus:
    checked: bool = False
    args_checked: bool = False
    start_value: Optional[float] = None
    checked_values: List[Tuple[float, float]] = field(default_factory=list)

    def reset(self) -> None:
        self.checked = False
        self.args_checked = False
        self.start_value = None
        self.checked_values = []
        assert len(self.checked_values) == 0, self.checked_values

    @property
    def status_available(self) -> bool:
        assert self.checked
        assert self.args_checked
        assert self.start_value is not None
        assert self.checked_values is not None
        return True


class ShiftFitResultEntry(NamedTuple):
    fit_case_name: str
    best_shift_values: Tuple[float, ...]
    best_chi2_values: Tuple[float, ...]
    start_chi_values: Tuple[float, ...]
    combined_chi2_value: float
    combined_start_chi2_value: float
    checked_values: Dict[Tuple[Union[int, float], ...], Tuple[Tuple[float, float], ...]]
    fit_case_info: Dict[str, Any]

    @property
    def as_dict(self) -> Dict[str, Any]:
        di = self._asdict()
        di["checked_values"] = {f"{k}": v for k, v in self.checked_values.items()}
        return di


class MM2ShiftFitter:
    chi2_methods = (
        "cowan",
        "pearson",
    )  # type: Tuple[str, ...]

    simple_search_method_name = "StupidSearch"  # type: str

    minimizer_methods_infos = (
        MinimizerMethodInfo(name=simple_search_method_name, bounds=True),  # <- Own simple check
    )  # type: Tuple[MinimizerMethodInfo, ...]

    _other_minimizer_methods_infos = (
        MinimizerMethodInfo(name="Nelder-Mead", bounds=False),
        MinimizerMethodInfo(name="Nelder-Mead", bounds=False),  # <- Cannot handle bounds or constraints
        MinimizerMethodInfo(name="Powell", bounds=True),
        MinimizerMethodInfo(name="CG", bounds=False),  # <- Cannot handle bounds or constraints
        MinimizerMethodInfo(name="BFGS", bounds=True),
        # MinimizerMethodInfo(name="Newton-CG", bounds=True),  # <- Requires Jacobian
        # MinimizerMethodInfo(name="L-BFGS-B", bounds=True),  # <- Does not work!
        # MinimizerMethodInfo(name="TNC", bounds=True),  # <- Does not work!
        MinimizerMethodInfo(name="COBYLA", bounds=True),
        # MinimizerMethodInfo(name="SLSQP", bounds=True),  # <- Does not work!
        # MinimizerMethodInfo(name="dogleg", bounds=False),  # <- Requires Jacobian and cannot handle bounds or constraints
        # MinimizerMethodInfo(name="trust-ncg", bounds=False),  # <- Requires Jacobian and cannot handle bounds or constraints
        # MinimizerMethodInfo(name="trust-exact", bounds=False),  # <- Requires Jacobian
        # MinimizerMethodInfo(name="trust-krylov", bounds=False),  # <- Requires Jacobian
    )  # type: Tuple[MinimizerMethodInfo, ...]

    minimizer_methods = tuple(v.name for v in minimizer_methods_infos)  # type: Tuple[str, ...]

    # Shift fit setups:
    main_setup_name = "All"  # type: str
    shift_fit_cases_dict = {v.name: v for v in shift_fit_cases}  # type: Dict[str, ShiftFitCase]
    assert main_setup_name in shift_fit_cases_dict, (main_setup_name, shift_fit_cases_dict.keys())

    def __init__(
        self,
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        weight_column_name: str,
        mm2_hist_var: HistVariable,
        minimizer_method: str,
        chi2_method: str,
    ) -> None:
        assert chi2_method in self.chi2_methods, (chi2_method, self.chi2_methods)
        self.chi2_method = chi2_method  # type: str

        assert minimizer_method in self.minimizer_methods, (minimizer_method, self.minimizer_methods)
        self.minimizer_method = minimizer_method  # type: str

        self._weight_col_name = weight_column_name  # type: str
        self._mm2_hist_var = mm2_hist_var  # type: HistVariable

        self._bin_edges = None  # type: Optional[np.ndarray]

        self._fit_results = []  # type: List[ShiftFitResultEntry]
        self._current_status = CurrentFitStatus()  # type: CurrentFitStatus

        self._base_mc_df = self._create_reduced_df(df=mc_df, is_data=False)  # type: pd.DataFrame
        self._base_data_df = self._create_reduced_df(df=data_df, is_data=True)  # type: pd.DataFrame

        self._mc_df = copy.deepcopy(self._base_mc_df)  # type: pd.DataFrame
        self._data_df = copy.deepcopy(self._base_data_df)  # type: pd.DataFrame
        self._data_hist = self._fill_hist_from_df(df=self._data_df, is_data=True)  # type: BinnedDistribution

    def reset_fitter(self) -> None:
        self._current_status.reset()

    @property
    def hist_var(self) -> HistVariable:
        return self._mm2_hist_var

    @property
    def weight_col_name(self) -> str:
        return self._weight_col_name

    @property
    def var_col_name(self) -> str:
        return self.hist_var.df_label

    @property
    def original_var_col_name(self) -> str:
        return f"original_{self.var_col_name}"

    @property
    def binning(self) -> Binning:
        return Binning(
            bins=self.hist_var.n_bins,
            dimensions=1,
            scope=self.hist_var.scope,
            log_scale=self.hist_var.use_log_scale,
        )

    @property
    def bin_edges(self) -> np.ndarray:
        if self._bin_edges is None:
            self._bin_edges = np.array(list(self.binning.bin_edges[0]))
        return self._bin_edges

    @property
    def data_hist(self) -> BinnedDistribution:
        return self._data_hist

    @property
    def data_bin_count(self) -> np.ndarray:
        return self.data_hist.bin_counts

    @property
    def mc_df(self) -> pd.DataFrame:
        return self._mc_df

    def get_required_cols(self, for_data: bool) -> Tuple[str, ...]:
        base_cols = [self.var_col_name]  # type: List[str]
        if not for_data:
            base_cols.append(self.weight_col_name)
        id_cols = list(set([v.id_column for v in self.shift_fit_cases_dict.values() if v.id_column]))  # type: List[str]
        base_cols.extend(id_cols)
        return tuple(base_cols)

    def _create_reduced_df(self, df: pd.DataFrame, is_data: bool) -> pd.DataFrame:
        required_cols = self.get_required_cols(for_data=is_data)  # type: Tuple[str, ...]

        assert all(c in df.columns for c in required_cols), (c for c in required_cols if c not in df.columns)
        reduced_df = copy.deepcopy(df[list(required_cols)])

        assert len(df.index) == len(reduced_df.index), (len(df.index), len(reduced_df.index))
        assert df[self.var_col_name].equals(reduced_df[self.var_col_name])

        if not is_data:
            reduced_df.loc[:, self.original_var_col_name] = df[self.var_col_name]
            assert reduced_df[self.original_var_col_name].equals(reduced_df[self.var_col_name])

        return reduced_df

    def _fill_hist_from_df(self, df: pd.DataFrame, is_data: bool) -> BinnedDistribution:
        return BinnedDistributionFromData(
            bins=self.hist_var.n_bins,
            dimensions=1,
            scope=self.hist_var.scope,
            log_scale_mask=self.hist_var.use_log_scale,
            name=None,
            data=df,
            weights=None if is_data else self.weight_col_name,
            systematics=None,
            data_column_names=self.var_col_name,
        )

    def _reset_for_case(
        self,
        fit_case: ShiftFitCase,
        sub_fit_case: Optional[Tuple[Union[int, float], ...]],
    ) -> None:

        if fit_case.has_cases:
            assert sub_fit_case is not None
            mc_condition: pd.Series = fit_case.get_condition(df=self._base_mc_df, id_case_values=sub_fit_case)
            data_condition: pd.Series = fit_case.get_condition(df=self._base_data_df, id_case_values=sub_fit_case)

            self._mc_df = copy.deepcopy(self._base_mc_df.loc[mc_condition])
            self._data_df = copy.deepcopy(self._base_data_df.loc[data_condition])
        else:
            self._mc_df = copy.deepcopy(self._base_mc_df)
            self._data_df = copy.deepcopy(self._base_data_df)

        self._data_hist = self._fill_hist_from_df(df=self._data_df, is_data=True)

        normalization_factor = self._calculate_mc_df_scaling_factor(mc_df=self._mc_df)  # type: float
        self._mc_df.loc[:, self.weight_col_name] = normalization_factor * self._mc_df.loc[:, self.weight_col_name]

    def _calculate_mc_df_scaling_factor(self, mc_df: pd.DataFrame) -> float:
        mc_hist = self._fill_hist_from_df(df=mc_df, is_data=False)  # type: BinnedDistribution
        sum_of_mc_weights = sum(mc_hist.base_data.weights)  # type: float
        assert isinstance(sum_of_mc_weights, float), type(sum_of_mc_weights)
        return self.data_hist.base_data.data.size / sum_of_mc_weights

    def _get_valid_bin_counts(self, mc_bin_counts: np.ndarray) -> Tuple[np.ndarray, np.ndarray, int]:
        zero_data_counts = self.data_bin_count == 0.0  # type: np.ndarray
        zero_mc_counts = mc_bin_counts == 0.0  # type: np.ndarray

        valid_bins_mask = (~zero_data_counts) & (~zero_mc_counts)  # type: np.ndarray
        valid_bins = sum(valid_bins_mask)  # type: int

        return self.data_bin_count[valid_bins_mask], mc_bin_counts[valid_bins_mask], valid_bins

    def __calc_data_mc_chi2_for_current(self) -> float:
        _mc_bin_counts = np.histogramdd(
            sample=self.mc_df[self.var_col_name].values,
            weights=self.mc_df[self.weight_col_name].values,
            bins=[self.bin_edges],
            range=self.binning.range,
        )[0]

        # mc_bin_errors_sq = np.histogramdd(
        #     sample=self.mc_df[self.var_col_name],
        #     weights=np.square(self.mc_df[self.weight_col_name]),
        #     bins=self.bin_edges,
        #     range=self.binning.range[0],
        # )[0]

        data_bin_counts, mc_bin_counts, n_valid_bins = self._get_valid_bin_counts(
            mc_bin_counts=_mc_bin_counts
        )  # type: np.ndarray, np.ndarray, int

        dof = n_valid_bins - 1  # type: int
        if dof < 3:
            raise RuntimeError(f"dof too small: dof = {dof}, number of bins = {self.binning.num_bins[0]}")

        if not self._current_status.checked:
            assert n_valid_bins <= self.binning.num_bins[0], (n_valid_bins, self.binning.num_bins[0])
            assert mc_bin_counts.shape == data_bin_counts.shape, (mc_bin_counts.shape, data_bin_counts.shape)
            # assert mc_bin_counts.shape == mc_bin_errors_sq.shape, (mc_bin_counts.shape, mc_bin_errors_sq.shape)
            assert len(mc_bin_counts.shape) == 1, mc_bin_counts
            assert mc_bin_counts.shape[0] == n_valid_bins, (mc_bin_counts.shape, n_valid_bins, self.binning.num_bins)
            self._current_status.checked = True

        if self.chi2_method == "pearson":
            chi2, ndf, p_val = pearson_chi2_test(data=data_bin_counts, expectation=mc_bin_counts, dof=dof)
            return chi2
        elif self.chi2_method == "cowan":
            chi2, ndf, p_val = cowan_binned_likelihood_gof(data=data_bin_counts, expectation=mc_bin_counts, dof=dof)
            return chi2
        else:
            raise ValueError(f"Unknown method {self.chi2_method}; Must be one of {self.chi2_methods}")

    def _data_mc_chi2(self, mc_mm2_shift: float) -> float:
        self.mc_df.loc[:, self.var_col_name] = mc_mm2_shift + self.mc_df[self.original_var_col_name]
        return self.__calc_data_mc_chi2_for_current()

    def _evaluate_cost_function_for(self, x: float) -> float:
        _chi2_value = self._data_mc_chi2(mc_mm2_shift=x)  # type: float

        assert np.isfinite(_chi2_value), (_chi2_value, self.hist_var.df_label, self.minimizer_method, self.chi2_method)

        if self._current_status.start_value is None:
            self._current_status.start_value = _chi2_value

        self._current_status.checked_values.append((x, _chi2_value))
        chi2_value = _chi2_value - self._current_status.start_value  # type: float

        return chi2_value

    def _cost_function_wrapper(self, x: np.ndarray) -> float:
        if not self._current_status.args_checked:
            assert len(x.shape) == 1 and x.shape[0] == 1.0, x.shape
            self._current_status.args_checked = True

        return self._evaluate_cost_function_for(x=x[0])

    def _is_minimizer_method_with_bounds(self, method_name: str) -> bool:
        assert method_name in self.minimizer_methods, method_name
        method_mapping = {v.name: v.bounds for v in self.minimizer_methods_infos}
        return method_mapping[method_name]

    def _find_shift_via_minimizer(self) -> float:
        bounds = ((-0.08, 0.02),)
        res = optimize.minimize(
            fun=self._cost_function_wrapper,
            x0=np.array([0.0]),
            method=self.minimizer_method,
            bounds=bounds if self._is_minimizer_method_with_bounds(method_name=self.minimizer_method) else None,
        )

        assert len(res.x) == 1, (len(res.x), res.x)
        return res.x[0]

    def _find_shift_via_simple_search(self) -> float:
        bounds = (-0.08, 0.02)  # type: Tuple[float, float]
        finer_bounds = (-0.001, 0.0)  # type: Tuple[float, float]

        best_test_value = 0.0  # type: float
        best_chi2_value = self._evaluate_cost_function_for(x=0.0)  # type: float

        test_values = np.linspace(bounds[0], bounds[1], 10000, endpoint=True)  # type: np.ndarray
        test_values_fine = np.linspace(finer_bounds[0], finer_bounds[1], 10000, endpoint=True)  # type: np.ndarray
        test_space = sorted(list(set(list(test_values) + list(test_values_fine))), key=abs)

        for _i, test_x in enumerate(test_space):

            try:
                test_result = self._evaluate_cost_function_for(x=test_x)
            except AssertionError as ae:
                raise AssertionError(f"Failed for {_i}th value test_x = {test_x} in test_space {test_space}") from ae

            if test_result < best_chi2_value:
                best_test_value = test_x
                best_chi2_value = test_result

        return best_test_value

    def _run_shift_fit(self) -> Tuple[float, float]:
        assert len(self._current_status.checked_values) == 0
        if self.minimizer_method == self.simple_search_method_name:
            mm2_shift_result = self._find_shift_via_simple_search()  # type: float
        else:
            mm2_shift_result = self._find_shift_via_minimizer()

        min_chi2_val = self._data_mc_chi2(mc_mm2_shift=mm2_shift_result)  # type: float
        return mm2_shift_result, min_chi2_val

    def _get_sampled_value_paris(self) -> Tuple[Tuple[float, float], ...]:
        assert self._current_status.checked_values
        return tuple(self._current_status.checked_values)

    def _evaluate_for_combined_case(
        self,
        current_fit_case: ShiftFitCase,
        shift_values: Tuple[float, ...],
    ) -> Tuple[float, float]:
        assert len(shift_values) == len(current_fit_case.id_cases), (current_fit_case.name, shift_values)

        main_setup = self.shift_fit_cases_dict[self.main_setup_name]  # type: ShiftFitCase
        assert len(main_setup.id_cases) <= 1, (main_setup.id_cases, main_setup.has_cases, self.main_setup_name)

        self._reset_for_case(fit_case=main_setup, sub_fit_case=main_setup.id_cases[0] if main_setup.has_cases else None)

        start_chi2 = self.__calc_data_mc_chi2_for_current()  # type: float

        for shift_value, id_case in zip(shift_values, current_fit_case.id_cases):
            cond = current_fit_case.get_condition(df=self.mc_df, id_case_values=id_case)  # type: pd.Series
            self.mc_df.loc[cond, self.var_col_name] = shift_value + self.mc_df.loc[cond, self.original_var_col_name]

        chi2 = self.__calc_data_mc_chi2_for_current()  # type: float
        return chi2, start_chi2

    def _perform_shift_fit_for_case(self, fit_case: ShiftFitCase) -> None:
        self.reset_fitter()
        if fit_case.has_cases:
            best_shift_values = []  # type: List[float]
            best_chi2_values = []  # type: List[float]
            start_chi_values = []  # type: List[float]
            checked_values = {}  # type: Dict[Tuple[Union[int, float], ...], Tuple[Tuple[float, float], ...]]

            for sub_fit_case in fit_case.id_cases:
                self.reset_fitter()
                self._reset_for_case(fit_case=fit_case, sub_fit_case=sub_fit_case)

                try:
                    shift, chi2 = self._run_shift_fit()  # type: float, float
                except AssertionError as e:
                    raise AssertionError(f"Assertion failed for {fit_case.name} and sub case {sub_fit_case}.") from e

                assert self._current_status.start_value is not None

                best_shift_values.append(shift)
                best_chi2_values.append(chi2)
                start_chi_values.append(self._current_status.start_value)
                checked_values.update({sub_fit_case: self._get_sampled_value_paris()})

            combined_chi2, combined_start_chi2 = self._evaluate_for_combined_case(
                current_fit_case=fit_case,
                shift_values=tuple(best_shift_values),
            )  # type: float, float

            shift_fit_result_entry = ShiftFitResultEntry(
                fit_case_name=fit_case.name,
                best_shift_values=tuple(best_shift_values),
                best_chi2_values=tuple(best_chi2_values),
                start_chi_values=tuple(start_chi_values),
                combined_chi2_value=combined_chi2,
                combined_start_chi2_value=combined_start_chi2,
                checked_values=checked_values,
                fit_case_info=fit_case.as_dict,
            )

        else:
            self.reset_fitter()
            self._reset_for_case(fit_case=fit_case, sub_fit_case=None)

            shift, chi2 = self._run_shift_fit()

            assert self._current_status.start_value is not None

            shift_fit_result_entry = ShiftFitResultEntry(
                fit_case_name=fit_case.name,
                best_shift_values=(shift,),
                best_chi2_values=(chi2,),
                start_chi_values=(self._current_status.start_value,),
                combined_chi2_value=chi2,
                combined_start_chi2_value=self._current_status.start_value,
                checked_values={(-999,): self._get_sampled_value_paris()},
                fit_case_info=fit_case.as_dict,
            )

        self._fit_results.append(shift_fit_result_entry)

    def get_fit_result_dict(self) -> Dict[str, Any]:
        assert self._fit_results
        result_dict = {
            "hist_var_col_name": self.hist_var.df_label,
            "chi2_calc_method": self.chi2_method,
            "minimizer_method": self.minimizer_method,
        }  # type: Dict[str, Any]

        # Adding combined results on top first, so that they are quickly accessible.
        result_dict.update(
            {
                "overview": {
                    v.fit_case_name: (v.combined_chi2_value, v.combined_start_chi2_value) for v in self._fit_results
                }
            }
        )

        result_dict.update({v.fit_case_name: v.as_dict for v in self._fit_results})
        return result_dict

    def find_shift(self) -> Dict[str, Any]:
        for fit_case_name, fit_case in self.shift_fit_cases_dict.items():
            self._perform_shift_fit_for_case(fit_case=fit_case)

        result_info = self.get_fit_result_dict()  # type: Dict[str, Any]

        return result_info


@dataclass(frozen=True)
class MCShiftInfo:
    shift_case: ShiftFitCase
    shift_var_col_name: str
    shift_values: Tuple[float, ...]
    shift_eval_info: ShiftEvaluationInfo
    full_shift_fit_result_dict: Dict[str, Any]

    def __post_init__(self) -> None:
        allowed_id_types = ("range", "values", "none")  # type: Tuple[str, ...]
        if self.id_type not in allowed_id_types:
            raise ValueError(f"MCShiftInfo {self.name}: id_type must be in {allowed_id_types}, but is {self.id_type}!")

        if self.id_type == "range":
            assert all(isinstance(v, tuple) and len(v) == 2 for v in self.id_cases), self.id_cases

        if not len(self.shift_values) == len(self.id_cases):
            raise ValueError(
                f"Number of shift_values must match number of shift cases,\n"
                f"but the {len(self.shift_values)} shift values {self.shift_values}\n"
                f"do not match the {len(self.id_cases)} id_cases {self.id_cases}\n"
                f"of ShiftFitCase '{self.name}'!"
            )

    @property
    def name(self) -> str:
        return self.shift_case.name

    @property
    def id_column(self) -> str:
        return self.shift_case.id_column

    @property
    def id_cases(self) -> Tuple[Tuple[Union[int, float], ...], ...]:
        return self.shift_case.id_cases

    @property
    def id_type(self) -> str:  # Can be "range", "values" or "none"
        return self.shift_case.id_type

    @property
    def has_cases(self) -> bool:
        return self.shift_case.has_cases

    @property
    def eval_info_key(self) -> str:
        return self.shift_eval_info.name

    @staticmethod
    def extra_info_key() -> str:
        return "mm2_shift_info"

    def get_condition(self, df: pd.DataFrame, id_case_values: Tuple[Union[int, float], ...]) -> pd.Series:
        return self.shift_case.get_condition(df=df, id_case_values=id_case_values)

    @property
    def new_col_name(self) -> str:
        return self.shift_eval_info.get_new_col_name(
            base_col_name=self.shift_var_col_name,
            shift_case_name=self.shift_case.name,
        )

    def add_column_with_shifts(self, df: pd.DataFrame, is_data: bool) -> None:
        assert self.shift_var_col_name in df.columns, self.shift_var_col_name

        new_col_name = self.new_col_name  # type: str

        assert new_col_name not in df.columns, new_col_name

        if len(np.unique(df.index)) == 0:
            df = df.assign(**{new_col_name: df[self.shift_var_col_name]})
        else:
            df.loc[:, new_col_name] = df[self.shift_var_col_name].values

        assert new_col_name in df.columns, new_col_name

        if is_data or len(np.unique(df.index)) == 0:
            return

        for shift_value, id_case in zip(self.shift_values, self.id_cases):
            cond = self.get_condition(df=df, id_case_values=id_case)  # type: pd.Series
            df.loc[cond, new_col_name] += shift_value

    def apply_shifts(self, df: pd.DataFrame) -> None:
        assert self.shift_var_col_name in df.columns, self.shift_var_col_name
        for shift_value, id_case in zip(self.shift_values, self.id_cases):
            cond = self.get_condition(df=df, id_case_values=id_case)  # type: pd.Series
            df.loc[cond, self.shift_var_col_name] += shift_value

    @staticmethod
    def get_default_file_name() -> str:
        return "mm2_shift_fit_info"

    @property
    def as_dict(self) -> Dict[str, Any]:
        res = {
            "shift_case": self.shift_case.as_dict,
            "shift_var_col_name": self.shift_var_col_name,
            "shift_values": self.shift_values,
            "shift_eval_info": self.shift_eval_info.as_dict,
            "full_shift_fit_result_dict": self.full_shift_fit_result_dict,
        }
        return res

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "MCShiftInfo":
        shift_case = ShiftFitCase.from_dict(input_dict=input_dict["shift_case"])
        shift_eval_info = ShiftEvaluationInfo.from_dict(input_dict=input_dict["shift_eval_info"])
        _input_dict = {k: tuple(v) if k == "shift_values" else v for k, v in input_dict.items()}
        _input_dict.update({"shift_case": shift_case})
        _input_dict.update({"shift_eval_info": shift_eval_info})
        return MCShiftInfo(**_input_dict)

    @staticmethod
    def from_file(file_path: PathType) -> "MCShiftInfo":
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"MCShiftInfo file {file_path} does not exist!")

        with open(file_path, "r") as in_file:
            input_dict = json.load(in_file)

        return MCShiftInfo.from_dict(input_dict=input_dict)

    def write_to_file(self, target_path: PathType) -> None:
        if os.path.exists(target_path):
            raise FileExistsError(f"Trying to overwrite {target_path}!")

        result_dict = self.as_dict  # type: Dict[str, Any]

        tmp_target_path = os.path.join(os.path.dirname(target_path), f"tmp_{os.path.basename(target_path)}")
        with open(tmp_target_path, "w") as output_file:
            json.dump(result_dict, output_file, indent=2)
        os.rename(tmp_target_path, target_path)


# First Setup used for tests...
first_test_shift_info = MCShiftInfo(
    shift_case=MM2ShiftFitter.shift_fit_cases_dict["BTagMbc"],
    shift_var_col_name="m2RecoilSignalSide_mbc_based_corrected_resolution",
    shift_values=(
        -0.018043804380438046,
        0.0031983198319831957,
        -0.0019121912191219087,
        -0.0016721672167216706,
        -0.010423042304230426,
    ),
    shift_eval_info=mm2_shift_eval_infos_dict["missing_mass_res_corr_extreme_fine_binning"],
    full_shift_fit_result_dict={},
)
