"""
Application of multi stage resolution correction with iterative shifting and smearing for the Squared Missing Mass

Felix Metzner 2021
"""

import numpy as np
import pandas as pd

from typing import Optional, Tuple, Callable

from rdstar.utilities import PathType
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit import MCShiftInfo
from rdstar.offline_analysis.mms_resolution_correction.multistage_resolution_evaluator import (
    MultistageResolutionCorrectionInfo,
)
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import (
    MissingMassResolutionCorrector,
    ResolutionInfo,
)
from rdstar.offline_analysis.mms_resolution_correction.multistage_setups import (
    MultistageResolutionCorrectionSetup,
    multi_stage_setups_dict,
)


__all__ = [
    "MultistageResolutionCorrector",
]


class MultistageResolutionCorrector:
    def __init__(
        self,
        correction_setup_name: str,
        setup_info_dir_path: PathType,
    ) -> None:
        self._correction_setup_name = correction_setup_name  # type: str
        self.corr_setup = multi_stage_setups_dict[
            self._correction_setup_name
        ]  # type: MultistageResolutionCorrectionSetup

        self._corr_eval_result = MultistageResolutionCorrectionInfo.read_from_directory(
            dir_path=setup_info_dir_path
        )  # type: MultistageResolutionCorrectionInfo

        assert self._corr_eval_result.multi_stage_correction_setup_name == self._correction_setup_name, (
            self._corr_eval_result.multi_stage_correction_setup_name,
            self._correction_setup_name,
        )

        assert self._corr_eval_result.has_second_iteration is self.corr_setup.has_second_iteration

    @property
    def original_col_name(self) -> str:
        return self.corr_setup.original_col_name

    @property
    def corrected_col_names(self) -> Tuple[str, ...]:
        return self.corr_setup.corrected_col_names

    def apply_resolution_correction(self, sample: pd.DataFrame, is_data: bool) -> None:
        self.add_required_secondary_res_corr_columns_to_df(stage=1, sample=sample)

        first_smear_info = self._corr_eval_result.first_resolution_fit  # type: ResolutionInfo
        first_res_corrector = MissingMassResolutionCorrector(
            resolution_correction_setup=self.corr_setup.first_resolution_corr_setup,
            resolution_info=first_smear_info,
        )

        second_res_corrector = None  # type: Optional[MissingMassResolutionCorrector]
        if self._corr_eval_result.has_second_iteration:
            assert self._corr_eval_result.second_resolution_fit is not None
            second_corr_setup = self.corr_setup.second_resolution_corr_setup
            assert second_corr_setup is not None

            self.add_required_secondary_res_corr_columns_to_df(stage=2, sample=sample)

            second_smear_info = self._corr_eval_result.second_resolution_fit  # type: ResolutionInfo
            second_res_corrector = MissingMassResolutionCorrector(
                resolution_correction_setup=second_corr_setup,
                resolution_info=second_smear_info,
            )

        assert self.original_col_name in sample.columns, self.original_col_name

        if len(np.unique(sample.index)) == 0:
            self.add_corrected_column_to_empty_data_frame(sample=sample)
        elif is_data:
            self._add_corrected_column_for_data(sample=sample)
        else:
            self._apply_resolution_correction(
                sample=sample,
                first_res_corrector=first_res_corrector,
                second_res_corrector=second_res_corrector,
            )

        assert all(c in sample.columns for c in self.corrected_col_names), (
            [c for c in self.corrected_col_names if c not in sample.columns],
            len(np.unique(sample.index)),
            is_data,
            len(sample[self.original_col_name].values),
        )

    def _apply_resolution_correction(
        self,
        sample: pd.DataFrame,
        first_res_corrector: MissingMassResolutionCorrector,
        second_res_corrector: Optional[MissingMassResolutionCorrector],
    ) -> None:
        # First Shift
        first_shift_info = self._corr_eval_result.first_shift_fit  # type: MCShiftInfo
        first_shift_info.add_column_with_shifts(df=sample, is_data=False)

        # First Smearing
        first_res_corrector.ensure_primary_col_available(sample=sample)
        first_res_corrector.apply_resolution_correction(sample=sample, correct_overflows=True, special_treatment=True)

        # Second Iteration
        if self._corr_eval_result.has_second_iteration:
            # Second Shift
            assert self._corr_eval_result.second_shift_fit is not None
            second_shift_info = self._corr_eval_result.second_shift_fit  # type: MCShiftInfo
            second_shift_info.add_column_with_shifts(df=sample, is_data=False)

            # Second Smearing
            assert second_res_corrector is not None
            second_res_corrector.ensure_primary_col_available(sample=sample)
            second_res_corrector.apply_resolution_correction(
                sample=sample,
                correct_overflows=True,
                special_treatment=True,
            )

    def _add_corrected_column_for_data(
        self,
        sample: pd.DataFrame,
    ) -> None:
        orig_column = self.original_col_name  # type: str
        new_columns = self.corrected_col_names  # type: Tuple[str, ...]

        assert orig_column in sample.columns, orig_column
        assert all(c not in sample.columns for c in new_columns), [c for c in new_columns if c in sample.columns]

        for new_column in new_columns:
            sample.loc[:, new_column] = sample[orig_column].values
            assert new_column in sample.columns, new_column

    def add_corrected_column_to_empty_data_frame(
        self,
        sample: pd.DataFrame,
    ) -> None:
        assert len(np.unique(sample.index)) == 0, len(np.unique(sample.index))
        orig_column = self.original_col_name  # type: str
        new_columns = self.corrected_col_names  # type: Tuple[str, ...]

        assert orig_column in sample.columns, orig_column
        assert all(c not in sample.columns for c in new_columns), [c for c in new_columns if c in sample.columns]

        for new_column in new_columns:
            sample.loc[:, new_column] = sample[orig_column].values
            assert new_column in sample.columns, new_column

        assert all(c in sample.columns for c in new_columns), [c for c in new_columns if c not in sample.columns]

    def add_required_secondary_res_corr_columns_to_df(self, stage: int, sample: pd.DataFrame) -> None:
        if stage == 1:
            res_corr_setup = self.corr_setup.first_resolution_corr_setup
        elif stage == 2:
            assert self._corr_eval_result.has_second_iteration
            res_corr_setup = self.corr_setup.second_resolution_corr_setup
        else:
            raise ValueError(f"No definition available for stage {stage}!")

        if res_corr_setup.variable_definition is not None:
            required_col = res_corr_setup.secondary_hist_var.df_label
            if required_col not in sample.columns:
                variable_definition_function = res_corr_setup.variable_definition  # type: Optional[Callable]
                assert variable_definition_function is not None
                variable_definition_function(df=sample, check_range=False, accept_existing_column=True)

    @staticmethod
    def get_correction_column_name_for(multi_stage_correction_setup_name: str) -> str:
        return multi_stage_setups_dict[multi_stage_correction_setup_name].final_corrected_col_name
