"""
Evaluation of Squared Missing Mass resolution with iterative shifting and smearing.

Felix Metzner 2021
"""

import os
import copy
import numpy as np
import pandas as pd

from dataclasses import dataclass
from typing import Optional, Tuple, List, Dict, Generator, Any

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.validation.sideband_validation import SidebandInfo
from rdstar.offline_analysis.validation import sideband_validation as sideband_validation_tools

from rdstar.comparison_tools.data_mc_comparer import DataMCComparer
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import ResolutionInfo
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit_setups import ShiftEvaluationInfo
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit import MM2ShiftFitter, MCShiftInfo
from rdstar.offline_analysis.mms_resolution_correction.resolution_evaluation import MissingMassResolutionEvaluator
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import ResolutionCorrectionSetup
from rdstar.offline_analysis.validation.correlation_calculator import corr_eval_info_additional_info_key, CorrCalcInfo
from rdstar.offline_analysis.mms_resolution_correction.smearing_technique_comparison_plot import SmearingComparisonPlot
from rdstar.offline_analysis.mms_resolution_correction.multistage_setups import (
    MultistageResolutionCorrectionSetup,
    multi_stage_setups_dict,
)


__all__ = [
    "MultistageResolutionCorrectionInfo",
    "MultistageResolutionEvaluator",
]


class MRCOutputManager:
    shift_sub_dir_base_name = "shift_fit_results"  # type: str
    smear_sub_dir_base_name = "resolution_fit_results"  # type: str

    shift_file_name_wo_ext = MCShiftInfo.get_default_file_name()  # type: str
    smear_file_name_wo_ext = MissingMassResolutionEvaluator.default_result_file  # type: str

    @staticmethod
    def get_stage_label(stage: int) -> str:
        return f"stage_{stage}"

    @staticmethod
    def get_shift_dir_path_for(base_dir_path: PathType, stage: int) -> PathType:
        MRCOutputManager.check_if_base_dir_path_is_valid(base_dir_path=base_dir_path)
        stage_label = MRCOutputManager.get_stage_label(stage=stage)  # type: str
        return os.path.join(base_dir_path, f"{stage_label}_{MRCOutputManager.shift_sub_dir_base_name}")

    @staticmethod
    def get_smear_dir_path_for(base_dir_path: PathType, stage: int) -> PathType:
        MRCOutputManager.check_if_base_dir_path_is_valid(base_dir_path=base_dir_path)
        stage_label = MRCOutputManager.get_stage_label(stage=stage)  # type: str
        return os.path.join(base_dir_path, f"{stage_label}_{MRCOutputManager.smear_sub_dir_base_name}")

    @staticmethod
    def create_dir_upto_stage(base_dir_path: PathType, max_stage: int) -> None:
        MRCOutputManager.check_if_base_dir_path_is_valid(base_dir_path=base_dir_path)
        for stage in [s + 1 for s in range(max_stage)]:
            shift_dir = MRCOutputManager.get_shift_dir_path_for(base_dir_path=base_dir_path, stage=stage)
            os.makedirs(shift_dir, exist_ok=True)
            smear_dir = MRCOutputManager.get_smear_dir_path_for(base_dir_path=base_dir_path, stage=stage)
            os.makedirs(smear_dir, exist_ok=True)

    @staticmethod
    def get_file_path_for(base_dir_path: PathType, stage: int, step: str) -> PathType:
        MRCOutputManager.check_if_base_dir_path_is_valid(base_dir_path=base_dir_path)
        allowed_step_values = ["shift", "smear"]  # type: List[str]

        if step == "shift":
            return os.path.join(
                MRCOutputManager.get_shift_dir_path_for(base_dir_path=base_dir_path, stage=stage),
                f"{MRCOutputManager.shift_file_name_wo_ext}.json",
            )
        elif step == "smear":
            return os.path.join(
                MRCOutputManager.get_smear_dir_path_for(base_dir_path=base_dir_path, stage=stage),
                f"{MRCOutputManager.smear_file_name_wo_ext}.json",
            )
        else:
            raise KeyError(f"Parameter step must be one of {allowed_step_values}, but {step} was provided!")

    @staticmethod
    def get_setup_name_from_base_dir_path(base_dir_path: PathType) -> str:
        setup_name = os.path.basename(os.path.normpath(base_dir_path))
        assert setup_name, setup_name
        assert "." not in setup_name, setup_name
        return setup_name

    @staticmethod
    def check_if_base_dir_path_is_valid(base_dir_path: PathType) -> None:
        setup_name = MRCOutputManager.get_setup_name_from_base_dir_path(base_dir_path=base_dir_path)  # type: str
        if setup_name not in multi_stage_setups_dict:
            raise RuntimeError(f"{setup_name}, {base_dir_path}")


@dataclass(frozen=True)
class MultistageResolutionCorrectionInfo:
    multi_stage_correction_setup_name: str
    first_shift_fit: MCShiftInfo
    first_resolution_fit: ResolutionInfo
    second_shift_fit: Optional[MCShiftInfo] = None
    second_resolution_fit: Optional[ResolutionInfo] = None

    def __post_init__(self) -> None:
        if int(self.second_shift_fit is None) + int(self.second_resolution_fit is None) == 1:
            sec_res_str = "second_resolution_fit" if self.second_shift_fit is None else "second_shift_fit"  # type: str
            raise ValueError(
                f"If a 2nd iteration was performed, both results are required, but only the {sec_res_str} result was!"
            )

    @property
    def has_second_iteration(self) -> bool:
        assert int(self.second_shift_fit is None) + int(self.second_resolution_fit is None) != 1
        return self.second_shift_fit is not None

    @staticmethod
    def read_from_directory(dir_path: PathType) -> "MultistageResolutionCorrectionInfo":
        if not os.path.isdir(dir_path):
            raise FileNotFoundError(
                f"The directory\n\t{dir_path}\nfrom which the MultistageResolutionCorrectionInfo shall be read does not exist!"
            )

        MRCOutputManager.check_if_base_dir_path_is_valid(base_dir_path=dir_path)
        multi_stage_correction_setup_name = MRCOutputManager.get_setup_name_from_base_dir_path(base_dir_path=dir_path)

        first_shift_file = MRCOutputManager.get_file_path_for(
            base_dir_path=dir_path, stage=1, step="shift"
        )  # type: PathType
        first_smear_file = MRCOutputManager.get_file_path_for(
            base_dir_path=dir_path, stage=1, step="smear"
        )  # type: PathType
        second_shift_file = MRCOutputManager.get_file_path_for(
            base_dir_path=dir_path, stage=2, step="shift"
        )  # type: PathType
        second_smear_file = MRCOutputManager.get_file_path_for(
            base_dir_path=dir_path, stage=2, step="smear"
        )  # type: PathType

        shift_2_in = second_shift_file if os.path.exists(second_shift_file) else None  # type: Optional[PathType]
        smear_2_in = second_smear_file if os.path.exists(second_smear_file) else None  # type: Optional[PathType]

        return MultistageResolutionCorrectionInfo(
            multi_stage_correction_setup_name=multi_stage_correction_setup_name,
            first_shift_fit=MCShiftInfo.from_file(file_path=first_shift_file),
            first_resolution_fit=ResolutionInfo.from_file(input_file_path=first_smear_file),
            second_shift_fit=None if shift_2_in is None else MCShiftInfo.from_file(file_path=shift_2_in),
            second_resolution_fit=None if smear_2_in is None else ResolutionInfo.from_file(input_file_path=smear_2_in),
        )

    @property
    def multi_stage_correction_setup(self) -> MultistageResolutionCorrectionSetup:
        return multi_stage_setups_dict[self.multi_stage_correction_setup_name]


class MultistageResolutionEvaluator:
    shift_minimizer_method = "StupidSearch"  # type: str
    shift_chi2_method = "pearson"  # type: str

    final_stage_output_key_appendix = "final__msrc_res_corr_stage"  # type: str
    last_res_output_key = (
        f"{MissingMassResolutionEvaluator.default_result_file}_{final_stage_output_key_appendix}"
    )  # type: str

    def __init__(
        self,
        correction_setup: MultistageResolutionCorrectionSetup,
        mc_sample: Optional[pd.DataFrame],
        data_sample: Optional[pd.DataFrame],
        selection_mva_output_col_name: Optional[str],
        base_output_dir_path: PathType,
    ) -> None:
        self.setup_info = correction_setup  # type: MultistageResolutionCorrectionSetup

        self._mc_sample = None if mc_sample is None else copy.deepcopy(mc_sample)  # type: Optional[pd.DataFrame]
        self._data_sample = None if data_sample is None else copy.deepcopy(data_sample)  # type: Optional[pd.DataFrame]

        self.selection_mva_output_col_name: Optional[str] = selection_mva_output_col_name

        self.output_dir_path = os.path.join(base_output_dir_path, correction_setup.name)  # type: PathType
        self._output_dict = dict()  # type: Dict[str, PathType]
        self._smearing_result_sigmas = dict()  # type: Dict[int, Tuple[float, ...]]

    @property
    def has_second_stage(self) -> bool:
        return self.setup_info.has_second_iteration

    @property
    def mc_sample(self) -> pd.DataFrame:
        assert self._mc_sample is not None
        return self._mc_sample

    @property
    def data_sample(self) -> pd.DataFrame:
        assert self._data_sample is not None
        return self._data_sample

    def get_mms_res_eval_output_key_appendix_for_stage(self, stage: int) -> str:
        is_final_stage = stage == 1 + int(self.has_second_stage)  # type: bool
        if is_final_stage:
            return self.final_stage_output_key_appendix
        else:
            return f"stage{stage}_msrc"

    def get_shift_fit_setup(self, stage: int) -> Tuple[str, ShiftEvaluationInfo]:
        if stage == 1:
            eval_info = self.setup_info.first_shift_fit  # type: ShiftEvaluationInfo
            shift_case_key = self.setup_info.first_shift_fit_case_key  # type: str
        elif stage == 2:
            assert self.setup_info.second_shift_fit_case_key is not None
            eval_info = self.setup_info.second_shift_fit
            shift_case_key = self.setup_info.second_shift_fit_case_key
        else:
            raise ValueError(f"No shift setup defined for a stage {stage}!")

        assert eval_info is not None and isinstance(eval_info, ShiftEvaluationInfo), (type(eval_info), eval_info)
        assert isinstance(shift_case_key, str), shift_case_key
        assert shift_case_key in MM2ShiftFitter.shift_fit_cases_dict, shift_case_key

        return shift_case_key, eval_info

    def get_resolution_corr_setup(self, stage: int) -> ResolutionCorrectionSetup:
        if stage == 1:
            setup_info = self.setup_info.first_resolution_corr_setup  # type: ResolutionCorrectionSetup
        elif stage == 2:
            _setup_info = self.setup_info.second_resolution_corr_setup
            assert _setup_info is not None
            setup_info = _setup_info
        else:
            raise ValueError(f"No resolution correction setup defined for a stage {stage}!")

        assert setup_info is not None
        assert isinstance(setup_info, ResolutionCorrectionSetup), (type(setup_info), setup_info)
        return setup_info

    def evaluate_mms_resolution(self, dry_run: bool = False) -> None:
        self.reset_output_dict()
        if not dry_run:
            n_stages = 1 + int(self.has_second_stage)  # type: int
            MRCOutputManager.create_dir_upto_stage(base_dir_path=self.output_dir_path, max_stage=n_stages)

        self.evaluate_and_apply_shift(stage=1, dry_run=dry_run)
        self.evaluate_and_apply_resolution_correction(stage=1, dry_run=dry_run)

        if self.has_second_stage:
            self.evaluate_and_apply_shift(stage=2, dry_run=dry_run)
            self.evaluate_and_apply_resolution_correction(stage=2, dry_run=dry_run)

        self.create_validation_plots(dry_run=dry_run)
        self.create_smearing_illustration_figure(dry_run=dry_run)

    def evaluate_and_apply_shift(self, stage: int, dry_run: bool = False) -> None:
        output_file_key = f"stage_{stage}_{MRCOutputManager.shift_file_name_wo_ext}"  # type: str
        output_file_path = MRCOutputManager.get_file_path_for(
            base_dir_path=self.output_dir_path,
            stage=stage,
            step="shift",
        )  # type: PathType

        self.add_to_output_dict(new_entries={output_file_key: output_file_path})

        if dry_run:
            return

        shift_case_key, shift_eval_setup = self.get_shift_fit_setup(stage=stage)  # type: str, ShiftEvaluationInfo

        shift_fitter = MM2ShiftFitter(
            mc_df=self.mc_sample,
            data_df=self.data_sample,
            weight_column_name=self.setup_info.shift_weight_column,
            mm2_hist_var=shift_eval_setup.hist_var,
            minimizer_method=self.shift_minimizer_method,
            chi2_method=self.shift_chi2_method,
        )

        shift_fit_result_dict = shift_fitter.find_shift()

        mm2_shift_info = MCShiftInfo(
            shift_case=MM2ShiftFitter.shift_fit_cases_dict[shift_case_key],
            shift_var_col_name=shift_eval_setup.hist_var.df_label,
            shift_values=shift_fit_result_dict[shift_case_key]["best_shift_values"],
            shift_eval_info=shift_eval_setup,
            full_shift_fit_result_dict=shift_fit_result_dict,
        )

        mm2_shift_info.write_to_file(target_path=output_file_path)
        mm2_shift_info.add_column_with_shifts(df=self.mc_sample, is_data=False)
        mm2_shift_info.add_column_with_shifts(df=self.data_sample, is_data=True)

    def evaluate_and_apply_resolution_correction(self, stage: int, dry_run: bool = False) -> None:
        output_dir_path = MRCOutputManager.get_smear_dir_path_for(base_dir_path=self.output_dir_path, stage=stage)

        resolution_corr_setup = self.get_resolution_corr_setup(stage=stage)  # type: ResolutionCorrectionSetup

        resolution_evaluator = MissingMassResolutionEvaluator(
            resolution_correction_setup=resolution_corr_setup,
            weight_column=self.setup_info.resolution_weight_column,
            results_path=output_dir_path,
            output_key_appendix=self.get_mms_res_eval_output_key_appendix_for_stage(stage=stage),
        )

        output_dict = resolution_evaluator.get_output_dict()  # type: Dict[str, PathType]
        self.add_to_output_dict(new_entries=output_dict)

        if not dry_run:
            resolution_evaluator.add_required_columns_to_dfs(mc_df=self.mc_sample, data_df=self.data_sample)
            resolution_evaluator.ensure_primary_col_available(df=self.mc_sample)
            resolution_evaluator.ensure_primary_col_available(df=self.data_sample)

            _output_dict = resolution_evaluator.evaluate_resolution(mc_df=self.mc_sample, data_df=self.data_sample)
            assert _output_dict == output_dict, (_output_dict, output_dict)

            res_corrector = MissingMassResolutionCorrector(
                resolution_correction_setup=resolution_corr_setup,
                resolution_info=resolution_evaluator.resolution_correction_result,
            )

            assert stage not in self._smearing_result_sigmas, (stage, self._smearing_result_sigmas)
            self._smearing_result_sigmas[stage] = copy.copy(resolution_evaluator.resolution_correction_result.sigmas)

            res_corrector.apply_resolution_correction(
                sample=self.mc_sample,
                correct_overflows=True,
                special_treatment=True,
            )
            res_corrector.add_corrected_column_for_data(sample=self.data_sample, special_treatment=True)

    def create_validation_plots(self, dry_run: bool) -> None:
        # stages from 0 (before correction) to 1 or 2, depending on self.has_second_stage
        stages = tuple(range(2 + int(self.has_second_stage)))  # type: Tuple[int, ...]

        for stage in stages:
            plot_dir_name = self.get_validation_output_dir_name(stage=stage)  # type: str
            plot_base_dir_path = os.path.join(self.output_dir_path, plot_dir_name)  # type: PathType

            for sb_info_dir_name, sb_info in self.get_sideband_infos_for_stage(stage=stage):
                plot_dir_path = os.path.join(plot_base_dir_path, sb_info_dir_name)  # type: PathType
                if not dry_run:
                    os.makedirs(plot_dir_path, exist_ok=True)

                data_mc_comparer = DataMCComparer(comparer_info=sb_info)
                for sb_plot_style in ["stacked", "summed"]:

                    plot_output_dict = data_mc_comparer.plot_data_mc_comparison(
                        mc_df=pd.DataFrame([]) if dry_run else self.mc_sample,
                        data_df=pd.DataFrame([]) if dry_run else self.data_sample,
                        dry_run=dry_run,
                        weight_col=WeightManager.final_weight_with_bf_corrections_col,
                        style=sb_plot_style,
                        normalize_mc_to_data=True,
                        ratio_type="vs_uncert",
                        gof_check_method="toys",
                        adaptive_binning=True,
                        show_plots=False,
                        export_path=plot_dir_path,
                    )  # type: Optional[Dict[str, PathType]]

                    assert plot_output_dict is not None
                    self.add_to_output_dict(new_entries=plot_output_dict)

    def get_result_column_names_map(self, stage: int) -> Dict[str, str]:
        fc = self.setup_info.final_corrected_col_name  # type: str
        fct = "FINAL_"  # type: str

        if stage == 0:
            return {"original_state": self.setup_info.original_col_name}
        elif stage == 1:
            first_stage_gaussian, first_stage_laplace = self.setup_info.first_res_corr_result_col_names
            _cols_map = {
                "s1_shift": self.setup_info.first_shift_fit_result_col_name,
                "s1_gaussian": first_stage_gaussian,
                "s1_laplace": first_stage_laplace,
            }  # type: Dict[str, str]
            return {f"{fct}{k}" if v == fc else k: v for k, v in _cols_map.items()}
        elif stage == 2:
            if not self.has_second_stage:
                return dict()
            second_stage_gaussian, second_stage_laplace = self.setup_info.second_res_corr_result_col_names
            _cols_map = {
                "s2_shift": self.setup_info.second_shift_fit_result_col_name,
                "s2_gaussian": second_stage_gaussian,
                "s2_laplace": second_stage_laplace,
            }
            return {f"{fct}{k}" if v == fc else k: v for k, v in _cols_map.items()}
        else:
            raise ValueError(f"Result columns for stage {stage} not available!")

    def get_sideband_infos_for_stage(self, stage: int) -> Generator[Tuple[str, SidebandInfo], None, None]:
        n_validation_bins_values = (25, 48)  # type: Tuple[int, ...]

        for n_validation_bins in n_validation_bins_values:
            for base_res_tag, result_col_name in self.get_result_column_names_map(stage=stage).items():
                _appendix = f"s{stage}_{n_validation_bins}bins_{base_res_tag}"  # type: str
                result_tag = f"{base_res_tag}_w{n_validation_bins}bins"  # type: str

                this_sideband_info_hist_var = HistVariable(
                    df_label=result_col_name,
                    n_bins=n_validation_bins,
                    scope=(-2.0, 4.0),
                    var_name=LatexObsNames.fitting_mm2,
                    unit=LatexUnits.gev_sq,
                )

                if self.selection_mva_output_col_name is not None:
                    corr_calc_info: CorrCalcInfo = CorrCalcInfo(
                        var_col_name=self.selection_mva_output_col_name,
                        var_latex_label=LatexObsNames.mva_sel_output_short,
                        weighted=True,
                        corr_calc_method="pearson",
                    )
                    additional_info: Optional[Dict[str, Any]] = {corr_eval_info_additional_info_key: corr_calc_info}
                else:
                    additional_info = None

                this_sideband_info = SidebandInfo(
                    key="q2_sideband",
                    hist_var=this_sideband_info_hist_var,
                    title=sideband_validation_tools.q2_sideband_title,
                    info_text=sideband_validation_tools.q2_sideband_text,
                    text_h_pos=(0.64, 0.03),
                    text_v_pos=(0.76, 0.97),
                    legend_loc=1,
                    legend_cols=1,
                    y_scale=1.1,
                    systematics=None,
                    data_component_label=sideband_validation_tools.sideband_data_component_label,
                    file_name_appendix=f"{_appendix}_wSelMVACorr",
                    additional_info=additional_info,
                )  # type: SidebandInfo

                yield result_tag, this_sideband_info

                _sideband_infos_with_sys = SidebandInfo.create_comparer_infos_with_systematics(
                    comparer_info_list=[this_sideband_info],
                    systematics_collection=full_systematics_cols,
                    sys_name_tag=full_systematics_name,
                )  # type: List[SidebandInfo]
                this_sideband_info_with_sys = _sideband_infos_with_sys[0]  # type: SidebandInfo

                yield f"{result_tag}_with_{full_systematics_name}", this_sideband_info_with_sys

    def create_smearing_illustration_figure(self, dry_run: bool) -> None:
        eval_stage = 1  # type: int

        base_dir_path = os.path.join(self.output_dir_path, self.get_smear_comparison_output_dir_name(stage=eval_stage))
        _base_hist_var = self.get_resolution_corr_setup(stage=eval_stage).primary_hist_var  # type: HistVariable
        base_hist_var = copy.copy(_base_hist_var)  # type: HistVariable
        base_hist_var.scope = (-2.5, 4.5)

        compare_plot = SmearingComparisonPlot(
            target_dir_path=base_dir_path,
            base_hist_var=base_hist_var,
        )

        output_dict = compare_plot.get_output_dict(n_sigmas=1)  # type: Dict[str, PathType]
        self.add_to_output_dict(new_entries=output_dict)

        if not dry_run:
            sigma_values = [v for v in self._smearing_result_sigmas[eval_stage] if np.isfinite(v)]  # type: List[float]
            for sigma_index, sigma_value in enumerate(sigma_values):
                compare_plot.create_plot(sigma_index=sigma_index, sigma_value=sigma_value)

    @staticmethod
    def get_validation_output_dir_name(stage: int) -> str:
        return f"mm2_res_correction_after_stage_{stage}_validation"

    @staticmethod
    def get_smear_comparison_output_dir_name(stage: int) -> str:
        return f"mm2_stage_{stage}_smearing_technique_comparison"

    def get_output_dict(self) -> Dict[str, PathType]:
        if not self._output_dict:
            self.evaluate_mms_resolution(dry_run=True)
        return self._output_dict

    def reset_output_dict(self) -> None:
        self._output_dict = dict()

    def add_to_output_dict(self, new_entries: Dict[str, PathType]) -> None:
        for k, v in new_entries.items():
            assert k not in self._output_dict, (k, v)
            assert v not in self._output_dict.values(), (k, v)
        self._output_dict.update(new_entries)
