"""
Definition of additional mc variables.

Felix Metzner 2021
"""

import copy

import numpy as np
import pandas as pd

from typing import Union, Tuple, List, Dict, Callable, Type, Optional

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.pdg_code_collections import PDGCodeCollection as PcC

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos, TagRecoModeInfos

from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MCSigIDNameToQueryMappingType

from rdstar.offline_analysis.observables.offline_observable_definition_base import OfflineObservableBase


__all__ = [
    "TruthSigSideBMesonCharge",
    "TruthSigSideBMesonIsCharged",
    "RDStarFitSignalID",
    "RDStarFitSignalIDwithRecoModeInfo",
    "RDStarTruthChargeSplitFitSignalID",
    "RDStarRecoChargeSplitFitSignalID",
    "RDStarRecoAndSigTruthChargeSplitFitSignalID",
    "RDStarRecoAndSigTruthChargeAndDFSplitFitSignalID",
    "RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID",
    "DssRecoChAndSigDTruthChargeFitSignalID",
    "DssRecoChAndMainSigDTruthChargeFitSignalID",
    "FEIBTagRecoChAndSignalID",
    "FEIBTagAndSigDRecoChAndSignalID",
    "OfflineMCObservables",
]


_old_sig_id_mapping: MCSigIDNameToQueryMappingType = copy.copy(OldMCMatchingIDsQueryInfo.get_sig_id_mapping())


class TruthSigSideBMesonCharge(OfflineObservableBase):
    new_column_name: str = "mc_truth_charge_of_signal_b_meson"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    charge_truth_1st_mc_b: str = "first_b_meson_charged_truth"
    charge_truth_2nd_mc_b: str = "second_b_meson_charged_truth"

    required_variables: Dict[str, str] = {
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        charge_truth_1st_mc_b: Cols.mc_b1_meson_charge,
        charge_truth_2nd_mc_b: Cols.mc_b2_meson_charge,
    }

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        charge_truth_1st_b: np.array = df[cls.required_variables[cls.charge_truth_1st_mc_b]].values
        charge_truth_2nd_b: np.array = df[cls.required_variables[cls.charge_truth_2nd_mc_b]].values

        b_meson_charge_truth_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[charge_truth_1st_b, charge_truth_2nd_b],
            default=cls._no_match_value,
        )

        return b_meson_charge_truth_array


class TruthSigSideBMesonIsCharged(OfflineObservableBase):
    new_column_name: str = "mc_truth_charge_of_signal_b_meson"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    first_mc_b_pdg: str = "first_b_meson_pdg"
    second_mc_b_pdg: str = "second_b_meson_pdg"

    required_variables: Dict[str, str] = {
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        first_mc_b_pdg: Cols.mc_b1_meson_pdg,
        second_mc_b_pdg: Cols.mc_b2_meson_pdg,
    }

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        pdg_1st_b: np.array = df[cls.required_variables[cls.first_mc_b_pdg]].values
        pdg_2nd_b: np.array = df[cls.required_variables[cls.second_mc_b_pdg]].values

        b_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_b, pdg_2nd_b],
            default=cls._no_match_value,
        )

        abs_b_pdg: np.ndarray = np.abs(b_meson_pdg_array)

        b_meson_is_charged_array: np.array = np.select(
            condlist=[np.isin(abs_b_pdg, PcC.b_meson_charged_pdg_id), np.isin(abs_b_pdg, PcC.b_meson_neutral_pdg_id)],
            choicelist=[np.ones_like(b_meson_pdg_array), np.zeros_like(b_meson_pdg_array)],
            default=cls._no_match_value,
        )

        return b_meson_is_charged_array


class RDStarFitSignalID(OfflineObservableBase):
    new_column_name: str = "__rdstar_fit_signal_component_id__"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    id_neutral_offset: int = 1000

    sig_id: str = "default_sig_id_col"
    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    first_mc_b_pdg: str = "first_b_meson_pdg"
    second_mc_b_pdg: str = "second_b_meson_pdg"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        first_mc_b_pdg: Cols.mc_b1_meson_pdg,
        second_mc_b_pdg: Cols.mc_b2_meson_pdg,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        valid_new_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values()]
        valid_new_sig_id_values.extend([v.id + cls.id_neutral_offset for v in _old_sig_id_mapping.values() if v.id < 5])
        return tuple(valid_new_sig_id_values)

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        pdg_1st_b: np.array = df[cls.required_variables[cls.first_mc_b_pdg]].values
        pdg_2nd_b: np.array = df[cls.required_variables[cls.second_mc_b_pdg]].values

        b_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_b, pdg_2nd_b],
            default=cls._no_match_value,
        )

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        _is_charged: np.ndarray = np.abs(b_meson_pdg_array) == PcC.b_meson_charged_pdg_id
        _is_neutral: np.ndarray = np.abs(b_meson_pdg_array) == PcC.b_meson_neutral_pdg_id

        unchanged_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values() if v.id >= 5]
        must_stay_unchanged: np.ndarray = np.isin(element=sig_id_array, test_elements=unchanged_sig_id_values)

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_charged & ~must_stay_unchanged,
                _is_neutral & ~must_stay_unchanged,
                must_stay_unchanged | ~(_is_charged | _is_neutral),
            ],
            choicelist=[
                sig_id_array,
                cls.id_neutral_offset + sig_id_array,
                sig_id_array,
            ],
            default=np.nan,
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            unchanged_sig_id_values,
            valid_new_sig_id_values,
            cls.id_neutral_offset,
        )

        return new_sig_id_array


class RDStarFitSignalIDwithRecoModeInfo(OfflineObservableBase):
    new_column_name: str = "__rdstar_fit_signal_component_id_with_reco_mode_info__"

    id_neutral_offset: int = 1000

    charged_b_to_d_reco_offset: int = 10000
    neutral_b_to_d_reco_offset: int = 20000
    charged_b_to_dst_reco_offset: int = 30000
    neutral_b_to_dst_reco_offset: int = 40000

    all_reco_offsets: Tuple[int, ...] = (
        charged_b_to_d_reco_offset,
        neutral_b_to_d_reco_offset,
        charged_b_to_dst_reco_offset,
        neutral_b_to_dst_reco_offset,
    )

    reco_id: str = "signal_b_reco_mode"
    required_variables: Dict[str, str] = RDStarFitSignalID.required_variables
    required_variables.update({reco_id: Cols.decay_mode_column_name})

    @classmethod
    def get_modified_sig_id(
        cls,
        sig_ids: Union[int, Tuple[int, ...]],
        is_neutral_signal: Optional[bool],
        reco_mode: Optional[str]
    ) -> Tuple[int, ...]:
        _reco_modes_map: Dict[Optional[str], Tuple[int, ...]] = {
            "other": (0, ),
            "cb_to_d": (cls.charged_b_to_d_reco_offset, ),
            "nb_to_d": (cls.neutral_b_to_d_reco_offset, ),
            "cb_to_dst": (cls.charged_b_to_dst_reco_offset, ),
            "nb_to_dst": (cls.neutral_b_to_dst_reco_offset, ),
            "cb": (cls.charged_b_to_d_reco_offset, cls.charged_b_to_dst_reco_offset),
            "nb": (cls.neutral_b_to_d_reco_offset, cls.neutral_b_to_dst_reco_offset),
            "d": (cls.charged_b_to_d_reco_offset, cls.neutral_b_to_d_reco_offset),
            "dst": (cls.charged_b_to_dst_reco_offset, cls.neutral_b_to_dst_reco_offset),
            None: (0, ) + cls.all_reco_offsets,
        }

        if reco_mode not in _reco_modes_map:
            raise ValueError(f"Invalid reco_mode value {reco_mode}, must be one of {list(_reco_modes_map.keys())}!")
        reco_offsets: Tuple[int, ...] = _reco_modes_map[reco_mode]

        if isinstance(sig_ids, tuple):
            _sig_ids: Tuple[int, ...] = sig_ids
        elif isinstance(sig_ids, int):
            _sig_ids = (sig_ids, )
        else:
            raise TypeError(f"Unknown type of sig_ids provided: type(sig_ids) = {type(sig_ids)}, sig_ids = {sig_ids}")

        if is_neutral_signal is None:
            _neutral_sig_ids: Tuple[int, ...] = tuple(v + cls.id_neutral_offset for v in _sig_ids if v <= 4)
            _new_sig_ids: Tuple[int, ...] = tuple(v for v in _sig_ids + _neutral_sig_ids)
        elif is_neutral_signal is True:
            _new_sig_ids = tuple(v + cls.id_neutral_offset if v <= 4 else v for v in _sig_ids)
        else:
            _new_sig_ids = _sig_ids

        new_sig_ids: Tuple[int, ...] = tuple(s + reco_offset for s in _new_sig_ids for reco_offset in reco_offsets)

        return new_sig_ids

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        _new_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values()]
        _new_sig_id_values.extend([v.id + cls.id_neutral_offset for v in _old_sig_id_mapping.values() if v.id < 5])

        valid_new_sig_id_values: List[int] = [v for v in _new_sig_id_values]
        for reco_offset in cls.all_reco_offsets:
            valid_new_sig_id_values.extend([v + reco_offset for v in _new_sig_id_values])

        return tuple(valid_new_sig_id_values)

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:
        new_sig_id_values: Union[pd.Series, np.array] = RDStarFitSignalID.calculate_new_column_values(
            df=df,
            check_range=check_range,
        )

        assert not np.any(np.isnan(new_sig_id_values)), np.sum(np.isnan(new_sig_id_values))

        is_cb2d_reco: pd.Series = df[cls.required_variables[cls.reco_id]].isin(RecoModeInfos.charged_b_to_d_modes.ids)
        is_nb2d_reco: pd.Series = df[cls.required_variables[cls.reco_id]].isin(RecoModeInfos.neutral_b_to_d_modes.ids)
        is_cb2dst_reco: pd.Series = df[cls.required_variables[cls.reco_id]].isin(RecoModeInfos.charged_b_to_ds_modes.ids)
        is_nb2dst_reco: pd.Series = df[cls.required_variables[cls.reco_id]].isin(RecoModeInfos.neutral_b_to_ds_modes.ids)

        assert not np.any(is_cb2d_reco & is_nb2d_reco & is_cb2dst_reco & is_nb2dst_reco)

        new_sig_id_values[is_cb2d_reco] += cls.charged_b_to_d_reco_offset
        new_sig_id_values[is_nb2d_reco] += cls.neutral_b_to_d_reco_offset
        new_sig_id_values[is_cb2dst_reco] += cls.charged_b_to_dst_reco_offset
        new_sig_id_values[is_nb2dst_reco] += cls.neutral_b_to_dst_reco_offset

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_values, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_values)),
            valid_new_sig_id_values,
            cls.id_neutral_offset,
            cls.all_reco_offsets,
        )

        return new_sig_id_values


class RDStarTruthChargeSplitFitSignalID(OfflineObservableBase):
    new_column_name: str = "__rdstar_fit_truth_charge_split_signal_component_id__"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    id_neutral_offset: int = 1000

    sig_id: str = "default_sig_id_col"
    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    first_mc_b_pdg: str = "first_b_meson_pdg"
    second_mc_b_pdg: str = "second_b_meson_pdg"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        first_mc_b_pdg: Cols.mc_b1_meson_pdg,
        second_mc_b_pdg: Cols.mc_b2_meson_pdg,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        valid_new_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values()]
        valid_new_sig_id_values.extend([v.id + cls.id_neutral_offset for v in _old_sig_id_mapping.values() if v.id != 9])
        return tuple(valid_new_sig_id_values)

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        pdg_1st_b: np.array = df[cls.required_variables[cls.first_mc_b_pdg]].values
        pdg_2nd_b: np.array = df[cls.required_variables[cls.second_mc_b_pdg]].values

        b_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_b, pdg_2nd_b],
            default=cls._no_match_value,
        )

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        _is_charged: np.ndarray = np.abs(b_meson_pdg_array) == PcC.b_meson_charged_pdg_id
        _is_neutral: np.ndarray = np.abs(b_meson_pdg_array) == PcC.b_meson_neutral_pdg_id

        unchanged_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values() if v.id == 9]
        must_stay_unchanged: np.ndarray = np.isin(element=sig_id_array, test_elements=unchanged_sig_id_values)

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_charged & ~must_stay_unchanged,
                _is_neutral & ~must_stay_unchanged,
                must_stay_unchanged | ~(_is_charged | _is_neutral),
            ],
            choicelist=[
                sig_id_array,
                cls.id_neutral_offset + sig_id_array,
                sig_id_array,
            ],
            default=np.nan,
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            unchanged_sig_id_values,
            valid_new_sig_id_values,
            cls.id_neutral_offset,
        )

        return new_sig_id_array


class RDStarRecoChargeSplitFitSignalID(OfflineObservableBase):
    new_column_name: str = "__rdstar_fit_reco_charge_split_signal_component_id__"

    _charged_reco_modes: Tuple[int, ...] = RecoModeInfos.charged_reco_modes.ids
    _neutral_reco_modes: Tuple[int, ...] = RecoModeInfos.neutral_reco_modes.ids

    id_neutral_offset: int = 1000

    sig_id: str = "default_sig_id_col"
    decay_mode_id: str = "decay_mode_id"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        decay_mode_id: Cols.decay_mode_column_name,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        valid_new_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values()]
        valid_new_sig_id_values.extend([v.id + cls.id_neutral_offset for v in _old_sig_id_mapping.values()])
        return tuple(valid_new_sig_id_values)

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        reco_mode_id_array: np.array = df[cls.required_variables[cls.decay_mode_id]].values

        _is_charged: np.ndarray = np.isin(element=reco_mode_id_array, test_elements=cls._charged_reco_modes)
        _is_neutral: np.ndarray = np.isin(element=reco_mode_id_array, test_elements=cls._neutral_reco_modes)

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_charged,
                _is_neutral,
            ],
            choicelist=[
                sig_id_array,
                cls.id_neutral_offset + sig_id_array,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either charged or neutral ones...
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            valid_new_sig_id_values,
            cls.id_neutral_offset,
        )

        return new_sig_id_array


class RDStarRecoAndSigTruthChargeSplitFitSignalID(OfflineObservableBase):
    new_column_name: str = "__rdstar_reco_sig_truth_charge_split_signal_component_id__"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    _charged_reco_modes: Tuple[int, ...] = RecoModeInfos.charged_reco_modes.ids
    _neutral_reco_modes: Tuple[int, ...] = RecoModeInfos.neutral_reco_modes.ids

    reco_id_neutral_offset: int = 1000
    sig_id_neutral_sign_flip: int = -1

    sig_id: str = "default_sig_id_col"
    decay_mode_id: str = "decay_mode_id"
    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    first_mc_b_pdg: str = "first_b_meson_pdg"
    second_mc_b_pdg: str = "second_b_meson_pdg"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        decay_mode_id: Cols.decay_mode_column_name,
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        first_mc_b_pdg: Cols.mc_b1_meson_pdg,
        second_mc_b_pdg: Cols.mc_b2_meson_pdg,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        _valid_values: List[int] = [v.id for v in _old_sig_id_mapping.values()]
        _valid_values.extend([v.id + cls.reco_id_neutral_offset for v in _old_sig_id_mapping.values()])
        _valid_values.extend([v.id * cls.sig_id_neutral_sign_flip for v in _old_sig_id_mapping.values() if v.id < 5])
        _valid_values.extend(
            [
                (v.id + cls.reco_id_neutral_offset) * cls.sig_id_neutral_sign_flip
                for v in _old_sig_id_mapping.values()
                if v.id < 5
            ]
        )
        return tuple(_valid_values)

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        pdg_1st_b: np.array = df[cls.required_variables[cls.first_mc_b_pdg]].values
        pdg_2nd_b: np.array = df[cls.required_variables[cls.second_mc_b_pdg]].values

        b_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_b, pdg_2nd_b],
            default=cls._no_match_value,
        )

        orig_sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        _is_sig_neutral: np.ndarray = np.abs(b_meson_pdg_array) == PcC.b_meson_neutral_pdg_id
        unchanged_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values() if v.id >= 5]
        must_stay_unchanged: np.ndarray = np.isin(element=orig_sig_id_array, test_elements=unchanged_sig_id_values)

        reco_mode_id_array: np.array = df[cls.required_variables[cls.decay_mode_id]].values
        _is_reco_charged: np.ndarray = np.isin(element=reco_mode_id_array, test_elements=cls._charged_reco_modes)
        _is_reco_neutral: np.ndarray = np.isin(element=reco_mode_id_array, test_elements=cls._neutral_reco_modes)

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_reco_charged,
                _is_reco_neutral,
            ],
            choicelist=[
                sig_id_array,
                cls.reco_id_neutral_offset + sig_id_array,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either charged or neutral ones...
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        elements_to_receive_sign_flip: np.ndarray = _is_sig_neutral & (~must_stay_unchanged)
        new_sig_id_array[elements_to_receive_sign_flip] *= cls.sig_id_neutral_sign_flip

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            unchanged_sig_id_values,
            valid_new_sig_id_values,
            cls.reco_id_neutral_offset,
            cls.sig_id_neutral_sign_flip,
        )

        return new_sig_id_array


class RDStarRecoAndSigTruthChargeAndDFSplitFitSignalID(OfflineObservableBase):
    new_column_name: str = "__rdstar_reco_sig_truth_charge_df_split_signal_component_id__"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    _charged_reco_modes: Tuple[int, ...] = RecoModeInfos.charged_reco_modes.ids
    _neutral_reco_modes: Tuple[int, ...] = RecoModeInfos.neutral_reco_modes.ids

    cross_feed_dict: Dict[int, Tuple[int, ...]] = {
        _old_sig_id_mapping["DSl"].id: RecoModeInfos.b_to_d_modes.ids,
        _old_sig_id_mapping["Dl"].id: RecoModeInfos.b_to_ds_modes.ids,
    }

    reco_id_neutral_offset: int = 1000
    sig_id_neutral_sign_flip: int = -1
    cross_feed_offset: int = 500

    sig_id: str = "default_sig_id_col"
    decay_mode_id: str = "decay_mode_id"
    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    first_mc_b_pdg: str = "first_b_meson_pdg"
    second_mc_b_pdg: str = "second_b_meson_pdg"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        decay_mode_id: Cols.decay_mode_column_name,
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        first_mc_b_pdg: Cols.mc_b1_meson_pdg,
        second_mc_b_pdg: Cols.mc_b2_meson_pdg,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        _valid_values: List[int] = [v.id for v in _old_sig_id_mapping.values()]
        _valid_values.extend([v.id + cls.reco_id_neutral_offset for v in _old_sig_id_mapping.values()])

        _valid_values.extend([v.id * cls.sig_id_neutral_sign_flip for v in _old_sig_id_mapping.values() if v.id < 5])
        _valid_values.extend(
            [
                (v.id + cls.reco_id_neutral_offset) * cls.sig_id_neutral_sign_flip
                for v in _old_sig_id_mapping.values()
                if v.id < 5
            ]
        )

        cf_values = [v + cls.cross_feed_offset for v in cls.cross_feed_dict]
        cf_values.extend([v + cls.cross_feed_offset + cls.reco_id_neutral_offset for v in cls.cross_feed_dict])

        _valid_values.extend(cf_values + [v * cls.sig_id_neutral_sign_flip for v in cf_values])

        return tuple(_valid_values)

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        pdg_1st_b: np.array = df[cls.required_variables[cls.first_mc_b_pdg]].values
        pdg_2nd_b: np.array = df[cls.required_variables[cls.second_mc_b_pdg]].values

        b_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_b, pdg_2nd_b],
            default=cls._no_match_value,
        )

        orig_sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        _is_sig_neutral: np.ndarray = np.abs(b_meson_pdg_array) == PcC.b_meson_neutral_pdg_id
        unchanged_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values() if v.id >= 5]
        must_stay_unchanged: np.ndarray = np.isin(element=orig_sig_id_array, test_elements=unchanged_sig_id_values)

        reco_mode_id_array: np.array = df[cls.required_variables[cls.decay_mode_id]].values
        _is_reco_charged: np.ndarray = np.isin(element=reco_mode_id_array, test_elements=cls._charged_reco_modes)
        _is_reco_neutral: np.ndarray = np.isin(element=reco_mode_id_array, test_elements=cls._neutral_reco_modes)

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_reco_charged,
                _is_reco_neutral,
            ],
            choicelist=[
                sig_id_array,
                cls.reco_id_neutral_offset + sig_id_array,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either charged or neutral ones...
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        for cf_sig_id, cf_reco_ch_ids in cls.cross_feed_dict.items():
            is_cf_sig: np.ndarray = orig_sig_id_array == cf_sig_id
            is_cf_reco_ch: np.ndarray = np.isin(element=reco_mode_id_array, test_elements=cf_reco_ch_ids)
            is_cf: np.ndarray = is_cf_sig & is_cf_reco_ch

            new_sig_id_array[is_cf] += cls.cross_feed_offset

        elements_to_receive_sign_flip: np.ndarray = _is_sig_neutral & (~must_stay_unchanged)
        new_sig_id_array[elements_to_receive_sign_flip] *= cls.sig_id_neutral_sign_flip

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            unchanged_sig_id_values,
            valid_new_sig_id_values,
            cls.reco_id_neutral_offset,
            cls.sig_id_neutral_sign_flip,
        )

        return new_sig_id_array


class RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID(OfflineObservableBase):
    new_column_name: str = "__rdstar_reco_ch_and_sig_truth_charge_and_df_split_signal_component_id__"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    cross_feed_dict: Dict[int, Tuple[int, ...]] = {
        _old_sig_id_mapping["DSl"].id: RecoModeInfos.d_reco_with_dss_modes.ids,
        _old_sig_id_mapping["Dl"].id: RecoModeInfos.d_star_reco_with_dss_modes.ids,
    }

    charged_d_offsets: int = 0
    charged_dst_offsets: int = 1000
    neutral_d_offsets: int = 2000
    neutral_dst_offsets: int = 3000

    reco_id_offsets: Dict[int, Tuple[int, ...]] = {
        charged_d_offsets: RecoModeInfos.charged_b_to_d_with_dss_modes.ids,
        charged_dst_offsets: RecoModeInfos.charged_b_to_ds_with_dss_modes.ids,
        neutral_d_offsets: RecoModeInfos.neutral_b_to_d_with_dss_modes.ids,
        neutral_dst_offsets: RecoModeInfos.neutral_b_to_ds_with_dss_modes.ids,
    }

    charged_offsets: Tuple[int, ...] = (charged_d_offsets, charged_dst_offsets)
    neutral_offsets: Tuple[int, ...] = (neutral_d_offsets, neutral_dst_offsets)

    sig_id_neutral_sign_flip: int = -1
    cross_feed_offset: int = 500

    sig_id: str = "default_sig_id_col"
    decay_mode_id: str = "decay_mode_id"
    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    first_mc_b_pdg: str = "first_b_meson_pdg"
    second_mc_b_pdg: str = "second_b_meson_pdg"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        decay_mode_id: Cols.decay_mode_column_name,
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        first_mc_b_pdg: Cols.mc_b1_meson_pdg,
        second_mc_b_pdg: Cols.mc_b2_meson_pdg,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        _valid_values: List[int] = [
            int((v.id + o) * (cls.sig_id_neutral_sign_flip if sf and v.id < 5 else 1.0))
            for v in _old_sig_id_mapping.values()
            for o in cls.reco_id_offsets
            for sf in (True, False)
        ]

        _valid_cf_values: List[int] = [
            int((v + o + cls.cross_feed_offset) * (cls.sig_id_neutral_sign_flip if sf else 1.0))
            for v, cf_reco_ids in cls.cross_feed_dict.items()
            for o, o_reco_ids in cls.reco_id_offsets.items()
            for sf in (True, False)
            if set(cf_reco_ids).intersection(set(o_reco_ids))
        ]

        return tuple(set(_valid_values + _valid_cf_values))

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        pdg_1st_b: np.array = df[cls.required_variables[cls.first_mc_b_pdg]].values
        pdg_2nd_b: np.array = df[cls.required_variables[cls.second_mc_b_pdg]].values

        b_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_b, pdg_2nd_b],
            default=cls._no_match_value,
        )

        orig_sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        _is_sig_neutral: np.ndarray = np.abs(b_meson_pdg_array) == PcC.b_meson_neutral_pdg_id
        unchanged_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values() if v.id >= 5]
        must_stay_unchanged: np.ndarray = np.isin(element=orig_sig_id_array, test_elements=unchanged_sig_id_values)

        reco_mode_id_array: np.array = df[cls.required_variables[cls.decay_mode_id]].values

        _is_reco_charged_d: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.charged_d_offsets])
        _is_reco_charged_dst: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.charged_dst_offsets])
        _is_reco_neutral_d: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.neutral_d_offsets])
        _is_reco_neutral_dst: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.neutral_dst_offsets])

        assert not any(_is_reco_charged_d + _is_reco_charged_dst + _is_reco_neutral_d + _is_reco_neutral_dst > 1)

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_reco_charged_d,
                _is_reco_charged_dst,
                _is_reco_neutral_d,
                _is_reco_neutral_dst,
            ],
            choicelist=[
                sig_id_array + cls.charged_d_offsets,
                sig_id_array + cls.charged_dst_offsets,
                sig_id_array + cls.neutral_d_offsets,
                sig_id_array + cls.neutral_dst_offsets,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either charged or neutral ones...
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        for cf_sig_id, cf_reco_ch_ids in cls.cross_feed_dict.items():
            is_cf_sig: np.ndarray = orig_sig_id_array == cf_sig_id
            is_cf_reco_ch: np.ndarray = np.isin(element=reco_mode_id_array, test_elements=cf_reco_ch_ids)
            is_cf: np.ndarray = is_cf_sig & is_cf_reco_ch

            new_sig_id_array[is_cf] += cls.cross_feed_offset

        elements_to_receive_sign_flip: np.ndarray = _is_sig_neutral & (~must_stay_unchanged)
        new_sig_id_array[elements_to_receive_sign_flip] *= cls.sig_id_neutral_sign_flip

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            unchanged_sig_id_values,
            valid_new_sig_id_values,
            tuple(cls.reco_id_offsets.keys()),
            cls.sig_id_neutral_sign_flip,
        )

        return new_sig_id_array


class DssRecoChAndSigDTruthChargeFitSignalID(OfflineObservableBase):
    new_column_name: str = "__dss_reco_ch_and_sigD_truth_charge_signal_component_id__"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    charged_d_offsets: int = 0
    charged_dst_offsets: int = 1000
    neutral_d_offsets: int = 2000
    neutral_dst_offsets: int = 3000

    reco_id_offsets: Dict[int, Tuple[int, ...]] = {
        charged_d_offsets: RecoModeInfos.charged_d_wwo_pi_reco_modes.ids,
        charged_dst_offsets: RecoModeInfos.charged_ds_wwo_pi_reco_modes.ids,
        neutral_d_offsets: RecoModeInfos.neutral_d_wwo_pi_reco_modes.ids,
        neutral_dst_offsets: RecoModeInfos.neutral_ds_wwo_pi_reco_modes.ids,
    }

    charged_offsets: Tuple[int, ...] = (charged_d_offsets, charged_dst_offsets)
    neutral_offsets: Tuple[int, ...] = (neutral_d_offsets, neutral_dst_offsets)

    sig_id_neutral_sign_flip: int = -1

    sig_id: str = "default_sig_id_col"
    decay_mode_id: str = "decay_mode_id"
    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    first_mc_d_pdg: str = "first_d_meson_pdg"
    second_mc_d_pdg: str = "second_d_meson_pdg"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        decay_mode_id: Cols.decay_mode_column_name,
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        first_mc_d_pdg: Cols.mc_b1_1st_degree_d_meson_pdg,
        second_mc_d_pdg: Cols.mc_b2_1st_degree_d_meson_pdg,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        _valid_values: List[int] = [
            int((v.id + o) * (cls.sig_id_neutral_sign_flip if sf and v.id < 7 else 1.0))
            for v in _old_sig_id_mapping.values()
            for o in cls.reco_id_offsets
            for sf in (True, False)
        ]

        return tuple(set(_valid_values))

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        pdg_1st_d: np.array = df[cls.required_variables[cls.first_mc_d_pdg]].values
        pdg_2nd_d: np.array = df[cls.required_variables[cls.second_mc_d_pdg]].values

        d_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_d, pdg_2nd_d],
            default=cls._no_match_value,
        )

        abs_d_meson_pdg: np.array = np.abs(d_meson_pdg_array)
        _is_sig_d_neutral: np.ndarray = np.isin(element=abs_d_meson_pdg, test_elements=PcC.all_neutral_d_pdg_ids)
        _is_sig_d_charged: np.ndarray = np.isin(element=abs_d_meson_pdg, test_elements=PcC.all_charged_d_pdg_ids)

        assert not np.any(_is_sig_d_neutral & _is_sig_d_charged), np.sum(_is_sig_d_neutral & _is_sig_d_charged)

        orig_sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        unchanged_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values() if v.id >= 7]
        must_stay_unchanged: np.ndarray = np.isin(element=orig_sig_id_array, test_elements=unchanged_sig_id_values)

        reco_mode_id_array: np.array = df[cls.required_variables[cls.decay_mode_id]].values

        _is_reco_charged_d: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.charged_d_offsets])
        _is_reco_charged_dst: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.charged_dst_offsets])
        _is_reco_neutral_d: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.neutral_d_offsets])
        _is_reco_neutral_dst: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.neutral_dst_offsets])

        assert not any(_is_reco_charged_d + _is_reco_charged_dst + _is_reco_neutral_d + _is_reco_neutral_dst > 1)

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_reco_charged_d,
                _is_reco_charged_dst,
                _is_reco_neutral_d,
                _is_reco_neutral_dst,
            ],
            choicelist=[
                sig_id_array + cls.charged_d_offsets,
                sig_id_array + cls.charged_dst_offsets,
                sig_id_array + cls.neutral_d_offsets,
                sig_id_array + cls.neutral_dst_offsets,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either charged or neutral ones...
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        elements_to_receive_sign_flip: np.ndarray = _is_sig_d_neutral & (~must_stay_unchanged)
        new_sig_id_array[elements_to_receive_sign_flip] *= cls.sig_id_neutral_sign_flip

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            unchanged_sig_id_values,
            valid_new_sig_id_values,
            tuple(cls.reco_id_offsets.keys()),
            cls.sig_id_neutral_sign_flip,
        )

        return new_sig_id_array


class DssRecoChAndMainSigDTruthChargeFitSignalID(OfflineObservableBase):
    new_column_name: str = "__dss_reco_ch_and_main_sigD_truth_charge_signal_component_id__"

    _sig_b_mdst_indices: Tuple[int, ...] = (1, 2)
    _no_match_index: int = -999
    _no_match_value: int = -999

    wide_dss_offset: int = 10000
    narrow_dss_offset: int = 20000

    no_dss_offset: int = 0
    all_dss_offsets: Tuple[int, ...] = (no_dss_offset, wide_dss_offset, narrow_dss_offset)

    true_cd_offset: int = 100
    true_nd_offset: int = 200
    true_cdst_offset: int = 300
    true_ndst_offset: int = 400

    true_charged_offsets: Tuple[int, ...] = (true_cd_offset, true_cdst_offset)
    true_neutral_offsets: Tuple[int, ...] = (true_nd_offset, true_ndst_offset)
    true_d_offsets: Tuple[int, ...] = (true_cd_offset, true_nd_offset)
    true_dst_offsets: Tuple[int, ...] = (true_cdst_offset, true_ndst_offset)

    true_d_id_offsets: Dict[int, int] = {
        true_cd_offset: PcC.d_charged_pdg_id,
        true_nd_offset: PcC.d_neutral_pdg_id,
        true_cdst_offset: PcC.dst_charged_pdg_id,
        true_ndst_offset: PcC.dst_neutral_pdg_id,
    }

    reco_d_charged_pi_offset: int = 0
    reco_dst_charged_pi_offset: int = 1000
    reco_d_neutral_pi_offset: int = 2000
    reco_dst_neutral_pi_offset: int = 3000
    reco_others_offset: int = 4000

    reco_id_offsets: Dict[int, Tuple[int, ...]] = {
        reco_d_charged_pi_offset: RecoModeInfos.dss_d_l_charged_pion_modes.ids,
        reco_dst_charged_pi_offset: RecoModeInfos.dss_ds_l_charged_pion_modes.ids,
        reco_d_neutral_pi_offset: RecoModeInfos.dss_d_l_neutral_pion_modes.ids,
        reco_dst_neutral_pi_offset: RecoModeInfos.dss_ds_l_neutral_pion_modes.ids,
        reco_others_offset: RecoModeInfos.main_reco_modes.ids + RecoModeInfos.pi_l_nu_reco_modes.ids,
    }

    reco_charged_pi_offsets: Tuple[int, ...] = (reco_d_charged_pi_offset, reco_dst_charged_pi_offset)
    reco_neutral_pi_offsets: Tuple[int, ...] = (reco_d_neutral_pi_offset, reco_dst_neutral_pi_offset)

    reco_d_offsets: Tuple[int, ...] = (reco_d_charged_pi_offset, reco_d_neutral_pi_offset)
    reco_dst_offsets: Tuple[int, ...] = (reco_dst_charged_pi_offset, reco_dst_neutral_pi_offset)

    reco_all_offsets: Tuple[int, ...] = reco_d_offsets + reco_dst_offsets

    sig_id_neutral_sign_flip: int = -1

    sig_id: str = "default_sig_id_col"
    decay_mode_id: str = "decay_mode_id"
    sig_b_mdst_index: str = "sig_b_mdst_index"
    sig_b_via_tau_mdst_index: str = "sig_b_via_tau_mdst_index"
    first_mc_d_pdg: str = "first_d_meson_pdg"
    second_mc_d_pdg: str = "second_d_meson_pdg"
    first_2nd_degree_mc_d_pdg: str = "first_2nd_d_meson_pdg"
    second_2nd_degree_mc_d_pdg: str = "second_2nd_d_meson_pdg"
    first_3rd_degree_mc_d_pdg: str = "first_3rd_d_meson_pdg"
    second_3rd_degree_mc_d_pdg: str = "second_3rd_d_meson_pdg"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        decay_mode_id: Cols.decay_mode_column_name,
        sig_b_mdst_index: Cols.mdst_index_via_lepton_col,
        sig_b_via_tau_mdst_index: Cols.mdst_index_via_tau_col,
        first_mc_d_pdg: Cols.mc_b1_1st_degree_d_meson_pdg,
        second_mc_d_pdg: Cols.mc_b2_1st_degree_d_meson_pdg,
        first_2nd_degree_mc_d_pdg: Cols.mc_b1_2nd_degree_d_meson_pdg,
        second_2nd_degree_mc_d_pdg: Cols.mc_b2_2nd_degree_d_meson_pdg,
        first_3rd_degree_mc_d_pdg: Cols.mc_b1_3rd_degree_d_meson_pdg,
        second_3rd_degree_mc_d_pdg: Cols.mc_b2_3rd_degree_d_meson_pdg,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        _valid_values: List[int] = [
            int(v.id + (t_o if v.id < 7 else 0) + r_o + dss_o)
            for v in _old_sig_id_mapping.values()
            for t_o in cls.true_d_id_offsets
            for r_o in cls.reco_id_offsets
            for dss_o in cls.all_dss_offsets
        ]

        return tuple(set(_valid_values))

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        sig_b_mdst: np.array = df[cls.required_variables[cls.sig_b_mdst_index]].values
        sig_b_via_tau_mdst: np.array = df[cls.required_variables[cls.sig_b_via_tau_mdst_index]].values

        is_sig_b: np.array = np.isin(element=sig_b_mdst, test_elements=cls._sig_b_mdst_indices)
        is_sig_b_via_tau: np.array = np.isin(element=sig_b_via_tau_mdst, test_elements=cls._sig_b_mdst_indices)

        assert not np.any(is_sig_b & is_sig_b_via_tau), np.sum(is_sig_b & is_sig_b_via_tau)

        b_mdst_index_array: np.array = np.select(
            condlist=[is_sig_b, is_sig_b_via_tau],
            choicelist=[sig_b_mdst, sig_b_via_tau_mdst],
            default=cls._no_match_index,
        )

        _unique_indices: np.array = np.unique(b_mdst_index_array)
        _valid_indices: List[int] = list(cls._sig_b_mdst_indices) + [cls._no_match_index]
        assert np.all(np.isin(element=_unique_indices, test_elements=_valid_indices)), (_unique_indices, _valid_indices)

        pdg_1st_primary_d: np.array = df[cls.required_variables[cls.first_mc_d_pdg]].values
        pdg_2nd_primary_d: np.array = df[cls.required_variables[cls.second_mc_d_pdg]].values

        pdg_1st_secondary_d: np.array = df[cls.required_variables[cls.first_2nd_degree_mc_d_pdg]].values
        pdg_2nd_secondary_d: np.array = df[cls.required_variables[cls.second_2nd_degree_mc_d_pdg]].values

        pdg_1st_tertiary_d: np.array = df[cls.required_variables[cls.first_2nd_degree_mc_d_pdg]].values
        pdg_2nd_tertiary_d: np.array = df[cls.required_variables[cls.second_2nd_degree_mc_d_pdg]].values

        primary_d_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_primary_d, pdg_2nd_primary_d],
            default=cls._no_match_value,
        )

        secondary_d_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_secondary_d, pdg_2nd_secondary_d],
            default=cls._no_match_value,
        )

        tertiary_d_meson_pdg_array: np.array = np.select(
            condlist=[b_mdst_index_array == 1, b_mdst_index_array == 2],
            choicelist=[pdg_1st_tertiary_d, pdg_2nd_tertiary_d],
            default=cls._no_match_value,
        )

        primary_abs_d_pdg: np.array = abs(primary_d_meson_pdg_array)
        secondary_abs_d_pdg: np.array = abs(secondary_d_meson_pdg_array)
        tertiary_abs_d_pdg: np.array = abs(tertiary_d_meson_pdg_array)

        primary_d_is_main: np.array = np.isin(element=primary_abs_d_pdg, test_elements=PcC.d_wwo_st_pdg_ids)
        secondary_d_could_be_main: np.array = np.isin(element=secondary_abs_d_pdg, test_elements=PcC.d_wwo_st_pdg_ids)
        secondary_d_is_main: np.array = secondary_d_could_be_main & ~primary_d_is_main
        tertiary_d_could_be_main: np.array = np.isin(element=tertiary_abs_d_pdg, test_elements=PcC.d_wwo_st_pdg_ids)
        tertiary_d_is_main: np.array = tertiary_d_could_be_main & ~secondary_d_is_main

        main_d_pdg: np.array = np.select(
            condlist=[
                primary_d_is_main,
                secondary_d_is_main,
                tertiary_d_is_main,
            ],
            choicelist=[
                primary_d_meson_pdg_array,
                secondary_d_meson_pdg_array,
                tertiary_d_meson_pdg_array,
            ],
            default=np.nan,  # This is the case, if no true d meson is available
        )

        _is_sig_cd: np.ndarray = np.abs(main_d_pdg) == cls.true_d_id_offsets[cls.true_cd_offset]
        _is_sig_nd: np.ndarray = np.abs(main_d_pdg) == cls.true_d_id_offsets[cls.true_nd_offset]
        _is_sig_cdst: np.ndarray = np.abs(main_d_pdg) == cls.true_d_id_offsets[cls.true_cdst_offset]
        _is_sig_ndst: np.ndarray = np.abs(main_d_pdg) == cls.true_d_id_offsets[cls.true_ndst_offset]

        _all_d_cond_and_combined: np.array = _is_sig_cd & _is_sig_nd & _is_sig_cdst & _is_sig_ndst
        assert not np.any(_all_d_cond_and_combined), np.sum(_all_d_cond_and_combined)

        orig_sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        unchanged_sig_id_values: List[int] = [v.id for v in _old_sig_id_mapping.values() if v.id >= 7]
        must_stay_unchanged: np.ndarray = np.isin(element=orig_sig_id_array, test_elements=unchanged_sig_id_values)

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        sig_id_array_with_true_d_info: np.array = np.select(
            condlist=[
                _is_sig_cd & ~must_stay_unchanged,
                _is_sig_nd & ~must_stay_unchanged,
                _is_sig_cdst & ~must_stay_unchanged,
                _is_sig_ndst & ~must_stay_unchanged,
            ],
            choicelist=[
                sig_id_array + cls.true_cd_offset,
                sig_id_array + cls.true_nd_offset,
                sig_id_array + cls.true_cdst_offset,
                sig_id_array + cls.true_ndst_offset,
            ],
            default=sig_id_array,  # If there is no true main D meson, no offset is added...
        )

        true_sig_d_meson_abs_pdg: np.array = np.abs(primary_d_meson_pdg_array)

        _is_wide_dss: np.ndarray = np.isin(element=true_sig_d_meson_abs_pdg, test_elements=PcC.dstst_wide_mass_pdg_ids)
        _is_narrow_dss: np.ndarray = np.isin(
            element=true_sig_d_meson_abs_pdg,
            test_elements=PcC.dstst_narrow_mass_pdg_ids,
        )

        assert not np.any(_is_wide_dss & _is_narrow_dss), np.sum(_is_wide_dss & _is_narrow_dss)

        sig_id_array_with_true_d_info[_is_wide_dss] += cls.wide_dss_offset
        sig_id_array_with_true_d_info[_is_narrow_dss] += cls.narrow_dss_offset

        reco_mode_id_array: np.array = df[cls.required_variables[cls.decay_mode_id]].values

        _is_reco_d_cpi: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.reco_d_charged_pi_offset])
        _is_reco_dst_cpi: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.reco_dst_charged_pi_offset])
        _is_reco_d_npi: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.reco_d_neutral_pi_offset])
        _is_reco_dst_npi: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.reco_dst_neutral_pi_offset])

        _is_reco_others: np.ndarray = np.isin(reco_mode_id_array, cls.reco_id_offsets[cls.reco_others_offset])

        _all_reco_conditions: np.array = _is_reco_d_cpi & _is_reco_dst_cpi & _is_reco_d_npi & _is_reco_dst_npi
        assert not np.any(_all_reco_conditions), np.sum(_all_reco_conditions)

        _all_and_others_reco_conditions: np.array = _all_reco_conditions & _is_reco_others
        assert not np.any(_all_and_others_reco_conditions), np.sum(_all_and_others_reco_conditions)

        sig_id_array_with_reco_info: np.array = np.select(
            condlist=[
                _is_reco_d_cpi,
                _is_reco_dst_cpi,
                _is_reco_d_npi,
                _is_reco_dst_npi,
                _is_reco_others,
            ],
            choicelist=[
                sig_id_array_with_true_d_info + cls.reco_d_charged_pi_offset,
                sig_id_array_with_true_d_info + cls.reco_dst_charged_pi_offset,
                sig_id_array_with_true_d_info + cls.reco_d_neutral_pi_offset,
                sig_id_array_with_true_d_info + cls.reco_dst_neutral_pi_offset,
                sig_id_array_with_true_d_info + cls.reco_others_offset,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either D** reco modes or others...
        )

        assert not np.any(np.isnan(sig_id_array_with_reco_info)), (
            sum(np.isnan(sig_id_array_with_reco_info)),
            np.unique(sig_id_array_with_reco_info),
        )

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=sig_id_array_with_reco_info, test_elements=valid_new_sig_id_values)), (
            list(np.unique(sig_id_array_with_reco_info)),
            unchanged_sig_id_values,
            valid_new_sig_id_values,
            tuple(cls.reco_id_offsets.keys()),
            cls.sig_id_neutral_sign_flip,
        )

        return sig_id_array_with_reco_info


class FEIBTagRecoChAndSignalID(OfflineObservableBase):
    new_column_name: str = "__fei_btag_reco_ch_and_signal_component_id__"

    sig_id_neutral_b_tag_sign_flip: int = -1

    b_to_d_1cpi_tag_offset: int = 0
    b_to_d_mcpi_tag_offset: int = 100
    b_to_ocd_pipi_tag_offset: int = 200
    b_to_dd_tag_offset: int = 300
    b_to_ds_1cpi_tag_offset: int = 400
    b_to_ds_mcpi_tag_offset: int = 500
    b_to_dsd_tag_offset: int = 600
    b_to_j_psi_tag_offset: int = 700

    reco_id_offsets: Dict[int, Tuple[int, ...]] = {
        b_to_d_1cpi_tag_offset: TagRecoModeInfos.b_to_d_1cpi_tag_modes.ids,
        b_to_d_mcpi_tag_offset: TagRecoModeInfos.b_to_d_mcpi_tag_modes.ids,
        b_to_ocd_pipi_tag_offset: TagRecoModeInfos.b_to_ocd_pipi_tag_modes.ids,
        b_to_dd_tag_offset: TagRecoModeInfos.b_to_dd_tag_modes.ids,
        b_to_ds_1cpi_tag_offset: TagRecoModeInfos.b_to_ds_1cpi_tag_modes.ids,
        b_to_ds_mcpi_tag_offset: TagRecoModeInfos.b_to_ds_mcpi_tag_modes.ids,
        b_to_dsd_tag_offset: TagRecoModeInfos.b_to_dsd_tag_modes.ids,
        b_to_j_psi_tag_offset: TagRecoModeInfos.b_to_j_psi_tag_modes.ids,
    }

    all_offset_values: Tuple[int, ...] = tuple(k for k in reco_id_offsets)

    sig_id: str = "default_sig_id_col"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        _valid_values: List[int] = [
            int((v.id + o) * (cls.sig_id_neutral_b_tag_sign_flip if sf else 1.0))
            for v in _old_sig_id_mapping.values()
            for o in cls.reco_id_offsets
            for sf in (True, False)
        ]

        return tuple(set(_valid_values))

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        fei_reco_mode_id: np.array = TagRecoModeInfos.calculate_new_fei_reco_mode_ids(df=df)

        _is_b_to_d_1cpi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_d_1cpi_tag_offset])
        _is_b_to_d_mcpi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_d_mcpi_tag_offset])
        _is_b_to_ocd_pipi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_ocd_pipi_tag_offset])
        _is_b_to_dd_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_dd_tag_offset])
        _is_b_to_ds_1cpi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_ds_1cpi_tag_offset])
        _is_b_to_ds_mcpi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_ds_mcpi_tag_offset])
        _is_b_to_dsd_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_dsd_tag_offset])
        _is_b_to_j_psi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_j_psi_tag_offset])

        assert not any(
            _is_b_to_d_1cpi_tag
            + _is_b_to_d_mcpi_tag
            + _is_b_to_ocd_pipi_tag
            + _is_b_to_dd_tag
            + _is_b_to_ds_1cpi_tag
            + _is_b_to_ds_mcpi_tag
            + _is_b_to_dsd_tag
            + _is_b_to_j_psi_tag
            > 1
        )

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_b_to_d_1cpi_tag,
                _is_b_to_d_mcpi_tag,
                _is_b_to_ocd_pipi_tag,
                _is_b_to_dd_tag,
                _is_b_to_ds_1cpi_tag,
                _is_b_to_ds_mcpi_tag,
                _is_b_to_dsd_tag,
                _is_b_to_j_psi_tag,
            ],
            choicelist=[
                sig_id_array + cls.b_to_d_1cpi_tag_offset,
                sig_id_array + cls.b_to_d_mcpi_tag_offset,
                sig_id_array + cls.b_to_ocd_pipi_tag_offset,
                sig_id_array + cls.b_to_dd_tag_offset,
                sig_id_array + cls.b_to_ds_1cpi_tag_offset,
                sig_id_array + cls.b_to_ds_mcpi_tag_offset,
                sig_id_array + cls.b_to_dsd_tag_offset,
                sig_id_array + cls.b_to_j_psi_tag_offset,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either charged or neutral ones...
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        elements_with_neutral_b_tag_reco_modes: np.ndarray = fei_reco_mode_id < 0
        new_sig_id_array[elements_with_neutral_b_tag_reco_modes] *= cls.sig_id_neutral_b_tag_sign_flip

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            valid_new_sig_id_values,
            tuple(cls.reco_id_offsets.keys()),
        )

        return new_sig_id_array


class FEIBTagAndSigDRecoChAndSignalID(OfflineObservableBase):
    new_column_name: str = "__fei_btag_and_sig_d_reco_ch_and_signal_component_id__"

    sig_id_neutral_b_tag_sign_flip: int = -1

    sig_d_star_offset: int = 1000

    b_to_d_1cpi_tag_offset: int = 0
    b_to_d_mcpi_tag_offset: int = 100
    b_to_ocd_pipi_tag_offset: int = 200
    b_to_dd_tag_offset: int = 300
    b_to_ds_1cpi_tag_offset: int = 400
    b_to_ds_mcpi_tag_offset: int = 500
    b_to_dsd_tag_offset: int = 600
    b_to_j_psi_tag_offset: int = 700

    reco_id_offsets: Dict[int, Tuple[int, ...]] = {
        b_to_d_1cpi_tag_offset: TagRecoModeInfos.b_to_d_1cpi_tag_modes.ids,
        b_to_d_mcpi_tag_offset: TagRecoModeInfos.b_to_d_mcpi_tag_modes.ids,
        b_to_ocd_pipi_tag_offset: TagRecoModeInfos.b_to_ocd_pipi_tag_modes.ids,
        b_to_dd_tag_offset: TagRecoModeInfos.b_to_dd_tag_modes.ids,
        b_to_ds_1cpi_tag_offset: TagRecoModeInfos.b_to_ds_1cpi_tag_modes.ids,
        b_to_ds_mcpi_tag_offset: TagRecoModeInfos.b_to_ds_mcpi_tag_modes.ids,
        b_to_dsd_tag_offset: TagRecoModeInfos.b_to_dsd_tag_modes.ids,
        b_to_j_psi_tag_offset: TagRecoModeInfos.b_to_j_psi_tag_modes.ids,
    }

    all_offset_values: Tuple[int, ...] = tuple(k for k in reco_id_offsets)

    sig_b_via_ds_reco_mode_ids: Tuple[int, ...] = RecoModeInfos.d_star_reco_with_dss_modes.ids

    sig_id: str = "default_sig_id_col"
    sig_b_reco_mode: str = "decay_mode_id"

    required_variables: Dict[str, str] = {
        sig_id: Cols.base_sig_id_col,
        sig_b_reco_mode: Cols.decay_mode_column_name,
    }

    @classmethod
    def _get_valid_new_values(cls) -> Tuple[int, ...]:
        _valid_values: List[int] = [
            int((v.id + o + ds_o) * (cls.sig_id_neutral_b_tag_sign_flip if sf else 1.0))
            for v in _old_sig_id_mapping.values()
            for o in cls.reco_id_offsets
            for ds_o in (0, cls.sig_d_star_offset)
            for sf in (True, False)
        ]

        return tuple(set(_valid_values))

    @classmethod
    def calculate_new_column_values(
        cls,
        df: pd.DataFrame,
        check_range: bool = True,
    ) -> Union[pd.Series, np.array]:

        fei_reco_mode_id: np.array = TagRecoModeInfos.calculate_new_fei_reco_mode_ids(df=df)

        _is_b_to_d_1cpi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_d_1cpi_tag_offset])
        _is_b_to_d_mcpi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_d_mcpi_tag_offset])
        _is_b_to_ocd_pipi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_ocd_pipi_tag_offset])
        _is_b_to_dd_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_dd_tag_offset])
        _is_b_to_ds_1cpi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_ds_1cpi_tag_offset])
        _is_b_to_ds_mcpi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_ds_mcpi_tag_offset])
        _is_b_to_dsd_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_dsd_tag_offset])
        _is_b_to_j_psi_tag: np.ndarray = np.isin(fei_reco_mode_id, cls.reco_id_offsets[cls.b_to_j_psi_tag_offset])

        assert not any(
            _is_b_to_d_1cpi_tag
            + _is_b_to_d_mcpi_tag
            + _is_b_to_ocd_pipi_tag
            + _is_b_to_dd_tag
            + _is_b_to_ds_1cpi_tag
            + _is_b_to_ds_mcpi_tag
            + _is_b_to_dsd_tag
            + _is_b_to_j_psi_tag
            > 1
        )

        sig_b_reco_mode_array: np.array = df[cls.required_variables[cls.sig_b_reco_mode]].values

        sig_id_array: np.array = df[cls.required_variables[cls.sig_id]].values

        new_sig_id_array: np.array = np.select(
            condlist=[
                _is_b_to_d_1cpi_tag,
                _is_b_to_d_mcpi_tag,
                _is_b_to_ocd_pipi_tag,
                _is_b_to_dd_tag,
                _is_b_to_ds_1cpi_tag,
                _is_b_to_ds_mcpi_tag,
                _is_b_to_dsd_tag,
                _is_b_to_j_psi_tag,
            ],
            choicelist=[
                sig_id_array + cls.b_to_d_1cpi_tag_offset,
                sig_id_array + cls.b_to_d_mcpi_tag_offset,
                sig_id_array + cls.b_to_ocd_pipi_tag_offset,
                sig_id_array + cls.b_to_dd_tag_offset,
                sig_id_array + cls.b_to_ds_1cpi_tag_offset,
                sig_id_array + cls.b_to_ds_mcpi_tag_offset,
                sig_id_array + cls.b_to_dsd_tag_offset,
                sig_id_array + cls.b_to_j_psi_tag_offset,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either charged or neutral ones...
        )

        assert not np.any(np.isnan(new_sig_id_array)), (sum(np.isnan(new_sig_id_array)), np.unique(new_sig_id_array))

        is_sig_b_via_ds_reco_mode: np.ndarray = np.isin(sig_b_reco_mode_array, cls.sig_b_via_ds_reco_mode_ids)
        new_sig_id_array[is_sig_b_via_ds_reco_mode] += cls.sig_d_star_offset

        elements_with_neutral_b_tag_reco_modes: np.ndarray = fei_reco_mode_id < 0
        new_sig_id_array[elements_with_neutral_b_tag_reco_modes] *= cls.sig_id_neutral_b_tag_sign_flip

        valid_new_sig_id_values: Tuple[int, ...] = cls._get_valid_new_values()

        assert np.all(np.isin(element=new_sig_id_array, test_elements=valid_new_sig_id_values)), (
            list(np.unique(new_sig_id_array)),
            valid_new_sig_id_values,
            tuple(cls.reco_id_offsets.keys()),
        )

        return new_sig_id_array


class OfflineMCObservables:

    definitions = {
        TruthSigSideBMesonCharge.get_new_column_name(): TruthSigSideBMesonCharge,
        TruthSigSideBMesonIsCharged.get_new_column_name(): TruthSigSideBMesonIsCharged,
        RDStarFitSignalID.get_new_column_name(): RDStarFitSignalID,
        RDStarFitSignalIDwithRecoModeInfo.get_new_column_name(): RDStarFitSignalIDwithRecoModeInfo,
        RDStarTruthChargeSplitFitSignalID.get_new_column_name(): RDStarTruthChargeSplitFitSignalID,
        RDStarRecoChargeSplitFitSignalID.get_new_column_name(): RDStarRecoChargeSplitFitSignalID,
        RDStarRecoAndSigTruthChargeSplitFitSignalID.get_new_column_name(): RDStarRecoAndSigTruthChargeSplitFitSignalID,
        RDStarRecoAndSigTruthChargeAndDFSplitFitSignalID.get_new_column_name(): RDStarRecoAndSigTruthChargeAndDFSplitFitSignalID,
        RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID.get_new_column_name(): RDStarRecoChAndSigTruthChargeAndDFSplitFitSignalID,
        DssRecoChAndSigDTruthChargeFitSignalID.get_new_column_name(): DssRecoChAndSigDTruthChargeFitSignalID,
        DssRecoChAndMainSigDTruthChargeFitSignalID.get_new_column_name(): DssRecoChAndMainSigDTruthChargeFitSignalID,
        FEIBTagRecoChAndSignalID.get_new_column_name(): FEIBTagRecoChAndSignalID,
        FEIBTagAndSigDRecoChAndSignalID.get_new_column_name(): FEIBTagAndSigDRecoChAndSignalID,
    }  # type: Dict[str, Type[OfflineObservableBase]]

    required_variables = tuple(
        [c for d in definitions.values() for c in d.get_required_columns()]
    )  # type: Tuple[str, ...]

    @staticmethod
    def add_kinematic_observables_definitions(df: pd.DataFrame, check_range: bool = True) -> None:
        assert all(c in df.columns for c in OfflineMCObservables.get_required_columns_list()), [
            c for c in OfflineMCObservables.get_required_columns_list() if c not in df.columns
        ]

        assert all(c not in df.columns for c in OfflineMCObservables.get_new_columns_list()), [
            c for c in OfflineMCObservables.get_new_columns_list() if c in df.columns
        ]

        for new_col, definition_class in OfflineMCObservables.definitions.items():
            new_values = definition_class.calculate_new_column_values(df=df)
            assert len(new_values) == len(df), (len(new_values), len(df), new_col)
            df.loc[:, new_col] = new_values

            assert not pd.isnull(df[new_col]).any(), new_col

    @staticmethod
    def get_definition_function(
        name: str,
        fail_when_exists: bool = True,
    ) -> Callable:
        def definition_function(
            df: pd.DataFrame,
            check_range: bool = True,
            accept_existing_column: bool = False,
        ) -> None:
            definition = OfflineMCObservables.definitions[name]  # type: Type[OfflineObservableBase]
            new_col_name = definition.get_new_column_name()

            if fail_when_exists and not accept_existing_column:
                assert new_col_name not in df.columns, new_col_name
            else:
                if new_col_name in df.columns:
                    return

            assert all(c in df.columns for c in definition.get_required_columns()), [
                c for c in definition.get_required_columns() if c not in df.columns
            ]

            if len(np.unique(df.index)) == 0:
                df = df.assign(**{new_col_name: df[definition.get_required_columns()[0]]})
            else:
                new_values = definition.calculate_new_column_values(df=df)
                assert len(new_values) == len(df), (len(new_values), len(df), new_col_name)
                df.loc[:, new_col_name] = new_values

            assert not pd.isnull(df[new_col_name]).any(), new_col_name

        return definition_function

    @staticmethod
    def get_required_columns_list() -> List[str]:
        return list(OfflineMCObservables.required_variables)

    @staticmethod
    def get_new_columns_list() -> List[str]:
        return list(OfflineMCObservables.definitions.keys())

    @staticmethod
    def get_mc_obs_definition(name: str) -> Type[OfflineObservableBase]:
        return OfflineMCObservables.definitions[name]

    @staticmethod
    def get_definition_names() -> Tuple[str, ...]:
        return tuple(OfflineMCObservables.definitions.keys())
