"""
Weights for RDStar analysis samples

Felix Metzner 2019
"""

import math
import pandas as pd

from typing import Optional, Dict, NamedTuple

from analysistools.offline_analysis import sample_weights as sw
from analysistools.batch_tools.file_handler import (
    Belle_url_mc_types,
    Belle_list_types,
    Belle_data_types,
    off_resonance_types,
)

from rdstar.utilities import PathType
from rdstar.branching_fractions import b_meson_bf_calculation as bf_calc
from rdstar.branching_fractions.bf_utilities import BFEntryType, BFSDType, update_branching_fraction


__all__ = [
    "scale_weight_column_name",
    "off_to_on_res_scale_weight_column_name",
    "signal_mc_bfs",
    "signal_mc_evts",
    "add_scaling_weights",
    "add_off_to_on_resonance_scaling_weights",
]


scale_weight_column_name = sw.scale_weight_column_name  # type: str
off_to_on_res_scale_weight_column_name = sw.off_to_on_resonance_scale_weight_column_name


class NEventEntryType(NamedTuple):
    prod: int
    meta: int
    summary: int


EvtNumDictType = Dict[str, NEventEntryType]


# fmt: off
sig_mc_bfs = {
    # B0
    "DSME": BFEntryType(bf_value=bf_calc.bf_b0_to_d_st_l_nu_avg, charged_b=False, decay_descr="B0 -> D*- e+ nu_e"),
    "DME": BFEntryType(bf_value=bf_calc.bf_b0_to_d_l_nu_avg, charged_b=False, decay_descr="B0 -> D- e+ nu_e"),
    "BZD1E": BFEntryType(bf_value=bf_calc.bf_b0_to_d_1_l_nu_our_cal, charged_b=False, decay_descr="B0 -> D_1- e+ nu_e"),
    "BZD0sE": BFEntryType(bf_value=bf_calc.bf_b0_to_d0st_l_nu_our_cal, charged_b=False, decay_descr="B0 -> D_0*- e+ nu_e"),
    "BZDp1E": BFEntryType(bf_value=bf_calc.bf_b0_to_dprime1_l_nu_our_cal, charged_b=False, decay_descr="B0 -> D'_1- e+ nu_e"),
    "BZD2sE": BFEntryType(bf_value=bf_calc.bf_b0_to_dst_2_l_nu_our_cal, charged_b=False, decay_descr="B0 -> D_2*- e+ nu_e"),
    "DSMMu": BFEntryType(bf_value=bf_calc.bf_b0_to_d_st_l_nu_avg, charged_b=False, decay_descr="B0 -> D*- mu+ nu_mu"),
    "DMMu": BFEntryType(bf_value=bf_calc.bf_b0_to_d_l_nu_avg, charged_b=False, decay_descr="B0 -> D- mu+ nu_mu"),
    "BZD1Mu": BFEntryType(bf_value=bf_calc.bf_b0_to_d_1_l_nu_our_cal, charged_b=False, decay_descr="B0 -> D_1- mu+ nu_mu"),
    "BZD0sMu": BFEntryType(bf_value=bf_calc.bf_b0_to_d0st_l_nu_our_cal, charged_b=False, decay_descr="B0 -> D_0*- mu+ nu_mu"),
    "BZDp1Mu": BFEntryType(bf_value=bf_calc.bf_b0_to_dprime1_l_nu_our_cal, charged_b=False, decay_descr="B0 -> D'_1- mu+ nu_mu"),
    "BZD2sMu": BFEntryType(bf_value=bf_calc.bf_b0_to_dst_2_l_nu_our_cal, charged_b=False, decay_descr="B0 -> D_2*- mu+ nu_mu"),
    "DSMTau": BFEntryType(bf_value=bf_calc.bf_b0_to_d_st_tau_nu_avg, charged_b=False, decay_descr="B0 -> D*- tau+ nu_tau"),
    "DMTau": BFEntryType(bf_value=bf_calc.bf_b0_to_d_tau_nu_avg, charged_b=False, decay_descr="B0 -> D- tau+ nu_tau"),
    "TBZD1Tau": BFEntryType(bf_value=bf_calc.bf_b0_to_d_1_tau_nu_cal, charged_b=False, decay_descr="B0 -> D_1- tau+ nu_tau"),
    "TBZD0sTau": BFEntryType(bf_value=bf_calc.bf_b0_to_d0st_tau_nu_cal, charged_b=False, decay_descr="B0 -> D_0*- tau+ nu_tau"),
    "TBZDp1Tau": BFEntryType(bf_value=bf_calc.bf_b0_to_dprime1_tau_nu_cal, charged_b=False, decay_descr="B0 -> D'_1- tau+ nu_tau"),
    "TBZD2sTau": BFEntryType(bf_value=bf_calc.bf_b0_to_dst_2_tau_nu_cal, charged_b=False, decay_descr="B0 -> D_2*- tau+ nu_tau"),
    "GapD1Mixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap, charged_b=False, decay_descr="B0 -> D_1- (-> D pi pi) l+ nu_l"),
    "GapDEtaMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_eta_l_nu_gap, charged_b=False, decay_descr="B0 -> D- eta l+ nu_l"),
    "GapDstEtaMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_st_eta_l_nu_gap, charged_b=False, decay_descr="B0 -> D*- eta l+ nu_l"),
    "GapDPiPiMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_pi_pi_l_nu_our_gap, charged_b=False, decay_descr="B0 -> D pi pi l+ nu_l"),
    "GapDstPiPiMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_st_pi_pi_l_nu_our_gap, charged_b=False, decay_descr="B0 -> D* pi pi l+ nu_l"),
    "GapDEtaViaD0sMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_eta_l_nu_gap_via_d_st_0, charged_b=False, decay_descr="B0 -> D_0* (-> D eta) l+ nu_l"),  # using BF(BZD0sE) / (BF(BZD0sE) + BF(BZDp1E)) * BF(B to D(st) eta l nu)
    "GapDstEtaViaDp1Mixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_st_eta_l_nu_gap_via_d_prime_1, charged_b=False, decay_descr="B0 -> D'1 (-> D* eta) l+ nu_l"),  # using BF(BZDp1E) / (BF(BZD0sE) + BF(BZDp1E)) * BF(B to D(st) eta l nu)
    "GapDPiPiViaDp1Mixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_pi_pi_l_nu_gap_via_d_prime_1, charged_b=False, decay_descr="B0 -> D'1 (-> D pi pi) l+ nu_l"),  # same calculation as above
    "GapDPiPiViaD0sMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_pi_pi_l_nu_gap_via_d_st_0, charged_b=False, decay_descr="B0 -> D_0s (-> D pi pi) l+ nu_l"),  # same calculation as above
    "GapDstPiPiViaDp1Mixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_prime_1, charged_b=False, decay_descr="B0 -> D'1 (-> D* pi pi) l+ nu_l"),  # same calculation as above
    "GapDstPiPiViaD0sMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_d_st_pi_pi_l_nu_gap_via_d_st_0, charged_b=False, decay_descr="B0 -> D_0s (-> D* pi pi) l+ nu_l"),  # same calculation as above
    "XuLNuIncMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_xu_l_nu_inclusive, charged_b=False, decay_descr="B0 -> Xu- l+ nu_l"),
    "XuLNuResMixed": BFEntryType(bf_value=2 * bf_calc.bf_b0_to_xu_l_nu_resonant, charged_b=False, decay_descr="B0 -> (pi- [0.000150], rho- [0.000294]) l+ nu_l"),
    # B+
    "DSZE": BFEntryType(bf_value=bf_calc.bf_bp_to_d_st_l_nu_avg, charged_b=True, decay_descr="B+ -> anti-D*0 e+ nu_e"),
    "DZE": BFEntryType(bf_value=bf_calc.bf_bp_to_d_l_nu_avg, charged_b=True, decay_descr="B+ -> anti-D0 e+ nu_e"),
    "BMD1E": BFEntryType(bf_value=bf_calc.bf_bp_to_d_1_l_nu_our_cal, charged_b=True, decay_descr="B+ -> anti-D_10 e+ nu_e"),
    "BMD0sE": BFEntryType(bf_value=bf_calc.bf_bp_to_d0st_l_nu_our_cal, charged_b=True, decay_descr="B+ -> anti-D_0*0 e+ nu_e"),
    "BMDp1E": BFEntryType(bf_value=bf_calc.bf_bp_to_dprime1_l_nu_our_cal, charged_b=True, decay_descr="B+ -> anti-D'_10 e+ nu_e"),
    "BMD2sE": BFEntryType(bf_value=bf_calc.bf_bp_to_dst_2_l_nu_our_cal, charged_b=True, decay_descr="B+ -> anti-D_2*0 e+ nu_e"),
    "DSZMu": BFEntryType(bf_value=bf_calc.bf_bp_to_d_st_l_nu_avg, charged_b=True, decay_descr="B+ -> anti-D*0 mu+ nu_mu"),
    "DZMu": BFEntryType(bf_value=bf_calc.bf_bp_to_d_l_nu_avg, charged_b=True, decay_descr="B+ -> anti-D0 mu+ nu_mu"),
    "BMD1Mu": BFEntryType(bf_value=bf_calc.bf_bp_to_d_1_l_nu_our_cal, charged_b=True, decay_descr="B+ -> anti-D_10 mu+ nu_mu"),
    "BMD0sMu": BFEntryType(bf_value=bf_calc.bf_bp_to_d0st_l_nu_our_cal, charged_b=True, decay_descr="B+ -> anti-D_0*0 mu+ nu_mu"),
    "BMDp1Mu": BFEntryType(bf_value=bf_calc.bf_bp_to_dprime1_l_nu_our_cal, charged_b=True, decay_descr="B+ -> anti-D'_10 mu+ nu_mu"),
    "BMD2sMu": BFEntryType(bf_value=bf_calc.bf_bp_to_dst_2_l_nu_our_cal, charged_b=True, decay_descr="B+ -> anti-D_2*0 mu+ nu_mu"),
    "DSZTau": BFEntryType(bf_value=bf_calc.bf_bp_to_d_st_tau_nu_avg, charged_b=True, decay_descr="B+ -> anti-D*0 tau+ nu_tau"),
    "DZTau": BFEntryType(bf_value=bf_calc.bf_bp_to_d_tau_nu_avg, charged_b=True, decay_descr="B+ -> anti-D0 tau+ nu_tau"),
    "TBMD1Tau": BFEntryType(bf_value=bf_calc.bf_bp_to_d_1_tau_nu_cal, charged_b=True, decay_descr="B+ -> anti-D_10 tau+ nu_tau"),
    "TBMD0sTau": BFEntryType(bf_value=bf_calc.bf_bp_to_d0st_tau_nu_cal, charged_b=True, decay_descr="B+ -> anti-D_0*0 tau+ nu_tau"),
    "TBMDp1Tau": BFEntryType(bf_value=bf_calc.bf_bp_to_dprime1_tau_nu_cal, charged_b=True, decay_descr="B+ -> anti-D'_10 tau+ nu_tau"),
    "TBMD2sTau": BFEntryType(bf_value=bf_calc.bf_bp_to_dst_2_tau_nu_cal, charged_b=True, decay_descr="B+ -> anti-D_2*0 tau+ nu_tau"),
    "GapD1Charged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_1_l_nu_with_d_1_to_d_pi_pi_gap, charged_b=True, decay_descr="B+ -> anti-D_10 (-> D pi pi) l+ nu_l"),
    "GapDEtaCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_eta_l_nu_gap, charged_b=True, decay_descr="B+ -> anti-D0 eta l+ nu_l"),
    "GapDstEtaCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_st_eta_l_nu_gap, charged_b=True, decay_descr="B+ -> anti-D*0 eta l+ nu_l"),
    "GapDPiPiCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_pi_pi_l_nu_our_gap, charged_b=True, decay_descr="B+ -> D pi pi l+ nu_l"),
    "GapDstPiPiCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_st_pi_pi_l_nu_our_gap, charged_b=True, decay_descr="B+ -> D* pi pi l+ nu_l"),
    "GapDEtaViaD0sCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_eta_l_nu_gap_via_d_st_0, charged_b=True, decay_descr="B+ -> D_0* (-> D eta) l+ nu_l"),  # using BF(BMD0sE) / (BF(BMD0sE) + BF(BMDp1E)) * BF(B to D(st) eta l nu)
    "GapDstEtaViaDp1Charged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_st_eta_l_nu_gap_via_d_prime_1, charged_b=True, decay_descr="B+ -> D'1 (-> D* eta) l+ nu_l"),  # using BF(BMDp1E) / (BF(BMD0sE) + BF(BMDp1E)) * BF(B to D(st) eta l nu)
    "GapDPiPiViaDp1Charged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_pi_pi_l_nu_gap_via_d_prime_1, charged_b=True, decay_descr="B+ -> D'1 (-> D pi pi) l+ nu_l"),  # same calculation as above
    "GapDPiPiViaD0sCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_pi_pi_l_nu_gap_via_d_st_0, charged_b=True, decay_descr="B+ -> D_0s (-> D pi pi) l+ nu_l"),  # same calculation as above
    "GapDstPiPiViaDp1Charged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_prime_1, charged_b=True, decay_descr="B+ -> D'1 (-> D* pi pi) l+ nu_l"),  # same calculation as above
    "GapDstPiPiViaD0sCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_d_st_pi_pi_l_nu_gap_via_d_st_0, charged_b=True, decay_descr="B+ -> D_0s (-> D* pi pi) l+ nu_l"),  # same calculation as above
    "XuLNuIncCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_xu_l_nu_inclusive, charged_b=True, decay_descr="B+ -> Xu0 l+ nu_l"),
    "XuLNuResCharged": BFEntryType(bf_value=2 * bf_calc.bf_bp_to_xu_l_nu_resonant, charged_b=True, decay_descr="B+ -> (pi0 [0.000078], rho0 [0.000158], eta [0.000039], eta' [0.000023], omega [0.000119]) l+ nu_l"),
}  # type: BFSDType
# fmt: on


# Adding branching fraction infos also for efficiency and MVA training samples.
# For the efficiency samples of the B -> D** decays these are not correct, as we have separate
# efficiency samples for the charged and neutral pion cases, but this does not affect the
# efficiency calculation performed in the EfficiencyEvaluationTask!
Sig_MVA_Info = {f"{k}MVA": v for k, v in sig_mc_bfs.items() if k.startswith("D")}  # type: BFSDType

Dss_PiZero_Eff_Tmp = {f"{k}Eff": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
Dss_PiZero_MVA_Tmp = {f"{k}MVA": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
Dss_PiPlus_Eff_Tmp = {f"{k}ChargedPiEff": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
Dss_PiPlus_MVA_Tmp = {f"{k}ChargedPiMVA": v for k, v in sig_mc_bfs.items() if k.startswith("B")}  # type: BFSDType
DssTau_PiZ_Eff_Tmp = {f"{k}Eff": v for k, v in sig_mc_bfs.items() if k.startswith("TB")}  # type: BFSDType
DssTau_PiP_Eff_Tmp = {f"{k}ChargedPiEff": v for k, v in sig_mc_bfs.items() if k.startswith("TB")}  # type: BFSDType

# Adapting branching fractions of charged and neutral pion D** decays:
Dss_PiZero_Eff_Info = update_branching_fraction(bf_dict=Dss_PiZero_Eff_Tmp, bf_factor=float(1.0 / 3.0))  # type: BFSDType
Dss_PiZero_MVA_Info = update_branching_fraction(bf_dict=Dss_PiZero_MVA_Tmp, bf_factor=float(1.0 / 3.0))  # type: BFSDType
Dss_PiPlus_Eff_Info = update_branching_fraction(bf_dict=Dss_PiPlus_Eff_Tmp, bf_factor=float(2.0 / 3.0))  # type: BFSDType
Dss_PiPlus_MVA_Info = update_branching_fraction(bf_dict=Dss_PiPlus_MVA_Tmp, bf_factor=float(2.0 / 3.0))  # type: BFSDType
DssTau_PiZ_Eff_Info = update_branching_fraction(bf_dict=DssTau_PiZ_Eff_Tmp, bf_factor=float(1.0 / 3.0))  # type: BFSDType
DssTau_PiP_Eff_Info = update_branching_fraction(bf_dict=DssTau_PiP_Eff_Tmp, bf_factor=float(2.0 / 3.0))  # type: BFSDType
assert all(
    math.isclose(Dss_PiZero_Eff_Info[f"{k}Eff"].bf + Dss_PiPlus_Eff_Info[f"{k}ChargedPiEff"].bf, v.bf)
    for k, v in sig_mc_bfs.items()
    if k.startswith("B")
)
assert all(
    math.isclose(Dss_PiZero_MVA_Info[f"{k}MVA"].bf + Dss_PiPlus_MVA_Info[f"{k}ChargedPiMVA"].bf, v.bf)
    for k, v in sig_mc_bfs.items()
    if k.startswith("B")
)
assert all(
    math.isclose(DssTau_PiZ_Eff_Info[f"{k}Eff"].bf + DssTau_PiP_Eff_Info[f"{k}ChargedPiEff"].bf, v.bf)
    for k, v in sig_mc_bfs.items()
    if k.startswith("TB")
)


assert all(k not in sig_mc_bfs for k in Sig_MVA_Info), [k for k in Sig_MVA_Info if k in sig_mc_bfs]
sig_mc_bfs.update(Sig_MVA_Info)
assert all(k not in sig_mc_bfs for k in Dss_PiZero_Eff_Info), [k for k in Dss_PiZero_Eff_Info if k in sig_mc_bfs]
sig_mc_bfs.update(Dss_PiZero_Eff_Info)
assert all(k not in sig_mc_bfs for k in Dss_PiZero_MVA_Info), [k for k in Dss_PiZero_MVA_Info if k in sig_mc_bfs]
sig_mc_bfs.update(Dss_PiZero_MVA_Info)
assert all(k not in sig_mc_bfs for k in Dss_PiPlus_Eff_Info), [k for k in Sig_MVA_Info if k in Dss_PiPlus_Eff_Info]
sig_mc_bfs.update(Dss_PiPlus_Eff_Info)
assert all(k not in sig_mc_bfs for k in Dss_PiPlus_MVA_Info), [k for k in Sig_MVA_Info if k in Dss_PiPlus_MVA_Info]
sig_mc_bfs.update(Dss_PiPlus_MVA_Info)
assert all(k not in sig_mc_bfs for k in DssTau_PiZ_Eff_Info), [k for k in DssTau_PiZ_Eff_Info if k in sig_mc_bfs]
sig_mc_bfs.update(DssTau_PiZ_Eff_Info)
assert all(k not in sig_mc_bfs for k in DssTau_PiP_Eff_Info), [k for k in DssTau_PiP_Eff_Info if k in sig_mc_bfs]
sig_mc_bfs.update(DssTau_PiP_Eff_Info)


signal_mc_bfs = sig_mc_bfs  # type: BFSDType


#  Number of simulated events in private MC samples
signal_mc_evts = {
    # B0
    "DSME": NEventEntryType(prod=2500000, meta=2500000, summary=2498465),
    "DSMEMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498399),
    "DME": NEventEntryType(prod=2500000, meta=2500000, summary=2498185),
    "DMEMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498257),
    "BZD1E": NEventEntryType(prod=72999988 - 150000, meta=72999988 - 150000, summary=72968227 - 147951),
    "BZD0sE": NEventEntryType(prod=72999988, meta=72999988, summary=72970500),
    "BZDp1E": NEventEntryType(prod=72999988, meta=72999988, summary=72969667),
    "BZD2sE": NEventEntryType(prod=72999988, meta=72999988, summary=72973227),
    "BZD1EEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498807),
    "BZD0sEEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498814),
    "BZDp1EEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498803),
    "BZD2sEEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498862),
    "BZD1EMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498830),
    "BZD0sEMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498738),
    "BZDp1EMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498908),
    "BZD2sEMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498819),
    "BZD1EChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499078),
    "BZDp1EChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499058),
    "BZD2sEChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499237),
    "BZD0sEChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499064),
    "BZD1EChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499057),
    "BZDp1EChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499001),
    "BZD2sEChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499215),
    "BZD0sEChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499110),
    "DSMMu": NEventEntryType(prod=2500000, meta=2500000, summary=2495946),
    "DSMMuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2496105),
    "DMMu": NEventEntryType(prod=2500000, meta=2500000, summary=2496282),
    "DMMuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2496267),
    "BZD1Mu": NEventEntryType(prod=72999988 - 50000, meta=72999988 - 50000, summary=72935556 - 49925),
    "BZD0sMu": NEventEntryType(prod=72999988 - 50000, meta=72999988 - 50000, summary=72933032 - 43151),
    "BZDp1Mu": NEventEntryType(prod=72999988 - 100000, meta=72999988 - 100000, summary=72925245 - 91485),
    "BZD2sMu": NEventEntryType(prod=72999988 - 50000, meta=72999988 - 50000, summary=72943989 - 49977),
    "BZD1MuEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497401),
    "BZD0sMuEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497464),
    "BZDp1MuEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497429),
    "BZD2sMuEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497434),
    "BZD1MuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497266),
    "BZD0sMuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497480),
    "BZDp1MuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497459),
    "BZD2sMuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497508),
    "BZD1MuChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497949),
    "BZDp1MuChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497831),
    "BZD2sMuChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498277),
    "BZD0sMuChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498196),
    "BZD1MuChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498034),
    "BZDp1MuChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497929),
    "BZD2sMuChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498398),
    "BZD0sMuChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498160),
    "DSMTau": NEventEntryType(prod=2500000, meta=2500000, summary=2496044),
    "DSMTauMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2496081),
    "DMTau": NEventEntryType(prod=2500000, meta=2500000, summary=2495966),
    "DMTauMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2495945),
    "TBZD1Tau": NEventEntryType(prod=17999995, meta=17999995, summary=17983826),
    "TBZD0sTau": NEventEntryType(prod=17999995, meta=17999995, summary=17984755),
    "TBZDp1Tau": NEventEntryType(prod=17999995, meta=17999995, summary=17983446),
    "TBZD2sTau": NEventEntryType(prod=17999995, meta=17999995, summary=17986986),
    "TBZD1TauEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497293),
    "TBZD0sTauEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497345),
    "TBZDp1TauEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497510),
    "TBZD2sTauEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497523),
    "TBZD1TauChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498041),
    "TBZD0sTauChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498040),
    "TBZDp1TauChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497759),
    "TBZD2sTauChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498489),
    "GapD1Mixed": NEventEntryType(prod=49999999, meta=49999999, summary=49954892),
    "GapDEtaMixed": NEventEntryType(prod=39999994 - 50000, meta=39999994 - 50000, summary=39978876 - 49977),
    "GapDstEtaMixed": NEventEntryType(prod=39999994, meta=39999994, summary=39979901),
    "GapDPiPiMixed": NEventEntryType(prod=9499997, meta=9499997, summary=9497036),
    "GapDstPiPiMixed": NEventEntryType(prod=45999994, meta=45999994, summary=45986274),
    "GapDEtaViaD0sMixed": NEventEntryType(prod=39999994, meta=39999994, summary=39970369),
    "GapDstEtaViaDp1Mixed": NEventEntryType(prod=39999994, meta=39999994, summary=39974088),
    "GapDPiPiViaDp1Mixed": NEventEntryType(prod=5999997, meta=5999997, summary=5995435),
    "GapDPiPiViaD0sMixed": NEventEntryType(prod=5999997, meta=5999997, summary=5994323),
    "GapDstPiPiViaDp1Mixed": NEventEntryType(prod=25000002, meta=25000002, summary=24978527),
    "GapDstPiPiViaD0sMixed": NEventEntryType(prod=25000002, meta=25000002, summary=24975471),
    "XuLNuIncMixed": NEventEntryType(prod=31633380, meta=31633380, summary=31603155),
    "XuLNuResMixed": NEventEntryType(prod=6851640, meta=6851640, summary=6842434),
    # B+
    "DSZE": NEventEntryType(prod=2500000, meta=2500000, summary=2498902),
    "DSZEMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499045),
    "DZE": NEventEntryType(prod=2500000, meta=2500000, summary=2498891),
    "DZEMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498861),
    "BMD1E": NEventEntryType(prod=72999988, meta=72999988, summary=72972061),
    "BMD0sE": NEventEntryType(prod=72999988, meta=72999988, summary=72970858),
    "BMDp1E": NEventEntryType(prod=72999988 - 1125239, meta=72999988 - 1125239, summary=72965689 - 1118869),
    "BMD2sE": NEventEntryType(prod=72999988, meta=72999988, summary=72975847),
    "BMD1EEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499211),
    "BMD0sEEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499129),
    "BMDp1EEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499198),
    "BMD2sEEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499248),
    "BMD1EMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499194),
    "BMD0sEMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499146),
    "BMDp1EMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499263),
    "BMD2sEMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499264),
    "BMD1EChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499030),
    "BMDp1EChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498962),
    "BMD2sEChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2499062),
    "BMD0sEChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498916),
    "BMD1EChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498913),
    "BMDp1EChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498921),
    "BMD2sEChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2499091),
    "BMD0sEChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498894),
    "DSZMu": NEventEntryType(prod=2500000, meta=2500000, summary=2497236),
    "DSZMuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497272),
    "DZMu": NEventEntryType(prod=2500000, meta=2500000, summary=2497364),
    "DZMuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497305),
    "BMD1Mu": NEventEntryType(prod=72999988, meta=72999988, summary=72939807),
    "BMD0sMu": NEventEntryType(prod=72999988, meta=72999988, summary=72939605),
    "BMDp1Mu": NEventEntryType(prod=72999988 - 1214612, meta=72999988 - 1214612, summary=72889052 - 1163954),
    "BMD2sMu": NEventEntryType(prod=72999988 - 100000, meta=72999988 - 100000, summary=72949175 - 99931),
    "BMD1MuEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498314),
    "BMD0sMuEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498148),
    "BMDp1MuEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498328),
    "BMD2sMuEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498388),
    "BMD1MuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498340),
    "BMD0sMuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498184),
    "BMDp1MuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498359),
    "BMD2sMuMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498384),
    "BMD1MuChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497710),
    "BMDp1MuChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497695),
    "BMD2sMuChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498204),
    "BMD0sMuChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497833),
    "BMD1MuChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497759),
    "BMDp1MuChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497635),
    "BMD2sMuChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2498219),
    "BMD0sMuChargedPiMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497807),
    "DSZTau": NEventEntryType(prod=2500000, meta=2500000, summary=2497350),
    "DSZTauMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497181),
    "DZTau": NEventEntryType(prod=2500000, meta=2500000, summary=2497049),
    "DZTauMVA": NEventEntryType(prod=2500000, meta=2500000, summary=2497120),
    "TBMD1Tau": NEventEntryType(prod=17999995, meta=17999995, summary=17985306),
    "TBMD0sTau": NEventEntryType(prod=17999995, meta=17999995, summary=17984176),
    "TBMDp1Tau": NEventEntryType(prod=17999995, meta=17999995, summary=17984511),
    "TBMD2sTau": NEventEntryType(prod=17999995, meta=17999995, summary=17988518),
    "TBMD1TauEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498261),
    "TBMD0sTauEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497998),
    "TBMDp1TauEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498380),
    "TBMD2sTauEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498310),
    "TBMD1TauChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497790),
    "TBMD0sTauChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497657),
    "TBMDp1TauChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2497636),
    "TBMD2sTauChargedPiEff": NEventEntryType(prod=2500000, meta=2500000, summary=2498382),
    "GapD1Charged": NEventEntryType(prod=49999999, meta=49999999, summary=49965108),
    "GapDEtaCharged": NEventEntryType(prod=39999994, meta=39999994, summary=39985913),
    "GapDstEtaCharged": NEventEntryType(prod=39999994, meta=39999994, summary=39987282),
    "GapDPiPiCharged": NEventEntryType(prod=9499997, meta=9499997, summary=9497676),
    "GapDstPiPiCharged": NEventEntryType(prod=45999994, meta=45999994, summary=45989597),
    "GapDEtaViaD0sCharged": NEventEntryType(prod=39999994, meta=39999994, summary=39980542),
    "GapDstEtaViaDp1Charged": NEventEntryType(prod=39999994, meta=39999994, summary=39983032),
    "GapDPiPiViaDp1Charged": NEventEntryType(prod=5999997, meta=5999997, summary=5996234),
    "GapDPiPiViaD0sCharged": NEventEntryType(prod=5999997, meta=5999997, summary=5995530),
    "GapDstPiPiViaDp1Charged": NEventEntryType(prod=25000002, meta=25000002, summary=24983424),
    "GapDstPiPiViaD0sCharged": NEventEntryType(prod=25000002, meta=25000002, summary=24982046),
    "XuLNuIncCharged": NEventEntryType(prod=34001299, meta=34001299, summary=33950980),
    # XuLNuIncCharged True info: (prod=34037540, meta=34001299, summary=33950980)  34037540 - 34001299 = 36241,
    # XuLNuIncCharged gsim summary info: 33977540 - 33950980 = 26560
    "XuLNuResCharged": NEventEntryType(prod=6434980, meta=6434980, summary=6427423),
}  # type: EvtNumDictType


# Checking consistency of the event numbers in the signal_mc_evts dictionary:
assert all(v.prod == v.meta for v in signal_mc_evts.values()), [
    (k, v.prod, v.meta) for k, v in signal_mc_evts.items() if not v.prod == v.meta
]
assert all(v.prod >= v.summary for v in signal_mc_evts.values()), [
    (k, v.prod, v.summary) for k, v in signal_mc_evts.items() if v.prod < v.summary
]


# Checking consistency of the dictionaries signal_mc_evts and signal_mc_bfs:
assert set(signal_mc_evts.keys()) == set(signal_mc_bfs.keys())


def add_scaling_weights(
    df: pd.DataFrame,
    data_type: str,
    global_file_dict_path: PathType,
    generic_stream_dict: Optional[Dict[str, int]] = None,
) -> None:
    if data_type in Belle_url_mc_types:
        applied_weight = sw.add_public_mc_weight(
            df=df,
            streams=generic_stream_dict[data_type] if generic_stream_dict is not None else 1.0,
            data_type=data_type,
            file_dict_dir_path=global_file_dict_path,
        )
    elif data_type in Belle_list_types:
        if data_type == "ulnuMixed":
            # Use correction factor obtained from number of failed ulnuMixed jobs over number of all ulnuMixed jobs.
            # failed_jobs = 37  # type: int  # production 04thJan2021
            # failed_jobs = 22  # type: int  # production 13thJan2021
            failed_jobs = 18  # type: int  # production 02ndMay2021
            ulnu_mixed_correction_factor = 1 - float(failed_jobs) / 1055.0  # type: float
            number_of_streams = ulnu_mixed_correction_factor * sw.FIXED_SPECIAL_WEIGHTS[data_type]
        else:
            number_of_streams = sw.FIXED_SPECIAL_WEIGHTS[data_type]

        applied_weight = sw.add_public_mc_weight(
            df=df,
            streams=number_of_streams,
            data_type=data_type,
        )
    elif data_type in signal_mc_evts:
        applied_weight = sw.add_private_mc_weight(
            df=df,
            produced_events=signal_mc_evts[data_type].prod,
            charged_B=signal_mc_bfs[data_type].charged_b,
            branching_fraction=signal_mc_bfs[data_type].bf,
        )
    elif data_type in Belle_data_types:
        # Also adding no scaling weight for actual data.
        applied_weight = 1.0
        df.loc[:, sw.scale_weight_column_name] = applied_weight
    else:
        raise RuntimeError(f"Data type {data_type} is not covered by add_scaling_weights!")

    assert applied_weight > 0.0

    print(f"Applied scaling factor for {data_type:>10}: {applied_weight:2.4f}")


def add_off_to_on_resonance_scaling_weights(
    df: pd.DataFrame,
    data_type: str,
) -> None:
    sw.add_off_resonance_scaling_weight(
        df=df,
        off_resonance_scaling_weight_column_name=sw.off_to_on_resonance_scale_weight_column_name,
        exp_number_column_name="expNum",
        run_number_column_name="runNum",
        add_neutral_weight=data_type not in off_resonance_types,
        use_updated_scale_values=False,
    )
