"""
Information on the decay reconstruction modes of this analysis.
"""

import copy
import numpy as np
import pandas as pd

from typing import Dict, Tuple

from rdstar.utilities.logging import log_to_default
from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.online_analysis.decay_channels import Dstar_decay_label_to_decay_id, D_decay_label_to_decay_id

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.fei_b_tagging.fei_reco_mode_interpreter import FEIDecayModeIDInterpreter

from rdstar.offline_analysis.reco_modes.reco_mode_utilities import (
    RecoModeIDToTitleDictType,
    RecoModeIDInfo,
    RecoModeGroupInfo,
    AbstractRecoModeInfosBaseClass,
)


__all__ = [
    "RecoModeInfos",
    "TagRecoModeInfos",
    "DMesonRecoModeInfos",
    "U4SRecoModeInfos",
]


logger = log_to_default(name=__name__)


# TODO: Fix charged B Meson charges in LaTeX labels: B- -> B+ everywhere!!!!
# TODO: Remove duplicates!?


class RecoModeInfos(AbstractRecoModeInfosBaseClass):
    _all_reco_modes_latex_label: str = LatexDecayProcesses.all_modes_label
    _all_reco_modes_latex_label_short: str = LatexDecayProcesses.all_modes_label_short

    # fmt: off
    # B -> D l  Reco Modes
    neutral_b_to_d_e_mode = RecoModeIDInfo(id=11, name="neutral_b_to_d_e_mode", latex_label=LatexDecayProcesses.reco_b0_to_dm_ep)  #                                               B0  to  D- e+                 # noqa: E262
    neutral_b_to_d_mu_mode = RecoModeIDInfo(id=12, name="neutral_b_to_d_mu_mode", latex_label=LatexDecayProcesses.reco_b0_to_dm_mup)  #                                            B0  to  D- mu+                # noqa: E262
    charged_b_to_d_e_mode = RecoModeIDInfo(id=13, name="charged_b_to_d_e_mode", latex_label=LatexDecayProcesses.reco_bm_to_d0_em)  #                                               B-  to  D0 e-                 # noqa: E262
    charged_b_to_d_mu_mode = RecoModeIDInfo(id=14, name="charged_b_to_d_mu_mode", latex_label=LatexDecayProcesses.reco_bm_to_d0_mum)  #                                            B-  to  D0 mu-                # noqa: E262

    # B -> D* l  Reco Modes
    neutral_b_to_ds_e_mode = RecoModeIDInfo(id=15, name="neutral_b_to_ds_e_mode", latex_label=LatexDecayProcesses.reco_b0_to_d_stm_ep)  #                                          B0  to  D*- e+                # noqa: E262
    neutral_b_to_ds_mu_mode = RecoModeIDInfo(id=16, name="neutral_b_to_ds_mu_mode", latex_label=LatexDecayProcesses.reco_b0_to_d_stm_mup)  #                                       B0  to  D*- mu+               # noqa: E262
    charged_b_to_ds_e_mode = RecoModeIDInfo(id=17, name="charged_b_to_ds_e_mode", latex_label=LatexDecayProcesses.reco_bm_to_d_st0_em)  #                                          B-  to  D*0 e-                # noqa: E262
    charged_b_to_ds_mu_mode = RecoModeIDInfo(id=18, name="charged_b_to_ds_mu_mode", latex_label=LatexDecayProcesses.reco_bm_to_d_st0_mum)  #                                       B-  to  D*0 mu-               # noqa: E262

    # B -> D** (-> D pi0) l  Reco Modes
    neutral_b_to_dss_e_via_d_pi0_mode = RecoModeIDInfo(id=61, name="neutral_b_to_dss_e_via_d_pi0_mode", latex_label=LatexDecayProcesses.reco_b0_to_dm_ep_pi0)  #                   B0  to  D- e+ pi0             # noqa: E262
    neutral_b_to_dss_mu_via_d_pi0_mode = RecoModeIDInfo(id=62, name="neutral_b_to_dss_mu_via_d_pi0_mode", latex_label=LatexDecayProcesses.reco_b0_to_dm_mup_pi0)  #                B0  to  D- mu+ pi0            # noqa: E262
    charged_b_to_dss_e_via_d_pi0_mode = RecoModeIDInfo(id=63, name="charged_b_to_dss_e_via_d_pi0_mode", latex_label=LatexDecayProcesses.reco_bm_to_d0_em_pi0)  #                   B-  to  D0 e- pi0             # noqa: E262
    charged_b_to_dss_mu_via_d_pi0_mode = RecoModeIDInfo(id=64, name="charged_b_to_dss_mu_via_d_pi0_mode", latex_label=LatexDecayProcesses.reco_bm_to_d0_mum_pi0)  #                B-  to  D0 mu- pi0            # noqa: E262

    # B -> D** (-> D* pi0) l  Reco Modes
    neutral_b_to_dss_e_via_ds_pi0_mode = RecoModeIDInfo(id=65, name="neutral_b_to_dss_e_via_ds_pi0_mode", latex_label=LatexDecayProcesses.reco_b0_to_d_stm_ep_pi0)  #              B0  to  D*- e+ pi0            # noqa: E262
    neutral_b_to_dss_mu_via_ds_pi0_mode = RecoModeIDInfo(id=66, name="neutral_b_to_dss_mu_via_ds_pi0_mode", latex_label=LatexDecayProcesses.reco_b0_to_d_stm_mup_pi0)  #           B0  to  D*- mu+ pi0           # noqa: E262
    charged_b_to_dss_e_via_ds_pi0_mode = RecoModeIDInfo(id=67, name="charged_b_to_dss_e_via_ds_pi0_mode", latex_label=LatexDecayProcesses.reco_bm_to_d_st0_em_pi0)  #              B-  to  D*0 e- pi0            # noqa: E262
    charged_b_to_dss_mu_via_ds_pi0_mode = RecoModeIDInfo(id=68, name="charged_b_to_dss_mu_via_ds_pi0_mode", latex_label=LatexDecayProcesses.reco_bm_to_d_st0_mum_pi0)  #           B-  to  D*0 mu- pi0           # noqa: E262

    # B -> D** (-> D pi+) l  Reco Modes
    neutral_b_to_dss_e_via_d_pip_mode = RecoModeIDInfo(id=71, name="neutral_b_to_dss_e_via_d_pip_mode", latex_label=LatexDecayProcesses.reco_b0_to_anti_d0_ep_pim)  #              B0  to  anti-D0 e+ pi-        # noqa: E262
    neutral_b_to_dss_mu_via_d_pip_mode = RecoModeIDInfo(id=72, name="neutral_b_to_dss_mu_via_d_pip_mode", latex_label=LatexDecayProcesses.reco_b0_to_anti_d0_mup_pim)  #           B0  to  anti-D0 mu+ pi-       # noqa: E262
    charged_b_to_dss_e_via_d_pip_mode = RecoModeIDInfo(id=73, name="charged_b_to_dss_e_via_d_pip_mode", latex_label=LatexDecayProcesses.reco_bm_to_dp_em_pim)  #                   B-  to  D+ e- pi-             # noqa: E262
    charged_b_to_dss_mu_via_d_pip_mode = RecoModeIDInfo(id=74, name="charged_b_to_dss_mu_via_d_pip_mode", latex_label=LatexDecayProcesses.reco_bm_to_dp_mum_pim)  #                B-  to  D+ mu- pi-            # noqa: E262

    # B -> D** (-> D* pi+) l  Reco Modes
    neutral_b_to_dss_e_via_ds_pip_mode = RecoModeIDInfo(id=75, name="neutral_b_to_dss_e_via_ds_pip_mode", latex_label=LatexDecayProcesses.reco_b0_to_anti_d_st0_ep_pim)  #         B0  to  anti-D*0 e+ pi-       # noqa: E262
    neutral_b_to_dss_mu_via_ds_pip_mode = RecoModeIDInfo(id=76, name="neutral_b_to_dss_mu_via_ds_pip_mode", latex_label=LatexDecayProcesses.reco_b0_to_anti_d_st0_mup_pim)  #      B0  to  anti-D*0 mu+ pi-      # noqa: E262
    charged_b_to_dss_e_via_ds_pip_mode = RecoModeIDInfo(id=77, name="charged_b_to_dss_e_via_ds_pip_mode", latex_label=LatexDecayProcesses.reco_bm_to_d_stp_em_pim)  #              B-  to  D*+ e- pi-            # noqa: E262
    charged_b_to_dss_mu_via_ds_pip_mode = RecoModeIDInfo(id=78, name="charged_b_to_dss_mu_via_ds_pip_mode", latex_label=LatexDecayProcesses.reco_bm_to_d_stp_mum_pim)  #           B-  to  D*+ mu- pi-           # noqa: E262

    # B -> pi l  Reco Modes
    neutral_b_to_pip_e_mode = RecoModeIDInfo(id=81, name="neutral_b_to_pip_e_mode", latex_label=LatexDecayProcesses.reco_b0_to_pim_ep)  #                                          B0 to pi+ e-                  # noqa: E262
    neutral_b_to_pip_mu_mode = RecoModeIDInfo(id=82, name="neutral_b_to_pip_mu_mode", latex_label=LatexDecayProcesses.reco_b0_to_pim_mup)  #                                       B0 to pi+ mu-                 # noqa: E262
    charged_b_to_pi0_e_mode = RecoModeIDInfo(id=83, name="charged_b_to_pi0_e_mode", latex_label=LatexDecayProcesses.reco_bp_to_pi0_ep)  #                                          B- to pi0 e-                  # noqa: E262
    charged_b_to_pi0_mu_mode = RecoModeIDInfo(id=84, name="charged_b_to_pi0_mu_mode", latex_label=LatexDecayProcesses.reco_bp_to_pi0_mup)  #                                       B- to pi0 mu-                 # noqa: E262

    #
    # Reco Mode Combinations:

    # B -> D(*) l  Reco Mode Groups
    neutral_b_to_d_modes = RecoModeGroupInfo(ids=(neutral_b_to_d_e_mode.id, neutral_b_to_d_mu_mode.id), name="neutral_b_to_d_modes", latex_label=LatexDecayProcesses.reco_b0_to_dm_lp)
    charged_b_to_d_modes = RecoModeGroupInfo(ids=(charged_b_to_d_e_mode.id, charged_b_to_d_mu_mode.id), name="charged_b_to_d_modes", latex_label=LatexDecayProcesses.reco_bm_to_d0_lm)

    neutral_b_to_ds_modes = RecoModeGroupInfo(ids=(neutral_b_to_ds_e_mode.id, neutral_b_to_ds_mu_mode.id), name="neutral_b_to_ds_modes", latex_label=LatexDecayProcesses.reco_b0_to_d_stm_lp)
    charged_b_to_ds_modes = RecoModeGroupInfo(ids=(charged_b_to_ds_e_mode.id, charged_b_to_ds_mu_mode.id), name="charged_b_to_ds_modes", latex_label=LatexDecayProcesses.reco_bm_to_d_st0_lm)

    neutral_b_to_d_wwo_star_to_e_modes = RecoModeGroupInfo(ids=(neutral_b_to_d_e_mode.id, neutral_b_to_ds_e_mode.id), name="neutral_b_to_d_wwo_star_to_e_modes", latex_label=LatexDecayProcesses.reco_b0_to_dm_wwos_e)
    charged_b_to_d_wwo_star_to_e_modes = RecoModeGroupInfo(ids=(charged_b_to_d_e_mode.id, charged_b_to_ds_e_mode.id), name="charged_b_to_d_wwo_star_to_e_modes", latex_label=LatexDecayProcesses.reco_bp_to_d0_wwos_e)
    neutral_b_to_d_wwo_star_to_mu_modes = RecoModeGroupInfo(ids=(neutral_b_to_d_mu_mode.id, neutral_b_to_ds_mu_mode.id), name="neutral_b_to_d_wwo_star_to_mu_modes", latex_label=LatexDecayProcesses.reco_b0_to_dm_wwos_mu)
    charged_b_to_d_wwo_star_to_mu_modes = RecoModeGroupInfo(ids=(charged_b_to_d_mu_mode.id, charged_b_to_ds_mu_mode.id), name="charged_b_to_d_wwo_star_to_mu_modes", latex_label=LatexDecayProcesses.reco_bp_to_d0_wwos_mu)

    neutral_b_to_d_wwo_star_modes = RecoModeGroupInfo(ids=neutral_b_to_d_modes.ids + neutral_b_to_ds_modes.ids, name="neutral_b_to_d_wwo_star_modes", latex_label=LatexDecayProcesses.reco_b0_to_dm_wwos_lepton)
    charged_b_to_d_wwo_star_modes = RecoModeGroupInfo(ids=charged_b_to_d_modes.ids + charged_b_to_ds_modes.ids, name="charged_b_to_d_wwo_star_modes", latex_label=LatexDecayProcesses.reco_bp_to_d0_wwos_lepton)

    b_to_d_e_modes = RecoModeGroupInfo(ids=(neutral_b_to_d_e_mode.id, charged_b_to_d_e_mode.id), name="b_to_d_e_modes", latex_label=LatexDecayProcesses.reco_b_to_d_e)
    b_to_d_mu_modes = RecoModeGroupInfo(ids=(neutral_b_to_d_mu_mode.id, charged_b_to_d_mu_mode.id), name="b_to_d_mu_modes", latex_label=LatexDecayProcesses.reco_b_to_d_mu)
    b_to_ds_e_modes = RecoModeGroupInfo(ids=(neutral_b_to_ds_e_mode.id, charged_b_to_ds_e_mode.id), name="b_to_ds_e_modes", latex_label=LatexDecayProcesses.reco_b_to_ds_e)
    b_to_ds_mu_modes = RecoModeGroupInfo(ids=(neutral_b_to_ds_mu_mode.id, charged_b_to_ds_mu_mode.id), name="b_to_ds_mu_modes", latex_label=LatexDecayProcesses.reco_b_to_ds_mu)

    b_to_d_wwo_star_e_modes = RecoModeGroupInfo(ids=b_to_d_e_modes.ids + b_to_ds_e_modes.ids, name="b_to_d_wwo_star_e_modes", latex_label=LatexDecayProcesses.reco_b_to_d_wwos_e)
    b_to_d_wwo_star_mu_modes = RecoModeGroupInfo(ids=b_to_d_mu_modes.ids + b_to_ds_mu_modes.ids, name="b_to_d_wwo_star_mu_modes", latex_label=LatexDecayProcesses.reco_b_to_d_wwos_mu)

    b_to_d_modes = RecoModeGroupInfo(ids=neutral_b_to_d_modes.ids + charged_b_to_d_modes.ids, name="b_to_d_modes", latex_label=LatexDecayProcesses.reco_b_to_d)
    b_to_ds_modes = RecoModeGroupInfo(ids=neutral_b_to_ds_modes.ids + charged_b_to_ds_modes.ids, name="b_to_ds_modes", latex_label=LatexDecayProcesses.reco_b_to_ds)

    b_to_d_wwo_star_modes = RecoModeGroupInfo(ids=b_to_d_modes.ids + b_to_ds_modes.ids, name="b_to_d_wwo_star_modes", latex_label=LatexDecayProcesses.all_modes_label[:-1] + r"\,\!$")

    # Main Reco Mode Groups with old Names
    _d_reco_modes: RecoModeGroupInfo = b_to_d_modes  # Duplicates!
    _d_star_reco_modes: RecoModeGroupInfo = b_to_ds_modes  # Duplicates!

    main_reco_modes = RecoModeGroupInfo(ids=_d_reco_modes.ids + _d_star_reco_modes.ids, name="main_reco_modes", latex_label=LatexDecayProcesses.all_modes_label)

    #
    # B -> D** l  Reco Mode Groups
    neutral_b_to_dss_e_via_dwwos_pi0_modes = RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_d_pi0_mode.id, neutral_b_to_dss_e_via_ds_pi0_mode.id), name="neutral_b_to_dss_e_via_dwwos_pi0_modes", latex_label=LatexDecayProcesses.reco_b0_to_dssm_e_with_dss_to_dm_wwos_pi0)
    charged_b_to_dss_e_via_dwwos_pi0_modes = RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_d_pi0_mode.id, charged_b_to_dss_e_via_ds_pi0_mode.id), name="charged_b_to_dss_e_via_dwwos_pi0_modes", latex_label=LatexDecayProcesses.reco_bp_to_dss0_e_with_dss_to_d0_wwos_pi0)
    neutral_b_to_dss_mu_via_dwwos_pi0_modes = RecoModeGroupInfo(ids=(neutral_b_to_dss_mu_via_d_pi0_mode.id, neutral_b_to_dss_mu_via_ds_pi0_mode.id), name="neutral_b_to_dss_mu_via_dwwos_pi0_modes", latex_label=LatexDecayProcesses.reco_b0_to_dssm_mu_with_dss_to_dm_wwos_pi0)
    charged_b_to_dss_mu_via_dwwos_pi0_modes = RecoModeGroupInfo(ids=(charged_b_to_dss_mu_via_d_pi0_mode.id, charged_b_to_dss_mu_via_ds_pi0_mode.id), name="charged_b_to_dss_mu_via_dwwos_pi0_modes", latex_label=LatexDecayProcesses.reco_bp_to_dss0_mu_with_dss_to_d0_wwos_pi0)

    neutral_b_to_dss_e_via_dwwos_pip_modes = RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_d_pip_mode.id, neutral_b_to_dss_e_via_ds_pip_mode.id), name="neutral_b_to_dss_e_via_dwwos_pip_modes", latex_label=LatexDecayProcesses.reco_b0_to_dssm_e_with_dss_to_d0_wwos_pim)
    charged_b_to_dss_e_via_dwwos_pip_modes = RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_d_pip_mode.id, charged_b_to_dss_e_via_ds_pip_mode.id), name="charged_b_to_dss_e_via_dwwos_pip_modes", latex_label=LatexDecayProcesses.reco_bp_to_dss0_e_with_dss_to_dp_wwos_pim)
    neutral_b_to_dss_mu_via_dwwos_pip_modes = RecoModeGroupInfo(ids=(neutral_b_to_dss_mu_via_d_pip_mode.id, neutral_b_to_dss_mu_via_ds_pip_mode.id), name="neutral_b_to_dss_mu_via_dwwos_pip_modes", latex_label=LatexDecayProcesses.reco_b0_to_dssm_mu_with_dss_to_d0_wwos_pim)
    charged_b_to_dss_mu_via_dwwos_pip_modes = RecoModeGroupInfo(ids=(charged_b_to_dss_mu_via_d_pip_mode.id, charged_b_to_dss_mu_via_ds_pip_mode.id), name="charged_b_to_dss_mu_via_dwwos_pip_modes", latex_label=LatexDecayProcesses.reco_bp_to_dss0_mu_with_dss_to_dp_wwos_pim)

    neutral_b_to_dss_e_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_e_via_dwwos_pi0_modes.ids + neutral_b_to_dss_e_via_dwwos_pip_modes.ids, name="neutral_b_to_dss_e_modes", latex_label=LatexDecayProcesses.reco_b0_to_dssm_e_with_dss_to_d_wwos_pi)
    charged_b_to_dss_e_modes = RecoModeGroupInfo(ids=charged_b_to_dss_e_via_dwwos_pi0_modes.ids + charged_b_to_dss_e_via_dwwos_pip_modes.ids, name="charged_b_to_dss_e_modes", latex_label=LatexDecayProcesses.reco_bp_to_dss0_e_with_dss_to_d_wwos_pi)
    neutral_b_to_dss_mu_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_mu_via_dwwos_pi0_modes.ids + neutral_b_to_dss_mu_via_dwwos_pip_modes.ids, name="neutral_b_to_dss_mu_modes", latex_label=LatexDecayProcesses.reco_b0_to_dssm_mu_with_dss_to_d_wwos_pi)
    charged_b_to_dss_mu_modes = RecoModeGroupInfo(ids=charged_b_to_dss_mu_via_dwwos_pi0_modes.ids + charged_b_to_dss_mu_via_dwwos_pip_modes.ids, name="charged_b_to_dss_mu_modes", latex_label=LatexDecayProcesses.reco_bp_to_dss0_mu_with_dss_to_d_wwos_pi)

    b_to_dss_e_via_dwwos_pi0_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_e_via_dwwos_pi0_modes.ids + charged_b_to_dss_e_via_dwwos_pi0_modes.ids, name="b_to_dss_e_via_dwwos_pi0_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_e_with_dss_to_d_wwos_pi0)
    b_to_dss_e_via_dwwos_pip_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_e_via_dwwos_pip_modes.ids + charged_b_to_dss_e_via_dwwos_pip_modes.ids, name="b_to_dss_e_via_dwwos_pip_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_e_with_dss_to_d_wwos_pim)
    b_to_dss_mu_via_dwwos_pi0_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_mu_via_dwwos_pi0_modes.ids + charged_b_to_dss_mu_via_dwwos_pi0_modes.ids, name="b_to_dss_mu_via_dwwos_pi0_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_mu_with_dss_to_d_wwos_pi0)
    b_to_dss_mu_via_dwwos_pip_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_mu_via_dwwos_pip_modes.ids + charged_b_to_dss_mu_via_dwwos_pip_modes.ids, name="b_to_dss_mu_via_dwwos_pip_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_mu_with_dss_to_d_wwos_pim)

    b_to_dss_e_modes = RecoModeGroupInfo(ids=b_to_dss_e_via_dwwos_pi0_modes.ids + b_to_dss_e_via_dwwos_pip_modes.ids, name="b_to_dss_e_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_e_with_dss_to_d_wwos_pi)
    b_to_dss_mu_modes = RecoModeGroupInfo(ids=b_to_dss_mu_via_dwwos_pi0_modes.ids + b_to_dss_mu_via_dwwos_pip_modes.ids, name="b_to_dss_mu_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_mu_with_dss_to_d_wwos_pi)

    e_reco_modes = RecoModeGroupInfo(ids=b_to_d_wwo_star_e_modes.ids + b_to_dss_e_modes.ids, name="e_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_d_w_wo_st_w_wo_pi_e)
    mu_reco_modes = RecoModeGroupInfo(ids=b_to_d_wwo_star_mu_modes.ids + b_to_dss_mu_modes.ids, name="mu_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_d_w_wo_st_w_wo_pi_mu)

    #
    neutral_b_to_dss_l_via_d_pi0_modes = RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_d_pi0_mode.id, neutral_b_to_dss_mu_via_d_pi0_mode.id), name="neutral_b_to_dss_l_via_d_pi0_modes", latex_label=LatexDecayProcesses.reco_b0_to_dm_pi0_lp)
    charged_b_to_dss_l_via_d_pi0_modes = RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_d_pi0_mode.id, charged_b_to_dss_mu_via_d_pi0_mode.id), name="charged_b_to_dss_l_via_d_pi0_modes", latex_label=LatexDecayProcesses.reco_bm_to_d0_pi0_lm)

    neutral_b_to_dss_l_via_ds_pi0_modes = RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_ds_pi0_mode.id, neutral_b_to_dss_mu_via_ds_pi0_mode.id), name="neutral_b_to_dss_l_via_ds_pi0_modes", latex_label=LatexDecayProcesses.reco_b0_to_d_stm_pi0_lp)
    charged_b_to_dss_l_via_ds_pi0_modes = RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_ds_pi0_mode.id, charged_b_to_dss_mu_via_ds_pi0_mode.id), name="charged_b_to_dss_l_via_ds_pi0_modes", latex_label=LatexDecayProcesses.reco_bm_to_d_st0_pi0_lm)

    neutral_b_to_dss_l_via_d_pip_modes = RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_d_pip_mode.id, neutral_b_to_dss_mu_via_d_pip_mode.id), name="neutral_b_to_dss_l_via_d_pip_modes", latex_label=LatexDecayProcesses.reco_b0_to_anti_d0_pim_lp)
    charged_b_to_dss_l_via_d_pip_modes = RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_d_pip_mode.id, charged_b_to_dss_mu_via_d_pip_mode.id), name="charged_b_to_dss_l_via_d_pip_modes", latex_label=LatexDecayProcesses.reco_bm_to_dp_pim_lm)

    neutral_b_to_dss_l_via_ds_pip_modes = RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_ds_pip_mode.id, neutral_b_to_dss_mu_via_ds_pip_mode.id), name="neutral_b_to_dss_l_via_ds_pip_modes", latex_label=LatexDecayProcesses.reco_b0_to_anti_d_st0_pim_lp)
    charged_b_to_dss_l_via_ds_pip_modes = RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_ds_pip_mode.id, charged_b_to_dss_mu_via_ds_pip_mode.id), name="charged_b_to_dss_l_via_ds_pip_modes", latex_label=LatexDecayProcesses.reco_bm_to_d_stp_pim_lm)

    neutral_b_to_dss_l_via_pi0_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_l_via_d_pi0_modes.ids + neutral_b_to_dss_l_via_ds_pi0_modes.ids, name="neutral_b_to_dss_l_via_pi0_modes", latex_label=LatexDecayProcesses.reco_dss_val_neutral_b_neutral_pi)
    charged_b_to_dss_l_via_pi0_modes = RecoModeGroupInfo(ids=charged_b_to_dss_l_via_d_pi0_modes.ids + charged_b_to_dss_l_via_ds_pi0_modes.ids, name="charged_b_to_dss_l_via_pi0_modes", latex_label=LatexDecayProcesses.reco_dss_val_charged_b_neutral_pi)

    neutral_b_to_dss_l_via_pip_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_l_via_d_pip_modes.ids + neutral_b_to_dss_l_via_ds_pip_modes.ids, name="neutral_b_to_dss_l_via_pip_modes", latex_label=LatexDecayProcesses.reco_dss_val_neutral_b_charged_pi)
    charged_b_to_dss_l_via_pip_modes = RecoModeGroupInfo(ids=charged_b_to_dss_l_via_d_pip_modes.ids + charged_b_to_dss_l_via_ds_pip_modes.ids, name="charged_b_to_dss_l_via_pip_modes", latex_label=LatexDecayProcesses.reco_dss_val_charged_b_charged_pi)

    neutral_b_to_dss_reco_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_l_via_pi0_modes.ids + neutral_b_to_dss_l_via_pip_modes.ids, name="neutral_b_to_dss_reco_modes", latex_label=LatexDecayProcesses.reco_b0_to_dss_l_with_dss_to_d_wwos_pi)
    charged_b_to_dss_reco_modes = RecoModeGroupInfo(ids=charged_b_to_dss_l_via_pi0_modes.ids + charged_b_to_dss_l_via_pip_modes.ids, name="charged_b_to_dss_reco_modes", latex_label=LatexDecayProcesses.reco_bp_to_dss_l_with_dss_to_d_wwos_pi)

    neutral_reco_modes = RecoModeGroupInfo(ids=neutral_b_to_d_wwo_star_modes.ids + neutral_b_to_dss_reco_modes.ids, name="neutral_reco_modes", latex_label=LatexDecayProcesses.reco_b0_to_d_w_wo_st_w_wo_pi_l)
    charged_reco_modes = RecoModeGroupInfo(ids=charged_b_to_d_wwo_star_modes.ids + charged_b_to_dss_reco_modes.ids, name="charged_reco_modes", latex_label=LatexDecayProcesses.reco_bp_to_d_w_wo_st_w_wo_pi_l)

    neutral_b_to_dss_l_via_d_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_l_via_d_pi0_modes.ids + neutral_b_to_dss_l_via_d_pip_modes.ids, name="neutral_b_to_dss_l_via_d_modes", latex_label=LatexDecayProcesses.reco_b0_to_d_pi_l)
    charged_b_to_dss_l_via_d_modes = RecoModeGroupInfo(ids=charged_b_to_dss_l_via_d_pi0_modes.ids + charged_b_to_dss_l_via_d_pip_modes.ids, name="charged_b_to_dss_l_via_d_modes", latex_label=LatexDecayProcesses.reco_bp_to_d_pi_l)
    neutral_b_to_dss_l_via_ds_modes = RecoModeGroupInfo(ids=neutral_b_to_dss_l_via_ds_pi0_modes.ids + neutral_b_to_dss_l_via_ds_pip_modes.ids, name="neutral_b_to_dss_l_via_ds_modes", latex_label=LatexDecayProcesses.reco_b0_to_ds_pi_l)
    charged_b_to_dss_l_via_ds_modes = RecoModeGroupInfo(ids=charged_b_to_dss_l_via_ds_pi0_modes.ids + charged_b_to_dss_l_via_ds_pip_modes.ids, name="charged_b_to_dss_l_via_ds_modes", latex_label=LatexDecayProcesses.reco_bp_to_ds_pi_l)

    # D** Reco Mode Groups

    # TODO: Rename to _with_neutral_pi
    _dss_via_d_neutral_b_reco_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_d_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_d_pi0_mode.id, neutral_b_to_dss_mu_via_d_pi0_mode.id), name="dss_via_d_neutral_b_reco_modes", latex_label=LatexDecayProcesses.reco_b0_to_dm_pi0_lp)
    _dss_via_ds_neutral_b_reco_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_ds_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_ds_pi0_mode.id, neutral_b_to_dss_mu_via_ds_pi0_mode.id), name="dss_via_ds_neutral_b_reco_modes", latex_label=LatexDecayProcesses.reco_b0_to_d_stm_pi0_lp)
    _dss_via_d_charged_b_reco_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_d_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_d_pi0_mode.id, charged_b_to_dss_mu_via_d_pi0_mode.id), name="dss_via_d_charged_b_reco_modes", latex_label=LatexDecayProcesses.reco_bm_to_d0_pi0_lm)
    _dss_via_ds_charged_b_reco_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_ds_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_ds_pi0_mode.id, charged_b_to_dss_mu_via_ds_pi0_mode.id), name="dss_via_ds_charged_b_reco_modes", latex_label=LatexDecayProcesses.reco_bm_to_d_st0_pi0_lm)

    # TODO: Rename to _with_charged_pi
    _dss_charged_via_d_neutral_b_reco_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_d_pip_modes  # Duplicates: RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_d_pip_mode.id, neutral_b_to_dss_mu_via_d_pip_mode.id), name="dss_charged_via_d_neutral_b_reco_modes", latex_label=LatexDecayProcesses.reco_b0_to_anti_d0_pim_lp)
    _dss_charged_via_ds_neutral_b_reco_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_ds_pip_modes  # Duplicates: RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_ds_pip_mode.id, neutral_b_to_dss_mu_via_ds_pip_mode.id), name="dss_charged_via_ds_neutral_b_reco_modes", latex_label=LatexDecayProcesses.reco_b0_to_anti_d_st0_pim_lp)
    _dss_charged_via_d_charged_b_reco_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_d_pip_modes  # Duplicates: RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_d_pip_mode.id, charged_b_to_dss_mu_via_d_pip_mode.id), name="dss_charged_via_d_charged_b_reco_modes", latex_label=LatexDecayProcesses.reco_bm_to_dp_pim_lm)
    _dss_charged_via_ds_charged_b_reco_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_ds_pip_modes  # Duplicates: RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_ds_pip_mode.id, charged_b_to_dss_mu_via_ds_pip_mode.id), name="dss_charged_via_ds_charged_b_reco_modes", latex_label=LatexDecayProcesses.reco_bm_to_d_stp_pim_lm)

    _dss_charged_b_neutral_pi_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=_dss_via_d_charged_b_reco_modes.ids + _dss_via_ds_charged_b_reco_modes.ids, name="dss_charged_b_neutral_pi_modes", latex_label=LatexDecayProcesses.reco_dss_val_charged_b_neutral_pi)
    _dss_charged_b_charged_pi_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_pip_modes  # Duplicates: RecoModeGroupInfo(ids=_dss_charged_via_d_charged_b_reco_modes.ids + _dss_charged_via_ds_charged_b_reco_modes.ids, name="dss_charged_b_charged_pi_modes", latex_label=LatexDecayProcesses.reco_dss_val_charged_b_charged_pi)
    _dss_neutral_b_neutral_pi_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=_dss_via_d_neutral_b_reco_modes.ids + _dss_via_ds_neutral_b_reco_modes.ids, name="dss_neutral_b_neutral_pi_modes", latex_label=LatexDecayProcesses.reco_dss_val_neutral_b_neutral_pi)
    _dss_neutral_b_charged_pi_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_pip_modes  # Duplicates: RecoModeGroupInfo(ids=_dss_charged_via_d_neutral_b_reco_modes.ids + _dss_charged_via_ds_neutral_b_reco_modes.ids, name="dss_neutral_b_charged_pi_modes", latex_label=LatexDecayProcesses.reco_dss_val_neutral_b_charged_pi)

    dss_d_l_neutral_pion_modes = RecoModeGroupInfo(ids=_dss_via_d_neutral_b_reco_modes.ids + _dss_via_d_charged_b_reco_modes.ids, name="dss_d_l_neutral_pion_modes", latex_label=LatexDecayProcesses.reco_dss_val_d_neutral_pi)
    dss_ds_l_neutral_pion_modes = RecoModeGroupInfo(ids=_dss_via_ds_neutral_b_reco_modes.ids + _dss_via_ds_charged_b_reco_modes.ids, name="dss_ds_l_neutral_pion_modes", latex_label=LatexDecayProcesses.reco_dss_val_ds_neutral_pi)
    dss_d_l_charged_pion_modes = RecoModeGroupInfo(ids=_dss_charged_via_d_neutral_b_reco_modes.ids + _dss_charged_via_d_charged_b_reco_modes.ids, name="dss_d_l_charged_pion_modes", latex_label=LatexDecayProcesses.reco_dss_val_d_charged_pi)
    dss_ds_l_charged_pion_modes = RecoModeGroupInfo(ids=_dss_charged_via_ds_neutral_b_reco_modes.ids + _dss_charged_via_ds_charged_b_reco_modes.ids, name="dss_ds_l_charged_pion_modes", latex_label=LatexDecayProcesses.reco_dss_val_ds_charged_pi)

    d_ss_neutral_pion_reco_modes = RecoModeGroupInfo(ids=_dss_charged_b_neutral_pi_modes.ids + _dss_neutral_b_neutral_pi_modes.ids, name="d_ss_neutral_pion_reco_modes", latex_label=LatexDecayProcesses.reco_dss_val_neutral_pi)
    d_ss_charged_pion_reco_modes = RecoModeGroupInfo(ids=_dss_charged_b_charged_pi_modes.ids + _dss_neutral_b_charged_pi_modes.ids, name="d_ss_charged_pion_reco_modes", latex_label=LatexDecayProcesses.reco_dss_val_charged_pi)

    dss_neutral_b_to_dss_pi_l_via_d_modes = RecoModeGroupInfo(ids=_dss_via_d_neutral_b_reco_modes.ids + _dss_charged_via_d_neutral_b_reco_modes.ids, name="dss_neutral_b_to_dss_pi_l_via_d_modes", latex_label=LatexDecayProcesses.reco_dss_val_neutral_b_d)
    dss_neutral_b_to_dss_pi_l_via_ds_modes = RecoModeGroupInfo(ids=_dss_via_ds_neutral_b_reco_modes.ids + _dss_charged_via_ds_neutral_b_reco_modes.ids, name="dss_neutral_b_to_dss_pi_l_via_ds_modes", latex_label=LatexDecayProcesses.reco_dss_val_neutral_b_d_star)
    dss_charged_b_to_dss_pi_l_via_d_modes = RecoModeGroupInfo(ids=_dss_via_d_charged_b_reco_modes.ids + _dss_charged_via_d_charged_b_reco_modes.ids, name="dss_charged_b_to_dss_pi_l_via_d_modes", latex_label=LatexDecayProcesses.reco_dss_val_charged_b_d)
    dss_charged_b_to_dss_pi_l_via_ds_modes = RecoModeGroupInfo(ids=_dss_via_ds_charged_b_reco_modes.ids + _dss_charged_via_ds_charged_b_reco_modes.ids, name="dss_charged_b_to_dss_pi_l_via_ds_modes", latex_label=LatexDecayProcesses.reco_dss_val_charged_b_d_star)

    dss_via_d_reco_modes = RecoModeGroupInfo(ids=dss_d_l_neutral_pion_modes.ids + dss_d_l_charged_pion_modes.ids, name="dss_via_d_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_l_with_dss_to_d_pi)
    dss_via_ds_reco_modes = RecoModeGroupInfo(ids=dss_ds_l_neutral_pion_modes.ids + dss_ds_l_charged_pion_modes.ids, name="dss_via_ds_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_l_with_dss_to_ds_pi)

    d_ss_reco_modes = RecoModeGroupInfo(ids=dss_via_d_reco_modes.ids + dss_via_ds_reco_modes.ids, name="d_ss_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_dss_l_with_dss_to_d_wwos_pi)

    d_reco_with_dss_modes = RecoModeGroupInfo(ids=_d_reco_modes.ids + dss_via_d_reco_modes.ids, name="d_reco_with_dss_modes", latex_label=LatexDecayProcesses.reco_b_to_d_w_wo_pi_l)
    d_star_reco_with_dss_modes = RecoModeGroupInfo(ids=_d_star_reco_modes.ids + dss_via_ds_reco_modes.ids, name="d_star_reco_with_dss_modes", latex_label=LatexDecayProcesses.reco_b_to_ds_w_wo_pi_l)

    neutral_b_to_d_with_dss_modes = RecoModeGroupInfo(ids=neutral_b_to_d_modes.ids + dss_neutral_b_to_dss_pi_l_via_d_modes.ids, name="neutral_b_to_d_with_dss_modes", latex_label=LatexDecayProcesses.reco_b0_to_d_w_wo_pi_l)
    neutral_b_to_ds_with_dss_modes = RecoModeGroupInfo(ids=neutral_b_to_ds_modes.ids + dss_neutral_b_to_dss_pi_l_via_ds_modes.ids, name="neutral_b_to_ds_with_dss_modes", latex_label=LatexDecayProcesses.reco_b0_to_ds_w_wo_pi_l)
    charged_b_to_d_with_dss_modes = RecoModeGroupInfo(ids=charged_b_to_d_modes.ids + dss_charged_b_to_dss_pi_l_via_d_modes.ids, name="charged_b_to_d_with_dss_modes", latex_label=LatexDecayProcesses.reco_bp_to_d_w_wo_pi_l)
    charged_b_to_ds_with_dss_modes = RecoModeGroupInfo(ids=charged_b_to_ds_modes.ids + dss_charged_b_to_dss_pi_l_via_ds_modes.ids, name="charged_b_to_ds_with_dss_modes", latex_label=LatexDecayProcesses.reco_bp_to_ds_w_wo_pi_l)

    # D(*(*)) Charge dependent reco mode groups
    _neutral_b_to_dss_via_neutral_d_pion_l_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_d_pip_modes  # Duplicates: RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_d_pip_mode.id, neutral_b_to_dss_mu_via_d_pip_mode.id), name="neutral_b_to_dss_via_neutral_d_pion_l_modes", latex_label="Label_neutral_b_to_dss_via_neutral_d_pion_l_modes")
    _neutral_b_to_dss_via_neutral_ds_pion_l_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_ds_pip_modes  # Duplicates: RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_ds_pip_mode.id, neutral_b_to_dss_mu_via_ds_pip_mode.id), name="neutral_b_to_dss_via_neutral_ds_pion_l_modes", latex_label="Label_neutral_b_to_dss_via_neutral_ds_pion_l_modes")
    _neutral_b_to_dss_via_charged_d_pion_l_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_d_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_d_pi0_mode.id, neutral_b_to_dss_mu_via_d_pi0_mode.id), name="neutral_b_to_dss_via_charged_d_pion_l_modes", latex_label="Label_neutral_b_to_dss_via_charged_d_pion_l_modes")
    _neutral_b_to_dss_via_charged_ds_pion_l_modes: RecoModeGroupInfo = neutral_b_to_dss_l_via_ds_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=(neutral_b_to_dss_e_via_ds_pi0_mode.id, neutral_b_to_dss_mu_via_ds_pi0_mode.id), name="neutral_b_to_dss_via_charged_ds_pion_l_modes", latex_label="Label_neutral_b_to_dss_via_charged_ds_pion_l_modes")

    _charged_b_to_dss_via_neutral_d_pion_l_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_d_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_d_pi0_mode.id, charged_b_to_dss_mu_via_d_pi0_mode.id), name="charged_b_to_dss_via_neutral_d_pion_l_modes", latex_label="Label_charged_b_to_dss_via_neutral_d_pion_l_modes")
    _charged_b_to_dss_via_neutral_ds_pion_l_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_ds_pi0_modes  # Duplicates: RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_ds_pi0_mode.id, charged_b_to_dss_mu_via_ds_pi0_mode.id), name="charged_b_to_dss_via_neutral_ds_pion_l_modes", latex_label="Label_charged_b_to_dss_via_neutral_ds_pion_l_modes")
    _charged_b_to_dss_via_charged_d_pion_l_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_d_pip_modes  # Duplicates: RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_d_pip_mode.id, charged_b_to_dss_mu_via_d_pip_mode.id), name="charged_b_to_dss_via_charged_d_pion_l_modes", latex_label="Label_charged_b_to_dss_via_charged_d_pion_l_modes")
    _charged_b_to_dss_via_charged_ds_pion_l_modes: RecoModeGroupInfo = charged_b_to_dss_l_via_ds_pip_modes  # Duplicates: RecoModeGroupInfo(ids=(charged_b_to_dss_e_via_ds_pip_mode.id, charged_b_to_dss_mu_via_ds_pip_mode.id), name="charged_b_to_dss_via_charged_ds_pion_l_modes", latex_label="Label_charged_b_to_dss_via_charged_ds_pion_l_modes")

    dss_via_neutral_d_l_pion_modes = RecoModeGroupInfo(ids=_neutral_b_to_dss_via_neutral_d_pion_l_modes.ids + _charged_b_to_dss_via_neutral_d_pion_l_modes.ids, name="dss_via_neutral_d_l_pion_modes", latex_label=LatexDecayProcesses.reco_b_to_d0_pi_l)
    dss_via_neutral_ds_l_pion_modes = RecoModeGroupInfo(ids=_neutral_b_to_dss_via_neutral_ds_pion_l_modes.ids + _charged_b_to_dss_via_neutral_ds_pion_l_modes.ids, name="dss_via_neutral_ds_l_pion_modes", latex_label=LatexDecayProcesses.reco_b_to_ds0_pi_l)
    dss_via_charged_d_l_pion_modes = RecoModeGroupInfo(ids=_neutral_b_to_dss_via_charged_d_pion_l_modes.ids + _charged_b_to_dss_via_charged_d_pion_l_modes.ids, name="dss_via_charged_d_l_pion_modes", latex_label=LatexDecayProcesses.reco_b_to_dm_pi_l)
    dss_via_charged_ds_l_pion_modes = RecoModeGroupInfo(ids=_neutral_b_to_dss_via_charged_ds_pion_l_modes.ids + _charged_b_to_dss_via_charged_ds_pion_l_modes.ids, name="dss_via_charged_ds_l_pion_modes", latex_label=LatexDecayProcesses.reco_b_to_dsm_pi_l)

    dss_via_neutral_dwwos_l_pion_modes = RecoModeGroupInfo(ids=dss_via_neutral_d_l_pion_modes.ids + dss_via_neutral_ds_l_pion_modes.ids, name="dss_via_neutral_dwwos_l_pion_modes", latex_label=LatexDecayProcesses.reco_b_to_d0_wwos_pi_l)
    dss_via_charged_dwwos_l_pion_modes = RecoModeGroupInfo(ids=dss_via_charged_d_l_pion_modes.ids + dss_via_charged_ds_l_pion_modes.ids, name="dss_via_charged_dwwos_l_pion_modes", latex_label=LatexDecayProcesses.reco_b_to_dm_wwos_pi_l)

    charged_d_wwo_pi_reco_modes = RecoModeGroupInfo(ids=neutral_b_to_d_modes.ids + dss_via_charged_d_l_pion_modes.ids, name="charged_d_wwo_pi_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_dm_w_wo_pi_l)
    charged_ds_wwo_pi_reco_modes = RecoModeGroupInfo(ids=neutral_b_to_ds_modes.ids + dss_via_charged_ds_l_pion_modes.ids, name="charged_ds_wwo_pi_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_dsm_w_wo_pi_l)
    neutral_d_wwo_pi_reco_modes = RecoModeGroupInfo(ids=charged_b_to_d_modes.ids + dss_via_neutral_d_l_pion_modes.ids, name="neutral_d_wwo_pi_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_d0_w_wo_pi_l)
    neutral_ds_wwo_pi_reco_modes = RecoModeGroupInfo(ids=charged_b_to_ds_modes.ids + dss_via_neutral_ds_l_pion_modes.ids, name="neutral_ds_wwo_pi_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_ds0_w_wo_pi_l)

    # TODO: Remove duplicates!
    _charged_dwwos_reco_modes: RecoModeGroupInfo = neutral_b_to_d_wwo_star_modes  # Duplicates!
    _neutral_dwwos_reco_modes: RecoModeGroupInfo = charged_b_to_d_wwo_star_modes  # Duplicates!

    charged_dwwos_wwoss_reco_modes = RecoModeGroupInfo(ids=neutral_b_to_d_wwo_star_modes.ids + dss_via_charged_dwwos_l_pion_modes.ids, name="charged_dwwos_wwoss_reco_modes", latex_label=LatexDecayProcesses.reco_b0_to_dm_wwos_wwos_l)
    neutral_dwwos_wwoss_reco_modes = RecoModeGroupInfo(ids=charged_b_to_d_wwo_star_modes.ids + dss_via_neutral_dwwos_l_pion_modes.ids, name="neutral_dwwos_wwoss_reco_modes", latex_label=LatexDecayProcesses.reco_bp_to_d0_wwos_wwos_l)

    #
    # B -> pi l  Reco Modes
    b_to_pi_e_modes = RecoModeGroupInfo(ids=(neutral_b_to_pip_e_mode.id, charged_b_to_pi0_e_mode.id), name="b_to_pi_e_modes", latex_label=LatexDecayProcesses.reco_b_to_pi_e)
    b_to_pi_mu_modes = RecoModeGroupInfo(ids=(neutral_b_to_pip_mu_mode.id, charged_b_to_pi0_mu_mode.id), name="b_to_pi_mu_modes", latex_label=LatexDecayProcesses.reco_b_to_pi_mu)

    b_to_pi0_l_modes = RecoModeGroupInfo(ids=(charged_b_to_pi0_e_mode.id, charged_b_to_pi0_mu_mode.id), name="b_to_pi0_l_modes", latex_label=LatexDecayProcesses.reco_bp_to_pi0_lp)
    b_to_pip_l_modes = RecoModeGroupInfo(ids=(neutral_b_to_pip_e_mode.id, neutral_b_to_pip_mu_mode.id), name="b_to_pip_l_modes", latex_label=LatexDecayProcesses.reco_b0_to_pim_lp)

    pi_l_nu_reco_modes = RecoModeGroupInfo(ids=b_to_pi0_l_modes.ids + b_to_pip_l_modes.ids, name="pi_l_nu_reco_modes", latex_label=LatexDecayProcesses.reco_b_to_pi_l)

    # fmt: on

    @classmethod
    def _get_additional_method_names(cls) -> Tuple[str, ...]:
        return ()

    @classmethod
    def run_dedicated_checks(cls) -> None:
        assert len(RecoModeInfos.charged_reco_modes.ids) == len(RecoModeInfos.neutral_reco_modes.ids), (
            len(RecoModeInfos.charged_reco_modes.ids),
            len(RecoModeInfos.neutral_reco_modes.ids),
        )

        reco_mode_id_to_label_map: RecoModeIDToTitleDictType = cls.get_reco_id_to_latex_label()
        len_all_main_mode_combos: int = len(RecoModeInfos.charged_reco_modes.ids) + len(
            RecoModeInfos.neutral_reco_modes.ids
        )
        len_all_mode_combos: int = len_all_main_mode_combos + len(RecoModeInfos.pi_l_nu_reco_modes.ids)
        assert len_all_mode_combos == len(reco_mode_id_to_label_map), (
            len(RecoModeInfos.charged_reco_modes.ids),
            len(RecoModeInfos.neutral_reco_modes.ids),
            len_all_main_mode_combos,
            len(RecoModeInfos.pi_l_nu_reco_modes.ids),
            len_all_mode_combos,
            len(reco_mode_id_to_label_map),
        )

        list_of_all_main_reco_modes = list(RecoModeInfos.charged_reco_modes.ids) + list(
            RecoModeInfos.neutral_reco_modes.ids
        )
        list_of_all_reco_modes = list_of_all_main_reco_modes + list(RecoModeInfos.pi_l_nu_reco_modes.ids)
        assert set(list_of_all_reco_modes) == set([v for v in reco_mode_id_to_label_map]), (
            set(list_of_all_reco_modes),
            set(reco_mode_id_to_label_map.keys()),
            [k for k in list_of_all_reco_modes if k not in reco_mode_id_to_label_map],
            [k for k in reco_mode_id_to_label_map if k not in list_of_all_reco_modes],
        )

        assert all(c in RecoModeInfos.neutral_reco_modes.ids for c in RecoModeInfos.neutral_b_to_d_modes.ids)
        assert all(c in RecoModeInfos.neutral_reco_modes.ids for c in RecoModeInfos.neutral_b_to_ds_modes.ids)
        assert all(c in RecoModeInfos.neutral_reco_modes.ids for c in RecoModeInfos.neutral_b_to_d_with_dss_modes.ids)
        assert all(c in RecoModeInfos.neutral_reco_modes.ids for c in RecoModeInfos.neutral_b_to_ds_with_dss_modes.ids)

        assert all(c in RecoModeInfos.charged_reco_modes.ids for c in RecoModeInfos.charged_b_to_d_modes.ids)
        assert all(c in RecoModeInfos.charged_reco_modes.ids for c in RecoModeInfos.charged_b_to_ds_modes.ids)
        assert all(c in RecoModeInfos.charged_reco_modes.ids for c in RecoModeInfos.charged_b_to_d_with_dss_modes.ids)
        assert all(c in RecoModeInfos.charged_reco_modes.ids for c in RecoModeInfos.charged_b_to_ds_with_dss_modes.ids)

        _all_dss_test1 = set(
            RecoModeInfos.charged_b_to_dss_l_via_pi0_modes.ids
            + RecoModeInfos.charged_b_to_dss_l_via_pip_modes.ids
            + RecoModeInfos.neutral_b_to_dss_l_via_pi0_modes.ids
            + RecoModeInfos.neutral_b_to_dss_l_via_pip_modes.ids
        )
        assert set(RecoModeInfos.d_ss_reco_modes.ids) == _all_dss_test1, (
            set(RecoModeInfos.d_ss_reco_modes.ids),
            _all_dss_test1,
        )

        _all_dss_test11 = set(
            RecoModeInfos.charged_b_to_dss_l_via_pi0_modes.ids + RecoModeInfos.neutral_b_to_dss_l_via_pi0_modes.ids
        )
        assert set(RecoModeInfos.d_ss_neutral_pion_reco_modes.ids) == _all_dss_test11, (
            set(RecoModeInfos.d_ss_neutral_pion_reco_modes.ids),
            _all_dss_test11,
        )

        _all_dss_test12 = set(
            RecoModeInfos.charged_b_to_dss_l_via_pip_modes.ids + RecoModeInfos.neutral_b_to_dss_l_via_pip_modes.ids
        )
        assert set(RecoModeInfos.d_ss_charged_pion_reco_modes.ids) == _all_dss_test12, (
            set(RecoModeInfos.d_ss_neutral_pion_reco_modes.ids),
            _all_dss_test12,
        )

        _all_dss_test2 = set(
            RecoModeInfos.dss_neutral_b_to_dss_pi_l_via_d_modes.ids
            + RecoModeInfos.dss_neutral_b_to_dss_pi_l_via_ds_modes.ids
            + RecoModeInfos.dss_charged_b_to_dss_pi_l_via_d_modes.ids
            + RecoModeInfos.dss_charged_b_to_dss_pi_l_via_ds_modes.ids
        )
        assert set(RecoModeInfos.d_ss_reco_modes.ids) == _all_dss_test2, (
            set(RecoModeInfos.d_ss_reco_modes.ids),
            _all_dss_test2,
        )

        _all_dss_test3 = set(
            RecoModeInfos.dss_d_l_neutral_pion_modes.ids
            + RecoModeInfos.dss_ds_l_neutral_pion_modes.ids
            + RecoModeInfos.dss_d_l_charged_pion_modes.ids
            + RecoModeInfos.dss_ds_l_charged_pion_modes.ids
        )
        assert set(RecoModeInfos.d_ss_reco_modes.ids) == _all_dss_test3, (
            set(RecoModeInfos.d_ss_reco_modes.ids),
            _all_dss_test3,
        )

        _all_dss_test31 = set(
            RecoModeInfos.dss_d_l_neutral_pion_modes.ids + RecoModeInfos.dss_ds_l_neutral_pion_modes.ids
        )
        assert set(RecoModeInfos.d_ss_neutral_pion_reco_modes.ids) == _all_dss_test31, (
            set(RecoModeInfos.d_ss_neutral_pion_reco_modes.ids),
            _all_dss_test31,
        )

        _all_dss_test32 = set(
            RecoModeInfos.dss_d_l_charged_pion_modes.ids + RecoModeInfos.dss_ds_l_charged_pion_modes.ids
        )
        assert set(RecoModeInfos.d_ss_charged_pion_reco_modes.ids) == _all_dss_test32, (
            set(RecoModeInfos.d_ss_neutral_pion_reco_modes.ids),
            _all_dss_test32,
        )


class TagRecoModeInfos(AbstractRecoModeInfosBaseClass):
    fei_reco_mode_col_name: str = "__advanced_fei_reco_mode_col__"
    fei_plus_sig_d_reco_mode_col_name: str = "__advanced_fei_plus_sig_d_reco_mode_col__"

    sig_b_via_ds_offset: int = 100

    _all_reco_modes_latex_label: str = LatexDecayProcesses.all_b_tag_modes_label
    _all_reco_modes_latex_label_short: str = LatexDecayProcesses.all_b_tag_modes_label_short

    @staticmethod
    def calculate_new_fei_reco_mode_ids(df: pd.DataFrame) -> np.ndarray:
        fei_reco_mode_values: np.ndarray = FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_ids(
            df=df,
            b_meson_pdg_column_name=Cols.b_tag_reco_pdg,
            fei_decay_mode_id_column_name=Cols.tag_reco_mode_column_name,
        )
        return fei_reco_mode_values

    @staticmethod
    def add_new_fei_reco_mode_col(df: pd.DataFrame, check_if_new: bool) -> None:
        if check_if_new:
            assert TagRecoModeInfos.fei_reco_mode_col_name not in df.columns, TagRecoModeInfos.fei_reco_mode_col_name

        fei_reco_mode_values: np.ndarray = TagRecoModeInfos.calculate_new_fei_reco_mode_ids(df=df)
        df.loc[:, TagRecoModeInfos.fei_reco_mode_col_name] = fei_reco_mode_values

    @staticmethod
    def add_sig_d_plus_fei_reco_mode_col(df: pd.DataFrame, check_if_new: bool) -> None:
        if check_if_new:
            assert (
                TagRecoModeInfos.fei_plus_sig_d_reco_mode_col_name not in df.columns
            ), TagRecoModeInfos.fei_plus_sig_d_reco_mode_col_name

        if TagRecoModeInfos.fei_reco_mode_col_name not in df.columns:
            TagRecoModeInfos.add_new_fei_reco_mode_col(df=df, check_if_new=False)

        b_tag_reco_mode_values: np.ndarray = df[TagRecoModeInfos.fei_reco_mode_col_name].values

        sig_b_reco_mode: np.ndarray = df[Cols.decay_mode_column_name].values
        is_sig_b_via_dst_reco_mode = np.isin(sig_b_reco_mode, RecoModeInfos.d_star_reco_with_dss_modes.ids)

        new_values: np.array = np.select(
            condlist=[
                is_sig_b_via_dst_reco_mode & (b_tag_reco_mode_values >= 0),
                is_sig_b_via_dst_reco_mode & (b_tag_reco_mode_values < 0),
                ~is_sig_b_via_dst_reco_mode,
            ],
            choicelist=[
                b_tag_reco_mode_values + TagRecoModeInfos.sig_b_via_ds_offset,
                b_tag_reco_mode_values - TagRecoModeInfos.sig_b_via_ds_offset,
                b_tag_reco_mode_values,
            ],
            default=np.nan,  # This should never happen, as all reco channels are either charged or neutral ones...
        )

        assert not np.any(np.isnan(new_values)), (sum(np.isnan(new_values)), np.unique(new_values))

        df.loc[:, TagRecoModeInfos.fei_plus_sig_d_reco_mode_col_name] = new_values

    @staticmethod
    def get_fei_reco_mode_id_to_latex_label_mapping() -> Dict[int, str]:
        try:
            return FEIDecayModeIDInterpreter.get_modified_fei_decay_mode_to_latex_label_dict(
                fei_settings_dict=None,
                include_b_to=True,
            )
        except TypeError:
            return FEIDecayModeIDInterpreter.get_backup_fei_reco_mode_latex_label_mapping()

    @staticmethod
    def get_ids_with_sig_ds_offsets(ids: Tuple[int, ...]) -> Tuple[int, ...]:
        return tuple(_id + ((-1 if _id < 0 else 1) * TagRecoModeInfos.sig_b_via_ds_offset) for _id in ids)

    @staticmethod
    def get_id_group_info_with_sig_ds_offsets(id_group: RecoModeGroupInfo) -> RecoModeGroupInfo:
        new_ids: Tuple[int, ...] = TagRecoModeInfos.get_ids_with_sig_ds_offsets(ids=id_group.ids)
        return RecoModeGroupInfo(ids=new_ids, name=id_group.name, latex_label=id_group.latex_label)

    # fmt: off
    # FEI Tag Reco Mode ID Groups:
    _charged_tag_reco_modes_ids: Tuple[int, ...] = tuple([int(i) for i in np.arange(0, 29, 1)])
    charged_tag_reco_modes = RecoModeGroupInfo(ids=_charged_tag_reco_modes_ids, name="charged_tag_reco_modes", latex_label=LatexDecayProcesses.charged_tag_reco_modes)
    _neutral_tag_reco_modes_ids: Tuple[int, ...] = tuple([int(-1 * i) for i in np.arange(1, 27, 1)])
    neutral_tag_reco_modes = RecoModeGroupInfo(ids=_neutral_tag_reco_modes_ids, name="neutral_tag_reco_modes", latex_label=LatexDecayProcesses.neutral_tag_reco_modes)

    _charged_b_to_d_1cpi_tag_modes_ids: Tuple[int, ...] = (0, 1, 2)
    charged_b_to_d_1cpi_tag_modes = RecoModeGroupInfo(ids=_charged_b_to_d_1cpi_tag_modes_ids, name="charged_b_to_d_1cpi_tag_modes", latex_label=LatexDecayProcesses.charged_b_to_d_1cpi_tag_modes)
    _neutral_b_to_d_1cpi_tag_modes_ids: Tuple[int, ...] = (-1, -2, -3)
    neutral_b_to_d_1cpi_tag_modes = RecoModeGroupInfo(ids=_neutral_b_to_d_1cpi_tag_modes_ids, name="neutral_b_to_d_1cpi_tag_modes", latex_label=LatexDecayProcesses.neutral_b_to_d_1cpi_tag_modes)
    _b_to_d_1cpi_tag_modes_ids: Tuple[int, ...] = charged_b_to_d_1cpi_tag_modes.ids + neutral_b_to_d_1cpi_tag_modes.ids
    b_to_d_1cpi_tag_modes = RecoModeGroupInfo(ids=_b_to_d_1cpi_tag_modes_ids, name="b_to_d_1cpi_tag_modes", latex_label=LatexDecayProcesses.b_to_d_1cpi_tag_modes)

    _charged_b_to_d_mcpi_tag_modes_ids: Tuple[int, ...] = (3, 4)
    charged_b_to_d_mcpi_tag_modes = RecoModeGroupInfo(ids=_charged_b_to_d_mcpi_tag_modes_ids, name="charged_b_to_d_mcpi_tag_modes", latex_label=LatexDecayProcesses.charged_b_to_d_mcpi_tag_modes)
    _neutral_b_to_d_mcpi_tag_modes_ids: Tuple[int, ...] = (-4, -5)
    neutral_b_to_d_mcpi_tag_modes = RecoModeGroupInfo(ids=_neutral_b_to_d_mcpi_tag_modes_ids, name="neutral_b_to_d_mcpi_tag_modes", latex_label=LatexDecayProcesses.neutral_b_to_d_mcpi_tag_modes)
    _b_to_d_mcpi_tag_modes_ids: Tuple[int, ...] = charged_b_to_d_mcpi_tag_modes.ids + neutral_b_to_d_mcpi_tag_modes.ids
    b_to_d_mcpi_tag_modes = RecoModeGroupInfo(ids=_b_to_d_mcpi_tag_modes_ids, name="b_to_d_mcpi_tag_modes", latex_label=LatexDecayProcesses.b_to_d_mcpi_tag_modes)

    _charged_b_to_ocd_pipi_tag_modes_ids: Tuple[int, ...] = (23, 24)
    charged_b_to_ocd_pipi_tag_modes = RecoModeGroupInfo(ids=_charged_b_to_ocd_pipi_tag_modes_ids, name="charged_b_to_ocd_pipi_tag_modes", latex_label=LatexDecayProcesses.charged_b_to_ocd_pipi_tag_modes)
    _neutral_b_to_ocd_pipi_tag_modes_ids: Tuple[int, ...] = (-6,)
    neutral_b_to_ocd_pipi_tag_modes = RecoModeGroupInfo(ids=_neutral_b_to_ocd_pipi_tag_modes_ids, name="neutral_b_to_ocd_pipi_tag_modes", latex_label=LatexDecayProcesses.neutral_b_to_ocd_pipi_tag_modes)
    _b_to_ocd_pipi_tag_modes_ids: Tuple[int, ...] = charged_b_to_ocd_pipi_tag_modes.ids + neutral_b_to_ocd_pipi_tag_modes.ids
    b_to_ocd_pipi_tag_modes = RecoModeGroupInfo(ids=_b_to_ocd_pipi_tag_modes_ids, name="b_to_ocd_pipi_tag_modes", latex_label=LatexDecayProcesses.b_to_ocd_pipi_tag_modes)

    _charged_b_to_dd_tag_modes_ids: Tuple[int, ...] = (5, 6, 7, 8, 9, 10, 11, 12, 13, 14)
    charged_b_to_dd_tag_modes = RecoModeGroupInfo(ids=_charged_b_to_dd_tag_modes_ids, name="charged_b_to_dd_tag_modes", latex_label=LatexDecayProcesses.charged_b_to_dd_tag_modes)
    _neutral_b_to_dd_tag_modes_ids: Tuple[int, ...] = (-7, -8, -9, -10, -11, -12, -13, -14, -15)
    neutral_b_to_dd_tag_modes = RecoModeGroupInfo(ids=_neutral_b_to_dd_tag_modes_ids, name="neutral_b_to_dd_tag_modes", latex_label=LatexDecayProcesses.neutral_b_to_dd_tag_modes)
    _b_to_dd_tag_modes_ids: Tuple[int, ...] = charged_b_to_dd_tag_modes.ids + neutral_b_to_dd_tag_modes.ids
    b_to_dd_tag_modes = RecoModeGroupInfo(ids=_b_to_dd_tag_modes_ids, name="b_to_dd_tag_modes", latex_label=LatexDecayProcesses.b_to_dd_tag_modes)

    _charged_b_to_ds_1cpi_tag_modes_ids: Tuple[int, ...] = (15, 16, 17)
    charged_b_to_ds_1cpi_tag_modes = RecoModeGroupInfo(ids=_charged_b_to_ds_1cpi_tag_modes_ids, name="charged_b_to_ds_1cpi_tag_modes", latex_label=LatexDecayProcesses.charged_b_to_ds_1cpi_tag_modes)
    _neutral_b_to_ds_1cpi_tag_modes_ids: Tuple[int, ...] = (-16, -17, -18)
    neutral_b_to_ds_1cpi_tag_modes = RecoModeGroupInfo(ids=_neutral_b_to_ds_1cpi_tag_modes_ids, name="neutral_b_to_ds_1cpi_tag_modes", latex_label=LatexDecayProcesses.neutral_b_to_ds_1cpi_tag_modes)
    _b_to_ds_1cpi_tag_modes_ids: Tuple[int, ...] = charged_b_to_ds_1cpi_tag_modes.ids + neutral_b_to_ds_1cpi_tag_modes.ids
    b_to_ds_1cpi_tag_modes = RecoModeGroupInfo(ids=_b_to_ds_1cpi_tag_modes_ids, name="b_to_ds_1cpi_tag_modes", latex_label=LatexDecayProcesses.b_to_ds_1cpi_tag_modes)

    _charged_b_to_ds_mcpi_tag_modes_ids: Tuple[int, ...] = (18, 19)
    charged_b_to_ds_mcpi_tag_modes = RecoModeGroupInfo(ids=_charged_b_to_ds_mcpi_tag_modes_ids, name="charged_b_to_ds_mcpi_tag_modes", latex_label=LatexDecayProcesses.charged_b_to_ds_mcpi_tag_modes)
    _neutral_b_to_ds_mcpi_tag_modes_ids: Tuple[int, ...] = (-19, -20)
    neutral_b_to_ds_mcpi_tag_modes = RecoModeGroupInfo(ids=_neutral_b_to_ds_mcpi_tag_modes_ids, name="neutral_b_to_ds_mcpi_tag_modes", latex_label=LatexDecayProcesses.neutral_b_to_ds_mcpi_tag_modes)
    _b_to_ds_mcpi_tag_modes_ids: Tuple[int, ...] = charged_b_to_ds_mcpi_tag_modes.ids + neutral_b_to_ds_mcpi_tag_modes.ids
    b_to_ds_mcpi_tag_modes = RecoModeGroupInfo(ids=_b_to_ds_mcpi_tag_modes_ids, name="b_to_ds_mcpi_tag_modes", latex_label=LatexDecayProcesses.b_to_ds_mcpi_tag_modes)

    _charged_b_to_dsd_tag_modes_ids: Tuple[int, ...] = (20, 21, 22)
    charged_b_to_dsd_tag_modes = RecoModeGroupInfo(ids=_charged_b_to_dsd_tag_modes_ids, name="charged_b_to_dsd_tag_modes", latex_label=LatexDecayProcesses.charged_b_to_dsd_tag_modes)
    _neutral_b_to_dsd_tag_modes_ids: Tuple[int, ...] = (-21, -22, -23)
    neutral_b_to_dsd_tag_modes = RecoModeGroupInfo(ids=_neutral_b_to_dsd_tag_modes_ids, name="neutral_b_to_dsd_tag_modes", latex_label=LatexDecayProcesses.neutral_b_to_dsd_tag_modes)
    _b_to_dsd_tag_modes_ids: Tuple[int, ...] = charged_b_to_dsd_tag_modes.ids + neutral_b_to_dsd_tag_modes.ids
    b_to_dsd_tag_modes = RecoModeGroupInfo(ids=_b_to_dsd_tag_modes_ids, name="b_to_dsd_tag_modes", latex_label=LatexDecayProcesses.b_to_dsd_tag_modes)

    _charged_b_to_j_psi_tag_modes_ids: Tuple[int, ...] = (25, 26, 27, 28)
    charged_b_to_j_psi_tag_modes = RecoModeGroupInfo(ids=_charged_b_to_j_psi_tag_modes_ids, name="charged_b_to_j_psi_tag_modes", latex_label=LatexDecayProcesses.charged_b_to_j_psi_tag_modes)
    _neutral_b_to_j_psi_tag_modes_ids: Tuple[int, ...] = (-24, -25, -26)
    neutral_b_to_j_psi_tag_modes = RecoModeGroupInfo(ids=_neutral_b_to_j_psi_tag_modes_ids, name="neutral_b_to_j_psi_tag_modes", latex_label=LatexDecayProcesses.neutral_b_to_j_psi_tag_modes)
    _b_to_j_psi_tag_modes_ids: Tuple[int, ...] = charged_b_to_j_psi_tag_modes.ids + neutral_b_to_j_psi_tag_modes.ids
    b_to_j_psi_tag_modes = RecoModeGroupInfo(ids=_b_to_j_psi_tag_modes_ids, name="b_to_j_psi_tag_modes", latex_label=LatexDecayProcesses.b_to_j_psi_tag_modes)

    _charged_b_to__all_dd__tag_modes_ids: Tuple[int, ...] = charged_b_to_dd_tag_modes.ids + charged_b_to_dsd_tag_modes.ids
    charged_b_to__all_dd__tag_modes = RecoModeGroupInfo(ids=_charged_b_to__all_dd__tag_modes_ids, name="charged_b_to__all_dd__tag_modes", latex_label=LatexDecayProcesses.charged_b_to__other__tag_modes)
    _neutral_b_to__all_dd__tag_modes_ids: Tuple[int, ...] = neutral_b_to_dd_tag_modes.ids + neutral_b_to_dsd_tag_modes.ids
    neutral_b_to__all_dd__tag_modes = RecoModeGroupInfo(ids=_neutral_b_to__all_dd__tag_modes_ids, name="neutral_b_to__all_dd__tag_modes", latex_label=LatexDecayProcesses.neutral_b_to__other__tag_modes)
    _b_to__all_dd__tag_modes_ids: Tuple[int, ...] = charged_b_to__all_dd__tag_modes.ids + neutral_b_to__all_dd__tag_modes.ids
    b_to__all_dd__tag_modes = RecoModeGroupInfo(ids=_b_to__all_dd__tag_modes_ids, name="b_to__all_dd__tag_modes", latex_label=LatexDecayProcesses.b_to__other__tag_modes)

    _charged_b_to__other__tag_modes_ids: Tuple[int, ...] = charged_b_to__all_dd__tag_modes.ids + charged_b_to_j_psi_tag_modes.ids
    charged_b_to__other__tag_modes = RecoModeGroupInfo(ids=_charged_b_to__other__tag_modes_ids, name="charged_b_to__other__tag_modes", latex_label=LatexDecayProcesses.charged_b_to__other__tag_modes + r"\,\!$")
    _neutral_b_to__other__tag_modes_ids: Tuple[int, ...] = neutral_b_to__all_dd__tag_modes.ids + neutral_b_to_j_psi_tag_modes.ids
    neutral_b_to__other__tag_modes = RecoModeGroupInfo(ids=_neutral_b_to__other__tag_modes_ids, name="neutral_b_to__other__tag_modes", latex_label=LatexDecayProcesses.neutral_b_to__other__tag_modes + r"\,\!$")
    _b_to__other__tag_modes_ids: Tuple[int, ...] = charged_b_to__other__tag_modes.ids + neutral_b_to__other__tag_modes.ids
    b_to__other__tag_modes = RecoModeGroupInfo(ids=_b_to__other__tag_modes_ids, name="b_to__other__tag_modes", latex_label=LatexDecayProcesses.b_to__other__tag_modes + r"\,\!$")
    # fmt: on

    @staticmethod
    def get_fei_reco_mode_name_for(fei_reco_id: int) -> str:
        return f"fei_reco_mode_{fei_reco_id}_wnv"  # wnv stands for "with negative values"

    @classmethod
    def get_reco_ids_dict(cls) -> Dict[str, RecoModeIDInfo]:
        return {
            TagRecoModeInfos.get_fei_reco_mode_name_for(fei_reco_id=_id): RecoModeIDInfo(
                id=_id,
                name=TagRecoModeInfos.get_fei_reco_mode_name_for(fei_reco_id=_id),
                latex_label=latex_label,
            )
            for _id, latex_label in cls.get_fei_reco_mode_id_to_latex_label_mapping().items()
        }

    @classmethod
    def get_all_reco_mode_ids(cls) -> Tuple[int, ...]:
        return tuple(_id for _id in cls.get_fei_reco_mode_id_to_latex_label_mapping().keys())

    @classmethod
    def _get_additional_method_names(cls) -> Tuple[str, ...]:
        return (
            "get_fei_reco_mode_name_for",
            "calculate_new_fei_reco_mode_ids",
            "add_new_fei_reco_mode_col",
            "add_sig_d_plus_fei_reco_mode_col",
            "get_fei_reco_mode_id_to_latex_label_mapping",
            "get_ids_with_sig_ds_offsets",
            "get_id_group_info_with_sig_ds_offsets",
        )

    @classmethod
    def run_dedicated_checks(cls) -> None:
        pass


class DMesonRecoModeInfos(AbstractRecoModeInfosBaseClass):
    d_meson_reco_mode_col_name: str = Cols.primary_d_meson_reco_mode

    _all_reco_modes_latex_label: str = LatexDecayProcesses.all_d_wwost_meson_reco_modes_label
    _all_reco_modes_latex_label_short: str = LatexDecayProcesses.all_d_wwost_meson_reco_modes_label_short

    # fmt: off

    # D*0 decay Channels
    dst0_to_d0_pi0: RecoModeIDInfo = RecoModeIDInfo(id=21, name="dst0_to_d0_pi0", latex_label=LatexDecayProcesses.dst0_to_d0_pi0)  #                                             D*0  to  D0 pi0                 # noqa: E262
    dst0_to_d0_gamma: RecoModeIDInfo = RecoModeIDInfo(id=22, name="dst0_to_d0_gamma", latex_label=LatexDecayProcesses.dst0_to_d0_gamma)  #                                       D*0  to  D0 gamma               # noqa: E262
    # D*+ decay Channels
    dstp_to_d0_pip: RecoModeIDInfo = RecoModeIDInfo(id=23, name="dstp_to_d0_pip", latex_label=LatexDecayProcesses.dstp_to_d0_pip)  #                                             D*+  to  D0 pi+                 # noqa: E262
    dstp_to_dp_pi0: RecoModeIDInfo = RecoModeIDInfo(id=24, name="dstp_to_dp_pi0", latex_label=LatexDecayProcesses.dstp_to_dp_pi0)  #                                             D*+  to  D+ pi0                 # noqa: E262
    # D+ decay Channels
    dp_to_kaon_m_pip_pip: RecoModeIDInfo = RecoModeIDInfo(id=31, name="dp_to_kaon_m_pip_pip", latex_label=LatexDecayProcesses.dp_to_kaon_m_pip_pip)  #                           D+  to  K- pi+ pi+              # noqa: E262
    dp_to_kaon_m_pip_pip_pi0: RecoModeIDInfo = RecoModeIDInfo(id=32, name="dp_to_kaon_m_pip_pip_pi0", latex_label=LatexDecayProcesses.dp_to_kaon_m_pip_pip_pi0)  #               D+  to  K- pi+ pi+ pi0          # noqa: E262
    dp_to_kaon_m_pip_pip_pip_pim: RecoModeIDInfo = RecoModeIDInfo(id=33, name="dp_to_kaon_m_pip_pip_pip_pim", latex_label=LatexDecayProcesses.dp_to_kaon_m_pip_pip_pip_pim)  #   D+  to  K- pi+ pi+ pi+ pi-      # noqa: E262
    dp_to_kaon_s0_pip: RecoModeIDInfo = RecoModeIDInfo(id=34, name="dp_to_kaon_s0_pip", latex_label=LatexDecayProcesses.dp_to_kaon_s0_pip)  #                                    D+  to  K0_short pi+            # noqa: E262
    dp_to_kaon_s0_pip_pi0: RecoModeIDInfo = RecoModeIDInfo(id=35, name="dp_to_kaon_s0_pip_pi0", latex_label=LatexDecayProcesses.dp_to_kaon_s0_pip_pi0)  #                        D+  to  K0_short pi+ pi0        # noqa: E262
    dp_to_kaon_s0_pip_pip_pim: RecoModeIDInfo = RecoModeIDInfo(id=36, name="dp_to_kaon_s0_pip_pip_pim", latex_label=LatexDecayProcesses.dp_to_kaon_s0_pip_pip_pim)  #            D+  to  K0_short pi+ pi+ pi-    # noqa: E262
    dp_to_kaon_s0_kaon_p: RecoModeIDInfo = RecoModeIDInfo(id=37, name="dp_to_kaon_s0_kaon_p", latex_label=LatexDecayProcesses.dp_to_kaon_s0_kaon_p)  #                           D+  to  K0_short K+             # noqa: E262
    dp_to_kaon_p_kaon_m_pip: RecoModeIDInfo = RecoModeIDInfo(id=38, name="dp_to_kaon_p_kaon_m_pip", latex_label=LatexDecayProcesses.dp_to_kaon_p_kaon_m_pip)  #                  D+  to  K+ K- pi+               # noqa: E262
    # D0 decay Channels
    d0_to_kaon_m_pip: RecoModeIDInfo = RecoModeIDInfo(id=41, name="d0_to_kaon_m_pip", latex_label=LatexDecayProcesses.d0_to_kaon_m_pip)  #                                       D0  to  K- pi+                  # noqa: E262
    d0_to_kaon_m_pip_pi0: RecoModeIDInfo = RecoModeIDInfo(id=42, name="d0_to_kaon_m_pip_pi0", latex_label=LatexDecayProcesses.d0_to_kaon_m_pip_pi0)  #                           D0  to  K- pi+ pi0              # noqa: E262
    d0_to_kaon_m_pip_pip_pim: RecoModeIDInfo = RecoModeIDInfo(id=43, name="d0_to_kaon_m_pip_pip_pim", latex_label=LatexDecayProcesses.d0_to_kaon_m_pip_pip_pim)  #               D0  to  K- pi+ pi+ pi-          # noqa: E262
    d0_to_kaon_m_pip_pip_pim_pi0: RecoModeIDInfo = RecoModeIDInfo(id=44, name="d0_to_kaon_m_pip_pip_pim_pi0", latex_label=LatexDecayProcesses.d0_to_kaon_m_pip_pip_pim_pi0)  #   D0  to  K- pi+ pi+ pi- pi0      # noqa: E262
    d0_to_kaon_s0_pi0: RecoModeIDInfo = RecoModeIDInfo(id=45, name="d0_to_kaon_s0_pi0", latex_label=LatexDecayProcesses.d0_to_kaon_s0_pi0)  #                                    D0  to  K0_short pi0            # noqa: E262
    d0_to_kaon_s0_pip_pim: RecoModeIDInfo = RecoModeIDInfo(id=46, name="d0_to_kaon_s0_pip_pim", latex_label=LatexDecayProcesses.d0_to_kaon_s0_pip_pim)  #                        D0  to  K0_short pi+ pi-        # noqa: E262
    d0_to_kaon_s0_pip_pim_pi0: RecoModeIDInfo = RecoModeIDInfo(id=47, name="d0_to_kaon_s0_pip_pim_pi0", latex_label=LatexDecayProcesses.d0_to_kaon_s0_pip_pim_pi0)  #            D0  to  K0_short pi+ pi- pi0    # noqa: E262
    d0_to_kaon_m_kaon_p: RecoModeIDInfo = RecoModeIDInfo(id=48, name="d0_to_kaon_m_kaon_p", latex_label=LatexDecayProcesses.d0_to_kaon_m_kaon_p)  #                              D0  to  K- K+                   # noqa: E262

    # Reco Mode Combinations:

    # B -> D(*) l  Reco Modes
    dstp_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dstp_to_d0_pip.id, dstp_to_dp_pi0.id), name="dstp_modes", latex_label=LatexDecayProcesses.dstpm_reco_modes_label)
    dst0_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dst0_to_d0_pi0.id, dst0_to_d0_gamma.id), name="dst0_modes", latex_label=LatexDecayProcesses.dst0_reco_modes_label)

    dst_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=dst0_modes.ids + dstp_modes.ids, name="dst_modes", latex_label=LatexDecayProcesses.all_dst_meson_reco_modes_label)

    dp_to_charged_kaon_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dp_to_kaon_m_pip_pip.id, dp_to_kaon_m_pip_pip_pi0.id, dp_to_kaon_m_pip_pip_pip_pim.id), name="dp_to_charged_kaon_modes", latex_label=LatexDecayProcesses.dpm_to_kaon_mp_reco_modes_label)
    dp_to_charged_kaon_pair_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dp_to_kaon_p_kaon_m_pip.id,), name="dp_to_charged_kaon_pair_modes", latex_label=LatexDecayProcesses.dp_to_kaon_p_kaon_m_pip[:-1] + r"\,\!$")
    dp_to_neutral_kaon_s_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dp_to_kaon_s0_pip.id, dp_to_kaon_s0_pip_pi0.id, dp_to_kaon_s0_pip_pip_pim.id, dp_to_kaon_s0_kaon_p.id), name="dp_to_neutral_kaon_s_modes", latex_label=LatexDecayProcesses.dpm_to_kaon_s0_reco_modes_label)

    dp_to_no_pi_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dp_to_kaon_s0_kaon_p.id,), name="dp_to_no_pi_modes", latex_label=LatexDecayProcesses.dpm_to_no_pi_reco_modes_label)
    dp_to_single_pi_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dp_to_kaon_s0_pip.id, dp_to_kaon_p_kaon_m_pip.id), name="dp_to_single_pi_modes", latex_label=LatexDecayProcesses.dpm_to_1_pi_reco_modes_label)
    dp_to_two_pi_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dp_to_kaon_m_pip_pip.id, dp_to_kaon_s0_pip_pi0.id), name="dp_to_two_pi_modes", latex_label=LatexDecayProcesses.dpm_to_2_pi_reco_modes_label)
    dp_to_multiple_pi_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(dp_to_kaon_m_pip_pip_pi0.id, dp_to_kaon_m_pip_pip_pip_pim.id, dp_to_kaon_s0_pip_pip_pim.id), name="dp_to_multiple_pi_modes", latex_label=LatexDecayProcesses.dpm_to_multiple_pi_reco_modes_label)

    d0_to_charged_kaon_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(d0_to_kaon_m_pip.id, d0_to_kaon_m_pip_pi0.id, d0_to_kaon_m_pip_pip_pim.id, d0_to_kaon_m_pip_pip_pim_pi0.id), name="d0_to_charged_kaon_modes", latex_label=LatexDecayProcesses.d0_to_kaon_pm_reco_modes_label)
    d0_to_charged_kaon_pair_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(d0_to_kaon_m_kaon_p.id,), name="d0_to_charged_kaon_pair_modes", latex_label=LatexDecayProcesses.d0_to_kaon_m_kaon_p[:-1] + r"\,\!$")
    d0_to_neutral_kaon_s_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(d0_to_kaon_s0_pi0.id, d0_to_kaon_s0_pip_pim.id, d0_to_kaon_s0_pip_pim_pi0.id), name="d0_to_neutral_kaon_s_modes", latex_label=LatexDecayProcesses.d0_to_kaon_s0_reco_modes_label)

    d0_to_no_pi_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(d0_to_kaon_m_kaon_p.id,), name="d0_to_no_pi_modes", latex_label=LatexDecayProcesses.d0_to_no_pi_reco_modes_label)
    d0_to_single_pi_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(d0_to_kaon_m_pip.id, d0_to_kaon_s0_pi0.id), name="d0_to_single_pi_modes", latex_label=LatexDecayProcesses.d0_to_1_pi_reco_modes_label)
    d0_to_two_pi_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(d0_to_kaon_m_pip_pi0.id, d0_to_kaon_s0_pip_pim.id), name="d0_to_two_pi_modes", latex_label=LatexDecayProcesses.d0_to_2_pi_reco_modes_label)
    d0_to_multiple_pi_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(d0_to_kaon_m_pip_pip_pim.id, d0_to_kaon_m_pip_pip_pim_pi0.id, d0_to_kaon_s0_pip_pim_pi0.id), name="d0_to_multiple_pi_modes", latex_label=LatexDecayProcesses.d0_to_multiple_pi_reco_modes_label)

    dp_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=dp_to_charged_kaon_modes.ids + dp_to_charged_kaon_pair_modes.ids + dp_to_neutral_kaon_s_modes.ids, name="dp_modes", latex_label=LatexDecayProcesses.dpm_reco_modes_label)
    d0_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=d0_to_charged_kaon_modes.ids + d0_to_charged_kaon_pair_modes.ids + d0_to_neutral_kaon_s_modes.ids, name="d0_modes", latex_label=LatexDecayProcesses.d0_reco_modes_label)

    d_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=dp_modes.ids + d0_modes.ids, name="d_modes", latex_label=LatexDecayProcesses.all_d_meson_reco_modes_label)

    d_wwo_st_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=d_modes.ids + dst_modes.ids, name="d_wwo_st_modes", latex_label=LatexDecayProcesses.all_d_wwost_meson_reco_modes_label)
    # fmt: on

    @classmethod
    def _get_additional_method_names(cls) -> Tuple[str, ...]:
        return (
            "d_meson_reco_mode_col_name",
            "d_wwo_st_meson_decay_name_to_id_dict",
        )

    @staticmethod
    def d_wwo_st_meson_decay_name_to_id_dict() -> Dict[str, int]:
        mapping: Dict[str, int] = copy.copy(Dstar_decay_label_to_decay_id)
        mapping.update(copy.copy(D_decay_label_to_decay_id))
        return mapping

    @classmethod
    def run_dedicated_checks(cls) -> None:
        reco_mode_id_infos: Dict[str, RecoModeIDInfo] = cls.get_reco_ids_dict()
        online_d_reco_mode_dict: Dict[str, int] = DMesonRecoModeInfos.d_wwo_st_meson_decay_name_to_id_dict()
        online_d_reco_mode_ids: Tuple[int, ...] = tuple(online_d_reco_mode_dict.values())

        assert all(v.id in online_d_reco_mode_ids for k, v in reco_mode_id_infos.items()), (
            [(k, v) for k, v in reco_mode_id_infos.items() if v.id not in online_d_reco_mode_ids],
            online_d_reco_mode_ids,
            online_d_reco_mode_dict,
        )
        assert len(online_d_reco_mode_dict) == len(reco_mode_id_infos), (
            len(online_d_reco_mode_dict),
            len(reco_mode_id_infos),
        )


class U4SRecoModeInfos(AbstractRecoModeInfosBaseClass):
    u4s_reco_mode_col_name: str = Cols.u4s_reco_mode_column_name

    _all_reco_modes_latex_label: str = LatexDecayProcesses.all_u4s_modes_label
    _all_reco_modes_latex_label_short: str = LatexDecayProcesses.all_u4s_modes_label_short

    # fmt: off
    u4s_from_charged_reco_mode: RecoModeIDInfo = RecoModeIDInfo(id=1001, name="u4s_from_charged_reco_mode", latex_label=LatexDecayProcesses.u4s_to_bp_tag_bm_sig)
    u4s_from_neutral_reco_mode: RecoModeIDInfo = RecoModeIDInfo(id=1000, name="u4s_from_neutral_reco_mode", latex_label=LatexDecayProcesses.u4s_to_b0_tag_anti_b0_sig)
    u4s_from_charged_reco_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(u4s_from_charged_reco_mode.id,), name="u4s_from_charged_reco_modes", latex_label=LatexDecayProcesses.u4s_to_bp_tag_bm_sig[:-1] + r"\,\!$")
    u4s_from_neutral_reco_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(u4s_from_neutral_reco_mode.id,), name="u4s_from_neutral_reco_modes", latex_label=LatexDecayProcesses.u4s_to_b0_tag_anti_b0_sig[:-1] + r"\,\!$")

    # Same flavour after mixing
    u4s_from_mixed_neutral_reco_mode: RecoModeIDInfo = RecoModeIDInfo(id=1002, name="u4s_from_mixed_neutral_reco_mode", latex_label=LatexDecayProcesses.u4s_to_b0_tag_b0_sig)
    u4s_from_mixed_neutral_reco_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(u4s_from_mixed_neutral_reco_mode.id,), name="u4s_from_mixed_neutral_reco_modes", latex_label=LatexDecayProcesses.u4s_to_b0_tag_b0_sig[:-1] + r"\,\!$")

    u4s_from_all_neutral_reco_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(u4s_from_neutral_reco_mode.id, u4s_from_mixed_neutral_reco_mode.id), name="u4s_from_all_neutral_reco_modes", latex_label=LatexDecayProcesses.u4s_to_neutral_bbbar)

    # Wrong Flavour Combinations
    u4s_from_charged_wt_reco_1_mode: RecoModeIDInfo = RecoModeIDInfo(id=1011, name="u4s_from_charged_wt_reco_1_mode", latex_label=LatexDecayProcesses.u4s_to_b0_tag_bm_sig)
    u4s_from_charged_wt_reco_2_mode: RecoModeIDInfo = RecoModeIDInfo(id=1021, name="u4s_from_charged_wt_reco_2_mode", latex_label=LatexDecayProcesses.u4s_to_b0_tag_bp_sig)
    u4s_from_neutral_wt_reco_1_mode: RecoModeIDInfo = RecoModeIDInfo(id=1010, name="u4s_from_neutral_wt_reco_1_mode", latex_label=LatexDecayProcesses.u4s_to_bp_tag_b0_sig)
    u4s_from_neutral_wt_reco_2_mode: RecoModeIDInfo = RecoModeIDInfo(id=1020, name="u4s_from_neutral_wt_reco_2_mode", latex_label=LatexDecayProcesses.u4s_to_bm_tag_b0_sig)

    u4s_from_charged_wt_reco_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(u4s_from_charged_wt_reco_1_mode.id, u4s_from_charged_wt_reco_2_mode.id), name="u4s_from_charged_wt_reco_modes", latex_label=LatexDecayProcesses.u4s_to_b_tag_bm_sig_with_wrong_tag)
    u4s_from_neutral_wt_reco_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=(u4s_from_neutral_wt_reco_1_mode.id, u4s_from_neutral_wt_reco_2_mode.id), name="u4s_from_neutral_wt_reco_modes", latex_label=LatexDecayProcesses.u4s_to_b_tag_0_sig_with_wrong_tag)

    # Additional Groups:
    u4s_from_all_correct_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=u4s_from_charged_reco_modes.ids + u4s_from_all_neutral_reco_modes.ids, name="u4s_from_all_correct_modes", latex_label=LatexDecayProcesses.u4s_to_all_correct_b_tag_b_sig)
    u4s_from_all_wrong_tag_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=u4s_from_charged_wt_reco_modes.ids + u4s_from_neutral_wt_reco_modes.ids, name="u4s_from_all_wrong_tag_modes", latex_label=LatexDecayProcesses.u4s_to_all_wrong_b_tag_b_sig)

    u4s_charge_sig_b_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=u4s_from_charged_reco_modes.ids + u4s_from_charged_wt_reco_modes.ids, name="u4s_charge_sig_b_modes", latex_label=LatexDecayProcesses.u4s_to_b_tag_bpm_sig)
    u4s_neutral_sig_b_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=u4s_from_all_neutral_reco_modes.ids + u4s_from_neutral_wt_reco_modes.ids, name="u4s_neutral_sig_b_modes", latex_label=LatexDecayProcesses.u4s_to_b_tag_b0_sig)

    u4s_charge_tag_b_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=u4s_from_charged_reco_modes.ids + u4s_from_neutral_wt_reco_modes.ids, name="u4s_charge_tag_b_modes", latex_label=LatexDecayProcesses.u4s_to_bpm_tag_b_sig)
    u4s_neutral_tag_b_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=u4s_from_all_neutral_reco_modes.ids + u4s_from_charged_wt_reco_modes.ids, name="u4s_neutral_tag_b_modes", latex_label=LatexDecayProcesses.u4s_to_b0_tag_b_sig)

    u4s_all_reco_modes: RecoModeGroupInfo = RecoModeGroupInfo(ids=u4s_from_all_correct_modes.ids + u4s_from_all_wrong_tag_modes.ids, name="u4s_all_reco_modes", latex_label=LatexDecayProcesses.all_u4s_modes_label)
    # fmt: on

    @classmethod
    def _get_additional_method_names(cls) -> Tuple[str, ...]:
        return (
            "u4s_reco_mode_col_name",
            #
        )


if __name__ == "__main__":

    RecoModeInfos.run_checks()
    TagRecoModeInfos.run_checks()
    DMesonRecoModeInfos.run_checks()
    U4SRecoModeInfos.run_checks()
