"""
Tools for the evaluation of ResultContainer files.

Felix Metzner 2021
"""

from typing import Dict, Any
from rdstar.branching_fractions.b_meson_bf_calculation import (
    bf_b0_to_d_l_nu_avg,
    bf_bp_to_d_l_nu_avg,
    bf_b0_to_d_st_l_nu_avg,
    bf_bp_to_d_st_l_nu_avg,
)


__all__ = [
    "sys_table_preparation",
]


def sys_table_preparation(input_dict: Dict[str, Any]) -> str:
    nom_val = input_dict["nominal_value"]

    output_order = [
        "std_dev_up",
        "std_dev_all_sys",
        ("systematic_uncertainties", "mc_stat_eff_sys"),
        ("systematic_uncertainties", "lepton_id"),
        ("systematic_uncertainties", "pion_id"),
        ("systematic_uncertainties", "pion_fake_id"),
        ("systematic_uncertainties", "kaon_id"),
        ("systematic_uncertainties", "kaon_fake_id"),
        ("systematic_uncertainties", "slow_pion"),
        ("systematic_uncertainties", "k_short"),
        ("systematic_uncertainties", "d_ff_sys"),
        ("systematic_uncertainties", "d_st_ff_sys"),
    ]

    output = [f"{nom_val:.3f}"]
    for k in output_order:
        if isinstance(k, str):
            v = input_dict[k]
            output.append(f"{100. * v / nom_val:.3f}")
        elif isinstance(k, tuple) and len(k) == 2 and all(isinstance(kk, str) for kk in k):
            v = input_dict[k[0]][k[1]]
            output.append(f"{100. * v / nom_val:.3f}")
        else:
            raise RuntimeError(f"Don't know what to do with {k}")

    return "\n".join(output)


def main():
    example_dict = {
        "rd_ratio_result": {
            "nominal_value": 0.2989661756679673,
            "std_dev_up": 0.0617844563045216,
            "std_dev_down": 0.0617844563045216,
            "std_dev_all_sys": 0.004559602055972202,
            "systematic_uncertainties": {
                "lepton_id": 0.002340480097956244,
                "pion_id": 8.005069919224926e-05,
                "pion_fake_id": 1.384137287755584e-05,
                "kaon_id": 0.00011280968202688604,
                "kaon_fake_id": 0.00010117144505401748,
                "slow_pion": 2.9647643717225183e-05,
                "k_short": 1.424663585187687e-05,
                "d_ff_sys": 0.00043708738922330773,
                "d_st_ff_sys": 4.567771903493368e-17,
                "AllSystematics": 0.004559602055972202,
                "mc_stat_eff_sys": 0.0038846409230416526,
                "all_but_mc_stat_sys": 0.002387370060936517,
            },
        },
        "rdstar_ratio_result": {
            "nominal_value": 0.25808632129876735,
            "std_dev_up": 0.031490542702024775,
            "std_dev_down": 0.031490542702024775,
            "std_dev_all_sys": 0.003115236482587995,
            "systematic_uncertainties": {
                "lepton_id": 0.001981230976916312,
                "pion_id": 8.31769595157122e-05,
                "pion_fake_id": 2.1479951772302514e-05,
                "kaon_id": 0.00010265040401309841,
                "kaon_fake_id": 3.770914476338441e-05,
                "slow_pion": 7.331854041047234e-05,
                "k_short": 5.6751104807802825e-06,
                "d_ff_sys": 3.999866326103369e-17,
                "d_st_ff_sys": 0.000400367370849174,
                "AllSystematics": 0.003115236482587995,
                "mc_stat_eff_sys": 0.002365244476095781,
                "all_but_mc_stat_sys": 0.002027391652036088,
            },
        },
    }

    for v in example_dict.values():
        print(sys_table_preparation(input_dict=v))

    print(
        f"\nBFs:\n"
        f"B0toDlnu/B0toD*lnu = {bf_b0_to_d_l_nu_avg / bf_b0_to_d_st_l_nu_avg}\n"
        f"B+toDlnu/B+toD*lnu = {bf_bp_to_d_l_nu_avg / bf_bp_to_d_st_l_nu_avg}\n"
    )


if __name__ == "__main__":
    main()
