"""
Containers for selection evaluation infos.

Felix Metzner 2021
"""

import os
import json
import copy
import pandas as pd

from typing import Union, Optional, Tuple, Sequence, Dict, Any, NamedTuple, Iterator, Mapping, Type, ClassVar

from rdstar.utilities import PathType
from rdstar.offline_analysis import rdstar_samples
from rdstar.offline_analysis.selection.selection_info_containers import BCSVariableInfo
from rdstar.offline_analysis.selection.selection_evaluation_utilities import EvalInfoUtils
from rdstar.offline_analysis.selection.selection_evaluation_information_base import EvalInfoContainerBase, EvalInfoBase
from rdstar.offline_analysis.selection.selection_evaluation_information import (
    BasicEvalInfo,
    MCEvalInfo,
    CutAndBCSEvalInfo,
    MultiplicityEvalInfo,
    MultiplicityMCEvalInfo,
)


__all__ = [
    "SelectionEvaluationContainerEntry",
    "SelectionEvaluationContainer",
]


class SelectionEvaluationContainerEntry(NamedTuple):
    level: int
    current_cut: str
    cut_list: Optional[Tuple[str, ...]]
    bcs_setup: Optional[Tuple[BCSVariableInfo, BCSVariableInfo]]
    basic_info: Union[BasicEvalInfo, EvalInfoContainerBase]
    mc_info: Optional[Union[MCEvalInfo, EvalInfoContainerBase]]
    cut_and_bcs_info: Optional[Union[CutAndBCSEvalInfo, EvalInfoContainerBase]]
    multiplicity_info: Union[MultiplicityEvalInfo, EvalInfoContainerBase]
    multiplicity_mc_info: Optional[Union[MultiplicityMCEvalInfo, EvalInfoContainerBase]]
    is_recorded_data: bool

    @staticmethod
    def get_entry_id(level: int) -> str:
        return str(level)

    @property
    def entry_id(self) -> str:
        return self.get_entry_id(level=self.level)

    def get_sub_info_by_key(self, key: str) -> Optional[EvalInfoContainerBase]:
        key_mapping: Dict[str, Optional[Union[EvalInfoContainerBase, EvalInfoBase]]] = {
            "basic_info": self.basic_info,
            "mc_info": self.mc_info,
            "cut_and_bcs_info": self.cut_and_bcs_info,
            "multiplicity_info": self.multiplicity_info,
            "multiplicity_mc_info": self.multiplicity_mc_info,
        }
        if key not in key_mapping.keys():
            raise KeyError(f"Unknown key {key} for sub info of SelectionEvaluationContainerEntry.")
        value = key_mapping[key]
        assert value is None or isinstance(value, EvalInfoContainerBase), type(value)
        return value

    def get_valid_sub_info_by_key(self, key: str) -> EvalInfoContainerBase:
        sub_info = self.get_sub_info_by_key(key=key)
        assert sub_info is not None
        return sub_info

    @classmethod
    def get_sub_info_class_by_key(cls, key: str) -> Type[EvalInfoContainerBase]:
        key_mapping: Dict[str, Union[Type[EvalInfoContainerBase], Type[EvalInfoBase]]] = {
            "basic_info": BasicEvalInfo,
            "mc_info": MCEvalInfo,
            "cut_and_bcs_info": CutAndBCSEvalInfo,
            "multiplicity_info": MultiplicityEvalInfo,
            "multiplicity_mc_info": MultiplicityMCEvalInfo,
        }
        if key not in key_mapping.keys():
            raise KeyError(f"Unknown key {key} for sub info class of SelectionEvaluationContainerEntry.")
        value = key_mapping[key]
        assert issubclass(value, EvalInfoContainerBase), value.__name__
        return value

    def as_dict(self) -> Dict[str, Any]:
        return {
            "level": self.level,
            "current_cut": self.current_cut,
            "cut_list": self.cut_list,
            "bcs_setup": self.bcs_setup,
            "basic_info": self.basic_info.as_dict(),
            "mc_info": None if self.mc_info is None else self.mc_info.as_dict(),
            "cut_and_bcs_info": None if self.cut_and_bcs_info is None else self.cut_and_bcs_info.as_dict(),
            "multiplicity_info": self.multiplicity_info.as_dict(),
            "multiplicity_mc_info": None if self.multiplicity_mc_info is None else self.multiplicity_mc_info.as_dict(),
            "is_recorded_data": self.is_recorded_data,
        }

    @staticmethod
    def create(
        level: int,
        current_cut: str,
        cut_list: Tuple[str, ...],
        bcs_setup: Optional[Tuple[BCSVariableInfo, BCSVariableInfo]],
        df: pd.DataFrame,
        data_type: str,
    ) -> "SelectionEvaluationContainerEntry":
        is_data: bool = data_type in rdstar_samples.recorded_data_samples
        if level == 0 or len(cut_list) == 0:
            assert level == len(cut_list), (level, cut_list)

        _new_entry = SelectionEvaluationContainerEntry(
            level=level,
            current_cut=current_cut,
            cut_list=None if len(cut_list) == 0 else copy.copy(cut_list),
            bcs_setup=None if bcs_setup is None else copy.copy(bcs_setup),
            basic_info=BasicEvalInfo(df=df, data_type=data_type),
            mc_info=None if is_data else MCEvalInfo(df=df, data_type=data_type),
            cut_and_bcs_info=None if is_data else CutAndBCSEvalInfo(df=df, data_type=data_type),
            multiplicity_info=MultiplicityEvalInfo(df=df, data_type=data_type),
            multiplicity_mc_info=None if is_data else MultiplicityMCEvalInfo(df=df, data_type=data_type),
            is_recorded_data=is_data,
        )
        return _new_entry

    @staticmethod
    def bcs_info_from_dict(
        input_value: Optional[Union[Sequence[Dict[str, Any]], Sequence[Sequence[Any]]]],
    ) -> Optional[Tuple[BCSVariableInfo, BCSVariableInfo]]:
        if input_value is None:
            return None
        else:
            assert isinstance(input_value, (tuple, list)), input_value
            assert len(input_value) == 2, input_value
            assert all(isinstance(v, (tuple, list, dict)) for v in input_value), input_value
            bcs_in_1 = input_value[0]
            bcs_in_2 = input_value[1]
            bcs_info_1 = BCSVariableInfo(**bcs_in_1) if isinstance(bcs_in_1, dict) else BCSVariableInfo(*bcs_in_1)
            bcs_info_2 = BCSVariableInfo(**bcs_in_2) if isinstance(bcs_in_2, dict) else BCSVariableInfo(*bcs_in_2)
            return bcs_info_1, bcs_info_2

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "SelectionEvaluationContainerEntry":
        is_recorded_data: bool = input_dict["is_recorded_data"]

        mc_di: Optional[Dict[str, Any]] = input_dict["mc_info"]
        cut_and_bcs_di: Optional[Dict[str, Any]] = input_dict["cut_and_bcs_info"]
        multiplicity_mc_di: Optional[Dict[str, Any]] = input_dict["multiplicity_mc_info"]

        if is_recorded_data:
            assert mc_di is None, mc_di
            assert cut_and_bcs_di is None, cut_and_bcs_di
            assert multiplicity_mc_di is None, multiplicity_mc_di

        _new = SelectionEvaluationContainerEntry(
            level=input_dict["level"],
            current_cut=input_dict["current_cut"],
            cut_list=None if input_dict["cut_list"] is None else tuple(input_dict["cut_list"]),
            bcs_setup=SelectionEvaluationContainerEntry.bcs_info_from_dict(input_value=input_dict["bcs_setup"]),
            basic_info=BasicEvalInfo.from_dict(input_dict=input_dict["basic_info"]),
            mc_info=None if mc_di is None else MCEvalInfo.from_dict(input_dict=mc_di),
            cut_and_bcs_info=None if cut_and_bcs_di is None else CutAndBCSEvalInfo.from_dict(input_dict=cut_and_bcs_di),
            multiplicity_info=MultiplicityEvalInfo.from_dict(input_dict=input_dict["multiplicity_info"]),
            multiplicity_mc_info=None
            if multiplicity_mc_di is None
            else MultiplicityMCEvalInfo.from_dict(input_dict=multiplicity_mc_di),
            is_recorded_data=is_recorded_data,
        )
        if _new.is_recorded_data:
            _mc_dicts = [mc_di, cut_and_bcs_di, multiplicity_mc_di]
            assert all(v is None for v in _mc_dicts), (v is None for v in _mc_dicts)
        return _new

    def merge(self, other: "SelectionEvaluationContainerEntry") -> "SelectionEvaluationContainerEntry":
        assert self.level == other.level, (self.level, other.level)
        assert self.current_cut == other.current_cut, (self.current_cut, other.current_cut)
        assert self.cut_list == other.cut_list, (self.cut_list, other.cut_list)
        assert self.bcs_setup == other.bcs_setup, (self.bcs_setup, other.bcs_setup)
        assert self.is_recorded_data == other.is_recorded_data, (self.is_recorded_data, other.is_recorded_data)

        if self.is_recorded_data:
            assert self.mc_info is None
            assert self.cut_and_bcs_info is None
            assert self.multiplicity_mc_info is None
            assert other.mc_info is None
            assert other.cut_and_bcs_info is None
            assert other.multiplicity_mc_info is None

        _mc_info: Optional[Union[MCEvalInfo, EvalInfoContainerBase]] = None
        if self.mc_info is not None:
            assert other.mc_info is not None
            _mc_info = self.mc_info.merge(other=other.mc_info)

        _cut_and_bcs_info: Optional[Union[CutAndBCSEvalInfo, EvalInfoContainerBase]] = None
        if self.cut_and_bcs_info is not None:
            assert other.cut_and_bcs_info is not None
            _cut_and_bcs_info = self.cut_and_bcs_info.merge(other=other.cut_and_bcs_info)

        _multiplicity_mc_info: Optional[Union[MultiplicityMCEvalInfo, EvalInfoContainerBase]] = None
        if self.multiplicity_mc_info is not None:
            assert other.multiplicity_mc_info is not None
            _multiplicity_mc_info = self.multiplicity_mc_info.merge(other=other.multiplicity_mc_info)

        return SelectionEvaluationContainerEntry(
            level=self.level,
            current_cut=self.current_cut,
            cut_list=self.cut_list,
            bcs_setup=self.bcs_setup,
            basic_info=self.basic_info.merge(other=other.basic_info),
            mc_info=_mc_info,
            cut_and_bcs_info=_cut_and_bcs_info,
            multiplicity_info=self.multiplicity_info.merge(other=other.multiplicity_info),
            multiplicity_mc_info=_multiplicity_mc_info,
            is_recorded_data=self.is_recorded_data,
        )


class SelectionEvaluationContainer(Mapping):
    no_cut_current_cut_str: ClassVar[str] = "no_cut"
    bcs_current_cut_str: ClassVar[str] = "bcs_application"

    def __init__(self) -> None:
        self._entries: Dict[str, SelectionEvaluationContainerEntry] = dict()
        self._level_to_entry_id_map: Dict[int, str] = dict()

    @property
    def level_to_entry_id_map(self) -> Dict[int, str]:
        assert len(self._entries.keys()) == len(self._level_to_entry_id_map.keys()), self._level_to_entry_id_map
        return self._level_to_entry_id_map

    @property
    def levels(self) -> Tuple[int, ...]:
        return tuple(self.level_to_entry_id_map.keys())

    def add(
        self,
        level: int,
        current_cut: str,
        cut_list: Tuple[str, ...],
        bcs_setup: Optional[Tuple[BCSVariableInfo, BCSVariableInfo]],
        df: pd.DataFrame,
        data_type: str,
    ) -> None:
        _new_entry_id: str = SelectionEvaluationContainerEntry.get_entry_id(level=level)
        assert _new_entry_id not in self._entries.keys(), (_new_entry_id, self._entries.keys())
        _new_entry = SelectionEvaluationContainerEntry.create(
            level=level,
            current_cut=current_cut,
            cut_list=cut_list,
            bcs_setup=bcs_setup,
            df=df,
            data_type=data_type,
        )
        self._add(new_entry=_new_entry)

    def _add(self, new_entry: SelectionEvaluationContainerEntry) -> None:
        assert isinstance(new_entry, SelectionEvaluationContainerEntry), type(new_entry)
        assert new_entry.entry_id not in self._entries.keys(), (new_entry.entry_id, self._entries.keys())

        assert new_entry.level not in self._level_to_entry_id_map.keys(), (new_entry.level, self._level_to_entry_id_map)
        self._level_to_entry_id_map.update({new_entry.level: new_entry.entry_id})

        self._entries.update({new_entry.entry_id: new_entry})

    def merge(self, other: "SelectionEvaluationContainer") -> "SelectionEvaluationContainer":
        _new = SelectionEvaluationContainer()
        _new._entries = {k: v.merge(other=other._entries[k]) for k, v in self._entries.items()}
        _new._level_to_entry_id_map = {v.level: k for k, v in _new._entries.items()}
        return _new

    def as_dict(self) -> Dict[str, Any]:
        return {k: v.as_dict() for k, v in self._entries.items()}

    def write(self, file_path: PathType) -> None:
        tmp_file_path = os.path.join(os.path.dirname(file_path), f"tmp_{os.path.basename(file_path)}")
        with open(tmp_file_path, "w") as out_file:
            json.dump(self.as_dict(), out_file, indent=2)
        os.rename(src=tmp_file_path, dst=file_path)

    @staticmethod
    def read(file_path: PathType) -> "SelectionEvaluationContainer":
        assert os.path.isfile(file_path), file_path

        with open(file_path, "r") as in_file:
            container_as_dict_from_file: Dict[str, Any] = json.load(in_file)

        container_as_dict: Dict[str, SelectionEvaluationContainerEntry] = {
            k: SelectionEvaluationContainerEntry.from_dict(input_dict=v) for k, v in container_as_dict_from_file.items()
        }

        new_container = SelectionEvaluationContainer()
        new_container._entries = container_as_dict
        new_container._level_to_entry_id_map = {v.level: k for k, v in new_container._entries.items()}
        return new_container

    @staticmethod
    def prepare_df(df: pd.DataFrame, data_type: str) -> None:
        is_data: bool = data_type in rdstar_samples.recorded_data_samples
        EvalInfoUtils.data_frame_preparation(df=df, is_data=is_data)

    @staticmethod
    def clean_df(df: pd.DataFrame, data_type: str) -> None:
        is_data: bool = data_type in rdstar_samples.recorded_data_samples
        EvalInfoUtils.post_evaluation_data_frame_cleaning(df=df, is_data=is_data)

    # Methods to evaluate filled container

    @property
    def max_selection_level(self) -> int:
        return max(self.levels)

    @property
    def full_cut_list(self) -> Tuple[str, ...]:
        full_cut_list = self._entries[self.level_to_entry_id_map[self.max_selection_level]].cut_list
        assert full_cut_list is not None
        return full_cut_list

    @property
    def info_entry_for_full_selection(self) -> SelectionEvaluationContainerEntry:
        return self._entries[self.level_to_entry_id_map[self.max_selection_level]]

    def get_entry_by_level(self, level: int) -> SelectionEvaluationContainerEntry:
        return self[self.level_to_entry_id_map[level]]

    def __getitem__(self, item: str) -> SelectionEvaluationContainerEntry:
        return self._entries[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self._entries)

    def __len__(self) -> int:
        return len(self._entries)

    def __copy__(self, memo) -> "SelectionEvaluationContainer":
        new_container = SelectionEvaluationContainer()
        new_container._entries = copy.deepcopy(self._entries)
        return new_container

    def __deepcopy__(self, memo) -> "SelectionEvaluationContainer":
        new_container = SelectionEvaluationContainer()
        new_container._entries = copy.deepcopy(self._entries)
        return new_container
