import copy
import inspect
import warnings
import numpy as np
import pandas as pd

from inspect import signature
from collections import defaultdict
from pandas.core.common import SettingWithCopyWarning
from typing import Union, Optional, Tuple, List, Sequence, Dict, Callable, Generator, Any

from rdstar.offline_analysis.tasks import utilities

from rdstar.offline_analysis.observables.common_observables import Cols

from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCSigIDs
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_utilities import MatchQueryInfo, MCSigIDNameToQueryMappingType

from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_info import GenMCMatchingInfo
from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name


__all__ = [
    "EvalInfoUtils",
    "ValueCounter",
    "ValuePerOtherCounter",
    "map_function_type_to_value_class",
    "map_function_type_to_value_interpretation",
]


class ValueTuple:
    internal_type = Tuple[Union[str, float, int], ...]

    def __init__(self, x: Sequence[Union[str, float, int]]) -> None:
        self._values: Optional[Tuple[Union[str, float, int], ...]] = None
        if isinstance(x, (list, tuple)):
            assert all(isinstance(v, (str, int, float)) for v in x), (x, [type(v) for v in x])
            assert len(x) == len(set(x)), (type(x), x, [type(v) for v in x])
            self._values = tuple(x)
        else:
            raise TypeError(f"Unexpected input type for x: {type(x)}")

    @property
    def values(self) -> Tuple[Union[int, float, str], ...]:
        assert self._values is not None
        return self._values

    @staticmethod
    def direct_merge_value_tuples(
        a: Tuple[Union[str, float, int], ...],
        b: Tuple[Union[str, float, int], ...],
    ) -> Tuple[Union[str, float, int], ...]:
        return tuple(set(list(a) + list(b)))

    @staticmethod
    def merge_value_tuples(a: "ValueTuple", b: "ValueTuple") -> "ValueTuple":
        return ValueTuple(x=ValueTuple.direct_merge_value_tuples(a=a.values, b=b.values))

    @staticmethod
    def interpret(x: Any) -> Tuple[Union[int, str], ...]:
        assert isinstance(x, (list, tuple)), (type(x), x)
        assert all(isinstance(v, (int, float, str)) for v in x), [type(v) for v in x]
        if all(isinstance(v, str) for v in x):
            try:
                return tuple([int(float(v)) for v in x])
            except ValueError:
                return tuple(x)
        else:
            return tuple([int(v) for v in x])


class ValueCounter:
    internal_type = Dict[Union[int, float, str], Union[int, float]]

    def __init__(self, x: Union[pd.Series, Dict[Union[int, float, str], Union[int, float]]]) -> None:
        self._value_counts: Optional[Dict[Union[int, float, str], Union[int, float]]] = None
        if isinstance(x, dict):
            assert all(isinstance(k, (str, int, float)) for k in x.keys()), [(k, type(k)) for k in x.keys()]
            assert all(isinstance(v, (int, float)) for v in x.values()), [(k, v, type(v)) for k, v in x.items()]
            self._value_counts = x
        elif isinstance(x, pd.Series):
            self._value_counts = self.value_counts_from_series(x=x)
        else:
            raise TypeError(f"Unexpected input type for x: {type(x)}")

    @property
    def value_counts(self) -> Dict[Union[int, float, str], Union[int, float]]:
        assert self._value_counts is not None
        return self._value_counts

    @property
    def value_counts_as_series(self) -> pd.Series:
        return pd.Series(self.value_counts)

    @staticmethod
    def value_counts_from_series(x: pd.Series) -> Dict[Union[int, float, str], Union[int, float]]:
        return {k: v for k, v in zip(x.index, x)}

    @staticmethod
    def merge_value_counters(a: "ValueCounter", b: "ValueCounter") -> "ValueCounter":
        return ValueCounter(x=a.value_counts_as_series.add(b.value_counts_as_series, fill_value=0))

    @staticmethod
    def interpret(x: Any) -> Dict[Union[int, str], int]:
        assert isinstance(x, dict), (type(x), x)
        assert all(isinstance(k, (int, float, str)) for k in x.keys()), [type(k) for k in x.keys()]
        assert all(isinstance(v, (int, float)) for v in x.values()), [type(v) for v in x.values()]

        if all(isinstance(k, str) for k in x.keys()):
            try:
                return {int(float(k)): int(v) for k, v in x.items()}
            except ValueError:
                return {k: int(v) for k, v in x.items()}
        else:
            return {int(k): int(v) for k, v in x.items()}


# Types for ValuePerOtherCounter class
ValuePerOtherCounterType = Dict[Union[int, float, str], Dict[Union[int, float, str], Union[int, float]]]
ValuePerOtherCounterInputType = Union[
    pd.Series,
    Tuple[Union[int, float, str], pd.Series],
    Tuple[Union[int, float, str], Dict[Union[int, float, str], Union[int, float]]],
    Dict[Union[int, float, str], pd.Series],
    ValuePerOtherCounterType,
]


class ValuePerOtherCounter:
    internal_type = Dict[Union[int, float, str], Dict[Union[int, float, str], Union[int, float]]]

    def __init__(self, x: ValuePerOtherCounterInputType) -> None:
        self._value_counts: Optional[ValuePerOtherCounterType] = None
        if isinstance(x, dict):
            self._value_counts = self.value_per_other_counts_from_dict(x=x)
        elif isinstance(x, tuple):
            self._value_counts = self.value_per_other_counts_from_tuple(x=x)
        elif isinstance(x, pd.Series):
            self._value_counts = self.value_per_other_counts_from_series(x=x)
        else:
            raise TypeError(f"Unexpected input type for x: {type(x)}")

    @property
    def value_counts(self) -> ValuePerOtherCounterType:
        assert self._value_counts is not None
        return self._value_counts

    @property
    def as_series(self) -> pd.Series:
        return pd.Series({(k, kk): vv for k, v in self.value_counts.items() for kk, vv in v.items()})

    @staticmethod
    def value_per_other_counts_from_series(x: pd.Series) -> ValuePerOtherCounterType:
        assert isinstance(x.index, pd.MultiIndex), type(x.index)
        assert x.index.nlevels == 2, x.index.nlevels
        return {k: x.xs(k).to_dict() for k in x.index.levels[0]}

    @staticmethod
    def value_per_other_counts_from_tuple(x: Tuple[Any, ...]) -> ValuePerOtherCounterType:
        assert isinstance(x, tuple), type(x)
        assert len(x) == 2, (len(x), x)
        assert isinstance(x[0], (str, int, float)), type(x[0])
        _x: Tuple[Union[int, float, str], Any] = (x[0], x[1])
        return ValuePerOtherCounter.value_per_other_counts_from_dict(x=dict([_x]))

    @staticmethod
    def value_per_other_counts_from_dict(x: Dict[Union[int, float, str], Any]) -> ValuePerOtherCounterType:
        assert isinstance(x, dict), type(x)
        assert all(isinstance(k, (str, int, float)) for k in x.keys()), [
            (k, type(k)) for k in x.keys() if not isinstance(k, (str, int, float))
        ]
        if all(isinstance(v, dict) for v in x.values()):
            assert all(isinstance(kk, (str, int, float)) for v in x.values() for kk in v.keys()), [
                (k, kk, type(kk)) for k, v in x.items() for kk in v.keys() if not isinstance(kk, (str, int, float))
            ]
            assert all(isinstance(vv, (int, float)) for v in x.values() for vv in v.values()), [
                (k, kk, vv, type(vv)) for k, v in x.items() for kk, vv in v.items() if not isinstance(vv, (int, float))
            ]
            return x
        elif all(isinstance(v, pd.Series) for v in x.values()):
            return ValuePerOtherCounter.value_per_other_counts_from_dict_w_series(x=x)
        else:
            raise TypeError(f"Unexpected input type for x: {type(x)}, {list(set([type(v) for v in x.values()]))}")

    @staticmethod
    def value_per_other_counts_from_dict_w_series(x: Dict[Union[int, float, str], pd.Series]) -> ValuePerOtherCounterType:
        assert isinstance(x, dict), type(x)
        assert all(isinstance(v, pd.Series) for v in x.values()), [type(v) for v in x.values()]
        return {k: v.to_dict() for k, v in x.items()}

    @staticmethod
    def merge_value_per_other_counters(a: "ValuePerOtherCounter", b: "ValuePerOtherCounter") -> "ValuePerOtherCounter":
        assert isinstance(a, ValuePerOtherCounter), type(a)
        assert isinstance(b, ValuePerOtherCounter), type(b)

        try:
            combined_counters: pd.Series = a.as_series.add(b.as_series, fill_value=0)
        except ValueError as ve:
            if len(a.as_series) == 0:
                new_values: ValuePerOtherCounterType = b.value_counts
                missing_keys: List[Union[int, str, float]] = [k for k in a.value_counts.keys() if k not in new_values]
            elif len(b.as_series) == 0:
                new_values = a.value_counts
                missing_keys = [k for k in b.value_counts.keys() if k not in new_values]
            else:
                raise ve
            assert all(k not in new_values.keys() for k in missing_keys)
            new_values.update({k: {} for k in missing_keys})
            return ValuePerOtherCounter(x=new_values)

        expected_keys: List[Union[int, str, float]] = list(set(list(a.value_counts.keys()) + list(b.value_counts.keys())))

        if len(combined_counters) == 0:
            new_values = {k: v for k, v in a.value_counts.items()}
            new_values.update({k: v for k, v in b.value_counts.items() if k not in new_values})
            assert len(expected_keys) == len(new_values), (len(expected_keys), len(new_values), expected_keys, new_values)
            return ValuePerOtherCounter(x=new_values)

        missing_keys = [k for k in expected_keys if k not in combined_counters.index.levels[0]]
        new_values = ValuePerOtherCounter.value_per_other_counts_from_series(combined_counters)
        new_values.update({k: {} for k in missing_keys})

        return ValuePerOtherCounter(x=new_values)

    @staticmethod
    def interpret(x: Any) -> Dict[Union[int, str], Dict[Union[int, str], int]]:
        assert isinstance(x, dict), (type(x), x)
        assert all(isinstance(k, (int, float, str)) for k in x.keys()), [type(k) for k in x.keys()]

        assert all(isinstance(v, dict) for v in x.values()), [type(v) for v in x.values()]
        assert all(isinstance(kk, (int, float, str)) for v in x.values() for kk in v.keys()), [
            type(kk) for v in x.values() for kk in v.keys()
        ]
        assert all(isinstance(vv, (int, float)) for v in x.values() for vv in v.values()), [
            type(vv) for v in x.values() for vv in v.values()
        ]
        _x: ValuePerOtherCounterType = x

        if all(isinstance(k, str) for k in _x.keys()):
            try:
                __x: Dict[Union[int, str], Dict[Union[int, float, str], Union[int, float]]] = {
                    int(float(k)): v for k, v in _x.items()
                }
            except ValueError:
                __x = {str(k): v for k, v in _x.items()}
        else:
            __x = {int(k): v for k, v in _x.items()}

        if all(isinstance(kk, str) for v in __x.values() for kk in v.keys()):
            try:
                _xx: Dict[Union[int, str], Dict[Union[int, str], int]] = {
                    k: {int(float(kk)): int(vv) for kk, vv in v.items()} for k, v in __x.items()
                }
            except ValueError:
                _xx = {k: {str(kk): int(vv) for kk, vv in v.items()} for k, v in __x.items()}
        else:
            _xx = {k: {int(kk): int(vv) for kk, vv in v.items()} for k, v in __x.items()}

        return _xx


def map_function_type_to_value_class(calc_function: Callable) -> type:
    calc_return_annotation = inspect.signature(calc_function).return_annotation
    signature_to_type_map = {
        int: int,
        ValueTuple.internal_type: ValueTuple,
        ValueCounter.internal_type: ValueCounter,
        ValuePerOtherCounter.internal_type: ValuePerOtherCounter,
    }
    if calc_return_annotation not in signature_to_type_map.keys():
        raise TypeError(f"Unknown return signature {calc_return_annotation} for calc function {calc_function.__name__}")
    return signature_to_type_map[calc_return_annotation]


def map_function_type_to_value_interpretation(calc_function: Callable) -> Callable:
    calc_return_annotation = inspect.signature(calc_function).return_annotation
    signature_to_interpretation_map = {
        int: lambda x: int(x),
        ValueTuple.internal_type: ValueTuple.interpret,
        ValueCounter.internal_type: ValueCounter.interpret,
        ValuePerOtherCounter.internal_type: ValuePerOtherCounter.interpret,
    }
    if calc_return_annotation not in signature_to_interpretation_map.keys():
        raise TypeError(f"Unknown return signature {calc_return_annotation} for calc function {calc_function.__name__}")
    return signature_to_interpretation_map[calc_return_annotation]


class EvalInfoUtils:
    decay_mode_id_col: str = Cols.decay_mode_column_name
    tag_decay_mode_id_col: str = Cols.tag_reco_mode_column_name
    u4s_decay_mode_id_col: str = Cols.u4s_reco_mode_column_name

    u_eid_col: str = utilities.UniqueEventIDColumnInfos.get_unique_event_id_col_name()

    sig_id_col: str = OldMCMatchingIDsQueryInfo.get_sig_col_name()
    is_sig_col: str = copy.copy(Cols.u4s_is_signal_col)
    is_b_sig_col: str = copy.copy(Cols.b_sig_is_signal_col)
    is_c_bkg_col: str = copy.copy(Cols.is_continuum_bkg_col)

    gen_b1_id_col: str = gen_b1_id_col_name
    gen_b2_id_col: str = gen_b2_id_col_name

    l_mother_pdg_col: str = copy.copy(Cols.lepton_mc_mother_pdg_col)
    mdst_index_col: str = copy.copy(Cols.mdst_index_via_lepton_col)
    mdst_index_tau_col: str = copy.copy(Cols.mdst_index_via_tau_col)

    gen_id_col: str = "tmp_gen_id"

    sig_id_queries: MCSigIDNameToQueryMappingType = {
        k: v for k, v in OldMCMatchingIDsQueryInfo.get_sig_id_mapping().items() if v.id != OldMCSigIDs.b_to_pi_l_nu_bkg.id
    }

    gen_b_ids_queries: Dict[str, MatchQueryInfo] = copy.copy(GenMCMatchingInfo.get_gen_id_mapping())

    reco_mode_to_sig_id_mapping: Dict[int, Tuple[int, ...]] = OldMCMatchingIDsQueryInfo.get_reco_mode_to_sig_id_mapping()
    reco_mode_ids: Tuple[int, ...] = copy.copy(tuple(reco_mode_to_sig_id_mapping.keys()))

    sig_id_to_reco_mode_ids_mapping: Dict[
        int, Tuple[int, ...]
    ] = OldMCMatchingIDsQueryInfo.get_sig_id_to_reco_mode_mapping()

    reco_mode_down_feed_mapping: Dict[int, Tuple[int, ...]] = {
        15: (11, 13),
        16: (12, 14),
        17: (11, 13),
        18: (12, 14),
    }

    # Meta methods:

    @staticmethod
    def evaluate_condition(cond_fcn: Callable, df: pd.DataFrame, data_type: str) -> pd.Series:
        EvalInfoUtils._check_cond_fcn(cond_fcn=cond_fcn)
        cond = cond_fcn(df=df, data_type=data_type)
        assert isinstance(cond, pd.Series), type(cond)
        return cond

    @staticmethod
    def _check_cond_fcn(cond_fcn: Callable) -> None:
        _expected_kw_args: Tuple[str, ...] = (
            "df",
            "data_type",
        )
        cond_fcn_signature = signature(cond_fcn)
        cond_fcn_parameters = cond_fcn_signature.parameters
        assert len(cond_fcn_parameters) == len(_expected_kw_args), (
            cond_fcn_parameters.keys(),
            _expected_kw_args,
            cond_fcn.__name__,
        )
        assert all(kwa in cond_fcn_parameters.keys() for kwa in _expected_kw_args), (
            cond_fcn_parameters.keys(),
            _expected_kw_args,
            cond_fcn.__name__,
        )
        assert cond_fcn_signature.return_annotation == pd.Series, (
            cond_fcn_signature.return_annotation,
            cond_fcn.__name__,
        )

    @staticmethod
    def get_datatype(df: pd.DataFrame, data_type: str) -> str:
        return data_type

    @staticmethod
    def get_candidate_count(
        df: pd.DataFrame,
        data_type: str,
        cond_fcn: Optional[Callable] = None,
    ) -> int:
        if cond_fcn is None:
            return len(df.index)
        else:
            cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
            return len(df.loc[cond].index)

    @staticmethod
    def get_event_count(
        df: pd.DataFrame,
        data_type: str,
        cond_fcn: Optional[Callable] = None,
    ) -> int:
        if cond_fcn is None:
            return len(df[EvalInfoUtils.u_eid_col].unique())
        else:
            cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
            return len(df.loc[cond, EvalInfoUtils.u_eid_col].unique())

    # Calculation methods:
    # Simple counts (return type = int)
    @staticmethod
    def candidate_count_fcn(cond_fcn: Optional[Callable] = None) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> int:
            return EvalInfoUtils.get_candidate_count(df=df, data_type=data_type, cond_fcn=cond_fcn)

        return fcn

    @staticmethod
    def event_count_fcn(cond_fcn: Optional[Callable] = None) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> int:
            return EvalInfoUtils.get_event_count(df=df, data_type=data_type, cond_fcn=cond_fcn)

        return fcn

    # Multiplicity counter: Returns frequency of number of candidates per event.
    @staticmethod
    def multiplicity_counter(cond_fcn: Optional[Callable] = None) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValueCounter.internal_type:
            if cond_fcn is None:
                query: pd.Series = df[EvalInfoUtils.u_eid_col].value_counts().value_counts()
            else:
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = df.loc[cond, EvalInfoUtils.u_eid_col].value_counts().value_counts()
            return ValueCounter(x=query).value_counts

        return fcn

    # Counts for values of column (return type from ValueCounter)
    @staticmethod
    def candidate_by_col_counter(col: str, cond_fcn: Optional[Callable] = None) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValueCounter.internal_type:
            if cond_fcn is None:
                query: pd.Series = df.loc[:, col].value_counts()
            else:
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = df.loc[cond, col].value_counts()
            return ValueCounter(x=query).value_counts

        return fcn

    @staticmethod
    def event_by_col_counter(col: str, cond_fcn: Optional[Callable] = None) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValueCounter.internal_type:
            if cond_fcn is None:
                query: pd.Series = df.groupby(col)[EvalInfoUtils.u_eid_col].nunique()
            else:
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = df.loc[cond].groupby(col)[EvalInfoUtils.u_eid_col].nunique()
            return ValueCounter(x=query).value_counts

        return fcn

    # Counts per fixed value, which is provided via key_fcn (return type from ValueCounter)
    @staticmethod
    def static_value_counter(key_fcn: Callable, counter_fcn: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValueCounter.internal_type:
            key: Union[str, int, float] = key_fcn(df=df, data_type=data_type)
            assert isinstance(key, (str, int, float)), (key, type(key))
            return ValueCounter(x={key: counter_fcn(df=df, data_type=data_type)}).value_counts

        return fcn

    # Counts number of events/candidates for each condition provided by the cond_fcn_generator function,
    # which should return the condition value and the condition function cond_fcn itself.
    # The result is of the form {cond_val: count}
    @staticmethod
    def candidates_multi_condition_counter(cond_fcn_generator: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValueCounter.internal_type:
            counter_dict: ValueCounter.internal_type = dict()
            for x_val, cond_fcn in cond_fcn_generator():
                assert x_val not in counter_dict.keys(), (x_val, counter_dict.keys())
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = len(df.loc[cond].index)
                counter_dict.update({x_val: query})
            return ValueCounter(x=counter_dict).value_counts

        return fcn

    @staticmethod
    def events_multi_condition_counter(cond_fcn_generator: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValueCounter.internal_type:
            counter_dict: ValueCounter.internal_type = dict()
            for x_val, cond_fcn in cond_fcn_generator():
                assert x_val not in counter_dict.keys(), (x_val, counter_dict.keys())
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = len(df.loc[cond, EvalInfoUtils.u_eid_col].unique())
                counter_dict.update({x_val: query})
            return ValueCounter(x=counter_dict).value_counts

        return fcn

    @staticmethod
    def multiplicity_multi_condition_counter(cond_fcn_generator: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            counter_dict: ValuePerOtherCounter.internal_type = defaultdict(dict)
            for x_val, cond_fcn in cond_fcn_generator():
                assert x_val not in counter_dict.keys(), (x_val, counter_dict.keys())
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query: pd.Series = df.loc[cond, EvalInfoUtils.u_eid_col].value_counts().value_counts()
                counter_dict.update({x_val: query})
            return ValuePerOtherCounter(x=counter_dict).value_counts

        return fcn

    # Counts for values of column x in subsets of values of column y (return type from ValuePerOtherCounter)
    @staticmethod
    def candidates_by_x_per_y_counter(col_x: str, col_y: str, cond_fcn: Optional[Callable] = None) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            if cond_fcn is None:
                query: pd.Series = df.groupby([col_y, col_x]).size()
            else:
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = df.loc[cond].groupby([col_y, col_x]).size()
            return ValuePerOtherCounter(x=query).value_counts

        return fcn

    @staticmethod
    def events_by_x_per_y_counter(col_x: str, col_y: str, cond_fcn: Optional[Callable] = None) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            if cond_fcn is None:
                query: pd.Series = df.groupby([col_y, col_x])[EvalInfoUtils.u_eid_col].nunique()
            else:
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = df.loc[cond].groupby([col_y, col_x])[EvalInfoUtils.u_eid_col].nunique()
            return ValuePerOtherCounter(x=query).value_counts

        return fcn

    # Counts values of col y per fixed value x, which is provided via x_key_fcn (return type from ValuePerOtherCounter)
    # The counter function should be a ..._by_col_counter method returning ValueCounter.
    # The arguments col_y and cond will be passed to the provided counter_fcn.
    @staticmethod
    def static_by_x_per_y_counter(
        x_key_fcn: Callable,
        counter_fcn: Callable,
        col_y: str,
        cond_fcn: Optional[Callable] = None,
    ) -> Callable:
        assert signature(counter_fcn).return_annotation == Callable, signature(counter_fcn).return_annotation

        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            x_key: Union[str, int, float] = x_key_fcn(df=df, data_type=data_type)
            assert isinstance(x_key, (str, int, float)), (x_key, type(x_key))
            query_fcn = counter_fcn(col=col_y, cond_fcn=cond_fcn)
            assert signature(query_fcn).return_annotation == ValueCounter.internal_type, signature(
                counter_fcn
            ).return_annotation
            query = query_fcn(df=df, data_type=data_type)
            return ValuePerOtherCounter(x={x_key: query}).value_counts

        return fcn

    # Counts ValueCounter objects per fixed value, which is provided via key_fcn (return type from ValuePerOtherCounter)
    @staticmethod
    def static_value_per_other_counter(key_fcn: Callable, counter_fcn: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            key: Union[str, int, float] = key_fcn(df=df, data_type=data_type)
            assert isinstance(key, (str, int, float)), (key, type(key))
            return ValuePerOtherCounter(x={key: counter_fcn(df=df, data_type=data_type)}).value_counts

        return fcn

    # Counts values of col for each condition provided by the cond_fcn_generator function,
    # which should return the condition value and the condition function cond_fcn itself.
    # The result is of the form {cond_val: {col_val: count}}
    @staticmethod
    def candidates_multi_condition_by_col_counter(col: str, cond_fcn_generator: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            counter_dict: ValuePerOtherCounter.internal_type = dict()
            for x_val, cond_fcn in cond_fcn_generator():
                assert x_val not in counter_dict.keys(), (x_val, counter_dict.keys())
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = df.loc[cond, col].value_counts()
                counter_dict.update({x_val: query})
            return ValuePerOtherCounter(x=counter_dict).value_counts

        return fcn

    @staticmethod
    def events_multi_condition_by_col_counter(col: str, cond_fcn_generator: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            counter_dict: ValuePerOtherCounter.internal_type = dict()
            for x_val, cond_fcn in cond_fcn_generator():
                assert x_val not in counter_dict.keys(), (x_val, counter_dict.keys())
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = df.loc[cond].groupby(col)[EvalInfoUtils.u_eid_col].nunique()
                counter_dict.update({x_val: query})
            return ValuePerOtherCounter(x=counter_dict).value_counts

        return fcn

    # Counts values per static value for each condition provided by the cond_fcn_generator function,
    # which should return the condition value and the condition function cond_fcn itself.
    # The static value is provided via y_key_fcn.
    # The result is of the form {cond_val: {static_val: count}}
    @staticmethod
    def candidates_multi_condition_by_static_counter(y_key_fcn: Callable, cond_fcn_generator: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            y_key: Union[str, int, float] = y_key_fcn(df=df, data_type=data_type)
            assert isinstance(y_key, (str, int, float)), (y_key, type(y_key))
            counter_dict: ValuePerOtherCounter.internal_type = dict()
            for x_val, cond_fcn in cond_fcn_generator():
                assert x_val not in counter_dict.keys(), (x_val, counter_dict.keys())
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = len(df.loc[cond].index)
                counter_dict.update({x_val: {y_key: query}})
            return ValuePerOtherCounter(x=counter_dict).value_counts

        return fcn

    @staticmethod
    def events_multi_condition_by_static_counter(y_key_fcn: Callable, cond_fcn_generator: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValuePerOtherCounter.internal_type:
            y_key: Union[str, int, float] = y_key_fcn(df=df, data_type=data_type)
            assert isinstance(y_key, (str, int, float)), (y_key, type(y_key))
            counter_dict: ValuePerOtherCounter.internal_type = dict()
            for x_val, cond_fcn in cond_fcn_generator():
                assert x_val not in counter_dict.keys(), (x_val, counter_dict.keys())
                cond = EvalInfoUtils.evaluate_condition(cond_fcn=cond_fcn, df=df, data_type=data_type)
                query = len(df.loc[cond, EvalInfoUtils.u_eid_col].unique())
                counter_dict.update({x_val: {y_key: query}})
            return ValuePerOtherCounter(x=counter_dict).value_counts

        return fcn

    # Tuple of values occurring in the data frame (return type from ValueTuple)
    @staticmethod
    def encountered_values_tuple(values_fcn: Callable) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> ValueTuple.internal_type:
            values: Sequence[Union[str, int, float]] = values_fcn(df=df, data_type=data_type)
            if isinstance(values, (list, tuple)):
                assert all(isinstance(v, (str, int, float)) for v in values), (
                    values,
                    [type(v) for v in values if not isinstance(v, (str, int, float))],
                )
                return ValueTuple(x=copy.copy(values)).values
            elif isinstance(values, (str, int, float)):
                return ValueTuple(x=tuple([values])).values
            else:
                raise TypeError(
                    f"Expected values of type (list, tuple, str, int, float)," f"but got {type(values)}\n{values}"
                )

        return fcn

    # Merger methods:

    @staticmethod
    def int_add(a: int, b: int) -> int:
        return a + b

    @staticmethod
    def value_counter_add(a: ValueCounter.internal_type, b: ValueCounter.internal_type) -> ValueCounter.internal_type:
        return ValueCounter.merge_value_counters(a=ValueCounter(a), b=ValueCounter(b)).value_counts

    @staticmethod
    def value_per_other_counter_add(a: ValuePerOtherCounterType, b: ValuePerOtherCounterType) -> ValuePerOtherCounterType:
        return ValuePerOtherCounter.merge_value_per_other_counters(
            a=ValuePerOtherCounter(a),
            b=ValuePerOtherCounter(b),
        ).value_counts

    @staticmethod
    def value_tuple_add(a: ValueTuple.internal_type, b: ValueTuple.internal_type) -> ValueTuple.internal_type:
        return ValueTuple.direct_merge_value_tuples(a=a, b=b)

    # Condition Functions
    @staticmethod
    def create_cond(col: str, val: Union[str, int, float], other_cond_fcn: Optional[Callable]) -> Callable:
        def fcn(df: pd.DataFrame, data_type: str) -> pd.Series:
            this_cond: pd.Series = df[col] == val

            if other_cond_fcn is None:
                return this_cond
            else:
                other_cond: pd.Series = EvalInfoUtils.evaluate_condition(
                    cond_fcn=other_cond_fcn,
                    df=df,
                    data_type=data_type,
                )
            return this_cond & other_cond

        return fcn

    @staticmethod
    def u4s_match() -> Callable:
        return EvalInfoUtils.create_cond(col=EvalInfoUtils.is_sig_col, val=1, other_cond_fcn=None)

    @staticmethod
    def no_u4s_match() -> Callable:
        return EvalInfoUtils.create_cond(col=EvalInfoUtils.is_sig_col, val=0, other_cond_fcn=None)

    @staticmethod
    def b_sig_match() -> Callable:
        return EvalInfoUtils.create_cond(col=EvalInfoUtils.is_b_sig_col, val=1, other_cond_fcn=None)

    @staticmethod
    def no_b_sig_match() -> Callable:
        return EvalInfoUtils.create_cond(col=EvalInfoUtils.is_b_sig_col, val=0, other_cond_fcn=None)

    @staticmethod
    def is_continuum() -> Callable:
        return EvalInfoUtils.create_cond(col=EvalInfoUtils.is_c_bkg_col, val=1, other_cond_fcn=None)

    @staticmethod
    def inconclusive_gen_id() -> Callable:
        return EvalInfoUtils.create_cond(col=EvalInfoUtils.gen_id_col, val=-999, other_cond_fcn=None)

    @staticmethod
    def generate_condition_functions(
        cond_col: str,
        query_dict: Dict[str, MatchQueryInfo],
        use_name: bool,
        other_cond_fcn: Optional[Callable],
    ) -> Callable:
        def fcn() -> Generator[Tuple[Union[str, int], Callable], None, None]:
            for k, match_query_info in query_dict.items():
                if use_name:
                    yield k, EvalInfoUtils.create_cond(
                        col=cond_col,
                        val=match_query_info.id,
                        other_cond_fcn=other_cond_fcn,
                    )
                else:
                    yield match_query_info.id, EvalInfoUtils.create_cond(
                        col=cond_col,
                        val=match_query_info.id,
                        other_cond_fcn=other_cond_fcn,
                    )

        return fcn

    @staticmethod
    def sig_id_cond_generator(use_name: bool) -> Callable:
        return EvalInfoUtils.generate_condition_functions(
            cond_col=EvalInfoUtils.sig_id_col,
            query_dict=EvalInfoUtils.sig_id_queries,
            use_name=use_name,
            other_cond_fcn=None,
        )

    @staticmethod
    def gen_id_cond_generator(use_name: bool) -> Callable:
        return EvalInfoUtils.generate_condition_functions(
            cond_col=EvalInfoUtils.gen_id_col,
            query_dict=EvalInfoUtils.gen_b_ids_queries,
            use_name=use_name,
            other_cond_fcn=None,
        )

    @staticmethod
    def create_condition_functions(
        cond_col: str,
        cond_values: Sequence[Union[str, int, float]],
        other_cond_fcn: Optional[Callable],
    ) -> Callable:
        def fcn() -> Generator[Tuple[Union[str, int, float], Callable], None, None]:
            for cond_value in cond_values:
                yield cond_value, EvalInfoUtils.create_cond(col=cond_col, val=cond_value, other_cond_fcn=other_cond_fcn)

        return fcn

    @staticmethod
    def decay_mode_id_cond_generator(other_cond_fcn: Optional[Callable]) -> Callable:
        return EvalInfoUtils.create_condition_functions(
            cond_col=EvalInfoUtils.decay_mode_id_col,
            cond_values=EvalInfoUtils.reco_mode_ids,
            other_cond_fcn=other_cond_fcn,
        )

    @staticmethod
    def _get_down_feed_sig_id_values(sig_ids: Union[int, Sequence[int]]) -> Tuple[int, ...]:
        sig_ids_with_down_feed: Tuple[int, ...] = (1, 3)
        assert isinstance(sig_ids, (int, tuple, list)), (type(sig_ids), sig_ids)
        if isinstance(sig_ids, (tuple, list)):
            assert all(isinstance(v, int) for v in sig_ids), [(v, type(v)) for v in sig_ids]

        _sig_ids: Sequence[int] = [sig_ids] if isinstance(sig_ids, int) else sig_ids
        current_down_feed_ids: List[int] = [s_id + 1 for s_id in _sig_ids if s_id in sig_ids_with_down_feed]
        return tuple(set(current_down_feed_ids + list(_sig_ids)))

    @staticmethod
    def decay_mode_id_with_sig_match_cond_generator(
        version: str,
        allow_down_feed: bool,
        other_cond_fcn: Optional[Callable],
    ) -> Callable:
        version_options: Tuple[str, ...] = ("sig_plus_norm", "sig", "norm", "bkg")
        assert version in version_options, (version, version_options)

        def cond_fcn() -> Generator[Tuple[Union[str, int, float], Callable], None, None]:
            for decay_mode_val, sig_id_values in EvalInfoUtils.reco_mode_to_sig_id_mapping.items():
                assert 0 < len(sig_id_values) <= 2, (sig_id_values, len(sig_id_values))

                def fcn(df: pd.DataFrame, data_type: str) -> pd.Series:
                    decay_mode_cond: pd.Series = df[EvalInfoUtils.decay_mode_id_col] == decay_mode_val
                    if version == "sig_plus_norm":
                        _sig_and_norm_id_values: Tuple[int, ...] = sig_id_values
                        if allow_down_feed:
                            _sig_and_norm_id_values = EvalInfoUtils._get_down_feed_sig_id_values(sig_ids=sig_id_values)
                        sig_cond: pd.Series = df[EvalInfoUtils.sig_id_col].isin(_sig_and_norm_id_values)
                    elif version == "bkg":
                        _not_bkg_id_values: Tuple[int, ...] = sig_id_values
                        if allow_down_feed:
                            _not_bkg_id_values = EvalInfoUtils._get_down_feed_sig_id_values(sig_ids=sig_id_values)
                        sig_cond = ~df[EvalInfoUtils.sig_id_col].isin(_not_bkg_id_values)
                    elif version == "sig":
                        if len(sig_id_values) == 2:
                            _only_sig_ids: Tuple[int, ...] = (sig_id_values[0],)
                            if allow_down_feed:
                                _only_sig_ids = EvalInfoUtils._get_down_feed_sig_id_values(sig_ids=sig_id_values[0])
                            sig_cond = df[EvalInfoUtils.sig_id_col].isin(_only_sig_ids)
                        else:
                            _sig_ids: Tuple[int, ...] = sig_id_values
                            if allow_down_feed:
                                _sig_ids = EvalInfoUtils._get_down_feed_sig_id_values(sig_ids=sig_id_values)
                            sig_cond = df[EvalInfoUtils.sig_id_col].isin(_sig_ids)
                    elif version == "norm":
                        if len(sig_id_values) == 2:
                            _norm_sig_ids: Tuple[int, ...] = (sig_id_values[1],)
                            if allow_down_feed:
                                _norm_sig_ids = EvalInfoUtils._get_down_feed_sig_id_values(sig_ids=sig_id_values[1])
                            sig_cond = df[EvalInfoUtils.sig_id_col].isin(_norm_sig_ids)
                        else:
                            sig_cond = pd.Series(data=False, index=df.index)
                    else:
                        raise RuntimeError(f"Invalid version {version}! Must be one of {version_options}!")

                    if other_cond_fcn is None:
                        return decay_mode_cond & sig_cond
                    else:
                        other_cond: pd.Series = EvalInfoUtils.evaluate_condition(
                            cond_fcn=other_cond_fcn,
                            df=df,
                            data_type=data_type,
                        )
                        if version == "bkg":
                            return decay_mode_cond & (sig_cond | ~other_cond)
                        else:
                            return decay_mode_cond & sig_cond & other_cond

                yield decay_mode_val, fcn

        return cond_fcn

    @staticmethod
    def sig_id_with_matching_decay_mode_id_condition_generator(
        allow_down_feed: bool,
        other_cond_fcn: Optional[Callable],
    ) -> Callable:
        def cond_fcn() -> Generator[Tuple[Union[str, int, float], Callable], None, None]:
            for sig_id_value, decay_mode_ids in EvalInfoUtils.sig_id_to_reco_mode_ids_mapping.items():

                def fcn(df: pd.DataFrame, data_type: str) -> pd.Series:
                    sig_id_cond: pd.Series = df[EvalInfoUtils.sig_id_col] == sig_id_value

                    if not allow_down_feed:
                        _decay_mode_ids: Tuple[int, ...] = decay_mode_ids
                    else:
                        _decay_mode_ids = tuple(
                            list(decay_mode_ids)
                            + [
                                dm_id
                                for v in decay_mode_ids
                                for dm_id in EvalInfoUtils.reco_mode_down_feed_mapping.get(v, [])
                            ]
                        )

                    decay_mode_cond: pd.Series = df[EvalInfoUtils.decay_mode_id_col].isin(_decay_mode_ids)

                    if other_cond_fcn is None:
                        return sig_id_cond & decay_mode_cond
                    else:
                        other_cond: pd.Series = EvalInfoUtils.evaluate_condition(
                            cond_fcn=other_cond_fcn,
                            df=df,
                            data_type=data_type,
                        )
                        return sig_id_cond & decay_mode_cond & other_cond

                yield sig_id_value, fcn

        return cond_fcn

    # Data frame preparation

    @staticmethod
    def data_frame_preparation(df: pd.DataFrame, is_data: bool) -> None:
        if not is_data:
            EvalInfoUtils.check_data_frame(df=df)
            EvalInfoUtils.add_gen_id_col(df=df)

    @staticmethod
    def check_data_frame(df: pd.DataFrame) -> None:
        is_continuum: pd.Series = df[EvalInfoUtils.is_c_bkg_col] == 1
        c_check = df.loc[is_continuum, EvalInfoUtils.gen_b1_id_col] == df.loc[is_continuum, EvalInfoUtils.gen_b2_id_col]
        assert isinstance(c_check, pd.Series)
        assert c_check.all()

    @staticmethod
    def add_gen_id_col(df: pd.DataFrame) -> None:
        if EvalInfoUtils.gen_id_col in df.columns:
            return

        is_continuum: pd.Series = df[EvalInfoUtils.is_c_bkg_col] == 1

        conditions: Tuple[Union[pd.Series, bool], ...] = (
            (df[EvalInfoUtils.mdst_index_col] == 1),
            (df[EvalInfoUtils.mdst_index_col] == 2),
            ((df[EvalInfoUtils.mdst_index_tau_col] == 1) & (abs(df[EvalInfoUtils.l_mother_pdg_col]) == 15)),
            ((df[EvalInfoUtils.mdst_index_tau_col] == 2) & (abs(df[EvalInfoUtils.l_mother_pdg_col]) == 15)),
            (is_continuum & (df[EvalInfoUtils.gen_b1_id_col] == df[EvalInfoUtils.gen_b2_id_col])),
        )

        assert all(pd.DataFrame({c: d for c, d in enumerate(conditions)}).sum(axis=1) <= 1), "Conditions are not unique"

        # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
        with warnings.catch_warnings(record=True) as warning:
            warnings.simplefilter("always")
            # Actual command:
            df.loc[:, EvalInfoUtils.gen_id_col] = np.select(
                condlist=conditions,
                choicelist=[
                    df[EvalInfoUtils.gen_b1_id_col],
                    df[EvalInfoUtils.gen_b2_id_col],
                    df[EvalInfoUtils.gen_b1_id_col],
                    df[EvalInfoUtils.gen_b2_id_col],
                    df[EvalInfoUtils.gen_b1_id_col],
                ],
                default=-999,
            )
            for w in warning:
                assert issubclass(w.category, SettingWithCopyWarning)
                assert "A value is trying to be set on a copy" in str(w.message)

    # Data frame post processing
    @staticmethod
    def post_evaluation_data_frame_cleaning(df: pd.DataFrame, is_data: bool) -> None:
        if not is_data:
            del df[EvalInfoUtils.gen_id_col]
