"""
Class collecting and evaluating information on Signal Selection

Felix Metzner 2020
"""

import pandas as pd

from typing import List, Tuple

from rdstar.offline_analysis.selection.selection_info_containers import BCSVariableInfo
from rdstar.offline_analysis.selection.selection_evaluation_info_container import SelectionEvaluationContainer


__all__ = [
    "SelectionEvaluator",
]


class SelectionEvaluator:
    def __init__(self, df: pd.DataFrame, data_type: str) -> None:
        self._data_type: str = data_type
        self._applied_cuts_list: List[str] = []

        self._is_finalized: bool = False
        self._evaluation_info: SelectionEvaluationContainer = SelectionEvaluationContainer()

        SelectionEvaluationContainer.prepare_df(df=df, data_type=data_type)

        self._evaluation_info.add(
            level=0,
            current_cut=SelectionEvaluationContainer.no_cut_current_cut_str,
            cut_list=tuple(),
            bcs_setup=None,
            df=df,
            data_type=self.data_type,
        )

    @property
    def data_type(self) -> str:
        return self._data_type

    @property
    def evaluation_info(self) -> SelectionEvaluationContainer:
        if not self._is_finalized:
            raise RuntimeError("Selection evaluation must be finalized first!")
        return self._evaluation_info

    def update_after_cut(
        self,
        df: pd.DataFrame,
        cut: str,
    ) -> None:
        assert not self._is_finalized
        assert isinstance(cut, str), (type(cut), cut)

        invalid_cut_strs: Tuple[str, ...] = (
            SelectionEvaluationContainer.no_cut_current_cut_str,
            SelectionEvaluationContainer.bcs_current_cut_str,
        )
        assert cut not in invalid_cut_strs, cut
        assert cut not in self._applied_cuts_list, (cut, self._applied_cuts_list)

        self._applied_cuts_list.append(cut)

        level: int = len(self._applied_cuts_list)
        self._evaluation_info.add(
            level=level,
            current_cut=cut,
            cut_list=tuple(self._applied_cuts_list),
            bcs_setup=None,
            df=df,
            data_type=self.data_type,
        )

    def update_after_best_candidate_selection(
        self,
        df: pd.DataFrame,
        bcs: Tuple[BCSVariableInfo, BCSVariableInfo],
    ) -> None:
        assert not self._is_finalized

        self._applied_cuts_list.append(SelectionEvaluationContainer.bcs_current_cut_str)
        level: int = len(self._applied_cuts_list)

        self._evaluation_info.add(
            level=level,
            current_cut=SelectionEvaluationContainer.bcs_current_cut_str,
            cut_list=tuple(self._applied_cuts_list),
            bcs_setup=bcs,
            df=df,
            data_type=self.data_type,
        )

    def finalize_selection_evaluation(
        self,
        df: pd.DataFrame,
    ) -> None:
        assert not self._is_finalized
        SelectionEvaluationContainer.clean_df(df=df, data_type=self.data_type)
        self._is_finalized = True
