"""
Fitting Tasks

These tasks will
    - perform the actual fit of the analysis (TODO)
    - perform toy studies (TODO)
    - perform profile scans (TODO)

Felix Metzner 2020
"""

import os
import shutil
import pandas as pd
import b2luigi as luigi

from collections import abc
from luigi.util import inherits
from typing import Tuple, List, Dict, Union, Generator, Any, Optional

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.logistics_tasks import DataCopyTask
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit import MCShiftInfo
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.tasks.efficiency_validation_task import EfficiencyValidationTask
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups, LooseMVASelection
from rdstar.offline_analysis.validation.efficiency_validation import EfficiencyDependencyEvaluator
from rdstar.offline_analysis.tasks.resolution_evaluation_tasks import ResolutionEvaluationManagerTask

from rdstar.offline_analysis.fitting.fit_case import FitCase
from rdstar.offline_analysis.fitting.mva_cut_info import MVACutInfo
from rdstar.offline_analysis.fitting.fit_model_creator import FitModelCreator
from rdstar.offline_analysis.fitting.rdstar_fit_models import rdstar_fit_cases
from rdstar.offline_analysis.fitting.fit_optimization import FitOptimizationValidator
from rdstar.offline_analysis.fitting.fit_routine import FitRoutine, FitResultContainer


__all__ = [
    "FitTask",
    "FittingTasksWrapperTask",
]


logger = log_to_default(name=__name__)


@inherits(ResolutionEvaluationManagerTask)
class FitInputCacheTask(utilities.BaseOfflineTask):
    mc_sample_keys = luigi.TupleParameter(positional=False, hashed=True)  # type: Tuple[str, ...]
    sample_selection_tag = luigi.Parameter(positional=False)  # type: str
    task_host_name = luigi.Parameter(positional=False)  # type: str
    production_tag = luigi.Parameter(positional=False)  # type: str

    mc_output_key = "mc_sample"  # type: str
    data_output_key = "data_sample"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if not os.path.isdir(self.machine_cache_dir):
            raise RuntimeError(
                f"Cache directory {self.machine_cache_dir} must be created manually "
                f"on host {self.task_host_name}, first!"
            )

        self.data_sample_key = "DataY4S"  # type: str
        keys_list = "_".join(self.mc_sample_keys)  # type: str

        self.mc_cache_dir_path = os.path.join(self.cache_dir, keys_list)  # type: PathType
        self.data_cache_dir_path = os.path.join(self.cache_dir, self.data_sample_key)  # type: PathType

    @property
    def machine_cache_dir(self) -> PathType:
        return utilities.cache_dir_mapping[self.task_host_name]

    @property
    def cache_dir(self) -> PathType:
        return os.path.join(self.machine_cache_dir, self.production_tag, self.sample_selection_tag)

    def load_mc_sample(self, input_dict: Dict[str, PathType]) -> pd.DataFrame:
        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in input_dict.items()
            if (k == f"{mc_key}_{self.sample_selection_tag}" and os.path.splitext(p)[1] == ".h5")
        ]  # type: List[PathType]
        assert len(mc_paths) == len(self.mc_sample_keys), (
            len(mc_paths),
            len(self.mc_sample_keys),
            mc_paths,
            self.mc_sample_keys,
            self.sample_selection_tag,
        )

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        return mc_sample

    def get_recorded_data_sample_path(self, input_dict: Dict[str, PathType]) -> PathType:
        key = f"{self.data_sample_key}_{self.sample_selection_tag}"  # type: str
        data_paths = [
            p for k, p in input_dict.items() if k == key and os.path.splitext(p)[1] == ".h5"
        ]  # type: List[PathType]
        assert len(data_paths) == 1, (len(data_paths), data_paths, key)
        data_path = data_paths[0]  # type: PathType
        assert os.path.exists(data_path), data_path
        return data_path

    def create_ssd_cache(
        self,
        dry_run: bool = False,
    ) -> Generator[Union[Dict[str, PathType], Dict[str, utilities.TaskType]], None, None]:
        mc_path = os.path.join(self.mc_cache_dir_path, f"{self.mc_output_key}.h5")  # type: PathType
        data_path = os.path.join(self.data_cache_dir_path, f"{self.data_output_key}.h5")  # type: PathType

        if not dry_run:
            if not (os.path.isfile(mc_path) and os.path.isfile(data_path)):
                run_time_requirements = self.run_time_requires()  # type: Dict[str, utilities.TaskType]
                runtime_input_dict = {
                    k: v.path
                    for t in run_time_requirements.values()
                    if isinstance(t.output(), dict)
                    for k, v in t.output().items()
                }  # type: Dict[str, PathType]

                assert all(isinstance(t.output(), dict) for t in run_time_requirements.values())
                incomplete_runtime_requirements = {
                    k: t
                    for k, t in run_time_requirements.items()
                    if not all(os.path.exists(p.path) for p in t.output().values())
                }  # type: Dict[str, utilities.TaskType]

                if incomplete_runtime_requirements:
                    for k, v in incomplete_runtime_requirements.items():
                        yield {k: v}

                else:
                    if not os.path.isfile(mc_path):
                        tmp_mc_path = os.path.join(
                            self.mc_cache_dir_path, f"tmp_{self.mc_output_key}.h5"
                        )  # type: PathType
                        mc_sample = self.load_mc_sample(input_dict=runtime_input_dict)  # type: pd.DataFrame
                        mc_sample.to_hdf(tmp_mc_path, key=self.mc_output_key)
                        os.rename(tmp_mc_path, mc_path)

                    if not os.path.isfile(data_path):
                        tmp_data_path = os.path.join(
                            self.data_cache_dir_path, f"tmp_{self.data_output_key}.h5"
                        )  # type: PathType
                        input_data_path = self.get_recorded_data_sample_path(
                            input_dict=runtime_input_dict
                        )  # type: PathType
                        shutil.copy(src=input_data_path, dst=tmp_data_path, follow_symlinks=True)
                        os.rename(tmp_data_path, data_path)

        yield {
            self.mc_output_key: mc_path,
            self.data_output_key: data_path,
        }

    def run(self) -> Generator[utilities.TaskType, None, None]:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.mc_cache_dir_path, exist_ok=True)
        os.makedirs(self.data_cache_dir_path, exist_ok=True)
        runtime_requirements = [
            (kk, vv)
            for t in self.create_ssd_cache()
            if isinstance(t, dict) and all(isinstance(v, utilities.TaskType) for v in t.values())
            for kk, vv in t.items()
        ]  # type: List[Tuple[str, utilities.TaskType]]

        if runtime_requirements:
            required_combo_names = [t[0] for t in runtime_requirements]  # type: List[str]
            yield self.clone(FitInputCacheRuntimeRequirementWrapperTask, required_combo_names=required_combo_names)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        _output_dict_list = list(self.create_ssd_cache(dry_run=True))
        assert len(_output_dict_list) == 1, _output_dict_list
        _output_dict = _output_dict_list[0]
        assert isinstance(_output_dict, dict), (type(_output_dict).__name__, _output_dict)
        output_dict = _output_dict  # type: Dict[str, PathType]
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def run_time_requires(self) -> Dict[str, utilities.TaskType]:
        task_dict = dict()  # type: Dict[str, utilities.TaskType]
        assert isinstance(self.config_combinations, abc.Sequence)

        for combo_name, data_types, relevant_configs in self.config_combinations:
            assert combo_name not in task_dict.keys(), (combo_name, task_dict.keys())
            task_dict.update(
                {
                    combo_name: ResolutionCorrectionTask(
                        combo_name=combo_name,
                        working_dir=self.working_dir,
                        selection_name=self.selection_name,
                        cs_tag=self.cs_tag,
                        config_params=relevant_configs,
                        config_combinations=self.config_combinations,
                    )
                }
            )

        return task_dict


@inherits(ResolutionEvaluationManagerTask)
class FitInputCacheRuntimeRequirementWrapperTask(luigi.WrapperTask):
    required_combo_names = luigi.ListParameter(positional=False, hashed=True)  # type: List[str]

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for combo_name, data_types, relevant_configs in self.config_combinations:
            if combo_name in self.required_combo_names:
                yield ResolutionCorrectionTask(
                    combo_name=combo_name,
                    working_dir=self.working_dir,
                    selection_name=self.selection_name,
                    cs_tag=self.cs_tag,
                    config_params=relevant_configs,
                    config_combinations=self.config_combinations,
                )


@inherits(ResolutionEvaluationManagerTask)
class FitTask(utilities.BaseOfflineTask):
    fit_case_key = luigi.Parameter(positional=False)  # type: str
    mva_cut_info_dict = luigi.DictParameter(positional=False, default={})  # type: Dict[str, Any]

    fit_base_dir_name = "FitResults"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.fit_case_name = self.fit_case_key  # type: str
        self.fit_case = FitModelCreator.fit_cases[self.fit_case_key]  # type: FitCase

        self.base_dir_name = self.fit_case.base_dir_name  # type: str

        self.mva_cut_info = MVACutInfo.from_dict(input_dict=self.mva_cut_info_dict)  # type: MVACutInfo
        if self.mva_cut_info.applies_cuts:
            self.base_dir_name = f"{self.fit_case.base_dir_name}_with_mva_cut_{self.mva_cut_info.get_cut_tag()}"

        input_sample_key = FitModelCreator.fit_cases[self.fit_case_key].sample_key  # type: str
        if input_sample_key != "":
            input_sample_key = f"_{input_sample_key}"

        if OptimizingByFitWrapperTask.is_relevant(fit_case=self.fit_case, selection_name=self.selection_name):
            _super_dir_name = f"{self.fit_case.base_dir_name}_mva_cut_probes"  # type: str
            _base_path = os.path.join(
                self.working_dir, self.fit_base_dir_name, self.main_sel_tag, _super_dir_name, self.base_dir_name
            )  # type: PathType
        else:
            _base_path = os.path.join(self.working_dir, self.fit_base_dir_name, self.main_sel_tag, self.base_dir_name)

        self.base_path = _base_path  # type: PathType

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.key = f"_{self.selection_name}_with_CS_{self.cs_tag}{input_sample_key}"  # type: str
        self.input_appendix = (
            f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}"
        )  # type: str

        self.sample_description_tag = self.input_appendix + self.key  # type: str

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
            if not self.fit_case.add_reco_mode_efficiency_correction:
                assert len(self._input_dict.keys()) == 2, self._input_dict.keys()
            else:
                assert len(self._input_dict.keys()) >= 2, self._input_dict.keys()
            assert FitInputCacheTask.mc_output_key in self._input_dict.keys()
            assert FitInputCacheTask.data_output_key in self._input_dict.keys()
        return self._input_dict

    @property
    def production_tag(self) -> str:
        tag = self.config_combinations[0][2][0]["reduced_production_tag"]
        assert isinstance(tag, str), (tag, type(tag))
        return tag

    def load_mc_sample(self) -> pd.DataFrame:
        mc_sample_path = self.input_dict[FitInputCacheTask.mc_output_key]  # type: PathType
        mc_sample = pd.read_hdf(mc_sample_path)
        assert isinstance(mc_sample, pd.DataFrame), type(mc_sample).__name__

        if self.fit_case.add_new_fei_calibration_factors:
            WeightManager.switch_fei_calibration(
                df=mc_sample,
                new_fei_calib_col=WeightManager.used_fei_calib_col,
                main_weight_col_name=self.fit_case.weight_column_name,
            )

        if self.fit_case.add_reco_mode_efficiency_correction:
            eff_reco_corr_input_key = EfficiencyDependencyEvaluator.get_eff_reco_correction_output_key(
                input_dict_keys=tuple(self.input_dict.keys())
            )  # type: str
            eff_reco_corr_container_path = self.input_dict[eff_reco_corr_input_key]  # type: PathType
            WeightManager.add_reco_mode_dependent_efficiency_correction(
                df=mc_sample,
                weight_column_name=self.fit_case.weight_column_name,
                reco_efficiency_correction_container_path=eff_reco_corr_container_path,
            )

        mm2_shifter = self.fit_case.get_additional_setting_with_default(
            setting_key=MCShiftInfo.extra_info_key(),
            default_setting=None,
        )  # type: Optional[MCShiftInfo]
        if mm2_shifter is not None:
            fit_var_cols = [v.variable.df_label for vs in self.fit_case.fit_variables_by_mode_dict.values() for v in vs]
            assert mm2_shifter.shift_var_col_name in fit_var_cols, (mm2_shifter.shift_var_col_name, fit_var_cols)
            mm2_shifter.apply_shifts(df=mc_sample)

        return mc_sample

    def load_recorded_data_sample(self) -> pd.DataFrame:
        data_sample_path = self.input_dict[FitInputCacheTask.data_output_key]  # type: PathType
        data_sample = pd.read_hdf(data_sample_path)
        assert isinstance(data_sample, pd.DataFrame), type(data_sample).__name__
        return data_sample

    def do_fit(self, dry_run: bool = False) -> Dict[str, PathType]:
        fit_routine = FitRoutine(
            fit_case=self.fit_case,
            base_output_dir_path=self.base_path,
            file_dict_dir_path=DataCopyTask.get_local_file_dict_path_from_main_dir(main_work_dir_path=self.working_dir),
            mva_cut_info=self.mva_cut_info if self.mva_cut_info.applies_cuts else None,
        )

        if dry_run:
            return fit_routine.get_fit_routine_output()
        else:
            fit_result_container = fit_routine.run_fit_routine(
                mc_sample=self.load_mc_sample(),
                data_sample=None if self.fit_case.asimov_data else self.load_recorded_data_sample(),
            )  # type: FitResultContainer
            return fit_result_container.file_output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_fit()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_fit(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        yield self.clone(
            FitInputCacheTask,
            mc_sample_keys=self.fit_case.mc_sample_keys,
            sample_selection_tag=self.sample_description_tag,
            task_host_name=utilities.get_host_name(),
            production_tag=self.production_tag,
        )

        if self.fit_case.add_reco_mode_efficiency_correction:
            sample_setup_key = SampleSetupHandler.identify_sample_setup_key(
                mc_sample_keys=self.fit_case.mc_sample_keys,
                data_sample_key="DataY4S",
            )
            yield self.clone(
                EfficiencyValidationTask,
                efficiency_evaluation_setup_key=EfficiencyValidationTask.correction_providing_setup_name,
                sample_setup_key=sample_setup_key,
                side_band_key="q2_sideband",  # We always correct from q2 sideband for fits...
                use_inclusive_fei_calibration=self.fit_case.add_new_fei_calibration_factors,
            )


@inherits(ResolutionEvaluationManagerTask)
class OptimizingByFitWrapperTask(utilities.BaseOfflineTask):
    fit_case_key = luigi.Parameter(positional=False)  # type: str

    mva_optimizing_selections = (LooseMVASelection.selection_name,)  # type: Tuple[str, ...]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.fit_case_name = self.fit_case_key  # type: str
        self.fit_case = FitModelCreator.fit_cases[self.fit_case_key]  # type: FitCase

        assert self.is_relevant(fit_case=self.fit_case, selection_name=self.selection_name), (
            self.fit_case_name,
            self.selection_name,
        )
        _base_mva_cut_info = self.fit_case.get_additional_setting(setting_key=MVACutInfo.probing_fit_case_setting_key)
        assert isinstance(_base_mva_cut_info, MVACutInfo), type(_base_mva_cut_info).__name__
        self.base_mva_cut_info = _base_mva_cut_info  # type: MVACutInfo

        self.base_dir_name = f"{self.fit_case.base_dir_name}_mva_optimization_result"  # type: str
        self.base_path = os.path.join(
            self.working_dir, FitTask.fit_base_dir_name, self.main_sel_tag, self.base_dir_name
        )  # type: PathType

        assert isinstance(self.input(), dict), (type(self.input()).__name__, self.input())
        assert all(isinstance(k, str) for k in self.input().keys()), [(k, type(k)) for k in self.input().keys()]
        assert all(isinstance(v, dict) for v in self.input().values()), [(v, type(v)) for v in self.input().values()]

        self._input_dict = None  # type: Optional[Dict[str, Dict[str, PathType]]]

    @property
    def input_dict(self) -> Dict[str, Dict[str, PathType]]:
        if self._input_dict is None:
            self._input_dict = {kk: {k: v.path for k, v in d.items()} for kk, d in self.input().items()}
        return self._input_dict

    @staticmethod
    def is_relevant(fit_case: FitCase, selection_name: str) -> bool:
        if selection_name not in OptimizingByFitWrapperTask.mva_optimizing_selections:
            return False
        if fit_case.has_additional_setting(setting_key=MVACutInfo.probing_fit_case_setting_key):
            assert not fit_case.create_plots
            assert not fit_case.create_projection_plots
            assert not fit_case.create_template_plots
            assert not fit_case.create_2d_template_plots
            assert not fit_case.run_significance_evaluation
            return True
        else:
            return False

    def get_probe_mva_cut_infos(self) -> Dict[str, MVACutInfo]:
        return self.base_mva_cut_info.get_probe_combinations_dict()

    def run_mva_optimization_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        fit_optimization_validator = FitOptimizationValidator(
            fit_case=self.fit_case,
            base_mva_cut_info=self.base_mva_cut_info,
            base_dir_path=self.base_path,
        )

        if not dry_run:
            fit_optimization_validator.validate(input_dict=self.input_dict)

        return fit_optimization_validator.get_output_dict()

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.run_mva_optimization_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.run_mva_optimization_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Dict[str, utilities.TaskType]:
        requires_dict = dict()  # type: Dict[str, utilities.TaskType]
        for p_signature, probe_mva_cut_info in self.get_probe_mva_cut_infos().items():
            requires_dict.update({p_signature: self.clone(FitTask, mva_cut_info_dict=probe_mva_cut_info.as_dict())})

        return requires_dict


@inherits(ResolutionEvaluationManagerTask)
class FittingTasksWrapperTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    @staticmethod
    def fit_case_with_mva_cut_not_relevant_for(fit_case: FitCase, selection_name: str) -> bool:
        if fit_case.has_additional_setting(setting_key=MVACutInfo.fit_case_setting_key):
            if selection_name not in OptimizingByFitWrapperTask.mva_optimizing_selections:
                return True
        return False

    def complete(self):
        # Catching special case of a FittingTasksWrapperTask for selection with which no output can be produced.
        if len(list(self.requires())) == 0:
            logger.warning(f"FittingTasksWrapperTask cannot produce output for selection name {self.selection_name}!")
            return True
        return super().complete()

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for fit_case_key, fit_case in rdstar_fit_cases.items():
            selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
            sample_tag = f"_{fit_case.sample_key}" if fit_case.sample_key else fit_case.sample_key
            required_tag = f"{self.selection_name}_with_CS_{self.cs_tag}{sample_tag}"  # type: str

            if not any(st == required_tag for st in selection_tags):
                continue

            if self.fit_case_with_mva_cut_not_relevant_for(fit_case=fit_case, selection_name=self.selection_name):
                continue

            yield self.clone(FitTask, fit_case_key=fit_case_key)

            if OptimizingByFitWrapperTask.is_relevant(fit_case=fit_case, selection_name=self.selection_name):
                yield self.clone(OptimizingByFitWrapperTask, fit_case_key=fit_case_key)
