"""
Tasks performing the Flow Analysis of the Analysis

Felix Metzner 2019
"""

import os
import copy
import multiprocessing
import b2luigi as luigi

from collections import abc
from functools import partial
from luigi.util import requires
from matplotlib import pyplot as plt
from typing import Dict, Tuple, List, Callable, Optional

from templatefitter.plotter.plot_utilities import export

from analysistools.plotting.latex_tables import df_to_split_latex_table

from rdstar.utilities import PathType
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.selection.selection_evaluation_info_container import SelectionEvaluationContainer
from rdstar.offline_analysis.validation import analysis_flow_validation, confusion_validation, multiplicity_validation

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.selection_tasks import DataSelectionTask
from rdstar.offline_analysis.tasks.combiner_tasks import SelectionInfoCombinerTask


__all__ = [
    "AnalysisEvaluationTask",
]


@requires(SelectionInfoCombinerTask)
class AnalysisEvaluationTask(utilities.BaseOfflineTask):
    cs_tag = luigi.Parameter(positional=False)  # type: str
    create_table_pdfs = luigi.BoolParameter(positional=False, default=False)  # type: bool

    base_dir_name = "AnalysisEvaluation"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(self.working_dir, self.base_dir_name, self.selection_name)  # type: PathType

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for k, v in self.input().items()}
            assert all(k.startswith(DataSelectionTask.evaluation_info_prefix) for k in self._input_dict)
            assert all(
                os.path.basename(p).startswith(DataSelectionTask.evaluation_info_prefix)
                for p in self._input_dict.values()
            )
            assert all(os.path.splitext(p)[1] == ".json" for p in self._input_dict.values())
        return self._input_dict

    def add_table_fraction_output_info(
        self,
        output_dict: Dict[str, PathType],
        name_prefix: str,
        table_fractions: List[int],
        subdir_name: str,
        dry_run: bool,
    ) -> None:
        dir_path = os.path.join(self.base_path, subdir_name)  # type: PathType

        if not dry_run:
            os.makedirs(dir_path, exist_ok=True)

        analysis_flow_validation.update_table_fraction_output_dict(
            output_dict=output_dict,
            name_prefix=name_prefix,
            base_path=dir_path,
            create_pdfs=self.create_table_pdfs,
            table_fractions=table_fractions,
        )

    def create_confusion_matrices(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]
        target_dir_name = os.path.join(self.base_path, "ConfusionMatrices")  # type: PathType

        if not dry_run:
            os.makedirs(target_dir_name, exist_ok=True)

        confusion_matrix_infos = copy.copy(confusion_validation.confusion_matrix_info_collection)

        for conf_mat_info in confusion_matrix_infos:
            assert isinstance(conf_mat_info, confusion_validation.ConfusionMatrixInfo)
            file_name = f"{selection_tag}_cm_{conf_mat_info.name}_confusion_matrix"
            if not dry_run:
                assert info_container is not None
                matrix = confusion_validation.get_confusion_matrix(
                    info_container=info_container,
                    confusion_matrix_info=conf_mat_info,
                )
                fig, _ = confusion_validation.create_confusion_matrix_plot(
                    confusion_matrix=matrix,
                    confusion_matrix_info=conf_mat_info,
                )
                export(fig=fig, filename=file_name, target_dir=target_dir_name)
                plt.close(fig=fig)
            assert file_name not in output_dict
            output_dict.update({file_name: os.path.join(target_dir_name, f"{file_name}.pdf")})

        return output_dict

    def do_simple_cut_flow_analysis(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = analysis_flow_validation.create_simple_cut_flow_table(
            info_container=info_container,
            selection_tag=selection_tag,
            selection_name=self.selection_name,
            base_path=self.base_path,
            create_pdfs=self.create_table_pdfs,
            dry_run=dry_run,
        )  # type: Dict[str, PathType]
        return output_dict

    def do_cut_flow_analysis(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        out_d = {}  # type: Dict[str, PathType]
        for compact_option in [True, False]:
            _od = analysis_flow_validation.create_cut_flow_table(
                info_container=info_container,
                selection_tag=selection_tag,
                selection_name=self.selection_name,
                base_path=self.base_path,
                compact_layout=compact_option,
                create_pdfs=self.create_table_pdfs,
                dry_run=dry_run,
            )  # type: Dict[str, PathType]
            assert all(k not in out_d.keys() for k in _od.keys()), [k for k in _od.keys() if k in out_d.keys()]
            assert all(v not in out_d.values() for v in _od.values()), [v for v in _od.values() if v in out_d.values()]
            out_d.update(_od)
        return out_d

    def do_cut_flow_analysis_for_data(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        out_d = {}  # type: Dict[str, PathType]
        for compact_option in [True, False]:
            _od = analysis_flow_validation.create_cut_flow_table_for_data(
                info_container=info_container,
                selection_tag=selection_tag,
                selection_name=self.selection_name,
                base_path=self.base_path,
                compact_layout=compact_option,
                create_pdfs=self.create_table_pdfs,
                dry_run=dry_run,
            )  # type: Dict[str, PathType]
            assert all(k not in out_d.keys() for k in _od.keys()), [k for k in _od.keys() if k in out_d.keys()]
            assert all(v not in out_d.values() for v in _od.values()), [v for v in _od.values() if v in out_d.values()]
            out_d.update(_od)
        return out_d

    def _do_event_multiplicity_evaluation(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        for_data: bool,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        if for_data:
            multiplicity_infos = copy.copy(multiplicity_validation.data_multiplicity_info_collection)
        else:
            multiplicity_infos = copy.copy(multiplicity_validation.multiplicity_info_collection)

        for multiplicity_info in multiplicity_infos:
            _new_output = multiplicity_validation.evaluate_multiplicity_info(
                info_container=info_container,
                multiplicity_info=multiplicity_info,
                selection_name=self.selection_name,
                selection_tag=selection_tag,
                base_target_dir_path=self.base_path,
                for_data=for_data,
                create_table_pdfs=self.create_table_pdfs,
                dry_run=dry_run,
            )
            assert all(k not in output_dict for k in _new_output), (_new_output, output_dict)
            output_dict.update(_new_output)

        return output_dict

    def do_event_multiplicity_evaluation(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        return self._do_event_multiplicity_evaluation(
            info_container=info_container,
            selection_tag=selection_tag,
            for_data=False,
            dry_run=dry_run,
        )

    def do_event_multiplicity_evaluation_for_data(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        return self._do_event_multiplicity_evaluation(
            info_container=info_container,
            selection_tag=selection_tag,
            for_data=True,
            dry_run=dry_run,
        )

    def do_bcs_validation(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        simple_bcs_table_output_dict = analysis_flow_validation.create_simple_bcs_table(
            info_container=info_container,
            selection_tag=selection_tag,
            base_path=self.base_path,
            create_pdfs=self.create_table_pdfs,
            dry_run=dry_run,
        )
        output_dict.update(simple_bcs_table_output_dict)

        bcs_infos = analysis_flow_validation.bcs_info_items  # type: Tuple[analysis_flow_validation.BCSInfoItem, ...]
        table_fractions = [2, 3, 4, 6]  # type: List[int]

        for bcs_info in bcs_infos:
            name_prefix = f"{selection_tag}_bcst_{bcs_info.name}_bcs"
            self.add_table_fraction_output_info(
                output_dict=output_dict,
                name_prefix=name_prefix,
                table_fractions=table_fractions,
                subdir_name="BCSValidation",
                dry_run=dry_run,
            )

        if not dry_run:
            assert info_container is not None
            bcs_df_dict = analysis_flow_validation.create_bcs_table(
                info_container=info_container,
                bcs_infos=bcs_infos,
            )

            row_formatters = {}  # type: Dict[str, Callable]
            for row_key in ["Sig", "Bkg", "Norm"]:
                row_formatters.update(
                    {
                        f"{row_key} before": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                        f"{row_key} after": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                        f"{row_key} Eff": lambda s: r"\multicolumn{1}{c}{---}"
                        if s == "---"
                        else f"{s * 100:0.2f}" + r"$\,\si{\percent}$",
                    }
                )

            for bcs_info_name, bcs_df in bcs_df_dict.items():
                name_prefix = f"{selection_tag}_bcst_{bcs_info_name}_bcs"

                df_to_split_latex_table(
                    df=bcs_df,
                    table_fractions=table_fractions,
                    name_prefix=name_prefix,
                    output_lookup_dict=output_dict,
                    placeholder_element="---",
                    remove_column_name_index=True,
                    create_pdf=self.create_table_pdfs,
                    row_formatters=row_formatters,
                    allow_smaller_sub_dicts=True,
                )

        return output_dict

    def do_bcs_validation_for_data(
        self,
        info_container: Optional[SelectionEvaluationContainer],
        selection_tag: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        table_fractions = [2, 5]  # type: List[int]

        name_prefix = f"{selection_tag}_bcst_data_bcs"
        self.add_table_fraction_output_info(
            output_dict=output_dict,
            name_prefix=name_prefix,
            table_fractions=table_fractions,
            subdir_name="BCSValidationOnData",
            dry_run=dry_run,
        )

        if not dry_run:
            assert info_container is not None
            bcs_df = analysis_flow_validation.create_bcs_table_for_data(info_container=info_container)

            row_formatters = {
                "Before": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                "After": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s:.0f}",
                "Eff": lambda s: r"\multicolumn{1}{c}{---}" if s == "---" else f"{s * 100:0.2f}" + r"$\,\si{\percent}$",
            }

            df_to_split_latex_table(
                df=bcs_df,
                table_fractions=table_fractions,
                name_prefix=name_prefix,
                output_lookup_dict=output_dict,
                placeholder_element="---",
                remove_column_name_index=True,
                create_pdf=self.create_table_pdfs,
                row_formatters=row_formatters,
                allow_smaller_sub_dicts=True,
            )

        return output_dict

    @staticmethod
    def do_validation_sub_task(
        sub_task_function: Callable,
        path: PathType,
        selection_tag: str,
        output_dict: Dict[str, PathType],
        dry_run: bool,
    ) -> Dict[str, PathType]:
        info_container = None  # type: Optional[SelectionEvaluationContainer]
        if not dry_run:
            info_container = SelectionEvaluationContainer.read(file_path=path)

        outputs = sub_task_function(
            info_container=info_container,
            selection_tag=selection_tag,
            dry_run=dry_run,
        )
        assert all(k not in output_dict for k in outputs)
        return outputs

    def do_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        mc_validation_sub_task_functions = [
            self.create_confusion_matrices,
            self.do_simple_cut_flow_analysis,
            self.do_cut_flow_analysis,
            self.do_event_multiplicity_evaluation,
            self.do_bcs_validation,
        ]  # type: List[Callable]

        data_validation_sub_task_functions = [
            self.do_cut_flow_analysis_for_data,
            self.do_event_multiplicity_evaluation_for_data,
            self.do_bcs_validation_for_data,
        ]  # type: List[Callable]

        output_dict = {}  # type: Dict[str, PathType]
        for key, path in self.input_dict.items():
            combo_key_and_sel_tag = key.split(DataSelectionTask.evaluation_info_prefix + "_")[1]
            combo_key, selection_tag = combo_key_and_sel_tag.split("_", 1)
            assert combo_key in SelectionInfoCombinerTask.info_names, combo_key
            selection_tags = SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag)
            assert selection_tag in selection_tags, (selection_tag, selection_tags)

            if combo_key == SelectionInfoCombinerTask.main_info_name:
                validation_sub_task_functions = mc_validation_sub_task_functions
            elif combo_key == SelectionInfoCombinerTask.data_info_name:
                validation_sub_task_functions = data_validation_sub_task_functions
            else:
                continue

            with multiprocessing.Pool(processes=1 if dry_run else max(len(validation_sub_task_functions), 1)) as pool:
                eval_outputs = pool.map(
                    partial(
                        AnalysisEvaluationTask.do_validation_sub_task,
                        path=path,
                        selection_tag=selection_tag,
                        output_dict=output_dict,
                        dry_run=dry_run,
                    ),
                    validation_sub_task_functions,
                )

            assert all(k not in output_dict for sub_dict in eval_outputs for k in sub_dict)
            assert all(v not in output_dict.values() for sub_dict in eval_outputs for v in sub_dict.values())
            for sub_dict in eval_outputs:
                output_dict.update(sub_dict)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}
