"""
Tasks for the Training of a MVA Classifier for the Selection

Patrick Ecker and Felix Metzner 2020
"""

import os
import numpy as np
import pandas as pd
import b2luigi as luigi

from PyFastBDT import FastBDT
from luigi.util import requires
from sklearn.model_selection import train_test_split
from typing import List, Tuple, Dict, Generator, Optional, Any

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.mva_selection.mva_selector_setups import MVASelectorSetups
from rdstar.mva_selection.mva_selector import MVASelector, MVASelectorValidation

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.selection.selection_info_containers import MVAInfo
from rdstar.offline_analysis.tasks.data_reduction_tasks import DataReductionTask


__all__ = [
    "MVASelectorTrainingDataPreparationTask",
    "MVASelectorTrainTask",
    "MVASelectorTestTask",
    "MVASelectionValidatorTask",
]


logger = log_to_default(name=__name__)


class MVASelectorTrainingDataPreparationTask(utilities.BaseOfflineTask):
    mva_selector_setup_name = luigi.Parameter(positional=False)  # type: str
    working_dir = luigi.Parameter(positional=False)  # type: str
    main_offline_config_path = luigi.Parameter(positional=False)  # type: str
    cs_tag = luigi.Parameter(positional=False)  # type: str
    cs_fastbdt_settings = luigi.DictParameter(positional=False, hashed=True, significant=False)

    base_dir_name = "MVASelectorsTrainings"  # type: str

    train_hdf_key = "selector_train_data"  # type: str
    test_hdf_key = "selector_test_data"  # type: str

    mva_sample_random_seed = 14747  # type: int

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.setup_name = str(self.mva_selector_setup_name)  # type: str
        mva_selector_dir_name = f"MVASelector_{self.setup_name}"
        self.base_path = os.path.join(self.working_dir, self.base_dir_name, mva_selector_dir_name)  # type: PathType

        self.base_n_tuple_type = "Y4S_ntuple"  # type: str

        self.mva_selector_setup = MVASelectorSetups.get_selection_mva_setup(setup_name=self.setup_name)

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @staticmethod
    def get_output_key(setup_name: str, train_or_test: str) -> str:
        valid_train_or_test_values = ("train", "test")
        if not (isinstance(train_or_test, str) and train_or_test in valid_train_or_test_values):
            raise ValueError(
                f"Parameter 'train_or_test' must be one of the strings {valid_train_or_test_values} "
                f"to be valid, but you provided {train_or_test} of type {type(train_or_test)}!"
            )
        key = f"mva_selector_{setup_name}_{train_or_test}_data"
        return key

    def get_key_pairs(self) -> Tuple[Tuple[str, str], ...]:
        key_pairs = []  # type: List[Tuple[str, str]]
        for data_type, stream in self.mva_selector_setup.get_training_data_info_tuples():
            stream_info = f"_{stream}" if stream != -999 else ""
            hdf5_key = f"{data_type}{stream_info}_{self.base_n_tuple_type}"
            file_key = f"{hdf5_key}_with_CS_{self.cs_tag}"
            key_pairs.append((file_key, hdf5_key))

        return tuple(key_pairs)

    def get_data_frames(self) -> Generator[Tuple[pd.DataFrame, str], None, None]:
        required_columns = self.mva_selector_setup.get_required_columns()
        for file_key, hdf5_key in self.get_key_pairs():
            logger.info(f"Loading data frame with key {file_key} for MVA training '{self.setup_name}'.")
            df = pd.read_hdf(path_or_buf=self.input_dict[file_key], key=hdf5_key)
            assert isinstance(df, pd.DataFrame), type(df)
            yield df.loc[:, required_columns], file_key

    def get_test_and_train_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        sig_dfs = []
        bkg_dfs = []

        random_state = np.random.RandomState(seed=self.mva_sample_random_seed)

        number_of_dfs = len(self.get_key_pairs())
        sample_size_per_df = int(np.ceil(self.mva_selector_setup.get_training_sample_size() / number_of_dfs))

        for df, df_file_key in self.get_data_frames():
            self.mva_selector_setup.data_preparation(df=df, for_application_only=False)
            target_var_col = self.mva_selector_setup.get_target_column_name()
            sig_cond = df[target_var_col] == 1
            bkg_cond = df[target_var_col] == 0

            sig_sample_size = sample_size_per_df if sig_cond.sum() >= sample_size_per_df else sig_cond.sum()
            bkg_sample_size = sample_size_per_df if bkg_cond.sum() >= sample_size_per_df else bkg_cond.sum()

            if sig_sample_size > 0:
                sig_ids = df.loc[sig_cond].index.to_series().sample(n=sig_sample_size, random_state=random_state).index
                sig_dfs.append(df.loc[sig_ids.values].copy(deep=True))
                logger.info(
                    f"Using {sig_sample_size} events from data frame with key {df_file_key} of signal "
                    f"sample of for MVA training '{self.setup_name}'."
                )
            else:
                logger.warning(
                    f"Not using data frame with key {df_file_key} of signal sample of for MVA training "
                    f"'{self.setup_name}' due to unavailability events."
                )

            if bkg_sample_size > 0:
                bkg_ids = df.loc[bkg_cond].index.to_series().sample(n=bkg_sample_size, random_state=random_state).index
                bkg_dfs.append(df.loc[bkg_ids.values].copy(deep=True))
                logger.info(
                    f"Using {bkg_sample_size} events from data frame with key {df_file_key} of background "
                    f"sample of for MVA training '{self.setup_name}'."
                )
            else:
                logger.warning(
                    f"Not using data frame with key {df_file_key} of background sample of for MVA training "
                    f"'{self.setup_name}' due to unavailability events."
                )

        sig_df = pd.concat(sig_dfs, ignore_index=True, sort=False)
        del sig_dfs
        sig_test_df, sig_train_df = train_test_split(sig_df, train_size=0.5, random_state=random_state)
        del sig_df

        bkg_df = pd.concat(bkg_dfs, ignore_index=True, sort=False)
        del bkg_dfs
        bkg_test_df, bkg_train_df = train_test_split(bkg_df, train_size=0.5, random_state=random_state)
        del bkg_df

        train_df = pd.concat([bkg_train_df, sig_train_df], ignore_index=True, sort=False)
        del sig_train_df
        del bkg_train_df
        test_df = pd.concat([bkg_test_df, sig_test_df], ignore_index=True, sort=False)

        return train_df, test_df

    def mva_selector_data_preparation_process(self, dry_run: bool = False) -> Dict[str, PathType]:
        train_data_output_key = self.get_output_key(setup_name=self.setup_name, train_or_test="train")
        train_data_path = os.path.join(self.base_path, f"{train_data_output_key}.h5")
        test_data_output_key = self.get_output_key(setup_name=self.setup_name, train_or_test="test")
        test_data_path = os.path.join(self.base_path, f"{test_data_output_key}.h5")

        if not dry_run:
            os.makedirs(self.base_path, exist_ok=True)

            train_sample, test_sample = self.get_test_and_train_sample()
            train_sample.to_hdf(path_or_buf=self.get_temp_file_path(path=train_data_path), key=self.train_hdf_key)
            test_sample.to_hdf(path_or_buf=self.get_temp_file_path(path=test_data_path), key=self.test_hdf_key)

            os.rename(src=self.get_temp_file_path(path=train_data_path), dst=train_data_path)
            os.rename(src=self.get_temp_file_path(path=test_data_path), dst=test_data_path)

        return {
            train_data_output_key: train_data_path,
            test_data_output_key: test_data_path,
        }

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.mva_selector_data_preparation_process()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_key, target_path in self.mva_selector_data_preparation_process(dry_run=True).items():
            output_dict.update({output_key: luigi.LocalTarget(target_path)})
        return output_dict

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for data_type, stream in self.mva_selector_setup.get_training_data_info_tuples():
            yield self.clone(
                DataReductionTask,
                data_type=data_type,
                stream=stream,
                tuple_type=self.base_n_tuple_type,
                offline_config_path=self.main_offline_config_path,
            )


@requires(MVASelectorTrainingDataPreparationTask)
class MVASelectorTrainTask(utilities.BaseOfflineTask):
    base_dir_name = MVASelectorTrainingDataPreparationTask.base_dir_name  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.setup_name = str(self.mva_selector_setup_name)  # type: str

        mva_selector_dir_name = f"MVASelector_{self.setup_name}"
        self.base_path = os.path.join(self.working_dir, self.base_dir_name, mva_selector_dir_name)  # type: PathType

        self.mva_selector_setup = MVASelectorSetups.get_selection_mva_setup(setup_name=self.setup_name)

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for k, v in self.input().items()}
        return self._input_dict

    @staticmethod
    def get_weight_output_key(setup_name: str) -> str:
        return f"mva_selector_{setup_name}_classifier_weights"

    @staticmethod
    def get_variables_output_key(setup_name: str) -> str:
        return f"mva_selector_{setup_name}_classifier_input_variables"

    def get_sample(self) -> pd.DataFrame:
        key = MVASelectorTrainingDataPreparationTask.get_output_key(
            setup_name=self.setup_name,
            train_or_test="train",
        )  # type: str
        assert key in self.input_dict, (key, self.input_dict)
        train_data_path = self.input_dict[key]
        df = pd.read_hdf(path_or_buf=train_data_path, key=MVASelectorTrainingDataPreparationTask.train_hdf_key)
        assert isinstance(df, pd.DataFrame), type(df).__name__
        return df

    def run_mva_selection_training(self, dry_run: bool = False) -> Dict[str, PathType]:
        weight_output_key = self.get_weight_output_key(setup_name=self.setup_name)
        weight_output_path = os.path.join(self.base_path, f"{weight_output_key}.xml")

        variables_output_key = self.get_variables_output_key(setup_name=self.setup_name)
        variables_output_path = os.path.join(self.base_path, f"{variables_output_key}.pkl")

        if not dry_run:
            train_df = self.get_sample()

            target_col = self.mva_selector_setup.get_target_column_name()

            assert (np.sort(train_df[target_col].unique()) == np.array([0, 1])).all(), (
                f"Training data contains values other than 0 and 1 in target column {target_col}:\n"
                f"{train_df[target_col].unique()}"
            )

            n_sig = len(train_df[train_df[target_col] == 1].index)
            n_bkg = len(train_df[train_df[target_col] == 0].index)

            weight_col_name = "__mva_selector_sample_weights__"
            assert weight_col_name not in train_df.columns, weight_col_name

            if n_sig == n_bkg:
                train_df.loc[:, weight_col_name] = 1.0
            else:
                train_df.loc[:, weight_col_name] = np.nan
                train_df.loc[train_df[target_col] == 1, weight_col_name] = 2.0 * n_bkg / (n_sig + n_bkg)
                train_df.loc[train_df[target_col] == 0, weight_col_name] = 2.0 * n_sig / (n_sig + n_bkg)

            assert not pd.isnull(train_df[weight_col_name]).any(), train_df[weight_col_name].unique()

            training_variables = list(self.mva_selector_setup.get_selection_variable_dict().keys())

            mva_selector = MVASelector(
                selection_mva_variables=training_variables,
                target_variable=target_col,
                weight=weight_col_name,
            )

            fast_bdt_settings = self.mva_selector_setup.get_fast_bdt_settings()

            mva_selector.add_classifier(classifier=FastBDT.Classifier(**fast_bdt_settings))
            mva_selector.train(data=train_df)

            mva_selector.save_classifier_weights(
                weight_file_path=self.get_temp_file_path(path=weight_output_path),
                overwrite=True,
            )

            mva_selector.dump_list_of_variables(
                variable_list=training_variables,
                pickle_file_path=self.get_temp_file_path(path=variables_output_path),
            )

            os.rename(src=self.get_temp_file_path(path=weight_output_path), dst=weight_output_path)
            os.rename(src=self.get_temp_file_path(path=variables_output_path), dst=variables_output_path)

        return {
            weight_output_key: weight_output_path,
            variables_output_key: variables_output_path,
        }

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_mva_selection_training()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {
            output_key: luigi.LocalTarget(target_path)
            for output_key, target_path in self.run_mva_selection_training(dry_run=True).items()
        }


@requires(
    MVASelectorTrainTask,
    MVASelectorTrainingDataPreparationTask,
)
class MVASelectorTestTask(utilities.BaseOfflineTask):
    mva_selector_working_point = luigi.FloatParameter(positional=False)  # type: float
    use_alternative_working_points = luigi.BoolParameter(positional=False, default=False)  # type: bool
    alternative_base_dir_name = luigi.OptionalParameter(positional=False, default="")  # type: Optional[str]

    base_dir_name = MVASelectorTrainingDataPreparationTask.base_dir_name  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.setup_name = str(self.mva_selector_setup_name)  # type: str

        mva_selector_dir_name = f"MVASelector_{self.setup_name}"

        _base_dir_name = self.base_dir_name  # type: str
        if self.alternative_base_dir_name is not None:
            _base_dir_name = self.alternative_base_dir_name

        self.base_path = os.path.join(self.working_dir, _base_dir_name, mva_selector_dir_name)  # type: PathType

        self.mva_selector_setup = MVASelectorSetups.get_selection_mva_setup(setup_name=self.setup_name)

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_train_and_test_data_paths(self) -> Tuple[PathType, PathType]:
        train_key = MVASelectorTrainingDataPreparationTask.get_output_key(
            setup_name=self.setup_name,
            train_or_test="train",
        )
        test_key = MVASelectorTrainingDataPreparationTask.get_output_key(
            setup_name=self.setup_name,
            train_or_test="test",
        )
        return self.input_dict[train_key], self.input_dict[test_key]

    def run_mva_selection_evaluation(self, dry_run: bool = False) -> Dict[str, PathType]:
        train_path, test_path = self.get_train_and_test_data_paths()

        var_file_path = self.input_dict[MVASelectorTrainTask.get_variables_output_key(setup_name=self.setup_name)]
        weight_file_path = self.input_dict[MVASelectorTrainTask.get_weight_output_key(setup_name=self.setup_name)]

        alt_working_points = tuple()  # type: Tuple[float, ...]
        if self.use_alternative_working_points:
            wp_tuple = MVASelectionValidatorTask.get_alternative_working_points(mva_setup_name=self.setup_name)
            if wp_tuple is not None:
                alt_working_points = wp_tuple

        mva_validator = MVASelectorValidation(
            path_train_df=train_path,
            path_test_df=test_path,
            path_mva_var_list=var_file_path,
            path_mva_bdt_weights=weight_file_path,
            mva_selector_setup_name=self.setup_name,
            target_dir_path=self.base_path,
            additional_options={
                "alternative_working_points": alt_working_points,
            },
        )

        working_point_value = self.mva_selector_working_point
        assert isinstance(working_point_value, float), type(working_point_value)
        output_info_list, test_roc_auc, train_roc_auc = mva_validator.run_validation(
            base_name=self.setup_name,
            working_point=working_point_value,
            mva_output_latex_name=MVASelectorValidation.default_selector_mva_output_latex_name,
            dry_run=dry_run,
        )

        return {k: v for k, v in output_info_list}

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.run_mva_selection_evaluation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {
            output_key: luigi.LocalTarget(target_path)
            for output_key, target_path in self.run_mva_selection_evaluation(dry_run=True).items()
        }


class MVASelectionValidatorTask(luigi.WrapperTask):
    main_offline_config_path = luigi.Parameter(positional=False)  # type: PathType

    selection_mva_validation_base_dir_name = "MVASelectorsValidations"  # type: str
    _alternative_working_points = {}  # type: Dict[str, Tuple[float, ...]]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        offline_config = utilities.load_yaml_config(yaml_config_path=self.main_offline_config_path)
        self.working_dir = offline_config["working_dir"]
        self.cs_tag = offline_config["cs_config"]["cs_tag"]  # type: str
        self.cs_fastbdt_settings = offline_config["cs_config"]["cs_fastbdt_settings"]  # type: Dict[str, Any]

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    @property
    def mva_selector_setup_infos(self) -> Generator[Tuple[str, float], None, None]:
        for sel_name in SelectionSetups.available_selections.keys():
            mva_info = SelectionSetups.get_selection_mva_info(selection_name=sel_name)  # type: Optional[MVAInfo]
            if mva_info is not None:
                yield mva_info.mva_setup_name, mva_info.mva_post_cut

    @staticmethod
    def get_alternative_working_points(mva_setup_name: str) -> Optional[Tuple[float, ...]]:
        return MVASelectionValidatorTask._alternative_working_points.get(mva_setup_name, None)

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for mva_selector_setup_name, mva_working_point_value in self.mva_selector_setup_infos:

            yield self.clone(
                MVASelectorTestTask,
                mva_selector_setup_name=mva_selector_setup_name,
                working_dir=self.working_dir,
                main_offline_config_path=self.main_offline_config_path,
                cs_tag=self.cs_tag,
                cs_fastbdt_settings=self.cs_fastbdt_settings,
                mva_selector_working_point=mva_working_point_value,
                use_alternative_working_points=True,
                alternative_base_dir_name=self.selection_mva_validation_base_dir_name,
            )
