"""
The objective of the tasks provided here is the application of the correction based on shift and smearing
determined in the resolution evaluation tasks by comparing mc and data in the q^2 sideband.

Felix Metzner 2019
"""

import os
import itertools

import numpy as np
import pandas as pd
import multiprocessing
import b2luigi as luigi

from functools import partial
from luigi.util import inherits
from typing import Union, Optional, Tuple, List, Dict, Generator, Any

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType
from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.data_handling.data_sample_tag_names import DataSampleTags
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.comparison_tools.data_mc_comparer import DataMCComparerInfo, DataMCComparer
from rdstar.offline_analysis.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler
from rdstar.offline_analysis.validation.observable_validation import plot_bin_composition, plot_bin_migration

from rdstar.offline_analysis.mms_resolution_correction.multistage_setups import multi_stage_setups_dict
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction import MissingMassResolutionCorrector
from rdstar.offline_analysis.mms_resolution_correction.multistage_resolution_corrector import (
    MultistageResolutionCorrector,
)
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetup,
    ResolutionCorrectionSetupManager,
)

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.utilities import InputDictType as IDType
from rdstar.offline_analysis.tasks.weight_tasks import WeightFinalizationTask, WeightFinalizationWrapperTask
from rdstar.offline_analysis.tasks.resolution_evaluation_tasks import (
    ResolutionEvaluationTask,
    ResolutionEvaluationManagerTask,
)


__all__ = [
    "ResolutionCorrectionTask",
    "ResolutionCorrectionWrapperTask",
    "ResolutionCorrectionValidationManagerTask",
]


@inherits(WeightFinalizationTask)
class ResolutionCorrectionTask(utilities.BaseOfflineTask):
    appendix: str = DataSampleTags.resolution_corrected_sample_tag

    input_data_paths_key: str = "input_from_weight_finalization_task"
    input_res_corr_infos_key: str = "input_from_resolution_correction_manager_task"

    memory_intensive_combo_names: Tuple[str, ...] = ("Charged", "Mixed")

    @property
    def resources(self) -> Dict[str, int]:
        generic_combo_names: Tuple[str, ...] = self.memory_intensive_combo_names + ("Charm", "UDS")
        if self.combo_name in generic_combo_names:
            return {utilities.offline_memory_resource_name: 5}
        return {utilities.offline_memory_resource_name: 3}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._data_input_dict: Optional[IDType] = None
        self._res_corr_info_input_dict: Optional[IDType] = None
        self._msrc_info_input_dict: Optional[IDType] = None

    @property
    def data_input_dict(self) -> IDType:
        if self._data_input_dict is None:
            self._data_input_dict = {
                k: v.path
                for kk, d in self.input().items()
                if isinstance(d, dict) and kk == self.input_data_paths_key
                for k, v in d.items()
            }
        return self._data_input_dict

    @property
    def res_corr_info_input_dict(self) -> IDType:
        if self._res_corr_info_input_dict is None:
            self._res_corr_info_input_dict = {
                k: v.path
                for kk, d in self.input().items()
                if isinstance(d, dict) and kk == self.input_res_corr_infos_key
                for k, v in d.items()
                if not ResolutionEvaluationManagerTask.is_multi_stage_res_corr_input(key=k)
            }
        return self._res_corr_info_input_dict

    @property
    def msrc_info_input_dict(self) -> IDType:
        if self._msrc_info_input_dict is None:
            self._msrc_info_input_dict = {
                ResolutionEvaluationManagerTask.get_original_multi_stage_input_key(key=k): v.path
                for kk, d in self.input().items()
                if isinstance(d, dict) and kk == self.input_res_corr_infos_key
                for k, v in d.items()
                if ResolutionEvaluationManagerTask.is_multi_stage_res_corr_input(key=k)
            }
        return self._msrc_info_input_dict

    def get_input_files(self) -> Dict[str, PathType]:
        key_path_dict = {
            key: path
            for key, path in self.data_input_dict.items()
            if WeightManager.finalized_weights_appendix in key and os.path.splitext(path)[1] == ".h5"
        }
        assert all(key.startswith(self.combo_name) for key in key_path_dict.keys()), key_path_dict
        return key_path_dict

    def get_resolution_correction_info_path(self, res_corr_setup_name: str) -> PathType:
        return self.res_corr_info_input_dict[res_corr_setup_name]

    def get_msrc_info_dir_path(self, setup_name: str) -> PathType:
        file_path: PathType = self.msrc_info_input_dict[setup_name]
        assert setup_name in str(file_path), (setup_name, file_path)

        setup_dir: PathType = os.path.dirname(os.path.dirname(file_path))
        assert os.path.basename(setup_dir) == setup_name, (setup_dir, setup_name, file_path)
        assert os.path.isdir(setup_dir), (setup_name, setup_dir)
        return setup_dir

    def apply_resolution_correction_for_key(
        self,
        key_path_tuple: Tuple[str, PathType],
        dry_run: bool = False,
    ) -> Tuple[str, PathType]:
        input_key, input_path = key_path_tuple
        output_key, temp_output_path, output_path = self.get_output_key_temp_path_path_triplet(
            input_key=input_key, input_path=input_path
        )

        if dry_run:
            return output_key, output_path

        assert os.path.isfile(input_path), key_path_tuple
        sample = pd.read_hdf(input_path)
        assert isinstance(sample, pd.DataFrame)

        for setup_name, setup_definition in ResolutionCorrectionSetupManager.setup_definitions.items():
            assert setup_name in self.res_corr_info_input_dict.keys(), (setup_name, self.res_corr_info_input_dict.keys())
            original_col_name: str = setup_definition.primary_col_name
            new_col_name = MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
                correction_setup_name=setup_name,
                use_laplace_smeared_version=False,
            )

            check_orig_name: str = MissingMassResolutionCorrector.get_single_stage_original_col_name(
                correction_setup_name=setup_name
            )
            assert original_col_name == check_orig_name, (original_col_name, check_orig_name)
            assert original_col_name in sample.columns, original_col_name
            assert new_col_name not in sample.columns, new_col_name

            res_cor_setup: ResolutionCorrectionSetup = ResolutionCorrectionSetupManager.get_setup_definition(
                name=setup_name
            )

            res_corrector = MissingMassResolutionCorrector(
                resolution_correction_setup=res_cor_setup,
                resolution_info_path=self.get_resolution_correction_info_path(res_corr_setup_name=setup_name),
            )
            res_corrector.add_required_secondary_columns(sample=sample)
            res_corrector.ensure_primary_col_available(sample=sample)
            res_corrector.check_corrected_col_name(expected_col_name=new_col_name, use_laplace_smeared_version=False)

            if len(np.unique(sample.index)) == 0:
                res_corrector.add_corrected_column_to_empty_data_frame(
                    sample=sample,
                    special_treatment=True,
                )
            elif input_key.startswith("Data"):
                res_corrector.add_corrected_column_for_data(
                    sample=sample,
                    special_treatment=True,
                )
            else:
                res_corrector.apply_resolution_correction(
                    sample=sample,
                    correct_overflows=True,
                    special_treatment=True,
                )

        # Apply Multi Stage Resolution Correction Results
        for msrc_setup in multi_stage_setups_dict.values():
            msrc_corrector = MultistageResolutionCorrector(
                correction_setup_name=msrc_setup.name,
                setup_info_dir_path=self.get_msrc_info_dir_path(setup_name=msrc_setup.name),
            )
            msrc_corrector.apply_resolution_correction(sample=sample, is_data=input_key.startswith("Data"))

        sample.to_hdf(temp_output_path, key=output_key)
        os.rename(temp_output_path, output_path)
        return output_key, output_path

    def apply_resolution_correction(self, dry_run: bool = False) -> Dict[str, PathType]:
        processes: int = 4 if self.memory_intensive_combo_names else 10

        with multiprocessing.Pool(processes=processes) as pool:
            output_tuples = pool.map(
                partial(ResolutionCorrectionTask.apply_resolution_correction_for_key, self, dry_run=dry_run),
                self.get_input_files().items(),
            )

        return {output_key: output_path for output_key, output_path in output_tuples}

    def get_output_key_temp_path_path_triplet(
        self,
        input_key: str,
        input_path: PathType,
    ) -> Tuple[str, PathType, PathType]:
        dir_path, file_name = os.path.split(input_path)
        output_name_1, output_name_2 = file_name.split(f"_{WeightManager.finalized_weights_appendix}_", 1)
        output_name = f"{output_name_1}_{WeightManager.finalized_weights_appendix}_{self.appendix}_{output_name_2}"
        output_path = os.path.join(dir_path, output_name)
        temp_output_path = os.path.join(dir_path, f"temp_{output_name}")
        output_key_1, output_key_2 = input_key.split(f"_{WeightManager.finalized_weights_appendix}_", 1)
        output_key = f"{output_key_1}_{WeightManager.finalized_weights_appendix}_{self.appendix}_{output_key_2}"

        return output_key, temp_output_path, output_path

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        self.apply_resolution_correction(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.apply_resolution_correction(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Dict[str, utilities.TaskType]:
        return {
            self.input_data_paths_key: self.clone(WeightFinalizationTask),
            self.input_res_corr_infos_key: ResolutionEvaluationManagerTask(
                working_dir=self.working_dir,
                selection_name=SelectionSetups.resolution_correction_mapping[self.selection_name],
                cs_tag=self.cs_tag,
                config_combinations=self.config_combinations,
            ),
        }


@inherits(WeightFinalizationWrapperTask)
class ResolutionCorrectionWrapperTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )


@inherits(ResolutionEvaluationTask)
class ResolutionCorrectionValidationTask(utilities.BaseOfflineTask):
    sample_setup_key: str = luigi.Parameter(positional=False)
    scope_setting: str = luigi.Parameter(positional=False)
    normed_to_data: bool = luigi.BoolParameter(positional=False)
    number_of_validation_bins: int = luigi.IntParameter(positional=False)

    sample_setup_key_option_name: str = "sample_setup_key"
    var_scope_option_name: str = "scope_setting"
    n_bins_option_name: str = "number_of_validation_bins"
    normed_option_name: str = "normed_to_data"
    normed_options: Tuple[bool, bool] = (True, False)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert self.number_of_validation_bins > 0, self.number_of_validation_bins

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )

        self.data_sample_key: str = SampleSetupHandler.get_data_sample_key(sample_name=self.sample_setup_key)
        self.mc_sample_keys: Tuple[str, ...] = SampleSetupHandler.get_mc_sample_keys(sample_name=self.sample_setup_key)

        self.__scope_tuple: Optional[Tuple[float, float]] = None

        self._res_corr_setup: ResolutionCorrectionSetup = ResolutionCorrectionSetupManager.get_setup_definition(
            name=self.resolution_correction_setup_name
        )

        self.before_col: str = MissingMassResolutionCorrector.get_single_stage_original_col_name(
            correction_setup_name=self.resolution_correction_setup_name
        )

        self.after_col: str = MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
            correction_setup_name=self.resolution_correction_setup_name,
            use_laplace_smeared_version=False,
        )

        self.weight_col_name: str = WeightManager.final_weight_with_additions_col

        self.sideband_key: str = "q2_sideband"

        self.key_prefix: str = (
            f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}_"
            f"{self.selection_name}_with_CS_{self.cs_tag}_"
        )

        self.figure_size: Tuple[float, float] = (8.0, 4.5)

        sub_dir_name1: str = (
            f"validation_{self.scope_tuple_str}_in_{self.number_of_validation_bins}_bins_on_{self.sample_setup_key}"
        )
        sub_dir_name2: str = ": strnormed_to_data" if self.normed_to_data else "unscaled"

        self.base_path: PathType = os.path.join(
            self.working_dir,
            ResolutionEvaluationTask.base_dir_name,
            self.main_sel_tag,
            self.resolution_correction_setup_name,
            sub_dir_name1,
        )

        self.comparison_plots_path: PathType = os.path.join(self.base_path, sub_dir_name2)
        self.composition_plots_path: PathType = os.path.join(self.base_path, "bin_composition")
        self.migration_plots_path: PathType = os.path.join(self.base_path, "bin_migration")

        self._input_dict: Optional[IDType] = None

    @property
    def input_dict(self) -> IDType:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def is_tight_scope_setting(self) -> bool:
        return self.scope_setting.startswith("tight")

    @property
    def scope_tuple(self) -> Tuple[float, float]:
        if self.__scope_tuple is not None:
            return self.__scope_tuple

        special_option_cases: Dict[
            Tuple[str, int], Tuple[float, float]
        ] = ResolutionCorrectionSetupManager.get_special_options_for_setup(
            setup_name=self.resolution_correction_setup_name
        )

        if (self.scope_setting, self.number_of_validation_bins) in special_option_cases.keys():
            self.__scope_tuple = special_option_cases[(self.scope_setting, self.number_of_validation_bins)]
        else:
            scope_options: Dict[str, Tuple[float, float]] = ResolutionCorrectionSetupManager.get_scope_options_for_setup(
                setup_name=self.resolution_correction_setup_name
            )
            self.__scope_tuple = scope_options[self.scope_setting]
        return self.__scope_tuple

    @property
    def scope_tuple_str(self) -> str:
        scope_str: str = f"from_{self.scope_tuple[0]}_to_{self.scope_tuple[1]}"
        scope_str = scope_str.replace(".", "p")
        scope_str = scope_str.replace("-", "minus")
        return scope_str

    def get_hist_variables(self) -> List[Tuple[HistVariable, bool]]:
        return [
            (
                HistVariable(
                    df_label=var,
                    n_bins=self.number_of_validation_bins,
                    scope=self.scope_tuple,
                    var_name=self._res_corr_setup.primary_hist_var.variable_name,
                    unit=self._res_corr_setup.primary_hist_var.unit,
                ),
                var == self.after_col,
            )
            for var in [self.before_col, self.after_col]
        ]

    def get_comparer_infos(self) -> List[Tuple[DataMCComparerInfo, bool]]:
        title_end = f"Correction via {self._res_corr_setup.latex_title_appendix}"
        q2_sideband_title = rf"{LatexObsNames.fitting_q2} Sideband"
        q2_sideband_text = q2_sideband_title + "\n" + LatexObsNames.fitting_q2[:-1] + r" < 4.0" + LatexUnits.gev_sq[1:]
        comparer_infos: List[Tuple[DataMCComparerInfo, bool]] = [
            (
                DataMCComparerInfo(
                    key="q2_sideband",
                    hist_var=var,
                    title=f"{'After' if is_corrected else 'Before'} {var.variable_name} {title_end}",
                    info_text=q2_sideband_text,
                    legend_loc=(1, 1),
                    legend_cols=(4, 1),
                    y_scale=(1.2, 1.6),
                    text_h_pos=(0.74, 0.56 if not self.is_tight_scope_setting else 0.02),
                    text_v_pos=(0.81, 0.98),
                    systematics=None,
                    data_component_label="Sideband Data",
                ),
                is_corrected,
            )
            for var, is_corrected in self.get_hist_variables()
        ]

        comparer_infos_with_full_sys: List[Tuple[DataMCComparerInfo, bool]] = [
            (
                DataMCComparerInfo.reset_systematics_in_comparer_info(
                    comparer_info=ci, new_systematics=full_systematics_cols, sys_name_tag=full_systematics_name
                ),
                is_corrected,
            )
            for ci, is_corrected in comparer_infos
        ]

        return comparer_infos + comparer_infos_with_full_sys

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_key = f"{self.key_prefix}{self.sideband_key}"
        mc_paths: List[PathType] = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (k == f"{mc_key}_{full_key}" and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths: List[PathType] = [
            p
            for k, p in self.input_dict.items()
            if (k == f"{self.data_sample_key}_{full_key}" and os.path.splitext(p)[1] == ".h5")
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame), type(data_sample)

        return mc_sample, data_sample

    def do_resolution_correction_evaluation(self, dry_run: bool = False) -> Dict[str, PathType]:
        if dry_run:
            mc_sample, data_sample = pd.DataFrame(), pd.DataFrame()
        else:
            mc_sample, data_sample = self.get_mc_and_data_sample()

        output_dict: Dict[str, PathType] = {}

        for comparer_info, is_corrected in self.get_comparer_infos():
            for style in comparer_info.styles:
                comparer = DataMCComparer(comparer_info=comparer_info)

                new_output = comparer.plot_data_mc_comparison(
                    mc_df=mc_sample,
                    data_df=data_sample,
                    weight_col=self.weight_col_name,
                    export_path=self.comparison_plots_path,
                    use_short_titles=True,
                    style=style,
                    ratio_type="vs_uncert",
                    gof_check_method="toys",
                    fig_size=self.figure_size,
                    normalize_mc_to_data=self.normed_to_data,
                    adaptive_binning=True,
                    dry_run=dry_run,
                )
                assert new_output is not None
                assert all(k not in output_dict for k in new_output.keys())
                assert all(v not in output_dict.values() for v in new_output.values())
                output_dict.update(new_output)

        if self.normed_to_data:
            rdstar_reco_modes = {k: v for k, v in BSigRecoModes.main_modes.main_reco_mode_infos_as_tuple}
            rdstar_reco_mode_labels = BSigRecoModes.main_modes.get_reco_mode_name_to_latex_label_mapping(short=False)

            for main_hist_var, is_corrected in self.get_hist_variables():
                new_output_entries = plot_bin_composition(
                    mc_df=mc_sample,
                    main_hist_var=main_hist_var,
                    secondary_hist_var=self._res_corr_setup.secondary_hist_var,
                    secondary_binning=self._res_corr_setup.secondary_var_binning,
                    secondary_hist_var_label_mapping=self._res_corr_setup.tick_label_mapping,
                    weight_col=self.weight_col_name,
                    plot_shape=True,
                    dec_modes=rdstar_reco_modes,
                    dec_titles=rdstar_reco_mode_labels,
                    file_name_suffix="corrected" if is_corrected else "uncorrected",
                    export_path=os.path.join(self.composition_plots_path, "corrected" if is_corrected else "uncorrected"),
                    dry_run=dry_run,
                )
                assert all(k not in output_dict for k in new_output_entries.keys())
                assert all(v not in output_dict.values() for v in new_output_entries.values())
                output_dict.update(new_output_entries)

            before_and_after_hist_vars: Tuple[HistVariable, ...] = tuple(v for v, _ in self.get_hist_variables())
            new_output_entries = plot_bin_migration(
                mc_df=mc_sample,
                from_hist_var=before_and_after_hist_vars[0],
                to_hist_var=before_and_after_hist_vars[1],
                from_to_label_appendix=("", "(Corrected)"),
                weight_col=self.weight_col_name,
                normalize_to_origin=True,
                dec_modes=rdstar_reco_modes,
                dec_titles=rdstar_reco_mode_labels,
                file_name_suffix=None,
                show_color_bar=False,
                fig_size=(12.0, 10.0),
                export_path=self.migration_plots_path,
                dry_run=dry_run,
            )
            assert all(k not in output_dict for k in new_output_entries.keys())
            assert all(v not in output_dict.values() for v in new_output_entries.values())
            output_dict.update(new_output_entries)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_resolution_correction_evaluation(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_resolution_correction_evaluation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )

    @staticmethod
    def get_parameter_combinations(res_corr_setup_name: str) -> Tuple[Dict[str, Any], ...]:
        sample_setup_key_options: Tuple[str, ...] = SampleSetupHandler.on_resonance_sample_setup_names
        n_bin_options = ResolutionCorrectionSetupManager.get_n_bin_options_for_setup(setup_name=res_corr_setup_name)
        scope_options = ResolutionCorrectionSetupManager.get_scope_options_for_setup(setup_name=res_corr_setup_name)

        options: Dict[str, Tuple[Union[str, int, bool], ...]] = {
            ResolutionCorrectionValidationTask.sample_setup_key_option_name: sample_setup_key_options,
            ResolutionCorrectionValidationTask.n_bins_option_name: n_bin_options,
            ResolutionCorrectionValidationTask.var_scope_option_name: tuple(scope_options.keys()),
            ResolutionCorrectionValidationTask.normed_option_name: ResolutionCorrectionValidationTask.normed_options,
        }

        value_combinations: List[Tuple[Any, ...]] = list(itertools.product(*options.values()))

        assert all(isinstance(vc, tuple) for vc in value_combinations)
        assert all(len(vc) == 4 for vc in value_combinations)
        assert all(isinstance(vc[0], str) for vc in value_combinations)
        assert all(isinstance(vc[1], int) for vc in value_combinations)
        assert all(isinstance(vc[2], str) for vc in value_combinations)
        assert all(isinstance(vc[3], bool) for vc in value_combinations)

        return tuple(
            [
                dict(zip(options.keys(), vc))
                for vc in value_combinations
                if abs(scope_options[vc[2]][1] - scope_options[vc[2]][0]) / (1.0 * vc[1]) >= 0.01
            ]
        )


@inherits(ResolutionEvaluationManagerTask)
class ResolutionCorrectionValidationManagerTask(luigi.WrapperTask):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for res_corr_setup_name in ResolutionCorrectionSetupManager.get_setup_definition_names():
            parameter_combinations: Tuple[
                Dict[str, Any], ...
            ] = ResolutionCorrectionValidationTask.get_parameter_combinations(res_corr_setup_name=res_corr_setup_name)
            for parameter_combination in parameter_combinations:
                yield self.clone(
                    ResolutionCorrectionValidationTask,
                    resolution_correction_setup_name=res_corr_setup_name,
                    **parameter_combination,
                )
