"""
The objective of the tasks provided by this module is the determination of the difference in resolution of the
observable missing mass squared on data and MC.
The resolution difference and shift factors are obtained from the q^2 sideband .

Felix Metzner 2019
"""

import os
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from typing import Optional, Tuple, List, Dict, Generator

from rdstar.utilities import PathType
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler

from rdstar.offline_analysis.mms_resolution_correction.resolution_evaluation import MissingMassResolutionEvaluator
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetup,
    ResolutionCorrectionSetupManager,
)

from rdstar.offline_analysis.mms_resolution_correction.multistage_resolution_evaluator import (
    MultistageResolutionEvaluator,
)
from rdstar.offline_analysis.mms_resolution_correction.multistage_setups import (
    multi_stage_setups_dict,
    MultistageResolutionCorrectionSetup,
)


from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.tasks.weight_tasks import WeightFinalizationTask, WeightFinalizationWrapperTask


__all__ = [
    "ResolutionEvaluationTask",
    "ResolutionEvaluationManagerTask",
]


@inherits(WeightFinalizationWrapperTask)
class ResolutionEvaluationTask(utilities.BaseOfflineTask):
    resolution_correction_setup_name = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "ResolutionEvaluation"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.main_sel_tag,
            self.resolution_correction_setup_name,
        )  # type: PathType

        self.key_prefix = f"_{self.selection_name}_with_CS_{self.cs_tag}_"  # type: str
        _sb_key = SelectionSetups.get_res_corr_source_sub_sample_key(selection_name=self.selection_name)  # type: str
        self.sideband_key = f"{_sb_key}" if _sb_key else _sb_key  # type: str

        self.res_corr_setup = ResolutionCorrectionSetupManager.get_setup_definition(
            name=self.resolution_correction_setup_name
        )  # type: ResolutionCorrectionSetup

        sample_setup_key = SampleSetupHandler.get_on_resonance_sample_key_for(
            use_alternative_gap=self.res_corr_setup.use_alternative_gap
        )  # type: str
        self.data_sample_key = SampleSetupHandler.get_data_sample_key(sample_name=sample_setup_key)  # type: str
        self.mc_sample_keys = SampleSetupHandler.get_mc_sample_keys(sample_name=sample_setup_key)  # type: Tuple[str, ...]

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def resolution_correction_setup(self) -> ResolutionCorrectionSetup:
        return ResolutionCorrectionSetupManager.get_setup_definition(name=self.resolution_correction_setup_name)

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_key = f"{self.key_prefix}{self.sideband_key}"  # type: str
        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{mc_key}_{WeightManager.finalized_weights_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]  # type: List[PathType]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths = [
            p
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{self.data_sample_key}_{WeightManager.finalized_weights_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame), type(data_sample)

        if self.resolution_correction_setup.fei_weight_col is not None:
            WeightManager.switch_fei_calibration(
                df=mc_sample,
                main_weight_col_name=WeightManager.final_weight_with_additions_col,
                new_fei_calib_col=WeightManager.used_fei_calib_col,
            )

        return mc_sample, data_sample

    def evaluate_resolution(self, dry_run: bool = False) -> Dict[str, PathType]:

        resolution_evaluator = MissingMassResolutionEvaluator(
            resolution_correction_setup=self.resolution_correction_setup,
            weight_column=WeightManager.final_weight_with_additions_col,
            results_path=self.base_path,
        )

        if dry_run:
            return resolution_evaluator.get_output_dict()

        mc_sample, data_sample = self.get_mc_and_data_sample()  # type: pd.DataFrame, pd.DataFrame
        resolution_evaluator.add_required_columns_to_dfs(mc_df=mc_sample, data_df=data_sample)

        resolution_evaluator.evaluate_resolution(mc_df=mc_sample, data_df=data_sample)

        return resolution_evaluator.get_output_dict()

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.evaluate_resolution(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.evaluate_resolution(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield WeightFinalizationTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )


@inherits(WeightFinalizationWrapperTask)
class MultiStageResolutionEvaluationTask(utilities.BaseOfflineTask):
    multi_stage_resolution_correction_setup_name = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "MultiStageResolutionEvaluation"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert (
            self.multi_stage_resolution_correction_setup_name in multi_stage_setups_dict
        ), self.multi_stage_resolution_correction_setup_name

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.main_sel_tag,
        )  # type: PathType

        self.key_prefix = f"_{self.selection_name}_with_CS_{self.cs_tag}"  # type: str
        _sb_key = SelectionSetups.get_res_corr_source_sub_sample_key(selection_name=self.selection_name)  # type: str
        self.sideband_key = f"_{_sb_key}" if _sb_key else _sb_key  # type: str

        _sample_setup_key = SampleSetupHandler.default_gap_sample_name  # type: str
        self.data_sample_key = SampleSetupHandler.get_data_sample_key(sample_name=_sample_setup_key)  # type: str
        self.mc_sample_keys = SampleSetupHandler.get_mc_sample_keys(
            sample_name=_sample_setup_key
        )  # type: Tuple[str, ...]

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def multi_stage_res_corr_setup(self) -> MultistageResolutionCorrectionSetup:
        return multi_stage_setups_dict[self.multi_stage_resolution_correction_setup_name]

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        full_key = f"{self.key_prefix}{self.sideband_key}"  # type: str
        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{mc_key}_{WeightManager.finalized_weights_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]  # type: List[PathType]
        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key)

        data_paths = [
            p
            for k, p in self.input_dict.items()
            if (
                k.startswith(f"{self.data_sample_key}_{WeightManager.finalized_weights_appendix}_")
                and k.endswith(full_key)
                and os.path.splitext(p)[1] == ".h5"
            )
        ]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame), type(data_sample)

        if self.multi_stage_res_corr_setup.fei_weight_col is not None:
            WeightManager.switch_fei_calibration(
                df=mc_sample,
                main_weight_col_name=WeightManager.final_weight_with_additions_col,
                new_fei_calib_col=WeightManager.used_fei_calib_col,
            )

        return mc_sample, data_sample

    def evaluate_shift_and_resolution(self, dry_run: bool = False) -> Dict[str, PathType]:
        if dry_run:
            mc_sample, data_sample = None, None  # type: Optional[pd.DataFrame], Optional[pd.DataFrame]
        else:
            mc_sample, data_sample = self.get_mc_and_data_sample()

        sel_mva_col: Optional[str] = SelectionSetups.get_selection_mva_output_col(selection_name=self.selection_name)
        msrc_evaluator = MultistageResolutionEvaluator(
            correction_setup=self.multi_stage_res_corr_setup,
            mc_sample=mc_sample,
            data_sample=data_sample,
            selection_mva_output_col_name=sel_mva_col,
            base_output_dir_path=self.base_path,
        )

        output_dict = {k: v for k, v in msrc_evaluator.get_output_dict().items()}  # type: Dict[str, PathType]

        if not dry_run:
            assert mc_sample is not None
            assert data_sample is not None
            msrc_evaluator.evaluate_mms_resolution(dry_run=False)

        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.evaluate_shift_and_resolution(dry_run=False)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.evaluate_shift_and_resolution(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield WeightFinalizationTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )


@inherits(WeightFinalizationWrapperTask)
class ResolutionEvaluationManagerTask(utilities.BaseOfflineTask):

    results_dir_name = "Results"  # type: str
    multi_stage_res_corr_key_prefix = "msrc_input_"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.main_sel_tag = SelectionSetups.get_main_selection_tag(
            selection_name=self.selection_name,
            cs_tag=self.cs_tag,
        )  # type: str

        self.base_path = os.path.join(
            self.working_dir,
            ResolutionEvaluationTask.base_dir_name,
            self.main_sel_tag,
            self.results_dir_name,
        )  # type: PathType

        self.multi_stage_base_path = os.path.join(
            self.working_dir,
            MultiStageResolutionEvaluationTask.base_dir_name,
            self.main_sel_tag,
        )  # type: PathType

        self._simple_res_corr_input_dict = None  # type: Optional[Dict[str, PathType]]
        self._msrc_input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def simple_res_corr_input_dict(self) -> Dict[str, PathType]:
        if self._simple_res_corr_input_dict is None:
            assert isinstance(self.input(), dict), self.input()
            self._simple_res_corr_input_dict = {
                k: v.path
                for k, d in self.input().items()
                if isinstance(d, dict) and not self.is_multi_stage_res_corr_input(key=k)
                for kk, v in d.items()
                if kk == MissingMassResolutionEvaluator.default_result_file
            }

        return self._simple_res_corr_input_dict

    @property
    def msrc_input_dict(self) -> Dict[str, PathType]:
        if self._msrc_input_dict is None:
            assert isinstance(self.input(), dict), self.input()
            self._msrc_input_dict = {
                self.get_original_multi_stage_input_key(key=k): v.path
                for k, d in self.input().items()
                if isinstance(d, dict) and self.is_multi_stage_res_corr_input(key=k)
                for kk, v in d.items()
                if kk == MultistageResolutionEvaluator.last_res_output_key
            }

        return self._msrc_input_dict

    def run(self) -> None:
        for setup_name, path in self.simple_res_corr_input_dict.items():
            os.symlink(src=path, dst=self.create_target_path(correction_setup_name=setup_name))

        for msrc_setup_name, path in self.msrc_input_dict.items():
            target_path = self.create_msrc_target_path(setup_name=msrc_setup_name, file_path=path)  # type: PathType
            os.makedirs(os.path.dirname(target_path), exist_ok=True)
            os.symlink(src=path, dst=target_path)

    def create_target_path(self, correction_setup_name: str) -> PathType:
        file_name = f"{MissingMassResolutionEvaluator.default_result_file}_{correction_setup_name}.json"  # type: str
        return os.path.join(self.base_path, file_name)

    def create_msrc_target_path(self, setup_name: str, file_path: Optional[PathType] = None) -> PathType:
        if file_path is not None:
            assert setup_name in str(file_path), (setup_name, file_path)
            test_setup_dir = os.path.dirname(os.path.dirname(file_path))  # type: PathType
            assert os.path.basename(test_setup_dir) == setup_name, (test_setup_dir, setup_name)

        file_name = f"{MissingMassResolutionEvaluator.default_result_file}.json"  # type: str
        return os.path.join(self.multi_stage_base_path, setup_name, self.results_dir_name, file_name)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = {}  # type: Dict[str, luigi.LocalTarget]
        for res_corr_setup_name in ResolutionCorrectionSetupManager.get_setup_definition_names():
            target_path = self.create_target_path(correction_setup_name=res_corr_setup_name)  # type: PathType
            output_dict.update({res_corr_setup_name: luigi.LocalTarget(target_path)})
        for msrc_setup_name in multi_stage_setups_dict:
            msrc_output_key = self.get_modified_multi_stage_input_key(key=msrc_setup_name)  # type: str
            assert msrc_output_key not in output_dict, (msrc_setup_name, msrc_output_key)
            target_path = self.create_msrc_target_path(setup_name=msrc_setup_name)
            output_dict.update({msrc_output_key: luigi.LocalTarget(target_path)})

        return output_dict

    def get_modified_multi_stage_input_key(self, key: str) -> str:
        return f"{self.multi_stage_res_corr_key_prefix}{key}"

    @staticmethod
    def get_original_multi_stage_input_key(key: str) -> str:
        return key.split(ResolutionEvaluationManagerTask.multi_stage_res_corr_key_prefix, 1)[1]

    @staticmethod
    def is_multi_stage_res_corr_input(key: str) -> bool:
        return key.startswith(ResolutionEvaluationManagerTask.multi_stage_res_corr_key_prefix)

    def requires(self) -> Dict[str, utilities.TaskType]:
        required_tasks_dict = {}  # type: Dict[str, utilities.TaskType]
        for res_corr_setup_name in ResolutionCorrectionSetupManager.get_setup_definition_names():
            required_tasks_dict.update(
                {
                    res_corr_setup_name: ResolutionEvaluationTask(
                        working_dir=self.working_dir,
                        selection_name=SelectionSetups.resolution_correction_mapping[self.selection_name],
                        cs_tag=self.cs_tag,
                        config_combinations=self.config_combinations,
                        resolution_correction_setup_name=res_corr_setup_name,
                    )
                }
            )

        for msrc_setup_name in multi_stage_setups_dict:
            required_tasks_dict.update(
                {
                    self.get_modified_multi_stage_input_key(key=msrc_setup_name): MultiStageResolutionEvaluationTask(
                        working_dir=self.working_dir,
                        selection_name=SelectionSetups.resolution_correction_mapping[self.selection_name],
                        cs_tag=self.cs_tag,
                        config_combinations=self.config_combinations,
                        multi_stage_resolution_correction_setup_name=msrc_setup_name,
                    )
                }
            )

        return required_tasks_dict
