"""
Tasks performing the Sideband Validation

Felix Metzner 2019
"""

import os
import copy
import pandas as pd
import b2luigi as luigi

from luigi.util import inherits
from collections import abc, defaultdict

from typing import Optional, List, Tuple, Dict, Generator, NamedTuple

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.comparison_tools.data_mc_comparer import DataMCComparer
from rdstar.offline_analysis.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.validation.sideband_validation import sideband_info_collection, SidebandInfo

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit import MCShiftInfo
from rdstar.offline_analysis.data_handling.selected_data_handling import DataSampleLoader
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.tasks.resolution_evaluation_tasks import ResolutionEvaluationManagerTask


__all__ = [
    "SidebandEvaluationWrapperTask",
    "SidebandEvaluationTask",
]


logger = log_to_default(name=__name__)


class SidebandInfoCollectionEntry(NamedTuple):
    sideband_key: str
    sample_setup_keys: Tuple[str, ...]
    sideband_infos: Tuple[SidebandInfo, ...]

    def matches_with(self, sideband_key: str, sample_setup_key: str) -> bool:
        return sideband_key == self.sideband_key and sample_setup_key in self.sample_setup_keys


@inherits(ResolutionEvaluationManagerTask)
class SidebandEvaluationWrapperTask(luigi.WrapperTask):
    systematics_info_names = ("no_sys", full_systematics_name)  # type: Tuple[str, ...]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    @staticmethod
    def get_sorted_sideband_infos(selection_name: str) -> Tuple[SidebandInfoCollectionEntry, ...]:
        sideband_infos = copy.copy(sideband_info_collection)  # type: List[SidebandInfo]
        sorted_sideband_info_dict = defaultdict(list)  # type: Dict[Tuple[str, Tuple[str, ...]], List[SidebandInfo]]
        for sideband_info in sideband_infos:
            is_relevant = SidebandEvaluationWrapperTask.sideband_info_relevant_for_selection(
                sideband_info=sideband_info,
                selection_name=selection_name,
            )  # type: bool

            if is_relevant:
                _key = (sideband_info.sideband_sample_key, sideband_info.valid_sample_setups)
                sorted_sideband_info_dict[_key].append(sideband_info)

        sorted_sideband_infos = tuple(
            SidebandInfoCollectionEntry(
                sideband_key=k[0],
                sample_setup_keys=k[1],
                sideband_infos=tuple(v),
            )
            for k, v in sorted_sideband_info_dict.items()
        )  # type: Tuple[SidebandInfoCollectionEntry, ...]

        return sorted_sideband_infos

    @staticmethod
    def sideband_info_relevant_for_selection(sideband_info: SidebandInfo, selection_name: str) -> bool:
        sb_info_selection_req = sideband_info.requires_selection  # type: Optional[Tuple[str, ...]]
        if sb_info_selection_req is not None:
            if selection_name not in sb_info_selection_req:
                return False
        return True

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for sideband_info_container in self.get_sorted_sideband_infos(selection_name=self.selection_name):
            for systematics_info_name in self.systematics_info_names:
                for sample_setup_key in sideband_info_container.sample_setup_keys:
                    yield self.clone(
                        SidebandEvaluationTask,
                        sample_setup_key=sample_setup_key,
                        side_band_key=sideband_info_container.sideband_key,
                        systematics_info_name=systematics_info_name,
                    )


@inherits(SidebandEvaluationWrapperTask)
class SidebandEvaluationTask(utilities.BaseOfflineTask):
    sample_setup_key = luigi.Parameter(positional=False)  # type: str
    side_band_key = luigi.Parameter(positional=False)  # type: str
    systematics_info_name = luigi.Parameter(positional=False)  # type: str

    base_dir_name = "SidebandValidation"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(
            self.working_dir,
            self.base_dir_name,
            self.selection_name,
            self.sample_setup_key,
        )  # type: PathType

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.side_band_infos = self.get_side_band_infos_w_or_wo_systematics()  # type: Tuple[SidebandInfo, ...]

        self._input_dict = None  # type: Optional[Dict[str, PathType]]
        self._data_loader = None  # type: Optional[DataSampleLoader]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    @property
    def data_loader(self) -> DataSampleLoader:
        if self._data_loader is None:
            sideband_identifiers = DataSampleLoader.get_identifiers_from_sideband_key(sideband_key=self.side_band_key)
            self._data_loader = DataSampleLoader(
                input_dict=self.input_dict,
                sample_setup_key=self.sample_setup_key,
                cs_tag=self.cs_tag,
                selection_name=self.selection_name,
                input_level=2,
                sideband_identifiers=sideband_identifiers,
            )

        return self._data_loader

    def get_side_band_infos_w_or_wo_systematics(self) -> Tuple[SidebandInfo, ...]:
        side_band_infos = SidebandEvaluationWrapperTask.get_sorted_sideband_infos(
            selection_name=self.selection_name
        )  # type: Tuple[SidebandInfoCollectionEntry, ...]

        assert self.systematics_info_name in SidebandEvaluationWrapperTask.systematics_info_names, (
            self.systematics_info_name,
        )

        relevant_side_band_infos_entries = [
            v
            for v in side_band_infos
            if v.matches_with(sideband_key=self.side_band_key, sample_setup_key=self.sample_setup_key)
        ]  # type: List[SidebandInfoCollectionEntry]
        assert len(relevant_side_band_infos_entries) == 1, (
            len(relevant_side_band_infos_entries),
            [(k.sideband_key, k.sample_setup_keys) for k in relevant_side_band_infos_entries],
        )
        relevant_side_band_infos_entry = relevant_side_band_infos_entries[0]  # type: SidebandInfoCollectionEntry

        side_band_info_list = relevant_side_band_infos_entry.sideband_infos  # type: Tuple[SidebandInfo, ...]
        if self.systematics_info_name == "no_sys":
            return side_band_info_list
        elif self.systematics_info_name == full_systematics_name:
            side_band_infos_with_full_sys = SidebandInfo.create_comparer_infos_with_systematics(
                comparer_info_list=list(side_band_info_list),
                systematics_collection=full_systematics_cols,
                sys_name_tag=full_systematics_name,
            )  # type: List[SidebandInfo]
            return tuple(side_band_infos_with_full_sys)
        else:
            raise ValueError(
                f"systematics_info_name '{self.systematics_info_name}' "
                f"is not known to SidebandEvaluationTask! "
                f"Available options are: {SidebandEvaluationWrapperTask.systematics_info_names}."
            )

    def produce_sideband_plot(
        self,
        mc_df: Optional[pd.DataFrame],
        data_df: Optional[pd.DataFrame],
        input_tuple: Tuple[SidebandInfo, str],
        dry_run: bool,
    ) -> Dict[str, PathType]:

        sideband_info, style = input_tuple  # type: SidebandInfo, str
        assert sideband_info.key == self.side_band_key, (sideband_info.key, self.side_band_key)

        output_path = os.path.join(self.base_path, sideband_info.hist_var.df_label, self.systematics_info_name)

        data_mc_comparer = DataMCComparer(comparer_info=sideband_info)

        _weight_column_name = WeightManager.final_weight_with_additions_col  # type: str

        if not dry_run:
            using_copy_of_mc_df = False  # type: bool

            os.makedirs(name=output_path, exist_ok=True)
            assert isinstance(mc_df, pd.DataFrame), type(mc_df)
            assert isinstance(data_df, pd.DataFrame), type(data_df)

            if data_mc_comparer.comparer_info.additional_selection_info is not None:
                _data_df = data_mc_comparer.apply_additional_selection(df=data_df)  # type: pd.DataFrame
                _mc_df = data_mc_comparer.apply_additional_selection(df=mc_df)  # type: pd.DataFrame
                using_copy_of_mc_df = True
            else:
                _data_df = data_df
                _mc_df = mc_df

            mm2_shifter = sideband_info.get_additional_setting_with_default(
                setting_key=MCShiftInfo.extra_info_key(),
                default_setting=None,
            )  # type: Optional[MCShiftInfo]
            if mm2_shifter is not None:
                if not using_copy_of_mc_df:
                    _mc_df = copy.deepcopy(_mc_df)
                    using_copy_of_mc_df = True
                assert using_copy_of_mc_df
                mm2_shifter.apply_shifts(df=_mc_df)

            if sideband_info.add_new_fei_calibration_factors:
                if not using_copy_of_mc_df:
                    _mc_df = copy.deepcopy(_mc_df)
                    using_copy_of_mc_df = True
                assert using_copy_of_mc_df

                WeightManager.switch_fei_calibration(
                    df=_mc_df,
                    new_fei_calib_col=WeightManager.used_fei_calib_col,
                    main_weight_col_name=_weight_column_name,
                )

            if data_mc_comparer.comparer_info.mc_pre_scaling is not None:
                assert data_mc_comparer.comparer_info.mc_pre_scaling.original_weight_column_name == _weight_column_name, (
                    data_mc_comparer.comparer_info.mc_pre_scaling.original_weight_column_name,
                    _weight_column_name,
                )
                rescaled_weight_column_name = f"__tmp_pre_scaled_{_weight_column_name}__"  # type: str
                _weight_column_name = rescaled_weight_column_name
                data_mc_comparer.apply_mc_sample_pre_scaling(
                    df=_mc_df,
                    new_weight_column_name=rescaled_weight_column_name,
                )

        else:
            _mc_df = pd.DataFrame()
            _data_df = pd.DataFrame()

        try:
            plot_output_dict = data_mc_comparer.plot_data_mc_comparison(
                mc_df=mc_df if dry_run else _mc_df,
                data_df=data_df if dry_run else _data_df,
                dry_run=dry_run,
                weight_col=_weight_column_name,
                style=style,
                normalize_mc_to_data=True,
                ratio_type="vs_uncert",
                gof_check_method="toys",
                adaptive_binning=True,
                export_path=output_path,
            )
            assert plot_output_dict is not None
        except Exception as e:
            logger.error(
                f"Exception occurred for variable {sideband_info.hist_var.df_label} "
                f"and sample setup {self.sample_setup_key}!"
            )
            raise e

        return plot_output_dict

    def produce_sideband_plot_with_check(
        self,
        mc_df: Optional[pd.DataFrame],
        data_df: Optional[pd.DataFrame],
        input_tuple: Tuple[SidebandInfo, str],
        dry_run: bool,
        overwrite_existing: bool = False,
    ) -> Dict[str, PathType]:

        if not overwrite_existing:
            eval_output = self.produce_sideband_plot(
                mc_df=mc_df,
                data_df=data_df,
                input_tuple=input_tuple,
                dry_run=True,
            )  # type: Dict[str, PathType]
            if all(os.path.exists(p) for p in eval_output.values()):
                return eval_output

        eval_output = self.produce_sideband_plot(
            mc_df=mc_df,
            data_df=data_df,
            input_tuple=input_tuple,
            dry_run=dry_run,
        )
        return eval_output

    def do_sideband_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        if not self.data_loader.files_are_available:
            return output_dict

        data_df, mc_df = (None, None)  # type: Optional[pd.DataFrame], Optional[pd.DataFrame]
        if not dry_run:
            data_df, mc_df = self.data_loader.load_data_and_mc()

            if any(sbi.reset_off_to_on_res_scaling for sbi in self.side_band_infos):
                assert all(sbi.reset_off_to_on_res_scaling for sbi in self.side_band_infos), (
                    self.side_band_infos[0].sideband_sample_key,
                    [sbi.hist_var.df_label for sbi in self.side_band_infos if sbi.reset_off_to_on_res_scaling],
                    [sbi.hist_var.df_label for sbi in self.side_band_infos if not sbi.reset_off_to_on_res_scaling],
                )
                WeightManager.remove_off_to_on_resonance_scaling_from_weight_with_adds(df=mc_df)

        sb_plot_infos = [(sb_info, style) for sb_info in self.side_band_infos for style in sb_info.styles]
        assert all(sb_info.key == self.side_band_key for sb_info, _ in sb_plot_infos), (
            self.side_band_key,
            [sbi.key for sbi, _ in sb_plot_infos],
        )

        excepted_errors = []  # type: List[Tuple[Exception, Optional[Tuple[str, int, str]]]]

        for sb_plot_info_tuple in sb_plot_infos:
            try:
                eval_output = self.produce_sideband_plot_with_check(
                    mc_df=mc_df,
                    data_df=data_df,
                    input_tuple=sb_plot_info_tuple,
                    dry_run=dry_run,
                    overwrite_existing=False,
                )
                assert all(k not in output_dict.keys() for k in eval_output.keys()), [
                    k for k in eval_output.keys() if k in output_dict.keys()
                ]
                assert all(v not in output_dict.values() for v in eval_output.values()), [
                    v for v in eval_output.values() if v in output_dict.values()
                ]
                output_dict.update(eval_output)

            except Exception as e:
                if e.__traceback__:
                    _tb_frame = e.__traceback__.tb_frame
                    _tb_info: Optional[Tuple[str, int, str]] = (
                        _tb_frame.f_code.co_filename,
                        _tb_frame.f_lineno,
                        str(_tb_frame.f_code),
                    )
                else:
                    _tb_info = None
                excepted_errors.append((e, _tb_info))

        if excepted_errors:
            raise Exception(excepted_errors)

        count_offset = sum(len(sb_info.modes_to_skip) for sb_info, _ in sb_plot_infos if sb_info.modes_to_skip)
        assert (len(output_dict) + count_offset) % len(sb_plot_infos) == 0, (len(output_dict), len(sb_plot_infos))
        return output_dict

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_sideband_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_sideband_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def complete(self):
        # Catching special case of a SidebandEvaluationTask which cannot produce output.
        if isinstance(self.output(), dict) and len(self.output()) == 0:
            logger.warning(
                f"SidebandEvaluationTask cannot produce output for "
                f"selection name {self.selection_name} and sideband key {self.side_band_key}!"
            )
            return True
        return super().complete()

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )
