"""
Tasks performing a evaluation of the impact of cuts on a given variable on a set of other variables.

Felix Metzner 2021
"""

import os
import pandas as pd
import b2luigi as luigi

from collections import abc
from luigi.util import inherits

from typing import Optional, List, Tuple, Dict, Generator

from rdstar.utilities import PathType
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.validation.sideband_validation import SidebandInfo
from rdstar.offline_analysis.validation.variable_impact_validation import (
    VariableCutImpactValidator,
    ValidationVariableSetups,
)

from rdstar.offline_analysis.tasks import utilities
from rdstar.offline_analysis.weight_management.weight_information import WeightManager
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler
from rdstar.offline_analysis.tasks.resolution_correction_task import ResolutionCorrectionTask
from rdstar.offline_analysis.tasks.resolution_evaluation_tasks import ResolutionEvaluationManagerTask


__all__ = [
    "VariableImpactValidationTask",
    "VariableImpactValidationWrapperTask",
]


@inherits(ResolutionEvaluationManagerTask)
class VariableImpactValidationWrapperTask(luigi.WrapperTask):

    relevant_selections = SelectionSetups.rdstar_selections  # type: Tuple[str, ...]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        assert self.selection_name in self.relevant_selections, (self.selection_name, self.relevant_selections)

    def get_log_file_dir(self) -> PathType:
        return self.working_dir

    def get_task_file_dir(self) -> PathType:
        return self.working_dir

    @staticmethod
    def sideband_info_relevant_for_selection(sideband_info: SidebandInfo, selection_name: str) -> bool:
        sb_info_selection_req = sideband_info.requires_selection  # type: Optional[Tuple[str, ...]]
        if sb_info_selection_req is not None:
            if selection_name not in sb_info_selection_req:
                return False
        return True

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        for sample_setup_key in SampleSetupHandler.on_resonance_sample_setup_names:
            for validation_variable_setup_name in ValidationVariableSetups.get_setup_mapping().keys():
                sideband_mapping = VariableCutImpactValidator.get_sideband_infos_by_key_mapping(
                    validation_setup_name=validation_variable_setup_name
                )  # type: Dict[str, List[SidebandInfo]]
                assert all(sample_setup_key in vv.valid_sample_setups for v in sideband_mapping.values() for vv in v), (
                    sample_setup_key,
                    validation_variable_setup_name,
                    [
                        (k, vv.hist_var.df_label)
                        for k, v in sideband_mapping.items()
                        for vv in v
                        if sample_setup_key not in vv.valid_sample_setups
                    ],
                )

                for side_band_key in sideband_mapping.keys():
                    yield self.clone(
                        VariableImpactValidationTask,
                        validation_variable_setup_name=validation_variable_setup_name,
                        sample_setup_key=sample_setup_key,
                        side_band_key=side_band_key,
                    )


@inherits(VariableImpactValidationWrapperTask)
class VariableImpactValidationTask(utilities.BaseOfflineTask):
    validation_variable_setup_name = luigi.Parameter(positional=False)  # type: str
    sample_setup_key = luigi.Parameter(positional=False)  # type: str
    side_band_key = luigi.Parameter(positional=False)  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.base_path = os.path.join(
            self.working_dir,
            VariableCutImpactValidator.base_dir_name,
            self.selection_name,
            self.sample_setup_key,
        )  # type: PathType

        assert isinstance(self.config_combinations, abc.Sequence)
        cs_tag_from_config = self.config_combinations[0][2][0]["cs_tag"]
        assert cs_tag_from_config == self.cs_tag, (cs_tag_from_config, self.cs_tag)
        assert all(all(c["cs_tag"] == self.cs_tag for c in cs) for _, _, cs in self.config_combinations)

        self.data_sample_key = SampleSetupHandler.get_data_sample_key(sample_name=self.sample_setup_key)  # type: str
        self.mc_sample_keys = SampleSetupHandler.get_mc_sample_keys(
            sample_name=self.sample_setup_key
        )  # type: Tuple[str, ...]

        self._input_dict = None  # type: Optional[Dict[str, PathType]]

    @property
    def input_dict(self) -> Dict[str, PathType]:
        if self._input_dict is None:
            self._input_dict = {k: v.path for d in self.input() if isinstance(d, dict) for k, v in d.items()}
        return self._input_dict

    def get_mc_and_data_file_paths(self) -> Tuple[List[PathType], PathType]:
        selection_sideband_id = f"{self.selection_name}_with_CS_{self.cs_tag}_{self.side_band_key}"  # type: str
        input_appendix = f"{WeightManager.finalized_weights_appendix}_{ResolutionCorrectionTask.appendix}"  # type: str
        full_key_end = f"{input_appendix}_{selection_sideband_id}"  # type: str

        mc_paths = [
            p
            for mc_key in self.mc_sample_keys
            for k, p in self.input_dict.items()
            if k == f"{mc_key}_{full_key_end}" and os.path.splitext(p)[1] == ".h5"
        ]  # type: List[PathType]

        assert len(mc_paths) == len(self.mc_sample_keys), (len(mc_paths), mc_paths, self.mc_sample_keys, full_key_end)

        data_key_str = f"{self.data_sample_key}_{full_key_end}"  # type: str
        data_paths = [p for k, p in self.input_dict.items() if k == data_key_str and os.path.splitext(p)[1] == ".h5"]
        assert len(data_paths) == 1, data_paths
        data_path = data_paths[0]

        return mc_paths, data_path

    def get_mc_and_data_sample(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        mc_paths, data_path = self.get_mc_and_data_file_paths()  # type: List[PathType], PathType

        mc_sample = pd.concat([pd.read_hdf(mc_path) for mc_path in mc_paths], ignore_index=True, sort=False)
        assert isinstance(mc_sample, pd.DataFrame)
        data_sample = pd.read_hdf(data_path)
        assert isinstance(data_sample, pd.DataFrame)

        return mc_sample, data_sample

    def do_variable_impact_validation(self, dry_run: bool = False) -> Dict[str, PathType]:
        validator = VariableCutImpactValidator(
            validation_variable=ValidationVariableSetups.get_setup(setup_name=self.validation_variable_setup_name),
            result_dir_path=self.base_path,
        )

        mc_df, data_df = (None, None)  # type: Optional[pd.DataFrame], Optional[pd.DataFrame]
        if not dry_run:
            mc_df, data_df = self.get_mc_and_data_sample()

        return validator.perform_impact_validation(
            mc_sample=mc_df,
            data_sample=data_df,
            dry_run=dry_run,
        )

    def run(self) -> None:
        utilities.host_check(hosts=utilities.GlobalParameters().work_host_list)
        os.makedirs(self.base_path, exist_ok=True)
        self.do_variable_impact_validation()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = self.do_variable_impact_validation(dry_run=True)
        return {output_key: luigi.LocalTarget(target_path) for output_key, target_path in output_dict.items()}

    def requires(self) -> Generator[utilities.TaskType, None, None]:
        assert isinstance(self.config_combinations, abc.Sequence)
        for combo_name, data_types, relevant_configs in self.config_combinations:
            yield ResolutionCorrectionTask(
                combo_name=combo_name,
                working_dir=self.working_dir,
                selection_name=self.selection_name,
                cs_tag=self.cs_tag,
                config_params=relevant_configs,
                config_combinations=self.config_combinations,
            )
