"""
Pre Selection Validator
    Will perform data-MC comparisons on a subset of the data sampled prior to the
    continuum suppression application and selection.
    This allows to evaluate the consistency of data and MC for observables used for the
    training of CS and Selection MVAs and the Selection Variables in general.

Felix Metzner 2021
"""


import os
import copy
import pandas as pd

from typing import NamedTuple, Union, Optional, List, Tuple, Dict

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.utilities import PathType
from rdstar.mva_methods.mva_methods import MVAVarInfo
from rdstar.cut_information_objects.cut_info import CutInfo
from rdstar.mva_selection.mva_selector_setups import MVASelectorSetupBase
from rdstar.continuum_suppression.continuum_suppression import get_cs_variable_dict
from rdstar.comparison_tools.data_mc_comparer import DataMCComparer, DataMCComparerInfo

from rdstar.offline_analysis.selection.selection_setups import SelectionSetups


__all__ = [
    "PreSelectionValidator",
]


class ObservableInfo(NamedTuple):
    df_label: str
    latex_label: str
    unit: Optional[str]
    n_bins: int = 10
    scope: Optional[Tuple[float, float]] = None

    def as_hist_var(self) -> HistVariable:
        return HistVariable(
            df_label=self.df_label,
            n_bins=self.n_bins,
            scope=self.scope,
            var_name=self.latex_label,
            unit=self.unit,
            use_log_scale=False,
        )

    @staticmethod
    def from_hist_var(hist_var: HistVariable) -> "ObservableInfo":
        return ObservableInfo(
            df_label=hist_var.df_label,
            latex_label=hist_var.variable_name,
            unit=hist_var.unit,
            n_bins=hist_var.n_bins,
            scope=hist_var.scope,
        )

    @staticmethod
    def from_cut_info(cut_info: CutInfo) -> "ObservableInfo":
        return ObservableInfo(
            df_label=cut_info.cut_var_root,
            latex_label=cut_info.cut_var_latex_label,
            unit=cut_info.latex_unit,
        )

    @staticmethod
    def from_mva_var_info(mva_var_info: MVAVarInfo) -> "ObservableInfo":
        return ObservableInfo(
            df_label=mva_var_info.var,
            latex_label=mva_var_info.latex,
            unit=mva_var_info.unit,
            n_bins=mva_var_info.bins,
            scope=(mva_var_info.min, mva_var_info.max),
        )


main_sel_name = SelectionSetups.used_selection_name  # type: str
_main_sel_mva_info = SelectionSetups.get_selection_mva_info(selection_name=main_sel_name)
assert _main_sel_mva_info is not None
main_sel_mva_setup = _main_sel_mva_info.mva_setup  # type: MVASelectorSetupBase
main_sel_mva_var_infos = main_sel_mva_setup.all_mva_variable_infos  # type: Tuple[MVAVarInfo, ...]
_mva_var_df_labels = [mva_i.var for mva_i in main_sel_mva_var_infos]  # type: List[str]

main_sel_bcs_infos = SelectionSetups.available_selections[main_sel_name].bcs_hist_vars  # type: List[HistVariable]

# Getting variables form selection cuts, but not adding the ones for which better versions are available!
_main_sel_cut_infos = SelectionSetups.available_selections[main_sel_name].cut_infos_for_validation  # type: List[CutInfo]
main_sel_cut_infos = [ci for ci in _main_sel_cut_infos if ci.cut_var not in _mva_var_df_labels]  # type: List[CutInfo]

observable_info_collection = (
    *[ObservableInfo.from_cut_info(cut_info=c) for c in main_sel_cut_infos],
    *[ObservableInfo.from_hist_var(hist_var=hv) for hv in main_sel_bcs_infos],
    *[ObservableInfo.from_mva_var_info(mva_var_info=mvi) for mvi in main_sel_mva_var_infos],
    *[ObservableInfo.from_hist_var(hist_var=hv) for hv in get_cs_variable_dict(signal_side=False).values()],
)  # type: Tuple[ObservableInfo, ...]


__all_main_sel_cut_strs = [_ci.cut_str for _ci in main_sel_cut_infos]  # type: List[str]
main_sel_subset_check_cut_stings = (
    "ROE_eextra__boCleanROE__bc < 3.",
    "abs(daughter__bo1__cm__spdaughter__bo0__cm__spdM__bc__bc) < 0.035",
    "daughter__bo0__cm__spMbc__bc > 5.27",
    "daughter__bo1__cm__spdaughter__bo1__cm__spp__bc__bc < 2.",
    "daughter__bo1__cm__spchiProb__bc > 0.0",
)
assert all(cs in __all_main_sel_cut_strs for cs in main_sel_subset_check_cut_stings), [
    c for c in main_sel_subset_check_cut_stings if c not in __all_main_sel_cut_strs
]


class PreSelectionValidator:
    plot_info_text = "Pre-Selection"  # type: str
    data_comparer_key = "pre_selection"

    default_fig_size = (6.5, 5.0)  # type: Tuple[float, float]

    subset_selection_name = main_sel_name  # type: str
    _main_sel_mva_setup = main_sel_mva_setup  # type: MVASelectorSetupBase
    _all_observable_infos = observable_info_collection  # type: Tuple[ObservableInfo, ...]
    subset_check_cut_stings = main_sel_subset_check_cut_stings  # type: Tuple[str, ...]

    def __init__(
        self,
        weight_column: str,
        export_path: PathType,
        adaptive_binning: bool = False,
        fig_size: Optional[Tuple[float, float]] = None,
    ) -> None:
        self._weight_column = weight_column  # type: str
        self._export_path = export_path  # type: PathType
        self._adaptive_binning = adaptive_binning  # type: bool
        self._fig_size = fig_size if fig_size is not None else self.default_fig_size  # type: Tuple[float, float]

    @property
    def comparer_infos(self) -> List[DataMCComparerInfo]:
        return [
            DataMCComparerInfo(
                key=self.data_comparer_key,
                hist_var=observable_info.as_hist_var(),
                title=observable_info.latex_label,
                info_text=self.plot_info_text,
                lumi_label=r"$5\,\%$ of " + DataMCComparerInfo.get_default_lumi_label(),
                legend_loc=(2, 2),
                legend_cols=(1, 2),
                text_v_pos=(0.97, 0.97),
                text_h_pos=(0.72, 0.7),
                y_scale=(1.2, 1.96),
                split_chi_info_text=True,
            )
            for observable_info in self._all_observable_infos
        ]

    def create_plots(
        self,
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        output_dict = dict()  # type: Dict[str, PathType]

        if not dry_run:
            self._main_sel_mva_setup.data_preparation(df=mc_df, for_application_only=True)
            self._main_sel_mva_setup.data_preparation(df=data_df, for_application_only=True)

        for comparer_info in self.comparer_infos:
            _output_dir_path = os.path.join(self._export_path, comparer_info.hist_var.df_label)

            _comparer_info = comparer_info  # type: DataMCComparerInfo
            if not comparer_info.hist_var.has_scope():
                if dry_run:
                    _hist_var_with_scope = HistVariable(
                        df_label=comparer_info.hist_var.df_label,
                        n_bins=comparer_info.hist_var.n_bins,
                        scope=(0.0, 1.0),
                        var_name=comparer_info.hist_var.variable_name,
                        unit=comparer_info.hist_var.unit if comparer_info.hist_var.unit else None,
                        use_log_scale=comparer_info.hist_var.use_log_scale,
                    )  # type: HistVariable
                else:
                    os.makedirs(_output_dir_path, exist_ok=True)

                    _hist_var_with_scope = HistVariable.get_scoped_histogram_variable(
                        base_hist_var=comparer_info.hist_var,
                        dfs=[mc_df, data_df],
                        round_scope_precision=0,
                        check_for_nans=True,
                    )

                _comparer_info = DataMCComparerInfo.create_modified_comparer_info(
                    comparer_info=comparer_info,
                    hist_var=_hist_var_with_scope,
                )

            comparer = DataMCComparer(comparer_info=_comparer_info)

            for style in comparer_info.styles:

                new_output_dict = self.create_validation_plot(
                    comparer=comparer,
                    mc_df=mc_df,
                    data_df=data_df,
                    output_dir_path=_output_dir_path,
                    style=style,
                    dry_run=True,
                )  # type: Dict[str, PathType]

                # Create plots for this style - comparer combination only if they do not exist already:
                if not dry_run and not all(os.path.exists(p) for p in new_output_dict.values()):
                    new_output_dict = self.create_validation_plot(
                        comparer=comparer,
                        mc_df=mc_df,
                        data_df=data_df,
                        output_dir_path=_output_dir_path,
                        style=style,
                        dry_run=dry_run,
                    )

                assert new_output_dict is not None
                assert all(k not in output_dict.keys() for k in new_output_dict.keys()), (
                    [k for k in new_output_dict.keys() if k in output_dict.keys()],
                    len(output_dict.keys()),
                )
                assert all(v not in output_dict.values() for v in new_output_dict.values()), (
                    [(k, v) for k, v in new_output_dict.items() if v in output_dict.values()],
                    len(output_dict.keys()),
                )
                output_dict.update(new_output_dict)

        return output_dict

    def create_validation_plot(
        self,
        comparer: DataMCComparer,
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        output_dir_path: PathType,
        style: str,
        dry_run: bool,
    ) -> Dict[str, PathType]:

        default_parameters = {
            "gof_check_method": "toys",
            "adaptive_binning": self._adaptive_binning,
        }  # type: Dict[str, Union[str, bool]]

        alternative_options = {
            "gof_check_method": "pearson",
            "adaptive_binning": False,
        }  # type: Dict[str, Union[str, bool]]

        _used_params = copy.copy(default_parameters)  # type: Dict[str, Union[str, bool]]

        if alternative_options["adaptive_binning"] == self._adaptive_binning:
            del alternative_options["adaptive_binning"]

        new_output_dict = {}  # type: Dict[str, PathType]
        excepted_errors = []  # type: List[Exception]
        try_to_plot = True  # type: bool
        while try_to_plot:
            try_to_plot = False
            try:
                _used_gof_check_method_option = _used_params["gof_check_method"]  # type: Union[str, bool]
                _used_adaptive_binning_option = _used_params["adaptive_binning"]  # type: Union[str, bool]
                assert isinstance(_used_gof_check_method_option, str), type(_used_gof_check_method_option).__name__
                assert isinstance(_used_adaptive_binning_option, bool), type(_used_adaptive_binning_option).__name__
                used_gof_check_method_option = _used_gof_check_method_option  # type: str
                used_adaptive_binning_option = _used_adaptive_binning_option  # type: bool

                new_output_dict = self._run_plotting(
                    comparer=comparer,
                    mc_df=mc_df,
                    data_df=data_df,
                    output_dir_path=output_dir_path,
                    gof_check_method=used_gof_check_method_option,
                    adaptive_binning=used_adaptive_binning_option,
                    style=style,
                    dry_run=dry_run,
                )
            except Exception as e:
                excepted_errors.append(e)
                if len(alternative_options.keys()) > 0:
                    try_to_plot = True
                    next_alternative_key = list(alternative_options.keys())[0]
                    _used_params.update({next_alternative_key: alternative_options.pop(next_alternative_key)})
                else:
                    raise Exception(excepted_errors)

        return new_output_dict

    def _run_plotting(
        self,
        comparer: DataMCComparer,
        mc_df: pd.DataFrame,
        data_df: pd.DataFrame,
        output_dir_path: PathType,
        style: str,
        gof_check_method: str,
        adaptive_binning: bool,
        dry_run: bool,
    ) -> Dict[str, PathType]:
        output_dict = comparer.plot_data_mc_comparison(
            mc_df=mc_df,
            data_df=data_df,
            weight_col=self._weight_column,
            lumi_label=True,
            draw_legend=True,
            legend_inside=True,
            export_path=output_dir_path,
            style=style,
            ratio_type="vs_uncert",
            fig_size=self._fig_size,
            normalize_mc_to_data=True,  # TODO: Add option to scale mc samples with fit result!
            gof_check_method=gof_check_method,
            adaptive_binning=adaptive_binning,
            dry_run=dry_run,
            show_plots=False,
        )  # type: Optional[Dict[str, PathType]]
        assert output_dict is not None
        return output_dict
