"""
Sideband Validation
    - Sideband Plot Definitions in sideband_info_collection
    - Plotting Function

Felix Metzner 2019
"""

import numpy as np

from dataclasses import dataclass
from typing import Tuple, List, Optional

from templatefitter.plotter.histogram_variable import HistVariable

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames
from rdstar.latex_labels.latex_particle_names import LatexParticleNames

from rdstar.mva_methods.mva_methods import MVAVarInfo
from rdstar.continuum_suppression.continuum_suppression import cs_vars_tag_side_name_space

from rdstar.comparison_tools.select_utilities import SimpleCutInfo
from rdstar.comparison_tools.data_mc_comparer import DataMCComparerInfo

from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos
from rdstar.offline_analysis.reco_modes.reco_mode_collections import BtoDssLRecoModes
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollections

from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.data_handling.selected_data_handling import DataSampleLoader
from rdstar.offline_analysis.systematics import full_systematics_cols, full_systematics_name
from rdstar.offline_analysis.data_handling.sample_setup_definitions import SampleSetupHandler
from rdstar.offline_analysis.mms_resolution_correction.mm2_shift_fit import MCShiftInfo, first_test_shift_info
from rdstar.offline_analysis.validation.correlation_calculator import CorrCalcInfo, corr_eval_info_additional_info_key


__all__ = [
    "SidebandInfo",
    "sideband_info_collection",
    "sideband_info_collection_plus_with_full_sys",
    "q2_sideband_title",
    "q2_sideband_text",
    "sideband_data_component_label",
    "cs_mva_sideband_title",
    "cs_mva_sideband_text",
    "cs_mva_anti_cut_selection",
    "mbc_sideband_title",
    "mbc_sideband_text",
    "dss_mms_selection",
    "dss_mms_tight_selection",
    "dss_mms_tight_m02_04_selection",
    "dss_mms_control_m02_08_selection",
    "dss_mms_control_m02_20_selection",
    "d_w_wo_star_pi_minus_d_w_wo_star_mass_selection",
]


class ModesToSkipInfo:
    b_to_d_reco_mode_names: List[str] = [
        RecoModeInfos.neutral_b_to_d_modes.name,
        RecoModeInfos.charged_b_to_d_modes.name,
    ]
    b_to_d_st_reco_mode_names: List[str] = [
        RecoModeInfos.neutral_b_to_ds_modes.name,
        RecoModeInfos.charged_b_to_ds_modes.name,
    ]


@dataclass(frozen=True)
class SidebandInfo(DataMCComparerInfo):
    @property
    def sideband_sample_key(self) -> str:
        return self.key

    @property
    def systematics_name_tag(self) -> str:
        if self.sys_name_tag is None:
            assert self.systematics is None
            return "no_sys"
        else:
            assert self.systematics is not None, self.sys_name_tag
            return self.sys_name_tag


modified_e_extra_hist_var: Optional[HistVariable] = None
modified_e_extra_binning: Tuple[float, ...] = tuple(
    round(x, 2) if not np.isclose(x, 0.6) else 0.6001
    for x in np.linspace(*RDStarHistVars.eextra.scope, 16)
    if x >= 0.6 or x == 0.0
)
__used_selection_mva_info = SelectionSetups.get_selection_mva_info(selection_name=SelectionSetups.used_selection_name)
if __used_selection_mva_info is not None:
    __used_selection_mva_setup = __used_selection_mva_info.mva_setup
    __e_extra_mva_info: Optional[MVAVarInfo] = getattr(__used_selection_mva_setup, "e_extra", None)
    if __e_extra_mva_info is not None:
        assert isinstance(__e_extra_mva_info, MVAVarInfo), type(__e_extra_mva_info)
        modified_e_extra_hist_var = HistVariable(
            df_label=__e_extra_mva_info.var,
            n_bins=RDStarHistVars.eextra.n_bins,
            scope=RDStarHistVars.eextra.scope,
            var_name=RDStarHistVars.eextra.variable_name,
            unit=RDStarHistVars.eextra.unit if RDStarHistVars.eextra.unit else None,
        )

sideband_data_component_label: str = "Sideband Data"

# q^2 Sideband
q2_sideband_title: str = rf"{LatexObsNames.fitting_q2} Sideband"
q2_sideband_text: str = LatexObsNames.fitting_q2[:-1] + r" < 4.0" + LatexUnits.gev_sq[1:]

_q2_sideband_info_collection: List[SidebandInfo] = [
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.neextra_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=RDStarHistVars.neextra_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=RDStarHistVars.neextra_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS",
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0",
        hist_var=RDStarHistVars.neextra_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_random_bcs",
        hist_var=RDStarHistVars.neextra_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_v_pos=(0.88, 0.62),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_no_bcs",
        hist_var=RDStarHistVars.neextra_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_v_pos=(0.88, 0.62),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.neextra_zoom_in_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.88, 0.64),
        legend_loc=(1, 9),
        y_scale=(1.1, 1.3),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=RDStarHistVars.neextra_zoom_in_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.88, 0.64),
        legend_loc=(1, 9),
        y_scale=(1.1, 1.3),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=RDStarHistVars.neextra_zoom_in_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.88, 0.64),
        legend_loc=(1, 9),
        y_scale=(1.1, 1.3),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0",
        hist_var=RDStarHistVars.neextra_zoom_in_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.88, 0.64),
        legend_loc=(1, 9),
        y_scale=(1.1, 1.3),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_random_bcs",
        hist_var=RDStarHistVars.neextra_zoom_in_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.88, 0.64),
        legend_loc=(1, 9),
        y_scale=(1.1, 1.3),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_no_bcs",
        hist_var=RDStarHistVars.neextra_zoom_in_fine,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.88, 0.64),
        legend_loc=(1, 9),
        y_scale=(1.1, 1.3),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.eextra,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=RDStarHistVars.eextra,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=RDStarHistVars.eextra,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0",
        hist_var=RDStarHistVars.eextra,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_random_bcs",
        hist_var=RDStarHistVars.eextra,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.1, 1.6),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_no_bcs",
        hist_var=RDStarHistVars.eextra,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        legend_loc=(1, 2),
        legend_cols=(2, 3),
        y_scale=(1.1, 1.6),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.eextra_zoom_in,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.18, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.6),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=RDStarHistVars.eextra_zoom_in,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=(0.18, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.6),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=RDStarHistVars.eextra_zoom_in,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=(0.18, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.6),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0",
        hist_var=RDStarHistVars.eextra_zoom_in,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.18, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.6),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_random_bcs",
        hist_var=RDStarHistVars.eextra_zoom_in,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.18, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.6),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_0_no_bcs",
        hist_var=RDStarHistVars.eextra_zoom_in,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"No cut on {LatexObsNames.roe_e_ecl_extra}",
        text_h_pos=(0.18, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.6),
        split_chi_info_text=(False, True),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="zoom_in",
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.mva_output_for_sbv_var,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        y_scale=(1.1, 1.6),
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        file_name_appendix="loose_cut",
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.cs_mva_output,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=RDStarHistVars.cs_mva_output,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=RDStarHistVars.cs_mva_output,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_2_random_bcs",
        hist_var=RDStarHistVars.loose_cs_mva_output,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"Loose cut on {LatexObsNames.mva_cs_output}",
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_2_no_bcs",
        hist_var=RDStarHistVars.loose_cs_mva_output,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"Loose cut on {LatexObsNames.mva_cs_output}",
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.sig_vertex_fit_prob,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.78),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.03, 0.66),
        text_v_pos=(0.97, 0.78),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 2.2),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.03, 0.66),
        text_v_pos=(0.97, 0.78),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 2.2),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="with_cs_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="cs_mva_output",
                var_latex_label=LatexObsNames.mva_cs_output,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_3",
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\n" + rf"No cut on {LatexObsNames.lepton_p}",
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.8),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 2.2),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_3_random_bcs",
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"No cut on {LatexObsNames.lepton_p}",
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.8),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.4, 2.6),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_3_no_bcs",
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"No cut on {LatexObsNames.lepton_p}",
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.8),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.4, 2.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.missing_mass_q2_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=RDStarHistVars.missing_mass_q2_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=RDStarHistVars.missing_mass_q2_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix=f"with_{first_test_shift_info.name}_mm2_shift_with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
            MCShiftInfo.extra_info_key(): first_test_shift_info,
        },
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="with_cs_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="cs_mva_output",
                var_latex_label=LatexObsNames.mva_cs_output,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS",
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.e_miss_minus_p_miss,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.76, 0.62),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.e_miss_minus_p_miss_cms,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.76, 0.62),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.mbc_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_v_pos=(0.97, 0.78),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.2, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1",
        hist_var=RDStarHistVars.mbc_full_range_normal_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.2, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1_random_bcs",
        hist_var=RDStarHistVars.mbc_full_range_normal_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.25, 1.7),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1_no_bcs",
        hist_var=RDStarHistVars.mbc_full_range_normal_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.25, 1.7),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1",
        hist_var=RDStarHistVars.mbc_full_range_rough_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.2, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="rough_binning",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1_random_bcs",
        hist_var=RDStarHistVars.mbc_full_range_rough_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.25, 1.7),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="rough_binning",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1_no_bcs",
        hist_var=RDStarHistVars.mbc_full_range_rough_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.25, 1.7),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="rough_binning",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1",
        hist_var=RDStarHistVars.mbc_full_range_fine_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.2, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="fine_binning",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1_random_bcs",
        hist_var=RDStarHistVars.mbc_full_range_fine_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.25, 1.7),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="fine_binning",
    ),
    SidebandInfo(
        key="q2_sideband_without_cut_1_no_bcs",
        hist_var=RDStarHistVars.mbc_full_range_fine_binning_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS\n" + rf"No cut on {LatexObsNames.b_tag_mbc}",
        text_v_pos=(0.75, 0.66),
        text_h_pos=(0.03, 0.03),
        legend_loc=2,
        legend_cols=(1, 3),
        y_scale=(1.25, 1.7),
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="fine_binning",
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.tag_delta_e,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_no_bcs",
        hist_var=RDStarHistVars.tag_delta_e,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nNo BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=RDStarHistVars.tag_delta_e,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.d_mass_for_sb_validation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.72, 0.73),
        y_scale=(1.2, 1.96),
        split_chi_info_text=True,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.d_mass_for_sb_validation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.72, 0.73),
        y_scale=(1.2, 1.96),
        split_chi_info_text=True,
        systematics=None,
        data_component_label=sideband_data_component_label,
        modes_to_skip=ModesToSkipInfo.b_to_d_reco_mode_names,
        file_name_appendix="only_d_mesons",
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.d_mass_for_sb_validation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.72, 0.73),
        y_scale=(1.2, 1.96),
        split_chi_info_text=True,
        systematics=None,
        data_component_label=sideband_data_component_label,
        modes_to_skip=ModesToSkipInfo.b_to_d_st_reco_mode_names,
        file_name_appendix="only_d_star_mesons",
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.d_star_daughter_mass_for_sb_validation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.72, 0.73),
        y_scale=(1.2, 1.96),
        split_chi_info_text=True,
        systematics=None,
        data_component_label=sideband_data_component_label,
        modes_to_skip=ModesToSkipInfo.b_to_d_reco_mode_names,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spR2__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spthrustBm__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.7),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spthrustOm__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.7),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spcosTBTO__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.03, 0.73),
        y_scale=(1.1, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spcosTBz__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.3, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo1__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo2__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo3__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo4__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo5__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo6__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo7__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo8__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo9__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        # hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__boet__bc__bc,
        hist_var=RDStarHistVars.special_cs_b_tag_ksfw_et,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.74),
        text_h_pos=(0.68, 0.68),
        y_scale=(1.3, 1.9),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        # hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo0__bc__bc,
        hist_var=RDStarHistVars.special_cs_b_tag_ksfw_hoo0,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.74),
        text_h_pos=(0.68, 0.68),
        y_scale=(1.3, 1.9),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo1__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo2__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo3__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo4__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso00__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.7),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso02__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso04__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso10__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso12__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso14__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(  # TODO: Optimize again
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso20__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.72),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso22__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso24__bc__bc,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bomm2__bc__bc,
        # hist_var=RDStarHistVars.special_cs_b_tag_ksfw_mm2,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
]

if modified_e_extra_hist_var:
    modified_e_extra_random_bcs_q2_sideband_info = SidebandInfo(
        key="q2_sideband_random_bcs",
        hist_var=modified_e_extra_hist_var,
        title=q2_sideband_title,
        info_text=q2_sideband_text + "\nRandom BCS",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.8, 0.66),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.1, 1.6),
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
        additional_info={
            "adaptive_binning": False,
            "special_binning": modified_e_extra_binning,
            # "gof_check_method": "pearson",
        },
    )
    _q2_sideband_info_collection.append(modified_e_extra_random_bcs_q2_sideband_info)

    modified_e_extra_q2_sideband_info = SidebandInfo(
        key="q2_sideband",
        hist_var=modified_e_extra_hist_var,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.8, 0.66),
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        y_scale=(1.1, 1.6),
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
        additional_info={
            "adaptive_binning": False,
            "special_binning": modified_e_extra_binning,
            # "gof_check_method": "pearson",
        },
    )
    _q2_sideband_info_collection.append(modified_e_extra_q2_sideband_info)


_q2_with_fei_corr_sideband_info_collection = (
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.missing_mass_q2_SidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="with_FEI_corr",
        add_new_fei_calibration_factors=True,
    ),
    SidebandInfo(
        key="q2_sideband",
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=q2_sideband_title,
        info_text=q2_sideband_text,
        text_h_pos=(0.64, 0.03),
        text_v_pos=(0.76, 0.97),
        legend_loc=1,
        legend_cols=1,
        y_scale=1.1,
        systematics=None,
        data_component_label=sideband_data_component_label,
        file_name_appendix="with_FEI_corr_with_sel_mva_corr",
        add_new_fei_calibration_factors=True,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
)


# Tag M_bc Sideband
mbc_sideband_title: str = rf"{LatexObsNames.b_tag_mbc} Sideband"
mbc_sideband_text: str = LatexObsNames.b_tag_mbc[:-1] + r" \leq 5.27\," + LatexUnits.gev[1:]

_tag_mbc_sideband_info_collection: List[SidebandInfo] = [
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.neextra_fine,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband_no_bcs",
        hist_var=RDStarHistVars.neextra_fine,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text + "\nNo BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband_random_bcs",
        hist_var=RDStarHistVars.neextra_fine,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text + "\nRandom BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(  # TODO Optimize
        key="mbc_sideband",
        hist_var=RDStarHistVars.eextra,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.2, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(  # TODO Optimize
        key="mbc_sideband_random_bcs",
        hist_var=RDStarHistVars.eextra,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text + "\nRandom BCS",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.2, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(  # TODO Optimize
        key="mbc_sideband_no_bcs",
        hist_var=RDStarHistVars.eextra,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text + "\nNo BCS",
        text_h_pos=(0.66, 0.68),
        text_v_pos=(0.88, 0.66),
        y_scale=(1.2, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.mbc_in_mbcsb_SidebandValidation,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.7),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.6),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.mbc_in_mbcsb_SidebandValidation_rough_binning,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.7),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.6),
        systematics=None,
        file_name_appendix="rough_binning",
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.sig_vertex_fit_prob,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.d_mass_for_sb_validation,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.72, 0.73),
        y_scale=(1.2, 1.96),
        split_chi_info_text=True,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.d_mass_for_sb_validation,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.72, 0.73),
        y_scale=(1.2, 1.96),
        split_chi_info_text=True,
        systematics=None,
        data_component_label=sideband_data_component_label,
        modes_to_skip=ModesToSkipInfo.b_to_d_reco_mode_names,
        file_name_appendix="only_d_mesons",
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.d_mass_for_sb_validation,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.72, 0.73),
        y_scale=(1.2, 1.96),
        split_chi_info_text=True,
        systematics=None,
        data_component_label=sideband_data_component_label,
        modes_to_skip=ModesToSkipInfo.b_to_d_st_reco_mode_names,
        file_name_appendix="only_d_star_mesons",
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.d_star_daughter_mass_for_sb_validation,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.72, 0.73),
        y_scale=(1.2, 1.96),
        split_chi_info_text=True,
        systematics=None,
        data_component_label=sideband_data_component_label,
        modes_to_skip=ModesToSkipInfo.b_to_d_reco_mode_names,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.mva_output_for_sbv_var,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(2, 3),
        text_h_pos=0.03,
        text_v_pos=(0.97, 0.7),
        y_scale=(1.4, 2.16),
        split_chi_info_text=False,
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(2, 3),
        text_h_pos=0.03,
        text_v_pos=(0.97, 0.7),
        y_scale=(1.4, 2.06),
        split_chi_info_text=False,
        systematics=None,
        file_name_appendix="loose_cut",
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=RDStarHistVars.cs_mva_output,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spR2__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spthrustBm__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.7),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spthrustOm__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.7),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spcosTBTO__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        text_v_pos=(0.97, 0.97),
        text_h_pos=(0.03, 0.73),
        y_scale=(1.1, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spcosTBz__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.3, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo1__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo2__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo3__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo4__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo5__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo6__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo7__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo8__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spCleoCone__bo9__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        # hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__boet__bc__bc,
        hist_var=RDStarHistVars.special_cs_b_tag_ksfw_et,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.74),
        text_h_pos=(0.68, 0.68),
        y_scale=(1.3, 1.9),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        # hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo0__bc__bc,
        hist_var=RDStarHistVars.special_cs_b_tag_ksfw_hoo0,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.74),
        text_h_pos=(0.68, 0.68),
        y_scale=(1.3, 1.9),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo1__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo2__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo3__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohoo4__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso00__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.7),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.4),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso02__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso04__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso10__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso12__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso14__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(  # TODO: Optimize again
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso20__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.97, 0.72),
        text_h_pos=(0.03, 0.03),
        y_scale=(1.1, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso22__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bohso24__bc__bc,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="mbc_sideband",
        hist_var=cs_vars_tag_side_name_space.daughter__bo0__cm__spKSFWVariables__bomm2__bc__bc,
        # hist_var=RDStarHistVars.special_cs_b_tag_ksfw_mm2,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.5),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
]

if modified_e_extra_hist_var:
    modified_e_extra_random_bcs_tag_mbc_sideband_info = SidebandInfo(
        key="mbc_sideband_random_bcs",
        hist_var=modified_e_extra_hist_var,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text + "\nRandom BCS",
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
        additional_info={
            "adaptive_binning": False,
            "special_binning": modified_e_extra_binning,
            # "gof_check_method": "pearson",
        },
    )
    _tag_mbc_sideband_info_collection.append(modified_e_extra_random_bcs_tag_mbc_sideband_info)

    modified_e_extra_tag_mbc_sideband_info = SidebandInfo(
        key="mbc_sideband",
        hist_var=modified_e_extra_hist_var,
        title=mbc_sideband_title,
        info_text=mbc_sideband_text,
        legend_loc=(1, 2),
        legend_cols=(1, 3),
        text_v_pos=(0.8, 0.74),
        y_scale=(1.1, 1.4),
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
        additional_info={
            "adaptive_binning": False,
            "special_binning": modified_e_extra_binning,
            # "gof_check_method": "pearson",
        },
    )
    _tag_mbc_sideband_info_collection.append(modified_e_extra_tag_mbc_sideband_info)

# CS MVA Sideband
cs_mva_sideband_title: str = rf"{LatexObsNames.mva_cs_output} Sideband"
cs_mva_sideband_text: str = r"0.2 < $" + LatexObsNames.mva_cs_output[1:-1] + r" < 0.5$"

cs_mva_anti_cut_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"{RDStarHistVars.cs_mva_output.df_label} < 0.5",
    cut_latex_label=LatexObsNames.mva_cs_output[:-1] + r" < 0.5$",
    cut_var_col_name=RDStarHistVars.cs_mva_output.df_label,
)


_cs_mva_sideband_info_collection: List[SidebandInfo] = [
    SidebandInfo(
        key="csmva_sideband",
        hist_var=RDStarHistVars.missing_mass_cs_mva_SidebandValidation,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.97, 0.74),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            "adaptive_binning": False,
            # "gof_check_method": "pearson",
        },
    ),
    SidebandInfo(
        key="csmva_sideband",
        hist_var=RDStarHistVars.missing_mass_res_corr_cs_mva_SidebandValidation,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        text_h_pos=(0.66, 0.66),
        text_v_pos=(0.97, 0.74),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
        additional_info={
            "adaptive_binning": False,
            # "gof_check_method": "pearson",
        },
    ),
    SidebandInfo(
        key="csmva_sideband",
        hist_var=RDStarHistVars.eextra,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        text_h_pos=(0.64, 0.66),
        text_v_pos=(0.97, 0.78),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.2, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="csmva_sideband",
        hist_var=RDStarHistVars.neextra_fine,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        text_h_pos=(0.64, 0.66),
        text_v_pos=(0.97, 0.78),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.2, 1.8),
        systematics=None,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="csmva_sideband",
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        text_h_pos=(0.64, 0.66),
        text_v_pos=(0.97, 0.78),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.2, 1.8),
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="csmva_sideband",
        hist_var=RDStarHistVars.mva_output_for_sbv_var,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        text_h_pos=(0.64, 0.66),
        text_v_pos=(0.97, 0.78),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 2.0),
        split_chi_info_text=True,
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
    ),
    SidebandInfo(
        key="csmva_sideband",
        hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        text_h_pos=(0.64, 0.66),
        text_v_pos=(0.97, 0.78),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 2.0),
        split_chi_info_text=True,
        systematics=None,
        file_name_appendix="loose_cut",
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
    ),
]

if modified_e_extra_hist_var:
    modified_e_extra_csmva_sideband_info = SidebandInfo(
        key="csmva_sideband",
        hist_var=modified_e_extra_hist_var,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        text_h_pos=(0.64, 0.66),
        text_v_pos=(0.97, 0.78),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.2, 1.8),
        systematics=None,
        requires_selection=SelectionSetups.used_selections_tuple,
        data_component_label=sideband_data_component_label,
        additional_info={
            "adaptive_binning": False,
            "special_binning": modified_e_extra_binning,
            # "gof_check_method": "pearson",
        },
    )
    _cs_mva_sideband_info_collection.append(modified_e_extra_csmva_sideband_info)


dss_control_channel_data_component_label: str = "Data"

# D** Control Channel
dss_control_channel_title: str = rf"{LatexParticleNames.d_st_st_str} Control Channels"
dss_control_channel_text: str = ""


modified_full_q2_range_binning: Tuple[float, ...] = (0.0, 1.0, 2.0, 3.0, 4.5, 7.0, 10.0)

d_w_wo_star_pi_minus_d_w_wo_star_mass_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"{Cols.diff_dstar_d_mass} > 0.2",
    cut_latex_label=rf"{LatexObsNames.d_st_st_delta_m[:-1]} > 0.2\,{LatexUnits.gev[1:]}",
    cut_var_col_name=Cols.diff_dstar_d_mass,
)

dss_mms_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"-0.4 < {RDStarHistVars.missing_mass_res_corr.df_label} < 2.4",
    cut_latex_label=rf"{LatexObsNames.fitting_mm2[:-1]} \in (-0.4, 2.4)\,{LatexUnits.gev_sq[1:]}",
    cut_var_col_name=RDStarHistVars.missing_mass_res_corr.df_label,
)

dss_mms_tight_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"-0.4 < {RDStarHistVars.missing_mass_res_corr.df_label} < 0.4",
    cut_latex_label=rf"{LatexObsNames.fitting_mm2[:-1]} \in (-0.4, 0.4)\,{LatexUnits.gev_sq[1:]}",
    cut_var_col_name=RDStarHistVars.missing_mass_res_corr.df_label,
)

dss_mms_tight_m02_04_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"-0.2 < {RDStarHistVars.missing_mass_res_corr.df_label} < 0.4",
    cut_latex_label=rf"{LatexObsNames.fitting_mm2[:-1]} \in (-0.2, 0.4)\,{LatexUnits.gev_sq[1:]}",
    cut_var_col_name=RDStarHistVars.missing_mass_res_corr.df_label,
)

dss_mms_control_m02_08_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"-0.2 < {RDStarHistVars.missing_mass_res_corr.df_label} < 0.8",
    cut_latex_label=rf"{LatexObsNames.fitting_mm2[:-1]} \in (-0.2, 0.8)\,{LatexUnits.gev_sq[1:]}",
    cut_var_col_name=RDStarHistVars.missing_mass_res_corr.df_label,
)

dss_mms_control_m02_20_selection: SimpleCutInfo = SimpleCutInfo(
    cut_str=f"-0.2 < {RDStarHistVars.missing_mass_res_corr.df_label} < 2.0",
    cut_latex_label=rf"{LatexObsNames.fitting_mm2[:-1]} \in (-0.2, 2.0)\,{LatexUnits.gev_sq[1:]}",
    cut_var_col_name=RDStarHistVars.missing_mass_res_corr.df_label,
)


_dss_reco_full_q2_sideband = [
    # Signal q2 Region (q2 > 4.0 GeV^2)
    SidebandInfo(
        key=DataSampleLoader.base_identifier,
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 2.2),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.base_identifier,
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.68, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        y_scale=(1.4, 1.9),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.base_identifier,
        hist_var=RDStarHistVars.q2_reco_in_signal_region,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.64, 0.76),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.56),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
            "adaptive_binning": False,
        },
    ),
    SidebandInfo(  # TODO: Optimize layout
        key=DataSampleLoader.base_identifier,
        hist_var=RDStarHistVars.diff_dstar_d_mass,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.64, 0.76),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.8),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            # corr_eval_info_additional_info_key: CorrCalcInfo(
            #     var_col_name="sec_sel_mva_output",
            #     var_latex_label=LatexObsNames.mva_sel_output_short,
            #     weighted=True,
            #     corr_calc_method="pearson",
            # ),
            "adaptive_binning": False,
        },
    ),
    #
    # Signal q2 Region (q2 > 4.0 GeV^2) and with M(D(*) + pi) - M(D(*)) > 0.2 Selection
    SidebandInfo(
        key=DataSampleLoader.base_identifier,
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 2.2),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_d_plus_pi_mass_selection_with_sel_mva_corr",
        additional_selection_info=d_w_wo_star_pi_minus_d_w_wo_star_mass_selection,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.base_identifier,
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.68, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(2, 2),
        legend_cols=(1, 2),
        y_scale=(1.4, 1.9),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_d_plus_pi_mass_selection_with_sel_mva_corr",
        additional_selection_info=d_w_wo_star_pi_minus_d_w_wo_star_mass_selection,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.base_identifier,
        hist_var=RDStarHistVars.q2_reco_in_signal_region,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.64, 0.72),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.56),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_d_plus_pi_mass_selection_with_sel_mva_corr",
        additional_selection_info=d_w_wo_star_pi_minus_d_w_wo_star_mass_selection,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
            "adaptive_binning": False,
        },
    ),
    #
    # Full q2 Region
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 2.2),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.68, 0.68),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.2, 1.7),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.missing_mass_res_corr_dss_control_SidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.68, 0.68),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.2, 1.7),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="full_range_mm2_with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.q2_reco_full_region,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.64, 0.76),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.56),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
            "adaptive_binning": False,
            "special_binning": modified_full_q2_range_binning,
        },
    ),
    SidebandInfo(  # TODO: Optimize layout
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.diff_dstar_d_mass,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.64, 0.76),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.8),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            # corr_eval_info_additional_info_key: CorrCalcInfo(
            #     var_col_name="sec_sel_mva_output",
            #     var_latex_label=LatexObsNames.mva_sel_output_short,
            #     weighted=True,
            #     corr_calc_method="pearson",
            # ),
            "adaptive_binning": False,
        },
    ),
    #
    # Full q2 Region and with M(D(*) + pi) - M(D(*)) > 0.2 Selection
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.75, 2.2),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_d_plus_pi_mass_selection_with_sel_mva_corr",
        additional_selection_info=d_w_wo_star_pi_minus_d_w_wo_star_mass_selection,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.68, 0.68),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.2, 1.7),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_d_plus_pi_mass_selection_with_sel_mva_corr",
        additional_selection_info=d_w_wo_star_pi_minus_d_w_wo_star_mass_selection,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.missing_mass_res_corr_dss_control_SidebandValidation,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.68, 0.68),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.2, 1.7),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="full_range_mm2_with_d_plus_pi_mass_selection_with_sel_mva_corr",
        additional_selection_info=d_w_wo_star_pi_minus_d_w_wo_star_mass_selection,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=(DataSampleLoader.base_identifier, "q2_sideband")),
        hist_var=RDStarHistVars.q2_reco_full_region,
        title=dss_control_channel_title,
        info_text=dss_control_channel_text,
        reco_mode_info_collection=BtoDssLRecoModes.b_to_dss_l_modes_in_various_combinations,
        mc_matching_collection=MCMatchingCollections.original__base__collection,
        text_h_pos=(0.64, 0.72),
        text_v_pos=(0.97, 0.76),
        legend_loc=(2, 2),
        legend_cols=(1, 3),
        y_scale=(1.3, 1.56),
        systematics=None,
        data_component_label=dss_control_channel_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        file_name_appendix="with_d_plus_pi_mass_selection_with_sel_mva_corr",
        additional_selection_info=d_w_wo_star_pi_minus_d_w_wo_star_mass_selection,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
            "adaptive_binning": False,
            "special_binning": modified_full_q2_range_binning,
        },
    ),
]
# Off-Resonance Sideband
off_resonance_data_component_label: str = "Off-Resonance Data"

off_resonance_sideband_title: str = "Off-Resonance Data"
off_resonance_sideband_text: str = ""


_off_resonance_sideband: List[SidebandInfo] = [
    SidebandInfo(
        key="mbc_sideband_without_cut_1",
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=(0.63, 0.70),
        text_v_pos=(0.97, 0.97),
        legend_loc=(2, 2),
        legend_cols=(1, 1),
        y_scale=(1.5, 2.2),
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key="mbc_sideband_without_cut_1",
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=(0.03, 0.03),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 1),
        legend_cols=(1, 1),
        y_scale=(1.3, 1.7),
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key="mbc_sideband_without_cut_1",
        hist_var=RDStarHistVars.missing_mass_res_corr_off_resonance_SidebandValidation,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=(0.03, 0.03),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 1),
        legend_cols=(1, 1),
        y_scale=(1.3, 1.7),
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="full_range_mm2_with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(  # TODO: Optimize Layout!
        key="mbc_sideband_without_cut_1",
        hist_var=RDStarHistVars.cs_mva_output,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(  # TODO: Optimize Layout!
        key="mbc_sideband_without_cut_1",
        hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        requires_selection=SelectionSetups.used_selections_tuple,
        file_name_appendix="with_cs_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="cs_mva_output",
                var_latex_label=LatexObsNames.mva_cs_output,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=("mbc_sideband", "q2_sideband_without_cut_1")),
        hist_var=RDStarHistVars.pl_star_SidebandValidation,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 1),
        y_scale=(1.5, 2.2),
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        split_chi_info_text=(False, True),
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=("mbc_sideband", "q2_sideband_without_cut_1")),
        hist_var=RDStarHistVars.missing_mass_res_corr_q2sbSidebandValidation,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 1),
        y_scale=(1.5, 1.9),
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=("mbc_sideband", "q2_sideband_without_cut_1")),
        hist_var=RDStarHistVars.missing_mass_res_corr_off_resonance_SidebandValidation,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=(0.63, 0.70),
        text_v_pos=(0.97, 0.97),
        legend_loc=(2, 2),
        legend_cols=(1, 1),
        y_scale=(1.5, 1.9),
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="full_range_mm2_with_sel_mva_corr",
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="sec_sel_mva_output",
                var_latex_label=LatexObsNames.mva_sel_output_short,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
    SidebandInfo(  # TODO: Optimize Layout!
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=("mbc_sideband", "q2_sideband_without_cut_1")),
        hist_var=RDStarHistVars.cs_mva_output,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="with_sel_mva_corr",
    ),
    SidebandInfo(  # TODO: Optimize Layout!
        key=DataSampleLoader.get_sideband_key_for(sideband_keys=("mbc_sideband", "q2_sideband_without_cut_1")),
        hist_var=RDStarHistVars.mva_output_loose_for_sbv_var,
        title=off_resonance_sideband_title,
        info_text=off_resonance_sideband_text,
        mc_matching_collection=MCMatchingCollections.original__off_resonance__collection,
        text_h_pos=0.03,
        legend_loc=2,
        split_chi_info_text=False,
        systematics=None,
        valid_sample_setups=(SampleSetupHandler.off_resonance_sample_name,),
        data_component_label=off_resonance_data_component_label,
        add_scaling_factor_to_info_text=True,
        use_short_titles=True,
        reset_off_to_on_res_scaling=True,
        file_name_appendix="with_cs_mva_corr",
        requires_selection=SelectionSetups.used_selections_tuple,
        additional_info={
            corr_eval_info_additional_info_key: CorrCalcInfo(
                var_col_name="cs_mva_output",
                var_latex_label=LatexObsNames.mva_cs_output,
                weighted=True,
                corr_calc_method="pearson",
            ),
        },
    ),
]


_sideband_info_collection: List[SidebandInfo] = [
    *_q2_sideband_info_collection,
    *_q2_with_fei_corr_sideband_info_collection,
    *_tag_mbc_sideband_info_collection,
    *_cs_mva_sideband_info_collection,
    *_dss_reco_full_q2_sideband,
    *_off_resonance_sideband,
]

_sideband_info_collection_with_full_sys: List[SidebandInfo] = SidebandInfo.create_comparer_infos_with_systematics(
    comparer_info_list=_sideband_info_collection,
    systematics_collection=full_systematics_cols,
    sys_name_tag=full_systematics_name,
)

sideband_info_collection: List[SidebandInfo] = _sideband_info_collection
sideband_info_collection_plus_with_full_sys: List[SidebandInfo] = (
    _sideband_info_collection + _sideband_info_collection_with_full_sys
)


_sbic_check_list: List[Tuple[str, List[SidebandInfo]]] = [
    ("q2_sideband_info_collection", _q2_sideband_info_collection),
    ("q2_with_fei_corr_sideband_info_collection", list(_q2_with_fei_corr_sideband_info_collection)),
    ("tag_mbc_sideband_info_collection", _tag_mbc_sideband_info_collection),
    ("cs_mva_sideband_info_collection", _cs_mva_sideband_info_collection),
    ("dss_reco_full_q2_sideband", _dss_reco_full_q2_sideband),
    ("off_resonance_sideband", _off_resonance_sideband),
]

for k, sbi_c in _sbic_check_list:
    assert all(v.reset_off_to_on_res_scaling == sbi_c[0].reset_off_to_on_res_scaling for v in sbi_c), k
