"""
Applies final correction to weights added post n-tuple production by us.

Felix Metzner 2020
"""

import os
import numpy as np
import pandas as pd

from typing import Optional, Tuple, Dict, Generator, Any

from rdstar.utilities import PathType
from rdstar.offline_analysis.tasks import utilities

from rdstar.mc_reweighting.BtoDlnu.BtoDlnu_weights import BtoDFFReweighting
from rdstar.mc_reweighting.BtoDstlnu.BtoDstlnu_weights import BtoDstarFFReweighting
from rdstar.mc_reweighting.BFReweighting.b_bf_reweighting import BMesonBFReweighter

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.rdstar_samples import data_type_combinations

from rdstar.offline_analysis.generator_level_mc_matching.gen_mc_matching_info import GenMCIDs
from rdstar.offline_analysis.generator_level_mc_matching import gen_b1_id_col_name, gen_b2_id_col_name

from rdstar.offline_analysis.selection.selection_setups import SelectionSetups
from rdstar.offline_analysis.selection.selection import SelectionSetupInterpreter
from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection

from rdstar.offline_analysis.mc_scaling_weights import MCSampleScaler
from rdstar.offline_analysis.weight_management.weight_information import WeightManager


__all__ = [
    "WeightFinalizingManager",
    "WeightFinalizer",
]


class WeightFinalizingManager:
    def __init__(
        self,
        selection_name: str,
        cs_tag: str,
        input_dict: Dict[str, PathType],
    ) -> None:
        self.selection_name: str = selection_name
        self.cs_tag: str = cs_tag

        self.input_dict = input_dict

    def get_input_files(self) -> Dict[str, PathType]:
        input_dict = {}
        for selection_tag in SelectionSetups.get_selection_tags(selection_name=self.selection_name, cs_tag=self.cs_tag):
            input_dict.update({k: v for k, v in self.input_dict.items() if k.endswith(selection_tag)})
            for ext in SelectionSetupInterpreter.get_possible_subsample_extensions(input_dict=self.input_dict):
                input_dict.update({k: v for k, v in self.input_dict.items() if k.endswith(selection_tag + ext)})
        return input_dict

    def get_flag_info_dict_path(self) -> PathType:
        return WeightManager.get_flag_info_dict_path_from_task_input_dict(
            input_dict=self.input_dict,
            selection_name=self.selection_name,
        )

    def get_input_key_path_pairs(self) -> Generator[Tuple[str, PathType], None, None]:
        for input_key, input_path in self.get_input_files().items():
            yield input_key, input_path


class WeightFinalizer:
    def __init__(
        self,
        input_key: str,
        input_path: PathType,
        pid_dict_path: PathType,
        file_dict_dir_path: PathType,
    ) -> None:
        self.input_key: str = input_key
        self.input_path: PathType = input_path
        self.pid_dict_path: PathType = pid_dict_path
        self._file_dict_dir_path: PathType = file_dict_dir_path

    @staticmethod
    def get_flag_info_dict(flag_info_dict_path: PathType) -> Dict[str, Any]:
        return utilities.load_dict_from_json(path=flag_info_dict_path)

    @staticmethod
    def reset_b_to_d_or_dst_tau_nu_ff_weights(df: pd.DataFrame) -> None:
        relevant_gen_ids: Tuple[int, ...] = (
            GenMCIDs.Dtau,
            GenMCIDs.Dsttau,
        )

        evt_cond_first_b: pd.Series = (df[gen_b1_id_col_name].isin(relevant_gen_ids)) & (
            df[Cols.b_via_tau_index_var] == 1
        )
        evt_cond_second_b: pd.Series = (df[gen_b2_id_col_name].isin(relevant_gen_ids)) & (
            df[Cols.b_via_tau_index_var] == 2
        )

        evt_cond: pd.Series = evt_cond_first_b | evt_cond_second_b

        ff_cols = [c for c in df.columns if c.startswith("weight_ff_") and c.endswith("nom")]
        assert len(ff_cols) == 6, (len(ff_cols), ff_cols)

        for nom_ff_col in ff_cols:
            df.loc[evt_cond, nom_ff_col] = 1.0
            for var_ff_col in [col for col in df.columns if col.startswith(nom_ff_col.replace("_nom", "_var_"))]:
                df.loc[evt_cond, var_ff_col] = 1.0

    @staticmethod
    def apply_weight_finalization_to_df(
        df: pd.DataFrame,
        flag_info_dict_path: PathType,
        file_dict_dir_path: PathType,
        combo_name: str,
        key_hint: Optional[str] = None,
    ) -> None:
        combo_hint = combo_name
        if key_hint is not None:
            combo_hint += f" with {key_hint} selection"

        assert WeightManager.final_weight_col not in df.columns, WeightManager.final_weight_col
        assert WeightManager.combined_ff_weight_col_name not in df.columns, WeightManager.combined_ff_weight_col_name

        # Particle ID Weights
        nom_pid_col, nom_pid_cols = ParticleIDCorrection.apply_pid_weight_finalization(
            df=df,
            flag_uncert_dict=WeightFinalizer.get_flag_info_dict(flag_info_dict_path=flag_info_dict_path),
        )
        assert nom_pid_col == ParticleIDCorrection.get_final_pid_weight_col_name(), (
            nom_pid_col,
            ParticleIDCorrection.get_final_pid_weight_col_name(),
        )

        # Reset FF Weights for B -> D tau nu case
        WeightFinalizer.reset_b_to_d_or_dst_tau_nu_ff_weights(df=df)

        # Form Factor Weights for B -> D l nu
        BtoDFFReweighting.finalize_weights(df=df, threshold=3.0, reset_special_to=0.0)

        # Form Factor Weights for B -> D* l nu
        BtoDstarFFReweighting.finalize_weights(
            df=df,
            threshold=5.0,
            set_to=1.0,
            reset_special_to=0.0,
            combo_name=combo_hint,
        )

        ff_cols = [c for c in df.columns if c.startswith("weight_ff_") and c.endswith("nom")]
        assert len(ff_cols) == 6, (len(ff_cols), ff_cols)

        for c in ff_cols:
            df.loc[np.isnan(df[c]) & df.data_type.isin(data_type_combinations["Special"]), c] = 1.0
            for cc in [col for col in df.columns if col.startswith(c.replace("_nom", "_var_"))]:
                df.loc[np.isnan(df[cc]) & df.data_type.isin(data_type_combinations["Special"]), cc] = 1.0

        df.loc[:, WeightManager.combined_ff_weight_col_name] = df[ff_cols].product(axis=1)

        # Application of new Scaling Weight Calculator and B Meson BF Reweighter
        my_mc_sample_scaler = MCSampleScaler(
            data_type_column_name=Cols.data_type,
            file_dict_dir_path=file_dict_dir_path,
        )
        my_mc_sample_scaler.add_scaling_weights(df=df)

        my_b_bf_reweighter = BMesonBFReweighter(
            data_type_column_name=Cols.data_type,
            rescale_others=True,
        )
        my_b_bf_reweighter.add_bf_weights(df=df)

        # Combining all weights
        WeightManager.add_final_weight(df=df)

        # Calculation of additional weight combinations
        WeightManager.add_additional_weights(df=df)
        WeightManager.add_bf_correction_weights(df=df)

        # Calculation of old weight definitions
        WeightManager.add_old_final_weight(df=df)
        WeightManager.add_old_additional_weights(df=df)
        WeightManager.add_old_bf_correction_weights(df=df)

    def add_finalized_weights(
        self,
        df: pd.DataFrame,
        combo_name: str,
        key_hint: Optional[str] = None,
    ) -> None:
        self.apply_weight_finalization_to_df(
            df=df,
            flag_info_dict_path=self.pid_dict_path,
            file_dict_dir_path=self._file_dict_dir_path,
            combo_name=combo_name,
            key_hint=key_hint,
        )

    @staticmethod
    def get_output_path(
        input_path: PathType,
        tmp: bool = False,
        intermediate: bool = False,
    ) -> PathType:
        dir_path, file_name = os.path.split(input_path)
        output_name_1, output_name_2 = file_name.split("_", 1)
        if tmp:
            output_name_1 = "temp_" + output_name_1
        if intermediate:
            output_name_1 = output_name_1 + "_intermediate"
        output_name = f"{output_name_1}_{WeightManager.finalized_weights_appendix}_{output_name_2}"
        return os.path.join(dir_path, output_name)

    @staticmethod
    def get_output_key(input_key: str, intermediate: bool = False) -> str:
        output_key_1, output_key_2 = input_key.split("_", 1)
        tag = "_intermediate" if intermediate else ""
        return f"{output_key_1}{tag}_{WeightManager.finalized_weights_appendix}_{output_key_2}"

    def apply_weight_finalization(self, combo_name: str, dry_run: bool = False) -> Tuple[str, PathType]:
        hdf_key = self.get_output_key(input_key=self.input_key, intermediate=False)
        output_key = self.get_output_key(input_key=self.input_key, intermediate=True)
        output_path = self.get_output_path(input_path=self.input_path, intermediate=True)

        if not dry_run:
            if self.input_key.startswith("Data"):
                if os.path.exists(output_path):
                    assert os.path.islink(output_path), output_path
                    os.remove(output_path)
                os.symlink(src=self.input_path, dst=output_path)
            else:
                df = pd.read_hdf(path_or_buf=self.input_path)
                assert isinstance(df, pd.DataFrame)
                self.add_finalized_weights(
                    df=df,
                    key_hint=self.input_key,
                    combo_name=combo_name,
                )

                df.to_hdf(path_or_buf=self.get_output_path(input_path=self.input_path, tmp=True), key=hdf_key)
                os.rename(src=self.get_output_path(input_path=self.input_path, tmp=True), dst=output_path)

        return output_key, output_path
