"""
Script to create a mapping of Markus' XuLNu Hybrid MC so that it can be used by the AnalysisTools.

Felix Metzner 2022
"""

import os

from typing import Dict, List

from rdstar.utilities import PathType

__all__ = [
    "HybridMCMapper",
]


class HybridMCMapper:

    markus_source_dir_path = "/group/belle2/users/mapr/BToXuLNu/gsim/mdst/"  # type: PathType
    my_path = "/home/belle2/metzner/MyXuLNuMapping"  # type: PathType

    dir_mapping = {
        "InclusiveCharged": "charged_inclusive",
        "InclusiveMixed": "mixed_inclusive",
        "ResonantCharged": "charged_resonances",
        "ResonantMixed": "mixed_resonances",
    }  # type: Dict[str, str]

    def __init__(
        self,
        overwrite: bool = False,
    ) -> None:
        self.overwrite = overwrite  # type: bool

    def create_symlinks(self) -> None:
        if not os.path.isdir(self.my_path):
            raise RuntimeError(f"Directory {self.my_path} does not exist! Please create it before running this script!")

        for my_dir_key, base_dir_key in self.dir_mapping.items():
            my_dir_path = os.path.join(self.my_path, my_dir_key)  # type: PathType
            os.makedirs(my_dir_path, exist_ok=True)

            sub_dir_paths = [
                os.path.join(self.markus_source_dir_path, f)
                for f in os.listdir(self.markus_source_dir_path)
                if f.startswith(base_dir_key)
            ]  # type: List[PathType]

            for sub_dir_path in sub_dir_paths:
                for mdst_file in os.listdir(sub_dir_path):
                    assert mdst_file.endswith(".mdst"), (sub_dir_path, mdst_file)

                    dest_path = os.path.join(my_dir_path, mdst_file)  # type: PathType

                    if self.overwrite:
                        if os.path.exists(dest_path):
                            assert os.path.islink(dest_path), dest_path
                            os.unlink(dest_path)
                    else:
                        assert not os.path.exists(dest_path), dest_path

                    os.symlink(os.path.join(sub_dir_path, mdst_file), dest_path)


if __name__ == "__main__":
    xulnu_mapper = HybridMCMapper(overwrite=False)
    xulnu_mapper.create_symlinks()
