"""
Steering file for testing the final state radiation correction.

Felix Metzner 2018
"""

import basf2

from typing import Tuple, List, Dict, Any

from modularAnalysis import setAnalysisConfigParams

from analysistools.processing import path_utilities
import analysistools.processing.final_state_particles_utils as fsp

from rdstar.utilities import PathType
from rdstar.online_analysis import b2bii_utils


__all__ = [
    "create_fsr_test_path",
]


def create_fsr_test_path(
    input_files: List[PathType],
    input_sequences: List[Tuple[int, int]],
    output_dict: Dict[str, PathType],
    config: Dict[str, Any],
    method: str,
    max_phi_angle: float,
    max_theta_angle: float,
    max_energy: float,
    momentum_factor: float,
) -> basf2.Path:
    path_utilities.setup_environment_vars()
    path_utilities.setup_databases(config)
    path_utilities.fei_backward_compatibility()

    main_path = basf2.create_path()  # type: basf2.Path

    b2bii_utils.b2bii_read_and_convert(
        path=main_path,
        input_files=input_files,
        input_sequences=input_sequences,
    )

    setAnalysisConfigParams({"mcMatchingVersion": "Belle"}, main_path)

    fsp_collection = fsp.FinalStateParticleCollection()

    photon = fsp.FinalStateParticle(
        label="gamma",
        particlelist="gamma:fsp",
        origin=["gamma:mdst", "gamma:v0mdst"],
        cuts=[
            "goodBelleGamma == 1",
        ],
    )

    fsp_collection.add_fsps(photon=photon)

    electron = fsp.FinalStateParticle(
        label="e",
        particlelist="e-:fsp",
        origin="reco",
        cuts=[
            "dr < 2",
            "abs(dz) < 4",
            "eIDBelle > 0.1",
            "useCMSFrame(pt) > 0.1",
        ],
        electron_fsr_correction=fsp.FSRCorrectionSettings(
            fsr_gamma_list_name="gamma:fsp",
            angle_threshold=max_phi_angle,
            theta_threshold=max_theta_angle,
            energy_threshold=max_energy,
            momentum_factor=momentum_factor,
            method=method,
        ),
    )

    fsp_collection.add_fsps(electron=electron)

    fsp_collection.get_final_state_particles(path=main_path, use_mc_info=config["use_mc_info"])

    ############
    # e and mu #
    ############

    lepton_vars = [
        "PDG",
        "matchedMC(PDG)",
        "mcMother(PDG)",
        "mcMother(mcMother(PDG))",
        "mcMother(mcDaughter(0,PDG))",
        "mcDaughter(0,PDG)",
        "isSignal",
        "M",
        "dM",
        "InvM",
        "daughterInvariantMass(0, 1)",
        "E",
        "Q",
        "dQ",
        "useCMSFrame(p)",
        "useCMSFrame(pt)",
        "p",
        "pt",
        "px",
        "py",
        "pz",
        "eIDBelle",
        "muIDBelle",
        "muIDBelleQuality",
        "atcPIDBelle(2,3)",
        "extraInfo(ksnbStandard)",
        "extraInfo(goodKs)",
        "dr",
        "dz",
        "pValue",
        "nCDCHits",
        "nSVDHits",
        "daughter(0, pValue)",
        "daughter(0, nCDCHits)",
        "daughter(0, nSVDHits)",
        "daughter(0, dr)",
        "daughter(0, dz)",
        "matchedMC(E)",
        "matchedMC(p)",
        "matchedMC(px)",
        "matchedMC(py)",
        "matchedMC(pz)",
        "extraInfo(fsrCorrectedE)",
        "extraInfo(fsrCorrectedPx)",
        "extraInfo(fsrCorrectedPy)",
        "extraInfo(fsrCorrectedPz)",
        "extraInfo(fsrCorrectedP)",
        "extraInfo(fsrGammaE)",
        "extraInfo(fsrGammaPx)",
        "extraInfo(fsrGammaPy)",
        "extraInfo(fsrGammaPz)",
        "extraInfo(fsrGammaP)",
        "extraInfo(fsrCorrected)",
    ]  # type: List[str]

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["e_ntuple"],
        particleList="e-:fsp",
        variables=lepton_vars,
    )

    #########
    # Gamma #
    #########

    gamma_vars = [
        "PDG",
        "matchedMC(PDG)",
        "mcMother(PDG)",
        "mcMother(mcMother(PDG))",
        "mcMother(mcMother(mcMother(PDG)))",
        "mcMother(mcMother(mcMother(mcMother(PDG))))",
        "mcMother(mcDaughter(0,PDG))",
        "isSignal",
        "goodBelleGamma",
        "clusterE9E25",
        "clusterNHits",
        "clusterHighestE",
        "M",
        "dM",
        "InvM",
        "abs(SigMBF)",
        "E",
        "Q",
        "dQ",
        "useCMSFrame(p)",
        "useCMSFrame(pt)",
        "p",
        "pt",
        "px",
        "py",
        "pz",
        "eIDBelle",
        "muIDBelle",
        "muIDBelleQuality",
        "atcPIDBelle(2,3)",
        "extraInfo(ksnbStandard)",
        "extraInfo(goodKs)",
        "dr",
        "dz",
        "matchedMC(E)",
        "matchedMC(px)",
        "matchedMC(py)",
        "matchedMC(pz)",
        "matchedMC(p)",
        "extraInfo(fsrCorrected)",
    ]  # type: List[str]

    main_path.add_module(
        "VariablesToNtuple",
        fileName=output_dict["gamma_ntuple"],
        particleList="gamma:fsp",
        variables=gamma_vars,
    )

    return main_path
