"""
Plotter class for low level cut study.

Patrick Ecker 2020
"""


import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, Tuple, Dict, List

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.plot_style import set_matplotlibrc_params, KITColors

from rdstar.utilities import PathType
from rdstar.cut_information_objects.cut_info import CutInfo

from rdstar.studies.low_level_cut_study.utilities.low_level_plotter_base import LowLevelPlotterBaseClass
from rdstar.studies.low_level_cut_study.utilities.special_observables_handler import SpecialVariableCollections


__all__ = [
    "LowLevelCutPlotter",
]


class LowLevelCutPlotter(LowLevelPlotterBaseClass):
    plot_name_suffix = "evaluation"  # type: str

    default_number_of_bins = 50  # type: int

    cut_line_color = KITColors.kit_red  # type: str
    bkg_color = KITColors.kit_cyan  # type: str
    sig_color = KITColors.kit_orange  # type: str

    def __init__(
        self,
        n_tuple: str,
        target_path: PathType,
        apply_pre_cuts: str,
        variable_set: str,
        remove_outliers: bool,
    ) -> None:
        super().__init__(
            n_tuple=n_tuple,
            target_path=target_path,
            apply_pre_cuts=apply_pre_cuts,
            variable_set=variable_set,
            remove_outliers=remove_outliers,
        )

    def get_output_key_path_pair(self, tuple_key: str, cut_info: CutInfo) -> Tuple[str, PathType]:
        escaped_key = self.escape_decay_string_in_tuple_key(tuple_key=tuple_key)  # type: str
        output_key = f"{escaped_key}_{cut_info.cut_name}_{self.plot_name_suffix}"  # type: str
        path = os.path.join(self.target_path, f"{output_key}.pdf")  # type: PathType
        return output_key, path

    def generate_output_dict(self) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]
        relevant_tuple_keys = [key for key in self.cut_dicts if key.startswith(self.n_tuple)]
        for tuple_key in relevant_tuple_keys:

            if self._variable_set_name == self.cut_variable_set_option:
                _cut_infos = self.cut_dicts[tuple_key]["cuts"]
                assert isinstance(_cut_infos, list), _cut_infos
                cut_infos = _cut_infos  # type: List[CutInfo]
            elif self._variable_set_name == self.validation_variable_set_option:
                hv_collection = SpecialVariableCollections.get_variable_collection_for_n_tuple(n_tuple_name=tuple_key)
                hist_vars = hv_collection.get_set_of_observables(root_compatible=False, use_mc_info=False)
                cut_infos = [
                    CutInfo.create_fake_cut_info_from_hist_variable(hist_var=hv, make_cut_name_root_compatible=True)
                    for hv in hist_vars
                ]
            else:
                raise RuntimeError(f"Invalid option {self.variable_set_option_name} = {self._variable_set_name}!")

            for cut_info in cut_infos:
                self.update_output_dict(
                    output_dict=output,
                    new_output_dict=dict([self.get_output_key_path_pair(tuple_key=tuple_key, cut_info=cut_info)]),
                )

        return output

    def create_plots(self) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]
        relevant_tuple_keys = [key for key in self.cut_dicts if key.startswith(self.n_tuple)]

        for tuple_key in relevant_tuple_keys:
            _channel = self.cut_dicts[tuple_key]["channel"]
            assert isinstance(_channel, str), (_channel, type(_channel).__name__)
            channel = _channel  # type: str

            if self._variable_set_name == self.cut_variable_set_option:
                new_output = self._create_cut_variables_plots(channel=channel, tuple_key=tuple_key)
                self.update_output_dict(output_dict=output, new_output_dict=new_output)
            elif self._variable_set_name == self.validation_variable_set_option:
                new_output = self._create_validation_variables_plots(channel=channel, tuple_key=tuple_key)
                self.update_output_dict(output_dict=output, new_output_dict=new_output)
            else:
                raise RuntimeError(f"Invalid option {self.variable_set_option_name} = {self._variable_set_name}!")

        return output

    def _create_cut_variables_plots(
        self,
        channel: str,
        tuple_key: str,
    ) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]
        cut_infos = self.cut_dicts[tuple_key]["cuts"]

        for cut_info in cut_infos:
            assert isinstance(cut_info, CutInfo), type(cut_info).__name__
            new_output = self._create_validation_plot(
                channel=channel,
                tuple_key=tuple_key,
                cut_info=cut_info,
                number_of_bins=self.default_number_of_bins,
            )
            self.update_output_dict(output_dict=output, new_output_dict=new_output)

        return output

    def _create_validation_variables_plots(
        self,
        channel: str,
        tuple_key: str,
    ) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]

        hist_var_collection = SpecialVariableCollections.get_variable_collection_for_n_tuple(n_tuple_name=tuple_key)

        for hist_var in hist_var_collection.get_set_of_observables(root_compatible=False, use_mc_info=False):
            cut_info = CutInfo.create_fake_cut_info_from_hist_variable(
                hist_var=hist_var,
                make_cut_name_root_compatible=True,
            )
            new_output = self._create_validation_plot(
                channel=channel,
                tuple_key=tuple_key,
                cut_info=cut_info,
                number_of_bins=hist_var.n_bins,
            )
            self.update_output_dict(output_dict=output, new_output_dict=new_output)

        return output

    def _create_validation_plot(
        self,
        channel: str,
        tuple_key: str,
        cut_info: CutInfo,
        number_of_bins: int,
    ) -> Dict[str, PathType]:
        output = {}  # type: Dict[str, PathType]

        ch_cut = f"{self.decay_mode_id_col_name} == {self.channel_name_to_id_mapping[channel]}" if channel else None
        _cut_infos = self.cut_dicts[tuple_key]["cuts"]
        assert isinstance(_cut_infos, list), type(_cut_infos).__name__
        cut_infos = _cut_infos  # type: List[CutInfo]

        set_matplotlibrc_params()

        output_key, output_path = self.get_output_key_path_pair(tuple_key=tuple_key, cut_info=cut_info)
        output.update({output_key: output_path})

        df = self.apply_cuts(df=self.df, pivot=cut_info, cut_infos=cut_infos, channel_cut_string=ch_cut)
        self.remove_outliers_from_dfs(pivot=cut_info, dfs=(df,))

        fig, ax = plt.subplots(figsize=self.default_figure_size)

        if len(np.unique(df.index)) == 0:
            ax.set_title(label=self.get_plot_title(channel=channel), loc="right")
            ax.set_title(label=cut_info.cut_label, loc="left")
            ax.text(x=0.5, y=0.5, s="No Data", fontsize=16)

        else:
            sig_data = df.loc[df[self.is_signal_col_name] == 1, cut_info.cut_var_root].values  # type: np.ndarray
            bkg_data = df.loc[df[self.is_signal_col_name] == 0, cut_info.cut_var_root].values  # type: np.ndarray
            if cut_info.cut_abs:
                sig_data = np.abs(sig_data)
                bkg_data = np.abs(bkg_data)

            var_range = (np.nanmin(np.concatenate((sig_data, bkg_data))), np.nanmax(np.concatenate((sig_data, bkg_data))))

            sig_counts, sig_bins = np.histogram(sig_data, bins=number_of_bins, range=var_range)
            bkg_counts, bkg_bins = np.histogram(bkg_data, bins=number_of_bins, range=var_range)

            factor = 0.5 / (sig_counts.max() / ((sig_counts + bkg_counts).max()))  # type: Union[float, int]
            if factor > 2.0 and np.isfinite(factor):
                factor = int(round(factor, 0))
            else:
                factor = 1

            ax.hist(
                [bkg_data, sig_data],
                bins=number_of_bins,
                range=var_range,
                histtype="stepfilled",
                stacked=True,
                lw=1.0,
                color=[self.bkg_color, self.sig_color],
                label=["Background", "Signal"],
                edgecolor=KITColors.kit_black,
            )
            ax.hist(
                x=sig_bins[:-1],
                bins=sig_bins,
                weights=sig_counts * factor,
                range=var_range,
                histtype="step",
                lw=1.2,
                edgecolor=KITColors.dark_grey,
                label=r"Signal Shape ($\times$" + f"{factor})" if factor != 1.0 else "Signal Shape",
            )

            self.plot_cut_lines(ax=ax, cut_info=cut_info)

            plt.grid(False)

            x_label_unit = rf" in {cut_info.latex_unit}" if cut_info.latex_unit is not None else ""
            ax.set_xlabel(xlabel=cut_info.cut_var_latex_label + x_label_unit)
            ax.set_ylabel(ylabel="Entries")

            ax.set_title(label=self.get_plot_title(channel=channel), loc="right")
            use_title = self.get_cut_string_title(cut_info=cut_info)
            if ax.get_ylim()[1] > 0.85e4:
                padding = " " * 12
                use_title = padding + cut_info.cut_label
            ax.set_title(label=use_title, loc="left")

            ax.legend(loc="best", frameon=False)

        target_dir, filename_w_ext = os.path.split(output_path)
        filename_wo_ext = os.path.splitext(filename_w_ext)[0]
        export(fig=fig, filename=filename_wo_ext, target_dir=target_dir, close_figure=True)

        return output


if __name__ == "__main__":
    plotter = LowLevelCutPlotter(
        n_tuple="dcharged_ntuple",
        target_path="",
        apply_pre_cuts=LowLevelPlotterBaseClass.apply_other_cuts_option,
        variable_set=LowLevelPlotterBaseClass.cut_variable_set_option,
        remove_outliers=False,
    )
    print(plotter.generate_output_dict())

    test_df_path = "/ceph/pecker/rdstar/low_level_cut_study/CombinedNTuples/dcharged_ntuple_total.h5"  # type: PathType
    my_df = pd.read_hdf(path_or_buf=test_df_path)
    assert isinstance(my_df, pd.DataFrame)

    plotter.add_df(df=my_df)
    plotter.create_plots()
