"""
Logging Methods

Inspired by Kilian Lieret's TagCal package.

Felix Metzner
"""

import copy
import logging
import colorlog

from dataclasses import dataclass
from pathlib import Path, PurePath
from typing import List, Optional, Union, Dict, Iterable, Type


__all__ = [
    "set_logger",
    "get_logger",
    "log_to_default",
    "OptionalLoggerType",
    "reset_third_party_logging_level",
    "add_ignore_log_from_files_filter_to",
]


OptionalLoggerType = Optional[logging.Logger]

_logger: OptionalLoggerType = None

logging.captureWarnings(True)


@dataclass(repr=True, eq=True, frozen=True)
class LoggingRecord:
    module: str
    level_no: int
    msg: str


class DuplicateFilter(logging.Filter):
    """Filter to filter out duplicate messages."""

    def __init__(self, logger: logging.Logger) -> None:
        super().__init__()
        self.logger: logging.Logger = logger
        self.last_log: Optional[LoggingRecord] = None
        self.last_log_repeated: int = 0
        self.last_log_level: int = 0

    def filter(self, record) -> bool:
        current_log = LoggingRecord(module=record.module, level_no=record.levelno, msg=record.msg)
        if str(current_log) == str(self.last_log):
            self.last_log_repeated += 1
            self.last_log_level = current_log.level_no
            return False
        else:
            # Need to already reset this here, because else we get into a
            # recursion with the logging message in the if statement
            self.last_log = current_log

            if self.last_log_repeated >= 1 and self.logger:
                self.logger.log(
                    msg=f"The last message was repeated {self.last_log_repeated} additional times",
                    level=self.last_log_level,
                )

            self.last_log_repeated = 0
            return True


LOG_COLORS: Dict[str, str] = {
    "DEBUG": "cyan",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "red",
    "CRITICAL": "red",
}


_default_format_str = "%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s"

#: Default formatter for logging stream handlers
DEFAULT_SH_FORMATTER = logging.Formatter(
    fmt=f"{_default_format_str}",
    datefmt="%Y-%m-%d:%H:%M:%S",
)
#: Color formatter for logging stream handlers
COLOR_SH_FORMATTER = colorlog.ColoredFormatter(
    fmt=f"%(log_color)s{_default_format_str}",
    datefmt="%Y-%m-%d:%H:%M:%S",
    log_colors=LOG_COLORS,
)
#: Default formatter for logging file handlers
DEFAULT_FH_FORMATTER = logging.Formatter(
    fmt=f"{_default_format_str}",
    datefmt="%Y-%m-%d:%H:%M:%S",
)


def set_logger(
    name: str = "Logger",
    level: int = logging.DEBUG,
    sh_level: int = logging.DEBUG,
    use_color: bool = False,
    overwrite: bool = False,
) -> logging.Logger:
    global _logger

    if _logger is not None and not overwrite:
        raise RuntimeError(f"Logger is already set up with the name '{_logger.name}'!")

    _logger = colorlog.getLogger(name)
    assert _logger is not None

    _logger.propagate = False

    _logger.setLevel(level)

    if use_color:
        sh = colorlog.StreamHandler()
        sh.setFormatter(COLOR_SH_FORMATTER)
    else:
        sh = logging.StreamHandler()
        sh.setFormatter(DEFAULT_SH_FORMATTER)
    sh.setLevel(sh_level)

    _logger.addHandler(sh)
    _logger.addFilter(filter=DuplicateFilter(logger=_logger))

    return _logger


def get_logger() -> logging.Logger:
    global _logger

    if _logger is None:
        raise RuntimeError("No Logger set up, yet! Call the set_logger method first!")

    return _logger


_ADD_LOG_FH_PROPERTY: str = "rdstar_file_handler_targets"


def reset_third_party_logging_level() -> None:
    mpl_logger = logging.getLogger("matplotlib")
    mpl_logger.setLevel(logging.WARNING)
    luigi_logger = logging.getLogger("luigi")
    luigi_logger.setLevel(logging.WARNING)
    luigi_logger.propagate = False
    urllib3_logger = logging.getLogger("urllib3")
    urllib3_logger.setLevel(logging.WARNING)


def add_log_fh(
    basedir: Union[str, PurePath],
    log: OptionalLoggerType = None,
) -> None:
    """
    Add file output to our logger: Creates
        - debug.log,
        - warn.log and
        - error.log
    at the target location.

    Args:
        basedir: Create files in this directory
        log: Use this log. If none, will use the default one.
    """

    logger = lon(log)

    basedir = Path(basedir).resolve()
    basedir.mkdir(parents=True, exist_ok=True)

    # Ensure that we don't add the same file handlers more than once
    log_fh_prop = getattr(logger, _ADD_LOG_FH_PROPERTY, [])
    if basedir in log_fh_prop:
        # Already called exactly this function
        return
    else:
        setattr(logger, _ADD_LOG_FH_PROPERTY, log_fh_prop + [basedir])

    logger.info(f"Log files will be written to {basedir.resolve()}.")

    levels_to_file_name_map: Dict[int, str] = {
        logging.DEBUG: "debug.log",
        logging.WARNING: "warning.log",
        logging.ERROR: "error.log",
    }

    for level, file_name in levels_to_file_name_map.items():
        _lfh = logging.FileHandler(filename=str(basedir / file_name))
        _lfh.setLevel(logging.DEBUG)
        _lfh.setFormatter(DEFAULT_FH_FORMATTER)
        logger.addHandler(_lfh)


def set_stream_level(
    level: str,
    log: OptionalLoggerType = None,
) -> None:
    """ Set the level of all StreamHandlers attached to logger """
    logger = lon(log)
    for handler in logger.handlers:
        if isinstance(handler, logging.StreamHandler) and not isinstance(handler, logging.FileHandler):
            handler.setLevel(getattr(logging, level))


def log_to_default(
    name: str,
) -> logging.Logger:
    if _logger:
        return get_logger()
    else:
        return set_logger(name=name)


def lon(
    log: OptionalLoggerType,
    logger_name: str = "rdstar",
) -> logging.Logger:
    """ Shortcut function to get default log if log is not specified """
    if log is None:
        return set_logger(name=logger_name)
    else:
        return log


def get_all_loggers() -> List[logging.Logger]:
    """ Return all loggers that we know of """
    loggers = [
        log
        for log in logging.root.manager.loggerDict.values()  # type: ignore
        if not isinstance(log, logging.PlaceHolder)
    ]
    return loggers


def create_ignore_log_from_files_filter_class(
    file_name_list: Iterable[str],
) -> Type[logging.Filter]:
    assert all(fn.endswith(".py") for fn in file_name_list)

    class FileNameFilterClass(logging.Filter):
        """Filter to filter out messages from specified files."""

        def __init__(self) -> None:
            super().__init__(name="file_name_filter")
            self.ignored_file_names: Iterable[str] = copy.deepcopy(file_name_list)

        def filter(self, record) -> bool:
            if str(record.filename) in self.ignored_file_names:
                return False
            return True

    return FileNameFilterClass


def add_ignore_log_from_files_filter_to(
    logger: logging.Logger,
    file_name_list: Iterable[str],
) -> None:
    _filter_class: Type[logging.Filter] = create_ignore_log_from_files_filter_class(
        file_name_list=file_name_list,
    )
    logger.addFilter(filter=_filter_class())


if __name__ == "__main__":

    # Test the color scheme for the logger.
    _lg = set_logger(
        name="test",
        use_color=True,
    )
    lg = get_logger()

    lg.setLevel(logging.DEBUG)
    lg.debug("Test debug message")
    lg.info("Test info message")
    lg.warning("Test warning message")

    for _ in range(10):
        lg.warning("Duplicate Test warning message")

    lg.error("Test error message")
    lg.critical("Test critical message")
