"""
Provides information about SPECIAL neutral and charged meson decays as used in Belle.

Felix Metzner 2021
"""

from typing import Tuple

from rdstar.utilities.logging import log_to_default
from rdstar.branching_fractions.bf_utilities import BelleBBFEntry

logger = log_to_default(name=__name__)

__all__ = [
    "sum_bf_bp_rare",
    "sum_bf_b0_rare",
    "sum_bf_bp_ulnu",
    "sum_bf_b0_ulnu",
]

# fmt: off
bm_rare_bfs = (
    BelleBBFEntry(decay="K*-       gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.0000403),
    BelleBBFEntry(decay="K_1-      gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.000043),
    BelleBBFEntry(decay="K_2*-     gamma", model="PHOTOS HELAMP 1.0 0.0 1.0 0.0;", bf=0.0000140),
    BelleBBFEntry(decay="anti-Xsu  gamma", model="BTOXSGAMMA 2;", bf=0.000257),
    BelleBBFEntry(decay="rho-      gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000088),
    BelleBBFEntry(decay="anti-Xdu-  gamma", model="BTOXDGAMMA 2;", bf=0.0000151),
    BelleBBFEntry(decay="K-        e+    e-", model="PHOTOS BTOSLLBALL;", bf=0.00000049),
    BelleBBFEntry(decay="K*-       e+    e-", model="PHOTOS BTOSLLBALL;", bf=0.00000080),
    BelleBBFEntry(decay="anti-Xsu  e+    e-", model="BTOXSLL;", bf=0.00000341),
    BelleBBFEntry(decay="K-        mu+   mu-", model="PHOTOS BTOSLLBALL;", bf=0.00000039),
    BelleBBFEntry(decay="K*-       mu+   mu-", model="PHOTOS BTOSLLBALL;", bf=0.00000080),
    BelleBBFEntry(decay="anti-Xsu  mu+   mu-", model="BTOXSLL;", bf=0.00000311),
    BelleBBFEntry(decay="K-   tau+  tau-", model="PHOTOS BTOSLLBALL;", bf=0.00000013),
    BelleBBFEntry(decay="K*-  tau+  tau-", model="PHOTOS BTOSLLBALL;", bf=0.00000019),
    BelleBBFEntry(decay="anti-Xsu  tau+  tau-", model="BTOXSLL;", bf=0.00000022),
    BelleBBFEntry(decay="K-       nu_e   anti-nu_e", model="PHOTOS PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*-      nu_e   anti-nu_e", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="anti-Xsu nu_e   anti-nu_e", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K-       nu_mu  anti-nu_mu", model="PHOTOS PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*-      nu_mu  anti-nu_mu", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="anti-Xsu nu_mu  anti-nu_mu", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K-       nu_tau anti-nu_tau", model="PHOTOS PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*-      nu_tau anti-nu_tau", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="anti-Xsu nu_tau anti-nu_tau", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K- K0", model="PHOTOS PHSP;", bf=0.00000136),
    BelleBBFEntry(decay="K- pi0", model="PHOTOS PHSP;", bf=0.00001290),
    BelleBBFEntry(decay="anti-K0 pi-", model="PHOTOS PHSP;", bf=0.00002310),
    BelleBBFEntry(decay="pi- pi0", model="PHOTOS PHSP;", bf=0.00000559),
    BelleBBFEntry(decay="K'*- pi0", model="PHOTOS PHSP;", bf=0.00000820),
    BelleBBFEntry(decay="anti-K'*0 pi-", model="PHOTOS PHSP;", bf=0.00004500),
    BelleBBFEntry(decay="anti-K_0*0 pi-", model="PHOTOS PHSP;", bf=0.00004710),
    BelleBBFEntry(decay="K_0*0 K-", model="PHOTOS PHSP;", bf=0.00000220),
    BelleBBFEntry(decay="f'_0 K-", model="PHOTOS PHSP;", bf=0.00001070),
    BelleBBFEntry(decay="f_0 K-", model="PHOTOS PHSP;", bf=0.00000920),
    BelleBBFEntry(decay="f_0(1500) K-", model="PHOTOS PHSP;", bf=0.00000440),
    BelleBBFEntry(decay="f'_0 pi-", model="PHOTOS PHSP;", bf=0.00000300),
    BelleBBFEntry(decay="f_0 pi-", model="PHOTOS PHSP;", bf=0.00000300),
    BelleBBFEntry(decay="f_0(600) pi-", model="PHOTOS PHSP;", bf=0.00000410),
    BelleBBFEntry(decay="anti-K_2*0 pi-", model="PHOTOS STS;", bf=0.00000690),
    BelleBBFEntry(decay="f'_2 K-", model="PHOTOS STS;", bf=0.00000340),
    BelleBBFEntry(decay="f_2 K-", model="PHOTOS STS;", bf=0.00000133),
    BelleBBFEntry(decay="a_00 K-", model="PHOTOS PHSP;", bf=0.00000250),
    BelleBBFEntry(decay="a_0- anti-K0", model="PHOTOS PHSP;", bf=0.00000390),
    BelleBBFEntry(decay="a_00 pi-", model="PHOTOS PHSP;", bf=0.00000580),
    BelleBBFEntry(decay="a_0- pi0", model="PHOTOS PHSP;", bf=0.00000140),
    BelleBBFEntry(decay="anti-K''*0 pi-", model="PHOTOS SVS;", bf=0.00001200),
    BelleBBFEntry(decay="K*- K0", model="PHOTOS SVS;", bf=0.00000030),
    BelleBBFEntry(decay="K*- pi0", model="PHOTOS SVS;", bf=0.00000690),
    BelleBBFEntry(decay="K*- f_0", model="PHOTOS SVS;", bf=0.00000520),
    BelleBBFEntry(decay="anti-K*0 pi-", model="PHOTOS SVS;", bf=0.00001070),
    BelleBBFEntry(decay="K*0 K-", model="PHOTOS SVS;", bf=0.00000110),
    BelleBBFEntry(decay="rho0 K-", model="PHOTOS SVS;", bf=0.00000425),
    BelleBBFEntry(decay="rho(2S)0 K-", model="PHOTOS SVS;", bf=0.00001170),
    BelleBBFEntry(decay="rho(3S)0 K-", model="PHOTOS SVS;", bf=0.00000500),
    BelleBBFEntry(decay="rho- anti-K0", model="PHOTOS SVS;", bf=0.00000800),
    BelleBBFEntry(decay="rho(2S)- anti-K0", model="PHOTOS SVS;", bf=0.00000500),
    BelleBBFEntry(decay="rho(3S)- anti-K0", model="PHOTOS SVS;", bf=0.00000250),
    BelleBBFEntry(decay="rho- pi0", model="PHOTOS SVS;", bf=0.00001090),
    BelleBBFEntry(decay="rho(2S)- pi0", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="rho(3S)- pi0", model="PHOTOS SVS;", bf=0.00000100),
    BelleBBFEntry(decay="rho0 pi-", model="PHOTOS SVS;", bf=0.00000870),
    BelleBBFEntry(decay="rho(2S)0 pi-", model="PHOTOS SVS;", bf=0.00000230),
    BelleBBFEntry(decay="rho(3S)0 pi-", model="PHOTOS SVS;", bf=0.00000100),
    BelleBBFEntry(decay="a_10 K-", model="PHOTOS SVS;", bf=0.00002700),
    BelleBBFEntry(decay="a_1- K0", model="PHOTOS SVS;", bf=0.00003490),
    BelleBBFEntry(decay="a_10 pi-", model="PHOTOS SVS;", bf=0.00002040),
    BelleBBFEntry(decay="a_1- pi0", model="PHOTOS SVS;", bf=0.00002640),
    BelleBBFEntry(decay="rho- f_0", model="PHOTOS SVS;", bf=0.00000190),
    BelleBBFEntry(decay="b_10 K-", model="PHOTOS SVS;", bf=0.00000910),
    BelleBBFEntry(decay="b_10 pi-", model="PHOTOS SVS;", bf=0.00000670),
    BelleBBFEntry(decay="f_2  pi-", model="PHOTOS STS;", bf=0.00000820),
    BelleBBFEntry(decay="a_20 K-", model="PHOTOS STS;", bf=0.00000110),
    BelleBBFEntry(decay="K- K- K+", model="PHOTOS PHSP;", bf=0.00002501),
    BelleBBFEntry(decay="K- K+ pi-", model="PHOTOS PHSP;", bf=0.00000268),
    BelleBBFEntry(decay="K- K_S0 K_S0", model="PHOTOS PHSP;", bf=0.00000944),
    BelleBBFEntry(decay="K- K0 pi0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="K- pi- pi+", model="PHOTOS PHSP;", bf=0.00000290),
    BelleBBFEntry(decay="anti-K0 pi- pi0", model="PHOTOS PHSP;", bf=0.00003086),
    BelleBBFEntry(decay="pi- pi- pi+", model="PHOTOS PHSP;", bf=0.00000460),
    BelleBBFEntry(decay="K- K- pi+", model="PHOTOS PHSP;", bf=0.00000130),
    BelleBBFEntry(decay="K+ pi- pi-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="K_S0 K_S0 pi-", model="PHOTOS PHSP;", bf=0.00000320),
    BelleBBFEntry(decay="K_L0 K_L0 pi-", model="PHOTOS PHSP;", bf=0.00000320),
    BelleBBFEntry(decay="K_S0 K_L0 pi-", model="PHOTOS PHSP;", bf=0.00000640),
    BelleBBFEntry(decay="f'_2 K*-", model="PHOTOS PHSP;", bf=0.00003700),
    BelleBBFEntry(decay="f_2 K*-", model="PHOTOS PHSP;", bf=0.00001800),
    BelleBBFEntry(decay="rho- anti-K_0*0", model="PHOTOS PHSP;", bf=0.00001200),
    BelleBBFEntry(decay="rho0 K_0*-", model="PHOTOS PHSP;", bf=0.00001200),
    BelleBBFEntry(decay="K*- K*0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00007100),
    BelleBBFEntry(decay="K*- rho0", model="PHOTOS SVV_HELAMP 0.0 0.0 1.0 0.0 0.0 0.0;", bf=0.00000610),
    BelleBBFEntry(decay="anti-K*0 rho-", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000920),
    BelleBBFEntry(decay="rho- rho0", model="PHOTOS SVV_HELAMP 0.0 0.0 1.0 0.0 0.0 0.0;", bf=0.00001820),
    BelleBBFEntry(decay="K*- pi- pi+", model="PHOTOS PHSP;", bf=0.00006650),
    BelleBBFEntry(decay="K*- pi0 pi0", model="PHOTOS PHSP;", bf=0.00000260),
    BelleBBFEntry(decay="anti-K*0 pi- pi0", model="PHOTOS PHSP;", bf=0.00000750),
    BelleBBFEntry(decay="rho- K- pi+", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="rho- anti-K0 pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="rho0 K- pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="rho0 anti-K0 pi-", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="K+ K- K*-", model="PHOTOS PHSP;", bf=0.00003130),
    BelleBBFEntry(decay="K0 anti-K0 K*-", model="PHOTOS PHSP;", bf=0.00001500),
    BelleBBFEntry(decay="K0 anti-K*0 K-", model="PHOTOS PHSP;", bf=0.00001500),
    BelleBBFEntry(decay="K*0 anti-K0 K-", model="PHOTOS PHSP;", bf=0.00001500),
    BelleBBFEntry(decay="K*+ K- K-", model="PHOTOS PHSP;", bf=0.00001500),
    BelleBBFEntry(decay="K*- K+ pi-", model="PHOTOS PHSP;", bf=0.00000000),
    BelleBBFEntry(decay="K*- K- pi+", model="PHOTOS PHSP;", bf=0.00000610),
    BelleBBFEntry(decay="K*+ K- pi-", model="PHOTOS PHSP;", bf=0.00000000),
    BelleBBFEntry(decay="eta   K-", model="PHOTOS PHSP;", bf=0.0000027),
    BelleBBFEntry(decay="eta'  K-", model="PHOTOS PHSP;", bf=0.0000702),
    BelleBBFEntry(decay="K_0*- eta", model="PHOTOS PHSP;", bf=0.0000180),
    BelleBBFEntry(decay="eta(2S)   K-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta(1405)  K-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta   pi-", model="PHOTOS PHSP;", bf=0.0000044),
    BelleBBFEntry(decay="eta'  pi-", model="PHOTOS PHSP;", bf=0.0000027),
    BelleBBFEntry(decay="eta(2S)   pi-", model="PHOTOS PHSP;", bf=0.0000020),
    BelleBBFEntry(decay="eta(1405)  pi-", model="PHOTOS PHSP;", bf=0.0000020),
    BelleBBFEntry(decay="omega K-", model="PHOTOS SVS;", bf=0.0000067),
    BelleBBFEntry(decay="phi   K-", model="PHOTOS SVS;", bf=0.0000083),
    BelleBBFEntry(decay="omega pi-", model="PHOTOS SVS;", bf=0.0000069),
    BelleBBFEntry(decay="phi   pi-", model="PHOTOS SVS;", bf=0.00000024),
    BelleBBFEntry(decay="phi(1680) K-", model="PHOTOS SVS;", bf=0.0000008),
    BelleBBFEntry(decay="K*-	 eta", model="PHOTOS SVS;", bf=0.0000193),
    BelleBBFEntry(decay="K*-   eta'", model="PHOTOS SVS;", bf=0.0000049),
    BelleBBFEntry(decay="K*-	 eta(2S)", model="PHOTOS SVS;", bf=0.0000020),
    BelleBBFEntry(decay="K*-   eta(1405)", model="PHOTOS SVS;", bf=0.0000020),
    BelleBBFEntry(decay="rho-  eta", model="PHOTOS SVS;", bf=0.0000054),
    BelleBBFEntry(decay="rho-  eta'", model="PHOTOS SVS;", bf=0.0000091),
    BelleBBFEntry(decay="rho-  eta(2S)", model="PHOTOS SVS;", bf=0.0000020),
    BelleBBFEntry(decay="rho-  eta(1405)", model="PHOTOS SVS;", bf=0.0000020),
    BelleBBFEntry(decay="omega K*-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000034),
    BelleBBFEntry(decay="K*-   phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000100),
    BelleBBFEntry(decay="omega rho-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000106),
    BelleBBFEntry(decay="phi   rho-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000160),
    BelleBBFEntry(decay="K_2*- eta", model="PHOTOS STS;", bf=0.0000091),
    BelleBBFEntry(decay="anti-Xsu eta'", model="PHOTOS BTOXSETAP;", bf=0.0003449),
    BelleBBFEntry(decay="anti-Xsu eta", model="PHOTOS BTOXSETAP;", bf=0.0003872),
    BelleBBFEntry(decay="eta  anti-K0 pi-", model="PHOTOS PHSP;", bf=0.0000145),
    BelleBBFEntry(decay="eta  K- pi0", model="PHOTOS PHSP;", bf=0.0000072),
    BelleBBFEntry(decay="eta  pi0 pi-", model="PHOTOS PHSP;", bf=0.0000072),
    BelleBBFEntry(decay="eta'  anti-K0 pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K- pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  pi0 pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta K-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta' K-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta' pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta' K-", model="PHOTOS PHSP;", bf=0.0000250),
    BelleBBFEntry(decay="eta'  eta' pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  anti-K*0 pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  anti-K0 rho-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  K*- pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  K- rho0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  rho0 pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  pi0 rho-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  anti-K*0 pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  anti-K0 rho-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K*- pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K- rho0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  rho0 pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  pi0 rho-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta K*-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta rho-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta' K*-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta' rho-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta' K*-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  eta' rho-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  anti-K0 pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  K- pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  pi0 pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  anti-K0 pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  K- pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  pi0 pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  eta K-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  eta pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  eta' K-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  eta' pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  eta K-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  eta pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  eta' K-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  eta' pi-", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  phi  K-", model="PHOTOS PHSP;", bf=0.0000042),
    BelleBBFEntry(decay="a_0- eta", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="a_0- eta'", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="a_0- eta(2S)", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="a_0- eta(1405)", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="a_0- a_00", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="a_0- f_0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="K*- a_00", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="anti-K*0 a_0-", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="rho- a_00", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="rho0 a_0-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="omega a_0-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="phi a_0-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_10 a_0-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_1- a_00", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="f_1 a_0-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_10 a_0-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_1- a_00", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="h_1 a_0-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_1- eta", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_1- eta'", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_1- eta(2S)", model="PHOTOS SVS;", bf=0.00000020),
    BelleBBFEntry(decay="a_1- eta(1405)", model="PHOTOS SVS;", bf=0.00000020),
    BelleBBFEntry(decay="a_1- f_0", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="K*- a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="anti-K*0 a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho- a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="rho0 a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="omega a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="phi a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="a_10 a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="f_1 a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_10 a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_1- a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="h_1 a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_1- f_0", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="f_1 K-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="f_1 pi-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="K*- f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="rho- f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_1- f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_1- anti-K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="b_1- pi0", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_1- eta", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_1- eta'", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_1- eta(2S)", model="PHOTOS SVS;", bf=0.00000020),
    BelleBBFEntry(decay="b_1- eta(1405)", model="PHOTOS SVS;", bf=0.00000020),
    BelleBBFEntry(decay="K*- b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="anti-K*0 b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho- b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="rho0 b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="omega b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="phi b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_10 b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="h_1  b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="h_1 K-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="h_1 pi-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="K*- h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00001310),
    BelleBBFEntry(decay="rho- h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="a_20  pi-", model="PHOTOS STS;", bf=0.00000260),
    BelleBBFEntry(decay="f'_2  pi-", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2-  pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2- eta", model="PHOTOS STS;", bf=0.00000029),
    BelleBBFEntry(decay="a_2- eta'", model="PHOTOS STS;", bf=0.00000130),
    BelleBBFEntry(decay="a_2- eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2- eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*- anti-K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*0 K-", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2- anti-K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*- pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*- eta'", model="PHOTOS STS;", bf=0.00000140),
    BelleBBFEntry(decay="K_2*- eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*- eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 rho-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000730),
    BelleBBFEntry(decay="f_2 rho-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000810),
    BelleBBFEntry(decay="f'_2 rho-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_2- rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_2- omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_2- phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*- K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*0 K*-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 K*-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000190),
    BelleBBFEntry(decay="a_2- anti-K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000450),
    BelleBBFEntry(decay="K_2*- rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000025),
    BelleBBFEntry(decay="K_2*- omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000011),
    BelleBBFEntry(decay="K_2*- phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000220),
    BelleBBFEntry(decay="anti-K_2*0 rho-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000025),
    BelleBBFEntry(decay="Sigma+  anti-Delta-- pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Lambda0 anti-p- pi0", model="PHOTOS PHSP;", bf=0.00000300),
    BelleBBFEntry(decay="Sigma0 anti-p- pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Sigma- anti-n0 pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Lambda0  anti-Delta-- pi+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Sigma0  anti-Delta-- pi+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Sigma- anti-p- pi+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Sigma+ anti-p- pi-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Lambda0 anti-n0 pi-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Sigma0 anti-n0 pi-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="p+ anti-Delta-- anti-K0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="n0 anti-p- anti-K0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Delta- anti-n0 anti-K0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Delta++  anti-Delta-- K-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="p+ anti-p- K-", model="PHOTOS PHSP;", bf=0.00000624),
    BelleBBFEntry(decay="n0 anti-n0 K-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Xi0 anti-Sigma- pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Lambda0 pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Sigma0 pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Sigma- pi+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-Lambda0 pi-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-Sigma0 pi-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Sigma+ pi-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-p- K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-Delta-- K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-n0 K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-p- K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Lambda0 anti-Sigma- anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma0 anti-Sigma- anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma+ anti-Sigma- K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma- anti-Lambda0 anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma- anti-Sigma0 anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0 K-", model="PHOTOS PHSP;", bf=0.00000290),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0 K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0 K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0 K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Xi0 pi0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Omega- anti-Xi+ pi-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Omega- anti-Lambda0 K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Omega- anti-Sigma0 K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Omega- anti-Sigma- K+", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Xi- anti-Xi0 anti-K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Xi0 anti-Xi0 K-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Sigma+  anti-Delta-- rho0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Lambda0 anti-p- rho0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Sigma0 anti-p- rho0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Sigma- anti-n0 rho0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Lambda0  anti-Delta-- rho+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Sigma0  anti-Delta-- rho+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Sigma- anti-p- rho+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Sigma+ anti-p- rho-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Lambda0 anti-n0 rho-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Sigma0 anti-n0 rho-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="p+ anti-Delta-- anti-K*0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="n0 anti-p- anti-K*0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Delta- anti-n0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Delta++  anti-Delta-- K*-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="p+ anti-p- K*-", model="PHOTOS PHSP;", bf=0.00000660),
    BelleBBFEntry(decay="n0 anti-n0 K*-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="Xi0 anti-Sigma- rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-Lambda0 rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-Sigma0 rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-Sigma- rho+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-Lambda0 rho-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-Sigma0 rho-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-Sigma+ rho-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-p- K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-Delta-- K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-n0 K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-p- K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Lambda0 anti-Sigma- anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma0 anti-Sigma- anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma+ anti-Sigma- K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma- anti-Lambda0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma- anti-Sigma0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0 K*-", model="PHOTOS PHSP;", bf=0.00000210),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0 K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0 K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0 K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Omega- anti-Xi0 rho0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Xi+ rho-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Lambda0 K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Sigma0 K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Sigma- K*+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Xi0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-Xi0 K*-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="p+ anti-Delta-- pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0 anti-p-	    pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-n0  pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="p+ anti-p- pi-", model="PHOTOS PHSP;", bf=0.00000169),
    BelleBBFEntry(decay="n0 anti-Delta-- pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta++ anti-Delta-- pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0 anti-n0	    pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-p-  pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-Delta+  pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Lambda0 anti-Sigma-	pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma-	pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Lambda0	pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Sigma0	pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0 pi-", model="PHOTOS PHSP;", bf=0.00000280),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0	pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0	pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0	pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Sigma-	pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma+ anti-Sigma-	pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Sigma+	pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-p-	K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-p-	K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Delta-- 	K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Delta-- 	K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-n0 	K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Delta- K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Sigma-	anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="p+ anti-Sigma- K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Delta- anti-Lambda0	anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Delta- anti-Sigma0	anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Lambda0	K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Sigma0	K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Xi- anti-Xi0	pi0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Xi+	pi-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi0 anti-Xi0	pi-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Lambda0	K0", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="Xi- anti-Sigma0	K0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Sigma-	K+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma- anti-Xi0	anti-K0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Lambda0 anti-Xi0	K-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma0 anti-Xi0	K-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="p+ anti-Delta-- rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0 anti-p-	    rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-n0  rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="p+ anti-p- rho-", model="PHOTOS PHSP;", bf=0.00000306),
    BelleBBFEntry(decay="n0 anti-Delta-- rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta++ anti-Delta-- rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0 anti-n0	    rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-p-  rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-Delta+  rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Lambda0 anti-Sigma-	rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma-	rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Lambda0	rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Sigma0	rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0 rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0	rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0	rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0	rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Sigma-	rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma+ anti-Sigma-	rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Sigma+	rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-p-	K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-p-	K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Delta-- 	K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Delta-- 	K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-n0 	K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Delta- K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Sigma-	anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="p+ anti-Sigma- K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Delta- anti-Lambda0	anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Delta- anti-Sigma0	anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Lambda0	K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Sigma0	K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Xi- anti-Xi0	rho0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Xi+	rho-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi0 anti-Xi0	rho-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Lambda0	K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Sigma0	K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Sigma-	K*+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma- anti-Xi0	anti-K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Lambda0 anti-Xi0	K*-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma0 anti-Xi0	K*-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma+  anti-Delta--", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Lambda0 anti-p-", model="PHOTOS PHSP;", bf=0.00000032),
    BelleBBFEntry(decay="Lambda0 anti-Delta-", model="PHOTOS PHSP;", bf=0.00000082),
    BelleBBFEntry(decay="Sigma0 anti-p-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Sigma*0 anti-p-", model="PHOTOS PHSP;", bf=0.00000047),
    BelleBBFEntry(decay="Sigma- anti-n0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Xi0 anti-Sigma-", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="Xi- anti-Lambda0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="Xi- anti-Sigma0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="Omega- anti-Xi0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="p+ anti-Delta--", model="PHOTOS PHSP;", bf=0.00000014),
    BelleBBFEntry(decay="n0 anti-p-", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="Delta- anti-n0", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="Lambda0 anti-Sigma-", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Sigma0 anti-Sigma-", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Sigma- anti-Lambda0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Sigma- anti-Sigma0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Xi- anti-Xi0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Lambda(1520)0 anti-p-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="D_s-  pi0", model="PHOTOS PHSP;", bf=0.00001080),
    BelleBBFEntry(decay="D_s-  eta", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="D_s-  eta'", model="PHOTOS PHSP;", bf=0.00000720),
    BelleBBFEntry(decay="rho0  D_s-", model="PHOTOS SVS;", bf=0.00002840),
    BelleBBFEntry(decay="D_s-  pi+  pi-", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="D_s-  pi0  pi0", model="PHOTOS PHSP;", bf=0.00000320),
    BelleBBFEntry(decay="a_10  D_s-", model="PHOTOS SVS;", bf=0.00002160),
    BelleBBFEntry(decay="omega D_s-", model="PHOTOS SVS;", bf=0.00002160),
    BelleBBFEntry(decay="D_s*- pi0", model="PHOTOS SVS;", bf=0.00000960),
    BelleBBFEntry(decay="D_s*- eta", model="PHOTOS SVS;", bf=0.00000680),
    BelleBBFEntry(decay="D_s*- eta'", model="PHOTOS SVS;", bf=0.00000640),
    BelleBBFEntry(decay="D_s*- rho0", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00002440),
    BelleBBFEntry(decay="D_s*- pi+ pi-", model="PHOTOS PHSP;", bf=0.00002880),
    BelleBBFEntry(decay="D_s*- pi0 pi0", model="PHOTOS PHSP;", bf=0.00000960),
    BelleBBFEntry(decay="D_s*- a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00004720),
    BelleBBFEntry(decay="D_s*- omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00004720),
    BelleBBFEntry(decay="D-  pi0", model="PHOTOS PHSP;", bf=0.00000054),
    BelleBBFEntry(decay="D-  eta", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="D-  eta'", model="PHOTOS PHSP;", bf=0.00000036),
    BelleBBFEntry(decay="rho0  D-", model="PHOTOS SVS;", bf=0.00000142),
    BelleBBFEntry(decay="D-  pi+  pi-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="D-  pi0  pi0", model="PHOTOS PHSP;", bf=0.00000016),
    BelleBBFEntry(decay="a_10  D-", model="PHOTOS SVS;", bf=0.00000108),
    BelleBBFEntry(decay="omega D-", model="PHOTOS SVS;", bf=0.00000108),
    BelleBBFEntry(decay="D*- pi0", model="PHOTOS SVS;", bf=0.00000048),
    BelleBBFEntry(decay="D*- eta", model="PHOTOS SVS;", bf=0.00000034),
    BelleBBFEntry(decay="D*- eta'", model="PHOTOS SVS;", bf=0.00000032),
    BelleBBFEntry(decay="D*- rho0", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000122),
    BelleBBFEntry(decay="D*- pi+ pi-", model="PHOTOS PHSP;", bf=0.00000144),
    BelleBBFEntry(decay="D*- pi0 pi0", model="PHOTOS PHSP;", bf=0.00000048),
    BelleBBFEntry(decay="D*- a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000236),
    BelleBBFEntry(decay="D*- omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000236),
    BelleBBFEntry(decay="D- anti-K0", model="PHOTOS PHSP;", bf=0.00000008),
    BelleBBFEntry(decay="anti-K*0  D-", model="PHOTOS SVS;", bf=0.00000022),
    BelleBBFEntry(decay="D- K- pi+", model="PHOTOS PHSP;", bf=0.00000002),
    BelleBBFEntry(decay="D- anti-K0 pi0", model="PHOTOS PHSP;", bf=0.00000001),
    BelleBBFEntry(decay="D*- anti-K0", model="PHOTOS SVS;", bf=0.00000007),
    BelleBBFEntry(decay="D*- anti-K*0", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000018),
    BelleBBFEntry(decay="D*- K- pi+", model="PHOTOS PHSP;", bf=0.00000015),
    BelleBBFEntry(decay="D*- anti-K0 pi0", model="PHOTOS PHSP;", bf=0.00000007),
    BelleBBFEntry(decay="phi D_s-", model="PHOTOS SVS;", bf=0.00000030),
    BelleBBFEntry(decay="phi D_s*-", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000027),
    BelleBBFEntry(decay="D_s- K0", model="PHOTOS PHSP;", bf=0.00000002),
    BelleBBFEntry(decay="D_s- K+ pi-", model="PHOTOS PHSP;", bf=0.00000001),
    BelleBBFEntry(decay="D_s- K0 pi0", model="PHOTOS PHSP;", bf=0.00000001),
    BelleBBFEntry(decay="K*0 D_s-", model="PHOTOS SVS;", bf=0.00000005),
    BelleBBFEntry(decay="D_s*- K0", model="PHOTOS SVS;", bf=0.00000004),
    BelleBBFEntry(decay="D_s*- K+ pi-", model="PHOTOS PHSP;", bf=0.00000012),
    BelleBBFEntry(decay="D_s*- K0 pi0", model="PHOTOS PHSP;", bf=0.00000004),
    BelleBBFEntry(decay="D_s*- K*0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="D_s*- gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000020),
    BelleBBFEntry(decay="D*-   gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000001),
    BelleBBFEntry(decay="mu-  anti-nu_mu", model="PHOTOS  SLN;", bf=0.00000042),
    BelleBBFEntry(decay="tau- anti-nu_tau", model="PHOTOS  SLN;", bf=0.00009300),
    BelleBBFEntry(decay="e-  anti-nu_e gamma", model="PHOTOS  LNUGAMMA 0.35 3.0 5.0 0;", bf=0.000005),
    BelleBBFEntry(decay="mu- anti-nu_mu gamma", model="PHOTOS  LNUGAMMA 0.35 3.0 5.0 0;", bf=0.000005),
)  # type: Tuple[BelleBBFEntry, ...]


bp_rare_bfs = (
    BelleBBFEntry(decay="K*+       gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.0000403),
    BelleBBFEntry(decay="K_1+      gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.000043),
    BelleBBFEntry(decay="K_2*+     gamma", model="PHOTOS HELAMP 1.0 0.0 1.0 0.0;", bf=0.0000140),
    BelleBBFEntry(decay="Xsu       gamma", model="BTOXSGAMMA 2;", bf=0.000257),
    BelleBBFEntry(decay="rho+      gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000088),
    BelleBBFEntry(decay="Xdu+      gamma", model="BTOXDGAMMA 2;", bf=0.0000151),
    BelleBBFEntry(decay="K+        e+    e-", model="PHOTOS BTOSLLBALL;", bf=0.00000049),
    BelleBBFEntry(decay="K*+       e+    e-", model="PHOTOS BTOSLLBALL;", bf=0.00000080),
    BelleBBFEntry(decay="Xsu       e+    e-", model="BTOXSLL;", bf=0.00000341),
    BelleBBFEntry(decay="K+        mu+   mu-", model="PHOTOS BTOSLLBALL;", bf=0.00000039),
    BelleBBFEntry(decay="K*+       mu+   mu-", model="PHOTOS BTOSLLBALL;", bf=0.00000080),
    BelleBBFEntry(decay="Xsu       mu+   mu-", model="BTOXSLL;", bf=0.00000311),
    BelleBBFEntry(decay="K+   tau+  tau-", model="PHOTOS BTOSLLBALL;", bf=0.00000013),
    BelleBBFEntry(decay="K*+  tau+  tau-", model="PHOTOS BTOSLLBALL;", bf=0.00000019),
    BelleBBFEntry(decay="Xsu  tau+  tau-", model="BTOXSLL;", bf=0.00000022),
    BelleBBFEntry(decay="K+       nu_e   anti-nu_e", model="PHOTOS PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*+      nu_e   anti-nu_e", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="Xsu      nu_e   anti-nu_e", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K+       nu_mu  anti-nu_mu", model="PHOTOS PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*+      nu_mu  anti-nu_mu", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="Xsu      nu_mu  anti-nu_mu", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K+       nu_tau anti-nu_tau", model="PHOTOS PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*+      nu_tau anti-nu_tau", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="Xsu      nu_tau anti-nu_tau", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K+ anti-K0", model="PHOTOS PHSP;", bf=0.00000136),
    BelleBBFEntry(decay="K+ pi0", model="PHOTOS PHSP;", bf=0.00001290),
    BelleBBFEntry(decay="K0 pi+", model="PHOTOS PHSP;", bf=0.00002310),
    BelleBBFEntry(decay="pi+ pi0", model="PHOTOS PHSP;", bf=0.00000559),
    BelleBBFEntry(decay="K'*+ pi0", model="PHOTOS PHSP;", bf=0.00000820),
    BelleBBFEntry(decay="K'*0 pi+", model="PHOTOS PHSP;", bf=0.00004500),
    BelleBBFEntry(decay="K_0*0 pi+", model="PHOTOS PHSP;", bf=0.00004710),
    BelleBBFEntry(decay="anti-K_0*0 K+", model="PHOTOS PHSP;", bf=0.00000220),
    BelleBBFEntry(decay="f'_0 K+", model="PHOTOS PHSP;", bf=0.00001070),
    BelleBBFEntry(decay="f_0 K+", model="PHOTOS PHSP;", bf=0.00000920),
    BelleBBFEntry(decay="f_0(1500) K+", model="PHOTOS PHSP;", bf=0.00000440),
    BelleBBFEntry(decay="f'_0 pi+", model="PHOTOS PHSP;", bf=0.00000300),
    BelleBBFEntry(decay="f_0 pi+", model="PHOTOS PHSP;", bf=0.00000300),
    BelleBBFEntry(decay="f_0(600) pi+", model="PHOTOS PHSP;", bf=0.00000410),
    BelleBBFEntry(decay="K_2*0 pi+", model="PHOTOS STS;", bf=0.00000690),
    BelleBBFEntry(decay="f'_2 K+", model="PHOTOS STS;", bf=0.00000340),
    BelleBBFEntry(decay="f_2 K+", model="PHOTOS STS;", bf=0.00000133),
    BelleBBFEntry(decay="a_00 K+", model="PHOTOS PHSP;", bf=0.00000250),
    BelleBBFEntry(decay="a_0+ K0", model="PHOTOS PHSP;", bf=0.00000390),
    BelleBBFEntry(decay="a_00 pi+", model="PHOTOS PHSP;", bf=0.00000580),
    BelleBBFEntry(decay="a_0+ pi0", model="PHOTOS PHSP;", bf=0.00000140),
    BelleBBFEntry(decay="K''*0 pi+", model="PHOTOS SVS;", bf=0.00001200),
    BelleBBFEntry(decay="K*+ anti-K0", model="PHOTOS SVS;", bf=0.00000030),
    BelleBBFEntry(decay="K*+ pi0", model="PHOTOS SVS;", bf=0.00000690),
    BelleBBFEntry(decay="K*+ f_0", model="PHOTOS SVS;", bf=0.00000520),
    BelleBBFEntry(decay="K*0 pi+", model="PHOTOS SVS;", bf=0.00001070),
    BelleBBFEntry(decay="anti-K*0 K+", model="PHOTOS SVS;", bf=0.00000110),
    BelleBBFEntry(decay="rho0 K+", model="PHOTOS SVS;", bf=0.00000425),
    BelleBBFEntry(decay="rho(2S)0 K+", model="PHOTOS SVS;", bf=0.00001170),
    BelleBBFEntry(decay="rho(2S)0 K+", model="PHOTOS SVS;", bf=0.00000500),
    BelleBBFEntry(decay="rho+ K0", model="PHOTOS SVS;", bf=0.00000800),
    BelleBBFEntry(decay="rho(2S)+ K0", model="PHOTOS SVS;", bf=0.00000500),
    BelleBBFEntry(decay="rho(3S)+ K0", model="PHOTOS SVS;", bf=0.00000250),
    BelleBBFEntry(decay="rho+ pi0", model="PHOTOS SVS;", bf=0.00001090),
    BelleBBFEntry(decay="rho(2S)+ pi0", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="rho(3S)+ pi0", model="PHOTOS SVS;", bf=0.00000100),
    BelleBBFEntry(decay="rho0 pi+", model="PHOTOS SVS;", bf=0.00000870),
    BelleBBFEntry(decay="rho(2S)0 pi+", model="PHOTOS SVS;", bf=0.00000230),
    BelleBBFEntry(decay="rho(3S)0 pi+", model="PHOTOS SVS;", bf=0.00000100),
    BelleBBFEntry(decay="a_10 K+", model="PHOTOS SVS;", bf=0.00002700),
    BelleBBFEntry(decay="a_1+ K0", model="PHOTOS SVS;", bf=0.00003490),
    BelleBBFEntry(decay="a_10 pi+", model="PHOTOS SVS;", bf=0.00002040),
    BelleBBFEntry(decay="a_1+ pi0", model="PHOTOS SVS;", bf=0.00002640),
    BelleBBFEntry(decay="rho+ f_0", model="PHOTOS SVS;", bf=0.00000190),
    BelleBBFEntry(decay="b_10 K+", model="PHOTOS SVS;", bf=0.00000910),
    BelleBBFEntry(decay="b_10 pi+", model="PHOTOS SVS;", bf=0.00000670),
    BelleBBFEntry(decay="f_2  pi+", model="PHOTOS STS;", bf=0.00000820),
    BelleBBFEntry(decay="a_20 K+", model="PHOTOS STS;", bf=0.00000110),
    BelleBBFEntry(decay="K+ K+ K-", model="PHOTOS PHSP;", bf=0.00002501),
    BelleBBFEntry(decay="K+ K- pi+", model="PHOTOS PHSP;", bf=0.00000268),
    BelleBBFEntry(decay="K+ K_S0 K_S0", model="PHOTOS PHSP;", bf=0.00000944),
    BelleBBFEntry(decay="K+ anti-K0 pi0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="K+ pi+ pi-", model="PHOTOS PHSP;", bf=0.00000290),
    BelleBBFEntry(decay="K0 pi+ pi0", model="PHOTOS PHSP;", bf=0.00003086),
    BelleBBFEntry(decay="pi+ pi+ pi-", model="PHOTOS PHSP;", bf=0.00000460),
    BelleBBFEntry(decay="K+ K+ pi-", model="PHOTOS PHSP;", bf=0.00000130),
    BelleBBFEntry(decay="K- pi+ pi+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="K_S0 K_S0 pi+", model="PHOTOS PHSP;", bf=0.00000320),
    BelleBBFEntry(decay="K_L0 K_L0 pi+", model="PHOTOS PHSP;", bf=0.00000320),
    BelleBBFEntry(decay="K_S0 K_L0 pi+", model="PHOTOS PHSP;", bf=0.00000640),
    BelleBBFEntry(decay="f'_2 K*+", model="PHOTOS PHSP;", bf=0.00003700),
    BelleBBFEntry(decay="f_2 K*+", model="PHOTOS PHSP;", bf=0.00001800),
    BelleBBFEntry(decay="rho+ K_0*0", model="PHOTOS PHSP;", bf=0.00001200),
    BelleBBFEntry(decay="rho0 K_0*+", model="PHOTOS PHSP;", bf=0.00001200),
    BelleBBFEntry(decay="K*+ anti-K*0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00007100),
    BelleBBFEntry(decay="K*+ rho0", model="PHOTOS SVV_HELAMP 0.0 0.0 1.0 0.0 0.0 0.0;", bf=0.00000610),
    BelleBBFEntry(decay="K*0 rho+", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000920),
    BelleBBFEntry(decay="rho+ rho0", model="PHOTOS SVV_HELAMP 0.0 0.0 1.0 0.0 0.0 0.0;", bf=0.00001820),
    BelleBBFEntry(decay="K*+ pi+ pi-", model="PHOTOS PHSP;", bf=0.00006650),
    BelleBBFEntry(decay="K*+ pi0 pi0", model="PHOTOS PHSP;", bf=0.00000260),
    BelleBBFEntry(decay="K*0 pi+ pi0", model="PHOTOS PHSP;", bf=0.00000750),
    BelleBBFEntry(decay="rho+ K+ pi-", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="rho+ K0 pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="rho0 K+ pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="rho0 K0 pi+", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="K+ K- K*+", model="PHOTOS PHSP;", bf=0.00003130),
    BelleBBFEntry(decay="K0 anti-K0 K*+", model="PHOTOS PHSP;", bf=0.00001500),
    BelleBBFEntry(decay="K0 anti-K*0 K+", model="PHOTOS PHSP;", bf=0.00001500),
    BelleBBFEntry(decay="K*0 anti-K0 K+", model="PHOTOS PHSP;", bf=0.00001500),
    BelleBBFEntry(decay="K+ K*- K+", model="PHOTOS PHSP;", bf=0.00001500),
    BelleBBFEntry(decay="K*+ K- pi+", model="PHOTOS PHSP;", bf=0.00000000),
    BelleBBFEntry(decay="K*+ K+ pi-", model="PHOTOS PHSP;", bf=0.00000610),
    BelleBBFEntry(decay="K*- K+ pi+", model="PHOTOS PHSP;", bf=0.00000000),
    BelleBBFEntry(decay="eta   K+", model="PHOTOS PHSP;", bf=0.0000027),
    BelleBBFEntry(decay="eta'  K+", model="PHOTOS PHSP;", bf=0.0000702),
    BelleBBFEntry(decay="K_0*+ eta", model="PHOTOS PHSP;", bf=0.0000180),
    BelleBBFEntry(decay="eta(2S)   K+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta(1405)  K+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta   pi+", model="PHOTOS PHSP;", bf=0.0000044),
    BelleBBFEntry(decay="eta'  pi+", model="PHOTOS PHSP;", bf=0.0000027),
    BelleBBFEntry(decay="eta(2S)   pi+", model="PHOTOS PHSP;", bf=0.0000020),
    BelleBBFEntry(decay="eta(1405)  pi+", model="PHOTOS PHSP;", bf=0.0000020),
    BelleBBFEntry(decay="omega K+", model="PHOTOS SVS;", bf=0.0000067),
    BelleBBFEntry(decay="phi   K+", model="PHOTOS SVS;", bf=0.0000083),
    BelleBBFEntry(decay="omega pi+", model="PHOTOS SVS;", bf=0.0000069),
    BelleBBFEntry(decay="phi   pi+", model="PHOTOS SVS;", bf=0.00000024),
    BelleBBFEntry(decay="phi(1680) K+", model="PHOTOS SVS;", bf=0.0000008),
    BelleBBFEntry(decay="K*+	 eta", model="PHOTOS SVS;", bf=0.0000193),
    BelleBBFEntry(decay="K*+   eta'", model="PHOTOS SVS;", bf=0.0000049),
    BelleBBFEntry(decay="K*+	 eta(2S)", model="PHOTOS SVS;", bf=0.0000020),
    BelleBBFEntry(decay="K*+   eta(1405)", model="PHOTOS SVS;", bf=0.0000020),
    BelleBBFEntry(decay="rho+  eta", model="PHOTOS SVS;", bf=0.0000054),
    BelleBBFEntry(decay="rho+  eta'", model="PHOTOS SVS;", bf=0.0000091),
    BelleBBFEntry(decay="rho+  eta(2S)", model="PHOTOS SVS;", bf=0.0000020),
    BelleBBFEntry(decay="rho+  eta(1405)", model="PHOTOS SVS;", bf=0.0000020),
    BelleBBFEntry(decay="omega K*+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000034),
    BelleBBFEntry(decay="K*+   phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000100),
    BelleBBFEntry(decay="omega rho+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000106),
    BelleBBFEntry(decay="phi   rho+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000160),
    BelleBBFEntry(decay="K_2*+ eta", model="PHOTOS STS;", bf=0.0000091),
    BelleBBFEntry(decay="Xsu eta'", model="PHOTOS BTOXSETAP;", bf=0.0003449),
    BelleBBFEntry(decay="Xsu eta", model="PHOTOS BTOXSETAP;", bf=0.0003872),
    BelleBBFEntry(decay="eta  K0 pi+", model="PHOTOS PHSP;", bf=0.0000145),
    BelleBBFEntry(decay="eta  K+ pi0", model="PHOTOS PHSP;", bf=0.0000072),
    BelleBBFEntry(decay="eta  pi0 pi+", model="PHOTOS PHSP;", bf=0.0000072),
    BelleBBFEntry(decay="eta'  K0 pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K+ pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  pi0 pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta K+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta' K+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta' pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta' K+", model="PHOTOS PHSP;", bf=0.0000250),
    BelleBBFEntry(decay="eta'  eta' pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  K*0 pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  K0 rho+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  K*+ pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  K+ rho0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  rho0 pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  pi0 rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  K*0 pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K0 rho+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K*+ pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K+ rho0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  rho0 pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  pi0 rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta K*+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta' K*+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta' rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta' K*+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  eta' rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  K0 pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  K+ pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  pi0 pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  K0 pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  K+ pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  pi0 pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  eta K+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  eta pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  eta' K+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  eta' pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  eta K+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  eta pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  eta' K+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  eta' pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  phi  K+", model="PHOTOS PHSP;", bf=0.0000042),
    BelleBBFEntry(decay="a_0+ eta", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="a_0+ eta'", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="a_0+ eta(2S)", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="a_0+ eta(1405)", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="a_0+ a_00", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="a_0+ f_0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="K*+ a_00", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="K*0 a_0+", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="rho+ a_00", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="rho0 a_0+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="omega a_0+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="phi a_0+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_10 a_0+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_1+ a_00", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="f_1 a_0+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_10 a_0+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_1+ a_00", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="h_1 a_0+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_1+ eta", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_1+ eta'", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="a_1+ eta(2S)", model="PHOTOS SVS;", bf=0.00000020),
    BelleBBFEntry(decay="a_1+ eta(1405)", model="PHOTOS SVS;", bf=0.00000020),
    BelleBBFEntry(decay="a_1+ f_0", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="K*+ a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="K*0 a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho+ a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="rho0 a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="omega a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="phi a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="a_10 a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="f_1 a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_10 a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_1+ a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="h_1  a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_1+ f_0", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="f_1 K+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="f_1 pi+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="K*+ f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="rho+ f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_1+ f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_1+ K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="b_1+ pi0", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_1+ eta", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_1+ eta'", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="b_1+ eta(2S)", model="PHOTOS SVS;", bf=0.00000020),
    BelleBBFEntry(decay="b_1+ eta(1405)", model="PHOTOS SVS;", bf=0.00000020),
    BelleBBFEntry(decay="K*+ b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="K*0 b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho+ b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="rho0 b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="omega b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="phi b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="b_10 b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="h_1  b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="h_1 K+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="h_1 pi+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="K*+ h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00001310),
    BelleBBFEntry(decay="rho+ h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="a_20  pi+", model="PHOTOS STS;", bf=0.00000260),
    BelleBBFEntry(decay="f'_2  pi+", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+  pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ eta", model="PHOTOS STS;", bf=0.00000029),
    BelleBBFEntry(decay="a_2+ eta'", model="PHOTOS STS;", bf=0.00000130),
    BelleBBFEntry(decay="a_2+ eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*+ anti-K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K_2*0 K+", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*+ pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*+ eta'", model="PHOTOS STS;", bf=0.00000140),
    BelleBBFEntry(decay="K_2*+ eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*+ eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 rho+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000730),
    BelleBBFEntry(decay="f_2 rho+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000810),
    BelleBBFEntry(decay="f'_2 rho+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*+ anti-K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K_2*0 K*+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 K*+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000190),
    BelleBBFEntry(decay="a_2+ K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000450),
    BelleBBFEntry(decay="K_2*+ rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000025),
    BelleBBFEntry(decay="K_2*+ omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000011),
    BelleBBFEntry(decay="K_2*+ phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000220),
    BelleBBFEntry(decay="K_2*0 rho+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000025),
    BelleBBFEntry(decay="anti-Sigma-  Delta++ pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Lambda0 p+ pi0", model="PHOTOS PHSP;", bf=0.00000300),
    BelleBBFEntry(decay="anti-Sigma0 p+ pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Sigma+ n0 pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Lambda0  Delta++ pi-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Sigma0  Delta++ pi-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Sigma+ p+ pi-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Sigma- p+ pi+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Lambda0 n0 pi+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Sigma0 n0 pi+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-p- Delta++ K0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-n0 p+ K0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Delta+ n0 K0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Delta--  Delta++ K+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-p- p+ K+", model="PHOTOS PHSP;", bf=0.00000624),
    BelleBBFEntry(decay="anti-n0 n0 K+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Xi0 Sigma+ pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Lambda0 pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Sigma0 pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Sigma+ pi-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 Lambda0 pi+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 Sigma0 pi+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Sigma- pi+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 p+ anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 Delta++ K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ n0 anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ p+ K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Lambda0 Sigma+ K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma0 Sigma+ K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma- Sigma+ K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma+ Lambda0 K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma+ Sigma0 K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0 K+", model="PHOTOS PHSP;", bf=0.00000290),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0 K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0 K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0 K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Xi0 pi0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Omega+ Xi- pi+", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Omega+ Lambda0 anti-K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Omega+ Sigma0 anti-K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Omega+ Sigma+ K-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Xi+ Xi0 K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Xi0 Xi0 K+", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Sigma-  Delta++ rho0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Lambda0 p+ rho0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Sigma0 p+ rho0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Sigma+ n0 rho0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Lambda0  Delta++ rho-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Sigma0  Delta++ rho-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Sigma+ p+ rho-", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Sigma- p+ rho+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Lambda0 n0 rho+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Sigma0 n0 rho+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-p- Delta++ K*0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-n0 p+ K*0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Delta+ n0 K*0", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Delta--  Delta++ K*+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-p- p+ K*+", model="PHOTOS PHSP;", bf=0.00000660),
    BelleBBFEntry(decay="anti-n0 n0 K*+", model="PHOTOS PHSP;", bf=0.00000570),
    BelleBBFEntry(decay="anti-Xi0 Sigma+ rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ Lambda0 rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ Sigma0 rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ Sigma+ rho-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 Lambda0 rho+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 Sigma0 rho+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ Sigma- rho+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 p+ anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 Delta++ K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ n0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ p+ K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Lambda0 Sigma+ K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma0 Sigma+ K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma- Sigma+ K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma+ Lambda0 K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma+ Sigma0 K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0 K*+", model="PHOTOS PHSP;", bf=0.00000210),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0 K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0 K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0 K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Omega+ Xi0 rho0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Xi- rho+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Lambda0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Sigma0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Sigma+ K*-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Xi0 K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 Xi0 K*+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-p- Delta++ pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0 p+	    pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ n0  pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-p- p+ pi+", model="PHOTOS PHSP;", bf=0.00000169),
    BelleBBFEntry(decay="anti-n0 Delta++ pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta-- Delta++ pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0 n0	    pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ p+  pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ Delta-  pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Lambda0 Sigma+	pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma+	pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Lambda0	pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Sigma0	pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0 pi+", model="PHOTOS PHSP;", bf=0.00000280),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0	pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0	pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0	pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Sigma+	pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma- Sigma+	pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Sigma-	pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 p+	anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 p+	anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Delta++ 	K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Delta++ 	K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ n0 	anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Delta+ K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Sigma+	anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-p- Sigma+ K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Delta+ Lambda0	K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Delta+ Sigma0	K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Lambda0	K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Sigma0	K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Xi+ Xi0	pi0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Xi-	pi+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi0 Xi0	pi+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Lambda0	anti-K0", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="anti-Xi+ Sigma0	anti-K0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Sigma+	K-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma+ Xi0	K0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Lambda0 Xi0	K+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma0 Xi0	K+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-p- Delta++ rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0 p+	    rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ n0  rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-p- p+ rho+", model="PHOTOS PHSP;", bf=0.00000306),
    BelleBBFEntry(decay="anti-n0 Delta++ rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta-- Delta++ rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0 n0	    rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ p+  rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ Delta-  rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Lambda0 Sigma+	rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma+	rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Lambda0	rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Sigma0	rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0 rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0	rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0	rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0	rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Sigma+	rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma- Sigma+	rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Sigma-	rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 p+	anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 p+	anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Delta++ 	K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Delta++ 	K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ n0 	anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Delta+ K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Sigma+	K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-p- Sigma+ K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Delta+ Lambda0	K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Delta+ Sigma0	K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Lambda0	K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Sigma0	K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Xi+ Xi0	rho0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Xi-	rho+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi0 Xi0	rho+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Lambda0	anti-K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Sigma0	anti-K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Sigma+	K*-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma+ Xi0	K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Lambda0 Xi0	K*+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma0 Xi0	K*+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma-  Delta++", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Lambda0 p+", model="PHOTOS PHSP;", bf=0.00000032),
    BelleBBFEntry(decay="anti-Lambda0 Delta+", model="PHOTOS PHSP;", bf=0.00000082),
    BelleBBFEntry(decay="anti-Sigma0 p+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Sigma*0 p+", model="PHOTOS PHSP;", bf=0.00000047),
    BelleBBFEntry(decay="anti-Sigma+ n0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Xi0 Sigma+", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="anti-Xi+ Lambda0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="anti-Xi+ Sigma0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="anti-Omega+ Xi0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-p- Delta++", model="PHOTOS PHSP;", bf=0.00000014),
    BelleBBFEntry(decay="anti-n0 p+", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="anti-Delta+ n0", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="anti-Lambda0 Sigma+", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Sigma0 Sigma+", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Sigma+ Lambda0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Sigma+ Sigma0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Xi+ Xi0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Lambda(1520)0 p+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="D_s+  pi0", model="PHOTOS PHSP;", bf=0.00001080),
    BelleBBFEntry(decay="D_s+  eta", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="D_s+  eta'", model="PHOTOS PHSP;", bf=0.00000720),
    BelleBBFEntry(decay="rho0  D_s+", model="PHOTOS SVS;", bf=0.00002840),
    BelleBBFEntry(decay="D_s+  pi+  pi-", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="D_s+  pi0  pi0", model="PHOTOS PHSP;", bf=0.00000320),
    BelleBBFEntry(decay="a_10  D_s+", model="PHOTOS SVS;", bf=0.00002160),
    BelleBBFEntry(decay="omega D_s+", model="PHOTOS SVS;", bf=0.00002160),
    BelleBBFEntry(decay="D_s*+ pi0", model="PHOTOS SVS;", bf=0.00000960),
    BelleBBFEntry(decay="D_s*+ eta", model="PHOTOS SVS;", bf=0.00000680),
    BelleBBFEntry(decay="D_s*+ eta'", model="PHOTOS SVS;", bf=0.00000640),
    BelleBBFEntry(decay="D_s*+ rho0", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00002440),
    BelleBBFEntry(decay="D_s*+ pi+ pi-", model="PHOTOS PHSP;", bf=0.00002880),
    BelleBBFEntry(decay="D_s*+ pi0 pi0", model="PHOTOS PHSP;", bf=0.00000960),
    BelleBBFEntry(decay="D_s*+ a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00004720),
    BelleBBFEntry(decay="D_s*+ omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00004720),
    BelleBBFEntry(decay="D+  pi0", model="PHOTOS PHSP;", bf=0.00000054),
    BelleBBFEntry(decay="D+  eta", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="D+  eta'", model="PHOTOS PHSP;", bf=0.00000036),
    BelleBBFEntry(decay="rho0  D+", model="PHOTOS SVS;", bf=0.00000142),
    BelleBBFEntry(decay="D+  pi+  pi-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="D+  pi0  pi0", model="PHOTOS PHSP;", bf=0.00000016),
    BelleBBFEntry(decay="a_10  D+", model="PHOTOS SVS;", bf=0.00000108),
    BelleBBFEntry(decay="omega D+", model="PHOTOS SVS;", bf=0.00000108),
    BelleBBFEntry(decay="D*+ pi0", model="PHOTOS SVS;", bf=0.00000048),
    BelleBBFEntry(decay="D*+ eta", model="PHOTOS SVS;", bf=0.00000034),
    BelleBBFEntry(decay="D*+ eta'", model="PHOTOS SVS;", bf=0.00000032),
    BelleBBFEntry(decay="D*+ rho0", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000122),
    BelleBBFEntry(decay="D*+ pi+ pi-", model="PHOTOS PHSP;", bf=0.00000144),
    BelleBBFEntry(decay="D*+ pi0 pi0", model="PHOTOS PHSP;", bf=0.00000048),
    BelleBBFEntry(decay="D*+ a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000236),
    BelleBBFEntry(decay="D*+ omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000236),
    BelleBBFEntry(decay="D+ K0", model="PHOTOS PHSP;", bf=0.00000008),
    BelleBBFEntry(decay="K*0  D+", model="PHOTOS SVS;", bf=0.00000022),
    BelleBBFEntry(decay="D+ K+ pi-", model="PHOTOS PHSP;", bf=0.00000002),
    BelleBBFEntry(decay="D+ K0 pi0", model="PHOTOS PHSP;", bf=0.00000001),
    BelleBBFEntry(decay="D*+ K0", model="PHOTOS SVS;", bf=0.00000007),
    BelleBBFEntry(decay="D*+ K*0", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000018),
    BelleBBFEntry(decay="D*+ K+ pi-", model="PHOTOS PHSP;", bf=0.00000015),
    BelleBBFEntry(decay="D*+ K0 pi0", model="PHOTOS PHSP;", bf=0.00000007),
    BelleBBFEntry(decay="phi D_s+", model="PHOTOS SVS;", bf=0.00000030),
    BelleBBFEntry(decay="phi D_s*+", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000027),
    BelleBBFEntry(decay="D_s+ anti-K0", model="PHOTOS PHSP;", bf=0.00000002),
    BelleBBFEntry(decay="D_s+ K- pi+", model="PHOTOS PHSP;", bf=0.00000001),
    BelleBBFEntry(decay="D_s+ anti-K0 pi0", model="PHOTOS PHSP;", bf=0.00000001),
    BelleBBFEntry(decay="anti-K*0 D_s+", model="PHOTOS SVS;", bf=0.00000005),
    BelleBBFEntry(decay="D_s*+ anti-K0", model="PHOTOS SVS;", bf=0.00000004),
    BelleBBFEntry(decay="D_s*+ K- pi+", model="PHOTOS PHSP;", bf=0.00000012),
    BelleBBFEntry(decay="D_s*+ anti-K0 pi0", model="PHOTOS PHSP;", bf=0.00000004),
    BelleBBFEntry(decay="D_s*+ anti-K*0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="D_s*+ gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000020),
    BelleBBFEntry(decay="D*+ gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000001),
    BelleBBFEntry(decay="mu+  nu_mu", model="PHOTOS  SLN;", bf=0.00000042),
    BelleBBFEntry(decay="tau+ nu_tau", model="PHOTOS  SLN;", bf=0.00009300),
    BelleBBFEntry(decay="e+  nu_e gamma", model="PHOTOS  LNUGAMMA 0.35 3.0 5.0 0;", bf=0.000005),
    BelleBBFEntry(decay="mu+ nu_mu gamma", model="PHOTOS  LNUGAMMA 0.35 3.0 5.0 0;", bf=0.000005),
)  # type: Tuple[BelleBBFEntry, ...]


anti_bz_rare_bfs = (
    BelleBBFEntry(decay="anti-K*0    gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.0000401),
    BelleBBFEntry(decay="anti-K_10   gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.000043),
    BelleBBFEntry(decay="anti-K_2*0  gamma", model="PHOTOS HELAMP 1.0 0.0 1.0 0.0;", bf=0.0000124),
    BelleBBFEntry(decay="anti-Xsd    gamma", model="BTOXSGAMMA 2;", bf=0.000258),
    BelleBBFEntry(decay="rho0        gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000093),
    BelleBBFEntry(decay="omega       gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000046),
    BelleBBFEntry(decay="Xdd    gamma", model="BTOXDGAMMA 2;", bf=0.0000146),
    BelleBBFEntry(decay="anti-K0   e+    e-", model="PHOTOS BTOSLLBALL;", bf=0.00000013),
    BelleBBFEntry(decay="anti-K*0  e+    e-", model="PHOTOS BTOSLLBALL;", bf=0.00000104),
    BelleBBFEntry(decay="anti-Xsd  e+    e-", model="BTOXSLL;", bf=0.00000353),
    BelleBBFEntry(decay="anti-K0   mu+   mu-", model="PHOTOS BTOSLLBALL;", bf=0.00000057),
    BelleBBFEntry(decay="anti-K*0  mu+   mu-", model="PHOTOS BTOSLLBALL;", bf=0.00000110),
    BelleBBFEntry(decay="anti-Xsd  mu+   mu-", model="BTOXSLL;", bf=0.00000263),
    BelleBBFEntry(decay="anti-K0   tau+  tau-", model="PHOTOS BTOSLLBALL;", bf=0.00000013),
    BelleBBFEntry(decay="anti-K*0  tau+  tau-", model="PHOTOS BTOSLLBALL;", bf=0.00000019),
    BelleBBFEntry(decay="anti-Xsd  tau+  tau-", model="BTOXSLL;", bf=0.00000022),
    BelleBBFEntry(decay="anti-K0   nu_e   anti-nu_e", model="PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="anti-K*0  nu_e   anti-nu_e", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="anti-Xsd  nu_e   anti-nu_e", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="anti-K0   nu_mu  anti-nu_mu", model="PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="anti-K*0  nu_mu  anti-nu_mu", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="anti-Xsd  nu_mu  anti-nu_mu", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="anti-K0   nu_tau anti-nu_tau", model="PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="anti-K*0  nu_tau anti-nu_tau", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="anti-Xsd  nu_tau anti-nu_tau", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K- K+", model="PHOTOS PHSP;", bf=0.00000015),
    BelleBBFEntry(decay="K- pi+", model="PHOTOS PHSP;", bf=0.00001940),
    BelleBBFEntry(decay="anti-K0 pi0", model="PHOTOS PHSP;", bf=0.00000990),
    BelleBBFEntry(decay="K_L0 K_L0", model="PHOTOS PHSP;", bf=0.00000048),
    BelleBBFEntry(decay="K_S0 K_S0", model="PHOTOS PHSP;", bf=0.00000048),
    BelleBBFEntry(decay="pi- pi+", model="PHOTOS PHSP;", bf=0.00000516),
    BelleBBFEntry(decay="pi0 pi0", model="PHOTOS PHSP;", bf=0.00000131),
    BelleBBFEntry(decay="K'*- pi+", model="PHOTOS PHSP;", bf=0.00008600),
    BelleBBFEntry(decay="anti-K'*0 pi0", model="PHOTOS PHSP;", bf=0.00001930),
    BelleBBFEntry(decay="K_0*- pi+", model="PHOTOS PHSP;", bf=0.00004970),
    BelleBBFEntry(decay="a_0- pi+", model="PHOTOS PHSP;", bf=0.00000280),
    BelleBBFEntry(decay="f'_0 anti-K0", model="PHOTOS PHSP;", bf=0.00000820),
    BelleBBFEntry(decay="f_0 anti-K0", model="PHOTOS PHSP;", bf=0.00000580),
    BelleBBFEntry(decay="f_0(1500) anti-K0", model="PHOTOS PHSP;", bf=0.00000340),
    BelleBBFEntry(decay="f_0 f_0", model="PHOTOS PHSP;", bf=0.00000016),
    BelleBBFEntry(decay="a_00 anti-K0", model="PHOTOS PHSP;", bf=0.00000780),
    BelleBBFEntry(decay="a_0+ K-", model="PHOTOS PHSP;", bf=0.00000160),
    BelleBBFEntry(decay="f_2 anti-K0", model="PHOTOS STS;", bf=0.00000250),
    BelleBBFEntry(decay="K_2*- pi+", model="PHOTOS STS;", bf=0.00000630),
    BelleBBFEntry(decay="anti-K_2*0 pi0", model="PHOTOS STS;", bf=0.00000315),
    BelleBBFEntry(decay="K*- K+", model="PHOTOS SVS;", bf=0.00000001),
    BelleBBFEntry(decay="K*- pi+", model="PHOTOS SVS;", bf=0.00000980),
    BelleBBFEntry(decay="K*+ K-", model="PHOTOS SVS;", bf=0.00000001),
    BelleBBFEntry(decay="anti-K*0 K0", model="PHOTOS SVS;", bf=0.00000190),
    BelleBBFEntry(decay="K*0 anti-K0", model="PHOTOS SVS;", bf=0.00000190),
    BelleBBFEntry(decay="anti-K*0 pi0", model="PHOTOS SVS;", bf=0.00000000),
    BelleBBFEntry(decay="anti-K*0 f_0", model="PHOTOS SVS;", bf=0.00000430),
    BelleBBFEntry(decay="rho(2S)0 anti-K0", model="PHOTOS SVS;", bf=0.00000900),
    BelleBBFEntry(decay="rho(3S)0 anti-K0", model="PHOTOS SVS;", bf=0.00000900),
    BelleBBFEntry(decay="rho- pi+", model="PHOTOS SVS;", bf=0.00002310),
    BelleBBFEntry(decay="rho+ K-", model="PHOTOS SVS;", bf=0.00001530),
    BelleBBFEntry(decay="rho+ pi-", model="PHOTOS SVS;", bf=0.00002310),
    BelleBBFEntry(decay="rho0 anti-K0", model="PHOTOS SVS;", bf=0.00000540),
    BelleBBFEntry(decay="rho0 pi0", model="PHOTOS SVS;", bf=0.00000180),
    BelleBBFEntry(decay="rho0 f_0", model="PHOTOS SVS;", bf=0.00000053),
    BelleBBFEntry(decay="rho(2S)- pi+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="rho(3S)- pi+", model="PHOTOS SVS;", bf=0.00000100),
    BelleBBFEntry(decay="rho(2S)+ pi-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="rho(3S)+ pi-", model="PHOTOS SVS;", bf=0.00000100),
    BelleBBFEntry(decay="rho(2S)+ K-", model="PHOTOS SVS;", bf=0.00000320),
    BelleBBFEntry(decay="rho(3S)+ K-", model="PHOTOS SVS;", bf=0.00000170),
    BelleBBFEntry(decay="K_1- pi+", model="PHOTOS SVS;", bf=0.00002520),
    BelleBBFEntry(decay="K'_1- pi+", model="PHOTOS SVS;", bf=0.00002180),
    BelleBBFEntry(decay="K''*- pi+", model="PHOTOS SVS;", bf=0.00001010),
    BelleBBFEntry(decay="anti-K''*0 pi0", model="PHOTOS SVS;", bf=0.00001010),
    BelleBBFEntry(decay="b_1+ K-", model="PHOTOS SVS;", bf=0.00000740),
    BelleBBFEntry(decay="b_1+ pi-", model="PHOTOS SVS;", bf=0.00001090),
    BelleBBFEntry(decay="b_1- pi+", model="PHOTOS SVS;", bf=0.00001090),
    BelleBBFEntry(decay="K- K+ anti-K0", model="PHOTOS PHSP;", bf=0.00001884),
    BelleBBFEntry(decay="K- K+ pi0", model="PHOTOS PHSP;", bf=0.00001886),
    BelleBBFEntry(decay="K- K0 pi+", model="PHOTOS PHSP;", bf=0.00001753),
    BelleBBFEntry(decay="K- pi+ pi0", model="PHOTOS PHSP;", bf=0.00000940),
    BelleBBFEntry(decay="anti-K0 pi- pi+", model="PHOTOS PHSP;", bf=0.00000048),
    BelleBBFEntry(decay="K_S0 K_S0 K_S0", model="PHOTOS PHSP;", bf=0.00000620),
    BelleBBFEntry(decay="K_L0 K_L0 K_L0", model="PHOTOS PHSP;", bf=0.00000620),
    BelleBBFEntry(decay="K_S0 K_S0 K_L0", model="PHOTOS PHSP;", bf=0.00000740),
    BelleBBFEntry(decay="K_S0 K_L0 K_L0", model="PHOTOS PHSP;", bf=0.00000740),
    BelleBBFEntry(decay="pi- pi+ pi0", model="PHOTOS PHSP;", bf=0.00000095),
    BelleBBFEntry(decay="f_0 pi+ pi-", model="PHOTOS PHSP;", bf=0.00000860),
    BelleBBFEntry(decay="f'_2 anti-K*0", model="PHOTOS PHSP;", bf=0.00003700),
    BelleBBFEntry(decay="f_2 anti-K*0", model="PHOTOS PHSP;", bf=0.00001800),
    BelleBBFEntry(decay="rho+ K_0*-", model="PHOTOS PHSP;", bf=0.00001200),
    BelleBBFEntry(decay="rho0 anti-K_0*0", model="PHOTOS PHSP;", bf=0.00001200),
    BelleBBFEntry(decay="a_1+ K-", model="PHOTOS SVS;", bf=0.00001630),
    BelleBBFEntry(decay="a_1+ pi-", model="PHOTOS SVS;", bf=0.00003170),
    BelleBBFEntry(decay="a_1- pi+", model="PHOTOS SVS;", bf=0.00003170),
    BelleBBFEntry(decay="K*- K*+", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00014100),
    BelleBBFEntry(decay="K*- rho+", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00001200),
    BelleBBFEntry(decay="anti-K*0 K*0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000049),
    BelleBBFEntry(decay="anti-K*0 anti-K*0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000018),
    BelleBBFEntry(decay="anti-K*0 rho0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000560),
    BelleBBFEntry(decay="rho- rho+", model="PHOTOS SVV_HELAMP 0.0 0.0 1.0 0.0 0.0 0.0;", bf=0.00002420),
    BelleBBFEntry(decay="rho0 rho0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000086),
    BelleBBFEntry(decay="rho- a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00006100),
    BelleBBFEntry(decay="rho+ a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00006100),
    BelleBBFEntry(decay="K*- pi+ pi0", model="PHOTOS PHSP;", bf=0.00000750),
    BelleBBFEntry(decay="anti-K*0 pi- pi+", model="PHOTOS PHSP;", bf=0.00005450),
    BelleBBFEntry(decay="anti-K*0 pi0 pi0", model="PHOTOS PHSP;", bf=0.00000260),
    BelleBBFEntry(decay="rho+ K- pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="rho+ anti-K0 pi-", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="rho0 K- pi+", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="rho0 anti-K0 pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="rho0 pi+ pi-", model="PHOTOS PHSP;", bf=0.00000630),
    BelleBBFEntry(decay="K+ K- anti-K*0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="anti-K0 K0 anti-K*0", model="PHOTOS PHSP;", bf=0.00004560),
    BelleBBFEntry(decay="anti-K0 K*0 anti-K0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="K+ K*- anti-K0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="K*+ K- anti-K0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="anti-K0  eta", model="PHOTOS PHSP;", bf=0.0000019),
    BelleBBFEntry(decay="anti-K0  eta'", model="PHOTOS PHSP;", bf=0.0000649),
    BelleBBFEntry(decay="anti-K0  eta(2S)", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="anti-K0  eta(1405)", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="anti-K_0*0 eta", model="PHOTOS PHSP;", bf=0.0000096),
    BelleBBFEntry(decay="pi0      eta", model="PHOTOS PHSP;", bf=0.0000013),
    BelleBBFEntry(decay="pi0      eta'", model="PHOTOS PHSP;", bf=0.0000015),
    BelleBBFEntry(decay="pi0      eta(2S)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="pi0      eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta      eta", model="PHOTOS PHSP;", bf=0.0000018),
    BelleBBFEntry(decay="eta      eta'", model="PHOTOS PHSP;", bf=0.0000017),
    BelleBBFEntry(decay="eta      eta(2S)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta      eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'     eta'", model="PHOTOS PHSP;", bf=0.0000024),
    BelleBBFEntry(decay="eta'     eta(2S)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'     eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta(2S)  eta(2S)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta(2S)  eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta(1405) eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="f_0 eta", model="PHOTOS PHSP;", bf=0.0000004),
    BelleBBFEntry(decay="f_0 eta'", model="PHOTOS PHSP;", bf=0.0000015),
    BelleBBFEntry(decay="omega    anti-K0", model="PHOTOS SVS;", bf=0.0000050),
    BelleBBFEntry(decay="phi      anti-K0", model="PHOTOS SVS;", bf=0.0000083),
    BelleBBFEntry(decay="phi      anti-K_0*0", model="PHOTOS PHSP;", bf=0.0000046),
    BelleBBFEntry(decay="phi	    anti-K''*0", model="PHOTOS PHSP;", bf=0.0000035),
    BelleBBFEntry(decay="phi      anti-K_3*0", model="PHOTOS PHSP;", bf=0.0000027),
    BelleBBFEntry(decay="phi	    anti-K_4*0", model="PHOTOS PHSP;", bf=0.0000153),
    BelleBBFEntry(decay="omega    pi0", model="PHOTOS SVS;", bf=0.0000012),
    BelleBBFEntry(decay="phi      pi0", model="PHOTOS SVS;", bf=0.00000028),
    BelleBBFEntry(decay="anti-K*0 eta", model="PHOTOS SVS;", bf=0.0000159),
    BelleBBFEntry(decay="rho0     eta", model="PHOTOS SVS;", bf=0.0000015),
    BelleBBFEntry(decay="omega    eta", model="PHOTOS SVS;", bf=0.0000019),
    BelleBBFEntry(decay="phi      eta", model="PHOTOS SVS;", bf=0.0000006),
    BelleBBFEntry(decay="anti-K*0 eta'", model="PHOTOS SVS;", bf=0.0000038),
    BelleBBFEntry(decay="rho0     eta'", model="PHOTOS SVS;", bf=0.0000013),
    BelleBBFEntry(decay="omega    eta'", model="PHOTOS SVS;", bf=0.0000022),
    BelleBBFEntry(decay="phi      eta'", model="PHOTOS SVS;", bf=0.0000005),
    BelleBBFEntry(decay="anti-K*0 eta(2S)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="rho0     eta(2S)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="omega    eta(2S)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="phi      eta(2S)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="anti-K*0 eta(1405)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="rho0     eta(1405)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="omega    eta(1405)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="phi      eta(1405)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="omega    f_0", model="PHOTOS SVS;", bf=0.0000015),
    BelleBBFEntry(decay="anti-K*0 omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000027),
    BelleBBFEntry(decay="anti-K*0 phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000095),
    BelleBBFEntry(decay="rho0     omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000015),
    BelleBBFEntry(decay="rho0     phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000130),
    BelleBBFEntry(decay="omega    omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000040),
    BelleBBFEntry(decay="omega    phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000012),
    BelleBBFEntry(decay="phi      phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000015),
    BelleBBFEntry(decay="anti-K_2*0 phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.0000078),
    BelleBBFEntry(decay="anti-K_2*0 eta", model="PHOTOS STS;", bf=0.0000096),
    BelleBBFEntry(decay="anti-Xsd eta'", model="PHOTOS BTOXSETAP;", bf=0.0003513),
    BelleBBFEntry(decay="anti-Xsd eta", model="PHOTOS BTOXSETAP;", bf=0.0004107),
    BelleBBFEntry(decay="eta  K- pi+", model="PHOTOS PHSP;", bf=0.0000115),
    BelleBBFEntry(decay="eta  pi- pi+", model="PHOTOS PHSP;", bf=0.0000062),
    BelleBBFEntry(decay="eta  anti-K0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  pi0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  K- pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  pi- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  anti-K0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  pi0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta anti-K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta' anti-K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta' pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta' anti-K0", model="PHOTOS PHSP;", bf=0.0000310),
    BelleBBFEntry(decay="eta'  eta' pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  K*- pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  K- rho+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  rho- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  pi- rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  anti-K*0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  anti-K0 rho0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  rho0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  K*- pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K- rho+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  rho- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  pi- rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  anti-K*0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  anti-K0 rho0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  rho0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta anti-K*0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta rho0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta' anti-K*0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta' rho0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta' anti-K*0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  eta' rho0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  K- pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  pi- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  anti-K0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  pi0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  K- pi+", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  pi- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  anti-K0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  pi0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  eta anti-K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  eta pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  eta' anti-K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  eta' pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  eta anti-K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  eta pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  eta' anti-K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  eta' pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  phi  anti-K0", model="PHOTOS PHSP;", bf=0.0000028),
    BelleBBFEntry(decay="a_00 pi0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="a_00 eta", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="a_00 eta'", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="a_00 eta(2S)", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="a_00 eta(1405)", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="a_00 a_00", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="a_0+ a_0-", model="PHOTOS PHSP;", bf=0.00000300),
    BelleBBFEntry(decay="a_00 f_0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="anti-K*0 a_00", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="K*- a_0+", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="rho0 a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="rho- a_0+", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="rho+ a_0-", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="omega a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="phi a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_1+ a_0-", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="a_1- a_0+", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="f_1 a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_1+ a_0-", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="b_1- a_0+", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="h_1  a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 anti-K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="a_10 pi0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 eta", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 eta'", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 eta(2S)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="a_10 eta(1405)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="a_10 f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="anti-K*0 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="K*- a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho0 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="omega a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="phi a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="a_1+ a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="f_1 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_1+ a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="b_1- a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="h_1  a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="f_0 pi0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="f_0 eta(2S)", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="f_0 eta(1405)", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="phi f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="h_1  f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 anti-K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="f_1 pi0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 eta", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 eta'", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 eta(2S)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="f_1 eta(1405)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K*0 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho0 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="omega f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="phi f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="h_1  f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 anti-K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="b_10 pi0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 eta", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 eta'", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 eta(2S)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="b_10 eta(1405)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K*0 b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000700),
    BelleBBFEntry(decay="K*- b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho0 b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="rho- b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho+ b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="omega b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="phi b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_10  b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_1+ b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="h_1  b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 anti-K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="h_1 pi0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 eta", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 eta'", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 eta(2S)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="h_1 eta(1405)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K*0 h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00001200),
    BelleBBFEntry(decay="rho0 h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="omega h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="phi h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="h_1  h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="a_2+ pi-", model="PHOTOS STS;", bf=0.00000490),
    BelleBBFEntry(decay="a_2- pi+", model="PHOTOS STS;", bf=0.00000490),
    BelleBBFEntry(decay="a_20 pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 eta", model="PHOTOS STS;", bf=0.00000014),
    BelleBBFEntry(decay="f_2  eta", model="PHOTOS STS;", bf=0.00000015),
    BelleBBFEntry(decay="f'_2 eta", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 eta'", model="PHOTOS STS;", bf=0.00000062),
    BelleBBFEntry(decay="f_2  eta'", model="PHOTOS STS;", bf=0.00000068),
    BelleBBFEntry(decay="f'_2 eta'", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*0 anti-K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K_2*0 K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ K-", model="PHOTOS STS;", bf=0.00000058),
    BelleBBFEntry(decay="a_20 anti-K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 anti-K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K_2*0 eta'", model="PHOTOS STS;", bf=0.00000130),
    BelleBBFEntry(decay="anti-K_2*0 eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K_2*0 eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ rho-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00001500),
    BelleBBFEntry(decay="a_2- rho+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00001500),
    BelleBBFEntry(decay="a_20 rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f_2 omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*0 anti-K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K_2*0 K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ K*-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000350),
    BelleBBFEntry(decay="a_20 anti-K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000210),
    BelleBBFEntry(decay="anti-K_2*0 rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000023),
    BelleBBFEntry(decay="K_2*- rho+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000023),
    BelleBBFEntry(decay="anti-K_2*0 omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma+ anti-p- pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Lambda0 anti-n0 pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Sigma0 anti-n0 pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Sigma- anti-Delta+ pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Lambda0 anti-p- pi+", model="PHOTOS PHSP;", bf=0.00000325),
    BelleBBFEntry(decay="Sigma0 anti-p- pi+", model="PHOTOS PHSP;", bf=0.00000380),
    BelleBBFEntry(decay="Sigma- anti-n0 pi+", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="Sigma+ anti-n0 pi-", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="Lambda0 anti-Delta+ pi-", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="Sigma0 anti-Delta+ pi-", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="p+ anti-p- anti-K0", model="PHOTOS PHSP;", bf=0.00000273),
    BelleBBFEntry(decay="n0 anti-n0 anti-K0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Delta- anti-Delta+ anti-K0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Delta++ anti-p- K-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="p+ anti-n0 K-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="n0 anti-Delta+ K-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="Xi0 anti-Lambda0 pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-Sigma0 pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Sigma+ pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Lambda0 pi+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Sigma0 pi+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-Sigma+ pi-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-Sigma- pi+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-n0 K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-p- K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Delta+ K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-n0 K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0 anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0 anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0 anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0 anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma+ anti-Lambda0 K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma+ anti-Sigma0 K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma- anti-Sigma+ anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Lambda0 anti-Sigma+ K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Sigma0 anti-Sigma+ K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Xi+ pi0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Omega- anti-Xi0 pi+", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Omega- anti-Sigma+ K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Omega- anti-Lambda0 K+", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Omega- anti-Sigma0 K+", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Xi- anti-Xi+ anti-K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Xi0 anti-Xi+ K-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="Sigma+ anti-p- rho0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Lambda0 anti-n0 rho0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Sigma0 anti-n0 rho0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Sigma- anti-Delta+ rho0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Lambda0 anti-p- rho+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Sigma0 anti-p- rho+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Sigma- anti-n0 rho+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Sigma+ anti-n0 rho-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Lambda0 anti-Delta+ rho-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Sigma0 anti-Delta+ rho-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="p+ anti-p- anti-K*0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0 anti-n0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Delta- anti-Delta+ anti-K*0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Delta++ anti-p- K*-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="p+ anti-n0 K*-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="n0 anti-Delta+ K*-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="Xi0 anti-Lambda0 rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-Sigma0 rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-Sigma+ rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-Lambda0 rho+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-Sigma0 rho+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-Sigma+ rho-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-Sigma- rho+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-n0 K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi0 anti-p- K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-Delta+ K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Xi- anti-n0 K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma+ anti-Lambda0 K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma+ anti-Sigma0 K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma- anti-Sigma+ anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Lambda0 anti-Sigma+ K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Sigma0 anti-Sigma+ K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="Omega- anti-Xi+ rho0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Sigma+ K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Lambda0 K*+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Omega- anti-Sigma0 K*+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi- anti-Xi+ anti-K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="Xi0 anti-Xi+ K*-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="p+	   anti-p- pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0	   anti-n0 pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-Delta+ pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="p+	   anti-n0 pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="p+	   anti-Delta-- pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta++ anti-p- pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0	   anti-p- pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0	   anti-Delta+ pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-n0 pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0 pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0 pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0 pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0 pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Sigma+ pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Sigma- pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma+ anti-Lambda0 pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Lambda0 pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma+ pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma- pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma+ anti-Sigma0 pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Sigma0 pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-p- K+", model="PHOTOS PHSP;", bf=0.00000082),
    BelleBBFEntry(decay="Sigma0 anti-p- K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-n0 K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-n0 K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="p+ anti-Lambda0 K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="p+ anti-Sigma0 K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Lambda0 anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Sigma0 anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Xi- anti-Xi+ pi0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi0 anti-Xi+ pi-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Xi0 pi+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma- anti-Xi+ anti-K0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Lambda0 anti-Xi+ K-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma0 anti-Xi+ K-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Sigma+ K0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Lambda0 K+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Sigma0 K+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="p+	   anti-p- rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0	   anti-n0 rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-Delta+ rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="p+	   anti-n0 rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0	   anti-p- rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Delta- anti-n0 rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="n0 anti-Delta+ rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0 rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0 rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0 rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0 rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Sigma+ rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-Sigma- rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma+ anti-Lambda0 rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Lambda0 rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma+ rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-Sigma- rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma+ anti-Sigma0 rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma- anti-Sigma0 rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-p- K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-p- K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Lambda0 anti-n0 K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Sigma0 anti-n0 K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="p+ anti-Lambda0 K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="p+ anti-Sigma0 K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Lambda0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="n0 anti-Sigma0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="Xi- anti-Xi+ rho0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi0 anti-Xi+ rho-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Xi0 rho+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma- anti-Xi+ anti-K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Lambda0 anti-Xi+ K*-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma0 anti-Xi+ K*-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Sigma+ K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Lambda0 K*+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Xi- anti-Sigma0 K*+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Sigma+ anti-p-", model="PHOTOS PHSP;", bf=0.00000026),
    BelleBBFEntry(decay="Lambda0 anti-n0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Sigma0 anti-n0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Sigma- anti-Delta+", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Xi0 anti-Lambda0", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="Xi0 anti-Sigma0", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="Xi- anti-Sigma+", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="Omega- anti-Xi+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="p+	   anti-p-", model="PHOTOS PHSP;", bf=0.00000011),
    BelleBBFEntry(decay="n0	   anti-n0", model="PHOTOS PHSP;", bf=0.00000027),
    BelleBBFEntry(decay="Delta- anti-Delta+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="Lambda0 anti-Lambda0", model="PHOTOS PHSP;", bf=0.00000032),
    BelleBBFEntry(decay="Lambda0 anti-Sigma0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Sigma0 anti-Lambda0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Sigma0 anti-Sigma0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="Xi- anti-Xi+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="D_s-  pi+", model="PHOTOS PHSP;", bf=0.00002200),
    BelleBBFEntry(decay="rho+  D_s-", model="PHOTOS SVS;", bf=0.00005700),
    BelleBBFEntry(decay="D_s-  pi+  pi0", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="a_1+  D_s-", model="PHOTOS SVS;", bf=0.00004300),
    BelleBBFEntry(decay="D_s*- pi+", model="PHOTOS SVS;", bf=0.00001150),
    BelleBBFEntry(decay="D_s*- rho+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00004900),
    BelleBBFEntry(decay="D_s*- pi+ pi0", model="PHOTOS PHSP;", bf=0.00005800),
    BelleBBFEntry(decay="D_s*- a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00009400),
    BelleBBFEntry(decay="D_s+ K-", model="PHOTOS PHSP;", bf=0.00003000),
    BelleBBFEntry(decay="K*-  D_s+", model="PHOTOS SVS;", bf=0.00008000),
    BelleBBFEntry(decay="D_s+ K- pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="D_s+ anti-K0 pi-", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="D_s*+ K-", model="PHOTOS SVS;", bf=0.00001980),
    BelleBBFEntry(decay="D_s*+ K*-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00006800),
    BelleBBFEntry(decay="D_s*+ K- pi0", model="PHOTOS PHSP;", bf=0.00002600),
    BelleBBFEntry(decay="D_s*+ anti-K0 pi-", model="PHOTOS PHSP;", bf=0.00005400),
    BelleBBFEntry(decay="J/psi D0", model="SVS;", bf=0.00000500),
    BelleBBFEntry(decay="J/psi D*0", model="SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000500),
    BelleBBFEntry(decay="J/psi D+ pi-", model="PHOTOS PHSP;", bf=0.00000500),
    BelleBBFEntry(decay="J/psi D0 pi0", model="PHSP;", bf=0.00000500),
    BelleBBFEntry(decay="psi(2S) D0", model="SVS;", bf=0.00000200),
    BelleBBFEntry(decay="psi(2S) D*0", model="SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="psi(2S) D+ pi-", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="psi(2S) D0 pi0", model="PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="D*0 gamma", model="SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="J/psi gamma", model="SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="psi(2S) gamma", model="SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000004),
    BelleBBFEntry(decay="tau+ tau-", model="PHOTOS  SLL;", bf=0.00000003),
    BelleBBFEntry(decay="gamma gamma", model="PHSP;", bf=0.00000003),
    BelleBBFEntry(decay="D0  anti-D0", model="PHSP;", bf=0.00002000),
    BelleBBFEntry(decay="D*0  anti-D0", model="SVS;", bf=0.00002400),
    BelleBBFEntry(decay="anti-D*0 D0", model="SVS;", bf=0.00002400),
    BelleBBFEntry(decay="D*0  anti-D*0", model="SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00003000),
    BelleBBFEntry(decay="D_s+  D_s-", model="PHOTOS PHSP;", bf=0.00002000),
    BelleBBFEntry(decay="D_s*+ D_s-", model="PHOTOS SVS;", bf=0.00002400),
    BelleBBFEntry(decay="D_s*- D_s+", model="PHOTOS SVS;", bf=0.00002400),
    BelleBBFEntry(decay="D_s*+ D_s*-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00003000),
)  # type: Tuple[BelleBBFEntry, ...]


bz_rare_bfs = (
    BelleBBFEntry(decay="K*0    gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.0000401),
    BelleBBFEntry(decay="K_10   gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.000043),
    BelleBBFEntry(decay="K_2*0  gamma", model="PHOTOS HELAMP 1.0 0.0 1.0 0.0;", bf=0.0000124),
    BelleBBFEntry(decay="Xsd    gamma", model="BTOXSGAMMA 2;", bf=0.000258),
    BelleBBFEntry(decay="rho0   gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000093),
    BelleBBFEntry(decay="omega  gamma", model="PHOTOS SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000046),
    BelleBBFEntry(decay="Xdd    gamma", model="BTOXDGAMMA 2;", bf=0.0000146),
    BelleBBFEntry(decay="K0   e+    e-", model="PHOTOS BTOSLLBALL;", bf=0.00000013),
    BelleBBFEntry(decay="K*0  e+    e-", model="PHOTOS BTOSLLBALL;", bf=0.00000104),
    BelleBBFEntry(decay="Xsd  e+    e-", model="BTOXSLL;", bf=0.00000353),
    BelleBBFEntry(decay="K0   mu+   mu-", model="PHOTOS BTOSLLBALL;", bf=0.00000057),
    BelleBBFEntry(decay="K*0  mu+   mu-", model="PHOTOS BTOSLLBALL;", bf=0.00000110),
    BelleBBFEntry(decay="Xsd  mu+   mu-", model="BTOXSLL;", bf=0.00000263),
    BelleBBFEntry(decay="K0   tau+  tau-", model="PHOTOS BTOSLLBALL;", bf=0.00000013),
    BelleBBFEntry(decay="K*0  tau+  tau-", model="PHOTOS BTOSLLBALL;", bf=0.00000019),
    BelleBBFEntry(decay="Xsd  tau+  tau-", model="BTOXSLL;", bf=0.00000022),
    BelleBBFEntry(decay="K0   nu_e   anti-nu_e", model="PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*0  nu_e   anti-nu_e", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="Xsd  nu_e   anti-nu_e", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K0   nu_mu  anti-nu_mu", model="PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*0  nu_mu  anti-nu_mu", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="Xsd  nu_mu  anti-nu_mu", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K0   nu_tau anti-nu_tau", model="PHSP;", bf=0.00000173),
    BelleBBFEntry(decay="K*0  nu_tau anti-nu_tau", model="PHOTOS PHSP;", bf=0.00000433),
    BelleBBFEntry(decay="Xsd  nu_tau anti-nu_tau", model="PHSP;", bf=0.00000574),
    BelleBBFEntry(decay="K+ K-", model="PHOTOS PHSP;", bf=0.00000015),
    BelleBBFEntry(decay="K+ pi-", model="PHOTOS PHSP;", bf=0.00001940),
    BelleBBFEntry(decay="K0 pi0", model="PHOTOS PHSP;", bf=0.00000990),
    BelleBBFEntry(decay="K_L0 K_L0", model="PHOTOS PHSP;", bf=0.00000048),
    BelleBBFEntry(decay="K_S0 K_S0", model="PHOTOS PHSP;", bf=0.00000048),
    BelleBBFEntry(decay="pi+ pi-", model="PHOTOS PHSP;", bf=0.00000516),
    BelleBBFEntry(decay="pi0 pi0", model="PHOTOS PHSP;", bf=0.00000131),
    BelleBBFEntry(decay="K'*+ pi-", model="PHOTOS PHSP;", bf=0.00008600),
    BelleBBFEntry(decay="K'*0 pi0", model="PHOTOS PHSP;", bf=0.00001930),
    BelleBBFEntry(decay="K_0*+ pi-", model="PHOTOS PHSP;", bf=0.00004970),
    BelleBBFEntry(decay="a_0+ pi-", model="PHOTOS PHSP;", bf=0.00000280),
    BelleBBFEntry(decay="f'_0 K0", model="PHOTOS PHSP;", bf=0.00000820),
    BelleBBFEntry(decay="f_0 K0", model="PHOTOS PHSP;", bf=0.00000580),
    BelleBBFEntry(decay="f_0(1500) K0", model="PHOTOS PHSP;", bf=0.00000340),
    BelleBBFEntry(decay="f_0 f_0", model="PHOTOS PHSP;", bf=0.00000016),
    BelleBBFEntry(decay="a_00 K0", model="PHOTOS PHSP;", bf=0.00000780),
    BelleBBFEntry(decay="a_0- K+", model="PHOTOS PHSP;", bf=0.00000160),
    BelleBBFEntry(decay="f_2 K0", model="PHOTOS STS;", bf=0.00000250),
    BelleBBFEntry(decay="K_2*+ pi-", model="PHOTOS STS;", bf=0.00000630),
    BelleBBFEntry(decay="K_2*0 pi0", model="PHOTOS STS;", bf=0.00000315),
    BelleBBFEntry(decay="K*+ K-", model="PHOTOS SVS;", bf=0.00000001),
    BelleBBFEntry(decay="K*+ pi-", model="PHOTOS SVS;", bf=0.00000980),
    BelleBBFEntry(decay="K*- K+", model="PHOTOS SVS;", bf=0.00000001),
    BelleBBFEntry(decay="K*0 anti-K0", model="PHOTOS SVS;", bf=0.00000190),
    BelleBBFEntry(decay="anti-K*0 K0", model="PHOTOS SVS;", bf=0.00000190),
    BelleBBFEntry(decay="K*0 pi0", model="PHOTOS SVS;", bf=0.00000000),
    BelleBBFEntry(decay="K*0 f_0", model="PHOTOS SVS;", bf=0.00000430),
    BelleBBFEntry(decay="rho(2S)0 K0", model="PHOTOS SVS;", bf=0.00000900),
    BelleBBFEntry(decay="rho(3S)0 K0", model="PHOTOS SVS;", bf=0.00000900),
    BelleBBFEntry(decay="rho+ pi-", model="PHOTOS SVS;", bf=0.00002310),
    BelleBBFEntry(decay="rho- K+", model="PHOTOS SVS;", bf=0.00001530),
    BelleBBFEntry(decay="rho- pi+", model="PHOTOS SVS;", bf=0.00002310),
    BelleBBFEntry(decay="rho0 K0", model="PHOTOS SVS;", bf=0.00000540),
    BelleBBFEntry(decay="rho0 pi0", model="PHOTOS SVS;", bf=0.00000180),
    BelleBBFEntry(decay="rho0 f_0", model="PHOTOS SVS;", bf=0.00000053),
    BelleBBFEntry(decay="rho(2S)+ pi-", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="rho(3S)+ pi-", model="PHOTOS SVS;", bf=0.00000100),
    BelleBBFEntry(decay="rho(2S)- pi+", model="PHOTOS SVS;", bf=0.00000200),
    BelleBBFEntry(decay="rho(3S)- pi+", model="PHOTOS SVS;", bf=0.00000100),
    BelleBBFEntry(decay="rho(2S)- K+", model="PHOTOS SVS;", bf=0.00000320),
    BelleBBFEntry(decay="rho(3S)- K+", model="PHOTOS SVS;", bf=0.00000170),
    BelleBBFEntry(decay="K_1+ pi-", model="PHOTOS SVS;", bf=0.00002520),
    BelleBBFEntry(decay="K'_1+ pi-", model="PHOTOS SVS;", bf=0.00002180),
    BelleBBFEntry(decay="K''*+ pi-", model="PHOTOS SVS;", bf=0.00001010),
    BelleBBFEntry(decay="K''*0 pi0", model="PHOTOS SVS;", bf=0.00001010),
    BelleBBFEntry(decay="b_1- K+", model="PHOTOS SVS;", bf=0.00000740),
    BelleBBFEntry(decay="b_1- pi+", model="PHOTOS SVS;", bf=0.00001090),
    BelleBBFEntry(decay="b_1+ pi-", model="PHOTOS SVS;", bf=0.00001090),
    BelleBBFEntry(decay="K+ K- K0", model="PHOTOS PHSP;", bf=0.00001884),
    BelleBBFEntry(decay="K+ K- pi0", model="PHOTOS PHSP;", bf=0.00001886),
    BelleBBFEntry(decay="K+ anti-K0 pi-", model="PHOTOS PHSP;", bf=0.00001753),
    BelleBBFEntry(decay="K+ pi- pi0", model="PHOTOS PHSP;", bf=0.00000940),
    BelleBBFEntry(decay="K0 pi+ pi-", model="PHOTOS PHSP;", bf=0.00000048),
    BelleBBFEntry(decay="K_S0 K_S0 K_S0", model="PHOTOS PHSP;", bf=0.00000620),
    BelleBBFEntry(decay="K_L0 K_L0 K_L0", model="PHOTOS PHSP;", bf=0.00000620),
    BelleBBFEntry(decay="K_S0 K_S0 K_L0", model="PHOTOS PHSP;", bf=0.00000740),
    BelleBBFEntry(decay="K_S0 K_L0 K_L0", model="PHOTOS PHSP;", bf=0.00000740),
    BelleBBFEntry(decay="pi+ pi- pi0", model="PHOTOS PHSP;", bf=0.00000095),
    BelleBBFEntry(decay="f_0 pi- pi+", model="PHOTOS PHSP;", bf=0.00000860),
    BelleBBFEntry(decay="f'_2 K*0", model="PHOTOS PHSP;", bf=0.00003700),
    BelleBBFEntry(decay="f_2 K*0", model="PHOTOS PHSP;", bf=0.00001800),
    BelleBBFEntry(decay="rho- K_0*+", model="PHOTOS PHSP;", bf=0.00001200),
    BelleBBFEntry(decay="rho0 K_0*0", model="PHOTOS PHSP;", bf=0.00001200),
    BelleBBFEntry(decay="a_1- K+", model="PHOTOS SVS;", bf=0.00001630),
    BelleBBFEntry(decay="a_1+ pi-", model="PHOTOS SVS;", bf=0.00003170),
    BelleBBFEntry(decay="a_1- pi+", model="PHOTOS SVS;", bf=0.00003170),
    BelleBBFEntry(decay="K*+ K*-", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00014100),
    BelleBBFEntry(decay="K*+ rho-", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00001200),
    BelleBBFEntry(decay="K*0 anti-K*0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000049),
    BelleBBFEntry(decay="K*0 K*0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000018),
    BelleBBFEntry(decay="K*0 rho0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000560),
    BelleBBFEntry(decay="rho+ rho-", model="PHOTOS SVV_HELAMP 0.0 0.0 1.0 0.0 0.0 0.0;", bf=0.00002420),
    BelleBBFEntry(decay="rho0 rho0", model="PHOTOS SVV_HELAMP 1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000086),
    BelleBBFEntry(decay="rho+ a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00006100),
    BelleBBFEntry(decay="rho- a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00006100),
    BelleBBFEntry(decay="K*+ pi- pi0", model="PHOTOS PHSP;", bf=0.00000750),
    BelleBBFEntry(decay="K*0 pi+ pi-", model="PHOTOS PHSP;", bf=0.00005450),
    BelleBBFEntry(decay="K*0 pi0 pi0", model="PHOTOS PHSP;", bf=0.00000260),
    BelleBBFEntry(decay="rho- K+ pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="rho- K0 pi+", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="rho0 K+ pi-", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="rho0 K0 pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="rho0 pi+ pi-", model="PHOTOS PHSP;", bf=0.00000630),
    BelleBBFEntry(decay="K+ K- K*0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="K0 anti-K0 K*0", model="PHOTOS PHSP;", bf=0.00004560),
    BelleBBFEntry(decay="K0 anti-K*0 K0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="K- K*+ K0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="K*- K+ K0", model="PHOTOS PHSP;", bf=0.00002280),
    BelleBBFEntry(decay="K0	 eta", model="PHOTOS PHSP;", bf=0.0000019),
    BelleBBFEntry(decay="K0	 eta'", model="PHOTOS PHSP;", bf=0.0000649),
    BelleBBFEntry(decay="K0	 eta(2S)", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="K0	 eta(1405)", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="K_0*0 eta", model="PHOTOS PHSP;", bf=0.0000096),
    BelleBBFEntry(decay="pi0   eta", model="PHOTOS PHSP;", bf=0.0000013),
    BelleBBFEntry(decay="pi0   eta'", model="PHOTOS PHSP;", bf=0.0000015),
    BelleBBFEntry(decay="pi0   eta(2S)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="pi0   eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta   eta", model="PHOTOS PHSP;", bf=0.0000018),
    BelleBBFEntry(decay="eta   eta'", model="PHOTOS PHSP;", bf=0.0000017),
    BelleBBFEntry(decay="eta   eta(2S)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta   eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta'", model="PHOTOS PHSP;", bf=0.0000024),
    BelleBBFEntry(decay="eta'  eta(2S)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta(2S)   eta(2S)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta(2S)   eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta(1405)  eta(1405)", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="f_0 eta", model="PHOTOS PHSP;", bf=0.0000004),
    BelleBBFEntry(decay="f_0 eta'", model="PHOTOS PHSP;", bf=0.0000015),
    BelleBBFEntry(decay="omega K0", model="PHOTOS SVS;", bf=0.0000050),
    BelleBBFEntry(decay="phi   K0", model="PHOTOS SVS;", bf=0.0000083),
    BelleBBFEntry(decay="phi	 K_0*0", model="PHOTOS PHSP;", bf=0.0000046),
    BelleBBFEntry(decay="phi	 K''*0", model="PHOTOS PHSP;", bf=0.0000035),
    BelleBBFEntry(decay="phi	 K_3*0", model="PHOTOS PHSP;", bf=0.0000027),
    BelleBBFEntry(decay="phi	 K_4*0", model="PHOTOS PHSP;", bf=0.0000153),
    BelleBBFEntry(decay="omega pi0", model="PHOTOS SVS;", bf=0.0000012),
    BelleBBFEntry(decay="phi   pi0", model="PHOTOS SVS;", bf=0.00000028),
    BelleBBFEntry(decay="K*0   eta", model="PHOTOS SVS;", bf=0.0000159),
    BelleBBFEntry(decay="rho0  eta", model="PHOTOS SVS;", bf=0.0000015),
    BelleBBFEntry(decay="omega eta", model="PHOTOS SVS;", bf=0.0000019),
    BelleBBFEntry(decay="phi   eta", model="PHOTOS SVS;", bf=0.0000006),
    BelleBBFEntry(decay="K*0   eta'", model="PHOTOS SVS;", bf=0.0000038),
    BelleBBFEntry(decay="rho0  eta'", model="PHOTOS SVS;", bf=0.0000013),
    BelleBBFEntry(decay="omega eta'", model="PHOTOS SVS;", bf=0.0000022),
    BelleBBFEntry(decay="phi   eta'", model="PHOTOS SVS;", bf=0.0000005),
    BelleBBFEntry(decay="K*0   eta(2S)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="rho0  eta(2S)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="omega eta(2S)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="phi   eta(2S)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="K*0   eta(1405)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="rho0  eta(1405)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="omega eta(1405)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="phi   eta(1405)", model="PHOTOS SVS;", bf=0.0000010),
    BelleBBFEntry(decay="omega f_0", model="PHOTOS SVS;", bf=0.0000015),
    BelleBBFEntry(decay="K*0   omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000027),
    BelleBBFEntry(decay="K*0   phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000095),
    BelleBBFEntry(decay="rho0  omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000015),
    BelleBBFEntry(decay="rho0  phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000130),
    BelleBBFEntry(decay="omega omega", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000040),
    BelleBBFEntry(decay="omega phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000012),
    BelleBBFEntry(decay="phi   phi", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.0000015),
    BelleBBFEntry(decay="K_2*0 phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.0000078),
    BelleBBFEntry(decay="K_2*0 eta", model="PHOTOS STS;", bf=0.0000096),
    BelleBBFEntry(decay="Xsd eta'", model="PHOTOS BTOXSETAP;", bf=0.0003513),
    BelleBBFEntry(decay="Xsd eta", model="PHOTOS BTOXSETAP;", bf=0.0004107),
    BelleBBFEntry(decay="eta  K+ pi-", model="PHOTOS PHSP;", bf=0.0000115),
    BelleBBFEntry(decay="eta  pi+ pi-", model="PHOTOS PHSP;", bf=0.0000062),
    BelleBBFEntry(decay="eta  K0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  pi0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  K+ pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  pi- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  K0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  pi0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta' K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta' pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta' K0", model="PHOTOS PHSP;", bf=0.0000310),
    BelleBBFEntry(decay="eta'  eta' pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  K*+ pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  K+ rho-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  rho- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  pi- rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  K*0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  K0 rho0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  rho0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  K*+ pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K+ rho-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  rho- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  pi- rho+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  K*0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  K0 rho0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  rho0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta K*0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta rho0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta  eta' K*0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta  eta' rho0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="eta'  eta' K*0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="eta'  eta' rho0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  K+ pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  pi- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  K0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  pi0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  K+ pi-", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  pi- pi+", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  K0 pi0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  pi0 pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  eta K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  eta pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="omega  eta' K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="omega  eta' pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  eta K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  eta pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  eta' K0", model="PHOTOS PHSP;", bf=0.0000030),
    BelleBBFEntry(decay="phi  eta' pi0", model="PHOTOS PHSP;", bf=0.0000010),
    BelleBBFEntry(decay="phi  phi  K0", model="PHOTOS PHSP;", bf=0.0000028),
    BelleBBFEntry(decay="a_00 pi0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="a_00 eta", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="a_00 eta'", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="a_00 eta(2S)", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="a_00 eta(1405)", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="a_00 a_00", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="a_0+ a_0-", model="PHOTOS PHSP;", bf=0.00000300),
    BelleBBFEntry(decay="a_00 f_0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="K*0 a_00", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="K*+ a_0-", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="rho0 a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="rho- a_0+", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="rho+ a_0-", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="omega a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="phi a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_1+ a_0-", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="a_1- a_0+", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="f_1 a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_1+ a_0-", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="b_1- a_0+", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="h_1  a_00", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="a_10 pi0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 eta", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 eta'", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 eta(2S)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="a_10 eta(1405)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="a_10 f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="K*0 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="K*+ a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho0 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="omega a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="phi a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="a_10 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="a_1+ a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="f_1 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_1+ a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="b_1- a_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="h_1  a_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="f_0 pi0", model="PHOTOS PHSP;", bf=0.00000050),
    BelleBBFEntry(decay="f_0 eta(2S)", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="f_0 eta(1405)", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="phi f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 f_0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="f_1 pi0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 eta", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 eta'", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 eta(2S)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="f_1 eta(1405)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="K*0 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho0 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="omega f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="phi f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="f_1 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 f_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="b_10 pi0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 eta", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 eta'", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 eta(2S)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="b_10 eta(1405)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="K*0 b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000700),
    BelleBBFEntry(decay="K*+ b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho0 b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="rho- b_1+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="rho+ b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="omega b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="phi b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_10 b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="b_1+ b_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000300),
    BelleBBFEntry(decay="h_1 b_10", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 K0", model="PHOTOS SVS;", bf=0.00000300),
    BelleBBFEntry(decay="h_1 pi0", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 eta", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 eta'", model="PHOTOS SVS;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 eta(2S)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="h_1 eta(1405)", model="PHOTOS SVS;", bf=0.00000010),
    BelleBBFEntry(decay="K*0 h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00001200),
    BelleBBFEntry(decay="rho0 h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="omega h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="phi h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="h_1 h_1", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000050),
    BelleBBFEntry(decay="a_2+ pi-", model="PHOTOS STS;", bf=0.00000490),
    BelleBBFEntry(decay="a_2- pi+", model="PHOTOS STS;", bf=0.00000490),
    BelleBBFEntry(decay="a_20 pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 pi0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 eta", model="PHOTOS STS;", bf=0.00000014),
    BelleBBFEntry(decay="f_2  eta", model="PHOTOS STS;", bf=0.00000015),
    BelleBBFEntry(decay="f'_2 eta", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 eta'", model="PHOTOS STS;", bf=0.00000062),
    BelleBBFEntry(decay="f_2  eta'", model="PHOTOS STS;", bf=0.00000068),
    BelleBBFEntry(decay="f'_2 eta'", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*0 anti-K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K_2*0 K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2- K+", model="PHOTOS STS;", bf=0.00000058),
    BelleBBFEntry(decay="a_20 K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 K0", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*0 eta'", model="PHOTOS STS;", bf=0.00000130),
    BelleBBFEntry(decay="K_2*0 eta(2S)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*0 eta(1405)", model="PHOTOS STS;", bf=0.00000010),
    BelleBBFEntry(decay="a_2+ rho-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00001500),
    BelleBBFEntry(decay="a_2- rho+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00001500),
    BelleBBFEntry(decay="a_20 rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f_2 omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_20 phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f_2  phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="f'_2 phi", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="K_2*0 anti-K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="anti-K_2*0 K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="a_2- K*+", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000350),
    BelleBBFEntry(decay="a_20 K*0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000210),
    BelleBBFEntry(decay="K_2*0 rho0", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000023),
    BelleBBFEntry(decay="K_2*+ rho-", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000023),
    BelleBBFEntry(decay="K_2*0 omega", model="PHOTOS HELAMP 1.0 0.0  1.0 0.0  1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma- p+ pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Lambda0 n0 pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Sigma0 n0 pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Sigma+ Delta- pi0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Lambda0 p+ pi-", model="PHOTOS PHSP;", bf=0.00000325),
    BelleBBFEntry(decay="anti-Sigma0 p+ pi-", model="PHOTOS PHSP;", bf=0.00000380),
    BelleBBFEntry(decay="anti-Sigma+ n0 pi-", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="anti-Sigma- n0 pi+", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="anti-Lambda0 Delta- pi+", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="anti-Sigma0 Delta- pi+", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="anti-p- p+ K0", model="PHOTOS PHSP;", bf=0.00000273),
    BelleBBFEntry(decay="anti-n0 n0 K0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Delta+ Delta- K0", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Delta-- p+ K+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-p- n0 K+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-n0 Delta- K+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-Xi0 Lambda0 pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 Sigma0 pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Sigma- pi0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Lambda0 pi-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Sigma0 pi-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 Sigma- pi+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 Sigma+ pi-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 n0 anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 p+ K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Delta- anti-K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ n0 K-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0 K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0 K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0 K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0 K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma- Lambda0 K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma- Sigma0 K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma+ Sigma- K0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Lambda0 Sigma- K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Sigma0 Sigma- K+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Xi- pi0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Omega+ Xi0 pi-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Omega+ Sigma- anti-K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Omega+ Lambda0 K-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Omega+ Sigma0 K-", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Xi+ Xi- anti-K0", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Xi0 Xi- K+", model="PHOTOS PHSP;", bf=0.00000020),
    BelleBBFEntry(decay="anti-Sigma- p+ rho0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Lambda0 n0 rho0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Sigma0 n0 rho0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Sigma+ Delta- rho0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Lambda0 p+ rho-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Sigma0 p+ rho-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Sigma+ n0 rho-", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Sigma- n0 rho+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Lambda0 Delta- rho+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Sigma0 Delta- rho+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-p- p+ K*0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0 n0 K*0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Delta+ Delta- K*0", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Delta-- p+ K*+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-p- n0 K*+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-n0 Delta- K*+", model="PHOTOS PHSP;", bf=0.00000560),
    BelleBBFEntry(decay="anti-Xi0 Lambda0 rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 Sigma0 rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ Sigma- rho0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ Lambda0 rho-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ Sigma0 rho-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 Sigma- rho+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 Sigma+ rho-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 n0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi0 p+ K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ Delta- anti-K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Xi+ n0 K*-", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0 K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0 K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0 K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0 K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma- Lambda0 K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma- Sigma0 K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma+ Sigma- K*0", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Lambda0 Sigma- K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Sigma0 Sigma- K*+", model="PHOTOS PHSP;", bf=0.00000180),
    BelleBBFEntry(decay="anti-Omega+ Xi- rho0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Sigma- anti-K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Lambda0 K*-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Omega+ Sigma0 K*-", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi+ Xi- anti-K*0", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-Xi0 Xi- K*+", model="PHOTOS PHSP;", bf=0.00000060),
    BelleBBFEntry(decay="anti-p-   p+ pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0   n0 pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ Delta- pi0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-p-   n0 pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-p-   Delta++ pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta-- p+ pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0   p+ pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ n0 pi-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0 Delta- pi+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0 pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0 pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0 pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0 pi0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Sigma- pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Sigma+ pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma- Lambda0 pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Lambda0 pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma- pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma+ pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma- Sigma0 pi+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Sigma0 pi-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 p+ K-", model="PHOTOS PHSP;", bf=0.00000082),
    BelleBBFEntry(decay="anti-Sigma0 p+ K-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 n0 anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 n0 anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-p- Lambda0 K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-p- Sigma0 K+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Lambda0 anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Sigma0 anti-K0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Xi+ Xi- pi0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi0 Xi- pi+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Xi0 pi-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma+ Xi- K0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Lambda0 Xi- K+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma0 Xi- K+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Sigma- anti-K0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Lambda0 K-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Sigma0 K-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-p-	   p+ rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0	   n0 rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ Delta- rho0", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-p-   n0 rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0   p+ rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Delta+ n0 rho-", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-n0 Delta- rho+", model="PHOTOS PHSP;", bf=0.00000150),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0 rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0 rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0 rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0 rho0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Sigma- rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 Sigma+ rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma- Lambda0 rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Lambda0 rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma- rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 Sigma+ rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma- Sigma0 rho+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma+ Sigma0 rho-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 p+ K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 p+ K*-", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Lambda0 n0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Sigma0 n0 anti-K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-p- Lambda0 K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-p- Sigma0 K*+", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Lambda0 K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-n0 Sigma0 K*0", model="PHOTOS PHSP;", bf=0.00000040),
    BelleBBFEntry(decay="anti-Xi+ Xi- rho0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi0 Xi- rho+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Xi0 rho-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma+ Xi- K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Lambda0 Xi- K*+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma0 Xi- K*+", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Sigma- anti-K*0", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Lambda0 K*-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Xi+ Sigma0 K*-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Sigma- p+", model="PHOTOS PHSP;", bf=0.00000026),
    BelleBBFEntry(decay="anti-Lambda0 n0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Sigma0 n0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Sigma+ Delta-", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Xi0 Lambda0", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="anti-Xi0 Sigma0", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="anti-Xi+ Sigma-", model="PHOTOS PHSP;", bf=0.00000030),
    BelleBBFEntry(decay="anti-Omega+ Xi-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-p-	   p+", model="PHOTOS PHSP;", bf=0.00000011),
    BelleBBFEntry(decay="anti-n0	   n0", model="PHOTOS PHSP;", bf=0.00000027),
    BelleBBFEntry(decay="anti-Delta+ Delta-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="anti-Lambda0 Lambda0", model="PHOTOS PHSP;", bf=0.00000032),
    BelleBBFEntry(decay="anti-Lambda0 Sigma0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Sigma0 Lambda0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Sigma0 Sigma0", model="PHOTOS PHSP;", bf=0.00000100),
    BelleBBFEntry(decay="anti-Xi+ Xi-", model="PHOTOS PHSP;", bf=0.00000010),
    BelleBBFEntry(decay="D_s+  pi-", model="PHOTOS PHSP;", bf=0.00002200),
    BelleBBFEntry(decay="rho-  D_s+", model="PHOTOS SVS;", bf=0.00005700),
    BelleBBFEntry(decay="D_s+  pi-  pi0", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="a_1-  D_s+", model="PHOTOS SVS;", bf=0.00004300),
    BelleBBFEntry(decay="D_s*+ pi-", model="PHOTOS SVS;", bf=0.00001150),
    BelleBBFEntry(decay="D_s*+ rho-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00004900),
    BelleBBFEntry(decay="D_s*+ pi- pi0", model="PHOTOS PHSP;", bf=0.00005800),
    BelleBBFEntry(decay="D_s*+ a_1-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00009400),
    BelleBBFEntry(decay="D_s- K+", model="PHOTOS PHSP;", bf=0.00003000),
    BelleBBFEntry(decay="K*+  D_s-", model="PHOTOS SVS;", bf=0.00008000),
    BelleBBFEntry(decay="D_s- K+ pi0", model="PHOTOS PHSP;", bf=0.00000400),
    BelleBBFEntry(decay="D_s- K0 pi+", model="PHOTOS PHSP;", bf=0.00000800),
    BelleBBFEntry(decay="D_s*- K+", model="PHOTOS SVS;", bf=0.00001980),
    BelleBBFEntry(decay="D_s*- K*+", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00006800),
    BelleBBFEntry(decay="D_s*- K+ pi0", model="PHOTOS PHSP;", bf=0.00002600),
    BelleBBFEntry(decay="D_s*- K0 pi+", model="PHOTOS PHSP;", bf=0.00005400),
    BelleBBFEntry(decay="J/psi anti-D0", model="SVS;", bf=0.00000500),
    BelleBBFEntry(decay="J/psi anti-D*0", model="SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000500),
    BelleBBFEntry(decay="J/psi D- pi+", model="PHOTOS PHSP;", bf=0.00000500),
    BelleBBFEntry(decay="J/psi anti-D0 pi0", model="PHSP;", bf=0.00000500),
    BelleBBFEntry(decay="psi(2S) anti-D0", model="SVS;", bf=0.00000200),
    BelleBBFEntry(decay="psi(2S) anti-D*0", model="SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="psi(2S) D- pi+", model="PHOTOS PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="psi(2S) anti-D0 pi0", model="PHSP;", bf=0.00000200),
    BelleBBFEntry(decay="anti-D*0 gamma", model="SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000200),
    BelleBBFEntry(decay="J/psi gamma", model="SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000010),
    BelleBBFEntry(decay="psi(2S) gamma", model="SVP_HELAMP 1.0 0.0 1.0 0.0;", bf=0.00000004),
    BelleBBFEntry(decay="tau+ tau-", model="PHOTOS  SLL;", bf=0.00000003),
    BelleBBFEntry(decay="gamma gamma", model="PHSP;", bf=0.00000003),
    BelleBBFEntry(decay="D0  anti-D0", model="PHSP;", bf=0.00002000),
    BelleBBFEntry(decay="D*0  anti-D0", model="SVS;", bf=0.00002400),
    BelleBBFEntry(decay="anti-D*0 D0", model="SVS;", bf=0.00002400),
    BelleBBFEntry(decay="D*0  anti-D*0", model="SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00003000),
    BelleBBFEntry(decay="D_s+  D_s-", model="PHOTOS PHSP;", bf=0.00002000),
    BelleBBFEntry(decay="D_s*+ D_s-", model="PHOTOS SVS;", bf=0.00002400),
    BelleBBFEntry(decay="D_s*- D_s+", model="PHOTOS SVS;", bf=0.00002400),
    BelleBBFEntry(decay="D_s*+ D_s*-", model="PHOTOS SVV_HELAMP  1.0 0.0 1.0 0.0 1.0 0.0;", bf=0.00003000),
)  # type: Tuple[BelleBBFEntry, ...]


bm_ulnu_bfs = (
    BelleBBFEntry(decay="pi0    e-    anti-nu_e", model="PHOTOS   SLPOLE 0.261 -2.03 1.293 1 0.261 -0.27 -0.752 1;", bf=0.000073),
    BelleBBFEntry(decay="rho0   e-    anti-nu_e", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000149),
    BelleBBFEntry(decay="eta    e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000084),
    BelleBBFEntry(decay="eta'   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000033),
    BelleBBFEntry(decay="omega  e-    anti-nu_e", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000115),
    BelleBBFEntry(decay="a_20   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000024),
    BelleBBFEntry(decay="f_2    e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000027),
    BelleBBFEntry(decay="b_10   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000082),
    BelleBBFEntry(decay="h_1    e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000087),
    BelleBBFEntry(decay="a_10   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000065),
    BelleBBFEntry(decay="f_1    e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000062),
    BelleBBFEntry(decay="a_00   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000003),
    BelleBBFEntry(decay="f_0    e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000005),
    BelleBBFEntry(decay="Xu0    e-    anti-nu_e", model="PHOTOS   VUB 4.54 7.5 0.22 1 0.28 1;", bf=0.001026),
    BelleBBFEntry(decay="pi0    mu-   anti-nu_mu", model="PHOTOS   SLPOLE 0.261 -2.03 1.293 1 0.261 -0.27 -0.752 1;", bf=0.000073),
    BelleBBFEntry(decay="rho0   mu-   anti-nu_mu", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000149),
    BelleBBFEntry(decay="eta    mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000084),
    BelleBBFEntry(decay="eta'   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000033),
    BelleBBFEntry(decay="omega  mu-   anti-nu_mu", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000115),
    BelleBBFEntry(decay="a_20   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000024),
    BelleBBFEntry(decay="f_2    mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000027),
    BelleBBFEntry(decay="b_10   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000082),
    BelleBBFEntry(decay="h_1    mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000087),
    BelleBBFEntry(decay="a_10   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000065),
    BelleBBFEntry(decay="f_1    mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000062),
    BelleBBFEntry(decay="a_00   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000003),
    BelleBBFEntry(decay="f_0    mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000005),
    BelleBBFEntry(decay="Xu0    mu-   anti-nu_mu", model="PHOTOS   VUB 4.54 7.5 0.22 1 0.28 1;", bf=0.001026),
)  # type: Tuple[BelleBBFEntry, ...]


bp_ulnu_bfs = (
    BelleBBFEntry(decay="pi0    e+    nu_e", model="PHOTOS   SLPOLE 0.261 -2.03 1.293 1 0.261 -0.27 -0.752 1;", bf=0.000073),
    BelleBBFEntry(decay="rho0   e+    nu_e", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000149),
    BelleBBFEntry(decay="eta    e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000084),
    BelleBBFEntry(decay="eta'   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000033),
    BelleBBFEntry(decay="omega  e+    nu_e", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000115),
    BelleBBFEntry(decay="a_20   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000024),
    BelleBBFEntry(decay="f_2    e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000027),
    BelleBBFEntry(decay="b_10   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000082),
    BelleBBFEntry(decay="h_1    e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000087),
    BelleBBFEntry(decay="a_10   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000065),
    BelleBBFEntry(decay="f_1    e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000062),
    BelleBBFEntry(decay="a_00   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000003),
    BelleBBFEntry(decay="f_0    e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000005),
    BelleBBFEntry(decay="Xu0    e+    nu_e", model="PHOTOS   VUB 4.54 7.5 0.22 1 0.28 1;", bf=0.001026),
    BelleBBFEntry(decay="pi0    mu+   nu_mu", model="PHOTOS   SLPOLE 0.261 -2.03 1.293 1 0.261 -0.27 -0.752 1;", bf=0.000073),
    BelleBBFEntry(decay="rho0   mu+   nu_mu", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000149),
    BelleBBFEntry(decay="eta    mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000084),
    BelleBBFEntry(decay="eta'   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000033),
    BelleBBFEntry(decay="omega  mu+   nu_mu", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000115),
    BelleBBFEntry(decay="a_20   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000024),
    BelleBBFEntry(decay="f_2    mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000027),
    BelleBBFEntry(decay="b_10   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000082),
    BelleBBFEntry(decay="h_1    mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000087),
    BelleBBFEntry(decay="a_10   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000065),
    BelleBBFEntry(decay="f_1    mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000062),
    BelleBBFEntry(decay="a_00   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000003),
    BelleBBFEntry(decay="f_0    mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000005),
    BelleBBFEntry(decay="Xu0    mu+   nu_mu", model="PHOTOS   VUB 4.54 7.5 0.22 1 0.28 1;", bf=0.001026),
)  # type: Tuple[BelleBBFEntry, ...]


anti_bz_ulnu_bfs = (
    BelleBBFEntry(decay="pi+    e-    anti-nu_e", model="PHOTOS   SLPOLE 0.261 -2.03 1.293 1 0.261 -0.27 -0.752 1;", bf=0.000136),
    BelleBBFEntry(decay="rho+   e-    anti-nu_e", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000277),
    BelleBBFEntry(decay="a_2+   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000047),
    BelleBBFEntry(decay="b_1+   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000154),
    BelleBBFEntry(decay="a_1+   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000123),
    BelleBBFEntry(decay="a_0+   e-    anti-nu_e", model="PHOTOS   ISGW2;", bf=0.000007),
    BelleBBFEntry(decay="Xu+    e-    anti-nu_e", model="PHOTOS   VUB 4.58 6.0 0.22 1 0.28 1;", bf=0.000965),
    BelleBBFEntry(decay="pi+    mu-   anti-nu_mu", model="PHOTOS   SLPOLE 0.261 -2.03 1.293 1 0.261 -0.27 -0.752 1;", bf=0.000136),
    BelleBBFEntry(decay="rho+   mu-   anti-nu_mu", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000277),
    BelleBBFEntry(decay="a_2+   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000047),
    BelleBBFEntry(decay="b_1+   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000154),
    BelleBBFEntry(decay="a_1+   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000123),
    BelleBBFEntry(decay="a_0+   mu-   anti-nu_mu", model="PHOTOS   ISGW2;", bf=0.000007),
    BelleBBFEntry(decay="Xu+    mu-   anti-nu_mu", model="PHOTOS   VUB 4.58 6.0 0.22 1 0.28 1;", bf=0.000965),
)  # type: Tuple[BelleBBFEntry, ...]


bz_ulnu_bfs = (
    BelleBBFEntry(decay="pi-    e+    nu_e", model="PHOTOS   SLPOLE 0.261 -2.03 1.293 1 0.261 -0.27 -0.752 1;", bf=0.000136),
    BelleBBFEntry(decay="rho-   e+    nu_e", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000277),
    BelleBBFEntry(decay="a_2-   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000047),
    BelleBBFEntry(decay="b_1-   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000154),
    BelleBBFEntry(decay="a_1-   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000123),
    BelleBBFEntry(decay="a_0-   e+    nu_e", model="PHOTOS   ISGW2;", bf=0.000007),
    BelleBBFEntry(decay="Xu-    e+    nu_e", model="PHOTOS   VUB 4.58 6.0 0.22 1 0.28 1;", bf=0.000965),
    BelleBBFEntry(decay="pi-    mu+   nu_mu", model="PHOTOS   SLPOLE 0.261 -2.03 1.293 1 0.261 -0.27 -0.752 1;", bf=0.000136),
    BelleBBFEntry(decay="rho-   mu+   nu_mu", model="PHOTOS   SLPOLE 0.261 -0.29 -0.415 1 0.223 -0.93 -0.092 1 0.338 -1.37 0.315 1 0.372 -1.40 0.437 1;", bf=0.000277),
    BelleBBFEntry(decay="a_2-   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000047),
    BelleBBFEntry(decay="b_1-   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000154),
    BelleBBFEntry(decay="a_1-   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000123),
    BelleBBFEntry(decay="a_0-   mu+   nu_mu", model="PHOTOS   ISGW2;", bf=0.000007),
    BelleBBFEntry(decay="Xu-    mu+   nu_mu", model="PHOTOS   VUB 4.58 6.0 0.22 1 0.28 1;", bf=0.000965),
)  # type: Tuple[BelleBBFEntry, ...]
# fmt: on


sum_belle_decay_b_minus_rare = sum([bf_e.bf for bf_e in bm_rare_bfs])  # type: float
sum_belle_decay_b_plus_rare = sum([bf_e.bf for bf_e in bp_rare_bfs])  # type: float
assert sum_belle_decay_b_minus_rare == sum_belle_decay_b_plus_rare, (
    sum_belle_decay_b_minus_rare,
    sum_belle_decay_b_plus_rare,
)

sum_belle_decay_b_zero_rare = sum([bf_e.bf for bf_e in bz_rare_bfs])  # type: float
sum_belle_decay_anti_b_zero_rare = sum([bf_e.bf for bf_e in anti_bz_rare_bfs])  # type: float
assert sum_belle_decay_b_zero_rare == sum_belle_decay_anti_b_zero_rare, (
    sum_belle_decay_b_zero_rare,
    sum_belle_decay_anti_b_zero_rare,
)

sum_belle_decay_b_minus_ulnu = sum([bf_e.bf for bf_e in bm_ulnu_bfs])  # type: float
sum_belle_decay_b_plus_ulnu = sum([bf_e.bf for bf_e in bp_ulnu_bfs])  # type: float
assert sum_belle_decay_b_minus_ulnu == sum_belle_decay_b_plus_ulnu, (
    sum_belle_decay_b_minus_ulnu,
    sum_belle_decay_b_plus_ulnu,
)

sum_belle_decay_b_zero_ulnu = sum([bf_e.bf for bf_e in bz_ulnu_bfs])  # type: float
sum_belle_decay_anti_b_zero_ulnu = sum([bf_e.bf for bf_e in anti_bz_ulnu_bfs])  # type: float
assert sum_belle_decay_b_zero_ulnu == sum_belle_decay_anti_b_zero_ulnu, (
    sum_belle_decay_b_zero_ulnu,
    sum_belle_decay_anti_b_zero_ulnu,
)


sum_bf_bp_rare = sum_belle_decay_b_minus_rare  # type: float
sum_bf_b0_rare = sum_belle_decay_b_zero_rare  # type: float
sum_bf_bp_ulnu = sum_belle_decay_b_minus_ulnu  # type: float
sum_bf_b0_ulnu = sum_belle_decay_b_zero_ulnu  # type: float


if __name__ == "__main__":
    special_b_charged_bf_sum = sum_bf_bp_rare + sum_bf_bp_ulnu
    special_b_mixed_bf_sum = sum_bf_b0_rare + sum_bf_b0_ulnu

    print(
        "\n sum_belle_decay_b_minus_rare",
        sum_belle_decay_b_minus_rare,
        "\n sum_belle_decay_b_plus_rare",
        sum_belle_decay_b_plus_rare,
        "\n sum_belle_decay_b_zero_rare",
        sum_belle_decay_b_zero_rare,
        "\n sum_belle_decay_anti_b_zero_rare",
        sum_belle_decay_anti_b_zero_rare,
        "\n sum_belle_decay_b_minus_ulnu",
        sum_belle_decay_b_minus_ulnu,
        "\n sum_belle_decay_b_plus_ulnu",
        sum_belle_decay_b_plus_ulnu,
        "\n sum_belle_decay_b_zero_ulnu",
        sum_belle_decay_b_zero_ulnu,
        "\n sum_belle_decay_anti_b_zero_ulnu",
        sum_belle_decay_anti_b_zero_ulnu,
        "\n special_b_charged_bf_sum",
        special_b_charged_bf_sum,
        2 * special_b_charged_bf_sum - special_b_charged_bf_sum ** 2,
        2 * special_b_charged_bf_sum - 2 * special_b_charged_bf_sum ** 2,
        "\n special_b_mixed_bf_sum",
        special_b_mixed_bf_sum,
        2 * special_b_mixed_bf_sum - special_b_mixed_bf_sum ** 2,
        2 * special_b_mixed_bf_sum - 2 * special_b_mixed_bf_sum ** 2,
    )
